//-----------------------------------------------------------------------------
//
// MBF_PLAYER.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//-----------------------------------------------------------------------------

#ifndef __MBF_PLAYER__
#define __MBF_PLAYER__


// The player data structure depends on a number
// of other structs: items (internal inventory),
// animation states (closely tied to the sprites
// used to represent them).
#include "doom/p_pspr.h"

// The player input is buffered within the player
// data struct, as commands per game tick.
#include "p_data.h"

// Weapon info: sprite frames, ammunition use.
typedef struct
{
    ammotype_t ammo;
	int		pershot;	// Ammo used per shot.
    int		upstate;
    int		downstate;
    int		readystate;
    int		atkstate;
    int		flashstate;

} weaponinfo_t;

extern  weaponinfo_t    weaponinfo[NUMWEAPONS];

void P_InitWeaponAndDEHInfo(void);


//
// Player states.
//
typedef enum
{
    // Playing or camping.
    PST_LIVE,
    // Dead on the ground, view follows killer.
    PST_DEAD,
    // Ready to restart/respawn???
    PST_REBORN		

}  playerstate_t;


//
// Player internal flags, for cheats and debug.
//
typedef enum
{
    CF_NOCLIP		= 1,	// No clipping, walk through barriers
    CF_GODMODE		= 2,	// No damage, no health loss
    CF_NOMOMENTUM	= 4		// Not really a cheat, just a debug aid

} cheat_t;

// v220-5 values moved out of player_t as these are common to
//        all players with doomednum = 1 (player + voodoo dolls)
typedef struct
{
	// Armor type is 0-2.
	int health, armorpoints, armortype;
} plhealth_t;

extern plhealth_t plhealth;


typedef struct player_s
{
	ddplayer_t		*plr;	// Pointer to the engine's player data.
	playerstate_t	playerstate;
	ticcmd_t		cmd;
	
	// bounded/scaled total momentum.
	fixed_t     bob;	
	
	// Power ups. invinc and invis are tic counters.
	int			powers[NUMPOWERS];
	int			cards[NUMCARDS];
	RDBOOL		backpack;
	
	int			frags[MAXPLAYERS];
	
	weapontype_t readyweapon;
	weapontype_t pendingweapon;	// set to wp_nochange if not changing.
	
	int			weaponowned[NUMWEAPONS];
	int			ammo[NUMAMMO];
	int			maxammo[NUMAMMO];
	
	// True if button down last tic.
	int			attackdown;
	int			usedown;
	
	// Bit flags, for cheats and debug.
	// See cheat_t, above.
	int			cheats;		
	
	// Refired shots are less accurate.
	int			refire;		
	
	// For intermission stats.
	int			killcount;
	int			itemcount;
	int			secretcount;
	
	// Hint messages.
	char*		hint_msg;	
	
	// For screen flashing (red or bright).
	int			damagecount;
	int			bonuscount;
	
	// Who did damage (NULL for floors/ceilings).
	mobj_t*		attacker;
	
	// Player skin colorshift,
	//  0-3 for which color to draw player.
	int			colormap;	
	
	// Overlay view sprites (gun, etc).
	pspdef_t	psprites[NUMPSPRITES];
	
	// True if secret level has been done.
	RDBOOL		didsecret;	
	
	// The player's view pitch is centering back to zero.
	RDBOOL		centering;
	
	// The player can jump if this counter is zero.
	int			jumptics;
	
	int			startspot; 
	
	// Target view to a player (0=disabled, 1=player zero, etc.).
	int			viewlock; // $democam
	
} mbf_player_t;

extern mbf_player_t mbf_players[MAXPLAYERS];

//
// INTERMISSION
// Structure passed e.g. to WI_Start(wb)
//
typedef struct
{
    RDBOOL	in;	// whether the player is in game
    
    // Player stats, kills, collected items etc.
    int		skills;
    int		sitems;
    int		ssecret;
    int		stime; 
    int		frags[MAXPLAYERS];
    int		score;	// current score on entry, modified on return
  
} wbplayerstruct_t;

typedef struct
{
    int		epsd;	// episode # (0-2)

    // if true, splash the secret level
    RDBOOL	didsecret;
    
    // if true show completed header
    RDBOOL	show_completed;
    
    // previous and next levels, origin 0
    int		last;
    int		next;	
    
    int		maxkills;
    int		maxitems;
    int		maxsecret;
    int		maxfrags;

	// b211-0 next map text to draw
	char	nextmapid[48];

	// v220-5 mapid for debrief
	char	inflid[8];
	RDBOOL	runCast;

    // the par time
    int		partime;
    
    // index of this player in game
    int		pnum;	

    wbplayerstruct_t	plyr[MAXPLAYERS];

} wbstartstruct_t;

// Intermission stats for world map / intermission.
extern  wbstartstruct_t		wminfo;	

#endif
//-----------------------------------------------------------------------------
//
// $Log:$
//
//-----------------------------------------------------------------------------
