//-----------------------------------------------------------------------------
//
// DD_SHARE.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __DD_SHARE_H__
#define __DD_SHARE_H__

// Note: where the __cdecl calling convention is used the caller
// must put the right number of arguments on the stack because,
// unlike __stdcall, the caller clears the stack not the function.

#include <stdlib.h>
#include "dd_types.h"
#include "p_think.h"
#include "dd_dfdat.h"
#include "doom/tables.h"

//------------------------------------------------------------------------
//
// General Definitions and Macros
//
//------------------------------------------------------------------------

#define MAX_OF(x, y)	((x) > (y)? (x) : (y))
#define MIN_OF(x, y)	((x) < (y)? (x) : (y))
#define FABS(x) ((x) >= 0.0f ? (x) : -(x))

// Used to replace /255 as *reciprocal255 is less expensive with CPU cycles.
// Note that this should err on the side of being < 1/255 to prevent result
// exceeding 1.0f (e.g. 255 * reciprocal255).
#define reciprocal255	0.003921568627f

typedef struct
{
	char		name[16];		// The name of the action.
	RDBOOL		on;				// True if action is active.
} action_t;

// DD_ indices
enum 
{
	// FI stuff
	DD_DEF_MOBJ = 1,
	DD_DEF_TEXT,
	DD_DEF_INFINE_DEBRIEF,
	DD_DEF_INFINE_BRIEF,

	// Sounds
	DD_DEF_MUSIC,
	DD_MUSIC_ID,
	DD_MUSIC_LUMP,
	DD_CD_TRACK,

	// general
	DD_DEF_STATE,
	DD_DEF_SOUND,
	DD_DEF_VALUE,
};


//------------------------------------------------------------------------
//
// Fixed-Point Math
//
//------------------------------------------------------------------------

#define	FRACBITS			16
#define	FRACUNIT			(1<<FRACBITS)

#define	FINEANGLES			8192
#define	FINEMASK			(FINEANGLES-1)
#define	ANGLETOFINESHIFT	19	// 0x100000000 to 0x2000

#define FIX2FLT(x)		( (x) * 0.0000152587890625f )	// was ( (x) / (float) FRACUNIT )

// Used where the loss of accuracy when discarding the fractional part
// (the low word) is acceptable or where it is known that the low word is zero
#define Q_FIX2FLT(x)	( (float)((x)>>FRACBITS) )

// Level 4: unreferenced inline/local function has been removed
#pragma warning (disable:4514)

// Note: with these inline asm functions a value is
//       returned in eax regardless of compiler warnings.
#pragma warning (disable:4035)

// b211-0 PointOnSide replacement asm code:
//  Returns negative (front), positive (back), or 0 (on).
//  validated against code in ZDoom (123beta33-src) with thanks
__inline int asmPointOnSide(int a, int b, int c, int d)
{
	__asm {
		mov		eax, a
		imul	b
		mov		ebx, eax
		mov		eax, c
		mov		esi, edx
		imul	d
		add		eax, ebx
		adc		edx, esi
		mov		eax, edx
	}
}

__inline fixed_t FixedMul (fixed_t a, fixed_t b)
{
	__asm {
		// Put the parameters in eax and ebx.
		mov		eax, a
		mov		ebx, b
		// multiply.
		imul	ebx
		shrd	eax, edx, 16
	}
}

__inline fixed_t FixedDiv2 (fixed_t a, fixed_t b)
{
	__asm {
		mov		eax, a
		mov		ebx, b
		// The operation.
		cdq
		shld	edx, eax, 16
		sal		eax, 16
		idiv	ebx
	}
}

__inline fixed_t FixedDiv (fixed_t a, fixed_t b)
{
	if((abs(a)>>14) >= abs(b))
	{
		return((a^b)<0 ? DDMININT : DDMAXINT);
	}
	return(FixedDiv2(a, b));
}

//------------------------------------------------------------------------
//
// Key Codes
//
//------------------------------------------------------------------------

// Most character keys are simple ASCII but translated keys (above 127) have a DDKEY number assigned.
enum 
{
	DDKEY_GRAVE				= 0x60,		// translated from 0x29 (Set to 96 given this is what everyone expects)
	DDKEY_RIGHTARROW		= 0x80,
	DDKEY_LEFTARROW,
	DDKEY_UPARROW,
	DDKEY_DOWNARROW,
	DDKEY_F1,
	DDKEY_F2,	
	DDKEY_F3,
	DDKEY_F4,
	DDKEY_F5,
	DDKEY_F6,
	DDKEY_F7,
	DDKEY_F8,
	DDKEY_F9,
	DDKEY_F10,
	DDKEY_F11,
	DDKEY_F12,
	DDKEY_NUMLOCK,
	DDKEY_SCROLL,
	DDKEY_NUMPAD0,
	DDKEY_NUMPAD1,
	DDKEY_NUMPAD2,
	DDKEY_NUMPAD3,
	DDKEY_NUMPAD4,
	DDKEY_NUMPAD5,
	DDKEY_NUMPAD6,
	DDKEY_NUMPAD7,
	DDKEY_NUMPAD8,
	DDKEY_NUMPAD9,
	DDKEY_DECIMAL,
	DDKEY_RSHIFT,
	DDKEY_LSHIFT = DDKEY_RSHIFT,
	DDKEY_RCTRL,
	DDKEY_DIVIDE,		// '/'
	DDKEY_RALT,
	DDKEY_LALT = DDKEY_RALT,
	DDKEY_INS,
	DDKEY_DEL,
	DDKEY_PGUP,
	DDKEY_PGDN,
	DDKEY_HOME,
	DDKEY_END,
	DDKEY_SUBTRACT,				    /* - on numeric keypad */
	DDKEY_ADD,						/* + on numeric keypad */
	DDKEY_BACKSPACE,
	DDKEY_TAB,
	DDKEY_ENTER,
	DDKEY_ESCAPE,
	DDKEY_CAPSLOCK,
	DD_HIGHEST_KEYCODE
};

//------------------------------------------------------------------------
//
// Miscellaneous definitions
//
//------------------------------------------------------------------------

#define RANGECHECK

// v220-4 although only one slot [0] is actually used keeping this
//        to 4 (must be at least 4) prevents having to modify any
//        other part of the code. Perhaps when time allows....
#define MAXPLAYERS		 4		// was 16
#define SAVEMAXPLR		16		// for backward savegame compatibility

#define PI				3.141592654f	// cast as a float
#define	SBARHEIGHT		39			// status bar height at bottom of screen

enum { VX, VY, VZ };				// Vertex indices.
enum { CR, CG, CB, CA };			// Color indices.

// tab_tables.c

//------------------------------------------------------------------------
//
// Map Data
//
//------------------------------------------------------------------------

// Each sector has two of these. A plane_t contains information about
// a plane's movement.
typedef struct
{
	int	target;					// Target height.
	int speed;					// Move speed.
} plane_t;

enum 
{
	PLN_FLOOR,
	PLN_CEILING
};

enum // Sector reverb data indices.
{
	SRD_PRESET,
	SRD_VOLUME,
	SRD_DECAY,
	SRD_DAMPING,
	NUM_REVERB_DATA
};

// Used with sgflags
// Transient flags
// Dlights are hitting this seg. Affects rendering order/blending.
#define DDSEGF_ISLIT			0x000F	// Number of torch lights (15 max)
#define DDSEGF_DLIGHT			0x0010
#define DDSEGF_FACING			0x0020

#define DDSEGF_CLEARMASK		0x00FF

// Fixed flags 0x0100 to 0xFF00 set during analysis
#define	R3DSEGF_IGNORE			0x0100		// Used in renderer to bypass a seg		
#define	R3DSEGF_CHKSKYFIX		0x0200		// Possible skyfix with moving floor
#define	R3DSEGF_SHORTSEG		0x0400		// Set if a short seg (analysis)

// Used with subsector->ssflags.
// DDSUBF_ are the original DD flags, the rest have been added for Risen3D
#define DDSUBF_DLIGHT_FLOOR		0x00000001
#define DDSUBF_DLIGHT_CEILING	0x00000002
#define SSF_FAKETRANS			0x00000004	// v220-3 set if the subsector has fake translucent
#define DDSUBF_CLEAR_MASK		0x00000007	// Flags that are cleared every frame.

#define DDSUBF_MIDPOINT			0x00000008	// Midpoint is tri-fan centre.

#define SSF_FOGUNSET			0x00000010	// used when rendering
#define SSF_NOGLOW				0x00000020	// No glow for whole subsector
#define SSF_SPARE1				0x00000040	// Unused
#define SSF_LOWFLOORSEC			0x00000080	// v220-10 indicates, conditionally, that the ssec's sector has R3Dlowsec set 

#define SSF_INHSECLIGHT			0x00000100	// rev170s
#define SSF_NOOCCLUDE			0x00000200
#define SSF_NOSKYFIX			0x00000400	// b211-0 inhibit skyfix in renderer
#define SSF_CHKFLOORDIFF		0x00000800	// b211-0 compare floor height of lfsec above a liquid floor

#define	SSF_IGNOREFAKE			0x00001000	// Ignore fake sector when rendering (special door types with masked textures spanning 1's lines)
#define	SSF_USEFSEC				0x00002000	// b211-0 force use in renderer
#define	SSF_CONTAINER			0x00004000	// v220-33 Used where a container has sector rings of same sector sided lines in different sectors
#define SSF_NORENDER			0x00008000	// v220-8 set if the subsector can be bypassed when rendering

// next four flags only used by analyser
#define SSF_NOCHANGECS			0x00010000	// rev170s
#define SSF_NOCHANGEFS			0x00020000	// rev170s
#define SSF_SECTSET				0x00040000	// Indicate if set (analyser).
#define SSF_NOCHANGE			0x00080000

#define SSF_FIXED				0x00100000
#define SSF_BASESSECFIX			0x00200000
#define SSF_CSECFIXED			0x00400000	// b211-0 don't reset csec
#define SSF_01INHIBIT			0x00800000	// b211-0 inhibits floor change with line type R3D_01

#define SSF_RENDLFSEC			0x01000000	// v220-4 set if the lfsec is always to be rendered
#define SSF_NOAMFLOOR			0x02000000	// v220-15 set if a floor texture should not be drawn in the automap
#define SSF_NOTRACEN			0x04000000	// v220-16 to be ignored in subfn_TraceNormal
#define SSF_REFONELINE			0x08000000	// v220-22 set if the ssec only references one line

#define SSF_CTRLONLY			0x10000000	// v220-22 set if a subsector references a line flagged SPF_CONTROLLINE
#define SSF_NOSKYFIX2			0x20000000	// v220-22 inhibit skyfix in renderer
#define SSF_FORCESKY			0x40000000	// v220-26 force drawing of sky
#define SSF_SPARE2				0x80000000	// Unused

// For PathTraverse.
#define PT_ADDLINES			1
#define	PT_ADDTHINGS		2
#define	PT_ADDSPEC			4
#define	PT_3DSWITCH			8		// b220-0 for stacked 3D switch handling
#define PT_ADDLINESPTC	   16		// b220-4 ptc line id

#define MAPBLOCKUNITS	128
#define	MAPBLOCKSIZE	(MAPBLOCKUNITS*FRACUNIT)
#define	MAPBLOCKSHIFT	(FRACBITS+7)
#define	MAPBMASK		(MAPBLOCKSIZE-1)
#define	MAPBTOFRAC		(MAPBLOCKSHIFT-FRACBITS)

typedef enum 
{
	ST_HORIZONTAL, ST_VERTICAL, ST_POSITIVE, ST_NEGATIVE
} slopetype_e;

// Frame flags.
#define	FF_FULLBRIGHT		0x00008000		// flag in thing->frame
#define FF_FRAMEMASK		0x00007fff

// mobj ddflags
#define DDMF_DONTDRAW		0x00000001
#define DDMF_SHADOW			0x00000002
#define DDMF_ALTSHADOW		0x00000004
#define DDMF_BRIGHTSHADOW	0x00000008
#define DDMF_VIEWALIGN		0x00000010
#define DDMF_FITTOP			0x00000020	// Identifies spawned ceiling dehacked sprite
#define DDMF_NOINFIGHT		0x00000040	// v220-13 was DDMF_NOFITBOTTOM
#define DDMF_LIGHTSCALE		0x00000180	// Light scale (0: full, 3: 1/4).
#define DDMF_LIGHTOFFSET	0x0000F000	// How to offset light (along Z axis)
#define DDMF_RESERVED		0x00030000	// don't touch these!! (translation class)
#define DDMF_ALWAYSLIT		0x00800000	// Always process DL even if hidden
#define	DDMF_TRANSLATION	0x0C000000	// use a translation table (>>DDMF_TRANSHIFT)

#define	DDMF_TRANSSHIFT			26			// table for player colormaps
#define DDMF_CLASSTRSHIFT		16
#define DDMF_LIGHTSCALESHIFT	7
#define DDMF_LIGHTOFFSETSHIFT	12

// Other unused and/or redirected DDMF_ flags
//#define DDMF_GOINGROUND	0x00040000	// Allow to go in the ground.
//#define DDMF_LOWGRAVITY	0x00080000	// 1/8th gravity (predict)
//#define DDMF_MISSILE		0x00100000	// for the missile scanner
//#define DDMF_FLY			0x00200000	// flying object (doesn't matter if airborne)
//#define DDMF_NOGRAVITY	0x00400000	// isn't affected by gravity (predict)
//#define DDMF_SOLID		0x20000000	// Solid on client side.
//#define DDMF_LOCAL		0x40000000	
//#define DDMF_REMOTE		0x80000000	// This mobj is really on the server.(v220-5 not used)
// Clear mask
//#define DDMF_CLEAR_MASK	0xc0000000

#define DDMF_VALID			0x0C83FFFF	// v220-13 For old savegame compatibility

// R3D redirected ddflags
#define DDF_FORCEFADE		0x00040000	// (was DDMF_GOINGROUND) v220-14 1 second corpsefade - overrides corpsetime
#define DDF_HOLDSTATE		0x00100000	// (was DDMF_MISSILE   ) v220-13 Do not activate in A_Look
#define DDF_AGGRESSIVE		0x00200000	// (was DDMF_FLY       ) v220-13 Increase aggression
#define DDF_IGNORE			0x00400000	// (was DDMF_NOGRAVITY ) v220-14 for backward compatibility. Ignore if health <= 0 (restricted to Amodels)
#define DDF_AMBIENT			0x10000000	// V220-33
#define DDF_DEATHSOUND		0x20000000	// V220-33 (was DDMF_SOLID)
#define DDF_SCPINFIGHT		0x40000000	// v220-35	


// The high byte of the selector is not used for modeldef selecting.
// 1110 0000 = alpha level (0: opaque => 7: transparent 7/8)
#define DDMOBJ_SELECTOR_MASK	0x00ffffff
#define DDMOBJ_SELECTOR_SHIFT	24

#define R3DMOBJ_ALPHA_SHIFT		29
#define R3DMOBJ_ALPHA_CHECK		0xE0000000

//------------------------------------------------------------------------
//
// Refresh
//
//------------------------------------------------------------------------

#define TICRATE			35			// number of tics / second
#define TICSPERSEC		35

#define SCREENWIDTH		320
#define SCREENHEIGHT	200


//------------------------------------------------------------------------
//
// Sound
//
//------------------------------------------------------------------------

// High order sound flags - can be applied on a per source basis
// These are or'ed with the soundid. This allows for up to 0x0FFF = 4095 id's
#define HSF_CANSUSPEND		0x01000000	// Sound can be suspended (door sounds usually)
#define HSF_AMBIENT			0x02000000	// Ambient sound
#define HSF_HALFATTN		0x04000000	// Use attenuation of 0.5
#define HSF_NO_ATTENUATION	0x08000000	// No attenuation

#define HSF_REPEAT			0x10000000
#define HSF_MENUSOUND		0x20000000

#define HSF_SUSPEND			0x40000000	// Only for use with node flags
#define HSF_SNDTHING		0x80000000	// *DO NOT CHANGE* Sound value points to a scripted array

#define HSF_FLAG_MASK		0xff000000

// For (un)linking.
#define DDLINK_SECTOR		0x1
#define DDLINK_BLOCKMAP		0x2
#define DDLINK_NOLINE		0x4

//------------------------------------------------------------------------
//
// Graphics
//
//------------------------------------------------------------------------

// a patch holds one or more columns
// patches are used for sprites and all masked pictures
typedef struct patch_s
{
	short	width;
	short	height;
	short	xoffset;		// pixels to the left of origin
	short	yoffset;		// pixels below the origin
	int		columnofs[8];	// only [width] used
} patch_t;

typedef struct
{
	int	splnum;			// ptr to spritelumps array.
	int realLump;		// Real lump number.
	int	flip;
	int xoffset;
	int yoffset;
	int width;
	int height;
	int numFrames;		// Number of frames the sprite has.
} spriteinfo_t;

//------------------------------------------------------------------------
//
// Console
//
//------------------------------------------------------------------------

// These correspond the good old text mode VGA colors.
#define CBLF_BLACK		0x00000001
#define CBLF_BLUE		0x00000002
#define CBLF_GREEN		0x00000004
#define CBLF_CYAN		0x00000008	
#define CBLF_RED		0x00000010
#define CBLF_MAGENTA	0x00000020
#define CBLF_YELLOW		0x00000040
#define CBLF_WHITE		0x00000080	
#define CBLF_LIGHT		0x00000100
#define CBLF_RULER		0x00000200
#define CBLF_CENTER		0x00000400

// Font flags.
#define DDFONT_WHITE		0x1			// The font data is white, can be colored.

typedef struct
{
	int			flags;
	float		sizeX, sizeY;			// The scale.
	int			height;
	void		(*uiTextOut)(char *text, int x, int y);
	int			(*Width)(char *text);
	void		(*Filter)(char *text);
} ddfont_t;

//------------------------------------------------------------------------
//
// Networking
//
//------------------------------------------------------------------------

// retained for event handling - R3D
typedef struct
{
	char	name[64];
	char	description[128];
	int		players, maxPlayers;
	int		canJoin;
	int		data[3];
} serverinfo_t;

//------------------------------------------------------------------------
//
// Player Data
//
//------------------------------------------------------------------------

// Player flags.
#define DDPF_FIXANGLES		0x01	// Server: send angle/pitch to client.
#define DDPF_FILTER			0x02	// Server: send filter to client.
#define DDPF_FIXPOS			0x04	// Server: send coords to client.
#define DDPF_DEAD			0x08	// Cl & Sv: player is dead.
#define DDPF_CAMERA			0x10	// Player is a cameraman.
#define DDPF_LOCAL			0x20	// Player is local (e.g. player zero).
// v220-35
#define DDPF_DELAY			0x40	// v220-35 Delay angle change with silent line TP

#define PLAYERNAMELEN		81

// Normally one for the weapon and one for the muzzle flash.
#define DDMAXPSPRITES 2

// Player sprite flags.
#define DDPSPF_RENDERED		0x1			// Was rendered.

#endif