//-----------------------------------------------------------------------------
//
// REND_LIST.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __REND_LIST_H__
#define __REND_LIST_H__

#include "p_data.h"
#include "dglib.h"
#include "r_things.h"

// Transient rendpoly flags (UINT)
#define RPF_OCCLUDE			0x00000001
#define RPF_QUADSET			0x00000002
#define RPF_MASKED			0x00000004	// Used to identify type to render (a vissprite).
#define RPF_SKY_MASK		0x00000008	// A sky mask polygon.
#define RPF_LIGHT			0x00000010	// A dynamic light.
#define RPF_DLIT			0x00000020	// Normal list: poly is dynamically lit.
#define RPF_PLANE_GLOW		0x00000040	// Multiply original vtx colors.
#define RPF_WALL_GLOW		0x00000080	// Used with RPF_LIGHT.
#define RPF_DETAIL			0x00000100	// Render with detail (incl. vtx distances)
#define RPF_SHADOW			0x00000200
#define RPF_TRANS			0x00000400	// Water trans with floors else fixed at 0x80 with walls 
#define RPF_ROTATE			0x00000800
#define RPF_TRILOW			0x00001000	// For triwall walls
#define RPF_TRIHI			0x00002000	// For triwall walls
#define RPF_NOFOG			0x00004000	// v220-3 inhibit fog
#define RPF_FOGWALL			0x00008000	// v220-3 fog wall
#define RPF_NOALPHACUT		0x00010000	// v220-9 use blend
#define RPF_NOSPLIT			0x00020000	// v220-10 don't apply seg splitter
#define RPF_SKY_MASKLO		0x00040000	// v220-16 A lower sky mask polygon.(ovflags & OVF_NOSKYWALL ignored in rend_,main)
#define RPF_PLRINSSEC		0x00080000	// v220-22 Flags if player is in the ssec poly (for detail texture purposes)

#define MAXPOLYSIDES		810


typedef enum {
	rp_none,
	rp_quad,			// Wall segment.
	rp_divquad,			// Divided wall segment.
	rp_flat				// Floor or ceiling.
} rendpolytype_e;

typedef enum {
	rl_quads,			// Normal or divided wall segments.
	rl_flats			// Planes (floors or ceilings).
} rendlisttype_e;

typedef struct {
	float	 pos[3];	// X, Y, Z coordinates.
	float	 vpdist;	// Distance to the vertex.
	DGLubyte rgba[4];
} rendpoly_vertex_t;

// rendpoly_t is only for convenience; the data written in the rendering
// flats list data buffer is taken from this struct.
typedef struct {
	rendpolytype_e	type;		
	uint			flags;				// RPF_*.
	float			texw, texh;			// Texture width and height.
	float			texoffx, texoffy;	// Texture coordinates for left/top (in real texcoords).
	DGLuint			tex;				// DG index.
	detailinfo_t	*detail;			// Detail texture name and dimensions.
	struct lumobj_s	*light;				// Used with RPF_LIGHT.
	float			top, bottom;			
	float			length;
	int				numvertices;		// Maximum possible number of vertices for the poly.
	rendpoly_vertex_t vertices[MAXPOLYSIDES];
} rendpoly_t;

// Used for quads - saves using unnecessarily large amounts of memory
// quads list data buffer is taken from this struct. R3D
typedef struct {
	rendpolytype_e	type;		
	uint			flags;				// RPF_*.
	float			texw, texh;			// Texture width and height.
	float			texoffx, texoffy;	// Texture coordinates for left/top (in real texcoords).
	DGLuint			tex;				// DG index.
	detailinfo_t	*detail;			// Detail texture name and dimensions.
	struct lumobj_s	*light;				// Used with RPF_LIGHT.
	float			top, bottom;			
	float			length;
	int				numvertices;		// Number of vertices for the poly.
	rendpoly_vertex_t vertices[2];		// All that's needed for a quad
	struct div_t {
		int num;
		float pos[MAX_DIVS];
	} divs[2];					// For wall segments (two vertices).
} rendpolyq_t;

// PrepareFlat directions.
#define RLPF_NORMAL		0
#define RLPF_REVERSE	1
#define RLPF_TILT		2
#define RLPF_NOTLIT		4

extern float	detailFactor, detailMaxDist, detailScale;

void RL_Init();
void RL_ClearLists();
void RL_DeleteLists();
void RL_AddPolyFlat(rendpoly_t *poly);
void RL_AddPolyWall(rendpolyq_t *poly, seg_t *sg);
void RL_PrepareFlat(rendpoly_t *poly, int numvrts, zvertex_t *vrts, 
					int flags, subsector_t *subsector);
// hassky (introduced in v220-5) removed v220-9
void RL_VertexColors(rendpoly_vertex_t *vtx, int numvertices, int lightlevel, RDBOOL masked);
void RL_RenderAllLists();

#endif