//-----------------------------------------------------------------------------
//
// Copyright (C) 2005 id Software, Inc., Jaakko Kernen
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: am_map.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified to work with R3D, reformatted, moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//  Rewritten to use floating point, bugs fixed
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 15 (v220-15) 08/Jun/2012 gmj
//  line colours revised, floor texture display option added
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
//  1's lines fixed white for reveal 4, 5, pan added to help
// Revision 2.2.0 build 17 (v220-17) 27/Oct/2012 gmj
//  Ssecs which can never be seen by the player suppressed
//  if mapcheat < 3;
// Revision 2.2.0 build 18 (v220-18) 05/Jan/2013 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
//  Reveal 3 cheat cyclic fade added plus some revisions (search CYCLE)
//  to highlight doors needing keys (added). Keys are now also drawn
//  using the hud graphics. Reveal 7 cheat added in order to show
//  subsector boundaries and function to delineate sector boundaries
//  Player can also be made to jump to automap co-ords.
// Revision 2.2.0 build 25 (v220-25) 13/May/2014 gmj
// Revision 2.2.0 build 26 (v220-26) 05/Jul/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll code
//
//
// DESCRIPTION:
//   DOOM Automap. Modified OGL version of the original.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdio.h>
#include <math.h>

#include "doom/mbf_data.h"
#include "doom/p_local.h"
#include "doom/p_start.h"
#include "doom/m_cheat.h"
#include "doom/am_map.h"
#include "doom/st_stuff.h"
#include "doom/Mn_def.h"
#include "doom/hu_stuff.h"
#include "doom/m_menu.h"
#include "doom/d_englsh.h"
#include "doom/tables.h"

#include "dd_wad.h"
#include "p_maputil.h"
#include "r_main.h"
#include "gl_tex.h"
#include "gl_draw.h"
#include "gl_font.h"
#include "gl_main.h"
#include "ui_main.h"
#include "net_main.h"
#include "r3d_xline.h"
#include "con_functions.h"
#include "r3d_compat.h"
#include "r3d_items.h"
#include "m_misc.h"
#include "cvars.h"
#include "r3d_sl3D.h"

// OGL
#include "r3d_OGLfunc.h"

// MACROS ------------------------------------------------------------------

// v220-15 Automap colors
enum {
	LINEBLACK = 0,
	LINEWHITE,
	LINEOFFWHITE,
	LINEGREY,
	LINEGRID,
	LINERED,
	LINEGREEN,
	LINEBLUE,
	LINECYAN,
	LINEYELLOW,
	LINEMAGENTA,
	LINEBROWN,
	LINEORANGE,
	LINEPALEORANGE,
	LINEPALEWHITE,
	LINEPALERED,
	LINEPALEGREEN,
	LINEPALEBLUE,
	LINEPALECYAN,
	LINEOFFYELLOW,
	LINEEND
}lcolor_e;

// Grid sizing
#define GRIDMAX			4096
#define GRIDDEFAULT		 128
#define GRIDMIN			  16

// Action keys
#define AM_PANDOWNKEY	DDKEY_DOWNARROW
#define AM_PANUPKEY		DDKEY_UPARROW
#define AM_PANRIGHTKEY	DDKEY_RIGHTARROW
#define AM_PANLEFTKEY	DDKEY_LEFTARROW
#define AM_ZOOMINKEY	'='
#define AM_ZOOMOUTKEY	'-'
#define AM_GOBIGKEY		'0'
#define AM_FOLLOWKEY	'f'
#define AM_GRIDKEY		'g'
#define AM_HELPKEY		'h'
#define AM_GRIDSIZEKEY	'k'
#define AM_LINESIZEKEY	'l'
#define AM_JUMP			'j'
#define AM_MARKKEY		'm'
#define AM_CLEARMARKKEY	'n'
#define AM_DEADKEY		'o'
#define AM_ROTATEKEY	'r'
#define AM_ALPHAKEY		'u'
#define AM_FLOORKEY		DDKEY_F1

#define AM_NUMMARKPOINTS 10

// pan and zoom defaults
#define F_PANINC		16.f
#define F_PANINCSLOW	1.6f
#define M_ZOOMIN		1.02f
#define M_ZOOMOUT       (1.f/ M_ZOOMIN)

#define FPPLAYERRADIUS	(8.0f*(PLAYERRADIUS>>FRACBITS))

#define FLIP(y)	(current_h - y)

// v220-22
#define CYCLE	0x80000000
#define	DOORKEYBLUE		1
#define	DOORKEYRED		2
#define	DOORKEYYELLOW	4

// TYPES -------------------------------------------------------------------

typedef struct {
    float x, y;
} mpoint_t;

typedef struct {
    mpoint_t a, b;
} mline_t;

typedef struct {
	float 	m_x, m_y;	// LL x,y where the window is on the map (map coords)
	float 	m_w, m_h;	// width/height of window on map (map coords)
} mapparms_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

char *getSpriteIDfromDoomednum(int dmedn);
void Con_Error (char *error, ...);
void subfn_DrawPatchQuads(float w, float h, float x, float y);
subsector_t *R_PointInSubsector(fixed_t x, fixed_t y);
void UI_DrawRectSolid(float x, float y, int w, int h, float brd,
				 ui_color_t *hi, ui_color_t *med, ui_color_t *low, float alpha);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

static void am_HelpDrawer(void);
static void am_ResetMarksec(RDBOOL active);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern int		maplimits[4];

// v220-22
extern GLfloat	linesizes[2];

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int 	mapcheat;
int		automap_active;
int		automapHelp;
float	current_scale_mtof;

// v220-16
int		linedetect;
byte	*lnamap;

// v220-22
RDBOOL	check_plr_in_bounds;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// debug see thidnum_e
char *th_name[TH_NUMTYPES] = {
	{ "ssec  " },
	{ "ssby  " },
	{ "partic" },
	{ "model " },
	{ "lod   " },
	{ "amodel" },
	{ "sprite" },
	{ "wall  " },
	{ "vswall" },
	{ "floor " },
	{ "shtble" },
	{ "ammo  " },
	{ "bottle" },
	{ "dead  " },
	{ "decor " },
	{ "equipm" },
	{ "gore  " },
	{ "keys  " },
	{ "pack  " },
	{ "pickup" },
	{ "sphere" },
	{ "torch " },
	{ "tree  " },
	{ "weapon" }
};

//
// The vector graphics for drawing things in the automap.
//

#define R (FPPLAYERRADIUS/7.0f)

//  Line defs for drawing the player pointing right,
//  (two types) starting from the middle.

static mline_t player_arrow[] = {
    { {   -R+R/8, 0 }, {      R,    0 } },	// -----
    { {        R, 0 }, {  R-R/2,  R/4 } },	// ----->
    { {        R, 0 }, {  R-R/2, -R/4 } },
    { {   -R+R/8, 0 }, { -R-R/8,  R/4 } },	// >---->
    { {   -R+R/8, 0 }, { -R-R/8, -R/4 } },
    { { -R+3*R/8, 0 }, { -R+R/8,  R/4 } },	// >>--->
    { { -R+3*R/8, 0 }, { -R+R/8, -R/4 } }
};

static mline_t cheat_player_arrow[] = {
    { {   -R+R/8,         0 }, {        R,    0 } },	// -----
    { {        R,         0 }, {    R-R/2,  R/6 } },	// ----->
    { {        R,         0 }, {    R-R/2, -R/6 } },
    { {   -R+R/8,         0 }, {   -R-R/8,  R/6 } },	// >----->
    { {   -R+R/8,         0 }, {   -R-R/8, -R/6 } },
    { { -R+3*R/8,         0 }, {   -R+R/8,  R/6 } },	// >>----->
    { { -R+3*R/8,         0 }, {   -R+R/8, -R/6 } },
    { {     -R/2,         0 }, {     -R/2, -R/6 } },	// >>-d--->
    { {     -R/2,      -R/6 }, { -R/2+R/6, -R/6 } },
    { { -R/2+R/6,      -R/6 }, { -R/2+R/6,  R/4 } },
    { {     -R/6,         0 }, {     -R/6, -R/6 } },	// >>-dd-->
    { {     -R/6,      -R/6 }, {        0, -R/6 } },
    { {        0,      -R/6 }, {        0,  R/4 } },
    { {      R/6,       R/4 }, {      R/6, -R/7 } },	// >>-ddt->
    { {      R/6,      -R/7 }, { R/6+R/32, -R/7-R/32 } },
    { { R/6+R/32, -R/7-R/32 }, { R/6+R/10, -R/7 } }
};

#undef R

// Lines for drawing Things
mline_t triangleThing[] = {
    { { -.5f, -.7f }, {  1.0f, 0.f } },
    { { 1.0f , 0.f }, { -.5f,  .7f } },
    { { -.5f,  .7f }, { -.5f, -.7f } }
};

// Lines for drawing crosshair
mline_t am_cross[] = {
    { { -4.f,  .0f }, {  4.f, .0f } },
    { {  .0f, -4.f }, {  .0f, 4.f } },
};

static int	NUMPLYRLINES	  = (sizeof(player_arrow) / sizeof(mline_t));
static int	NUMCHEATPLYRLINES = (sizeof(cheat_player_arrow) / sizeof(mline_t));
static int	NUMCROSSLINES     = (sizeof(am_cross) / sizeof(mline_t));

static unsigned char cheat_amap_seq[] = { 0xb2, 0x26, 0x26, 0x2e, 0xff };	// iddt
static cheatseq_t cheat_amap = { cheat_amap_seq, 0 };

static char buffer[64];
static char num[20];

// Vars based on level size. b211-0 placed in a structure
static struct mapbounds_t {
	float 	min_x, min_y; 
	float 	max_x, max_y;
	float 	max_w, max_h;
} mb;

static mapparms_t	mp;			// current scaling values
static mapparms_t	oldmp;		// used by bigstate to store current scale
static mpoint_t		m_paninc;	// current pan increment

static struct amlimit_t {
	int		*amlinelist;
	int		amlinenum;
	byte	*ammarkssec;
	int		*amsseclist;
	int		amssecnum;
} amlm;

// x,y points used for marking by the automap
static mpoint_t	markpoints[AM_NUMMARKPOINTS];	// where the points are
static int		markpnums[AM_NUMMARKPOINTS];	// lump ptr indices for display
static int		markpointnum = 0;				// next point to be assigned

// Player related
static mbf_player_t *plr;		// the player represented by an arrow

// size of window on screen
static float 	f_w, f_h;

static float 	min_scale_mtof; // Zoom out limit
static float 	max_scale_mtof; // Zoom in  limit

static float 	zoommul_factor;	// Zoom in/out  factor per tic (map coords)

static float	scale_mtof;		// Last scale selected - maximised on init
static float	scale_ftom;		// scales frame-buffer to map-coords (=1/scale_mtof)

static float	mlen;

// User settings
static int	 	gridsize = GRIDDEFAULT;
static RDBOOL 	grid = 0;
static RDBOOL	bigstate ;
static RDBOOL	followplayer;	// Used in automap, set true when game first loaded

static RDBOOL	drawFloor = 0;		// if set draw subsector floor texture

// v220-15
static int		sz_ssamap;
static byte		*ssamap;

// v220-16
static int		sz_lnamap;

// v220-22
static const GLubyte lcv[LINEEND][3] =
{
	{   0,   0,   0 },		// LINEBLACK
	{ 255, 255, 255 },		// LINEWHITE
	{ 180, 180, 180 },		// LINEOFFWHITE
	{ 120, 120, 120 },		// LINEGREY
	{  85,  85,  85 },		// LINEGRID
	{ 255,   0,   0 },		// LINERED
	{   0, 255,   0 },		// LINEGREEN
	{   0,   0, 255 },		// LINEBLUE
	{   0, 255, 255 },		// LINECYAN
	{ 255, 255,   0 },		// LINEYELLOW
	{ 255,   0, 255 },		// LINEMAGENTA
	{ 153, 102,  31 },		// LINEBROWN
	{ 189, 120,  22 },		// LINEORANGE 
	{  90,  60,  11 },		// LINEPALEORANGE 
	{  76,  76,  76 },		// LINEPALEWHITE
	{ 132,   0,   0 },		// LINEPALERED
	{   0, 132,   0 },		// LINEPALEGREEN
	{   0,   0, 132 },		// LINEPALEBLUE
	{   0, 132, 132 },		// LINEPALECYAN
	{ 140, 140,   0 },		// LINEOFFYELLOW
};

static byte		*door_array;
static RDBOOL	solidalpha;
static int		oldcheat;

// v220-35
mobj_t	*map8mobj;


#define		KEYSTUFFNUM	6

static mobj_t	**keythings;
static int		keythingcount = 0;
static int		maxkeythingcount = 0;

typedef struct {
	int		keylnum;
	float	keyscale;
	float	xoff;
	float	yoff;
	RDBOOL	hires;
} keystuff_t;

static keystuff_t keystuff[KEYSTUFFNUM];

static char		*marksecmsg[] = { "sec", "fsec", "sec + fsec", "lfsec", "sec + lfsec", "all" };

static float	maxlinesz;
static int		markspotnum = 0;
static int		markkeynum = 0;

static RDBOOL	set_sector;
static char		str_secn[8];
static int		set_secnpos;
static int		marksecnum;
static sector_t	*am_marksec;
static int		am_marksec_type;
static RDBOOL	show_additional_info;


// CODE --------------------------------------------------------------------

// Use __int64 to prevent overflows
static fixed_t am_MulINT64_4(__int64 a, __int64 b, __int64 c, __int64 d)
{
	return (fixed_t)( ((a * b) + (c * d)) >> FRACBITS );
}

// Get the length of the diagonal from viewable width and height
static float am_GetMLEN()
{
	double a = (double)mp.m_w, b = (double)mp.m_h;
	return (float) sqrt(a*a + b*b);
}

// adds a marker at the current location
static void am_AddMark(void)
{
	markpoints[markpointnum].x = fvx;
	markpoints[markpointnum].y = fvz;

    markpointnum = (++markpointnum) % AM_NUMMARKPOINTS;
}

static void am_ChangeWindowLoc(void)
{
    if (m_paninc.x || m_paninc.y)
	{
		followplayer = false;
		mp.m_x += m_paninc.x;
		mp.m_y += m_paninc.y;
	}
	
	// Ensure values reflect the map limits
    if (mp.m_x + mp.m_w/2.f > mb.max_x) mp.m_x = mb.max_x - mp.m_w/2.f;
    else
	if (mp.m_x + mp.m_w/2.f < mb.min_x) mp.m_x = mb.min_x - mp.m_w/2.f;
	
    if (mp.m_y + mp.m_h/2.f > mb.max_y) mp.m_y = mb.max_y - mp.m_h/2.f;
    else
	if (mp.m_y + mp.m_h/2.f < mb.min_y) mp.m_y = mb.min_y - mp.m_h/2.f;
}

static void am_SetScale()
{
	mp.m_w = f_w * scale_ftom;
	mp.m_h = f_h * scale_ftom;
	mlen = am_GetMLEN();
}

static void am_ClearMarks(void)
{
    int i;

    for (i = 0; i < AM_NUMMARKPOINTS; i++)
		markpoints[i].x = 65536.0f; // means empty

    markpointnum = 0;
}

// v220-15
static void am_Set_ssamap(int recalc)
{
	static int	sscalc = 0;
	subsector_t	*ss = subsectors;
	line_t		*ld;
	int			i, k;

	if (mapcheat > 1) return;
	if (!recalc && sscalc == linedetect) return;
	
	sscalc = linedetect;

	for (i = 0; i < numsubsectors; i++, ss++)
	{
		if (ss->ssflags & (SSF_NORENDER|SSF_NOAMFLOOR) )
		{
			ssamap[i] = false;
			continue;
		}

		if (!ssamap[i] && ss->sector->siflags & SIF_WASSEEN)
		{
			seg_t *sg = &segs[ss->firstline];

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if ( (ld = sg->linedef) && lnamap[ld - lines])
				{
					ssamap[i] = true;
					break;
				}
			}
		}
	}
}

static void am_Start (void)
{
    static int lastwidth = -1, lastheight = -1;

	if (!ppt.listset)
	{
		automap_active = false;
		return;
	}

	automap_active = true;

    if (ma_inf.mapchange)
	{
		// Routines to be called at the start of every level
		current_scale_mtof = .5f;	// Set the default scale multiplier
		followplayer = true;

		// Load the extent of the map (already done in p_blockmap.c)
		mb.min_x = (float)maplimits[BOXLEFT ];
		mb.max_x = (float)maplimits[BOXRIGHT];
		mb.min_y = (float)maplimits[BOXBASE ];
		mb.max_y = (float)maplimits[BOXTOP  ];

		mb.max_w = mb.max_x - mb.min_x;
		mb.max_h = mb.max_y - mb.min_y;

		// Reset the marker array
		// 2 indicates that the markpoints have been restored from a savegame
		if ( !(ma_inf.mapchange & 2) ) am_ClearMarks();

		if (!markpnums[0])
		{
			// Load the marker texture lumps
			int i;
			char namebuf[10];

			for (i = 0; i < 10; i++)
			{
				sprintf(namebuf, "STTNUM%d", i);
				markpnums[i] = W_CheckIWADNumForName(namebuf);
				if (markpnums[i] < 0) Con_Error("Unknown name \"%s\" in IWAD", namebuf);
			}
		}

		ma_inf.mapchange = 0;
		lastwidth = 0;
	}

    if (lastwidth != current_w || lastheight != current_h)
    {
		f_w = (float)current_w;
		f_h = (float)current_h;
		
		min_scale_mtof = MIN_OF(f_w, f_h)/MAX_OF(mb.max_w, mb.max_h)*0.5f;
		max_scale_mtof = f_h * 2 / FPPLAYERRADIUS;
    
		scale_mtof = MAX_OF(current_scale_mtof, min_scale_mtof);
		scale_ftom = 1.0f / scale_mtof;

		m_paninc.x = m_paninc.y = 0;
		zoommul_factor = 1.0f;
		am_SetScale();
							
		P_SetMessage(plr, "FOR HELP PRESS H");
		am_ChangeWindowLoc();
		
		// Save current values
		memcpy(&oldmp, &mp, sizeof(oldmp));

		lastwidth = current_w;
		lastheight = current_h;
    }

	am_Set_ssamap(true);
}

static void am_ActivateNewScale(float scale)
{
    scale_mtof = scale;

	if (scale_mtof < min_scale_mtof)
	{
		scale_mtof = min_scale_mtof;
		zoommul_factor = 1.0f;
	}

    if (scale_mtof > max_scale_mtof)
	{
		scale_mtof = max_scale_mtof;
		zoommul_factor = 1.0f;
	}

	scale_ftom = 1.0f / scale_mtof;

    mp.m_x += mp.m_w/2.f;
    mp.m_y += mp.m_h/2.f;
	am_SetScale();
    mp.m_x -= mp.m_w/2.f;
    mp.m_y -= mp.m_h/2.f;
}

static void am_GoToMarks(void)
{
	if (mapcheat == 7)
	{
		if (followplayer) return;

		// Move player to screen centre cross point
		float px = mp.m_x + mp.m_w/2.f;
		float py = mp.m_y + mp.m_h/2.f;

		ED_MovePlayer( (fixed_t)(px*FRACUNIT), (fixed_t)(py*FRACUNIT));
		am_ResetMarksec(false);
		automap_active = false;
		return;
	}
	
	if (mapcheat == 3)
	{
		if (!keythingcount)
		{
			P_SetMessage(plr, "NO KEYS FOUND");
		}
		else
		{
			static char	buf[32];

			if (markkeynum >= keythingcount) markkeynum = 0;

			m_paninc.x = m_paninc.y = 0;
			mp.m_x = (keythings[markkeynum  ]->x>>FRACBITS) - mp.m_w/2.f;
			mp.m_y = (keythings[markkeynum++]->y>>FRACBITS) - mp.m_h/2.f;

			sprintf(buf, "JUMPED TO KEY %d\n", markkeynum);
			P_SetMessage(plr, buf);

			followplayer = false;
		}
		return;
	}

	if (markpoints[0].x == 65536.0f)
	{
		P_SetMessage(plr, "NO MARKSPOTS SET");
		markspotnum = 0;
	}
	else
	{
		if (markpoints[markspotnum].x == 65536.0f)
			markspotnum = 0;

		mp.m_x = markpoints[markspotnum].x - mp.m_w/2.f;
		mp.m_y = markpoints[markspotnum].y - mp.m_h/2.f;

		if (R3D_edit)
		{
			fixed_t x = (int)(markpoints[markspotnum].x*FRACUNIT);
			fixed_t y = (int)(markpoints[markspotnum].y*FRACUNIT);

			sprintf(buffer, "EDIT VIEW MARKSPOT %d", markspotnum);
			followplayer = true;
			ED_MovePlayer(x, y);
		}
		else
		{
			sprintf(buffer, "AUTOMAP CENTRE MARKSPOT %d", markspotnum);
			followplayer = false;
		}
		
		P_SetMessage(plr, buffer);

		if ((markspotnum++) >= AM_NUMMARKPOINTS) markspotnum = 0;
	}
}

// v220-22 modified to allow cyclical fade
static RDBOOL am_SetColour(int color)
{
	if (color & CYCLE)
	{
		float	lt;

		byte intensity = (leveltime & 0x1F) << 3;

		lt = intensity/256.f;
		if (lt > 1.f ) lt = 1.f;
		if (lt < 0.5f) lt = 1.f - lt;

		if (solidalpha)
		{
			float	at[3] = {123.f/255.f, 99.f/255.f, 79.f/255.f};
			float	m = 2.f, n = 1.f - lt;

			at[0] *= n;
			at[1] *= n;
			at[2] *= n;

			if (color == (LINEBLUE|CYCLE))
			{
				glColor3f(at[0]*m, at[1]*m, at[2] + lt);
			}
			else if (color == (LINERED|CYCLE))
			{
				glColor3f(at[0] + lt, at[1]*m, at[2]*m);
			}
			else if (color == (LINEYELLOW|CYCLE))
			{
				glColor3f(at[0] + lt, at[1] + lt, at[2]*m);
			}
			else if (color == (LINEMAGENTA|CYCLE))
			{
				glColor3f(at[0] + lt, at[1]*m, at[2] + lt);
			}
			else if (color == (LINECYAN|CYCLE))
			{
				glColor3f(at[0]*m, at[0] + lt, at[0] + lt);
			}
			else
			{
				return false;
			}
		}
		else
		{
			if (color == (LINEBLUE|CYCLE))
			{
				glColor3f(0.f, 0.f, lt);
			}
			else if (color == (LINERED|CYCLE))
			{
				glColor3f(lt, 0.f, 0.f);
			}
			else if (color == (LINEYELLOW|CYCLE))
			{
				glColor3f(lt, lt, 0.f);
			}
			else if (color == (LINEMAGENTA|CYCLE))
			{
				glColor3f(lt, 0.f, lt);
			}
			else if (color == (LINECYAN|CYCLE))
			{
				glColor3f(0.f, lt, lt);
			}
			else
			{
				return false;
			}
		}

		return true;
	}

	if (color >= LINEBLACK && color < LINEEND)
	{
		glColor3ubv(lcv[color]);
		return true;
	}

	return false;
}

// Draw grid/character lines. Use fp R3D
static void am_DrawMline(mline_t* ml, int color)
{
	if (am_SetColour(color))
	{
		glVertex2f( (ml->a.x - mp.m_x) * scale_mtof, f_h - (ml->a.y - mp.m_y) * scale_mtof );
		glVertex2f( (ml->b.x - mp.m_x) * scale_mtof, f_h - (ml->b.y - mp.m_y) * scale_mtof );
	}
}

// Draw wall lines. Use fp R3D
static void am_DrawLine(line_t* ld, int color)
{
	if (am_SetColour(color))
	{
		glVertex2f( (FIX2FLT(ld->v1->x) - mp.m_x) * scale_mtof, f_h - (FIX2FLT(ld->v1->y) - mp.m_y) * scale_mtof );
		glVertex2f( (FIX2FLT(ld->v2->x) - mp.m_x) * scale_mtof, f_h - (FIX2FLT(ld->v2->y) - mp.m_y) * scale_mtof );
	}
}

// Rotate player arrow line character.
static void am_Rotate(float *x, float *y, angle_t a)
{
	fixed_t tx = (fixed_t)(*x*FRACUNIT);
	fixed_t ty = (fixed_t)(*y*FRACUNIT);

	a >>= ANGLETOFINESHIFT;
	*x = FIX2FLT(am_MulINT64_4(tx, finecosine[a], ty,  -finesine[a]));
	*y = FIX2FLT(am_MulINT64_4(tx,   finesine[a], ty, finecosine[a]));
}

// v220-15
static void am_DrawFloorFlats()
{
	subsector_t		*ss;
	sector_t		*sc;
	float			ofx, ofy;
	int				*flatslist;
	int				i, k, p;
	byte			ll;

	am_Set_ssamap(false);

	if (am_marksec && mapcheat == 7) glColor3ubv(lcv[LINECYAN]);

	for (i = 0; i < amlm.amssecnum; i++)
	{
		ofx = ofy = 0.f;
		p = amlm.amsseclist[i];
		ss = &subsectors[p];

		if (am_marksec && mapcheat == 7)
		{
			RDBOOL	a = ss->sector == am_marksec;
			int		n = am_marksec - sectors;

			if (am_marksec_type == 1)
			{
				if (ss->floorsec != n) continue;
			}
			else if (am_marksec_type == 2)
			{
				if (!a && ss->floorsec != n) continue;
			}
			else if (am_marksec_type == 3)
			{
				if (ss->lfsec != n) continue;
			}
			else if (am_marksec_type == 4)
			{
				if (!a && ss->lfsec != n) continue;
			}
			else if (am_marksec_type == 5)
			{
				if (!a && ss->floorsec != n && ss->lfsec != n) continue;
			}
			else
			{
				if (!a) continue;
			}

			glBegin(GL_POLYGON);
			
			for (k = 0; k < ss->numverts; k++)
			{
				glTexCoord2f(ss->verts[k].x/64.f, -ss->verts[k].y/64.f);
				glVertex2f( (ss->verts[k].x - mp.m_x) * scale_mtof, f_h - (ss->verts[k].y - mp.m_y) * scale_mtof );
			}

			glEnd();
		}
		else if (mapcheat > 1 || ssamap[p])
		{
			float	hasscaleX = 1.f;
			float	hasscaleY = 1.f;	// used with 3D flats
			int		fpic = -1;

			sc = NULL;
			ll = 0;

			// If necessary find floor and ceiling player is between
			if ( (flatslist = ss->flats3DArray) != NULL)
			{
				sector_t	*lsec = NULL;
				int			floor = ss->sector->floorheight;
				int			n = 0, count = *(int*)flatslist++;

				ll = (byte)ss->sector->lightlevel;

				while(n < count)
				{
					// If the viewpoint is between floors then the lower floor values are used
					if (viewz >= floor && viewz <= sectors[flatslist[1]].floorheight + flatslist[2] + flatslist[3])
						break;

					lsec = &sectors[flatslist[0]];

					if (++n == count) break;

					floor = sectors[flatslist[1]].floorheight + flatslist[2] + flatslist[3];
					flatslist += FLIST_ENTRIES;
				}

				if (lsec)
				{
					fpic = lsec->floorpic;

					hasscaleX = lsec->scalingValueX;
					hasscaleY = lsec->scalingValueY;

					ofx = FIX2FLT(lsec->flooroffx);
					ofy = FIX2FLT(lsec->flooroffy);
				}
				else
				{
					fpic = ss->sector->floorpic;
					ofx = FIX2FLT(ss->sector->flooroffx);
					ofy = FIX2FLT(ss->sector->flooroffy);
				}
			}

			if (fpic < 0)
			{
				if (!sc) sc = ss->floorsec >= 0 ? &sectors[ss->floorsec] : ss->sector;
				if (!ll) ll = (byte)(sc->floorlightsec >= 0 ? sectors[sc->floorlightsec].lightlevel : sc->lightlevel);

				if (viewz > sc->withheight)	// Note: DDMAXINT if unset
				{
					fpic = sc->texwithheight[0];
				}
				else
				{
					fpic = sc->floorpic;

					if (sc->heightsec >= 0)
					{
						int fht = sectors[sc->heightsec].floorheight;

						if (fht <= sc->floorheight) sc = &sectors[sc->heightsec];
					}
					else if (ss->lfsec >= 0
						&& viewz >= sectors[ss->lfsec].floorheight
						&& sectors[ss->lfsec].floorheight > sc->floorheight)
					{
						fpic = sectors[ss->lfsec].floorpic;
					}
				}

				ofx = FIX2FLT(sc->flooroffx), ofy = FIX2FLT(sc->flooroffy);
			}

			if (fpic < 0 || !CheckFlatSet(fpic)) continue;	// not to be set or used here

			GL_Bind(GL_PrepareFlat(fpic));
			glBegin(GL_POLYGON);
			glColor3ub(ll, ll, ll);

			float h = ctex->rendh;
			float w = ctex->rendw;

			w *= hasscaleX, h *= hasscaleY;

			for (k = 0; k < ss->numverts; k++)
			{
				glTexCoord2f((ofx + ss->verts[k].x)/w, -(ofy + ss->verts[k].y)/h);
				glVertex2f( (ss->verts[k].x - mp.m_x) * scale_mtof, f_h - (ss->verts[k].y - mp.m_y) * scale_mtof );
			}

			glEnd();
		}
	}
}

// Draws grid lines to extent of map boundaries.
static void am_DrawGrid()
{
	int		gs = gridsize;
	int		x, y;
	int		startx, endx;
	int		starty, endy;

	mline_t		ml;

	//------------------------------------------------------

	// Set limits for horizontal gridlines
	startx = (int)mb.min_x;
	endx   = (int)mb.max_x;

	// Set the line vtxs y value
	ml.a.y = (float)mb.min_y;
	ml.b.y = (float)mb.max_y;

	// draw vertical gridlines
	for (x = startx - (startx % 64); x <= endx; x += gs)
	{
		if (x >= mb.min_x && x <= mb.max_x)
		{
			ml.a.x = ml.b.x = (float)x;
			am_DrawMline(&ml, LINEGRID);
		}
	}
	
	// Set limits for vertical gridlines
	starty = (int)mb.min_y;
	endy   = (int)mb.max_y;

	// Set the line vtxs x value
	ml.a.x = (float)mb.min_x;
	ml.b.x = (float)mb.max_x;

	// draw horizontal gridlines
	for (y = starty - (starty % 64); y <= endy; y += gs)
	{
		if (y >= mb.min_y && y <= mb.max_y)
		{
			ml.a.y = ml.b.y = (float)y;
			am_DrawMline (&ml,LINEGRID);
		}
	}
}

// Determine and draws lines as required.
// Clipping is left up to the graphics card as
// this is faster than calculating it.
static void am_DrawWalls(void)
{
extern byte walkCheck[];

    line_t	*ld;
	ushort	flags;
    int		i, n;
	RDBOOL	cdiff, fdiff;
	
	linedetect = 0;

	if (num_xlsame && mapcheat >= 3)
	{
		xlinemem_t	*lm;

		am_SetColour(LINEPALEORANGE);

		for (i = 0; i < num_xlsame; i++)
		{
			lm = &xlinemem[i];

			glVertex2f( (lm->startx - mp.m_x) * scale_mtof, f_h - (lm->starty - mp.m_y) * scale_mtof );
			glVertex2f( (lm->endx   - mp.m_x) * scale_mtof, f_h - (lm->endy   - mp.m_y) * scale_mtof );
		}
	}

	for (i = 0; i < amlm.amlinenum; i++)
	{
		n = amlm.amlinelist[i];
		ld = &lines[n];
		flags = ld->ml_flags;

		if (mapcheat || (flags & ML_MAPPED && !(flags & ML_DONTDRAW)))
		{
			if (mapcheat > 3 && !ld->backsector)
			{
				am_DrawLine(ld, LINEWHITE);
				continue;
			}

			if (mapcheat == 6 || mapcheat == 7) continue;
			
			if (   ld->frontsector->siflags & SIF_WASSECRET
				|| (ld->backsector && ld->backsector->siflags & SIF_WASSECRET))
			{
				am_DrawLine(ld, LINEMAGENTA);
				if (mapcheat < 2) { lnamap[n] = true; linedetect++; }
				continue;
			}

			if (mapcheat < 4 && ld->spflags & SPF_CONTROLLINE)
			{
				if (mapcheat > 1 || ld->ml_flags & ML_MAPPED) am_DrawLine(ld, LINEGREY);
				continue;
			}

			// check for locked door line (bug fix v220-26)
			if (mapcheat == 3)
			{
				if (door_array[n])
				{
					if (door_array[n] == DOORKEYBLUE  ) { am_DrawLine(ld, LINEBLUE|CYCLE  ); continue; }
					if (door_array[n] == DOORKEYRED   ) { am_DrawLine(ld, LINERED|CYCLE   ); continue; }
					if (door_array[n] == DOORKEYYELLOW) { am_DrawLine(ld, LINEYELLOW|CYCLE); continue; }
				}

				if (   ld->frontsector->special == 9
					|| ld->frontsector->special & SECRET_MASK
					|| (ld->backsector && ld->backsector->special == 9)
					|| (ld->backsector && ld->backsector->special & SECRET_MASK))
				{
					am_DrawLine(ld, LINEMAGENTA|CYCLE);
					continue;
				}
			}

			if (mapcheat < 2)
			{
				if (flags & ML_DONTDRAW)
				{
					// v220-16 needed for floor flats
					if (!lnamap[ld - lines])
					{
						if (ld->ml_flags & ML_DRAWFLATS)
						{
							lnamap[n] = true; linedetect++;
						}
						else if (ld->frontsector->siflags & SIF_WASSECRET)
						{
							lnamap[n] = true; linedetect++;
						}
						else if (ld->backsector
							&& ld->backsector->siflags & SIF_WASSECRET)
						{
							lnamap[n] = true; linedetect++;
						}
					}

					continue;
				}

				if (flags & ML_MAPPED)
				{
					// v220-16 needed for floor flats
					lnamap[n] = true; linedetect++;
				}
				else if (mapcheat)
				{
					am_DrawLine(ld, LINEGREY); continue;
				}
			}

			if (!ld->backsector || ld->ml_flags & ML_SECRET)
			{
				am_DrawLine(ld, solidalpha ? LINEBLACK : LINEPALERED); continue;
			}

			// Check for walkover tp line
			if (   mapcheat >= 3
				&& ld->special > 0
				&& ld->special < 270
				&& ld->tag)
			{
				byte n = walkCheck[ld->special];

				if (n == 2)
				{
					am_DrawLine(ld, LINEPALECYAN);
					continue;
				}

				if (n == 3)
				{
					am_DrawLine(ld, LINEPALEBLUE);
					continue;
				}
			}

			if (   ld->specType >= R3D_40
				|| ld->frontsector->stptr >= 0
				|| ld->backsector->stptr >= 0)
			{
				am_DrawLine(ld, LINEPALEGREEN);
				continue;
			}

			fdiff = ld->backsector->floorheight != ld->frontsector->floorheight;
			cdiff = ld->backsector->ceilingheight != ld->frontsector->ceilingheight;

			if (mapcheat == 8)
			{
				if (!fdiff) continue;
			}

			if (fdiff || cdiff) // floor and/or ceiling level change
			{
				am_DrawLine(ld, fdiff && cdiff ? LINEOFFYELLOW :
								  fdiff ? LINEPALERED : LINEBROWN);
			}
			else if (mapcheat == 8)
			{
				// do nothing
			}
			else if (ld->ovflags & OVF_MAPPED)
			{
				am_DrawLine(ld, LINERED);
			}
			else if (mapcheat >= 3)
				am_DrawLine(ld, LINEORANGE);
		}
    }
}

static line_t	**msclines = NULL;
static int		maxmsclines = 0;

static void am_DrawSubsectors(void)
{
	subsector_t	*ss;
	zvertex_t	*vtx;
	seg_t		*sg;
	int			numvrts;
	int			i, k, n;
	int			nummsclines = 0;

	validcount++;
	glBegin(GL_LINES);

	for (i = 0; i < amlm.amssecnum; i++)
	{
		ss = &subsectors[amlm.amsseclist[i]];
		vtx = ss->verts;
		numvrts = ss->numverts;
		n = numvrts - 1;
		
		if (ss->ssflags & SSF_REFONELINE)
			glColor3ubv(lcv[LINEOFFYELLOW]);
		else
			glColor3ubv(lcv[LINEPALEWHITE]);

		for (k = 0; k < numvrts; k++)
		{
			glVertex2f( (vtx->x - mp.m_x) * scale_mtof, f_h - (vtx->y - mp.m_y) * scale_mtof );
			if (k == n) vtx = ss->verts; else vtx++;
			glVertex2f( (vtx->x - mp.m_x) * scale_mtof, f_h - (vtx->y - mp.m_y) * scale_mtof );
		}

		for (k = 0, sg = &segs[ss->firstline]; k < ss->linecount; k++, sg++)
		{
			if (am_marksec && (sg->frontsector == am_marksec || sg->backsector == am_marksec))
			{
				if (sg->linedef && sg->linedef->validcount != validcount)
				{
					sg->linedef->validcount = validcount;

					if (maxmsclines <= nummsclines)
					{
						maxmsclines += 16;
						msclines = (line_t**)W_Realloc(msclines, maxmsclines * sizeof(line_t*), STATICID, 0);
					}
					msclines[nummsclines++] = sg->linedef;
				}

				continue;
			}
			else if (sg->frontsector == sg->backsector)
			{
				glColor3ubv(lcv[LINEPALERED]);
			}
			else if (!sg->backsector)
			{
				glColor3ubv(lcv[LINEWHITE]);
			}
			else
			{
				continue;
			}

			glVertex2f( (sg->fv1[VX] - mp.m_x) * scale_mtof, f_h - (sg->fv1[VY] - mp.m_y) * scale_mtof );
			glVertex2f( (sg->fv2[VX] - mp.m_x) * scale_mtof, f_h - (sg->fv2[VY] - mp.m_y) * scale_mtof );
		}
	}

	glEnd();

	if (nummsclines)
	{
		// v220-25 draw last as show up better
		glLineWidth(cfg.linesz + 1.f);
		glBegin(GL_LINES);
		glColor3ubv(lcv[LINECYAN]);

		for (i = 0; i < nummsclines; i++)
		{
			line_t *ld = msclines[i];

			glVertex2f( (FIX2FLT(ld->v1->x) - mp.m_x) * scale_mtof, f_h - (FIX2FLT(ld->v1->y) - mp.m_y) * scale_mtof );
			glVertex2f( (FIX2FLT(ld->v2->x) - mp.m_x) * scale_mtof, f_h - (FIX2FLT(ld->v2->y) - mp.m_y) * scale_mtof );
		}
		glEnd();
 		glLineWidth(cfg.linesz);
	}
}

static void am_DrawLineCharacter(mline_t *lineguy, int lineguylines,
				float scale, angle_t angle, int color, float x, float y)
{
    int		i;
    mline_t	l;

    for (i = 0; i < lineguylines; i++)
    {
		l.a.x = lineguy[i].a.x;
		l.a.y = lineguy[i].a.y;

		if (scale)
		{
			l.a.x = scale * l.a.x;
			l.a.y = scale * l.a.y;
		}

		if (angle) am_Rotate(&l.a.x, &l.a.y, angle);

		l.a.x += x;
		l.a.y += y;

		l.b.x = lineguy[i].b.x;
		l.b.y = lineguy[i].b.y;

		if (scale)
		{
			l.b.x = scale * l.b.x;
			l.b.y = scale * l.b.y;
		}

		if (angle) am_Rotate(&l.b.x, &l.b.y, angle);
	
		l.b.x += x;
		l.b.y += y;

		am_DrawMline(&l, color);
    }
}

static void am_DrawThingBox(float x, float y, float radius, int color)
{
	mline_t		ml;
	// Draw left and right box sides
	ml.a.y = y - radius;
	ml.b.y = y + radius;
	ml.a.x = ml.b.x = x - radius;
	am_DrawMline(&ml, color);
	ml.a.x = ml.b.x = x + radius;
	am_DrawMline(&ml, color);
	// Draw bottom and top box line
	ml.a.x = x - radius;
	ml.b.x = x + radius;
	ml.a.y = ml.b.y = y - radius;
	am_DrawMline(&ml, color);
	ml.a.y = ml.b.y = y + radius;
	am_DrawMline(&ml, color);
}

// v220-35
static void am_DrawThings(int show_all)
{
	thinker_t	*th;
    mobj_t		*t;
	
	float	px, py;
	fixed_t	ftpxpos, ftpypos;
	fixed_t	ftpxneg, ftpyneg;

	if (mapcheat == 8)
	{
		px = mp.m_x + mp.m_w/2.f;
		py = mp.m_y + mp.m_h/2.f;
		ftpxpos = (fixed_t)(px*FRACUNIT) + 4*FRACUNIT;
		ftpypos = (fixed_t)(py*FRACUNIT) + 4*FRACUNIT;
		ftpxneg = (fixed_t)(px*FRACUNIT) - 4*FRACUNIT;
		ftpyneg = (fixed_t)(py*FRACUNIT) - 4*FRACUNIT;
	}

	for (th = thinkercap.next ; th != &thinkercap ; th = th->next)
	{
		t = (mobj_t*)th;

		if (t->thinker.function != (void (__cdecl *)(void))P_MobjThinker) continue;
		if ((t->dplayer && t->dplayer->mo == t) || t->flags2 & MF2_NORENDER) continue;

		if (mapcheat == 8)
		{
			if (   (t->health > 0 && t->flags & MF_SHOOTABLE && t->info->seestate)	// enemy
				|| (t->flags3 & MF3_AMODEL && t->flags & MF_SOLID && !(t->flags & MF_NOCLIP)) )
			{
				if (amlm.ammarkssec[t->subsector - subsectors])	// $limitcheck
				{
					am_DrawThingBox(FIX2FLT(t->x), FIX2FLT(t->y), FIX2FLT(t->radius),
						t->flags3 & MF3_AMODEL ? LINEBLUE : LINEYELLOW);

				/*	am_DrawLineCharacter(triangleThing, 3, 16.0f,	// &remove
							// angle
							t->angle,
							// colour
							t->flags3 & MF3_AMODEL ? LINEBLUE : LINEYELLOW,
							FIX2FLT(t->x),
							FIX2FLT(t->y));
				*/

					if (   !followplayer
						&& t->x >= ftpxneg && t->x <= ftpxpos
						&& t->y >= ftpyneg && t->y <= ftpypos)
					{
						map8mobj = t;
					}
				}
			}
			continue;
		}
		
		if (   !(!cfg.show_dead && t->health <= 0)	// v220-22
			&& (show_all || (t->flags & (MF_SPECIAL|MF_SHOOTABLE|MF_CORPSE) && !(t->flags3 & MF3_AMODEL))) )
		{
			// v220-22 
			if (mapcheat == 3 && (t->sprite >= SPR_BKEY && t->sprite <= SPR_YSKU))
			{
				if (keythingcount >= maxkeythingcount)
				{
					maxkeythingcount += 8;
					keythings = (mobj_t**)W_Realloc(keythings, maxkeythingcount * sizeof(mobj_t*), STATICID, 0); 
				}

				keythings[keythingcount++] = (mobj_t*)t;
			}
			else if (mapcheat != 3)
			{
				if (amlm.ammarkssec[t->subsector - subsectors])	// $limitcheck
				{
					am_DrawLineCharacter(triangleThing, 3, 16.0f,
						// angle
						t->flags  & MF_COUNTITEM ? 0 :
						t->flags  & (MF_SPECIAL|MF_DROPPED) ? ANG90 : t->angle,
						// colour
						!t->type ? LINEOFFWHITE :
						t->flags3 & MF3_AMODEL   ? LINEBLUE :
						t->flags  & MF_MISSILE   ? LINEYELLOW :
						t->flags  & MF_COUNTITEM ? LINECYAN|CYCLE :
						t->flags  & (MF_SPECIAL|MF_DROPPED) ? LINECYAN :
						t->flags  & MF_CORPSE    ? LINEGREEN :
						t->flags  & MF_SHOOTABLE ? LINERED : LINEBROWN,
						FIX2FLT(t->x),
						FIX2FLT(t->y));
				}
			}
		}
    }
}

// Modified for R3D
static void am_DrawMarks(void)
{
	float	w, h, fx, fy;
	int		i;

	w  = 40.0f * scale_mtof;
	h  = 48.0f * scale_mtof;
	if (w < 5.f) w = 5.f, h = 6.0f;

	for (i = 0; i < AM_NUMMARKPOINTS; i++)
	{
		if (markpoints[i].x != 65536.0f)
		{
			fx = (markpoints[i].x - 20.f - mp.m_x) * scale_mtof;
			if (fx + w < 0 || fx - w > f_w) continue;

			fy = f_h - (markpoints[i].y + 24.f - mp.m_y) * scale_mtof;
			if (fy + h < 0 || fy - h > f_h) continue;
			
			if (GL_SetPatch(markpnums[i], 0))
				subfn_DrawPatchQuads(w, h, fx, fy);
		}
	}
}

// v220-22
static void am_DrawKeys(void)
{
	static char *sname[] = { "BKEYA0" , "RKEYA0" , "YKEYA0" , "BSKUA0" , "RSKUA0" , "YSKUA0" , NULL};
	static char *hname[] = { "STKEYS0", "STKEYS2", "STKEYS1", "STKEYS3", "STKEYS5", "STKEYS4", NULL};

	keystuff_t	*ks;
	int			i, lump, snum;
	float		fx, fy, w, h;

	//----------------------------------------
	
	for (i = 0; i < keythingcount; i++)
	{
		lump = -1;
		snum = keythings[i]->sprite - SPR_BKEY;

		if (snum < 0 || snum >= KEYSTUFFNUM) continue;	// insurance

		ks = &keystuff[snum];

		if ( (lump = ks->keylnum) == -2)
		{
			lump = ks->keylnum = W_CheckNumForName(sname[snum]);

			// set defaults
			ks->keyscale = 2.f;
			ks->xoff = isarctic ? -64.f : (snum > 2) ?  -13.f : -14.f;
			ks->yoff = isarctic ? 116.f : (snum > 2) ?   16.f :  16.f;
			
			// check for standard hires texture if needed
			if (lump >= 0 && hiresTextures && lumpinfo[lump].handle == IWADhandle)
			{
				lump = W_CheckNumForName(hname[snum]);

				if (lump >= 0)
				{
					ks->keylnum = lump;
					ks->keyscale = 5.0f;
					ks->xoff = -18.f;
					ks->yoff = (snum > 2) ? 18.f : 14.f;
					ks->hires = true;
				}
				else
					lump = ks->keylnum;
			}
		}

		if (lump < 0 || !GL_SetPatch(lump, 0)) continue;

		// $limitcheck
		w = lumptexinfo[lump].origw * ks->keyscale * scale_mtof;
		fx = (FIX2FLT(keythings[i]->x) + ks->xoff - mp.m_x) * scale_mtof;
		if (fx + w < 0 || fx - w > f_w) continue;

		h = lumptexinfo[lump].origh * ks->keyscale * scale_mtof;
		fy = f_h - (FIX2FLT(keythings[i]->y) + ks->yoff - mp.m_y) * scale_mtof;
		if (fy + h < 0 || fy - h > f_h) continue;

		subfn_DrawPatchQuads(w, h, fx, fy);
	}
}

static fixed_t cbox[4];	// cursor box

static RDBOOL PIT_GetLine(line_t *ld)
{
	if (   cbox[BOXRIGHT] <= ld->bbox[BOXLEFT ]
		|| cbox[BOXLEFT ] >= ld->bbox[BOXRIGHT]
		|| cbox[BOXTOP  ] <= ld->bbox[BOXBASE ]
		|| cbox[BOXBASE ] >= ld->bbox[BOXTOP  ]
		|| P_BoxOnLineSide(cbox, ld) != -1)
	{
		// Line does not cross the ldbox box.
		return true;
	}

	return false;
}

static int P_GetLineIterator(int x, int y, RDBOOL(*func)(line_t*,void*))
{
	if (x >= 0 && y >= 0 && x < bmapwidth && y < bmapheight)
	{
		int			offset = *(blockmap + y*bmapwidth + x);
		const long	*list = blockmaplump + offset;
		line_t		*ld;

		// skip starting delimiter
		while( *(++list) >= 0)
		{
			ld = &lines[*list];

			if (ld->validcount != validcount)
			{
				ld->validcount = validcount;
				
				if (!PIT_GetLine(ld)) return ld - lines;
			}
		}
	}
	
	return -1;
}

// validcount must be set by the calling code
static int am_FindLine(fixed_t x, fixed_t y, fixed_t radius)
{
	int		bx, by, xl, yl, xh, yh, lnum = -1;

	// Set up a line iterator for doing the linking.
	M_setBoxOffsets  (cbox, x, y, radius);
	M_getBlockOffsets(cbox, MAXRADIUS, &xl, &xh, &yl, &yh);

	for(bx = xl; bx <= xh; bx++)
	{
		for(by = yl; by <= yh; by++)
		{
			if ( (lnum = P_GetLineIterator(bx, by, (int (__cdecl *)(struct line_s *,void *))PIT_GetLine)) >= 0 )
				return lnum;
		}
	}
	return -1;
}

static void am_HelpDrawer(void)
{
	if (!R3D_edit || automapHelp)
	{
		subsector_t	*ss = NULL;

		char	buf[40];
		int		angle = (int)vang;
		int		y = 10, w = 80;
		int		h = FR_TextHeight("A");
		int		x = current_w - w - 10;

		// ------------------------------

		FR_SetFont(0);

		// Go into screen projection mode.
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(0, current_w, current_h, 0, -1, 1);

		if (!R3D_edit)
		{
			char	str[3][40];
			float	px, py;
			fixed_t	ftpx, ftpy;
			int		xoff, yoff = h + 8, yinc = h + 2;
			int		d = yinc * 6 + 14, i;
			uint	e = 10;
			int		lnum[4] = { -1, -1, -1, -1 };

			if (mapcheat == 7 || mapcheat == 8)
			{
				px = mp.m_x + mp.m_w/2.f;
				py = mp.m_y + mp.m_h/2.f;
				e = 14;
			}

			if (mapcheat == 7)
			{
				ftpx = (fixed_t)(px*FRACUNIT);
				ftpy = (fixed_t)(py*FRACUNIT);

				if (am_marksec) d += yinc*5; else d += yinc*3;

				if (!followplayer && ge_inf.devmap)
				{
					validcount++;
					lnum[0] = am_FindLine(ftpx, ftpy, 0x5000);

					d += yinc * 2;
					if (lnum[0] >= 0) d += yinc * 20;
				}
			}
			else if (map8mobj)	// only set if mapcheat == 8
			{
				int ln;
				int		lcount = 0;

				d += yinc;

				validcount++;

				while (lcount < 3)
				{
					if ( (ln = am_FindLine(map8mobj->x, map8mobj->y, map8mobj->radius)) < 0)
						break;

					if (lines[ln].ml_flags & (ML_BLOCKING|ML_BLOCKMONSTERS))
					{
						lnum[lcount++] = ln;
					}
					else if (lines[ln].frontsector->floorheight > map8mobj->z)
					{
						lnum[lcount++] = ln;
					}
					else if (lines[ln].backsector
						&& lines[ln].backsector->floorheight > map8mobj->z)
					{
						lnum[lcount++] = ln;
					}
				}

				if (lcount) d += (lcount * yinc);
			}
			else if (cfg.counterCheat && mapcheat != 8)
			{
				sprintf(str[0], "Kills : %i/%i ", plr->killcount  , totalkills);
				sprintf(str[1], "Items : %i/%i ", plr->itemcount  , totalitems);
				sprintf(str[2], "Secret: %i/%i ", plr->secretcount, totalsecret);

				d += yinc*3;

				for (i = 0; i < 3; i++)
				{
					if (strlen(str[i]) > e)
						e = strlen(str[i]);
				}
			}

			xoff = current_w - e*9;

			if ((mapcheat == 7 || mapcheat == 8) && lnum[0] >= 0)
			{
				UI_DrawRectSolid((float)(xoff - 10), (float)y, e*9 + 4, d, 4, 
					UI_COL(UIC_BRD_HI),
					UI_COL(UIC_BG_DARK),
					UI_COL(UIC_BRD_LOW), 0.9f);
			}
			else
			{
				UI_DrawRectFill((float)(xoff - 10), (float)y, e*9 + 4, d, 4, 
					UI_COL(UIC_BRD_HI),
					UI_COL(UIC_BRD_MED),
					UI_COL(UIC_BRD_LOW), 0.3f);
			}

			if (map8mobj)
			{
				sprintf(buf, "Thing %i", MAPNUM(map8mobj));
			}
			else
			{
				if (commercial_game)
					sprintf(buf, "map%02i", gamemap);
				else
					sprintf(buf,"E%iM%i", gameepisode, gamemap);
			}

			UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);

			if (map8mobj)
			{
				sprintf(buf, "ID %s", getSpriteIDfromDoomednum(map8mobj->info->doomednum));
				UI_TextOutEx(buf, xoff, (yoff += yinc), false, true, UI_COL(UIC_TEXT), 1);
			}

			if (mapcheat == 7 || mapcheat == 8)
			{
				static char *ptr[] = { "PLAYER:", "CURSOR:" };

				yoff += yinc;

				UI_TextOutEx(ptr[!followplayer], xoff, (yoff += yinc), false, true, UI_COL(UIC_TEXT), 1);
				sprintf(buf, "X % 6i", followplayer ? viewx>>16 : (int)px);
				UI_TextOutEx(buf, xoff, (yoff += yinc), false, true, UI_COL(UIC_TEXT), 1);
				sprintf(buf, "Y % 6i", followplayer ? viewy>>16 : (int)py);
				UI_TextOutEx(buf, xoff, (yoff += yinc), false, true, UI_COL(UIC_TEXT), 1);
				yoff += yinc;
			}

			if (mapcheat == 7)
			{
				ss = followplayer ? ppt.view_subsector : R_PointInSubsector(ftpx, ftpy);

				if (set_sector)
				{
					sprintf(buf, "SECN >%s", str_secn);
					if (leveltime & 0x10) strcat(buf, "_");
				}
				else
				{
					sprintf(buf, "SEC  %6i", ss->sector - sectors);
				}

				yoff += yinc;
				UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);

				sprintf(buf, "SSEC %6i", ss - subsectors); yoff += yinc;
				UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);

				if (!followplayer && ge_inf.devmap)
				{
					extern char	*globalskytex;

					char *name = lumpinfo[ss->ssfloorpic].name;
				
					if (!stricmp("F_SKY1", name)) name = globalskytex;

					sprintf(buf, "F %s", name); yoff += yinc;
					UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);

					name = lumpinfo[ss->ssceilpic].name;
					if (!stricmp("F_SKY1", name)) name = globalskytex;

					sprintf(buf, "C %s", name); yoff += yinc;
					UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);
				}

				if (lnum[0] >= 0)
				{
					int sd0 = lines[lnum[0]].sidenum[0];
					int sd1 = lines[lnum[0]].sidenum[1];
					
					char *name;
					int  idx;

					yoff += yinc;
					sprintf(buf, "LINE % 6i", lnum[0]); yoff += yinc;
					UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);

					sprintf(buf, "Type R3D_%.2X", lines[lnum[0]].specType); yoff += yinc;
					UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);

					sprintf(buf, "FS %.6i", sides[sd0].sector - sectors); yoff += yinc;
					UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);

					if (sd1 >= 0)
						sprintf(buf, "BS %.6i", sides[sd1].sector - sectors);
					else
						sprintf(buf, "BS  NONE");
					
					yoff += yinc;
					UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);

					sprintf(buf, "mlf %.4X", lines[lnum[0]].ml_flags); yoff += yinc;
					UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);

					sprintf(buf, "lnf %.8X", lines[lnum[0]].lnflags2); yoff += yinc;
					UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);

					sprintf(buf, "mtf %.2X", lines[lnum[0]].mtflags); yoff += yinc;
					UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);

					sprintf(buf, "spf %.2X", lines[lnum[0]].spflags); yoff += yinc;
					UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);

					sprintf(buf, "tyf %.2X", lines[lnum[0]].tyflags); yoff += yinc;
					UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);

					sprintf(buf, "ovf %.2X", lines[lnum[0]].ovflags); yoff += yinc;
					UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);

					sprintf(buf, "SN0 % 6i", sd0); yoff += yinc;
					UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);

					idx = sides[sd0].toptexture;
					name = idx ? texArray[idx]->name : "-";

					sprintf(buf, "T %s", name); yoff += yinc;
					UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);

					idx = sides[sd0].midtexture;
					name = idx ? texArray[idx]->name : "-";

					sprintf(buf, "M %s", name); yoff += yinc;
					UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);

					idx = sides[sd0].bottomtexture;
					name = idx ? texArray[idx]->name : "-";

					sprintf(buf, "B %s", name); yoff += yinc;
					UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);

					sprintf(buf, "slfl %.4X", sides[sd0].lsiflags); yoff += yinc;
					UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);

					sprintf(buf, "SN1 % 6i", sd1); yoff += yinc;
					UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);

					if (sd1 > 0)
					{
						idx = sides[sd1].toptexture;
						name = idx ? texArray[idx]->name : "-";

						sprintf(buf, "T %s", name); yoff += yinc;
						UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);

						idx = sides[sd1].midtexture;
						name = idx ? texArray[idx]->name : "-";

						sprintf(buf, "M %s", name); yoff += yinc;
						UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);

						idx = sides[sd1].bottomtexture;
						name = idx ? texArray[idx]->name : "-";

						sprintf(buf, "B %s", name); yoff += yinc;
						UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);

						sprintf(buf, "slfl %.4X", sides[sd1].lsiflags); yoff += yinc;
						UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);
					}
				}

				if (am_marksec)
				{
					yoff += yinc;

					if (drawFloor && show_additional_info)
						UI_TextOutEx(marksecmsg[am_marksec_type], xoff, yoff, false, true, UI_COL(UIC_RED), 1);

					sprintf(buf, "SECT %6i", am_marksec - sectors); yoff += yinc;
					UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_CYAN), 1);
				}
			}
			else if (mapcheat == 8)
			{
				for (i = 0; i < 4; i++)
				{
					if (lnum[i] >= 0)
					{
						sprintf(buf, "L%5i", lnum[i]); yoff += yinc;
						UI_TextOutEx(buf, xoff, yoff, false, true,
							lines[lnum[i]].ml_flags & (ML_BLOCKING|ML_BLOCKMONSTERS) ? UI_COL(UIC_WARN) : UI_COL(UIC_TEXT),
							1);
					}
				}
			}
			else
			{
				sprintf(buf, "X % 6i", viewx>>16); yoff += yinc;
				UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);
				sprintf(buf, "Y % 6i", viewy>>16); yoff += yinc;
				UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);
				sprintf(buf, "Z % 6i", viewz>>16); yoff += yinc;
				UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);
				sprintf(buf, "A %6i", angle)    ; yoff += yinc;
				UI_TextOutEx(buf, xoff, yoff, false, true, UI_COL(UIC_TEXT), 1);

				if (cfg.counterCheat)
				{
					UI_TextOutEx(str[0], xoff, (yoff += yinc), false, true, UI_COL(UIC_TEXT), 1);
					UI_TextOutEx(str[1], xoff, (yoff += yinc), false, true, UI_COL(UIC_TEXT), 1);
					UI_TextOutEx(str[2], xoff, (yoff += yinc), false, true, UI_COL(UIC_TEXT), 1);
				}
			}
		}

		// v220-31 bug fixed in edit mode by checking state of ss
		if (automapHelp)
		{
			int		d = 0;
			uint	e = 20;
			RDBOOL	no_dead_things = !R3D_edit && (mapcheat == 2 || mapcheat == 5 || mapcheat == 8);
			char	inf[32];

			w = e*9, h += 2;
			d += h*14 + 4;
			if (!followplayer) d += h;
			if (maxlinesz > linesizes[0]) d += h;

			if (ss && !followplayer) d += h;
			if (ss && show_additional_info) d += (h*24);

			if (no_dead_things) d += h;	// accomodate TOGGLE DEAD option

			x = 18;
			y = (current_h/200) * 30;

			// draw rectangle
			UI_DrawRectSolid(10.f, (float)y, w, d, 4, 
				UI_COL(UIC_BRD_HI), UI_COL(UIC_BRD_MED), UI_COL(UIC_BRD_LOW), .6f);

			// draw options
			UI_TextOutEx("=  - ZOOM IN"			, x, (y += h), false, true, UI_COL(UIC_TEXT), 1);
			UI_TextOutEx("-  - ZOOM OUT"		, x, (y += h), false, true, UI_COL(UIC_TEXT), 1);
			UI_TextOutEx("0  - TOGGLE GOBIG"	, x, (y += h), false, true, UI_COL(UIC_TEXT), 1);
			UI_TextOutEx("F  - TOGGLE FOLLOW"	, x, (y += h), false, true, followplayer ? UI_COL(UIC_TEXT) : UI_COL(UIC_GREY), 1);
			UI_TextOutEx("G  - TOGGLE GRID"		, x, (y += h), false, true, UI_COL(UIC_TEXT), 1);
			UI_TextOutEx("H  - TOGGLE HELP"		, x, (y += h), false, true, UI_COL(UIC_TEXT), 1);

			if (mapcheat == 3)
			{
				UI_TextOutEx("J  - JUMP TO KEYS", x, (y += h), false, true, UI_COL(UIC_TEXT), 1);
			}
			else if (mapcheat == 7)
			{
				if (!followplayer)
					UI_TextOutEx("J  - PLYR TO CURSOR", x, (y += h), false, true, UI_COL(UIC_TEXT), 1);
			}
			else if (R3D_edit)
			{
				UI_TextOutEx("J  - JUMP TO SPOTS", x, (y += h), false, true, UI_COL(UIC_TEXT), 1);
			}
			else
			{
				UI_TextOutEx("J  - CENTRE MSPOTS", x, (y += h), false, true, UI_COL(UIC_TEXT), 1);
			}

			UI_TextOutEx("K  - CYCLE GRID SIZE"	, x, (y += h), false, true, UI_COL(UIC_TEXT), 1);

			if (maxlinesz > linesizes[0])
				UI_TextOutEx("L  - CYCLE LINE SIZE", x, (y += h), false, true, UI_COL(UIC_TEXT), 1);

			UI_TextOutEx("M  - SET MARKSPOT"	, x, (y += h), false, true, UI_COL(UIC_TEXT), 1);
			UI_TextOutEx("N  - CLEAR MARKSPOTS"	, x, (y += h), false, true, UI_COL(UIC_TEXT), 1);

			if (no_dead_things)
				UI_TextOutEx("O  - TOGGLE DEAD"		, x, (y += h), false, true, cfg.show_dead ? UI_COL(UIC_TEXT) : UI_COL(UIC_GREY), 1);

			UI_TextOutEx("R  - TOGGLE ROTATE"	, x, (y += h), false, true, cfg.automapRotate ? UI_COL(UIC_TEXT) : UI_COL(UIC_GREY), 1);
			UI_TextOutEx("U  - CYCLE ALPHA"		, x, (y += h), false, true, UI_COL(UIC_TEXT), 1);
			UI_TextOutEx("F1 - TOGGLE FLOOR"	, x, (y += h), false, true, UI_COL(UIC_TEXT), 1);

			if (!followplayer)
				UI_TextOutEx("ARROW KEYS - PAN"	, x, (y += h), false, true, UI_COL(UIC_TEXT), 1);

			if (ss && show_additional_info)
			{
				y += h;

				sprintf(inf, "SEC %9i", ss->sector - sectors);
				UI_TextOutEx(inf, x, (y += h), false, true, UI_COL(UIC_GREY), 1);
				sprintf(inf, "SECFH %7i", ss->sector->floorheight>>16);
				UI_TextOutEx(inf, x, (y += h), false, true, UI_COL(UIC_GREY), 1);
				sprintf(inf, "SECCH %7i", ss->sector->ceilingheight>>16);
				UI_TextOutEx(inf, x, (y += h), false, true, UI_COL(UIC_GREY), 1);
				sprintf(inf, "SFL  %.8X", ss->sector->specFlags);
				UI_TextOutEx(inf, x, (y += h), false, true, UI_COL(UIC_GREY), 1);
				sprintf(inf, "HSEC %8i", ss->sector->heightsec);
				UI_TextOutEx(inf, x, (y += h), false, true, UI_COL(UIC_GREY), 1);
				sprintf(inf, "SCH %9i", ss->sector->R3Dceilheight>>16);
				UI_TextOutEx(inf, x, (y += h), false, true, UI_COL(UIC_GREY), 1);
				sprintf(inf, "SCSEC %7i", ss->sector->R3Dceilsec);
				UI_TextOutEx(inf, x, (y += h), false, true, UI_COL(UIC_GREY), 1);
				sprintf(inf, "SFH %9i", ss->sector->R3Dfloorheight>>16);
				UI_TextOutEx(inf, x, (y += h), false, true, UI_COL(UIC_GREY), 1);
				sprintf(inf, "SFSEC %7i", ss->sector->R3Dfloorsec);
				UI_TextOutEx(inf, x, (y += h), false, true, UI_COL(UIC_GREY), 1);
				sprintf(inf, "SLSEC %7i", ss->sector->R3Dlowsec);
				UI_TextOutEx(inf, x, (y += h), false, true, UI_COL(UIC_GREY), 1);
				sprintf(inf, "SWH %9i", ss->sector->withheight>>16);
				UI_TextOutEx(inf, x, (y += h), false, true, UI_COL(UIC_GREY), 1);

				y += h;

				sprintf(inf, "SSEC %8i vg %i", ss - subsectors, ss->viewgroup);
				UI_TextOutEx(inf, x, (y += h), false, true, UI_COL(UIC_GREY), 1);
				sprintf(inf, "SSFL %.8X", ss->ssflags);
				UI_TextOutEx(inf, x, (y += h), false, true, UI_COL(UIC_GREY), 1);
				sprintf(inf, "CSEC %8i", ss->ceilsec >= 0 ? ss->ceilsec : -1);
				UI_TextOutEx(inf, x, (y += h), false, true, UI_COL(UIC_GREY), 1);
				sprintf(inf, "LCSEC %7i", ss->lcsec >= 0 ? ss->lcsec : -1);
				UI_TextOutEx(inf, x, (y += h), false, true, UI_COL(UIC_GREY), 1);
				sprintf(inf, "FSEC %8i", ss->floorsec >= 0 ? ss->floorsec : -1);
				UI_TextOutEx(inf, x, (y += h), false, true, UI_COL(UIC_GREY), 1);
				sprintf(inf, "LFSEC %7i", ss->lfsec >= 0 ? ss->lfsec : -1);
				UI_TextOutEx(inf, x, (y += h), false, true, UI_COL(UIC_GREY), 1);
				sprintf(inf, "LTSEC %7i", ss->sslightsec >= 0 ? ss->sslightsec : -1);
				UI_TextOutEx(inf, x, (y += h), false, true, UI_COL(UIC_GREY), 1);
				sprintf(inf, "3DFLRS %6i", ss->flats3DArray ? *(int*)ss->flats3DArray : 0);
				UI_TextOutEx(inf, x, (y += h), false, true, UI_COL(UIC_GREY), 1);

				y += h;

				sprintf(inf, "FLTS %7i", ss->sector->floorlightsec);
				UI_TextOutEx(inf, x, (y += h), false, true, UI_COL(UIC_GREY), 1);
				sprintf(inf, "CLTS %7i", ss->sector->ceilinglightsec);
				UI_TextOutEx(inf, x, (y += h), false, true, UI_COL(UIC_GREY), 1);
				if (followplayer)
				{
					sprintf(inf, "PLRL %7i", ppt.viewer_lightlevel);
					UI_TextOutEx(inf, x, (y += h), false, true, UI_COL(UIC_GREY), 1);
				}
			}
		}

		// Restore original matrix.
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
	}
}

static void am_ResetMarksec(RDBOOL active)
{
	if (active)
	{
		marksecnum = str_secn[0] ? atoi(str_secn) : -1;
	}
	else
	{
		marksecnum = -1;
		am_marksec_type = 0;
	}

	am_marksec = NULL;
	set_sector = false;
	str_secn[0] = 0;
	set_secnpos = 0;
}

// v220-25 clip ssecs not in the viewing area.
//
// Note: When the map is rotated this algo is not as efficient as it could be
//       but it is a lot better than just allowing all subsecs to be drawn,
//       especially when zoomed in. The alternative is to rotate all the ssec
//       box limits to establish whether they are contained in the rotated
//       view but, mother, is it worth it?
//
static void am_SetLimits()
{
	float	fh = mp.m_h;
	float	fw = mp.m_w;
	int		i;

	if (fw >= mb.max_w * .8f && fh >= mb.max_h * .8f)
	{
		// don't bother with checking if ssecs are in window
		memset(amlm.ammarkssec, 1, numsubsectors);
		amlm.amlinenum = numlines;
		amlm.amssecnum = numsubsectors;

		for (i = 0; i < numlines; i++) amlm.amlinelist[i] = i;
		for (i = 0; i < numsubsectors; i++) amlm.amsseclist[i] = i;
		return;
	}

	subsector_t	*ss = subsectors;
	line_t		*ld;
	seg_t		*sg;
	float		px = followplayer ? fvx : mp.m_x + mp.m_w/2.f;
	float		py = followplayer ? fvz : mp.m_y + mp.m_h/2.f;

	memset(amlm.ammarkssec, 0, numsubsectors);
	amlm.amlinenum = 0;
	amlm.amssecnum = 0;

	if (cfg.automapRotate && followplayer)	// limits need to be adjusted
	{
		// Between steps of 45deg an increment is needed
		// based on the current screen's diagonal value
		// maxing out at the 45 deg points.

		float an = (float)(fabs(sin(vang*DPI/90)));

		an *= 1.25f;	// fiddle factor

		fh += (mlen - mp.m_h) * an;
		fw += (mlen - mp.m_w) * an;

		// At 90deg and 270 deg fh and fw will be transposed
		// At  0deg and 180 deg fh and fw will remain unchanged
		// This gives the same results as when rotation is off.

		an = (fw - fh) * (float)(fabs(sin(vang*DPI/180)));

		fh += an;
		fw -= an;
	}

	fh *= 0.5f;
	fw *= 0.5f;

	validcount++;

	for (i = 0; i < numsubsectors; i++, ss++)
	{
		// $limitcheck
		if (FIX2FLT(ss->ssbox[BOXRIGHT]) < px - fw) continue;
		if (FIX2FLT(ss->ssbox[BOXLEFT ]) > px + fw) continue;
		if (FIX2FLT(ss->ssbox[BOXTOP  ]) < py - fh) continue;
		if (FIX2FLT(ss->ssbox[BOXBASE ]) > py + fh) continue;

		amlm.ammarkssec[i] = true;
		amlm.amsseclist[amlm.amssecnum++] = i;
		sg = &segs[ss->firstline];

		for (int k = 0; k < ss->linecount; k++, sg++)
		{
			if ( (ld = sg->linedef) && ld->validcount != validcount)
			{
				ld->validcount = validcount;
				amlm.amlinelist[amlm.amlinenum++] = ld - lines;
			}
		}
	}
}

// public functions ---------------------------------------------

void AM_Init()
{
	sector_t	*sc;
	int			i, k, s, t, v;
	byte		type;

	plr = &mbf_players[0];

	// set limits
	amlm.amlinelist = (int *)W_Malloc(numlines * sizeof(int), LEVELID, 50);
	amlm.amsseclist = (int *)W_Malloc(numsubsectors * sizeof(int ), LEVELID, 51);
	amlm.ammarkssec = (byte*)W_Malloc(numsubsectors * sizeof(byte), LEVELID, 52);
	amlm.amlinenum = 0;
	amlm.amssecnum = 0;

	for (i = 0; i < KEYSTUFFNUM; i++) keystuff[i].keylnum = -2;

	markspotnum = 0;
	markkeynum = 0;
	
	oldcheat = 0;
	
	show_additional_info = ArgExists("-amaddinf");

	if (sz_ssamap < numsubsectors)
	{
		sz_ssamap = numsubsectors;
		ssamap = (byte*)W_Realloc(ssamap, sz_ssamap, STATICID, 0);
	}

	if (sz_lnamap < numlines)
	{
		sz_lnamap = numlines;
		lnamap = (byte*)W_Realloc(lnamap, sz_lnamap, STATICID, 0);
	}

	memset(ssamap, 0, sz_ssamap);
	memset(lnamap, 0, sz_lnamap);

	solidalpha = cfg.automapAlpha < 0.f;

	if (cfg.linesz < 1.f) cfg.linesz = 1.f;
	if (cfg.linesz > linesizes[1] || cfg.linesz > 3.f) cfg.linesz = 3.f;

	maxlinesz = (linesizes[1] >= 3.f) ? 3.f : linesizes[1];

	// v220-22	Setup automap door ids
	door_array = (byte*)W_Calloc(numlines, LEVELID, 0);

	for (i = 0; i < numlines; i++)
	{
		if ( (v = lines[i].special) )
		{
			type = 0;

			if (    v == 99
				|| (v >= 26  && v <= 28)
				|| (v >= 32  && v <= 34)
				|| (v >= 133 && v <= 137) )
			{
				if      (v == 26 || v == 32 || v ==  99 || v == 133) type = DOORKEYBLUE;
				else if (v == 28 || v == 33 || v == 134 || v == 135) type = DOORKEYRED;
				else if (v == 27 || v == 34 || v == 136 || v == 137) type = DOORKEYYELLOW;
			}
			else if (v >= GenLockedBase && v < GenDoorBase)
			{
				switch((v & LockedKey) >> LockedKeyShift)
				{
				case RCard: case RSkull: type = DOORKEYRED   ; break;
				case BCard: case BSkull: type = DOORKEYBLUE  ; break;
				case YCard: case YSkull: type = DOORKEYYELLOW; break;
				case AnyKey: break;
				case AllKeys: break;
				}
			}

			if (!type) continue;

			if  ( (t = lines[i].tag) )
			{
				for (s = -1; (s = P_FindSectorFromTag(t, s)) >= 0;)
				{
					sc = &sectors[s];

					for (k = 0; k < sc->linecount; k++)
					{
						if (   sc->lines[k]->backsector
							&& !(sc->lines[k]->spflags & SPF_CONTROLLINE))
						{
							door_array[ (sc->lines[k] - lines) ] = type;
						}
					}
				}
			}
			else
			{
				door_array[i] = type;
			}
		}
	}
}

void AM_Drawer (void)
{
    if (!automap_active) return;

	GLboolean tx2D;

	keythingcount = 0;
	map8mobj = NULL;

	// Setup.
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	// Black background.
	GL_SetNoTexture();
	if ( (tx2D = glIsEnabled(GL_TEXTURE_2D)) ) glDisable(GL_TEXTURE_2D);

	if (solidalpha)
		GL_DrawRect(0, 0, (float)SCREENWIDTH, (float)SCREENHEIGHT, 123.f/255.f, 99.f/255.f, 79.f/255.f, 1.f); 
	else
		GL_DrawRect(0, 0, (float)SCREENWIDTH, (float)SCREENHEIGHT, 0, 0, 0, 
			cfg.automapAlpha);

	// Setup state.
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, current_w, current_h, 0, -1, 1);

	if (cfg.automapRotate && followplayer) // Rotate map?
	{	
		float tx = f_w/2.f, ty = f_h/2.f;

		glTranslatef(tx, ty, 0.0f);
		glRotatef(plr->plr->clAngle/(float)ANGMAX * 360 - 90, 0.0f, 0.0f, 1.0f);
		glTranslatef(-tx, -ty, 0.0f);
	}

	if (drawFloor)
	{
		glEnable(GL_TEXTURE_2D);
		am_DrawFloorFlats();
		glDisable(GL_TEXTURE_2D);
	}

    if (grid)
	{
		glBegin(GL_LINES);
		am_DrawGrid();
		glEnd();
	}

	glLineWidth(cfg.linesz);

	if (mapcheat == 7)
	{
		am_DrawSubsectors();
	}
	else
	{
		glBegin(GL_LINES);
		am_DrawWalls();
		glEnd();
	}


	if (cfg.linesz > 1.f) glLineWidth(1.f);

	glColor3f(1.0f, 1.0f, 1.0f);
	
	if (mapcheat != 7)
	{
		glEnable(GL_TEXTURE_2D);
		am_DrawMarks();
		glDisable(GL_TEXTURE_2D);
	}
	
	glBegin(GL_LINES);

    if (mapcheat == 2 || mapcheat == 3 || mapcheat == 5 || mapcheat == 8)
		am_DrawThings(mapcheat == 5 || mapcheat == 8);

	RDBOOL	marktype = mapcheat >= 7 && (!followplayer || marksecnum >= 0);

	if (marktype)
	{
		if (marksecnum >= 0 && marksecnum < numsectors)
		{
			int c, i, n = -1;

			am_marksec = &sectors[marksecnum];

			for (c = i = 0; i < numsubsectors; i++)
			{
				if (subsectors[i].sector == am_marksec)
				{
					n = i;

					if ( !(subsectors[i].ssflags & SSF_CTRLONLY) )
					{
						mp.m_x = subsectors[i].midpoint.x - mp.m_w/2.f;
						mp.m_y = subsectors[i].midpoint.y - mp.m_h/2.f;
						c = followplayer; n = -1; break;
					}
				}
			}

			if (n >= 0)	// Fallback if all ssec's found are control types
			{			// (which, until now, have all been ignored)
				mp.m_x = subsectors[n].midpoint.x - mp.m_w/2.f;
				mp.m_y = subsectors[n].midpoint.y - mp.m_h/2.f;
				c = followplayer;
			}

			if (c)
			{
				followplayer = false;
				if (!R3D_edit && !automapHelp)	// help screen shows state
					P_SetMessage(plr, AMSTR_FOLLOWOFF);
			}
		}
		
		marksecnum = -1;
	}

	float px = mp.m_x + mp.m_w/2.f;
	float py = mp.m_y + mp.m_h/2.f;

	glColor3f(1.0f, 1.0f, 1.0f);

	// draw standard player line character
	am_DrawLineCharacter(
		mapcheat ? cheat_player_arrow : player_arrow,
		mapcheat ? NUMCHEATPLYRLINES  : NUMPLYRLINES,
		0, plr->plr->clAngle,
		marktype ? LINEMAGENTA : LINEWHITE,
		followplayer ? px : fvx,
		followplayer ? py : fvz);

	if (marktype)
	{
		// Draw cross
		am_DrawLineCharacter(am_cross, NUMCROSSLINES,
			scale_ftom, 0, LINEWHITE, px, py);
	}

	glEnd();

	glEnable(GL_TEXTURE_2D);
    am_DrawKeys();

	// Restore.
	if (!tx2D) glDisable(GL_TEXTURE_2D);
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	am_HelpDrawer();
}

// v220-0 added. v220-9 revised
void Airbar_Drawer(void)
{
	char	buf[160];
	int		airbar = ( !(thinker_uwater.tuwflags & TUWF_RESET) );

	if (automap_active) return;

	if ( !(vidShowFPS || airbar) ) return;

	FR_SetFont(0);

	// Go into screen projection mode.
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, current_w, current_h, 0, -1, 1);
	
	if (airbar && !ge_inf.devmap)
	{
		static char *air[] = { "Air   ", "Air   GOD MODE", "Air   INVULN MODE" };
		mbf_player_t *player = &mbf_players[0];
		int x, w, h, idx = 0;
		int t = thinker_uwater.breath_tics/TICRATE;

		w = 194;

		if (t >= UWMAXTIME)
		{
			t = UWMAXTIME;

			if (mbf_players[0].cheats & CF_GODMODE)
			{
				idx = 1;
			}
			else if (mbf_players[0].powers[pw_invulnerability])
			{
				idx = 2;
			}
		}

		h = FR_TextHeight(air[idx]) + 16;
		x = current_w - w - 6;

		UI_DrawRectEx((float)x, 4.0f, w, h, 4, 
			UI_COL(UIC_BRD_HI), UI_COL(UIC_BRD_MED), UI_COL(UIC_BRD_LOW), .5f);
		UI_Gradient(x + 4, 8, w - 8, h - 8, 
			UI_COL(UIC_BG_MEDIUM), UI_COL(UIC_BG_LIGHT), .5f, .5f);
		UI_TextOutEx(air[idx], x + 8, h/2 + 4, false, true, UI_COL(UIC_TEXT), 1);

		if (!idx)
		{
			x += FR_TextWidth(air[idx]);

			UI_DrawRect(x, h/2 + 2, 120, 8, 4, 
				UI_COL(UIC_SHADOW), UI_COL(UIC_SHADOW), UI_COL(UIC_SHADOW));

			if ( (t = UWMAXTIME - t) > 0)
			{
				UI_DrawRect(x, h/2 + 2, t*4, 8, 4, 
					UI_COL(UIC_TEXT), UI_COL(UIC_BG_DARK), UI_COL(UIC_BG_LIGHT));
			}
		}
	}
	else
	{
		int ch, x, y = 4, w, h;

		if (ma_inf.activethinkers > 4000)
		{
			w = 240;
			sprintf(buf, "%4i FPS %5i E %5i M", stats.fpsnum,
				stats.global_mobj_count, stats.missile_mobj_count);
		}
		else
		{
			w = 80;
			sprintf(buf, "%i FPS", stats.fpsnum);
		}

		ch = FR_TextHeight(buf);

		h = ch + 16;
		x = current_w - w - 10;

		UI_DrawRectEx((float)x, (float)y, w, h, 4, 
			UI_COL(UIC_BRD_HI), UI_COL(UIC_BRD_MED), UI_COL(UIC_BRD_LOW), .5f);
		UI_Gradient(x + 4, y + 4, w - 8, h - 8, 
			UI_COL(UIC_BG_MEDIUM), UI_COL(UIC_BG_LIGHT), .5f, .5f);

		UI_TextOutEx(buf, x + 8, y + h/2, false, true, UI_COL(UIC_TEXT), 1);

		if (ge_inf.devmap)
		{
			int i = 0;
			int col = UIC_TEXT;

			y += 40;
			w = cfg.amDevboxWidth;
			x = current_w - w - 10;
			h = (ch + 2)*(TH_NUMTYPES + 1) + 16;

			UI_DrawRectEx((float)x, (float)y, w, h, 4, 
				UI_COL(UIC_BRD_HI), UI_COL(UIC_BRD_MED), UI_COL(UIC_BRD_LOW), .5f);
			UI_Gradient(x + 4, y + 4, w - 8, h - 8, 
				UI_COL(UIC_BG_MEDIUM), UI_COL(UIC_BG_LIGHT), .5f, .8f);

			y += 12;

			while (1)
			{
				sprintf(buf, "%s %6i", th_name[i], th_items[i]);
				UI_TextOutEx(buf, x + 8, y, false, true, UI_COL(col), 1);

				if (++i >= TH_NUMTYPES) break;

				if (i == 11) { col = UIC_LIST; y += (ch + 2); }

				y += (ch + 2);
			}
		}
	}

	// Restore original matrix.
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
}

// Activate or deactivate automap.
int CCmdAutoMap(int argc, char **argv)
{
	// v220-3 ignore if altdown is true
	if (gamestate != GS_LEVEL || altDown) return true;

 	am_ResetMarksec(false);

	if ( (automap_active = !automap_active) ) am_Start();

	return true;
}

//
// Handle events (user inputs) in automap mode
//
RDBOOL AM_Responder(event_t *ev)
{
    int rc = false;
	
	if (check_plr_in_bounds)
	{
		if (!followplayer)
		{
			float fx = (fvx - mp.m_x) * scale_mtof;
			float fy = f_h - (fvz - mp.m_y) * scale_mtof;

			followplayer = (fx < 0 || fx > f_w || fy < 0 || fy > f_h);

			if (followplayer) P_SetMessage(plr, "PLAYER OUT OF BOUNDS: FOLLOW MODE RESET TO ON");
		}
	
		check_plr_in_bounds = false;
	}

	if (ev->type == ev_keydown)
    {
		rc = true;

		if (set_sector)
		{
			switch(ev->data1)
			{
			case '0': case '1': case '2': case '3': case '4':
			case '5': case '6': case '7': case '8': case '9':
				if (set_secnpos < 6)
				{
					str_secn[set_secnpos++] = ev->data1;
					str_secn[set_secnpos] = 0;
				}
				break;
			case DDKEY_ENTER:
				am_ResetMarksec(true);
				break;
			case DDKEY_BACKSPACE:
				if (set_secnpos) str_secn[--set_secnpos] = 0; 
				break;
			}

			return true;
		}

		switch(ev->data1)
		{
		case DDKEY_ENTER:
			if (mapcheat == 7 && !R3D_edit) set_sector = true; else rc = false;
			break;
		case DDKEY_F2:
			if (am_marksec && drawFloor && show_additional_info)
			{
				am_marksec_type++;
				if (am_marksec_type > 5) am_marksec_type = 0;
			}
			else rc = false;
			break;
		case AM_ALPHAKEY:
			cfg.automapAlpha += (shiftdown ? -.1f : .1f);
			if (cfg.automapAlpha >  1.f) cfg.automapAlpha = -.1.f;
			if (cfg.automapAlpha < -.1f) cfg.automapAlpha = 1.f;
			solidalpha = cfg.automapAlpha < 0.f;
			break;
		case AM_CLEARMARKKEY:
			if (mapcheat != 7)
			{
				am_ClearMarks();
				P_SetMessage(plr, AMSTR_MARKSCLEARED);
			}
			break;
		case AM_FLOORKEY:
			drawFloor = !drawFloor;
			break;
		case AM_FOLLOWKEY:
			followplayer = !followplayer;
			if (!R3D_edit && !automapHelp)	// help screen shows state
				P_SetMessage(plr, followplayer ? AMSTR_FOLLOWON : AMSTR_FOLLOWOFF);
			break;
		case AM_GOBIGKEY:
			if ( (bigstate = !bigstate) )
			{
				memcpy(&oldmp, &mp, sizeof(oldmp));	// Save current position
				am_ActivateNewScale(min_scale_mtof);
			}
			else	// Restore previous values
			{
				memcpy(&mp, &oldmp, sizeof(mp));	// Restore values

				if (followplayer)
				{
					mp.m_x = fvx - mp.m_w/2.f;
					mp.m_y = fvz - mp.m_h/2.f;
				}

				mlen = am_GetMLEN();

				// Reset the scaling multipliers
				scale_mtof = f_w / mp.m_w;
				scale_ftom = 1.0f / scale_mtof;
			}
			break;
		case AM_GRIDKEY:
			if ( (grid = !grid) )
			{
				itoa(gridsize, num, 10);
				sprintf(buffer, "%s SIZE %s", AMSTR_GRIDON, num);
				P_SetMessage(plr, buffer); 
			}
			else
			{
				P_SetMessage(plr, AMSTR_GRIDOFF);
			}
			break;
		case AM_GRIDSIZEKEY:
			if (shiftdown)
			{
				if (!grid) grid = true; else gridsize *= 2;	
				if (gridsize > GRIDMAX) gridsize = GRIDMIN;
			}
			else
			{
				if (!grid) grid = true; else gridsize /= 2;	
				if (gridsize < GRIDMIN) gridsize = GRIDMAX;
			}
			itoa(gridsize, num, 10);
			sprintf(buffer, "GRID SIZE %s", num);
			P_SetMessage(plr, buffer);
			break;
		case AM_HELPKEY:
			automapHelp = !automapHelp;
			break;
		case AM_JUMP:
			am_GoToMarks();
			break;
		case AM_LINESIZEKEY:
			if (maxlinesz > linesizes[0])
			{
				cfg.linesz += 1.f;

				if (cfg.linesz > maxlinesz) cfg.linesz = 1.f;

				sprintf(buffer, "LINE THICKNESS SET = %d", (int)cfg.linesz);
			}
			break;
		case AM_MARKKEY:
			if (mapcheat != 7 && followplayer)
			{
				sprintf(buffer, "%s %d", AMSTR_MARKEDSPOT, markpointnum);
				P_SetMessage(plr, buffer);
				am_AddMark();
			}
			else
			{
				if (mapcheat == 7)
					P_SetMessage(plr, "CHEAT MODE 7: MARK IGNORED");
				else
					P_SetMessage(plr, "FOLLOW MODE IS OFF: MARK IGNORED");
			}
			break;
		case AM_DEADKEY:
			cfg.show_dead = !cfg.show_dead;
			break;
		case AM_PANDOWNKEY: // pan down
			if (!followplayer) m_paninc.y = shiftdown ? -F_PANINCSLOW * scale_ftom : -F_PANINC * scale_ftom;
			else rc = false;
			break;
		case AM_PANLEFTKEY: // pan left
			if (!followplayer) m_paninc.x = shiftdown ? -F_PANINCSLOW * scale_ftom : -F_PANINC * scale_ftom;
			else rc = false;
			break;
		case AM_PANRIGHTKEY: // pan right
			if (!followplayer) m_paninc.x = shiftdown ? F_PANINCSLOW * scale_ftom : F_PANINC * scale_ftom;
			else rc = false;
			break;
		case AM_PANUPKEY: // pan up
			if (!followplayer) m_paninc.y = shiftdown ? F_PANINCSLOW * scale_ftom : F_PANINC * scale_ftom;
			else rc = false;
			break;
		case AM_ROTATEKEY:
			cfg.automapRotate = !cfg.automapRotate;

			if (cfg.automapRotate && !followplayer)
				followplayer = true;
			if (!automapHelp)
					P_SetMessage(plr, cfg.automapRotate ? "ROTATE ON" : "ROTATE OFF");

			break;
		case AM_ZOOMOUTKEY: // zoom out
			zoommul_factor = M_ZOOMOUT;
			break;
		case AM_ZOOMINKEY: // zoom in
			zoommul_factor = M_ZOOMIN;
			break;
		default:
			rc = false;
		}

		if (cht_CheckCheat(&cheat_amap, (char) ev->data1))
		{
			int *p = &mbf_players[0].powers[pw_allmap];

			if (mapcheat > 2 || ++(*p) > 2)
			{
				*p = 0;
				check_plr_in_bounds = true;
			}
			mapcheat = *p;
			rc = false;
		}
    }
    else if (ev->type == ev_keyup)
    {
		zoommul_factor = 1.0f;

		switch (ev->data1)
		{
		case AM_PANRIGHTKEY:
			if (!followplayer) m_paninc.x = 0;
			break;
		case AM_PANLEFTKEY:
			if (!followplayer) m_paninc.x = 0;
			break;
		case AM_PANUPKEY:
			if (!followplayer) m_paninc.y = 0;
			break;
		case AM_PANDOWNKEY:
			if (!followplayer) m_paninc.y = 0;
			break;
		case AM_ZOOMOUTKEY:
		case AM_ZOOMINKEY:
			zoommul_factor = 1.0f;
			break;
		default:
			rc = false;
		}
    }
	else if(ev->type == ev_keyrepeat)
	{
		return true;
	}
	
    return rc;
}

void AM_Ticker() // Updates on Game Tick
{
	if (!automap_active) return;
	
	if (set_sector && mapcheat != 7)
		am_ResetMarksec(false);
	
	if (mapcheat != oldcheat)
	{
		oldcheat = mapcheat;
		am_ResetMarksec(false);
	}
	
	if (followplayer)
	{
		mp.m_x = fvx - mp.m_w/2.f;
		mp.m_y = fvz - mp.m_h/2.f;
	}
	else if (cfg.automapRotate)
	{
		cfg.automapRotate = false;
	}

	// Change the zoom if necessary
	if (zoommul_factor != 1.0f)
	{
		// Change the scaling multipliers
		current_scale_mtof = scale_mtof *= zoommul_factor;
		am_ActivateNewScale(scale_mtof);
	}
	
	// Change x,y location
	if (m_paninc.x || m_paninc.y) am_ChangeWindowLoc();
	
	am_SetLimits();	// added v220-15
}

// v220-17	Functions for storing and resetting markpoints
void AM_SetMarkPoints(byte *mp, uint nm)
{
	memcpy(markpoints, mp, sizeof(markpoints));
	markpointnum = (int)nm;
	ma_inf.mapchange = 2;	// prevents clearing on load
}

byte *AM_GetMarkPoints(int *sz, int *nm)
{
	byte *mp = (byte*)malloc(sizeof(markpoints));
	*sz = sizeof(markpoints);
	if (nm)
	{
		*nm = markpointnum;
		memcpy(mp, markpoints, sizeof(markpoints));
	}
	return mp;
}
