//-----------------------------------------------------------------------------
//
// Copyright (C) 1998 - 2005 id Software, Inc., Chi Hoang, Lee Killough,
//  Jim Flynn, Rand Phares, Ty Halderman, Jaakko Kernen
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: dp_mobj.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  BOOM and R3D extensions added, reformatted, moved to main engine.
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
// Revision 2.2.0 build 2 17/Jun/2008 gmj (b220-2)
//  Mod for sliding doors
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj mobjcarry_t etc. removed
// Revision 2.2.0 build 11 (v220-11) 08/Aug/2011 gmj
// Revision 2.2.0 build 12 (v220-12) 01/Oct/2011 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 14 (v220-14) 01/May/2012 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 18 (v220-18) 05/Jan/2013 gmj
// Revision 2.2.0 build 29 (v220-29) 22/Nov/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll code and BOOM v2.01
//
//
// DESCRIPTION:
//	 Moving object handling. Spawn functions.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <string.h>
#include <math.h>

#include "doom/mbf_data.h"
#include "doom/p_local.h"
#include "doom/p_start.h"
#include "doom/tables.h"

#include "AcFnLink.h"
#include "p_maputil.h"
#include "net_main.h"
#include "m_misc.h"
#include "m_bams.h"
#include "r_data.h"
#include "r_main.h"
#include "r_things.h"
#include "r_model.h"
#include "r_world.h"
#include "s_main.h"
#include "con_console.h"

#include "r3d_edit.h"
#include "r3d_sl3D.h"
#include "r3d_utils.h"
#include "r3d_MAINspec.h"
#include "r3d_slopes.h"

		
// MACROS ------------------------------------------------------------------

#define VANISHTICS	(2*TICSPERSEC) // $vanish

enum
{
	ST_WATER,
	ST_LAVA,
	ST_SLIME,
	ST_BLOOD,
	ST_NUKE,
	ST_TYPENUM
} splashtime_t;

#define ST_SOUND	6

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void	P_RegisterPlayerStart(mapthing_t *mthing, int mapnum);
void	P_ApplyTorque(mobj_t *mo);	// killough $dropoff_fix

void	P_SRVOAngleTicker(mobj_t *mo);
void	P_UpdateHealthBits(mobj_t *mobj);
void	P_UpdateAlpha(mobj_t *mo);
RDBOOL	P_SoundSight(mobj_t *t1, mobj_t *t2);

void	EnactAction(int action, mobj_t *mobj);
short	identifySpecial(line_t *line);
void	subfn_ResetHeight(mobj_t *iter);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

fixed_t	P_GetFriction(mobj_t *mo, fixed_t *frictionfactor);

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// For height adjustment - any set will be removed following analysis

extern byte		*spawnArray;
extern byte		*linkSectors;
extern int		corpsetime;
extern mobj_t	*teleport_flash; // v220-9


// PUBLIC DATA DEFINITIONS -------------------------------------------------

int		respawnmonsters;
int		playerdrag;		// player deceleration when stopping

// For use in waterlevel
int		waterlevel;
int		onsurface;

// Ambients, effects and map music
int			musChange;
int			music_array[MAX_MUS_ENTRIES];
reverb_t	reverbThing[MAX_REVERBTHING];
stats_t		stats;
mobj_t		mobjarray[MOBJ_ENTRIES];
mpos_t		mpos;
tpfake_t	tpfake[100];

int	ambientfound;
int ambientcount[64];

// v220-10 dehacked ambients
// v220-13 changed to 1 and 1 from .4 and 2
//         Entries in the bin file R3D_modify changed accordingly
float		dehamb_vol = 1.f, dehamb_attn = 1.f;

gamedata_t	gdt;
fixed_t		mapgravity;

// v220-18
dmarray_t	dmarray[DM_MAXCOUNT];
int			dmcount;

// v220-35
linetrig_t	ltarray[LT_MAXCOUNT];
int			forceDynLights;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static mobj_t*	P_SpawnSplashMobj( fixed_t x, fixed_t y, fixed_t z, mobjtype_t type );

static int	maxnumstates;
static int	norandomrotate;
static int	splashtime[ST_TYPENUM];

// 65536 (MOBJ_ENTRIES) bits telling which IDs are in use
static int		idtable[2048];
static thid_t	iddealer = 0;

fixed_t	stepfriction;

// CODE --------------------------------------------------------------------

// v220-11
static void CheckDesync(mobj_t *mo)
{
	static int seed = 0;

	if (   R3D_ScriptedMap
		&& mo->state
		&& mo->state->stflags & ST_DESYNC)
	{
		state_t	*st = mo->state;
		int		ms = mo->state - states, n = 0, sn;

		while (1)
		{
			if ( (sn = st->nextstate) <= 0) return;
			if (n++ > 12) return;
			if (sn == ms) break;
			st = &states[sn];
		}

		if (n < 2) return;

		seed += 3;
		sn = ms + (seed & 0xff) % n;
		mo->state = &states[sn];

		if (ge_inf.devmap)
			pr_OFinsert("Desynced N%i T%i\n", n, MAPNUM(mo));
	}
}

mobj_t *getArrayMobj(int id)
{
	if (id > 0 && id < MOBJ_ENTRIES)
		return &mobjarray[id];

	return NULL;
}

// b220 COD support
void setCODflag(int id)
{
	if (id > 0 && id < MOBJ_ENTRIES)
		mobjarray[id].flags4 |= MF4_CODENDTYPE;
}

//==========================================================================
// P_ClearThingSRVO
//	The thing's timer has run out, which means the thing has completed its
//	step. Or there has been a teleport.
//==========================================================================
void P_ClearThingSRVO(mobj_t *mo, RDBOOL clrmpos)
{
	memset(mo->srvo, 0, sizeof(mo->srvo));
}

//==========================================================================
// P_ClearMobjIDs
//==========================================================================
void P_ClearMobjIDs(void)
{
	memset(idtable, 0, sizeof(idtable));
	iddealer = 0;
	idtable[0] |= 1;	// ID zero is always "used" (it's not a valid ID).
}

//==========================================================================
// P_IsUsedMobjID
//  Note: sending an id of 0 is invalid (as the first entry is never used)
//  The onus is on the sender to ensure the id is > 0
//==========================================================================
RDBOOL P_IsUsedMobjID(thid_t id)
{
	return (idtable[id >> 5] & (1 << (id & 0x1f))) != 0;	
}

//==========================================================================
// P_SetMobjID	id must not = 0
//  See P_ClearMobjID(). If the iddealer wraps to 0 then P_IsUsedMobjID()
//  should always return true. Note that no linked list is used to get an
//  unused mobj ASAP since it may still be being referenced by a particle
//  generaor (e.g. smoke rising from a dead mobj). Since the iddealer is
//  always incremented then it would have to wrap round before encountering
//  the mobj entry. As such entries are only cleared when next needed. Also
//  see Lee Killough's update P_RemoveThinkerDelayed();
//==========================================================================
void P_SetMobjID(thid_t id, RDBOOL set)
{
	if (id)	// R3D Trap for idtable[0] bit 1 (must not be cleared)
	{
		int c = id >> 5, bit = 1 << (id & 0x1f);

		if(set)
			idtable[c] |= bit;
		else
			idtable[c] &= ~bit;
	}
#ifdef _DEBUG	// R3D - left here although, as expected, never happened
	else
		Con_Error("P_SetMobjID: sent id of 0");
#endif
}

//==========================================================================
// P_SetMpos   v220-5 Set the mpos value
//==========================================================================
void P_SetMpos(mobj_t *mo)
{
	memcpy(&mpos.mp[mo - mobjarray].ox, &mo->x, sizeof(fixed_t) * 3);
}

//==========================================================================
// P_NewMobjID
//==========================================================================
thid_t P_NewMobjID(void)
{
	thid_t check = iddealer;

	// Increment the ID dealer until a free ID is found.
	while(P_IsUsedMobjID( (++iddealer) ) && check != iddealer);

	if (check == iddealer) Con_ErrorD("Mobjs exceed 65534");

	// Mark this ID as used.
	P_SetMobjID(iddealer, true);

	return iddealer;
}

static mobj_t *GetFreeMobj()
{
	thid_t	freeid = P_NewMobjID();
	mobj_t	*nextfree = &mobjarray[freeid];

	memset(nextfree, 0, sizeof(mobj_t));

	nextfree->thinker.id = freeid;
	nextfree->thinker.function = (void (__cdecl *)(void))P_MobjThinker;

	ma_inf.mobjsCreated++;	// For stats on exit

	return nextfree;
}

int P_CountMobjThinkers()
{
	int c, i;

	for (c = i = 0; i < 65535; i++)
	{
		if (mobjarray[i].thinker.function == (void (__cdecl *)(void))P_MobjThinker) c++;
	}

	return c;
}

void ResetArrays(RDBOOL initblockrings)
{
	int i;

	if (initblockrings)
	{
		// re-initialise linkmobj blockrings.
		// Memory is allocated and these are first done in P_CreateBlockmap
		for (i = 0; i < bmapwidth * bmapheight; i++)
			blockrings[i].next = blockrings[i].prev = (mobj_t*) &blockrings[i];
	}


	P_InitThinkers();	// resets all bar the particle thinkers. THINKIDs freed here
	R_InitLinks();		// initialize node piles and line rings. NODESIDs freed here
}

// Used only with savegames
mobj_t *NewReferencedMobj(thid_t idnum)
{
	if (idnum && !P_IsUsedMobjID(idnum))
	{
		mobj_t *refmobj = &mobjarray[idnum];

		P_SetMobjID(idnum, true);

		if (idnum > iddealer) iddealer = idnum;

		memset(refmobj, 0, sizeof(mobj_t));
		refmobj->thinker.id = idnum;
		refmobj->thinker.function = (void (__cdecl *)(void))P_MobjThinker;
		ma_inf.mobjsCreated++;	// For stats on exit
		return refmobj;
	}
	return NULL;
}

mobj_t *SetReferencedMobj(thid_t idnum)
{
	if (idnum && P_IsUsedMobjID(idnum))
	{
		return &mobjarray[idnum];
	}
	return NULL;
}

void P_SetMobjAbort()
{
	Con_ErrorD("P_SetMobjState: Thing undefined.\n"
	"Probable cause: savegame and load state have different dehacked states.\n"
	"E.g. game saved with -nodeh switch but loaded with it not set");
}

//===========================================================================
// P_SetState
//	'statenum' must be valid
//===========================================================================
void P_SetState(mobj_t *mobj, int statenum)
{
	RDBOOL	spawning = !mobj->state;
	state_t	*st;

	// 220-9
	if (statenum < 0 || statenum >= maxnumstates)
		Con_Error(	"ERROR: bad state with T%i type %i [%s %s]\n"
					"Probable cause: incompatible savegame used\n",
					MAPNUM(mobj),
					mobj->type,
					mapinfo->id,
					mapinfo->mapname);

	mobj->state  = st = states + statenum;
	mobj->tics   = st->tics;
	mobj->sprite = (spritenum_t)st->sprite;
	mobj->frame  = st->frame;

	// Check for a ptcgen trigger.
	if (   r_use_particles
		&& statenum
		&& st->ptrigger
		&& mobj->thinker.id
		&& !(statenum == S_GIBS && mobj->type == MT_SKULL)
		&& !(mobj->flags2 & MF2_NOAIRSUPPLY)
		// v220-6 prevent particle generation with dehacked items
		&& !(hasDEH
		&& mobj->type < DH_NUMOBJECTS
		&& sprites[mobj->sprite].sprflags & SPR_NOMODEL)
		&& (spawning
		|| !(((ded_ptcgen_t*)st->ptrigger)->flags & PGF_SPAWN_ONLY)))
	{
		// We are allowed to spawn the generator.
		P_SpawnParticleGen((ded_ptcgen_t*)st->ptrigger, mobj);
	}
}

//
// P_SetMobjState
//  Returns true if the mobj is still present.
//  Modified to fix an original doom bug that
//  could lead to recursion
//  Method devised by Lee Killough 4/9/98
//

static void callFunc(int num, mobj_t *mo)
{
	if (num)
	{
		if (num >= nmBFGSpray && num < nmEnd)
		{
			actionlinks[num].func(mo);
		}
		// v220-35 check for shared functions called by an enemy
		else if (num == nmOpenShotgun2)
		{
			S_StartSound(sfx_dbopn, mo);
		}
		else if (num == nmLoadShotgun2)
		{
			S_StartSound(sfx_dbload, mo);
		}
		else
		{
			int    sprnum = states[mo->info->spawnstate].sprite;
			RDBOOL dh = hasDEH
						&& sprnum >= 0 && sprnum < DH_NUMSPRITES
						&& mo->type < DH_NUMOBJECTS
						&& (DEHCheck[mo->type] & DEHF_THING || DEHCheck[sprnum] & DEHF_SPRITE);

			Con_Error(	"%s error: action num %i %s\n"
						" player function %s set with invalid type [doomednum %i]\n%s",
						dh ? "DEH" : "PROGRAM",
						num,
						num < nmEnd ? actionlinks[num].name : "Invalid",
						mo->info->doomednum,
						dh && CFX ? "Try running in basic doom mode" : "");
		}
	}
}


// killough 4/9/98: used to remember states seen, to detect cycles:
static int	*seenstate_tab; // fast transition table
static int	*tempstate;		// for use with recursion

// Only called once when number of states loaded is known
void P_SetStateTable(int num)
{
	maxnumstates = num;
	seenstate_tab = (int*)W_Calloc(num * sizeof(int*), STATICID, 0);
	tempstate = (int*)W_Calloc(num * sizeof(int*), STATICID, 0);
}

// v220-35
RDBOOL P_SetSmokeState(mobj_t *mobj, int state)
{
	if (R3D_edit) return false;

    do
    {
		if (state == S_NULL) break;
		P_SetState(mobj, state);	// spawn particle gens

		state = states[state].nextstate;
    }
	while (!mobj->tics);

	return true;
}

RDBOOL P_SetMobjState(mobj_t *mobj, int state)
{
	static int	recursion = 0;	// detects recursion
	int			*seenstate = seenstate_tab;      // pointer to table
	int			i = state;					// initial state
	RDBOOL		ret = true;					// return value
    state_t		*st;

	//-----------------------

	if (R3D_edit) return false;

	if (recursion++)				// if recursion detected,
		memset(seenstate=tempstate, 0, maxnumstates * sizeof(statenum_t)); // clear state table

    do
    {
		if (state == S_NULL)
		{
			mobj->state = (state_t *) S_NULL;
			P_RemoveMobj(mobj, 6);
			ret = false;
			break;
		}
		
		P_SetState(mobj, state);	// Plus, if needed, spawn particle gens

		st = &states[state];

		// v220-13 Call action revised. An action_num of 0 (A_Null) has been added
		if (st->action_num) callFunc(st->action_num, mobj);
		
		seenstate[state] = 1 + st->nextstate;

		state = st->nextstate;

    }
	while (!mobj->tics && !seenstate[state]);
		
	if (ret && !mobj->tics)			// detect state cycles
		Con_Message("Warning: State Cycle Detected with mobj T%i doomednum %i\n", MAPNUM(mobj), mobj->info->doomednum);

	if (!--recursion)
	{
		for ( ; (state = seenstate[i]) ; i = state - 1)
			seenstate[i] = 0;		// erase memory of states
	}

	return ret;

/*	//original code

    state_t*	st;
	
    do
    {
		if (state == S_NULL)
		{
			mobj->state = (state_t *) S_NULL;
			P_RemoveMobj(mobj);
			return false;
		}
		
		P_SetState(mobj, state);
		st = &states[state];

		mobj->turntime = false;		// $visangle-facetarget

		// Modified handling.
		// Call action functions when the state is set
		if (st->action) st->action(mobj);	
		
		state = st->nextstate;
    } while (!mobj->tics);
				
    return true;*/
}

// Returns the ground friction factor for the mobj. 
fixed_t P_GetMobjFriction(mobj_t* mo)
{
	return P_GetFriction(mo, NULL);// use P_GetFriction for time being										  
}

int SplashRND()
{
	int val = M_RandomD() % 8;
	if (val < 4) return val + 1;
	return -(val - 3);
}

fixed_t getZ(mobj_t *mo)
{
	int *flatslist = mo->subsector->flats3DArray;
	if (flatslist)
	{
		int c = *(int*)flatslist++;
		int p = mo->flat3Dlevel[1];
		if (p <= 0) return mo->subsector->sector->floorheight;
		if (p >= c) p = c - 1;
		flatslist += FLIST_ENTRIES * p;
		return sectors[flatslist[1]].floorheight +
									flatslist[2] +	// offset
									flatslist[3];	// depth
	}
	
	fixed_t ht = mo->subsector->s_waterplane;

	if (ht == DDMININT)
	{
		ht = mo->subsector->dlfloorheight;	// v220-31 $splash
	}

	return ht;
}

// v220-4 $stepsounds
// Note: should only be called after verifying that the player's z
//       is on the floor and that waterlevel is either 0 or 1
#include "dd_wad.h"
static void P_PlayerStep(mobj_t *mo)
{
int S_FootstepSound(int sndid, float vol, int shiftup);

	static int shiftup = 0;

	fixed_t	wp = mo->subsector->s_waterplane;
	float	vol;
	int		sndid;
	int		flags;

	//-------------------------

	if (!waterlevel && mo->z < wp) return;

	fixed_t ht = DDMAXINT;

	// Default values
	sndid = r3ds.jumpdn_id;
	vol = 1.f;

	flags = R_GetFlatFlags(gdt.fpic);

	if (flags & FLT_SPLASH)
	{
		if (mo->z < wp || (wp == DDMININT && mo->z == (ht = getZ(mo))) )
		{
			if (flags & (FLT_WATER|FLT_BLKWATER))
			{
				sndid = mo->z < wp ? r3ds.wade_id : sfx_gloop;
				vol = mo->z < wp ? 0.4f : 0.3f * cfg.splashvol;
			}
			else if (flags & FLT_LAVA)
			{
				if (waterlevel == 2) return;
				sndid = sfx_burn; vol = 0.6f * cfg.splashvol;
			}
			else if (flags & (FLT_SLIME|FLT_BLOOD|FLT_NUKE))
			{
				if (waterlevel == 2) return;
				sndid = sfx_sludg; vol = 0.3f * cfg.splashvol;
			}
		}
	}

	if (sounds[sndid].flags & SFX_BADSOUND) return;

	S_FootstepSound(sndid, vol, (shiftup = !shiftup));
}

float	last_vol = 0;
int		last_id = 0;
int		Sys_GetTime (void);

// b211-0 Added model test
// v220-33 only play sound with player footsteps
int P_SpawnSplash(int x, int y, int floorh, float vol, int flags, RDBOOL chkmo)
{
	mobj_t		*mo;
	int			delay;
	int			scp_flag = (flags & F3D_3DFLAT) ? (MJF_SCP_SPLASH|MJF_SCP_3DFLOOR) : MJF_SCP_SPLASH;

	last_vol = 0, last_id = 0;

	if (!cfg.footsteps || !BDMFX) return false;

	delay = ST_SOUND + (M_RandomD() & 0x7);

	if (!(flags & FLT_LAVA) && (chkmo || !cfg.splashsprite))
	{
		static 	degenmobj_t	soundorg;	// for sound positioning

		if (flags & FLT_BLKWATER || flags & FLT_WATER)
		{
			if (abs(leveltime - splashtime[ST_WATER]) > delay)
			{
				last_vol = vol, last_id = sfx_gloop;
				soundorg.x = x, soundorg.y = y, soundorg.z = floorh;
				S_StartSoundAtVolume(sfx_gloop, (mobj_t*)&soundorg, vol, 1.0f);
				splashtime[ST_WATER] = leveltime;
			}
		}
		else if (flags & (FLT_NUKE|FLT_SLIME|FLT_BLOOD) )
		{
			int id = flags & FLT_NUKE ? ST_NUKE : flags & FLT_SLIME ? ST_SLIME : ST_BLOOD;

			if (abs(leveltime - splashtime[id]) > delay)
			{
				last_vol = vol, last_id = sfx_sludg;
				soundorg.x = x, soundorg.y = y, soundorg.z = floorh;
				S_StartSoundAtVolume(sfx_sludg, (mobj_t*)&soundorg, vol, 1.0f);
				splashtime[id] = leveltime;
			}
		}

		if (!cfg.splashsprite && !(flags & FLT_LAVA)) return false;
	}
	else
	if (flags & FLT_WATER)
	{
		mo = P_SpawnSplashMobj(x, y, floorh, MT_SPLASHBASE);
		mo->splashbase = floorh;
		mo->misc |= scp_flag;

		if (abs(leveltime - splashtime[ST_WATER]) > delay)
		{
			last_vol = vol, last_id = sfx_gloop;
			S_StartSoundAtVolume(sfx_gloop, mo, vol, 1.0f);
			splashtime[ST_WATER] = leveltime;
		}

		if (R_CheckModelExists(mo))
		{
			int hs = mo->subsector->sector->heightsec;

			// v220-4 adjust model z in a trans fake sector
			if (hs >= 0 && sectors[hs].wvar && sectors[hs].wvar->opacity < 0xFF)
			{
				mo->z = mo->floorz = mo->splashbase = floorh - 4*FRACUNIT;
			}

			return true; // Note: unlike a sprite a model's splash has only one part
		}

		// droplet momentum values
		mo->momx = SplashRND()<<14;
		mo->momy = SplashRND()<<14;
		mo->momz = 5*FRACUNIT + (SplashRND()<<14);

		mo = P_SpawnSplashMobj(x, y, floorh, MT_SPLASH);
		mo->splashbase = floorh;
		mo->misc |= scp_flag;

		return true;
	}
	else
	if (flags & FLT_BLKWATER)
	{
		mo = P_SpawnSplashMobj(x, y, floorh, MT_BLACKSPLASHBASE);
		mo->splashbase = floorh;
		mo->misc |= scp_flag;

		if (abs(leveltime - splashtime[ST_WATER]) > delay)
		{
			last_vol = vol, last_id = sfx_gloop;
			S_StartSoundAtVolume(sfx_gloop, mo, vol, 1.0f);
			splashtime[ST_WATER] = leveltime;
		}

		if (R_CheckModelExists(mo))
		{
			int hs = mo->subsector->sector->heightsec;

			// v220-4 adjust model z in a trans fake sector
			if (hs >= 0 && sectors[hs].wvar && sectors[hs].wvar->opacity < 0xFF)
			{
				mo->z = mo->floorz = mo->splashbase = floorh - 4*FRACUNIT;
			}

			return true; // Note: unlike a sprite a model's splash has only one part
		}

		// droplet momentum values
		mo->momx = SplashRND()<<14;
		mo->momy = SplashRND()<<14;
		mo->momz = 5*FRACUNIT + (SplashRND()<<14);

		mo = P_SpawnSplashMobj(x, y, floorh, MT_BLACKSPLASH);
		mo->splashbase = floorh;
		mo->misc |= scp_flag;

		return true;
	}
	if (flags & FLT_LAVA)
	{
		if (cfg.splashsprite)
		{
			mo = P_SpawnSplashMobj(x, y, floorh, MT_LAVASPLASH);
			mo->splashbase = floorh;
			mo->misc |= scp_flag;
		}

		mo = P_SpawnSplashMobj(x, y, floorh, MT_LAVASMOKE);
		mo->splashbase = floorh;
		mo->momz = (FRACUNIT>>3)*2;

		if (abs(leveltime - splashtime[ST_LAVA]) > delay)
		{
			S_StartSound(sfx_burn, mo);
			splashtime[ST_LAVA] = leveltime;
		}

		return true;
	}
	if (flags & FLT_SLIME)
	{
		mo = P_SpawnSplashMobj(x, y, floorh, MT_SLUDGESPLASHBASE);
		mo->splashbase = floorh;
		mo->momx = SplashRND()<<13;
		mo->momy = SplashRND()<<13;
		mo->misc |= scp_flag;

		if (abs(leveltime - splashtime[ST_SLIME]) > delay)
		{
			last_vol = vol, last_id = sfx_sludg;
			S_StartSoundAtVolume(sfx_sludg, mo, vol, 1.0f);
			splashtime[ST_SLIME] = leveltime;
		}

		if (R_CheckModelExists(mo)) return true;

		mo->momz = 4*FRACUNIT;

		mo = P_SpawnSplashMobj(x, y, floorh, MT_SLUDGESPLASH);
		mo->splashbase = floorh;
		mo->misc |= scp_flag;

		return true;
	}
	if (flags & FLT_BLOOD)
	{
		mo = P_SpawnSplashMobj(x, y, floorh, MT_BLOODSPLASHBASE);
		mo->splashbase = floorh;
		mo->momx = SplashRND()<<14;
		mo->momy = SplashRND()<<14;
		mo->misc |= scp_flag;

		if (abs(leveltime - splashtime[ST_BLOOD]) > delay)
		{
			last_vol = vol, last_id = sfx_sludg;
			S_StartSoundAtVolume(sfx_sludg, mo, vol, 1.0f);
			splashtime[ST_BLOOD] = leveltime;
		}

		if (R_CheckModelExists(mo)) return true;

		mo->momz = 6*FRACUNIT;

		mo = P_SpawnSplashMobj(x, y, floorh, MT_BLOODSPLASH);
		mo->splashbase = floorh;
		mo->misc |= scp_flag;

		return true;
	}
	if (flags & FLT_NUKE)
	{
		mo = P_SpawnSplashMobj(x, y, floorh, MT_NUKESPLASHBASE);
		mo->splashbase = floorh;
		mo->momx = SplashRND()<<14;
		mo->momy = SplashRND()<<14;
		mo->misc |= scp_flag;

		if (abs(leveltime - splashtime[ST_NUKE]) > delay)
		{
			last_vol = vol, last_id = sfx_sludg;
			S_StartSoundAtVolume(sfx_sludg, mo, vol, 1.0f);
			splashtime[ST_NUKE] = leveltime;
		}

		if (R_CheckModelExists(mo)) return true;

		mo->momz = 4*FRACUNIT;

		mo = P_SpawnSplashMobj(x, y, floorh, MT_NUKESPLASH);
		mo->splashbase = floorh;
		mo->misc |= scp_flag;

		return true;
	}
	last_vol = 0, last_id = 0;
	return false;
}

//
// P_ExplodeMissile  
//
static void P_ExplodeMissile (mobj_t* mo, int from)
{	
	mo->momx = mo->momy = mo->momz = 0;

	P_SetMobjState (mo, mobjinfo[mo->type].deathstate);
	
	mo->tics -= (M_RandomD() & 3);
	
	if (mo->tics < 1) mo->tics = 1;
	
	if(mo->flags & MF_MISSILE)
	{
		mo->flags &= ~MF_MISSILE;
		mo->flags |= MF_VIEWALIGN;

		// Remove or set as appropriate the brightshadow flag.
		if(mo->flags & MF_BRIGHTEXPLODE) mo->flags |= MF_BRIGHTSHADOW;
		else
		if(mo->flags & MF_BRIGHTSHADOW ) mo->flags &= ~MF_BRIGHTSHADOW;

		mo->x = mo->old_x;
		mo->y = mo->old_y;
	}
	
	if (mo->info->deathsound)
		S_StartSound(mo->info->deathsound, mo);
}

// v220-31 Various tests to check for sky collision
static void P_CheckExplodeOrRemove(mobj_t *mo, sector_t *s)
{
	line_t *li;

	if (   gdt.tmz <= s->floorheight
		&& (s->floorpic == skyflatnum || mo->subsector->ssfloorpic == skyflatnum))
	{
		P_RemoveMobj(mo, 7); return;
	}
	else if ( (li = gdt.floorline)
		&& li->spflags & SPF_NOTEX
		&& li->backsector
		&& gdt.tmz < MAX_OF(li->frontsector->floorheight, li->backsector->floorheight)
		&& li->backsector->floorpic == skyflatnum
		&& li->frontsector->floorpic == skyflatnum)
	{
		P_RemoveMobj(mo, 8); return;
	}
	else if ( (li = gdt.ceilingline)
		&& li->spflags & SPF_NOTEX
		&& li->backsector
		&& gdt.moheight > MIN_OF(li->frontsector->ceilingheight, li->backsector->ceilingheight)
		&& li->backsector->ceilingpic == skyflatnum
		&& li->frontsector->ceilingpic == skyflatnum)
	{
		P_RemoveMobj(mo, 9); return;
	}
	else if (gdt.moheight > s->ceilingheight
		&& (s->ceilingpic == skyflatnum || mo->subsector->ssceilpic == skyflatnum))
	{
		P_RemoveMobj(mo, 10); return;
	}
	else if (s->ceilingpic == skyflatnum
		&& gdt.moheight >= SL_PointHeight(gdt.newsubsec->sector, gdt.tmx, gdt.tmy, HISLOPE))
	{
		P_RemoveMobj(mo, 11); return;
	}
	else if ( (li = gdt.ceilingline) 
		&& li->backsector
		&& li->frontsector != li->backsector
		&& gdt.moheight >= MIN_OF(li->frontsector->ceilingheight, li->backsector->ceilingheight))
	{
		if (   li->frontsector->stptr < 0
			&& li->backsector->stptr < 0
			&& li->frontsector->ceilingpic == skyflatnum
			&& li->backsector->ceilingpic == skyflatnum)
		{
			P_RemoveMobj(mo, 12); return;	// must have hit a sky wall
		}

		int side = P_PointOnLineSide(gdt.tmx, gdt.tmy, li);
		
		if (   (!side && li->frontsector->ceilingpic == skyflatnum && li->frontsector->stptr >= 0)
			|| ( side && li->backsector->ceilingpic == skyflatnum  && li->backsector->stptr >= 0) )
		{
			P_RemoveMobj(mo, 13); return;	// Hit a line referencing sloped sectors in a sky ceiling
		}
	}

	P_ExplodeMissile(mo,2);
}

//
// P_XYMovement  
//  b211-0 - some revisions made
//

extern fixed_t playermaxmove;

static fixed_t clampTo (fixed_t val, fixed_t max, fixed_t min)
{
	return val <= min ? min : val >= max ? max : val;
}

static void P_XYMovement (mobj_t* mo) 
{
	if (!mo->momx && !mo->momy)
	{
		if (mo->flags & MF_SKULLFLY)
		{
			// the skull slammed into something
			mo->flags &= ~MF_SKULLFLY;
			mo->momz = 0;
			
			P_SetMobjState (mo, mo->info->seestate);
			return;
		}

		int lf3d;

		if (   !mo->momz
			&& mo->z == mo->subsector->dlfloorheight
			&& !mo->subsector->sector->floordata
			&& !mo->subsector->sector->ceilingdata
			&& !( (lf3d = mo->subsector->sector->liftsec3D) >= 0 && sectors[lf3d].floordata))
		{
			return;
		}
	}

	mbf_player_t *player = mo->player;
	RDBOOL		 isplayer = player && player->plr->mo == mo;
	RDBOOL		 set_missile_pos = false;
	
	fixed_t		ptryx;
	fixed_t		ptryy;
	fixed_t		xmove;
	fixed_t		ymove;

	fixed_t		x = mo->x;
	fixed_t		y = mo->y;
	
	int			maxmove;
	int			loopn;
	int			stptr, grad;

	//----------------------------
	
	// $democam: cameramen have their own movement code
//	if (P_CameraXYMovement(mo)) return;

	if (isplayer)
	{
		if (gdt.onladder)	// R3D
		{
			if (   !flying
				&& (mo->z > gdt.ladderbase
				&&  mo->z < gdt.laddertop ) )
			{
				mo->momx = mo->momy = 0;
				return;
			}
			
			gdt.onladder = false;
		}

	/*	maxmove =  (waterlevel == 2 ? MAXMOVE>>1 :
		waterlevel == 3 ? MAXMOVE>>2 : MAXMOVE);*/

		// v220-4
		//   lowered maxmove limit if waterlevel == 1
		//   ensure maxmove is still restricted if in fly mode
		maxmove = playermaxmove;
	}
	else
	{
		// v220-4 Abort only if on ground with no x, y or z momentum
		//        else even if no x or y momentum the molist must be set
		if (mo->z == mo->subsector->sector->floorheight && !mo->momx && !mo->momy && !mo->momz)
		{
			return;
		}
		
		maxmove = MAXMOVE;
	}
	
	loopn = 1;
	xmove = mo->momx;
	ymove = mo->momy;

	if (!(mo->flags & MF_MISSILE) && (abs(xmove) > maxmove || abs(ymove) > maxmove))	// added v220-14
	{
		xmove = clampTo(mo->momx, maxmove, -maxmove);
		ymove = clampTo(mo->momy, maxmove, -maxmove);

		fixed_t frac = MIN_OF(FixedDiv(xmove, mo->momx), FixedDiv(ymove, mo->momy));

		xmove = mo->momx = FixedMul(mo->momx, frac);
		ymove = mo->momy = FixedMul(mo->momy, frac);
	}
	
	// v220-13 To prevent fallthrough into a sector with a lower floor bounded
	//         by vertical or horizontal lines exactly 32 map units apart ensure
	//         the player, if moving, has momentum in both axes.
	if (isplayer && !(mo->misc & MJMISC_TRANS_TC) && mo->floorz != mo->dropoffz)
	{
		if (mo->momx && !mo->momy) mo->momy = 4;
		else
		if (mo->momy && !mo->momx) mo->momx = 4;
	}

	// v220-33 apply to player also
	if (isplayer || mo->subsector->ssflags & SSF_NORENDER)
	{
		// v220-13 Edge a thing in a control sector nearer to blocking
		//         lines in case it should have crossed a trigger or
		//         teleport line.
		if (abs(xmove) > 4*FRACUNIT || abs(ymove) > 4*FRACUNIT)
		{		
			loopn += MAX_OF(abs(xmove), abs(ymove)) / (4*FRACUNIT);

			xmove = FixedDiv(xmove,loopn*FRACUNIT);
			ymove = FixedDiv(ymove,loopn*FRACUNIT);
		}
	}
	else
	{
		// v220-12 Fix bug in original Doom source that failed to detect lines
		//         being crossed when the Thing's displacement was >= its radius
		if (mo->radius && (abs(xmove) >= mo->radius || abs(ymove) >= mo->radius))
		{		
			loopn += MAX_OF(abs(xmove), abs(ymove)) / mo->radius;

			xmove = FixedDiv(xmove,loopn*FRACUNIT);
			ymove = FixedDiv(ymove,loopn*FRACUNIT);
		}
	}
	
	do
	{
		RDBOOL res;

		ptryx = mo->x + xmove;
		ptryy = mo->y + ymove;

		// killough $dropoff_fix
		if ( (res = P_TryMove (mo, ptryx, ptryy, 0)) )
			set_missile_pos = false;
		
		if (!res)
		{
			line_t	*lb;
			
			if (gdt.player && gdt.onladder) return;
			
			if (ge_inf.grenade_id && mo->type == ge_inf.grenade_id) mo->flags3 |= MF3_ARMED;

			if (mo->flags & MF_SKULLFLY)
			{
				// The skull's slammed nto something
				mo->momx = mo->momy = 0;
				return;
			}

			lb = gdt.blockline;
			if (!lb) lb = gdt.floorline;
			if (!lb) lb = gdt.ceilingline;

			// v220-8 kill x,y momentum if floater's trying to go into a wall
			if (   lb
				&& mo->target															// v220-8
				&& !((mo->flags ^ MF_FLOAT) & (MF_FLOAT | MF_SKULLFLY | MF_INFLOAT)) )
			{
				mo->momx = mo->momy = 0;
				break;
			}

			// blocked move
			// killough 8/11/98: bouncing off walls
			// killough 10/98:
			// Add ability for objects other than players to bounce on ice
			// v220-13 support dehacked spawnshot
			if (   lb
				&& !(mo->flags & MF_MISSILE)
				&& (mo->flags3 & MF3_BOUNCES
				|| (!player && mo->z <= mo->floorz && P_GetFriction(mo, NULL) > ORIG_FRICTION)))
			{
				int		dx = lb->dx >> FRACBITS;
				int		dy = lb->dy >> FRACBITS;

				fixed_t f = (dx * mo->momx + dy * mo->momy) / (dx * dx + dy * dy);
				fixed_t x = FixedMul(f, lb->dx);
				fixed_t y = FixedMul(f, lb->dy);

				//-------------------------------
					
				// reflect momentum away from wall
				mo->momx = x*2 - mo->momx;
				mo->momy = y*2 - mo->momy;

				gdt.blockline = NULL;

			//	if (mo->type != MT_SPAWNSHOT)
				{
					// if under gravity, slow down in direction perpendicular to wall.
					if (!(mo->flags & MF_NOGRAVITY))
					{
						mo->momx = (mo->momx + x)/2;
						mo->momy = (mo->momy + y)/2;
					}
				}
			}
			else if (mo->flags & MF_MISSILE)
			{
				sector_t *s = mo->subsector->sector;

				// If a missile is blocked by a Thing explode it
				if (gdt.thingblocked)
				{
					P_ExplodeMissile(mo,1);
					return;
				}

				if (lb && lb->tyflags & TYF_SKYWALL)
				{
					P_RemoveMobj(mo, 14);		// Don't explode against sky wall
					return;
				}

				if (gdt.explodemissile)
				{
					P_CheckExplodeOrRemove(mo, s);
					return;
				}

				// v220-3 remove missiles hitting sloped sky ceilings
				if (   s->ceilingpic == skyflatnum
					&& s->specFlags2 & SF2_HISLOPE
					&& mo->z > SL_LowCeilHeight(s))
				{
					P_RemoveMobj(mo, 15);
					return;
				}

				// Check to prevent missiles exploding against the sky. (sky floors ignored)
				if (gdt.ceilingline && gdt.ceilingline->backsector)
				{
					RDBOOL fsky = gdt.ceilingline->frontsector->ceilingpic == skyflatnum;
					RDBOOL bsky = gdt.ceilingline->backsector->ceilingpic == skyflatnum;
					
					if (fsky && bsky)
					{
						int ht = MIN_OF(gdt.ceilingline->frontsector->ceilingheight,
								gdt.ceilingline->backsector->ceilingheight);

						if ( !(origTEX[gdt.ceilingline - lines] & 0x44) )
							ht += gdt.ceilingline->frontsector->skyfix;
							
						if (mo->z > ht - 4*FRACUNIT)
						{
							P_RemoveMobj(mo, 16);
							return;
						}

						if (gdt.floorline && gdt.floorline->backsector)
						{
							ht = MAX_OF(gdt.floorline->frontsector->floorheight,
								gdt.floorline->backsector->floorheight);

							if (mo->z > ht + 4*FRACUNIT)
							{
								set_missile_pos = true;
								goto end_loop;
							}
						}
						else
						{
							set_missile_pos = true;
							goto end_loop;
						}
					}
					else if (fsky || bsky)
					{
						s = P_PointOnLineSide(mo->x, mo->y, gdt.ceilingline) ?
							gdt.ceilingline->backsector : gdt.ceilingline->frontsector;
						
						if (skyflatnum == s->ceilingpic && mo->z > s->ceilingheight - 4*FRACUNIT)
						{
							// Prevent missiles exploding against the sky. (sky floors ignored)
							P_RemoveMobj(mo, 17);
							return;
						}
					}
				}
				
				// v220-10 rewritten to check validity of blockline sidenum
				if (mo->subsector->sector->ceilingpic == skyflatnum)
				{
					if (gdt.blockline)
					{
						// v220-13 ammended given new step reductiom momentum
						sector_t *s = mo->subsector->sector;

						if (mo->z > s->ceilingheight || s->ceilingheight == s->floorheight)
						{
							P_RemoveMobj(mo, 18);
							return;
						}
					}
					else
					{
						if (mo->z >= mo->subsector->sector->ceilingheight + mo->subsector->sector->skyfix)
						{
							P_RemoveMobj(mo, 19); return;
						}

						set_missile_pos = true;
						goto end_loop;
					}
				}

				P_ExplodeMissile(mo,3);
				return;
			}
			else if (isplayer)	// b220-2 revised
			{
				line_t	*lt = gdt.plrlineblocked;

				// Increment the counter used with punch and chainsaw
				if (gdt.plrblocked < 5) gdt.plrblocked++;

				// v220-5 swim up when thing blocked test added
				if (  (lt || gdt.thingblocked)
					&& !(mo->flags3 & MF3_ONGROUND)
					&& player->cmd.forwardmove
					&& player->jumptics <= 0
					&& waterlevel		
					&& waterlevel < 4
					&& player->plr->clLookDir >= 0
					&& ppt.viewer_fake_sector >= 0
					&& sectors[ppt.viewer_fake_sector].ceilingheight > sectors[ppt.viewer_fake_sector].floorheight)
				{
					float movefactor = player->plr->clLookDir/110;
					// If in water auto swim up if blocked
					mo->momz = FRACUNIT + (int)(movefactor*3*FRACUNIT/2);
				}
				else if (lt && lt->ml_flags & ML_TWOSIDED)
				{
					// v220-14 simplify jump detection - see also $jumping
					if (  !waterlevel
						&& player->jumptics > 0
						&& mo->momz > 0)
					{
						gdt.plrisjumping = true;
						break;
					}

					gdt.plrisjumping = false;
				}
			}

			if (   isplayer
				|| (mo->misc & MJMISC_TRANS_TC
				&& (!player || mo->health <= 0)))
			{
				P_SlideMove(mo);		// try to slide along it
			}
			// v220-5 if blocked kill x,y momentum
			else if (gdt.blockline)
			{
				if (mo->flags & MF_NOGRAVITY && sentient(mo))
				{
					P_SlideMove(mo);	// try to slide along it
				}
				else
					mo->momx = mo->momy = 0;
			}

			break;
		}
		else if (isplayer)
		{
			if (gdt.plrblocked) gdt.plrblocked--;
		}
end_loop:;

	} while (--loopn);

	// R3D: set true if a missile can continue based on skyfix values
	if (set_missile_pos)
	{
		P_UnlinkThing(mo);
		mo->x = x + mo->momx, mo->y = y + mo->momy;
		P_SetThingPosition(mo);
	}

	// debug option for no sliding at all
	if (player && player->cheats & CF_NOMOMENTUM)
	{
		mo->momx = mo->momy = 0;
		return;
	}

	// $slopes - player/grenade dynamics R3D
	if (  (player && mo->servoz)
		|| mo->flags3 & MF3_BOUNCES
		|| mo->flags & (MF_SHOOTABLE|MF_CORPSE)
		|| mo->type == MT_BLOOD)
	{
		int fz = gdt.slopefht == DDMAXINT ? mo->floorz : gdt.slopefht;	// b220-1

		if (   !mo->onmobj	// If standing on a Thing on a slope then not actually on the slope
			&& (stptr = mo->subsector->sector->stptr) >= 0
			&& (grad = slopes[stptr].slope[0].gradient) != 0
			&& mo->z - 3*FRACUNIT <= fz)
		{
			if (mo->momx || mo->momy)
			{
				// b211-0 check for corpse
				if (mo->flags & MF_CORPSE)
				{
					mo->momx = mo->momx ? FixedMul(mo->momx, 0xCFFF) : 0;
					mo->momy = mo->momy ? FixedMul(mo->momy, 0xCFFF) : 0;
				}
				else 
				if (player || mo->flags3 & MF3_BOUNCES)
				{
					int mapangle = slopes[stptr].slope[0].mapangle;
					int movangle, grav;
					
					// Calculate mapangle of movement		
					float mangle = R_PointToAngle2(mo->x, mo->y,
						mo->x + mo->momx, mo->y + mo->momy)/(float)ANGMAX * 360 - 90;
					
					if (mangle < 0) mangle += 360;
					
					movangle = (int)mangle;
					
					if (grad > 0x32000) grad = 0x32000;
					
					grad >>= 1;
					
					grav = (int)((float)(cos((double)(movangle - mapangle)/180 * DPI))*(float)grad);
					
					if ( !(grav < 0 && waterlevel) )
					{
						mo->momx -= mo->momx ? FixedMul(mo->momx, grav) : 0;
						mo->momy -= mo->momy ? FixedMul(mo->momy, grav) : 0;
	
						if (mo->momx > maxmove)
							mo->momx = maxmove;
						else if (mo->momx < -maxmove)
							mo->momx = -maxmove;
						
						if (mo->momy > maxmove)
							mo->momy = maxmove;
						else if (mo->momy < -maxmove)
							mo->momy = -maxmove;
					}
				}
			}
			
			mo->flags2 |= MF2_ONSLOPE;
			mo->flags3 &= ~MF3_FALLING;
			
			goto chk_pnomove;
		}
		mo->flags2 &= ~MF2_ONSLOPE;
	}

	if (mo->flags & (MF_MISSILE | MF_SKULLFLY) )
		return; 	// no friction for missiles; ever
	
	if (   mo->z > mo->floorz
		&& !mo->onmobj
		&& !(player && waterlevel))
	{
		return;		// no friction when airborne unless in water or flying
	}

chk_pnomove:;

	// killough 8/11/98: add bouncers
	// killough 9/15/98: add objects falling off ledges
	// killough   11/98: only include bouncers hanging off ledges
	if ( ((mo->flags3 & MF3_BOUNCES && mo->z > mo->dropoffz)
		|| mo->flags3 & MF3_FALLING)
		&& (abs(mo->momx) >  FRACUNIT/4
		||  abs(mo->momy) >  FRACUNIT/4)
		&& mo->floorz != mo->subsector->sector->floorheight)
	{
		return;  // do not stop sliding if halfway off a step with some momentum
	}

	// Stop any thing where it's momentum is < STOPSPEED including
	// the player unless the player is being given a move command.
	// killough 11/98: ensure a player's voodoo doll does not affect the player
	if (   abs(mo->momx) < STOPSPEED
		&& abs(mo->momy) < STOPSPEED)
	{
		if (isplayer && !(player->cmd.forwardmove || player->cmd.sidemove) )
		{
			// if in a walking frame, stop moving
			if ( (unsigned)(player->plr->mo->state - states - S_PLAY_RUN1) < 4)
			{
				P_SetMobjState(player->plr->mo, S_PLAY);
			}
		}

		// v220-22 Don't slow things on scrolling sectors
		if (!(mo->misc & MJMISC_TRANS_TC))
			mo->momx = mo->momy = 0;
	}
	else
	{
		// phares 3/17/98
		//
		// Friction will have been adjusted by friction thinkers for
		// icy or muddy floors. Otherwise it was never touched and
		// remained set at ORIG_FRICTION
		
		fixed_t friction = P_GetFriction(mo, NULL);

		if (isplayer) stepfriction = friction;

		// v220-4 if player and no input then apply user selected drag
		if (   isplayer
			&& friction == ORIG_FRICTION
			&& mo->health > 0
			&& !(player->cmd.forwardmove || player->cmd.sidemove)
			&& !(mo->misc & MJMISC_TRANS_TC) )
		{
			mo->momx = FixedMul(mo->momx, ORIG_FRICTION - playerdrag*1024);
			mo->momy = FixedMul(mo->momy, ORIG_FRICTION - playerdrag*1024);
		}
		else
		{
			// v220-10 ZDoom scroll compatibility
			//         voodoo dolls in scroll floor controlling
			//         sectors seem to have slightly more friction
			if (   player
				&& !isplayer
				&& ma_inf.modifier & MODF_SCROLL
				&& friction == ORIG_FRICTION
				&& mo->misc & MJMISC_TRANS_TC
				&& mo->subsector->ssflags & SSF_NORENDER)
			{
				friction -= 256;
			}

			mo->momx = FixedMul(mo->momx, friction);
			mo->momy = FixedMul(mo->momy, friction);
		}
	}
}

// v220-33 Make ladder lines more user friendly
static fixed_t lmolow, lmohigh;

static RDBOOL PIT_LadderLines(line_t* ld, RDBOOL *up)// v220-33 added for ladder movement check
{
	if (   gdt.tmbbox[BOXRIGHT] <= ld->bbox[BOXLEFT ]
		|| gdt.tmbbox[BOXLEFT ] >= ld->bbox[BOXRIGHT]
		|| gdt.tmbbox[BOXTOP  ] <= ld->bbox[BOXBASE ]
		|| gdt.tmbbox[BOXBASE ] >= ld->bbox[BOXTOP  ]
		|| P_BoxOnLineSide(gdt.tmbbox, ld) != -1)
	{
		return true;	// ignore
	}

	if (!ld->backsector) return false;
	if (ld->ovflags & OVF_3DLADDERLINE) return true;

	sector_t *s = ppt.view_subsector->sector;

	if (ld->specType >= R3D_40)	// 3D lines
	{
		int		lift = ld->frontsector->liftsec3D;
		fixed_t z_offset = lift >= 0 ? sectors[lift].planes[0].speed : 0;
		fixed_t	cf = ld->backsector->ceilingheight;
		fixed_t	lf = ld->backsector->floorheight;
		fixed_t	height = lift >= 0 ? sectors[lift].floorheight : lf;
		fixed_t zh = height + ld->d3d->offset3D + ld->d3d->totalheight;

		if (lmohigh > zh && lmolow < zh)
		{
			if (!(*up)) gdt.onladder = false;
			return false;
		}
	}
	else
	{
		if (*up)
		{
			if (ld->frontsector->ceilingheight <= lmohigh || ld->backsector->ceilingheight <= lmohigh)
				return false;
		}
		else
		{
			if (ld->frontsector->floorheight >= lmolow || ld->backsector->floorheight >= lmolow)
			{
				gdt.onladder = false;
				return false;
			}
		}
	}

	return true;
}

RDBOOL P_CheckLines(mobj_t *mo, fixed_t z, fixed_t nz, int dir)
{
	lmolow = z, lmohigh = nz;

	if (mo->subsector->flats3DArray)
	{
		int *flatslist = mo->subsector->flats3DArray;
		int		count = *(int*)flatslist++;

		flatslist += (count - 1)*FLIST_ENTRIES;

		// Remember - floor's higher than ceiling
		while (--count >= 0)
		{
			fixed_t zh = sectors[flatslist[1]].floorheight +
								flatslist[2];	// offset
								;
			if (dir)
			{

				if (lmohigh > zh && lmolow < zh)
				{
					return false;
				}
			}
			else
			{
				zh += flatslist[3];	// depth

				if (lmohigh > zh && lmolow < zh)
				{
					gdt.onladder = false;
					return false;
				}
			}

			flatslist -= FLIST_ENTRIES;
		}
	}

	int		xl, xh;
	int		yl, yh;
	int		bx, by;

	gdt.tmthing = mo;

	M_setBoxOffsets(gdt.tmbbox, mo->x, mo->y, mo->radius);
	M_getBlockOffsets(gdt.tmbbox, 0, &xl, &xh, &yl, &yh);

	validcount++;

	// Set thing's line list
	for (bx = xl ; bx <= xh ; bx++)
	{
		for (by = yl ; by <= yh ; by++)
		{
			if (!P_BlockLinesIterator (bx, by, (int (__cdecl *)(struct line_s *,void *))PIT_LadderLines, &dir))
				return false;
		}
	}
	return true;
}

//
// P_ZMovement
//
// v220-4 splash code reworked
// v220-5 noclip is only used by boss spit cubes which never get
//        here so MF_NOCLIP test with missiles has been removed.
//
static void P_ZMovement (mobj_t* mo)
{
	RDBOOL	is_grenade;		// b211-0
	RDBOOL  mf_missile = mo->flags & MF_MISSILE;
	fixed_t gravity = mapgravity;
	fixed_t	cht;
	int		zflags;
	int		hmo;

	// $democam: cameramen get special z movement
	// if (P_CameraZMovement(mo)) return;

	if (mf_missile && mo->subsector->ssceilpic == skyflatnum)
		cht = mo->subsector->sector->ceilingheight + mo->subsector->sector->skyfix;
	else
		cht = mo->ceilingz;
	
	zflags = R_GetFlatFlags(gdt.fpic) & FLT_SPLASH;

	if (   zflags
		&& mo->flags & (MF_SPECIAL|MF_DROPPED|MF_PICKUP)	// MF_SPECIAL indicates it's the player
		&& mo->momz < 0
		&& waterlevel < 2
		&& !(mo->flags & MF_NOCLIP))
	{
		fixed_t	ht = getZ(mo);

		if (mo->z + mo->momz <= ht)	// v220-31 $splash
		{
			float vol = gdt.player ? mo->momz <= -18*FRACUNIT ? cfg.splashvol : FIX2FLT(FixedDiv(mo->momz, -18*FRACUNIT)) : 1.0f;

			P_SpawnSplash(mo->x, mo->y, ht, vol, zflags, ppt.viewplrmo == mo);
		}
	}

	// Set if valid grenade
	is_grenade = ge_inf.grenade_id
			&& mo->type == ge_inf.grenade_id
			&& mo->health > 0;

	// killough 7/11/98:
	// BFG fireballs bounced on floors and ceilings in Pre-Beta Doom
	// killough 8/9/98: added support for non-missile objects bouncing
	// (e.g. grenade, mine, pipebomb)
	
	if (mo->flags3 & MF3_BOUNCES && mo->momz)
	{
		fixed_t fz = mo->floorz;

		mo->z += mo->momz;

		if (mo->onmobj) fz = mo->onmobj->z + mo->onmobj->height;

		if (mo->z <= fz)						// bounce off floors
		{
			if (!zflags) S_StartSound(sfx_bounce, mo);

			mo->z = fz;

			if (mo->momz < 0)
			{
				mo->momz = -mo->momz;

				if (!(mo->flags & MF_NOGRAVITY))		// bounce back with decay
				{
					mo->momz = mo->flags & MF_FLOAT ?	// floaters fall slowly
						mo->flags & MF_DROPOFF ?		// DROPOFF indicates rate
						FixedMul(mo->momz, (fixed_t)(FRACUNIT*.85)) :
						FixedMul(mo->momz, (fixed_t)(FRACUNIT*.70)) :
						FixedMul(mo->momz, (fixed_t)(FRACUNIT*.45)) ;
					
					// Bring it to rest below a certain speed
					if (abs(mo->momz) <= mo->info->mass*(FRACUNIT*4/256))
						mo->momz = 0;
				}
				
				// killough 11/98: touchy objects explode on impact
				if (   mo->flags3 & MF3_TOUCHY
					&& mo->flags3 & MF3_ARMED
					&& mo->type != ge_inf.grenade_id	// v220-31 these only explode when touched
					&& mo->health > 0)
				{
					P_DamageMobj(mo, NULL, NULL, mo->health, 5);
				}
				else
				{
					if (mo->flags & MF_FLOAT && sentient(mo))
						goto floater;

					mo->momz = 0;
				}
				return;
			}
		}
		else
		{
			if (mo->z >= cht - mo->height)   // bounce off ceilings
			{
				mo->z = cht - mo->height;

				if (mo->momz > 0)
				{
					if (mo->subsector->sector->ceilingpic != skyflatnum)
					{
						mo->momz = -mo->momz;    // always bounce off non-sky ceiling
						S_StartSound(sfx_bounce, mo);
					}
					else
					{
						if (mf_missile || is_grenade)	// b211-0
						{
							// missiles don't bounce off skies whether or not clipped
							P_RemoveMobj(mo, 20);
						}
						else
						{
							if (mo->flags & MF_NOGRAVITY)
								mo->momz = -mo->momz; // bounce unless under gravity
							
							if (mo->flags & MF_FLOAT && sentient(mo))
								goto floater;
							// else fall to floor
						}
					}
					return;
				}
			}
			else
			{
				if (!(mo->flags & MF_NOGRAVITY))      // free-fall under gravity
					mo->momz -= mo->info->mass*(FRACUNIT/256);
				if (mo->flags & MF_FLOAT && sentient(mo))
					goto floater;
				return;
			}
			
			// came to a stop
			mo->momz = 0;
			
			if (mf_missile)
			{
				if (gdt.ceilingline &&
					gdt.ceilingline->backsector &&
					gdt.ceilingline->backsector->ceilingpic == skyflatnum &&
					mo->z > (mf_missile ?
					gdt.ceilingline->backsector->ceilingheight + gdt.ceilingline->backsector->skyfix :
					gdt.ceilingline->backsector->ceilingheight) )
				{
					P_RemoveMobj(mo, 21);
					return;
				}
				else
				{
					P_ExplodeMissile(mo,4);
					return;
				}	
			}

			if (mo->flags & MF_FLOAT && sentient(mo))
				goto floater;

			return;
		}
	}

	// b211-0 exclude voodoo dolls	
	if (mo->dplayer && mo->dplayer->mo == mo)
	{
		// check if hit ceiling above fake sector - R3D
		if (gdt.plrceilblocked)
		{
			mo->momx = mo->momy = 0;
			mo->momz = -6*FRACUNIT;
			gdt.plrceilblocked = false;
		}
		// check for smooth step up;
		else if (mo->z < mo->floorz)
		{
			if (mo->flags2 & MF2_ONSLOPE)	// reduce adjustment
				mo->dplayer->viewheight -= (mo->floorz - mo->z)>>1;
			else
				mo->dplayer->viewheight -= mo->floorz - mo->z;
			
			mo->dplayer->deltaviewheight
				= (plrviewheight - mo->dplayer->viewheight)>>3;
		}
    
		if (gdt.onladder)	// R3D
		{
			ticcmd_t*	cmd = &mo->player->cmd;

			if (mo->z < gdt.laddertop && cmd->forwardmove)
			{
				byte dir = (cmd->forwardmove > 0)^(mo->dplayer->clLookDir >= 0);

				if (dir)
				{
					if (   mo->z > gdt.ladderbase
						&& P_CheckLines(mo, mo->z-4*FRACUNIT, mo->z + mo->height, 0))
					{
						mo->z -= 4*FRACUNIT;
					}
				}
				else
				{
					if (   mo->z < gdt.laddertop
						&& mo->z + mo->height < mo->subsector->sector->ceilingheight
						&& P_CheckLines(mo, mo->z, mo->z + mo->height + 4*FRACUNIT, 1))
					{
						mo->z += 4*FRACUNIT;
					}
				}
			}
			mo->momx = mo->momy = mo->momz = 0;
			mo->floorz = mo->z;
			return;
		}
	}

    // adjust height
    mo->z += mo->momz;

floater:

    // adjust z if necessary
	if (   mo->target &&
		!((mo->flags ^ MF_FLOAT) & (MF_FLOAT | MF_SKULLFLY | MF_INFLOAT)) )	// killough 11/98
	{
		fixed_t delta = mo->target->z + (mo->target->height>>1) - (mo->z + (mo->height>>1));
		fixed_t dist  = P_ApproxDistance(mo->x - mo->target->x, mo->y - mo->target->y);

		if (   dist < mo->radius + mo->target->radius 
			&& abs(delta) < mo->height + mo->target->height)
		{
			// Don't go INTO the target (i.e. do nothing further)
			gdt.adjustz = false;
		}
		else if (delta < 0)
		{
			if (   dist < -(delta*3)
				&& checkStepupClear(mo, mo->z - FLOATSPEED))
			{
				mo->srvo[VZ] = (short)(FLOATSPEED >> 8);
				mo->z -= FLOATSPEED;
			}
		}
		else
		{
			if (   dist < (delta*3)
				&& checkStepupClear(mo, mo->z + FLOATSPEED))
			{
				mo->srvo[VZ] = (short)(-FLOATSPEED >> 8);
				mo->z += FLOATSPEED;
			}
		}
	}

	// Clip movement. Another thing?
	if (mo->onmobj)
	{
		hmo = mo->onmobj->z + mo->onmobj->height;

		if (mo->z <= hmo || mo->z < mo->floorz)
		{
			if (mf_missile)
			{
				P_ExplodeMissile (mo,5);
				return;
			}
			
			if (mo->momz < 0)
			{
				if (mo->dplayer && mo->dplayer->mo == mo && mo->momz < -gravity*8)	
				{
					// Squat down.
					// Decrease viewheight for a moment
					// after hitting the ground (hard),
					// and utter appropriate sound.
					mo->dplayer->deltaviewheight = mo->momz>>3;

					if (!underwater) S_StartSound(sfx_oof, mo);
				}
				mo->momz = 0;
			}
			else
			if (   gdt.floorline
				&& gdt.floorline->specType >= R3D_40	// if floorline set as a 3D line
				&& mo->misc & MJF_LINE3D)				// and mo is on or hitting a 3D line
			{
				fixed_t adj = gdt.floorline->d3d->offset3D - mo->onmobj->height;

				if (adj > 0 && adj <= 24*FRACUNIT) hmo += adj;
			}

			mo->z = mo->floorz = hmo;
		}
	}

	// Slope servoz setting - also see in P_TryMove() R3D
	if (mo->servoz && !mo->player)
	{
		if (mo->z > mo->floorz)
		{
			mo->momz -= mo->z - mo->floorz;
			mo->z = mo->floorz;
		}
		mo->srvo[VZ] = -(mo->servoz >> 8);
		gdt.adjustz = false;
	}

	// Hitting the floor.
    if (mo->z <= mo->floorz)
    {
		// if skull hits floor then reverse z-momentum
		if (mo->flags & MF_SKULLFLY) mo->momz = -mo->momz;
		
		if (mo->momz < 0)
		{
			// killough 11/98: touchy objects explode on impact
			if (   mo->flags3 & MF3_TOUCHY
				&& mo->flags3 & MF3_ARMED
				&& mo->health > 0)
			{
				P_DamageMobj(mo, NULL, NULL, mo->health, 6);
			}
			else
			if (mo->dplayer && mo->dplayer->mo == mo)
			{
				if (mo->momz < -gravity*8)	
				{
					// Squat down.
					// Decrease viewheight for a moment
					// after hitting the ground (hard),
					mo->dplayer->deltaviewheight = mo->momz>>3;

					if (!underwater)
					{
						// v220-3 falling damage - uses same values as Hexen
						if (cfg.fallingDamage && mo->momz < -gravity*23)
						{
							int damage;

							if (mo->momz < -gravity*63)
								damage = 10000;	// smashed to bits
							else
							{
								int factor = FixedMul(-mo->momz, 16*FRACUNIT/23);
								  
								damage = ((FixedMul(factor, factor)/10)>>FRACBITS)-24;
							}

							if (waterlevel) damage >>= 1;

							// no-death threshold
							if (mo->momz > -39*FRACUNIT && damage > mo->health && mo->health != 1)
							{
								damage = mo->health - 1;
							}

							P_DamageMobj(mo, NULL, NULL, damage, 7);
						}
					}
				}

				if (waterlevel != 4)	// No sounds in fly mode even if on ground
				{
					if (!zflags && mo->momz < -gravity*6 && gdt.fpic != skyflatnum)
					{
						// b211-0 if really hard utter appropriate sound.
						if (mo->momz < -gravity*11)
							S_StartSound(sfx_oof, mo);
						else
							S_StartSound(r3ds.jumpdn_id, mo);
					}
				}
			}

			if (!zflags && (mo->type == MT_BARREL || mo->type == MT_MISC77) )
			{
				if (   M_GroupVis(mo->subsector)
					&& !(hasDEH && DEHCheck[mo->type] & DEHF_THING))
				{
					S_StartSoundAtVolume(r3ds.bardrop_id, mo, 0.5f, 2.0f);
				}
			}

			mo->momz = 0;
		}

		mo->z = mo->floorz;
		
		if (is_grenade || mf_missile) // b211-0 add grenade
		{
			if (mo->subsector->sector->floorpic == skyflatnum)
				P_RemoveMobj(mo, 22);
			else if (!is_grenade)	// b211-0 grenades just come to rest
				P_ExplodeMissile(mo,6);
			return;
		}
    }
	else if (mo->type == MT_SPLASH)	// R3D
	{
		if(mo->momz == 0)
			mo->momz = -gravity;
		else
			mo->momz -= gravity/2;
	}
	else if (mo->flags2 & MF2_LOGRAV)
	{
		if (!mo->momz == 0) mo->momz = -(gravity>>3);

		mo->momz -= gravity>>3;
	}
    else if ( !(mo->flags & MF_NOGRAVITY) )
    {
		if (!mo->momz) mo->momz = -gravity;

		mo->momz -= gravity;
    }
	
    if (mo->z + mo->height > cht)
    {
		if (mo->momz > 0) mo->momz = 0;	// hit the ceiling

		mo->z = cht - mo->height;
		
		if (mo->flags & MF_SKULLFLY)	// the skull slammed into something
		{
			mo->momz = -mo->momz;
		}

		if (is_grenade || mf_missile) // b211-0
		{
			if (mo->subsector->sector->ceilingpic == skyflatnum)
			{
				P_RemoveMobj(mo, 23);
			}
			else if (!is_grenade)	// b211-0
			{
				P_ExplodeMissile(mo,7);
			}
		}
    }
}

mobj_t *P_SpawnTeleFog(int x, int y)
{
	mobj_t *mo = P_SpawnMobj (x, y, 
		(R_PointInSubsector(x, y))->sector->floorheight, 
		MT_TFOG, 0);

	P_SetMpos(mo);

	return mo;
}

//===========================================================================
// P_NightmareRespawn
//  Rewritten v220-31 to use existing mobj and support 3D lifts
//===========================================================================

static fixed_t l3dheightCheck(subsector_t *ss, mobj_t *mo)
{
	line_t	*ld;
	seg_t	*sg = &segs[ss->firstline];

	for (int i = 0; i < ss->linecount; i++, sg++)
	{
		if (  (ld = sg->linedef)
			&& ld->specType >= R3D_40
			&& XYonLine(mo->x, mo->y, ld, mo->radius))
		{
			return ld->d3d->totalheight;
		}
	}

	return 0;
}

static RDBOOL P_NightmareRespawn (mobj_t* mobj)
{
	if (   mobj->thinker.function != (void (__cdecl *)(void))P_MobjThinker
		|| !(mobj->info->flags & MF_COUNTKILL)	// ?
		|| !mobj->info->spawnstate)				// must be respawnable
	{
		return false;
	}

	fixed_t		x = mobj->spawnpoint.x<<FRACBITS, y = mobj->spawnpoint.y<<FRACBITS;
	fixed_t		h = mobj->height, r = mobj->radius;
	int			pcptest = 0;
	// store some current vars
	int			ctics = mobj->corpsetics;	// probably 0 but...
	int			flags = mobj->flags;
	int			flags2 = mobj->flags2;
	state_t		*st = mobj->state;

    mobj->momx = mobj->momy = mobj->momz = 0;	// should be 0 but just in case

	// set to starting position
	mobj->x = x, mobj->y = y;

	// Restore starting values
    mobj->radius = mobj->info->radius;
    mobj->height = mobj->info->height;
    mobj->flags  = mobj->info->flags;
	mobj->flags2 = mobj->info->flags2 & MF2_LOADMASK;		// R3D - filter flags2
	mobj->flags3 = mobj->info->flags3 & MF3_TRANSFERMASK;	// R3D - filter flags3
    mobj->health = mobj->info->spawnhealth;
	mobj->player = NULL;
	mobj->dplayer = NULL;
	mobj->corpsetics = 0;
	mobj->angle = ANG45 * (mobj->spawnpoint.angle/45);
	mobj->state = &states[mobj->info->spawnstate];

	// Now determine the z height when first spawned
	subsector_t *ss = R_PointInSubsector(x, y);
	sector_t	*sc = ss->sector;

	mobj->subsector = ss;

	mobj->floorz = mobj->dropoffz = mobj->z = SL_PointHeight(sc, mobj->x, mobj->y, LOSLOPE);
	mobj->ceilingz = SL_PointHeight(sc, mobj->x, mobj->y, HISLOPE);

	if (mobj->flags & MF_SPAWNCEILING)
	{
		mobj->z = mobj->ceilingz - mobj->height;
	}
	else if (ss->flats3DArray)	// This *must* be tested first (else 3D line checks would muck things up)
	{
		int *flatslist = ss->flats3DArray;
		int count = *(int*)flatslist++;
		sector_t *rsec = &sectors[flatslist[1]];
		int v = mobj->flat3Dlevel[0];

		if (v < 1)
		{
			mobj->ceilingz = rsec->floorheight;

			if (rsec->floordata) mobj->ceilingz += rsec->planes[0].speed;

			if (mobj->z + mobj->height > mobj->ceilingz) goto abort;
		}

		if (v > 0 && v <= count)	// valid entry
		{
			flatslist += (v - 1)*FLIST_ENTRIES;
			rsec = &sectors[flatslist[1]];
			int ceil  = rsec->floorheight + flatslist[2] + rsec->planes[0].speed;	// offset (smoothfloor only valid when rendering)
			int floor = ceil + flatslist[3];	// depth

			mobj->z = mobj->floorz = mobj->dropoffz = floor;

			if (v < count)
			{
				flatslist += FLIST_ENTRIES;
				rsec = &sectors[flatslist[1]];
				mobj->ceilingz = rsec->floorheight + flatslist[2] + rsec->planes[0].speed;
			}
			else
			{
				mobj->ceilingz = sc->ceilingheight;
			}
		}

		if (mobj->ceilingz - mobj->floorz < mobj->height) goto abort;

		mobj->flat3Dlevel[1] = mobj->flat3Dlevel[0];
		pcptest = -1;
	}
	else
	{
		// This assumes that the Thing is spawned on any 3D line that it may be
		// in contact with and on top of any 3D lift sector the line may be ref-
		// erencing. To specifically render under a line or lift sector then a
		// 16000 Thing has to be used to set the height. To do this then -32000
		// should be used to force the Thing to be spawned on the sector's floor.

		int of = mobj->spawnz << FRACBITS;
		int zp = sc->liftsec3D >= 0 ? sectors[sc->liftsec3D].floorheight : sc->floorheight;

		if (mobj->spawnz == -32000)
		{
			// do nothing
		}
		else
		{
			if (!of) of = l3dheightCheck(ss, mobj);	// check if on a 3D line

			zp += of;

			if (of <= 0 && zp - mobj->height < sc->floorheight)
			{
				if (sc->liftsec3D >= 0) goto abort;	// Wait for the height to change
				else zp = sc->floorheight;
			}
			if (zp + mobj->height > sc->ceilingheight) goto abort;

			mobj->floorz = mobj->dropoffz = mobj->z = zp;
			pcptest = -1;
		}
	}

	// Check not blocked by another Thing or a moving 3D lift
	if ( !(P_CheckPosition(mobj, x, y, mobj->z, pcptest)) )
	{
abort:;
		// restore corpse settings
		mobj->height = h, mobj->radius = r;
		mobj->flags = flags;
		mobj->flags2 = flags2;
		mobj->corpsetics = ctics;
		mobj->health = 0;
		mobj->radius = 0;
		mobj->height = 0;
		mobj->state = st;
		return false;	// can't be resurrected ATM
	}

	ma_inf.mobjsResurrected++;

	P_SetTarget(&mobj->target,    NULL);
	P_SetTarget(&mobj->tracer,    NULL);
	P_SetTarget(&mobj->lastenemy, NULL);

	P_UnlinkThing(mobj);

	// start tp sound and spawn a teleport fog	
	// Must be done first so as not to wake up anything
	mobj_t *mo = P_SpawnMobj(
		// Spawn slightly off position to prevent texture rendering fight 
		mobj->x + ((M_RandomD() - M_RandomD()) << 12),
		mobj->y + ((M_RandomD() - M_RandomD()) << 12),
		mobj->z, MT_TFOG, 0);

	P_SetMpos(mo);
	S_StartSound(sfx_telept, mo);

    mobj->flags  = mobj->info->flags;
	mobj->flags2 = mobj->info->flags2 & MF2_LOADMASK;		// R3D - filter flags2
	mobj->flags3 = mobj->info->flags3 & MF3_TRANSFERMASK;	// R3D - filter flags3
	mobj->validcount = 0;
	mobj->ddflags = 0;
	mobj->old_x = mobj->x;
	mobj->old_y = mobj->y;
	mobj->inverseAlpha = 0;
	mobj->gear = 0;
	mobj->servoz = 0;
	mobj->turntime = 0;
	mobj->lastlook = 0;
	mobj->threshold = 0;
	mobj->pursuecount = 0;
	mobj->strafecount = 0;
	mobj->movecount = 0;
	mobj->movedir = 0;
	mobj->visangle = mobj->angle >> 16;
	P_SetDDFlags(mobj);
	P_SetState(mobj, mobj->info->spawnstate);

	if(gameskill != sk_nightmare)
		mobj->reactiontime = mobj->info->reactiontime;
	else
		mobj->reactiontime = 0;

	P_SetThingPosition(mobj);
	P_UpdateThinker(&mobj->thinker);	// v220-31 placed here
	P_SetMpos(mobj);		// update the mobjarray
	return true;
}

// v220-33 for mobjs that are stuck due to conflict with dropoff
typedef struct {
	int box[4];
	fixed_t floorz;
	fixed_t dropoffz;
	fixed_t height;
} nhf_t;

nhf_t nhfdata;

static RDBOOL PIT_NextHighestFloor(line_t* ld, void *data)
{
	if (  !ld->backsector
		|| ld->frontsector == ld->backsector
		|| nhfdata.box[BOXRIGHT] <= ld->bbox[BOXLEFT ]
		|| nhfdata.box[BOXLEFT ] >= ld->bbox[BOXRIGHT]
		|| nhfdata.box[BOXTOP  ] <= ld->bbox[BOXBASE ]
		|| nhfdata.box[BOXBASE ] >= ld->bbox[BOXTOP  ]
		|| P_BoxOnLineSide(nhfdata.box, ld) != -1)
	{
		return true;	// ignore
	}

	if (   nhfdata.height < ld->frontsector->floorheight
		&& ld->frontsector->floorheight < nhfdata.floorz
		&& ld->frontsector->floorheight > nhfdata.dropoffz)
	{
		nhfdata.height = ld->frontsector->floorheight;
	}

	if (   nhfdata.height < ld->backsector->floorheight
		&& ld->backsector->floorheight < nhfdata.floorz
		&& ld->backsector->floorheight > nhfdata.dropoffz)
	{
		nhfdata.height = ld->backsector->floorheight;
	}

	return true;
}

//===========================================================================
// P_MobjThinker
//  Rewritten for R3D
//===========================================================================
void P_MobjThinker(mobj_t* mobj)
{
	RDBOOL	allow_type = true;
	RDBOOL	armable = false;
	int		snd;

	mobj->lastdx = mobj->lastdy = 0;

	if (mobj->modelboxnum > 0 && !(mobj->flags3 & MF3_INTERREACT))
		return;
	
	if (teleport_flash && ppt.thinkers_have_run)
	{
		P_SpawnMobj(teleport_flash->x, teleport_flash->y, teleport_flash->z, MT_TFOG, 0);
		S_LocalSound(sfx_telept, teleport_flash);
		teleport_flash = NULL;
	}

	// R3D specials
	if (mobj->flags3 & MF3_DEHAMBIENT && !(leveltime & 0x7))
	{
		if (compat[c_DEHamb])
			S_SoundRepeat(mobj->info->painsound|HSF_REPEAT, mobj, dehamb_vol, dehamb_attn);
		return;
	}

	if (mobj->flags2 & MF2_DONTDRAW)
	{
		if (mobj->type == MT_FAKESOUNDSOURCE)
		{
			mobj->z = mobj->subsector->sector->floorheight;
		}
		else
		if (mobj->type == MT_TPUNDERFAKE || mobj->type == MT_TPOVERFAKE)
		{
			// Note; if viewer_fake_sector >= 0 then it guarantees
			//       view_subsector and viewplayer are set.
			if (ppt.viewer_fake_sector >= 0 && !(ppt.view_subsector->sector->specFlags2 & SF2_PSEUDOFAKE))
			{
				int	under = mobj->type == MT_TPUNDERFAKE ? 1 : 0;
				int num = TIDNUM(mobj);
				
				if (tpfake[num].secnum[under] == ppt.view_subsector->sector - sectors)
				{
					mobj_t		*plrmo = ppt.viewplayer->mo;
					fixed_t		newx, newy;

					//-----------------------------------------

					if (    plrmo->player->cheats & CF_NOCLIP
						||	( under && viewz >= ppt.fake_floorheight)
						||  (!under && viewz <  ppt.fake_floorheight))
					{
						return;
					}

					// Add the player's offset from the current Thing
					// to the position of the Thing's pair.
					newx = tpfake[num].x_pos[!under] + plrmo->x - mobj->x;
					newy = tpfake[num].y_pos[!under] + plrmo->y - mobj->y;

					// setting check - commented out v220-3
					// Can fail with stacked subsectors referencing different sectors
					// in a 3D map.  This was only a belt-and-braces check so is not
					// strictly necessary.  If the player's head is over in an under
					// fake sector (or under in an over) then it will always be moved.
				/*	{
						subsector_t	*ss = R_PointInSubsector(newx, newy);

						if (ss->sector - sectors != tpfake[num].secnum[!under])
							return;
					}*/

					P_UnlinkThing (plrmo);

					plrmo->x = newx;
					plrmo->y = newy;

					P_SetThingPosition (plrmo);

					if (   mobj->type == MT_TPUNDERFAKE
						&& BDMFX
						&& cfg.footsteps
						&& last_vol	// id and volume of last footstep
						&& last_id)
					{
						S_StartSoundAtVolume(last_id, plrmo, last_vol, 1.0f);
					}

					last_vol = 0, last_id = 0;
				}
			}
		}
		else
		if (mobj->type == MT_SOUNDSOURCE)
		{
			S_CheckAmbientSound(mobj);
		}
		else
		if (mobj->type == MT_MUSICSOURCE)
		{
			SetMAPMusic(mobj);
		}
		return;
	}

	if (mobj->flags2 & MF2_PUSHABLE && (mobj->momz || (!mobj->momx && !mobj->momy)) && SFXBF_playing(mobj))
	{
		int	sndid = AS_GetPushSound(mobj->info->doomednum);

		if (sndid) S_StopSound(sndid, mobj);
	}

	// R3D report stats when stressed
	if (mobj->flags & MF_MISSILE)
	{
		// Note: previous code in v220 removed as problem of missiles
		// passing through walls has been fixed in v220-4

		stats.missile_mobj_count++;
	}
	else
	if (   mobj->flags & MF_SHOOTABLE
		&& mobj->flags & MF_COUNTKILL
		&& mobj->type)	// I.e. not a player
	{
		stats.global_mobj_count++;
	}

	if (mobj->misc & MJF_SCP_SPLASH )
	{
		if (!(mobj->misc & MJF_SCP_3DFLOOR))
			mobj->z = mobj->floorz = mobj->splashbase = mobj->subsector->dlfloorheight;
	}
	else
	if (mobj->type == MT_PUFF)
	{
		// Note min and max limits are ignored as it matters little.
		mobj->z += mobj->momz;
		allow_type = false;
	}
	else
	if (mobj->type == MT_SHADOWS)
	{
		mobj->selector = (mobj->selector & ~DDMOBJ_SELECTOR_MASK) | 1;
	}
	else
	// v220-31 Deal with crushed Skulls/pipebombs or whatever directly
	if (   mobj->state == &states[S_GIBS]
		&& !(mobj->flags & MF_COUNTKILL)		// could need to be resurrected
		&& mobj->type							// not the player
		&& mobj->info->spawnstate != S_GIBS)	// not spawned in this state
	{
		// 1 tic - time for the particle generators to be dealt with
		// For MT_SKULL this is dealt with in P_SetState
		if (mobj->movecount++) P_RemoveMobj(mobj, 24);
		return;
	}
	else
	if (mobj->type == MT_PIPEBOMB)
	{
		 allow_type = ((mobj->state - states) == S_PIPEBOMB
			 || (mobj->state - states) == S_PIPEBOMB2);
		 armable = true;
	}
	else
/*	if (mobj->type == MT_GRENADE)
	{
		 allow_type = (mobj->state - states) == S_GRENADE;
		 armable = true;
	}
	else*/
	if (mobj->type == MT_GRENADEX)
	{
		 allow_type = (mobj->state - states) == S_GRENADEX;
		 armable = true;
	}
	else
	if (mobj->type == MT_MINE)
	{
		 allow_type = ((mobj->state - states) == S_MINE
					|| (mobj->state - states) == S_MINE2);
		 armable = true;
	}

	memcpy(&mpos.mp[mobj - mobjarray].ox, &mobj->x, sizeof(fixed_t) * 3);
	mobj->misc &= ~(MJF_LINE3D|MJF_FLAT3D|MJF_ONLIFT);	// b211-0

	// v220-31
	if (mobj->subsector->flats3DArray)
	{
		int n = mobj->flat3Dlevel[1];

		if (n > 0 && (mobj->subsector->flats3DArray[5] & (F3D_MOVFLAG << (n - 1))) > 0)
			mobj->misc |= MJF_ONLIFT;
	}
		// Check if the target was a voodoo doll. If so kill the player
	P_UpdateHealthBits(mobj);
	P_UpdateAlpha(mobj); // $mobjalpha

	if (mobj->flags2 & MF2_3DTELEPLINE)	// R3D
	{
		if (mobj->floorz > mobj->subsector->sector->floorheight)
		{
			mobj->floorz = mobj->subsector->sector->floorheight;
			mobj->momx = mobj->momy = mobj->momz = 0;
		}
		mobj->flags2 &= ~MF2_3DTELEPLINE;
	}

	if (mobj->flags2 & MF2_DONTDRAW)	// R3D
	{
		if (mobj->tics != -1)
		{
			if (!mobj->tics--)
				P_SetMobjState(mobj, mobj->state->nextstate);

			// Check for cyclic architectural dummy model sound
			// If needed this will have been placed in seesound
			if (   mobj->flags3 & MF3_AMODEL
				&& mobj->info->seesound > 0
				&& mobj->health > 0)
			{
				// These sounds are attenuated at 1/3rd max distance
				S_StartSoundAtVolume(mobj->info->seesound, mobj, 1.0f, 3.0f);
			}
		}
		return;
	}
	else	// v220-31 Move amodels in 3D lift sectors
	{
		if (   mobj->flags3 & MF3_AMODEL
			&& mobj->subsector->sector - sectors == mobj->subsector->sector->liftsec3D)
		{
			mobj->z = mobj->subsector->sector->floorheight;
		}
	}
	
	if (allow_type)
	{
		gdt.adjustz = true;
		allow_type = !(mobj->flags3 & (MF3_AMODEL|MF3_STATIC))
				  && !(mobj->flags2 & MF2_DONTDRAW);
	}

	// v220-5 keep voodoo dolls health in step with the main player
	if (mobj->player) mobj->health = plhealth.health;

	// v220-5 unclipped spawnshots just go directly from a to b
	if (mobj->type == MT_SPAWNSHOT && mobj->flags & MF_NOCLIP)
	{
		P_UnlinkThing(mobj);
		mobj->x += mobj->momx;
		mobj->y += mobj->momy;
		mobj->z += mobj->momz;
		P_SetThingPosition(mobj);	
	}
	// b211-0 - if grenade etc. and false then explosions stay put
	else if (allow_type)
	{
		int cptime = mobj->ddflags & DDF_FORCEFADE ? TICSPERSEC : corpsetime;

		P_XYMovement(mobj);
	
		// Check mobj has not been removed
		if (mobj->thinker.function != (void (__cdecl *)(void))P_MobjThinker) return;

		// v220-35 make sure pickups remain tied to a moving floor
		if (!R3D_ScriptedMap && mobj->flags == 0x00000001 && mobj->subsector->sector->specFlags & SF_MOVFLOOR)
			mobj->z = mobj->floorz = mobj->dropoffz = mobj->subsector->sector->floorheight;

		// v220-31 check stressed limit dropped items
		if (mobj->flags & MF_DROPPED)
		{
			if (   ma_inf.activedropped
				&& ma_inf.isdropped > ma_inf.activedropped
				&& mobj->movecount > 30*35)	// 30 seconds (assuming game can keep up)
			{
				P_RemoveMobj(mobj, 25);
				return;
			}
			else
			{
				mobj->movecount++;
				ma_inf.isdropped++;
			}
		}

		if (!(mobj->misc & MJF_SCP_SPLASH) && !(mobj->flags3 & MF3_AMODEL))
		{
			// v220-4 dead things float to the surface...
			if (   mobj->flags & MF_CORPSE
				&& !(mobj->flags & MF_SHADOW)
				&& mobj->corpsetics < 2
				&& mobj->subsector->s_waterplane > mobj->z
				&& mobj->info->height <= 56*FRACUNIT
				// v220-29 ...but only if splash flats have been set
				&& mobj->subsector->sector->heightsec >= 0	// should be
				&& sectors[mobj->subsector->sector->heightsec].specFlags2 & SF2_HASWATERFLATS
				// v220-31 and only if the depth of water is > 64
				&& mobj->subsector->s_waterplane - mobj->subsector->sector->floorheight > 64*FRACUNIT)
			{
				mobj->momx = mobj->momy = 0;

				// v220-31 sprites stay put
				if ( !(mobj->state->stflags & ST_NOMODEL) )
				{
				//	if ((mobj->subsector->s_waterplane - mobj->z) >= 5*FRACUNIT)
						mobj->corpsetics = 0;	// Don't fade out at all, previously until reaching the surface

					if (mobj->subsector->s_waterplane > mobj->z + 4*FRACUNIT)
					{
						mobj->z += (FRACUNIT >> 3);			
						mobj->srvo[VZ] = -(FRACUNIT >> 11);
					}
					else	// move with undulating water
					{
						mobj->z = mobj->subsector->s_waterplane - 4*FRACUNIT;
					}
				}
			}
			else
			if (   (mobj->z != mobj->floorz && !(mobj->flags2 & MF2_NOCHECK))	// $mod
				|| mobj->momz
				|| (gdt.onladder && mobj->dplayer && mobj->dplayer->mo == mobj)
				|| (  mobj->target															// v220-8
				&& !((mobj->flags ^ MF_FLOAT) & (MF_FLOAT | MF_SKULLFLY | MF_INFLOAT)) ))
			{
				// if unarmed touchy thing
				if (  (mobj->flags3 & (MF3_ARMED|MF3_TOUCHY)) == MF3_TOUCHY
					&& !armable					// that is not a mine/grenade
					&& mobj->momz < 0			// is falling
					&& mobj->z > mobj->floorz)	// and is above the floor
				{
					mobj->flags3 |= MF3_ARMED;	// then arm it
				}

				P_ZMovement(mobj);

				// Check mobj has not been removed
				if (mobj->thinker.function != (void (__cdecl *)(void))P_MobjThinker) return;
			}
			// Note: playerslide test removed v220-5
			else if (!mobj->momx
				&& !mobj->momy
				&& !sentient(mobj)		// non-sentient objects at rest
				&& !mobj->player)		// skip players
			{
				// arm a mine that has come to rest
				if (armable) mobj->flags3 |= MF3_ARMED;

				// killough 9/12/98: objects fall off ledges if they are hanging off
				// slightly push off of ledge if hanging more than halfway off
				if ( !(mobj->flags & (MF_FLOAT|MF_NOGRAVITY))	// v220-14 exclude
					&& mobj->z == mobj->floorz			// must be on the floor
					&& mobj->z > mobj->dropoffz)		// and contact dropoff
				{
					P_ApplyTorque(mobj);
				}
				else
				{
					mobj->flags3 &= ~MF3_FALLING;
					mobj->gear = 0;  // Reset torque
				}
			}
			// v220-33 could have been on a mobj and is now stuck in the air
			else if (sentient(mobj)
				&& mobj->state->action_num == nmChase
				&& !mobj->momx && !mobj->momy && !mobj->momz
				&& !(mobj->flags & (MF_FLOAT | MF_INFLOAT | MF_SKULLFLY | MF_NOGRAVITY))
				&& !(mobj->flags3 & MF3_FALLING)
				&& !mobj->onmobj
				&& mobj->floorz > mobj->subsector->sector->floorheight
				&& mobj->floorz == mobj->z )
			{
				int bx, by, xl, xh, yl, yh;

				M_setBoxOffsets(nhfdata.box, mobj->x, mobj->y, mobj->radius);
				M_getBlockOffsets(nhfdata.box, 0, &xl, &xh, &yl, &yh);

				validcount++;

				nhfdata.floorz = mobj->floorz;
				nhfdata.dropoffz = mobj->dropoffz;
				nhfdata.height = DDMININT;

				// Check thing's line list
				for (bx = xl ; bx <= xh ; bx++)
				{
					for (by = yl ; by <= yh ; by++)
						P_BlockLinesIterator (bx, by, (int (__cdecl *)(struct line_s *,void *))PIT_NextHighestFloor, NULL);
				}

				if (nhfdata.height > DDMININT) mobj->floorz = nhfdata.height;
				else if (mobj->dropoffz > mobj->subsector->sector->floorheight) mobj->dropoffz = mobj->subsector->sector->floorheight;
			}
		}

		// $vanish: dead monsters disappear after some time
		if (   cptime
			&& mobj->flags & MF_CORPSE
			&& !(mobj->misc & MJMISC_TRANS_TC)
			&& !(mobj->flags2 & (MF2_CORPSEFADE|MF2_NOFADE)))	// R3D
		{
			if (++mobj->corpsetics >= cptime)
			{
				if (mobj->corpsetics < cptime + VANISHTICS)
				{
					// Translucent during vanishing.
					mobj->inverseAlpha = ((mobj->corpsetics - cptime) * 255) / VANISHTICS;
				}
				else	// Too long; get rid of the corpse
				{
					if (   mobjinfo[mobj->type].raisestate != S_NULL
						&& (compat[c_resurrect] || respawnmonsters))
					{
						mobj->flags2 |= MF2_CORPSEFADE;
					}
					else if ((respawnmonsters && mobj->flags & MF_COUNTKILL) || wadani.mobj == mobj)
					{
						mobj->flags2 |= MF2_CORPSEFADE;
					}
					else
					{
						P_RemoveMobj(mobj, 26);
						return;
					}
				}
			}
		}
	}
	else
	{
		// prevent accidental incrementation.
		mobj->momx = mobj->momy = mobj->momz = 0;
	}

	// v220-4 $stepsounds
	if (   mobj == ppt.viewplrmo
		&& mobj->player
		&& BDMFX
		&& cfg.footsteps
		&& !(mobj->flags & MF_NOCLIP))
	{
		int Sys_GetTime (void);

		static	int step_interval;
		int		mov;
		int		nowtime = Sys_GetTime();
		int		period = nowtime - step_interval;

		if (  !mobj->momz
			&& mobj->z == mobj->floorz
			&& gdt.fpic != skyflatnum
			&& waterlevel <= 2
			&& (mov = FixedDiv(P_ApproxDistance(mobj->momx, mobj->momy), 0x13000)) > (FRACUNIT >> 2)
			&& period > 0x7f0000/mov)
		{
			step_interval = nowtime;

			// check if on ice or floor carry sector if
			// so no footstep unless being moved by user
			if (   (stepfriction > ORIG_FRICTION || mobj->misc & MJMISC_TRANS_TC)
				&& !(mobj->player->cmd.forwardmove|mobj->player->cmd.sidemove) )
			{
				// do nothing
			}
			else
				P_PlayerStep(mobj);
		}
		else if (period > 0x7f)
			step_interval = nowtime;
	}

	// v220-13 Kill sound immediately on death (i.e. don't wait for sound to complete)
	if (sthing && (snd = mobj->info->painsound) & HSF_SNDTHING)
	{
		soundthing_t *sth = &sthing[ (snd & ~HSF_SNDTHING) ];

		if (sth->type == mobj->type)
		{
			if (mobj->health <= 0)
			{
				if (S_IsPlaying(sth->soundid, mobj))
					S_StopSound(sth->soundid, mobj);
			}
			else
			if (!(leveltime&0x7))
			{
			// R3D - SOUNDTHING painsound's repeat (set in an SCP_DEFN script)
				S_SoundRepeat(sth->soundid|HSF_REPEAT, mobj, sth->volume, sth->attenuation);
			}
		}
	}

	// v220-13 Scripted types with doomednum between 6070 and 6099 are dealt
	//         with differently if the player is dead and they are not.
	if (ma_inf.playerdead && mobj->health > 0)
	{
		if (mobj->info->doomednum >= 6070 && mobj->info->doomednum <= 6099)
		{
			P_SetTarget(&mobj->target, NULL);	// stop any infighting

			// types with no gravity are held at their current frame
			if (!(mobj->flags & MF_NOGRAVITY)) return;
		}
	}

    // cycle through states, calling action functions at transitions
    if (mobj->tics != -1)
    {
		mobj->tics--;

		if (CFX && mobj->state - states == S_BFGLAND3)	// b211-0
			S_StartSound(r3ds.xtrabfg_id|HSF_NO_ATTENUATION, NULL);

		P_SRVOAngleTicker(mobj);	// "angle-servo"; smooth actor turning
		
		// you can cycle through multiple states in a tic
		if (!mobj->tics)
		{
			mobj_t		 *pl;
			int			 stnum;

			P_ClearThingSRVO(mobj, false);

			if (!mobj->tics && !P_SetMobjState (mobj, mobj->state->nextstate) )
				return;     // freed itself
			
			// R3D - Check for cyclic architectural model sound
			// If needed this will have been placed in seesound
			if (   mobj->flags3 & MF3_AMODEL
				&& (stnum = mobj->state - states) > 0
				&& mobj->health > 0
				&& (pl = ppt.viewplrmo))
			{
				if (   stnum == mobj->info->seestate
					&& (snd = mobj->info->seesound) > 0)
				{
					RDBOOL isp = S_IsPlaying(snd, mobj);

					if (   !isp
						&& S_CheckSoundDistance(mobj, NULL, 1.5f)
						&& P_SoundSight(mobj, pl))
					{
						S_StartSoundAtVolume(snd, mobj, 1.0f, 1.5f);
					}
					else if (isp)
					{
						S_StopSound(snd, mobj);
					}
				}
				else if (stnum == mobj->info->painstate
					&& (snd = mobj->info->painsound) > 0)
				{
					RDBOOL isp = S_IsPlaying(snd, mobj);

					if (   !isp
						&& S_CheckSoundDistance(mobj, NULL, 1.5f)
						&& P_SoundSight(mobj, pl))
					{
						S_StartSoundAtVolume(snd, mobj, 1.0f, 1.5f);
					}
					else if (isp)
					{
						S_StopSound(snd, mobj);
					}
				}
			}
		}
    }
	// check for nightmare respawn.
	//  Note: checking order important because of movecount increment
    else if (respawnmonsters
		&& mobj->flags & MF_COUNTKILL
		&& (mobj->movecount++) >= 12*35
		&& !(leveltime & 31)
		&& M_RandomD() <= 4)
    {
		P_NightmareRespawn (mobj);
    }
	else if (wadani.mobj)
	{
		for (int i = 0; i < 3; i++)
		{
			if (wadani.trigset[i] > 0)
			{
				mobj_t *th = wadani.mobj;

				if (   th								// was set
					&& th->flags & MF_CORPSE			// a dead monster
					&& th->tics == -1					// lying still
					&& th->info->raisestate != S_NULL	// has a raise state
					&& !(th->flags3 & MF3_AMODEL))		// not an amodel
				{
					if (P_NightmareRespawn(th))
						wadani.trigset[i] = 0;
				}
			}
		}
	}
}

//
// P_SpawnMobj
//
mobj_t *P_SpawnMobj( fixed_t x, fixed_t y, fixed_t z, mobjtype_t type, int iden_num )
{
    mobj_t		*mobj = GetFreeMobj();
    mobjinfo_t	*info = &mobjinfo[type];
	sector_t	*sec;
	int			id = iden_num & 0xFFFF;

    mobj->type = type;
    mobj->info = info;
    mobj->x = x;
    mobj->y = y;
    mobj->radius = info->radius;
    mobj->height = info->height;
    mobj->flags  = info->flags;
	mobj->flags2 = info->flags2 & MF2_LOADMASK;		// R3D - filter flags2
	mobj->flags3 = info->flags3 & MF3_TRANSFERMASK;	// R3D - filter flags3
    mobj->health = info->spawnhealth;
	mobj->scplightlevel = info->scplightlevel;
	mobj->player = NULL;
	mobj->dplayer = NULL;
	mobj->flat3Dlevel[0] = mobj->flat3Dlevel[1] = -1;
	
	if (R3D_ScriptedMap && id >= 7000 && id < 7700)
	{
		mobj->flags |= MF_LOCAL;	// R3D - ensure flag is set

		if (!mobj->health)	// v220-13 backward compatibility for soundthing
			mobj->health = info->spawnhealth = 1;
	}

	mobj->iden_nums = iden_num;

	mobj->startangle = norandomrotate ? type % 360 : rand() % 360;

	subfn_ResetHeight(mobj);	// v220-31
	P_SetDDFlags(mobj);

	P_AddThinker(&mobj->thinker);		// Link before setting state
	P_SetState(mobj, info->spawnstate);	// Do not use P_SetMobjState as action routines cannot be called yet
    P_SetThingPosition(mobj);			// Set subsector and sector and/or block links

	sec = mobj->subsector->sector;

	if(gameskill != sk_nightmare)
		mobj->reactiontime = info->reactiontime;
    
	if (R3D_ScriptedMap)
	{
		switch (mobj->sprite)
		{
		case SPR_TRED:
		case SPR_TGRN:
		case SPR_TBLU:
		case SPR_SMRT:
		case SPR_SMGT:
		case SPR_SMBT:
		case SPR_CAND:
		case SPR_CBRA:

		R_SetModelTorchCentre(mobj);

		default:
			break;
		}
	}
	
	//killough $dropoff_fix
	mobj->dropoffz = mobj->floorz = sec->floorheight;
    mobj->ceilingz = sec->ceilingheight;
	
	// R3D - Slopes are set in LS_SetTiltOffsets()
	//       So this must be run before spawning anything
	//       SL_PointHeight used to get correct height if on a slope
    if (z == ONFLOORZ)
	{
		fixed_t lheight = SL_PointHeight(sec, mobj->x, mobj->y, LOSLOPE);

		if (lheight == sec->floorheight)
		{
			subsector_t *ss = mobj->subsector;
			seg_t		*sg = &segs[ss->firstline];
			line_t		*ld;
			int			i, h;

			// v220-4 sometimes a mobj is placed exactly on a two's line with different sectors
			// meaning R_PointInSubsector has a 50/50 chance of returning one or the other. If
			// the sectors have different floorheights then this can cause a problem.

			validcount++;

			for (i = 0; i < ss->linecount; i++, sg++)
			{
				if (  (ld = sg->linedef)
					&& ld->ml_flags & ML_TWOSIDED
					// v220-5 eliminate things on a line with the same front and
					// back sector floorheights
					&& ld->frontsector->floorheight != ld->backsector->floorheight
					&& ld->validcount != validcount)
				{
					ld->validcount = validcount;

					if (ld->specType >= R3D_40)
					{
						lheight = sec->floorheight;
						break;	// mobjs on 3D lines are set by the analyser
					}
					else
					{
						h = MAX_OF(ld->frontsector->floorheight, ld->backsector->floorheight);

						if (   h > lheight
							&& XYonLine(mobj->x,  mobj->y, ld, 1)
							&& h + mobj->height <= MIN_OF(ld->frontsector->ceilingheight, ld->backsector->ceilingheight))
						{
							lheight = h;
						}
					}
				}
			}
		}

		if (lheight > sec->floorheight)
			mobj->flags2 |= MF2_HEIGHTSET;		// R3D - mark already set for analyser

		mobj->z = mobj->floorz = mobj->dropoffz = lheight;
	}
    else if (z == ONCEILINGZ)
	{
		// b220-1 heightsec test removed
		//        Note: heights are checked when pre-caching
		mobj->ceilingz = SL_PointHeight(sec, mobj->x, mobj->y, HISLOPE);
		mobj->z = mobj->ceilingz - mobj->height;

		if (R3D_ScriptedMap && id >= 7000 && id < 7700)
			mobj->flags2 |= (MF2_GLUECEIL|MF2_HEIGHTSET);

		if (   R3D_linemap
			&& !(mobj->flags2 & MF2_HEIGHTSET)
			&& mobj->z == mobj->subsector->sector->ceilingheight - mobj->height)
		{
			// check if on a 3D line and hook to lowest where it will fit between the line and the floor
			subsector_t *ss = mobj->subsector;
			seg_t		*sg = &segs[ss->firstline];
			line_t		*ld;
			int			i, h = mobj->z;

			validcount++;

			for (i = 0; i < ss->linecount; i++, sg++)
			{
				if (  (ld = sg->linedef)
					&& ld->validcount != validcount
					&& ld->specType == R3D_40
					&& ld->frontsector == ld->backsector
					&& ld->d3d
					&& ld->frontsector->liftsec3D < 0
					&& XYonLine(mobj->x,  mobj->y, ld, 1))
				{
					int bottomStep = ss->sector->floorheight + ld->d3d->offset3D;

					if (   h > bottomStep
						&& bottomStep - ss->sector->floorheight > mobj->height)
					{
						mobj->z = bottomStep - mobj->height;
					}

					ld->validcount = validcount;
				}
			}
		}

		// R3D - Don't adjust in analyser if not contacting the sector's ceiling
		if (mobj->ceilingz < sec->ceilingheight) mobj->flags2 |= MF2_HEIGHTSET;
	}
    else
	{
		mobj->z = z;
	}

	mobj->old_x = mobj->x;
	mobj->old_y = mobj->y;
	
	if (mobj->flags & MF_SPECIAL)				// a pickup
		mobj->pickupRadius = mobj->radius;

	CheckDesync(mobj); // v220-11

	return mobj;
}


//
// R3D - P_SpawnSplashMobj
//
static mobj_t* P_SpawnSplashMobj( fixed_t x, fixed_t y, fixed_t z, mobjtype_t type)
{
    mobj_t		*mobj = GetFreeMobj();
    mobjinfo_t	*info = &mobjinfo[type];
	
    mobj->type = type;
    mobj->info = info;
    mobj->x = x;
    mobj->y = y;
    mobj->radius = info->radius;
    mobj->height = info->height;
    mobj->flags = info->flags;
	mobj->flags2 = info->flags2 & MF2_LOADMASK;
    mobj->health = info->spawnhealth;

	P_SetDDFlags(mobj);

	mobj->reactiontime = info->reactiontime;
    
	// Link before setting state
	P_AddThinker (&mobj->thinker);
	
	P_SetState(mobj, info->spawnstate);

    // set subsector and sector and/or block links
    P_SetThingPosition (mobj);
	mobj->z = mobj->dropoffz = mobj->floorz = z;
    mobj->ceilingz = mobj->subsector->dlceilheight;

	P_SetMpos(mobj);

    return mobj;
}


//===========================================================================
// P_RemoveMobj
//===========================================================================
void P_RemoveMobj(mobj_t* mobj, int from)
{
	// stop any playing sound (unless death sound)
	if (!(mobj->ddflags & DDF_DEATHSOUND))
		S_StopAllSounds(mobj);
	
	// unlink from sector and block lists
	P_UnlinkThing (mobj);

	P_SetTarget(&mobj->target,    NULL);
	P_SetTarget(&mobj->tracer,    NULL);
	P_SetTarget(&mobj->lastenemy, NULL);

	// free block
	P_RemoveThinker (&mobj->thinker);
}

// R3D - Player height positioning check. If a player is straddling
// a line then a check is made of the set height.  A map may have
// a player positioned so as to fall when starting.  If the wrong
// height's been set then player will be stuck in the line.  This
// happens when a nodes builder does not select the correct side
// for the player mobj.
void subfn_CheckPlayerHeight(mobj_t *mobj, int x, int y, int z)
{
	subsector_t *ss = mobj->subsector;
	seg_t *seg = &segs[ss->firstline];
	int i, h1, h2;

	for (i = 0; i < ss->linecount; i++, seg++)
	{
		line_t *ld = seg->linedef;

		if (ld == NULL
			|| ld->backsector == NULL
			// continue if same height each side
			|| (h1 = ld->frontsector->floorheight)
			== (h2 = ld->backsector->floorheight))
		{
			continue;
		}

		if (h1 < h2) h1 = h2;	// set h1 to highest

		if (z < h1 && PointIsOnLine(x, y, ld, mobj->radius>>16))
		{
			// Note: no need to link/unlink thing as only z being changed.
			mobj->z = h1;
			break;
		}
	}
}
void	G_PlayerReborn (int pnum);
//
// P_SpawnPlayer
// Called when a player is spawned on the level.
// Currently only called with pnum == 0

mobj_t *P_SpawnPlayer (mapthing_t* mthing, int pnum, int iden_num)
{
    mbf_player_t	*p;
	mobj_t			*mobj;
    fixed_t			x, y, z;
    int				i;
	
	if(pnum) pnum = 0;	// Warning printed in G_PlayerReborn()

    // check playing
    if (!mbf_players[pnum].plr->ingame) return NULL;					
	
    p = &mbf_players[pnum];

    if (p->playerstate == PST_REBORN)
		G_PlayerReborn(pnum);

    x = mthing->x << FRACBITS;
    y = mthing->y << FRACBITS;
    z = ONFLOORZ;
    mobj = P_SpawnMobj(x, y, z, MT_PLAYER, iden_num);
	ma_inf.numthinkers++;

	subfn_CheckPlayerHeight(mobj, x, y, z);	// R3D

	P_SetMpos(mobj);

	// set color translations for player sprites
	i = cfg.PlayerColor[pnum]; 
	if(i > 0) mobj->flags |= i << MF_TRANSSHIFT;

    p->plr->clAngle = mobj->angle = ANG45 * (mthing->angle/45);
	p->plr->clLookDir = 0;
	p->plr->flags |= DDPF_FIXANGLES | DDPF_FIXPOS;
    mobj->player = p;
	mobj->dplayer = p->plr;
    mobj->health = plhealth.health;
	mobj->visangle = mobj->angle >> 16;

    p->plr->mo = mobj;
    p->playerstate = PST_LIVE;	
    p->refire = 0;
    p->hint_msg = NULL;
    p->damagecount = 0;
    p->bonuscount = 0;
    p->plr->extralight = 0;
    p->plr->fixedcolormap = 0;
	p->plr->lookdir = 0;
    p->plr->viewheight = PLR_VIEWHT;

    P_SetupPsprites(p);		// setup pistol psprite

	return mobj;
}

//================= R3D P_SpawnScriptThing =======================

static int		radius_limit;
static int		spawn_posx;
static int		spawn_posy;
static int		spawn_refh;
static int		check_flag;

static RDBOOL PIT_CheckSpot(mobj_t *thing, int *bbox)
{
    fixed_t		blockdist;

	if ( !(thing->flags & MF_SOLID)
		|| thing->flags3 & MF3_AMODEL
		|| thing->flags & MF_CORPSE
		|| thing->flags2 & MF2_CORPSEFADE
		|| thing->flags3 & check_flag/*
		|| P_IsCamera(thing)*/)
	{
		return true;
	}

	if (bbox)
	{
		sector_t *s1, *s2;
		int		 diff;

		// Where it is
		if (thing->z < spawn_refh && thing->z + thing->height > spawn_refh)
		{
			if (   thing->x + thing->radius > bbox[BOXLEFT ]
				&& thing->x - thing->radius < bbox[BOXRIGHT]
				&& thing->y + thing->radius > bbox[BOXBASE ]
				&& thing->y - thing->radius < bbox[BOXTOP  ])
			{
				return false;
			}
		}

		s1 = thing->subsector->sector;
		s2 = R_PointInSubsector(thing->x + thing->momx, thing->y + thing->momy)->sector;
		diff = s2->floorheight - s1->floorheight;

		// Where it could be
		if (thing->z + diff < spawn_refh && thing->z + diff + thing->height > spawn_refh)
		{
			if (   thing->x + thing->momx + thing->radius > bbox[BOXLEFT ]
				&& thing->x + thing->momx - thing->radius < bbox[BOXRIGHT]
				&& thing->y + thing->momy + thing->radius > bbox[BOXBASE ]
				&& thing->y + thing->momy - thing->radius < bbox[BOXTOP  ])
			{
				return false;
			}
		}

		return true;	// didn't hit it
	}

    blockdist = thing->radius + radius_limit;

    if (   (abs(thing->x - spawn_posx) < blockdist
		&&  abs(thing->y - spawn_posy) < blockdist)					// Where it is now
		|| (abs(thing->x + thing->momx - spawn_posx) < blockdist
		&&  abs(thing->y + thing->momy - spawn_posy) < blockdist) )	// Where it could be
	{
		return false;
	}

	return true;
}

// check if a spawn spot is occupied
RDBOOL spotIsOccupied(fixed_t x, fixed_t y, fixed_t z, fixed_t r, int *bbox, int flags)
{
    int			xl, xh, yl, yh;
    int			bx, by;
	int			box[4];
    	
	radius_limit = r + 8*FRACUNIT;

	M_setBoxOffsets(box, x, y, MAX_OF(MAXRADIUS, radius_limit) + MAXMOVE);
	M_getBlockOffsets(box, 0, &xl, &xh, &yl, &yh);

	spawn_posx = x;
	spawn_posy = y;
	spawn_refh = z;
	check_flag = flags;

    for (bx = xl; bx <= xh; bx++)
	{
		for (by = yl; by <= yh; by++)
		{
			if (!P_BlockThingsIterator(bx, by, (int (__cdecl *)(struct mobj_s *,void *))PIT_CheckSpot, bbox))
				return true;
		}
	}
    return false;
}

// v220-33 add support for spawning torches
void TS_CheckModelLink(mobj_t *iter, int sprnum);

RDBOOL P_SpawnScriptThing(spawnspot_t *sspot, int *monum, int tXXnum, RDBOOL chkspawn, RDBOOL fromscp)
{
	mobj_t		*mobj;
	int			x, y, z, type, res = 0, dmed_num;
	RDBOOL		chkmove = false;

	if (monum) *monum = 0;

	if (!sspot || (type = sspot->type) < 0 || type == MT_TELEPORTMAN) return false;

	if (nomonsters && (type == MT_SKULL || mobjinfo[type].flags & MF_COUNTKILL))
	{
		return false;
	}

	// If endlevel then the SM_NOSPAWN flag will have been set
	if (chkspawn && scpMessage.smflags & SM_NOSPAWN) return true;

	x = sspot->x << FRACBITS;
	y = sspot->y << FRACBITS;
	dmed_num = mobjinfo[type].doomednum;

	if (dmed_num >= 7000 && dmed_num < 7700)
	{
		if (sspot->mo) return false;	// Already set
	}
	else if (fromscp)
	{
		SR_SCPErr("Type spawned with %i is not an AMODEL", tXXnum);
	}
	else if (dmed_num <  7000
		||  dmed_num >= 17100
		|| (dmed_num >=  8000 && dmed_num < 16000))
	{
		if (mobjinfo[type].flags & MF_SOLID)
		{
			chkmove = chkspawn;
		}
	}
	else
	{
		Con_Error("SCRIPTED Thing type %i is invalid", type);
	}

	if (mobjinfo[type].flags & MF_SPAWNCEILING)
		z = ONCEILINGZ;
	else
		z = ONFLOORZ;
	
	norandomrotate = true;

	mobj = P_SpawnMobj(x, y, z, (mobjtype_t)type, dmed_num|(tXXnum<<16));	// spawn it

	norandomrotate = false;

	mobj->angle = ANG45 * (sspot->angle/45);
	mobj->visangle = mobj->angle >> 16; // "angle-servo"; smooth actor turning

	if (dmed_num >= 7000 && dmed_num < 7700)
	{
		mobj->flags3 |= MF3_AMODEL;

		if (mobj->flags3 & MF3_INTERREACT)
		{
			mobj->z = mobj->subsector->sector->floorheight + sspot->z_off * FRACUNIT;

			validcount++;

			MB_Update(mobj);

			if (spotIsOccupied(x, y, mobj->z, mobj->radius, MB_GetBox(mobj->modelboxnum - 1, NULL), 0) )
			{
				MB_Remove(mobj);
				P_RemoveMobj(mobj, 27);
				return false;
			}
		}
		sspot->mo = mobj;
	}

	if (chkmove)
	{
		validcount++;

		mobj->validcount = validcount;

		if (spotIsOccupied(x, y, 0, mobj->radius, NULL, mobj->flags3 & MF3_PAIREDSPAWN) )
		{
			P_RemoveMobj(mobj, 28);
			return false;
		}
	}

	if (mobj->flags & MF_COUNTKILL)
	{
		if (mobj->flags3 & MF3_FRIEND)	// R3D count friends
			totalfriends++;
		else
			totalkills++;
	}

	mobj->type = sspot->type;
	mobj->spawnpoint.x = sspot->x;
	mobj->spawnpoint.y = sspot->y;
	mobj->spawnpoint.doomednum = (short)dmed_num;
	mobj->spawnpoint.angle = sspot->angle;
	mobj->spawnpoint.options = sspot->options;

	if (mobj->tics > 0 && mobj->tics < 30 && !(mobj->flags3 & (MF3_AMODEL|MF3_STATIC)))
	{
		mobj->tics = 1 + (M_RandomD() % mobj->tics);
	}

	if (   (z = sspot->z_off)
		&& !(mobjinfo[type].flags & MF_SPAWNCEILING)
		&& ((mobj->flags & MF_SOLID && mobj->flags & MF_SHOOTABLE)
		||  mobj->flags & MF_SPECIAL) )
	{
		mobj->floorz = mobj->subsector->sector->floorheight;

		if (z > 0)
		{
			if ( !(mobjinfo[type].flags3 & MF3_INTERREACT) )
				mobj->z += (z * FRACUNIT);

			if (sspot->z_nofall || mobj->flags & MF_SPECIAL || !(mobj->flags & MF_SOLID))
			{
				mobj->floorz = mobj->dropoffz = mobj->z;

				if (sspot->z_nofall) mobj->flags3 |= (MF3_STATIC|MF3_NOROTATE);	// v220-10
			}
		}
		else if (mobj->flags & MF_SPECIAL)
		{
			if (z < -10) z = -10;

			mobj->z = mobj->floorz;
			mobj->misc |= ((-z) << NEGOFF_SHIFT);
		}
	}

	if (chkmove)
	{
		if (mobj->flags & MF_SHOOTABLE)
		{
			if (monum) *monum = mobj - mobjarray;

			if (sspot->options & MTF_AMBUSH)
				mobj->flags |= MF_AMBUSH;

			if (sspot->options & MTF_MULTIPLAYER)
			{
				mobj->flags2 |= MF2_NOSEESOUND;
			}

			/* not currently used
			if (sspot->sspflags & SSPF_FRIEND)
			{
				mobj->flags3 |= MF3_FRIEND;
				// change thread
				P_UpdateThinker(&mobj->thinker);
			}
			if (sspot->sspflags & SSPF_TOUCHY)
				mobj->flags3 |= MF3_TOUCHY;
			if (sspot->sspflags & SSPF_BOUNCES)
				mobj->flags3 |= MF3_BOUNCES;
			*/
		}

		if (sspot->spawn_fog)
		{
			S_StartSound(sfx_telept, P_SpawnMobj(mobj->x, mobj->y, mobj->z , MT_TFOG, 0));
		}
	}

	P_SetMpos(mobj);
	
	TS_CheckModelLink(mobj, mobj->sprite);
	
	CheckDesync(mobj); // v220-11

	return true;
}

// =============== End R3D P_SpawnScriptThing

// R3D routines used with P_SpawnMapThing

// R3D script error header
static void checkScriptedError()
{
	if (!(scriptedMapError & 1))
	{
		PrintProgInfo (
			"In General;\n"
			"  There must only be one instance in the map of any SCRIPT THING.\n"
			"  THING types 7700 to 7799 must always be on a line.\n\n"
			"ERRORS;\n\n");

		scriptedMapError |= 1;
	}
}

// R3D - get sector num Thing is on
static int getMapthingSectorNumber(short mx, short my)
{
	int	x = mx << FRACBITS, y = my << FRACBITS;

	return R_PointInSubsector(x, y)->sector - sectors;
}

// R3D - check if thing is to be spawned on 3d flat lifts or sectors
//       If 3dflats check the height is valid and whether it's a static or
//       lift flat as different rules apply (and the two types can be stacked)
static RDBOOL checkMapthing3DSector(short mx, short my, short z_off, int *pos, int *error)
{
	subsector_t	*ss = R_PointInSubsector((mx << FRACBITS), (my << FRACBITS));
	sector_t	*sc = ss->sector;

	*error = false;
	*pos = -1;
	int *flatslist = ss->flats3DArray;

	if (flatslist)
	{
		int c = *(int*)flatslist++;
		int s = sc - sectors;
		int z = ((int)z_off << FRACBITS) + sc->floorheight;
		int floor = sc->floorheight;
		int ceil = sectors[flatslist[1]].floorheight + flatslist[2];

		for (int i = 0; i <= c; i++)
		{
			if (z >= floor && z < ceil)
			{
				*pos = i;
				return (sectors[s].specFlags & SF_MOVFLOOR) != 0;
			}

			if (i == c) break;

			s = flatslist[1];
			floor = ceil + flatslist[3];

			if (i == c - 1)
			{
				ceil = sc->ceilingheight;
			}
			else
			{
				flatslist += FLIST_ENTRIES;
				ceil = sectors[flatslist[1]].floorheight + flatslist[2];
			}
		}

		*error = true;
		return false;
	}

	return sc->liftsec3D >= 0 || (sc->specFlags & SF_MOVFLOOR) != 0;
}

// R3D - get line Thing is on
static int getMapthingLineNumber(short mx, short my, int thingtype, int mapnum, RDBOOL error)
{
	int			x = mx << FRACBITS, y = my << FRACBITS;
	subsector_t	*ss = R_PointInSubsector(x, y);
	sector_t	*sc = ss->sector;
	seg_t		*sg = &segs[ss->firstline];
	line_t		*ld;
	int			i, lnum = -1;

	//------------------------

	validcount++;

	for (i = 0; i < ss->linecount; i++, sg++)
	{
		if ( (ld = sg->linedef) && ld->validcount != validcount)
		{
			ld->validcount = validcount;

			if (ld->frontsector != sc && ld->backsector != sc)
				continue;

			// b211-0 Eliminate getting wrong line where only one mapunit offset
			if (XYonLine(x, y, ld, FRACUNIT/2))
			{
				if (lnum < 0)
					lnum = ld - lines;
				else
				{
					PrintProgInfo("THING %i mapnum %i is on multiple lines [%i %i]\n"
						"Lines at x,y must have a min separation of 1 mapunit\n",
						thingtype, mapnum, lnum, ld - lines);
					break;
				}
			}
		}
	}

	if (lnum < 0 && error)
	{
		line_t	*ld;
		int		rnum = -1;

		for (i = 0, ld = lines; i < numlines; i++, ld++)
		{
			if (XYonLine(x, y, ld, FRACUNIT/2))
			{
				if (ld->frontsector == sc || ld->backsector == sc)
					return i;

				rnum = i;
			}
		}

		checkScriptedError();

		if (rnum >= 0)
		{
			sector_t *fs = lines[rnum].frontsector;
			sector_t *bs = lines[rnum].backsector;

			PrintProgInfo("THING %i mapnum %i was found on line %i\nThe line's sector", thingtype, mapnum, rnum);

			if (bs && bs != fs)
				PrintProgInfo("s are %i and %i", fs - sectors, bs - sectors);
			else
				PrintProgInfo(" is %i", fs - sectors);

			PrintProgInfo(" but R_PointInSubsector() returned sector %i\n", sc - sectors);
		}
		else
			PrintProgInfo("THING %i mapnum %i centre [x %i y %i] is not on a line in sector %i\n", thingtype, mapnum, mx, my, sc - sectors);

		PrintProgInfo("This indicates a nodes build or map design problem. If the former try building Normal and GL nodes\n");
	}

	return lnum;
}

// R3D get scripted spot
static int getSPOTnum(int thingtype, int *spottype)
{
	if (R3D_ScriptedMap)
	{
		*spottype = 6;

		if (thingtype >= 8000 && thingtype <= 8010) return thingtype;

		*spottype = 5;

		if (thingtype >= 7500 && thingtype <= 7599) return thingtype - 7500;

		*spottype = 4;

		if (thingtype >= 7400 && thingtype <= 7499) return thingtype - 7400;

		*spottype = 3;

		if (thingtype >= 7900 && thingtype <= 7999) return thingtype - 7900;
		if (thingtype >= 6900 && thingtype <= 6999) return thingtype - 6900 + 100;
		if (thingtype >= 5900 && thingtype <= 5999) return thingtype - 5900 + 200;
		if (thingtype >= 4900 && thingtype <= 4999) return thingtype - 4900 + 300;

		*spottype = 2;

		if (thingtype >= 7800 && thingtype <= 7899) return thingtype - 7800;
		if (thingtype >= 6800 && thingtype <= 6899) return thingtype - 6800 + 100;
		if (thingtype >= 5800 && thingtype <= 5899) return thingtype - 5800 + 200;
		if (thingtype >= 4800 && thingtype <= 4899) return thingtype - 4800 + 300;

		*spottype = 1;

		if (thingtype >= 7700 && thingtype <= 7799) return thingtype - 7700;
		if (thingtype >= 6700 && thingtype <= 6799) return thingtype - 6700 + 100;
		if (thingtype >= 5700 && thingtype <= 5799) return thingtype - 5700 + 200;
		if (thingtype >= 4700 && thingtype <= 4799) return thingtype - 4700 + 300;
	}

	*spottype = 0;

	return 0;
}

static int checkCoincident(mobj_t *mobj)
{
	thinker_t	*th;
	mobj_t		*mo;

	for (th = thinkercap.next; th != &thinkercap; th = th->next)
	{
		if (   th->function == (void (__cdecl *)(void))P_MobjThinker
			&& (mo = (mobj_t*)th ) != mobj
			&& mo->flags3 & MF3_AMODEL
			&& mo->type == mobj->type
			&& mo->x == mobj->x
			&& mo->y == mobj->y
			&& mo->z == mobj->z)
		{
			int mapnum = MAPNUM(mo);
			if (spawnArray[mapnum]) return mapnum;
		}
	}

	return -1;
}

void Sys_GeneralMessage(char *msg);
void SCP_ZoffWarn(int mapnum, int thingtype);

//
// P_SpawnMapThing
// The fields of the mapthing should
// already be in host byte order.
//
void P_SpawnMapThing(mapthing_t* mthing, int mapnum, int *typecheck, int codflags, RDBOOL is_iwad)
{
	int			mt_type;
	int			spotnum, spottype;
	int			set_flags3 = 0;
	mobj_t		*mobj;
	fixed_t		x;
	fixed_t		y;
	fixed_t		z;
	fixed_t		z_offset = 0;
	short		scpllevel = -1;
	short		scpzoff_ptr = 0;

	int			iden_num = 0;
	int			thingtype = (ushort) mthing->doomednum;
	int			flat3Dpos = -1;
	int			checkMonster = 0;
	RDBOOL		zoff16xxx = false;

    // discard deathmatch and/or extended co-op start positions
    if (    thingtype == 11
		|| (thingtype >= 4001 && thingtype <= 4004))
	{
		// save the co-ords for analysis purposes
		if (dmcount < DM_MAXCOUNT)
		{
			dmarray[dmcount].x = (int)((short)mthing->x);
			dmarray[dmcount].y = (int)((short)mthing->y);
			dmcount++;
		}
		return;
	}

    // Check for players specially.
	if (thingtype >= 1 && thingtype <= 4)
    {
		P_RegisterPlayerStart(mthing, mapnum);
		return;
    }

	// ================ R3D check thing types and/or spot nums

	if (thingtype >= 15500 && thingtype <= 15550)
		return;

	if (!R3D_lightextns)
	{
		if (thingtype >= 5500 && thingtype <= 5505) return;
	}

	if (R3D_scpzoff)
	{
		if (thingtype >= 16000 && thingtype <= 16899)
		{
			scpzoff_t *sc = &scpzoff[thingtype - 16000];

			if (sc->thingnum)
			{
				// v220-31 allow active monsters to be spawned
				int		err, check = P_GetTypeFromDoomednum(sc->thingnum);

				if (check == stats.num_mobjs)
				{
					Con_Error("Unrecognised type %i set in SCP_ZOFF script\n", sc->thingnum);
				}

				checkMonster = (check == MT_SKULL || mobjinfo[check].flags & MF_COUNTKILL);

				thingtype = sc->thingnum;
				scpllevel = sc->lightlevel;

				if (sc->intensity)
					scpzoff_ptr = (short)(thingtype - 16000 + 1);

				mthing->doomednum = thingtype;

				if (!checkMonster)
				{
					set_flags3 |= MF3_NOROTATE;

					if (R3D_ScriptedMap && sc->zoflags & ZO_OFFSET5)
					{
						set_flags3 |= MF3_OFFSET5;
					}
				}

				if ( (z_offset = sc->z_off) != 0)
				{
					RDBOOL	ismov = checkMapthing3DSector(mthing->x, mthing->y, z_offset, &flat3Dpos, &err);

					if (err)
					{
						pr_OFinsert("WARNING: T%i type %i set to invalid height %i\n"
							"with a 3D floor in SCP_ZOFF script\n", mapnum, thingtype, z_offset);
					}

					if (!ismov) zoff16xxx = true;
				}

				// NOTE: if (z_offset == -32000)	// Put a thing under a 3D line only lift
				//       do nothing further down. Height will be set later by LS_CheckThingsHeight
			}
			else
			{
				SCP_ZoffWarn(mapnum, thingtype);
				// Allow spawning as condition may be intentional
			}
		}
		else
		if (thingtype >= 16900 && thingtype <= 16999)
		{
			int soundid = scplinesound[thingtype - 16900][0];

			if (soundid)
			{
				int lnum = getMapthingLineNumber(mthing->x, mthing->y, thingtype, mapnum, true);

				if (lnum >= 0)
				{
					// v220-31 walkover once sound for standard map type
					if (lines[lnum].special == 0x1FCC || lines[lnum].special == 0x1FCD)
					{
						lines[lnum].linesound = soundid;
					}
					else
					{
						byte bvnum = (byte)scplinesound[thingtype - 16900][1];

						if (bvnum == 100 || bvnum == 101)
						{
							short n = identifySpecial(lines + lnum);

							if (n && n != 5)	// use types return 0 and 5
							{
								Con_Error("SETLINETRIGGERSOUND error: L%i does not have a use trigger type", lnum);
							}
							lines[lnum].linesound = soundid;
							byte val =  (bvnum == 100 ? XT_PLAYLSONCE : XT_PLAYLSREPEAT);
							lines[lnum].xtraflags |= val;
						}
						else if (bvnum == 102)
						{
							short chk = identifySpecial(lines + lnum);
							if (chk < 1 || chk > 4)	// walkover types return 1,2,3 and 4
							{
								Con_Error("SETLINESOUNDOFF error: L%i does not have a walkover trigger type", lnum);
							}
							lines[lnum].linesound = soundid;
							lines[lnum].xtraflags |= XT_TURNLSOFF;
						}
						else
						{
							if (lines[lnum].special == 8144)
								lines[lnum].lnflags2 |= LNF2_LINESOUND;
							else if (lines[lnum].special != 8093)
								lines[lnum].lnflags2 |= (LNF2_LINESOUND|LNF2_BLOCKALL);

							if (!lines[lnum].special)
								lines[lnum].special = 8106;

							lines[lnum].linesound = soundid;
							lines[lnum].bv = bvnum;
						}
					}
				}
			}
			else
			{
				SCP_ZoffWarn(mapnum, thingtype);
			}

			return;	// Type not spawned - info only
		}

		if (thingtype >= 17000 && thingtype <= 17099)
		{
			int num = thingtype - 17000;

			if (!R3D_linemap)
			{
				Con_Error ("ERROR: T%i type %i invalid.\n"
							"Map not defined as a 3D linetype\n"
							"Thing type 15501 or 15504 required in map\n"
							"See Risen3D\\R3D_Docs\\Editing\\R3D_Types.txt\n",
							mapnum, thingtype);
			}

			if (pseudofake[num][PF_TAG])	// refsec needs to be set
			{
				pseudofake[num][PF_REFSEC] = getMapthingSectorNumber(mthing->x, mthing->y);
				if (sectors[pseudofake[num][PF_REFSEC]].heightsec >= 0)
					Con_Error("T%i dmednum %i is referencing S%i already tagged to type 242",
							mapnum, thingtype, pseudofake[num][PF_REFSEC]);
				scpofftagsset = true;
			}

			if (elevspeed[num][ELV_SPEED])	// if speed has been set
			{
				if (!elevspeed[num][ELV_TAG])
				{
					int scn = getMapthingSectorNumber(mthing->x, mthing->y);

					if ( !(sectors[scn].specFlags2 & SF2_ELEVATOR3D) || sectors[scn].liftsec3D < 0 )
						Con_Error("T%i dmednum %i is not referencing a 3D moving floor's control sector", mapnum, thingtype);
					
					elevspeed[num][ELV_SECN] = scn;
					elevspeedset = true;
				}
				else
				{
					int sn, t = elevspeed[num][ELV_TAG];
					// Note: ELV_HILIMT at this point just has the number of steps up
					__int64	max_amount = (__int64)elevspeed[num][ELV_HILIMIT] * (__int64)elevspeed[num][ELV_HEIGHT];
					__int64 fh = (__int64)P_GetHighestCompositeFloor(t) + max_amount;

					if (fh > (__int64)DDMAXINT)
						Con_Error("SETELEVSPEEDHEIGHT: n-up max value of %i if added in group tagged %i\n"
								  "would cause an overflow error", (int)(max_amount>>16), t);

					if ( (sn = P_CheckCompositeCeilings(t, (fixed_t)max_amount)) >= 0)
					{
						Con_Error("SETELEVSPEEDHEIGHT: n-up max value of %i if added would send S%i's floor\n"
								  "in group tagged %i through its ceiling. This is not allowed.", (int)(max_amount>>16), sn, t);
					}

					elevspeed[num][ELV_SECN] = -1;	// sectors will be established from tags
					elevspeed[num][ELV_HILIMIT] = (fixed_t)fh;

					// Note: ELV_LOLIMT at this point just has the number of steps down
					max_amount = (__int64)elevspeed[num][ELV_LOLIMIT] * (__int64)elevspeed[num][ELV_HEIGHT];
					fh = (__int64)P_GetLowestCompositeFloor(t) - max_amount;

					if (fh < (__int64)DDMININT)
					{
						Con_Error("SETELEVSPEEDHEIGHT: n-down max value of %i if deducted would\n"
								  "in group tagged %i cause an underflow error of < -32768", (int)(max_amount>>16), t);
					}

					elevspeed[num][ELV_LOLIMIT] = (fixed_t)fh;
					elevheightset = true;
				}
			}
			else if (!pseudofake[num][PF_TAG])
			{
				SCP_ZoffWarn(mapnum, thingtype);
			}

			return;	// Type not spawned - info only
		}
	}

	// v220-31
	if (R3D_maptype)
	{
		if (thingtype >= 18000 && thingtype <= 18006)
		{
			int lnum = getMapthingLineNumber(mthing->x, mthing->y, thingtype, mapnum, true);

			if (lnum >= 0)
			{
				short chk = identifySpecial(lines + lnum);

				if ( !(chk == 2 || chk == 3 || chk == 5) )	// walkover tp types return 2 and 3 and tp use 5
				{
					Con_ErrorDev("T%i error: L%i is not a teleport trigger type", thingtype, lnum);
				}

				if (thingtype == 18006)
				{
					lines[lnum].xtraflags |= XT_PLRTRIGONLY;
				}
				else	// can only be 18000 to 18005
				{
					lines[lnum].xtraflags &= ~XT_LIGHTFLAGS;
					lines[lnum].xtraflags |= XT_SETLIGHT;

					if (thingtype == 18000) { /* do nothing */ }
					else
					if (thingtype == 18001) lines[lnum].xtraflags |= XT_TURNSINEON;
					else
					if (thingtype == 18002) lines[lnum].xtraflags |= XT_TURNVTXON;
					else
					if (thingtype == 18003) lines[lnum].xtraflags |= (XT_TURNVTXON|XT_TURNSINEON);
					else
					if (thingtype == 18004) lines[lnum].xtraflags |= XT_TURNTORCHON;
					else
					if (thingtype == 18005) lines[lnum].xtraflags |= (XT_TURNTORCHON|XT_TURNSINEON);
				}
			}
			else
			{
				Con_ErrorDev("T%i error: Thing is not on a line", thingtype);
			}

			return;	// Type not spawned - info only
		}
		else if (thingtype >= 18010 && thingtype <= 18310)
		{
			int nnum = (thingtype - 18010) / 4;
			int nptr = (thingtype - 18010) % 4;
			int lnum = getMapthingLineNumber(mthing->x, mthing->y, thingtype, mapnum, true);
			int sp = lines[lnum].special;

			if (!ma_inf.report18010)
			{
				ma_inf.report18010 = true;
				if (ge_inf.devmap)
				{
					pr_OFinsert("\n============ 18010 - 18310 report ==============\n");
					pr_OFinsert("Thing    div mod  lnum      L0    L1    L2    L3\t\n");
				}
			}

			if (!nptr)
			{
				if (sp != 8142 && sp != 8143) Con_Error("The line %i with Model link type %i must have its special set to 8142 or 8143", lnum, thingtype);
				if (lines[lnum].tag) Con_Error("The line %i with Model link type %i must not be tagged", lnum, thingtype);
			}

			if (nptr && (sp == 8142 || sp == 8143)) Con_Error("The line %i with linked trigger type %i must *not* have its special set to 8142 or 8143", thingtype, lnum);

			ma_inf.ambtrigger = true;

			if (nptr == 0) { if (ltarray[nnum].line0 < 0) { ltarray[nnum].line0 = lnum; lines[lnum].xtraflags |= XT_TRIGGER; } else goto lterror; }
			else
			if (nptr == 1) { if (ltarray[nnum].line1 < 0) ltarray[nnum].line1 = lnum; else goto lterror; }
			else
			if (nptr == 2) { if (ltarray[nnum].line2 < 0) ltarray[nnum].line2 = lnum; else goto lterror; }
			else
			if (nptr == 3) ltarray[nnum].line3 = lnum;	// trigger in other group if needed

			if (ge_inf.devmap)
				pr_OFinsert(" %6i%6i%6i%6i\n",
					ltarray[nnum].line0, ltarray[nnum].line1, ltarray[nnum].line2, ltarray[nnum].line3);

			return;	// Type not spawned - info only
lterror:;
		Con_Error("T%i already exists\n", thingtype);
		}
	}

	// v220-4 scripted map 73xx types.  Info types only
	if (R3D_ScriptedMap)
	{
		if (thingtype >= 7300 && thingtype <= 7309)
		{
			int dmednum[] = {7, 16, 67, 68, 3003, 6085, 6086, 6087, 6088, 6089};
			int tnum = thingtype - 7300;
			int lnum = getMapthingLineNumber(mthing->x, mthing->y, thingtype, mapnum, true);

			if (deathmonitor[tnum].linenum) Con_Error("TYPE %i has already been set", thingtype);
			deathmonitor[tnum].linenum = &lines[lnum];
			deathmonitor[tnum].type = P_GetTypeFromDoomednum(dmednum[tnum]);

			return;
		}

		if (thingtype == 7310)
		{
			int  x = mthing->x;
			int  y = mthing->y;
			sector_t *s = R_PointInSubsector(x << FRACBITS, y << FRACBITS)->sector;

			s->specFlags2 |= SF2_NOMOVSOUND;

			if (ge_inf.devmap)
				pr_OFinsert("7310 type in sec %5i\n", s - sectors);

			return;
		}
	}

	// scripted spot types
	spotnum = getSPOTnum(thingtype, &spottype);

	if (spottype)
	{
		if (R3D_edit) return;

		if (spottype == 1)	// Link line spots
		{
			th7700_t *t77 = &th7700[spotnum];
			int		 lnum = getMapthingLineNumber(mthing->x, mthing->y, thingtype, mapnum, true);
			short	 options = mthing->options;

			if (lnum < 0) return;

			if (t77->spawn_count >= t77->maxspawn_count)
			{
				t77->maxspawn_count += 4;
				t77->info = (t77info_t*)W_ReallocCLR(t77->info, t77->maxspawn_count * sizeof(*t77->info), LEVELID, 64);
			}

			memcpy(&t77->info[t77->spawn_count], mthing, sizeof(mapthing_t));

			if (options & MTF_EASY)
			{
				// v220-3 modified
				if (options & MTF_ALIVE)
					lines[lnum].ml_flags |= ML_BLOCKING;
				else
					lines[lnum].lnflags2 |= LNF2_BLOCKALL;

				t77->blocked = true;
			}

			// 77nn things are directly linked to 79nn things by nn
			if (options & MTF_HARD)
				th7900[spotnum].t79flags |= T79_TRIGGER;
				
			if (   options & MTF_NORMAL
				&& (options & (MTF_NORMAL|MTF_MULTIPLAYER)) != (MTF_NORMAL|MTF_MULTIPLAYER))
			{
				t77->info[t77->spawn_count].options |= MTF_NOKILL;
			}
			
			t77->info[t77->spawn_count  ].t77lnum = lnum;
			t77->info[t77->spawn_count++].t77special = lines[lnum].special;

			ma_inf.scriptthings77++;
		}

		else if (spottype == 2)	// Spawn spots
		{
			th7800_t *t78 = &th7800[spotnum];

			if (t78->setflags & T_SET)
			{
				checkScriptedError();

				t78->setflags |= T_MULTISET;
			}
			else
			{
				t78->setflags |= T_SET;

				memcpy(t78, mthing, sizeof(mapthing_t));

				ma_inf.scriptthings78++;
			}

			// May or may not be on a line
			t78->refline78 = getMapthingLineNumber(mthing->x, mthing->y, thingtype, mapnum, false);
		}

		else if (spottype == 3)	// Script spots
		{
			th7900_t *t79 = &th7900[spotnum];

			if (t79->setflags & T_SET)
			{
				checkScriptedError();

				t79->setflags |= T_MULTISET;
			}
			else
			{
				t79->setflags |= T_SET;
				memcpy(t79, mthing, sizeof(mapthing_t));
				ma_inf.scriptthings79++;
			}
		}

		else if (spottype == 4)	// 7400 spots
		{
			th7400_t *t74 = &th7400[spotnum];

			if (t74->refline)
			{
				checkScriptedError();
				PrintProgInfo("Multiple instance of THING 74%i\n", spotnum);
			}
			else
			{
				int lnum = getMapthingLineNumber(mthing->x, mthing->y, thingtype, mapnum, true);

				if (lnum >= 0) t74->refline = &lines[lnum];
			}
		}

		else if (spottype == 5)	// 7500 special line spots
		{
			th7400_t *t74 = &th7400[spotnum];

			if (t74->ptrnum < 10)
			{
				int lnum = getMapthingLineNumber(mthing->x, mthing->y, thingtype, mapnum, true);

				if (lnum >= 0) t74->lptr[t74->ptrnum++] = &lines[lnum];
			}
			else
			{
				PrintProgInfo("Max instance of type 75%i exceeded. Max allowed is 10\n", spotnum);
			}
		}
		else	// Door message spot
		{
			int lnum = getMapthingLineNumber(mthing->x, mthing->y, thingtype, mapnum, true);

			if (lnum >= 0)
			{
				if (maxnumdoorspots <= numdoorspots)
				{
					maxnumdoorspots += 32;
					doorspotlnums = (int*) W_Realloc(doorspotlnums, maxnumdoorspots*sizeof(int), STATICID, 0);
				}

				if (thingtype >= 8001) lnum |= ((thingtype - 8000) << 28);

				doorspotlnums[numdoorspots++] = lnum;

			}
		}

		return;	// Information types are not spawned
	}

	if (R3D_ScriptedMap)
	{
		// Things to do with sound
		if (thingtype == 15600)
		{
			subsector_t *ss = R_PointInSubsector(mthing->x<<FRACBITS,mthing->y<<FRACBITS);
			int secn = ss->sector - sectors;

			if (!linkSectors)
				linkSectors = (byte*)W_Calloc(numsectors, LEVELID, 65);
			if (linkSectors[secn])
				Con_Message("WARNING: More than one Thing 15600 in sector %i\n", secn);
			linkSectors[secn] = true;

			return;	// Information only - not spawned
		}
		
		if (thingtype >= 20000 && thingtype <= 20127)
		{
			sector_t *s1, *s2;
			int		 rvptr = thingtype - 20000;
			int		 lnum = getMapthingLineNumber(mthing->x, mthing->y, thingtype, mapnum, false);
			int		 newcount;
			int		 i, k, n = -1;

			if (lnum >= 0)
			{
				s1 = lines[lnum].frontsector;
				s2 = lines[lnum].backsector;

				if (s1 == s2) s2 = NULL;
			}
			else
			{
				s1 = (R_PointInSubsector(mthing->x << FRACBITS, mthing->y << FRACBITS))->sector;
				s2 = NULL;
			}

			for (i = 0; i < MAX_REVERBTHING; i++)
			{
				for (k = 0; k < reverbThing[i].seccount; k++)
				{
					if (s1 && reverbThing[i].secnum[k] == s1 - sectors) { s1 = NULL; n = i; }
					if (s2 && reverbThing[i].secnum[k] == s2 - sectors) { s2 = NULL; n = i; }
				}
				if (!s1 && !s2) break;
			}

			if (!s1 && s2) { s1 = s2; s2 = NULL; }

			if (!s1 && !s2)
			{
				if (n < 0) Con_Error("Program error: ReverbTHING T%i", mapnum);

				Con_Message("WARNING: ReverbTHING T%i (type %i) sector already referenced by type %i and has been ignored\n",
					mapnum, thingtype, n + 20000);

				return;
			}

			ma_inf.reverbthings++;

			newcount = s2 ? 2 : 1;

			if (reverbThing[rvptr].seccount + newcount >= reverbThing[rvptr].maxseccount)
			{
				reverbThing[rvptr].maxseccount += 32;

				reverbThing[rvptr].secnum = (int*)W_Realloc(reverbThing[rvptr].secnum,
						sizeof(int*) * reverbThing[rvptr].maxseccount, LEVELID, 66);
			}

			reverbThing[rvptr].num = thingtype;
			reverbThing[rvptr].thcount++;
			reverbThing[rvptr].secnum[reverbThing[rvptr].seccount++] = s1 - sectors;

			if (s2)
				reverbThing[rvptr].secnum[reverbThing[rvptr].seccount++] = s2 - sectors;

			return;
		}
	}

	if (thingtype >= 6000 && thingtype <= 6099)
	{
		if (R3D_ScriptedMap) iden_num = thingtype;	// model that can drop any item
	}
	else if (thingtype >= 6100 && thingtype <= 6199)
	{
		if (R3D_ScriptedMap)
		{
			int n = thingtype - 6100;

			if (fakeidentifier[n] != -1)
			{
				PrintProgInfo ("ERROR: Multiple instance of THING type %i\n", thingtype);

				scriptedMapError |= 2;
				return;
			}

			fakeidentifier[n] = R_PointInSubsector(mthing->x << FRACBITS, mthing->y << FRACBITS)->sector - sectors;
			return;	// info only
		}
	}
	else if (thingtype >= 7000 && thingtype < 7700)
	{
		if (R3D_maptype)
		{
			if (!loadModels || R3D_edit) return;	// No sprite equivalent or not drawn in edit mode

			ma_inf.architmodels++;
			iden_num = thingtype;
			set_flags3 = MF3_AMODEL;
		}
	}
	else if (ge_inf.addon_flags & AF_COD
		&& thingtype >= 4988 && thingtype <= 4996)
	{
		iden_num = thingtype - 4988;		// COD sound hack
		thingtype = 14064;					// MT_SOUNDSOURCE
		set_flags3 = MF3_STATIC;
	}

    // find which type to spawn
	else if (thingtype >= 13901 && thingtype <= 13964)
	{
		if (R3D_maptype)
		{
			iden_num = thingtype - 13901;
			thingtype = 13964;				// MT_TRIGAMBIENT
			set_flags3 = MF3_STATIC;
		}
	}

	else if (thingtype >= 14001 && thingtype <= 14064)
	{
		if (R3D_edit) return;
		iden_num = thingtype - 14001;
		thingtype = 14064;					// MT_SOUNDSOURCE
		set_flags3 = MF3_STATIC;
		ambientfound++;
		ambientcount[iden_num]++;
	}

	else if (thingtype >= 14101 && thingtype <= 14164)
	{
		if (R3D_edit) return;
		// Use the ambient number
		iden_num = thingtype - 14100;	// Mus change
		thingtype = 14164;				// MT_MUSICSOURCE
		musChange = true;
		set_flags3 = MF3_STATIC;
	}

	else if (thingtype >= 14200 && thingtype <= 14299)
	{
		iden_num = thingtype - 14200;
		thingtype = 14200;				// MT_TPUNDERFAKE
		set_flags3 = MF3_STATIC;
	}

	else if (thingtype >= 14300 && thingtype <= 14399)
	{
		iden_num = thingtype - 14300;
		thingtype = 14300;				// MT_TPOVERFAKE
		set_flags3 = MF3_STATIC;
	}

	else if (thingtype == 15551)		// MT_FAKESOUNDSOURCE
	{
		ma_inf.hasfakewater = true;
		set_flags3 = MF3_STATIC;
	}

	// ================ End R3D check thing types and/or spot nums

	// Don't spawn things flagged for Multiplayer
	if (!set_flags3 && mthing->options & MTF_MULTIPLAYER)
	{
		// v220-5 TNT iwad yellow key map31 fix
		if (   gamemission == pack_tnt
			&& is_iwad
			&& thingtype == 6
			&& !strcmp(P_GetCurrentMapid(), "map31"))
		{
			mthing->options &= ~MTF_MULTIPLAYER;
		}
		else
		{
			return;
		}
	}

    // b211-0 Translate doomednum to mt_type
	//     Check for Doombuilder id
	if ( (mt_type = P_GetTypeFromDoomednum(thingtype)) == stats.num_mobjs)
	{
		if (thingtype == 32000)
			ma_inf.doombuilder = true;
		else
			*typecheck = thingtype;	// unrecognised type
		return;
	}

	// noobjects should only be set if in edit mode but filter anyway
	if (R3D_edit && ge_inf.noobjects && mt_type != MT_TELEPORTMAN) return;

	spawnArray[mapnum] = true;

	if (mobjinfo[mt_type].flags3 & MF3_DEHAMBIENT)
	{
		if (R3D_edit) return;
	}
	else if (!R3D_lightextns && thingtype >= 5500 && thingtype <= 5505)
	{
		return;	// Not for this map
	}
	else if (thingtype == 5000 || (thingtype >= 5500 && thingtype <= 5505))
	{
		if ( !(mthing->options & (MTF_EASY|MTF_NORMAL|MTF_HARD)) ) return;	// Will do nothing
	}
	else if (!set_flags3 || checkMonster)	// R3D - In case of ded flag error unless known to be a monster
	{
		// check for appropriate skill level
		if (!(mthing->options & ge_inf.skill))
		{
			// Any set will be removed following analysis.
			spawnArray[mapnum] = false;	// Remove this.
		}

		// R3D - defer removal of any monsters even if -nomonsters
		// They are first spawned in case they are needed for analysis checks
		// The map may be loaded at later time with -nomonsters unset and this
		// prevents having to re-analyse the map.
		// If not required then they will be removed before the game starts
		else if (nomonsters
			&& ( mt_type == MT_SKULL || (mobjinfo[mt_type].flags & MF_COUNTKILL)) )
		{
			if (R3D_edit) return;	// analysis file not stored and Thing not wanted in edit mode
			spawnArray[mapnum] = false;	// Remove this.
		}
	}

	x = mthing->x << FRACBITS;
	y = mthing->y << FRACBITS;

	if (mobjinfo[mt_type].flags & MF_SPAWNCEILING)
		z = ONCEILINGZ;
	else
		z = ONFLOORZ;
	
	norandomrotate = true;

	// SPAWN IT ================================================

	mobj = P_SpawnMobj(x, y, z, (mobjtype_t)mt_type, iden_num|(mapnum<<16));	// spawn it

	// =========================================================

	norandomrotate = false;

	// v220-31
	if (thingtype == 5000 || (thingtype >= 5500 && thingtype <= 5505) )
	{
		mobj->inverseAlpha = 0;	// Use this as mobj is never rendered (so is unused)

		// dynlights use difficulty settings to set the colour
		if (mthing->options & MTF_EASY  ) mobj->inverseAlpha |= 0x00FF0000;	// red
		if (mthing->options & MTF_NORMAL) mobj->inverseAlpha |= 0x0000FF00;	// green
		if (mthing->options & MTF_HARD  ) mobj->inverseAlpha |= 0x000000FF;	// blue

		if (!useDynLights && CFX && !forceDynLights)
		{
			forceDynLights = true;
			pr_OFinsert("ADVISORY: DYN LIGHTS DETECTED - rend-light 0 has been forced to rend-light 1\n");
		}
	}

	// v220-31 if MF3_NOROTATE is set from a 16xxx Thing with an item that has
	//         a lod model with no spin then this can upset the lod code.  As
	//         such it is cleared here if set with any model that does not spin.
	if (set_flags3 & MF3_NOROTATE)
	{
		int			numf;

		// modified $v2.0.11
		if (  !(numf = sprites[mobj->sprite].numframes)
			|| (mobj->frame & FF_FRAMEMASK) >= numf )
		{
			pr_OFinsert("Sprite error: T%i has been ignored\n", mapnum);

			spawnArray[mapnum] = false;	// Remove this.
			return;
		}

		// Check for a model.
		if (CFX && !(sprites[mobj->sprite].sprflags & SPR_NOMODEL))
		{
			modeldef_t	*mf = NULL, *nextmf = NULL;

			R_CheckModelFor(mobj, &mf, &nextmf);
			
			if (mf && !(mf->sub[0].flags & MODEL_SPINMASK))
			{
				set_flags3 &= ~MF3_NOROTATE;
			}
		}
	}

	if (set_flags3) mobj->flags3 |= set_flags3;	// Architectural/static models/script offset/rotate

	if (!mobj->scplightlevel && scpllevel > 0)	// R3D: if not set in the deds..
		mobj->scplightlevel = scpllevel;		// ..diminished light fix (set in SCP_ZOFF)

	// if > 0 then scpzoff_t *scpzoff[scpzoff_ptr - 1].intensity is used
	// (other scpzoff_t entries are dealt with elsewhere)
	mobj->scpzoff_ptr = scpzoff_ptr;

	if (z_offset && mobj->flags & MF_SPAWNCEILING)
	{
		mobj->z = mobj->dropoffz = mobj->floorz = mobj->subsector->sector->floorheight + (z_offset<<FRACBITS);
		mobj->flags2 |= MF2_HEIGHTSET;
	}
	else
	if (z_offset && flat3Dpos == -1)
	{
		if (z_offset == -32000)
		{
			mobj->spawnz = z_offset; // heights reset in LS_CheckThingsHeight
		}
		else
		if (z_offset > 0)
		{
			// R3D - use SL_PointHeight
			fixed_t ht = SL_PointHeight(mobj->subsector->sector, mobj->x, mobj->y, LOSLOPE);
			mobj->z = mobj->dropoffz = mobj->floorz = ht + (z_offset<<FRACBITS);
		}
		else if (mobj->flags & MF_SPECIAL)
		{
			if (z_offset < -10) z_offset = -10;
			mobj->misc |= ((-z_offset) << NEGOFF_SHIFT);
		}
		else
		{
			pr_OFinsert("SCP_ZOFF error with T%i: negative offsets can only be used with Things flagged mf_special\n", mapnum);
		}
	}
	else if (z_offset > 0 && z == ONFLOORZ)
	{
		fixed_t zh = mobj->subsector->sector->floorheight + (z_offset << FRACBITS);

		if (flat3Dpos < 0 && mobj->subsector->sector->liftsec3D >= 0)
		{
			mobj->z = mobj->floorz = mobj->dropoffz = zh;
			mobj->flags2 |= MF2_HEIGHTSET;
			// v220-31 spawnz is needed if mobj is respawned (e.g. nightmare mode)
			mobj->spawnz = z_offset;
		}
		else
		// Note: flat3Dpos is only set with 3D flats
		if (flat3Dpos >= 0)
		{
			mobj->flags2 |= MF2_HEIGHTSET;
			mobj->z = mobj->floorz = mobj->dropoffz = zh;
			
			mobj->flat3Dlevel[0] = mobj->flat3Dlevel[1] = flat3Dpos;
			
			if (ge_inf.devmap)
				pr_OFinsert("3D floor ========= T%i level %i\n", mapnum, mobj->flat3Dlevel[1]);
		}
	}

	if (zoff16xxx && !mobj->info->seestate)
	{
		mobj->flags2 |= MF2_NOCHECK;
	}

	// R3D - prevent lights being drawn with custom dehacked things
	//       Lighting can be unacceptable with unknown sprites
	if (mobj->sprite > 0 && mobj->sprite < numsprites)
	{
		if (   hasDEHactors
			&& mt_type < DH_NUMOBJECTS
			&& sprites[mobj->sprite].sprflags & SPR_NOMODEL)
		{
			mobj->flags2 |= MF2_NOTLIT;
		}

		// v220-3 align any sprite with no rotations,
		//        unless a missile, to the viewplane
		// v220-4 removed
	}

	memcpy(&mobj->spawnpoint, mthing, sizeof(mapthing_t));

	if (mobj->tics > 0 && !(mobj->flags3 & (MF3_AMODEL|MF3_STATIC)))
	{
		mobj->tics = 1 + (M_RandomD() % mobj->tics);
	}

	if (ge_inf.MBFfriendallow && mthing->options & MTF_FRIEND)
	{
		ma_inf.friendsfound++;
		mobj->flags3 |= MF3_FRIEND;
		ge_inf.isfriendly = true;
		P_UpdateThinker(&mobj->thinker);	// R3D - put friends on correct thread
	}

	// Increment totals only if not to be removed
	if (spawnArray[mapnum])
	{
		if (mobj->flags & MF_COUNTKILL)
		{
			if (mobj->flags3 & MF3_FRIEND)	// R3D count friends
				totalfriends++;
			else
				totalkills++;
		}

		if (mobj->flags & MF_COUNTITEM)
			totalitems++;
	}

	if (mobj->flags2 & MF2_RANDOM)
	{
		mobj->angle = ANG45 * (rand() % 8);
		mobj->startangle = 0;
	}
	else
		mobj->angle = ANG45 * (mthing->angle/45);

	mobj->visangle = mobj->angle >> 16; // "angle-servo"; smooth actor turning

	if (mthing->options & MTF_AMBUSH)
	{
		if (R3D_Secret_Pickups && mobj->flags & MF_SPECIAL)
		{
			mobj->flags3 |= MF3_SECRET;
			totalsecret++;
			if (ge_inf.scp_devpu)
				Con_Message("SECRET PICKUP mapnum %i\n", MAPNUM(mobj));
		}	
		else
		{
			mobj->flags |= MF_AMBUSH;
		}
	}

	if (ge_inf.addon_flags & AF_COD)
	{
		// codflag mods
		if (codflags == 1) mobj->flags4 |= MF4_LOSTSOUL;
		else if (codflags == 2)
		{
			mobj->health = 1500;
			mobj->flags4 |= MF4_BOSSSPID;
		}
		else if (codflags == 3) mobj->flags4 |= MF4_CODENDTYPE;
		else if (codflags == 4)
		{
			P_SetMobjState(mobj, mobj->info->deathstate);
			mobj->height = 16*FRACUNIT;
			mobj->radius = mobj->info->radius;
			mobj->health = 5;
			mobj->flags3 |= MF3_RESURRECT;
		}
		else if (codflags == 5)
		{
			// hardasszombie
			mobj->health = 80;
			mobj->flags4 |= MF4_BADASSZOMBIE;
		}
		else if (codflags == 6)
		{
			P_UnlinkThing(mobj);
			mobj->flags |= MF_NOSECTOR;
			P_SetThingPosition(mobj);
		}
		else if (codflags == 7)
		{
			mobj->flags4 |= MF4_VILESPAWNER;
			mobj->flags |= MF_SHADOW;
			mobj->ddflags |= DDMF_SHADOW;
			mobj->flags2 |= MF2_NOSHADOW;
		}
		else if (codflags == 8)
		{
			mobj->health = 4000;
		}
		else if (codflags == 8)
		{
			mobj->health = 4000;
		}
	}

	P_SetMpos(mobj);
	
	CheckDesync(mobj); // v220-11
}



//
// GAME SPAWN FUNCTIONS
//


//===========================================================================
// P_SpawnPuff
//===========================================================================
void P_SpawnPuff(fixed_t x, fixed_t y, fixed_t z, RDBOOL ricoh)
{
	static void *trigptr = NULL;

	if (R3D_edit) return;

	mobj_t *th;

	z += ((M_RandomD()-M_RandomD()) << 10);
	th = P_SpawnMobj(x, y, z, MT_PUFF, 0);
	th->momz = FRACUNIT;
	th->tics -= (M_RandomD() & 3);	
	if (th->tics < 1) th->tics = 1;

	P_SetMpos(th);

	if (r_use_particles)
	{
		state_t	*st = states + S_PUFF3;

		// ensure ptrigger is stored.
		if (!trigptr)
		{
			if (st->ptrigger)
			{
				trigptr = st->ptrigger;
				st->ptrigger = 0;	// prevent first use creating ricochet
			}
		}
		else
		{
			if (ricoh && !(rand()%cfg.ricochet) )
			{
				st->ptrigger = trigptr;
				S_StartSound( (rand() & 1) ? r3ds.ricoh1_id : r3ds.ricoh2_id, th);
			}
			else
				// Disable this state until required for richochet particles
				st->ptrigger = 0;
		}
	}
}


//===========================================================================
// P_SpawnBlood
//===========================================================================
void P_SpawnBlood(fixed_t x, fixed_t y, fixed_t z, int damage, int type)
{
	mobj_t*	th;
	int		bloodtype = MT_BLOOD;	//red default

	if (CFX)
	{
		if (type == MT_HEAD)
			bloodtype = MT_BLUEBLOOD;
		else if (type == MT_KNIGHT || type == MT_BRUISER)
			bloodtype = MT_GREENBLOOD;
	}

	z += ((M_RandomD() - M_RandomD()) << 10);

	th = P_SpawnMobj (x,y,z, (mobjtype_t)bloodtype, 0);
	th->momz = FRACUNIT*2;
	th->tics -= (M_RandomD() & 3);
	
	if (th->tics < 1)
		th->tics = 1;
	
	if (damage <= 12 && damage >= 9)
		P_SetMobjState (th,S_BLOOD2);
	else if (damage < 9)
		P_SetMobjState (th,S_BLOOD3);

	P_SetMpos(th);
}


//
// P_CheckMissileSpawn
// Moves the missile forward a bit
//  and possibly explodes it right there.
//
static RDBOOL P_CheckMissileSpawn(mobj_t* th)
{
	th->tics -= (M_RandomD() & 3);

	if (th->tics < 1) th->tics = 1;
	
	// v220-10 replaced P_TryMove
	if (!P_CheckPosition(th, th->x + th->momx, th->y + th->momy, th->z + th->momz, 0))
	{
		// move forward slightly so an angle can be computed
		th->x += (th->momx>>1);
		th->y += (th->momy>>1);
		th->z += (th->momz>>1);
		P_ExplodeMissile(th,8); return false;
	}
	
	return true;
}

//
// P_SpawnMissile
//  v220-5 modified
//  v220-31 speed made constant in 3D distance for missiles (other than boss cubes)
//
fixed_t M_MobjDistance(fixed_t dx, fixed_t dy)
{
	if (!dx) return abs(dy);
	if (!dy) return abs(dx);

	double fx = (double)dx, fy = (double)dy;
	return (fixed_t)sqrt(fx*fx + fy*fy);
}

mobj_t *P_SpawnMissile(mobj_t* source, mobj_t* dest, mobjtype_t type)
{
	angle_t	an;
	int		dist;
	mobj_t	*th = P_SpawnMobj(source->x, source->y, source->z + 32*FRACUNIT, type, 0);
	
	if (type == MT_SPAWNSHOT && th->flags & MF_NOCLIP)	// MT_BOSSCUBE would have been a better name
	{
		// v220-5 if source z is below dest z then change it.
		if (th->z < dest->z)
			th->z = dest->subsector->sector->ceilingheight - 32 * FRACUNIT;
	}
	else
	{
		// v220-4 if necessary adjust firing height
		if (source->info->weapon_offset)
			th->z += source->info->weapon_offset;

		// b211-0 Make stealth monsters spawn stealth missiles
		if (source->flags3 & MF3_STEALTH) th->flags3 |= MF3_STEALTH;
	}

	if (th->info->seesound)
		S_StartSound(th->info->seesound, th);
	
	P_SetTarget(&th->target, source);    // where it came from // killough 11/98
	an = R_PointToAngle2 (source->x, source->y, dest->x, dest->y);
	
	// introduce error for partially visible actor
	if (dest->flags & MF_SHADOW) an += ((M_RandomD() - M_RandomD()) << 20);
	
	th->angle = an;

	// Calculate momx, momy
	an >>= ANGLETOFINESHIFT;
	th->momx = FixedMul (th->info->speed, finecosine[an]);
	th->momy = FixedMul (th->info->speed, finesine[an]);

	if (type == MT_SPAWNSHOT && th->flags & MF_NOCLIP)
	{
		// v220-5 Calculate momz using the true x,y distance
		dist = (int)((M_AccurateDistance(dest->x - source->x, dest->y - source->y) / (float)th->info->speed) * FRACUNIT);
		if (dist < 1) dist = 1;
		th->momz = (dest->z - th->z) / dist;
	}
	else
	{
		// v220-31 Calculate using the true x, y, z distance
		dist = M_MobjDistance(dest->x - th->x, dest->y - th->y);

		// v220-5 aim slightly higher up the thing
		fixed_t distz = dest->z + FixedMul(dest->height, 0xA000) - th->z;
		fixed_t dist3d = distz ? M_MobjDistance(dist, distz) : dist;

		dist = FixedDiv(dist, th->info->speed);
		dist3d = FixedDiv(dist3d, th->info->speed);

		if (dist < 1)
		{
			th->momz = distz;
		}
		else
		{
			th->momx = FixedMul(th->momx, FixedDiv(dist, dist3d));
			th->momy = FixedMul(th->momy, FixedDiv(dist, dist3d));
			th->momz = FixedDiv(distz, dist3d);
		}
	}

	// Check it can be spawned successfully (only MT_SPAWNSHOT is unclipped unless it's been dehacked)
	if (th->flags & MF_NOCLIP || P_CheckMissileSpawn(th))
	{
		P_SetMpos(th);
		return th;
	}

	return NULL;
}

//
// P_SpawnMCPMissile - For mushroom code pointer compatibility
// Note: dest's x, y, z values are the only ones set
//
mobj_t *P_SpawnMCPMissile(mobj_t* source, mobj_t* dest, mobjtype_t type)
{
	mobj_t	*th  = P_SpawnMobj(source->x, source->y, source->z + 32*FRACUNIT, type, 0);
	angle_t	an   = th->angle = R_PointToAngle2 (source->x, source->y, dest->x, dest->y);
	int		dist = P_ApproxDistance (dest->x - source->x, dest->y - source->y)/th->info->speed;
	//-----------------------
	
	// Calculate momx, momy
	if (dist < 1) dist = 1;
	an >>= ANGLETOFINESHIFT;
	th->momx = FixedMul (th->info->speed, finecosine[an]);
	th->momy = FixedMul (th->info->speed, finesine[an]);
	th->momz = (dest->z - source->z) / dist;
	
	P_SetTarget(&th->target, source);    // where it came from // killough 11/98
	// Check it can be spawned successfully (only MT_SPAWNSHOT is unclipped unless it's been dehacked)
	if (!P_CheckMissileSpawn(th)) return NULL;

	P_SetMpos(th);
	return th;
}

//
// P_SpawnPlayerMissile
//  Possibly tries to aim at a nearby monster
//
// v220-10 v220-16 revised v220-31 revised
void P_SpawnPlayerMissile(mobj_t *source, mobjtype_t type)
{
	mobj_t		*mo;
	fixed_t		dist;
	fixed_t		slope = 0;
	angle_t		an = source->angle;
	RDBOOL		isgnd;
	int			snd;

	// ------------------------------------

	// b211-0 Check for grenade
	isgnd = type == MT_ROCKET && ge_inf.grenade_id;	// grenade/pipebomb
	if (isgnd) type = (mobjtype_t)ge_inf.grenade_id;

	// v220-16 Don't bother with P_AimLineAttack if not autoaiming
	//         Note: autoaiming is never used with grenades or pipebombs
	if (!isgnd && autoAim)
	{
		int	mask = (autoAim && ge_inf.isfriendly) ? MF3_FRIEND : 0;	// killough 8/2/98: select enemy order
														// v220-16 Don't bother if there are no friends
		while(1)	// try to find an enemy
		{
			slope = P_AimLineAttack(source, an, MEDIUMRANGE, mask);
			if (mobjtarget) break;
			slope = P_AimLineAttack(source, an += 1<<26, MEDIUMRANGE, mask);
			if (mobjtarget) break;
			slope = P_AimLineAttack(source, an -= 2<<26, MEDIUMRANGE, mask);
			if (mobjtarget) break;
			an = source->angle, slope = 0;
			if (mask) mask = 0; else break;
		}
	}
	else
	{
		float x = (float)(source->dplayer->lookdir * RATIO * PI / 180);

		slope = (fixed_t)((FRACUNIT * RATIO_FACTOR) * tan(x));
	}

	mo = P_SpawnMobj(source->x, source->y, source->z + plrfiringheight, type, 0);

	// b211-0 Transfer originator's id to object's spawnid
	//     This is used to explode an, as yet, unarmed mine that hits an enemy.
	//     The id is used to stop a grenade/pipebomb exloding when first fired.
	// Note: transient, not saved.  This does not cause a problem because when
	//       armed, which happens when it comes to rest, it overrides.  So a
	//       grenade, for example, will still explode in a restored game.
	//       If thinker.id is 0 then the early explode test will not be done so
	//       any grenades just fired will not explode prematurely as the id's
	//       are all cleared to 0 when a game is restored.

	if (isgnd)
		mo->thinker.spawnid = source->thinker.id;	// b211-0

    if ( (snd = mo->info->seesound) > 0 )
		S_StartSound(snd, mo);
	
    P_SetTarget(&mo->target, source);   // killough 11/98

    mo->angle = an;
	an >>= ANGLETOFINESHIFT;
    mo->momx = FixedMul(mo->info->speed, finecosine[an]);
    mo->momy = FixedMul(mo->info->speed, finesine[an]);
    mo->momz = FixedMul(mo->info->speed, slope);

	// Make sure the speed is right (in 3D) - JK.
	dist = P_ApproxDistance(P_ApproxDistance(mo->momx, mo->momy), mo->momz);
	dist = (dist != 0) ? FixedDiv(mo->info->speed, dist) : mo->info->speed;

	mo->momx = FixedMul(mo->momx, dist);
	mo->momy = FixedMul(mo->momy, dist);
	mo->momz = FixedMul(mo->momz, dist);

	if (P_CheckMissileSpawn(mo))
	{
		// Move it to the new position
		mo->old_x = mo->x;
		mo->old_y = mo->y;
		P_UnlinkThing(mo);
		mo->x += mo->momx;
		mo->y += mo->momy;
		mo->z += mo->momz;
		P_SetThingPosition (mo);

		// v220-10 prevent model rocket from being drawn too
		//         close by moving it further from the player
		if (type == MT_ROCKET && loadModels && !(sprites[mo->type].sprflags & SPR_NOMODEL) )
		{
			if (!P_CheckPosition(mo, mo->x + mo->momx, mo->y + mo->momy, mo->z + mo->momz, 0))
			{
				P_ExplodeMissile(mo, 8); return;
			}

			mo->old_x = mo->x;
			mo->old_y = mo->y;
			P_UnlinkThing(mo);
			mo->x += mo->momx;
			mo->y += mo->momy;
			mo->z += mo->momz;
			P_SetThingPosition (mo);
		}
		P_SetMpos(mo);
	}
}

