//-----------------------------------------------------------------------------
//
// Copyright (C) 2005 id Software, Inc., Jaakko Kernen
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: g_game.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  modified for R3D, moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 15 (v220-15) 08/Jun/2012 gmj
// Revision 2.2.0 build 25 (v220-25) 13/May/2014 gmj
// Revision 2.2.0 build 28 (v220-28) 04/Oct/2014 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll code
//
//
// DESCRIPTION:  Top-level game routines.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "doom/p_local.h"
#include "doom/p_start.h"
#include "doom/am_map.h"
#include "doom/f_infine.h"
#include "doom/mbf_player.h"
#include "doom/p_setup.h"
#include "doom/m_menu.h"
#include "doom/hu_stuff.h"
#include "doom/st_stuff.h"
#include "doom/wi_stuff.h"
#include "doom/g_game.h"
#include "doom/d_main.h"
#include "doom/SndIdx.h"
#include "doom/p_saveg.h"
#include "doom/d_englsh.h"
#include "doom/tables.h"
#include "doom/wi_stuff.h"
#include "doom/am_map.h"
#include "p_data.h"
#include "con_action.h"
#include "con_console.h"
#include "con_functions.h"
#include "dd_main.h"
#include "dd_wad.h"
#include "gl_main.h"
#include "m_misc.h"
#include "p_maputil.h"
#include "r_data.h"
#include "r_main.h"
#include "sys_timer.h"
#include "sys_system.h"
#include "s_main.h"
#include "con_action.h"
#include "net_main.h"
#include "sys_input.h"
#include "rend_dyn.h"

#include "r3d_compat.h"
#include "r3d_mapinfo.h"

// MACROS ------------------------------------------------------------------

#define LOCKF_FULL		0x10000
#define LOCKF_MASK		0xff

#define MAXPLMOVE		(forwardmove[1]) 
#define TURBOTHRESHOLD	0x32

#define SLOWTURNTICS		6 
#define NUMGKEYS			256 
#define	BODYQUESIZE			32
#define JOY(x)				(x /*-cfg.joydead*/) / (100 /*-cfg.joydead*/)

// TYPES -------------------------------------------------------------------

// Joystick axes.
enum 
{ 
	JA_X, JA_Y, JA_Z, JA_RX, JA_RY, JA_RZ, JA_SLIDER0, JA_SLIDER1, 
	NUM_JOYSTICK_AXES 
};

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void	P_InitPlayerValues(mbf_player_t *p);
void	P_RunPlayers();
void	DP_Ticker();

void	HU_UpdatePsprites(void);
void	checkSCP_ModifyFlags(void);	// v220-9
void	redirect_fakeSND(void);		// v220-10
int		Def_GetMobjNum(char *id);
int		FI_StartScriptDeferred(int gameaction);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

static void	G_DoNewGame (void); 
static void	G_DoCompleted (void); 
static void	G_DoWorldDone (void); 
static void	G_DoSaveGame (void); 

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern int		demo_lnum;
extern int		wimap_pic;
extern int		pagetic;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

wbstartstruct_t	wminfo;				// parms for world map / intermission 
config_t		cfg;
mbf_player_t	mbf_players[MAXPLAYERS];

mobj_t		*bodyque[BODYQUESIZE]; 
char		savename[256];
char		savedescription[SAVESTRINGSIZE]; 
int			savegameslot; 
int			bodyqueslot; 

int			gameepisode; 
int			gamemap; 
int			levelstarttic;          // gametic at level start 

// for intermission
int			totalkills, totalfriends, totalitems, totalsecret;

RDBOOL		gamekeydown[NUMGKEYS]; 
RDBOOL		secretexit; 

// Looking around.
int			povangle = -1;			// -1 means centered (really 0 - 7).
float		targetLookOffset=0;
float		lookOffset = 0;

fixed_t		angleturn[3] = {640, 1280, 320};	// + slow turn 
fixed_t		forwardmove[2] = {0x19, 0x38};		// was 0x19, 0x32
fixed_t		sidemove[2] = {0x18, 0x28}; 

// Mouse stuff
int			mousebuttons[NUM_MOUSE_BUTTONS]; 

int			turnheld;				// for accelerative turning 
 
// mouse values are used once . Eh? 
int			mousex;
int			mousey;

int			dclicktime;
int			dclickstate;
int			dclicks; 
int			dclicktime2;
int			dclickstate2;
int			dclicks2;

// joystick values are repeated 
int			joymove[NUM_JOYSTICK_AXES];	// X, Y, Z, Rx, Ry, Rz, S1, S2
RDBOOL		joyarray[33]; 
RDBOOL		*joybuttons = &joyarray[1];		// allow [-1] 
 
int			zoomweapontype = false;
int			setup_active;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

RDBOOL	CheckStrafe()
{
	return actions[A_STRAFE].on;
}
 
// Offset is in 'angles', where 110 corresponds 85 degrees.
// The delta has higher precision with small offsets.
char G_MakeLookDelta(float offset)
{
	RDBOOL minus = offset < 0;

	offset = (float)(sqrt(fabs((double)offset)) * DELTAMUL);
	if(minus) offset = -offset;
	// It's only a char...
	if(offset > 127) offset = 127;
	if(offset < -128) offset = -128;
	return (signed char) offset;
}

void G_AdjustAngle(mbf_player_t *player, int turn)
{
	if(!player->plr->mo) return;	// Sorry, can't help you, pal.

	// Turn around.
	if(player->playerstate != PST_DEAD)
		player->plr->clAngle += (turn << FRACBITS);
}

void G_AdjustLookDir(mbf_player_t *player, int look)
{
	ddplayer_t *ddplr = player->plr;

	if(look)
	{
		if(look == TOCENTER)
		{
			player->centering = true;
		}
		else
		{
			int spd = cfg.lookSpeed;
			ddplr->clLookDir += spd*look;
		}
	}
	if(player->centering)
	{
		if (ddplr->clLookDir > 0.f)
		{
			ddplr->clLookDir -= 8.f;
		}
		else if (ddplr->clLookDir < 0.f)
		{
			ddplr->clLookDir += 8.f;
		}
		if (FABS(ddplr->clLookDir) < 8.f)
		{
			ddplr->clLookDir = 0.f;
			player->centering = false;
		}
	}
}

void G_SetCmdViewAngles(ticcmd_t *cmd, mbf_player_t *pl)
{
	// These will be sent to the server (or P_MovePlayer).
	cmd->cmd_angle = pl->plr->clAngle;

	// 110 corresponds 85 degrees (see RATIO).
	if(pl->plr->clLookDir >  110.f) pl->plr->clLookDir =  110.f;
	if(pl->plr->clLookDir < -110.f) pl->plr->clLookDir = -110.f;

	cmd->cmd_lookdir = (short)(pl->plr->clLookDir/110.f * DDMAXSHORT);
}

#pragma optimize("g",off)

#define GOTWPN(x)	(cplr->weaponowned[x])
#define ISWPN(x)	(cplr->readyweapon == x)

//
// G_BuildTiccmd
// Builds a ticcmd from all of the available inputs
// or reads it from the demo buffer. 
// If recording a demo, write it out 
// 
void G_BuildTiccmd (ticcmd_t* cmd) 
{
	mbf_player_t *cplr = &mbf_players[0];
	RDBOOL	mlook_pressed = false;
	RDBOOL	pausestate = menu_active || con_freezegame || (paused > 0);	// v220-28 add con_freezegame
	RDBOOL	strafe;
	RDBOOL	bstrafe; 
	int		i; 
	int		speed;
	int		tspeed; 
	int		forward;
	int		side;
	int		turn = 0;
	int		joyturn = 0, joystrafe = 0, joyfwd = 0, joylook = 0;
	int		*axes[5] = { 0, &joyfwd, &joyturn, &joystrafe, &joylook };
	int		num, oldstate;	// b211-0

	memset(cmd, 0, sizeof(*cmd));

	if (R3D_edit && !pausestate &&
		Edit_CheckEditSelect(actions[A_FIRE].on, &mousex, &mousey))
	{
		G_SetCmdViewAngles(cmd, cplr);
		return;
	}

	// b211-0 revised: Check the joystick axes.
	for(i = 0; i < 8; i++)
	{
		if( (num = cfg.joyaxis[i]) > 0 && num < 5)
			*axes[num] += joymove[i];
	}

	strafe = actions[A_STRAFE].on; 
	speed = actions[A_SPEED].on;
	forward = side = 0;

	// Walk -> run, run -> walk.
	if (cfg.alwaysRun) speed = !speed;

	// Use two stage accelerative turning on the keyboard and joystick.
	if (   joyturn < -0 /*cfg.joydead*/ 
		|| joyturn > 0 /*cfg.joydead*/ 
		|| actions[A_TURNRIGHT].on 
		|| actions[A_TURNLEFT].on)
		turnheld++; 
	else 
		turnheld = 0; 
	
	if(turnheld < SLOWTURNTICS) 
		tspeed = 2;             // slow turn 
	else 
		tspeed = speed;
	
	// let movement keys cancel each other out
	if(strafe) 
	{ 
		if(actions[A_TURNRIGHT].on) side += sidemove[speed]; 
		if(actions[A_TURNLEFT].on) side -= sidemove[speed]; 
		// Swap strafing and turning.
		i = joystrafe;
		joystrafe = joyturn;
		joyturn = i;
	} 
	else 
	{ 
		if(actions[A_TURNRIGHT].on) turn -= angleturn[tspeed]; 
		if(actions[A_TURNLEFT].on) turn += angleturn[tspeed]; 
	} 
	
	// Joystick turn.
	if (joyturn > 0 /*cfg.joydead*/)
		turn -= angleturn[tspeed] * JOY(joyturn);
	if (joyturn < -0 /*cfg.joydead*/)
		turn += angleturn[tspeed] * JOY(-joyturn);
	
	// Joystick strafe.
	if(joystrafe < -0 /*cfg.joydead*/)
		side -= sidemove[speed] * JOY(-joystrafe);
	if(joystrafe > 0 /*cfg.joydead*/)
		side += sidemove[speed] * JOY(joystrafe);
	
	if(actions[A_FORWARD].on) 
	{
		forward += forwardmove[speed]; 
	}
	if(actions[A_BACKWARD].on) 
	{
		forward -= forwardmove[speed]; 
	}
	if(actions[A_USE].on) cmd->use = true;
	
	if(joyfwd < -0 /*cfg.joydead*/)
		forward += forwardmove[speed] * JOY(-joyfwd);
	if(joyfwd > 0 /*cfg.joydead*/)
		forward -= forwardmove[speed] * JOY(joyfwd);
	
	if(actions[A_STRAFERIGHT].on) side += sidemove[speed]; 
	if(actions[A_STRAFELEFT].on) side -= sidemove[speed];
	
	if(joystrafe < -0 /*cfg.joydead*/)
		side -= sidemove[speed] * JOY(-joystrafe);
	if(joystrafe > 0 /*cfg.joydead*/)
		side += sidemove[speed] * JOY(joystrafe);
	
	
	// Buttons
	
	if(actions[A_FIRE].on) cmd->buttons |= BT_ATTACK; 
	
	if(actions[A_USE].on) 
	{ 
		cmd->buttons |= BT_USE;
		// clear double clicks if hit use button 
		dclicks = 0;                   
	} 
	
	if(actions[A_JUMP].on) cmd->buttons |= BT_JUMP;
	
	// Determine whether a weapon change should be done.
	if (R3D_edit) oldstate = cmd->buttons;
	if(actions[A_WEAPONCYCLE1].on) // Fist/chainsaw.
	{
		if(ISWPN(wp_fist) && GOTWPN(wp_chainsaw))
			i = wp_chainsaw;
		else if(ISWPN(wp_chainsaw))
			i = wp_fist;
		else if(GOTWPN(wp_chainsaw))
			i = wp_chainsaw;
		else
			i = wp_fist;
		
		cmd->buttons &= ~BT_JUMP;
		cmd->buttons |= BT_CHANGE; 
		cmd->buttons |= i << BT_WEAPONSHIFT; 
	}
	else if(actions[A_WEAPONCYCLE2].on) // Shotgun/super sg.
	{
		if(ISWPN(wp_shotgun) && GOTWPN(wp_supershotgun)
			&& commercial_game)
			i = wp_supershotgun;
		else if(ISWPN(wp_supershotgun))
			i = wp_shotgun;
		else if(GOTWPN(wp_supershotgun) && commercial_game)
			i = wp_supershotgun;
		else
			i = wp_shotgun;
		
		cmd->buttons &= ~BT_JUMP;
		cmd->buttons |= BT_CHANGE; 
		cmd->buttons |= i << BT_WEAPONSHIFT; 
	}
	else
	{
		// Take the first weapon action.
		for(i = 0; i < NUMWEAPONS; i++)
		{
			if(actions[A_WEAPON1+i].on) 
			{ 
				cmd->buttons &= ~BT_JUMP;
				cmd->buttons |= BT_CHANGE; 
				cmd->buttons |= i << BT_WEAPONSHIFT; 
				break;
			}
		}
	}
	if(actions[A_NEXTWEAPON].on || actions[A_PREVIOUSWEAPON].on)
	{
		cmd->buttons = (byte)(BT_SPECIAL 
			| (actions[A_NEXTWEAPON].on? BTS_NEXTWEAPON : BTS_PREVWEAPON));
	}
	if (R3D_edit && cmd->buttons != oldstate)
	{
		cmd->buttons &= oldstate;
		cmd->buttons &= ~BT_JUMP;
	}
	
	// forward double click
	if(actions[A_FORWARD].on != dclickstate 
		&& dclicktime > 1 
		&& cfg.dclickuse)  
	{ 
		dclickstate = actions[A_FORWARD].on; 
		
		if (dclickstate) 
			dclicks++; 
		if (dclicks == 2) 
		{ 
			cmd->buttons |= BT_USE; 
			dclicks = 0; 
		} 
		else 
			dclicktime = 0; 
	} 
	else 
	{ 
		dclicktime++; 
		if (dclicktime > 4) 
		{ 
			dclicks = 0; 
			dclickstate = 0; 
		} 
	}
	
	// strafe double click
	bstrafe = strafe; 
	if(bstrafe != dclickstate2 && dclicktime2 > 1 && cfg.dclickuse) 
	{ 
		dclickstate2 = bstrafe; 
		if (dclickstate2) 
			dclicks2++; 
		if (dclicks2 == 2) 
		{ 
			cmd->buttons |= BT_USE; 
			dclicks2 = 0; 
		} 
		else 
			dclicktime2 = 0; 
	} 
	else 
	{ 
		dclicktime2++; 
		if (dclicktime2 > 4) 
		{ 
			dclicks2 = 0; 
			dclickstate2 = 0; 
		} 
	}
	
	// Mouse strafe and turn (X axis).
	if(strafe) 
		side += mousex*2; 
	else 
		turn -= mousex*0x8; 
	
	if (!pausestate)
	{
		if (   strafe
			|| (!cfg.usemlook && !actions[A_MLOOK].on)
			|| mbf_players[0].playerstate == PST_DEAD) 
		{
			forward += mousey;
		}
		else
		{
			float adj = mousey ?
				180.0f * RATIO * (float)((mousey*0x8)<<16)/(float)ANG180 : 0;

			if(cfg.mlookInverseY) adj = -adj;
			cplr->plr->clLookDir += adj;
		}

		if (cfg.usejlook)
		{
			int sign = cfg.jlookInverseY ? -1 : 1;

			if(cfg.jlookDeltaMode)
				cplr->plr->clLookDir += joylook/20.0f * cfg.lookSpeed * sign;
			else
				cplr->plr->clLookDir = joylook * 1.1f * sign;
		}
	}
	
	mousex = mousey = 0; 
	
	if (abs(forward) > MAXPLMOVE)
		forward = forward > 0 ? MAXPLMOVE : -MAXPLMOVE;

	if (abs(side) > MAXPLMOVE)
		side = side > 0 ? MAXPLMOVE : -MAXPLMOVE;
	
	cmd->forwardmove = (char)(cmd->forwardmove + forward); 
	cmd->sidemove    = (char)(cmd->sidemove + side);
	
	if (   cfg.lookSpring
		&& !(cfg.usemlook)	// v220-15 bug fix
		&& !actions[A_MLOOK].on 
		&& (mlook_pressed
		|| abs(cmd->forwardmove) > MAXPLMOVE/3
		|| abs(cmd->sidemove   ) > MAXPLMOVE/3))
	{
		// Center view when mlook released w/lookspring, or when moving.
		cplr->centering = true;
	}
	
	if (!pausestate && cplr->playerstate == PST_LIVE)
	{
		i = 0;
		// Key look.
		if(actions[A_LOOKCENTER].on)
		{
			i = TOCENTER;
		}
		else 
		{
			if(actions[A_LOOKUP].on)
			{
				i += cfg.lookSpeed;
			}
			if(actions[A_LOOKDOWN].on)
			{
				i -= cfg.lookSpeed;
			}	
		}
		// Adjust the look direction.
		G_AdjustLookDir(cplr, i);
	}
	
	if(!pausestate) G_AdjustAngle(cplr, turn);
	
	// Store the current mlook key state.
	mlook_pressed = actions[A_MLOOK].on;
	
	G_SetCmdViewAngles(cmd, cplr);
}

#undef ISWPN
#undef GOTWPN

// clear cmd building stuff
void G_ClrCommands(void)
{
	action_t	*act;
	
	memset(gamekeydown, 0, sizeof(gamekeydown)); 
	memset(joymove, 0, sizeof(joymove));
	mousex = mousey = 0; 
	paused = false; 
	memset(mousebuttons, 0, sizeof(mousebuttons)); 
	memset(joybuttons, 0, sizeof(joybuttons)); 
	
	// Deactivate all action keys.
	for(act = actions; act->name[0]; act++) act->on = false;
}
 
//
// G_DoLoadLevel 
//

void G_DoLoadLevel (void) 
{
	if (FI_StartScriptDeferred(gameaction)) return;

	mbf_players[0].plr->mo = NULL;	// NULL the player's mobj

	mbf_players[0].plr->ingame = true;

	// v220-16 if necessary reset the player/player's inventory 
	if (mbf_players[0].plr->ingame)
	{
		int pd1 = (ma_inf.playerdamage
			  && ma_inf.playerdamage >= plhealth.health)
			  || mbf_players[0].playerstate == PST_DEAD;
		int pd2 = false;

		if (!pd1)
		{
			ded_mapinfo_t *mi;
			char id[8];

			if (commercial_game)
				sprintf(id, "map%02i", gamemap);
			else
				sprintf(id,"E%iM%i", gameepisode, gamemap);

			if ( (mi = Def_GetMapInfo(id)) && mi->R3Dflags & resetinventory)	// Good as dead
			{
				pd2 = true;
			}
		}

		if (pd1 || pd2)
		{
			mbf_players[0].playerstate = PST_REBORN;

			if (pd1)
				memset(mbf_players[0].frags, 0, sizeof(mbf_players[0].frags));
		}
	}

    automap_active = false; 
	setup_active = true;
	P_SetupLevel(&gameepisode, &gamemap);
	setup_active = false;

	redirect_fakeSND();	// v220-10
	checkSCP_ModifyFlags();
	wimap_pic = -1;
	
	G_ClrCommands();
	
	gamestate = GS_LEVEL; 
	gameaction = ga_nothing;

	// Start a briefing, if there is one.
	if (ge_inf.first_map)
		ge_inf.first_map = false;
	else
		FI_Briefing(ge_inf.nextname[0] ? ge_inf.nextname : mapinfo->id);

	levelstarttic = gametic = 0;	// for time calculation
	leveltime = actual_leveltime = 0;
}

//
// G_Responder  
// Get info needed to make ticcmd_ts for the players.
// Return false if the event should be checked for bindings.
// 
RDBOOL G_Responder (event_t* ev) 
{ 
	int		i;

	if (   gameaction == ga_nothing
		&& ev->data1 != DDKEY_RALT
		&& gamestate == GS_DEMOSCREEN)
	{
		if (   ev->type == ev_keydown 
			|| ev->type == ev_mousebdown 
			|| ev->type == ev_joybdown) 
		{ 
			M_StartControlPanel (); 
		} 
		return false; 
	}

	if (ma_inf.exitlevel)
	{
		gamestate = GS_LEVEL;
		gameaction = ga_nothing;
	}
	
	if (gamestate == GS_LEVEL)
	{
		if (HU_Responder(ev)) 
			return true;	// chat ate the event
		
		if (ST_Responder(ev)) 
			return true;	// status window ate it 
		
		if (automap_active && AM_Responder(ev)) 
			return true;	// automap ate it 

		if (ma_inf.exitlevel & 0x88800000)
		{
			int delay = (ma_inf.exitlevel) & 0xFF;

			if ( !(--delay) )
			{
				if (ma_inf.exitlevel & 0x88000000)
				{
					G_SecretExitLevel();
				}
				else	// 0x00800000
				{
					G_ExitLevel();
				}
			}
			else
			{
				ma_inf.exitlevel &= ~0xFF;
				ma_inf.exitlevel |= delay;
			}
		}
	}
	else if (gamestate == GS_INFINE)
	{
		if (FI_Responder(ev)) return true; 
	}
	
	// Check keyboard first
	if (ev->type == ev_keydown)
	{
		if(ev->data1 < NUMGKEYS) gamekeydown[ev->data1] = true; 
		return false;
	}

	if (ev->type == ev_keyup)
	{
		if(ev->data1 < NUMGKEYS) gamekeydown[ev->data1] = false; 
		return false;   // always let key up events filter down
	}	

	if (ev->type == ev_keyrepeat)
	{
		return false;
	}
	
	// Check mouse
	if (mouseInited)
	{
		if (ev->type == ev_mouse)
		{
			mousex += (ev->data1 * (cfg.mouseSensiX*2+5)) / 6;
			mousey += (ev->data2 * (cfg.mouseSensiY*2+5)) / 6;
			return true; // eat events
		}
		
		if (ev->type == ev_mousebdown)
		{
			for(i = 0; i < NUM_MOUSE_BUTTONS; i++)
			{
				if (ev->data1 & (1<<i))
					mousebuttons[i] = true;
			}
			
			if (mousebuttons[1]) Edit_SetMB2();

			return false;
		}
			
		if (ev->type == ev_mousebup)
		{
			for(i = 0; i < NUM_MOUSE_BUTTONS; i++)
			{
				if(ev->data1 & (1<<i))
					mousebuttons[i] = false;
			}

			return false;
		}
	}

	// Check joystick
	if (joyInited)
	{		
		if (ev->type == ev_joystick) // Joystick movement
		{
			joymove[JA_X ] = ev->data1;
			joymove[JA_Y ] = ev->data2;
			joymove[JA_Z ] = ev->data3;
			joymove[JA_RX] = ev->data4;
			joymove[JA_RY] = ev->data5;
			joymove[JA_RZ] = ev->data6;
			return true; // eat events
		}
			
		if (ev->type == ev_joyslider) // Joystick slider movement
		{
			joymove[JA_SLIDER0] = ev->data1;
			joymove[JA_SLIDER1] = ev->data2;
			return true;
		}
			
		if (ev->type == ev_joybdown)
		{
			for(i = 0; i < 32; i++)
			{
				if(ev->data1 & (1<<i)) 
					joybuttons[i] = true;
			}

			return false; // eat events
		}
				
		if (ev->type == ev_joybup)
		{
			for(i = 0; i < 32; i++)
			{
				if(ev->data1 & (1<<i)) 
					joybuttons[i] = false;
			}

			return false; // eat events
		}
				
		if (ev->type == ev_povup)
		{
			povangle = -1;
			// If looking around with PoV, don't allow bindings.
			if (cfg.povLookAround) return true;
		}	
		else if (ev->type == ev_povdown)
		{
			povangle = ev->data1;
			if (cfg.povLookAround) return true;
		}
	}

	// The event wasn't used. 
	return false; 
} 

// v220-5 moved here
static void updatePlayer(void)
{
	mbf_player_t	*plr = &mbf_players[0];
	RDBOOL			berzerk = false;

	// v220-4 Set the overlay filters for (players) the console player.
	// Note: code that duplicated this in st_stuff.c has been removed
	// v220-5 revised. Hit flash alpha fixed, Hit and pickup flash optioned
	if (plr->plr)
	{
		byte	pal[4] = { 0, 0, 0, 0 };
		int		val, cnt;

		if (!plr->plr->ingame) return;

		// if underwater and dead then fakefilter takes precedence
		if (plhealth.health <= 0 && fakefilter && underwater)
		{
			goto set_pal;	// No plr filters wanted
		}
		
		cnt = plr->damagecount;

	// berzerk pack
		if (plr->powers[pw_strength])
		{
			// v220-6 always allow doom bug
			int bzk = 12 - (plr->powers[pw_strength] >> 6);

			if (bzk < 0) bzk = 0;

			if (!bzk)
			{
				plr->powers[pw_strength] = 0;
			}
			else if (!cnt && bzk)
			{
				//  This forces the doom bug which causes pw_strength, 
				//  once attained, to persist; unless scripted otherwise
				cfg.berserkBug = !cfg.noberserkBug;

				// v220-4 slowly fade the berzerk out
				pal[2] = 0xff;
				pal[3] = (byte)((900 - plr->powers[pw_strength]) >> 3);
				goto set_pal;
			}
			
			if (bzk > cnt) cnt = bzk;
		}
		
	// player hurt
		if (cnt && cfg.flashdamage)
		{
			if ((val = (cnt+7)>>3) >= NUMREDPALS) val = NUMREDPALS-1;

			if (berzerk && val > 2) val = 2;

			if ( !(berzerk && val == 1 && leveltime & 0x10) )
			{
				// v220-5 flash intensity.
				// Start at .4 as anything less is barely visible
				float d = 9.0f / (0.3f + cfg.flashdamage * 0.7f);

				val += STARTREDPALS;
				pal[2] = 0xff;
				pal[3] = (byte)(0xff * (float)val/d);
				goto set_pal;
			}
		}
		
	// pickup flash	v220-33 kill flash if fakefilter is set
		if (plr->bonuscount && cfg.flashpickup && !fakefilter)
		{
			if ((val = (plr->bonuscount+7)>>3) >= NUMBONUSPALS)
				val = NUMBONUSPALS-1;

			if (val)
			{
				// v220-5 flash intensity.
				// Start at .4 as anything less is barely visible
				float d = 0.2f + cfg.flashpickup * 0.8f;

				val <<= 4;

				pal[0] = 0x80, pal[1] = 0xcc, pal[2] = 0xff;
				pal[3] = (byte)((float)val * d);
				goto set_pal;
			}
		}

	// radiation suit
		if ( plr->powers[pw_ironfeet] > 4*32 || plr->powers[pw_ironfeet]&8)
		{
			pal[1] = 0xb2;
			pal[3] = 0x26;
		}

	set_pal:;
		plr->plr->filter = *(int*)&pal;
	}

	zoomweapontype = plr->pendingweapon == wp_nochange
		&& plr->readyweapon != wp_fist
		&& plr->readyweapon != wp_chainsaw;
}

//===========================================================================
// G_Ticker
//===========================================================================
void G_Ticker (void) 
{
	int i;

	// do player rebirths if needed
	for (i = 0; i < MAXPLAYERS; i++) 
	{
		if(mbf_players[i].plr->ingame && mbf_players[i].playerstate == PST_REBORN) 
			G_DoReborn (i);
	}
	
	// do things to change the game state
	while (gameaction != ga_nothing) 
	{
		yesno_exit = 0;

		switch (gameaction) 
		{ 
		case ga_loadlevel: 
			ge_inf.oldgamenum = -1;
			gamestate = GS_LOADING; 
			G_DoLoadLevel (); 
			break;
			
		case ga_newgame: 
			ge_inf.oldgamenum = -1;
			G_DoNewGame (); 
			break;
			
		case ga_loadgame: 
			curfilter = fakefilter = 0;	
			G_DoLoadGame (); 
			break;
			
		case ga_savegame: 
			G_DoSaveGame (); 
			break; 
			
		case ga_playdemo: 
			break;
			
		case ga_completed: 
			ge_inf.oldgamenum = -1;
			G_DoCompleted (); 
			break;
			
		case ga_victory: 
			ge_inf.oldgamenum = -1;
			curfilter = fakefilter = 0;	
			gameaction = ga_nothing;
			break;
			
		case ga_worlddone: 
			ge_inf.oldgamenum = -1;
			G_DoWorldDone (); 
			break;
			
		case ga_screenshot: 
			{
				char	buff[40];
				int		num = M_DoScreenShot();

				if (num == -1)
					strcpy(buff, "NO FREE FILES - SCREENSHOT NOT TAKEN");
				else
					sprintf(buff, "SCREENSHOT DOOM%.2i.TGA TAKEN", num);

				P_playSetMessage(&mbf_players[0], buff);
			}
			gameaction = ga_nothing; 
			break;
			
		case ga_nothing: 
			break;
		} 
	}
	
	// Look around.
	if(povangle != -1)
	{
		targetLookOffset = povangle/8.0f;
		if(targetLookOffset == .5f)
		{
			if(lookOffset < 0) targetLookOffset = -.5f;
		}
		else if(targetLookOffset > .5)
			targetLookOffset -= 1;
	}
	else targetLookOffset = 0;
	
	if (targetLookOffset != lookOffset && cfg.povLookAround)
	{
		float diff = (targetLookOffset - lookOffset)/2;
		if(diff >  .075f) diff =  .075f;
		if(diff < -.075f) diff = -.075f; 
		lookOffset += diff;
	}
	
	// Must be called on every tick.
	P_RunPlayers();

	// Do main actions.
	switch (gamestate) 
	{ 
	case GS_LEVEL: 
		DL_InitForNewFrame();	// v220-35 preliminary function need only be done on gametic
		DP_Ticker (); 
		HU_UpdatePsprites();
		ST_Ticker (); 
		AM_Ticker (); 
		HU_Ticker ();
		break; 
		
	case GS_INTERMISSION: 
		WI_Ticker (); 
		break; 
		
	case GS_INFINE:
		FI_Ticker();
		break;
		
	case GS_DEMOSCREEN:
		if (music_status >= 0)
			advancedemo = music_status == 0;
		else
			if (--pagetic < 0) advancedemo = true;
		break; 
	}        
	
	updatePlayer();
} 
 
//
// PLAYER STRUCTURE FUNCTIONS
// also see P_SpawnPlayer in P_Things
//

//
// G_InitPlayer 
// Called at the start.
// Called by the game initialization functions.
//
// R3D not used
/*
void G_InitPlayer (int pnum) 
{ 
	mbf_player_t*	p; 
	
	// set up the saved info         
	p = &mbf_players[player]; 
	
	// clear everything else to defaults 
	G_PlayerReborn (pnum); 
} 
*/

//
// G_PlayerFinishLevel
// Can when a player completes a level.
//
void G_PlayerFinishLevel (int player) 
{ 
	mbf_player_t *p = &mbf_players[player]; 	 
	
	memset (p->powers, 0, sizeof (p->powers)); 
	memset (p->cards, 0, sizeof (p->cards));
	
	p->plr->mo->flags &= ~MF_SHADOW;	// cancel invisibility 
	p->plr->extralight = 0;				// cancel gun flashes 
	p->plr->fixedcolormap = 0;			// cancel goggles 
	p->damagecount = 0;					// no palette changes 
	p->bonuscount = 0; 
	p->plr->lookdir = 0;

	// v220-4 Reset to default in case was changed by the script
	ge_inf.customhits = 100;
	ge_inf.customdamage = 9999;
} 

//==========================================================================
//
// ClearPlayer
//
// Safely clears the player data structures.
//
//==========================================================================

void ClearPlayer(mbf_player_t *p)
{
	ddplayer_t	*ddplayer = p->plr;
	int			playeringame = ddplayer->ingame;
	int			flags = ddplayer->flags;
	int			start = p->startspot;
	
	memset(p, 0, sizeof(*p));
	// Restore the pointer to ddplayer.
	p->plr = ddplayer;
	// Also clear ddplayer.
	memset(ddplayer, 0, sizeof(*ddplayer));
	// Restore the pointer to this player.
	ddplayer->extradata = p;
	// Restore the playeringame data.
	ddplayer->ingame = playeringame;
	ddplayer->flags = flags;
	// Don't clear the start spot.
	p->startspot = start;
}

//
// G_PlayerReborn
// Called after a player dies 
// almost everything is cleared and initialized 
//
void G_PlayerReborn (int pnum) 
{ 
	static haswarned = false;
	mbf_player_t*	p; 
	int			frags[MAXPLAYERS]; 
	int			killcount;
	int			itemcount;
	int			secretcount; 
	int			i; 

	if (pnum)
	{
		if (!haswarned)
		{
			Con_Message("WARNING: No playerstart type1 spots found in map\n"
				"Reverting to next found type%i\n", pnum);
			haswarned = true;
		}
		pnum = 0;
	}
	
	memcpy(frags, mbf_players[pnum].frags, sizeof(frags)); 
	killcount	= mbf_players[pnum].killcount; 
	itemcount	= mbf_players[pnum].itemcount; 
	secretcount	= mbf_players[pnum].secretcount; 
	
	p = &mbf_players[pnum]; 
	
	// Clears (almost) everything.
	ClearPlayer(p);
	
	memcpy(mbf_players[pnum].frags, frags, sizeof(mbf_players[pnum].frags)); 
	mbf_players[pnum].killcount	= killcount; 
	mbf_players[pnum].itemcount	= itemcount; 
	mbf_players[pnum].secretcount	= secretcount; 
	
	p->usedown = p->attackdown = true;	// don't do anything immediately 
	p->playerstate = PST_LIVE;       
	plhealth.health = MAXHEALTH;
	plhealth.armorpoints = 0;
	plhealth.armortype = 0;
	p->readyweapon = p->pendingweapon = wp_pistol; 
	p->weaponowned[wp_fist] = true; 
	p->weaponowned[wp_pistol] = true; 
	p->ammo[am_clip] = 50; 
	
	// See if the Values specify anything.
	P_InitPlayerValues(p);
	
	// Reset maxammo.
	for (i=0 ; i<NUMAMMO ; i++) p->maxammo[i] = maxammo[i]; 
	
	// We'll need to update almost everything.
	p->plr->flags &= ~DDPF_DEAD;
}

//
// G_DoReborn 
// 
void G_DoReborn (int playernum) 
{
	brief_disabled = -1;		// We've just died, don't do a briefing now.
	gameaction = ga_loadlevel; 	// reload the level from scratch
} 

void G_ScreenShot (void) 
{ 
    gameaction = ga_screenshot; 
} 
 
// DOOM Par Times
int pars[4][10] = 
{ 
    {0, 30, 75,120, 90,165,180,180, 30,165},	// Epi 1 games 1 - 9
    {0, 90, 90, 90,120, 90,360,240, 30,170},	// Epi 2
    {0, 90, 45, 90,150, 90, 90,165, 30,135},	// Epi 3
    {0, 90, 45, 90,150, 90, 90,165, 30,135}		// Epi 4 (Just duplicated E3 - correct values need setting)
}; 

// DOOM II Par Times
int cpars[32] =
{
     30, 90,120,120, 90,150,120,120,270, 90,	//  1-10
    210,150,150,150,210,150,420,150,210,150,	// 11-20
    240,150,180,150,150,300,330,420,300,180,	// 21-30
    120,30										// 31-32
};

//
// G_DoCompleted 
//
 
void G_ExitLevel (void) 
{ 
	ma_inf.exitlevel = 0;

	if (R3D_edit) return;

    secretexit = false; 
    gameaction = ga_completed; 
}

// v220-33 ammended to fix fault with exiting to Doom secret levels
void G_SecretExitLevel (void) 
{
	if (R3D_edit)
	{
		ma_inf.exitlevel = 0;
		return;
	}

	if (commercial_game)
	{
		char *map = ma_inf.exitlevel & 0x80000000 ? "map32" :
					ma_inf.exitlevel & 0x08000000 ? "map31" : "";
		
		// IF NO WOLF3D LEVELS, NO SECRET EXIT!
		secretexit = *map && W_CheckNumForName(map) >= 0;
	}
	else	// doom ExMx maps.
	{
		secretexit = ma_inf.exitlevel & 0x1000;
	}

	ma_inf.exitlevel = 0;
	gameaction = ga_completed; 
}

// v220-4 carryover stats code ==========================================

static struct addstats_t {
	int		mitems;
	int		mkills;
	int		msecret;
	int		mcount;
	int		friends;
	int		emassacre;
	int		eremaining;
	int		ecollateral;
	int		fremaining;
	int		iremaining;

	int		psecret;
	int		pkills;
	int		pitems;
	int		ptime;

	int		ftype[64];

} ads;

void G_ClearAddstats()
{
	memset(&ads, 0, sizeof(ads));
}

byte *G_SetAddstats(int *size)
{
	*size = sizeof(ads);
	return (byte*)&ads;
}

// User must clear returned ptr with M_Free()
byte *G_CopyAddstats(int *size)
{
	byte	*ptr;

	*size = sizeof(ads);
	ptr = (byte*)M_Malloc(*size, 0);
	memcpy(ptr, &ads, *size);
	return ptr;
}

void G_GetFriendStats()
{
	ma_inf.friendscarriedover = ads.fremaining;
}

void G_DeductFriendStats()
{
	ads.fremaining--;
	totalfriends--;
}

int G_GetFriendType(int n)
{
	return ads.ftype[n];
}

// v220-3 endgame screen stats (single player)
// v220-4 support for "R3Dtype carrystatsover" added
void R3D_EndLevelStats(char *levelid, int addstats)
{
	thinker_t *th;
	char	p0[8], p1[8], p2[8];
	int		n = 0;	// wminfo.plyr index (always 0)
	int		tsec;
	int		cleft = ma_inf.collateraldamage_count;
	int		ileft = 0;
	int		eleft = 0;
	int		fleft = 0;
	int		mleft = ma_inf.massacre_count;

	for (th = thinkercap.next; th != &thinkercap; th = th->next)
	{
		if ((void (__cdecl *)(void))th->function == (void (__cdecl *)(void))P_MobjThinker)
		{
			int flags = ((mobj_t*)th)->flags;

			if (flags & MF_COUNTITEM) ileft++;

			if (   flags & MF_COUNTKILL
				&& ((mobj_t*)th)->health > 0)
			{
				if (((mobj_t*)th)->flags3 & MF3_FRIEND)
					ads.ftype[fleft++] = ((mobj_t*)th)->type;
				else if (((mobj_t*)th)->info->missilestate
					|| ((mobj_t*)th)->info->meleestate)
					eleft++;
			}
		}
	}

	if (addstats)
	{
		ads.mitems  += wminfo.maxitems ; wminfo.maxitems  = ads.mitems ;
		ads.mkills  += wminfo.maxkills ; wminfo.maxkills  = ads.mkills ;
		ads.msecret += wminfo.maxsecret; wminfo.maxsecret = ads.msecret;

		ads.friends += totalfriends; totalfriends = ads.friends;

		ads.pitems  += wminfo.plyr[n].sitems ; wminfo.plyr[n].sitems  = ads.pitems;
		ads.pkills  += wminfo.plyr[n].skills ; wminfo.plyr[n].skills  = ads.pkills;
		ads.psecret += wminfo.plyr[n].ssecret; wminfo.plyr[n].ssecret = ads.psecret;
		ads.ptime   += wminfo.plyr[n].stime  ; wminfo.plyr[n].stime   = ads.ptime;

		ads.ecollateral += cleft; cleft = ads.ecollateral;
		ads.eremaining  += eleft; eleft = ads.eremaining;
		ads.fremaining  += fleft; fleft = ads.fremaining;
		ads.iremaining  += ileft; ileft = ads.iremaining;
		ads.emassacre   += mleft; mleft = ads.emassacre;
		ads.mcount++;
	}
	else
	{
		G_ClearAddstats();
	}

	// Setting the strings is unnecessary if nointermission
	if (mapinfo->R3Dflags & nointermission) return;

	if (ads.mcount > 1)
		sprintf(scpMessage.info[0], "FOR %i MAPS", ads.mcount);	/* AGGREGATED STATS*/
	else
		scpMessage.info[0][0] = 0;

	itoa(wminfo.plyr[n].skills, scpMessage.info[1], 10);	/* PLAYER KILLS */

	if (cleft)
		itoa(cleft, scpMessage.info[2], 10);				/* COLLATERAL KILLS */
	else
		scpMessage.info[2][0] = 0;

	if (mleft)
		itoa(mleft, scpMessage.info[3], 10);				/* MASSACRED */
	else
		scpMessage.info[3][0] = 0;

	itoa(eleft, scpMessage.info[4], 10);					/* ENEMIES LEFT		*/

	if (totalfriends)
	{
		itoa(fleft, p0, 10);								/* FRIENDS LEFT		*/
		itoa(totalfriends, p1, 10);							/* of TOTAL FRIENDS	*/
		sprintf(scpMessage.info[5], "%s of %s", p0, p1);
	}
	else
	{
		scpMessage.info[5][0] = 0;
	}

	itoa(ileft, scpMessage.info[6], 10);					/* ITEMS UNUSED		*/

	itoa(wminfo.plyr[n].ssecret, p0, 10);					/* SECRETS FOUND	*/
	itoa(wminfo.maxsecret, p1, 10);							/* of TOTAL SECRETS	*/
	sprintf(scpMessage.info[7], "%s of %s", p0, p1);

	tsec = wminfo.plyr[n].stime;
	itoa(tsec/3600, p0, 10);		// hours
	itoa((tsec%3600)/60, p1, 10);	// minutes
	itoa(tsec%60, p2, 10);			// seconds

	if (!p1[1]) p1[1] = p1[0], p1[0] = '0', p1[2] = 0;	// add leading 0
	if (!p2[1]) p2[1] = p2[0], p2[0] = '0', p2[2] = 0;	// add leading 0

	sprintf(scpMessage.info[8], "%s:%s:%s", p0, p1, p2);	/* TIME				*/
}

// v220-4 end carryover stats code ==================================================

//==========================================================================
//  Starts any mapinfo intermission routines and sets up next map.
//==========================================================================
static void G_DoCompleted (void) 
{
	// b211-0 used to access next map
	ded_mapinfo_t *dmt = NULL;

	int		gmap;
	RDBOOL	hasmapinfo = (mapinfo->R3Dflags & R3Dhasmapinfo) != 0;
	RDBOOL	checkpwads = commercial_game && haspwads;

	//-------------------------------

	// v220-5 add test for showing interpic in place of titlepic
	wimap_pic = !R3D_ScriptedMap ? interpic_num : titlepic_num;

	curfilter = fakefilter = 0;

	S_StopAllSounds(NULL);

	// v220-13 Ensure underwater reverb is off
	waterlevel = underwater = 0;

/*	if (ppt.view_subsector)
	{
		// We need a way of updating the listener. If underwater
		// then the listener will be in a fake sector. If so then
		// set the listener's sector to the fake sector.
		int hs = ppt.view_subsector->sector->heightsec;

		if (hs >= 0)
		{
			ppt.view_subsector->sector = &sectors[hs];
		}
	}*/

	G_PrepareWIData();
	R3D_EndLevelStats(mapinfo->id, mapinfo->R3Dflags & R3Dcarrystatsover);

	wminfo.show_completed = false;
	wminfo.nextmapid[0] = 0;	// b211-0 store patch nums and strings
	wminfo.inflid[0] = 0;
	wminfo.runCast = false;

	if (!commercial_game && gamemap == 8)
	{
		wminfo.show_completed = true;
	}
	else
	if (hasmapinfo && mapinfo->R3Dflags & (R3Dendgame|R3DendgameC))
	{
		wminfo.show_completed = true;
	}
	else	// Deal with pwad map range in the absence of a MAPINFO definition
	if (  !hasmapinfo
		&& commercial_game
		&& gamemap > 0 && gamemap < 31
		&& haslevelMapxx[gamemap]
		&& checkpwads
		&& !R3D_ScriptedMap
		&& !secretexit)
	{
		// If no more maps then treat as completed
		if (gamemap == 30 || !haslevelMapxx[gamemap + 1])
		{
			wminfo.show_completed = true;
		}
	}
	// b220-0 Check flags set by MAPINFO (if exists)
	else if (hasmapinfo && !R3D_ScriptedMap)
	{
		if (mapinfo->R3Dflags & R3Dendmaps)
		{
			wminfo.show_completed = true;
		}
		else
		{
			// If exiting one cluster and entering another check for exittext (debrief)
			 if (   (dmt = Def_GetMapInfo(mapinfo->nextmap))
				 && mapinfo->cluster != dmt->cluster
				 && ((mapinfo->cluster > 0 && dmt->cluster > 0)
				 || (!dmt->cluster && W_CheckPWADNumForName(dmt->nextmap) >= 0)) )
			{
				// Actually a debriefing but game has not ended
				if (FI_Briefing(NULL)) return;
			}
		}
	}

	// v220-5 add test to invoke the cast infine script if the game is over
	if (   commercial_game
		&& mapinfo->R3Dflags & R3DendgameC)
	{
		wminfo.show_completed = true;
		wminfo.runCast = true;
	}

	// Debriefings are only run if the game has ended
	if (!R3D_ScriptedMap && wminfo.show_completed)
	{
		if (commercial_game)
			sprintf(wminfo.inflid, "MAP%.2i", gamemap);
		else
			sprintf(wminfo.inflid, "E%iM%i", gameepisode, gamemap);
	}
		
	gameaction = ga_nothing; 
	automap_active = false;
	
	if (mbf_players[0].plr->ingame) 
		G_PlayerFinishLevel(0);        // take away cards and stuff 
	
	wminfo.didsecret = mbf_players[0].didsecret; 
	// wminfo.next starts from 0, unlike gamemap which starts from 1
	wminfo.last = gamemap - 1;
	wminfo.next = gamemap; // default, could be changed
	
	// Check map name
	if ((gmap = MI_NextMap(secretexit, &gameepisode)) > 0)
	{
		wminfo.next = gmap - 1;

		if (!commercial_game && gmap == 9)
			wminfo.show_completed = false;
	}
	else if (commercial_game)
	{
		if (!R3D_ScriptedMap)
		{
			if (secretexit)
			{
				switch(gamemap)
				{
				case 15: wminfo.next = 30; break;
				case 31: wminfo.next = 31; break;
				case 32:
					// v220-7 emulate Doom2 behaviour where a line special 186 has been crossed
					//        in map32 (this would normally be considered a map error)
					if (!haslevelMapxx[gamemap + 1])
						wminfo.next = 31;
					break;
				default:
					if (!hasmapinfo)
					{
						sprintf(ge_inf.nextname, "MAP%02i", gamemap + 1);
						secretexit = 0;
						wminfo.next = 0;
					}
					break;
				}
			}
			else
			{
				switch(gamemap)
				{
				case 30: wminfo.show_completed = true; break;
				case 31:
				case 32: wminfo.next = 15; break;
				}
			}
		}
	}
	else
	{
		if (secretexit) 
			wminfo.next = 8; 	// go to secret level 
		else if (gamemap == 8)
			wminfo.show_completed = true;
		else if (gamemap == 9) 
		{
			// returning from secret level 
			switch (gameepisode) 
			{ 
			case 1: wminfo.next = 3; break; 
			case 2: wminfo.next = 5; break; 
			case 3: wminfo.next = 6; break; 
			case 4: wminfo.next = 2; break;
			}
		}
	}

	if (commercial_game)
	{
		if (wminfo.next + 1 < 99)
			sprintf(wminfo.nextmapid, "MAP%02i", wminfo.next + 1);
		else
		{
			wminfo.show_completed = true;
			strcpy(wminfo.nextmapid, "MAP01");
		}

		if (checkpwads && !wminfo.show_completed)
			wminfo.show_completed = !haslevelMapxx[wminfo.next + 1];
	}
	else
	{
		if (wminfo.next + 1 < 10)
			sprintf(wminfo.nextmapid, "E%iM%i", gameepisode, wminfo.next + 1);
		else
		{
			wminfo.show_completed = true;
			sprintf(wminfo.nextmapid, "E%iM1", gameepisode);
		}
	}

	if (!wminfo.show_completed)
	{
		if ( (dmt = Def_GetMapInfo(wminfo.nextmapid)) )
			strcpy(wminfo.nextmapid, dmt->mapname);
		else
			wminfo.show_completed = true;
	}

	if (!R3D_ScriptedMap)
	{
		wminfo.plyr[0].skills += ma_inf.massacre_count;
	}

	if (mapinfo->R3Dflags & nointermission)
	{
		G_WorldDone();
	}
	else
	{
		gamestate = GS_INTERMISSION; 
		WI_Start(&wminfo);
	}
}

void G_PrepareWIData(void)
{
	int	i;

	wminfo.epsd = gameepisode-1;
	wminfo.maxfrags = 0;
	wminfo.maxkills = totalkills; 
	wminfo.maxitems = totalitems; 
	wminfo.maxsecret = totalsecret; 
	wminfo.partime = mapinfo->partime;

	if (!wminfo.partime)
	{
		if (commercial_game)
		{
			if (gamemap <= 32) wminfo.partime = cpars[gamemap-1];
		}
		else if (gameepisode > 0 && gameepisode < 5 && gamemap < 10)
		{
			wminfo.partime = pars[gameepisode - 1][gamemap];
		}

		if (!wminfo.partime)
			wminfo.partime = 180;	// Default
	}

	wminfo.pnum = 0;

	for (i = 0 ; i < MAXPLAYERS ; i++) 
	{ 
		wminfo.plyr[i].in		= mbf_players[i].plr->ingame; 
		wminfo.plyr[i].skills	= mbf_players[i].killcount; 
		wminfo.plyr[i].sitems	= mbf_players[i].itemcount; 
		wminfo.plyr[i].ssecret	= mbf_players[i].secretcount; 
		wminfo.plyr[i].stime	= leveltime / TICRATE; 
		memcpy(wminfo.plyr[i].frags, mbf_players[i].frags, 
			sizeof(wminfo.plyr[i].frags)); 
	} 
}

//
// G_WorldDone 
//
void G_WorldDone (void) 
{
    gameaction = ga_worlddone; 

    if (secretexit) 
		mbf_players[0].didsecret = true; 
} 

//============================================================================
//
// G_DoWorldDone
//
//============================================================================

static void G_DoWorldDone (void) 
{   
	gamestate = GS_LOADING; 
    gamemap = wminfo.next + 1;
	ge_inf.loadlevel = true;
    G_DoLoadLevel(); 
    gameaction = ga_nothing; 
} 

//==========================================================================
//
// G_DoLoadGame
//
// Called by G_Ticker based on gameaction.
//
//==========================================================================

void G_DoLoadGame (void) 
{ 
    gameaction = ga_nothing; 
	SV_LoadGame();
	redirect_fakeSND();	// v220-10
	checkSCP_ModifyFlags();
} 

//
// G_SaveGame
// Called by the menu task.
// Description points to the slot
//
void G_SaveGame(int slot, char* description) 
{ 
    savegameslot = slot; 
    strcpy (savedescription, description); 
	gameaction = ga_savegame;
} 
 
//==========================================================================
//
// G_DoSaveGame
//
// Called by G_Ticker based on gameaction.
//
//==========================================================================

static void G_DoSaveGame (void) 
{
	SV_SaveGame(savegameslot, savedescription);		

    gameaction = ga_nothing; 
    savedescription[0] = 0;		 
	
    P_SetMessage(mbf_players/* + 0*/, GGSAVED);
} 
 
//
// G_InitNew
// Can be called by the startup code or the menu task,
//
int		d_skill; 
int     d_episode; 
int     d_map; 

// v220-6 type added.
//	0 = cheat
//  1 = Doom  (new game from menu)
//  2 = Doom2 (new game from menu)
//
void G_DeferedInitNew(int skill, int episode, int map) 
{
	d_skill = skill; 
	d_episode = episode; 
	d_map = map;

	con_freezegame = false;
	gameaction = ga_newgame; 
} 

static void G_DoNewGame (void) 
{
	G_InitNew(d_skill, d_episode, d_map); 
	gameaction = ga_nothing; 
} 

// v220-25 game map selection corrected
void G_InitNew(int skill, int episode, int map) 
{ 
	int		i;
	
	gamestate = GS_LOADING;
	
	if (skill < sk_baby) skill = sk_baby;
	if (skill > sk_nightmare) skill = sk_nightmare;
	
	if (gamemode == commercial)
	{
		episode = 0;
		if (map < 1) map = 1;
	}
	else
	{
		if (episode < 1) episode = 1;
		if (map < 0) map = 0;	// ExM0 etc. is allowed
		if (map > 9) map = 9;
		
		if (gamemode == shareware)
		{
			episode = 1;	// only start episode 1 on shareware
			if (map < 1) map = 1;
		}
		else if (episode > 9)
		{
			episode = 9;
		}
	}
	
	ge_inf.first_map = true;
	
	if (haspwads)
	{
		RDBOOL	exists = commercial_game ? haslevelMapxx[map] : haslevelExMx[episode*10 + map];
		
		if (!exists)
		{
			char	name[8], reqname[8], buf[256];
			byte	*ptr = commercial_game ? haslevelMapxx : haslevelExMx;
			int		n = commercial_game ? map : episode*10 + map;
			int		c, i;
			
			for (i = n, c = -1; i < 100; i++)
			{
				if (ptr[i]) { c = i; break; }
			}
			
			if (c < 0)
			{
				for (i = 0, c = -1; i < n; i++)
				{
					if (ptr[i]) { c = i; break; }
				}
			}
			
			strcpy(name, commercial_game ? "MAP01" : "E1M1");
			strcpy(reqname, name);
			
			if (commercial_game)
			{
				sprintf(reqname + 3, "%02d", map);
			}
			else
			{
				*(reqname + 1) = '0' + episode;
				*(reqname + 3) = '0' + map;
			}
			
			if (c >= 0)
			{ 
				if (commercial_game)
				{
					sprintf(name + 3, "%02d", c);
					d_map = map = c;
					d_episode = episode = 1;
				}
				else
				{
					d_map = map = c%10;
					d_episode = episode = c/10;
					*(name + 1) = '0' + d_episode;
					*(name + 3) = '0' + map;
				}
				
			}
			else
			{
				d_map = map = 1;
				d_episode = episode = 1;
			}
			
				sprintf(buf, "\n!!!! Requested or default start map %s is not in pwad. !!!!\n"
				"Defaulting to first %s map %s\n\n"
				"Continue (Y/N)?",
				reqname, c >= 0 ? "pwad" : "iwad", name);
			
			if (!Sys_CriticalMessage(buf, true)) Sys_Quit(false);
		}
	}
	
	G_ClearAddstats();
		
	if (fastparm || (skill == sk_nightmare && gameskill != sk_nightmare) )
	{ 
		for(i=S_SARG_RUN1; i<=S_SARG_RUN8 ; i++) states[i].tics = 1;
		for(i=S_SARG_ATK1; i<=S_SARG_ATK3 ; i++) states[i].tics = 4;
		for(i=S_SARG_PAIN; i<=S_SARG_PAIN2; i++) states[i].tics = 1;

		mobjinfo[MT_BRUISERSHOT].speed = 20*FRACUNIT; 
		mobjinfo[MT_HEADSHOT   ].speed = 20*FRACUNIT; 
		mobjinfo[MT_TROOPSHOT  ].speed = 20*FRACUNIT; 
	} 
	else
	{ 
		for(i=S_SARG_RUN1; i<=S_SARG_RUN8 ; i++) states[i].tics = 2;
		for(i=S_SARG_ATK1; i<=S_SARG_ATK3 ; i++) states[i].tics = 8;
		for(i=S_SARG_PAIN; i<=S_SARG_PAIN2; i++) states[i].tics = 2;

		mobjinfo[MT_BRUISERSHOT].speed = 15*FRACUNIT; 
		mobjinfo[MT_HEADSHOT   ].speed = 10*FRACUNIT; 
		mobjinfo[MT_TROOPSHOT  ].speed = 10*FRACUNIT; 
	} 
	
	// force players to be initialized upon first level load         
	for (i=0 ; i<MAXPLAYERS ; i++) 
	{
		mbf_players[i].playerstate = PST_REBORN; 
		mbf_players[i].didsecret = false;
	}
	
	flying = 0;
	usergame = true;                // will be set false if a demo 
	paused = false; 
	gameepisode = episode; 
	gameskill = skill;
	gamemap = map; 
	
	G_DoLoadLevel(); 
} 

//===========================================================================
// P_Thrust3D
//===========================================================================
void P_Thrust3D(mbf_player_t *player, angle_t angle, float lookdir, 
				int forwardmove, int sidemove)
{
	angle_t pitch = (angle_t)((lookdir * RATIO)/360 * ANGMAX);
	angle_t sideangle = angle - ANG90;
	mobj_t *mo = player->plr->mo;
	int zmul;
	int x, y, z;

	angle >>= ANGLETOFINESHIFT;
	sideangle >>= ANGLETOFINESHIFT;
	pitch >>= ANGLETOFINESHIFT;
	
	x = FixedMul(forwardmove, finecosine[angle]);
	y = FixedMul(forwardmove, finesine[angle]);
	z = FixedMul(forwardmove, finesine[pitch]);
	
	zmul = finecosine[pitch];
	x = FixedMul(x, zmul) + FixedMul(sidemove, finecosine[sideangle]);
	y = FixedMul(y, zmul) + FixedMul(sidemove, finesine[sideangle]);

	mo->momx += x;
	mo->momy += y;
	mo->momz += z;
}

int CCmdSetCamera(int argc, char **argv)
{
	Con_FPrintf(CBLF_CYAN, "Camera not supported\n");
	return false;
}

int CCmdPrintPlayerCoords(int argc, char **argv)
{
	mobj_t *mo = mbf_players[0].plr->mo;

	if(!mo || gamestate != GS_LEVEL) return false;
	Con_FPrintf(CBLF_CYAN, "X=%g Y=%g\n", FIX2FLT(mo->x), FIX2FLT(mo->y));
	return true;
}

int CCmdDynIntensity(int argc, char **argv)
{
	extern float set_intensity;

	set_intensity += 0.05f;
	if (set_intensity > 0.95f) set_intensity = 0;
	// Within rend_model.c 0 is taken as 1
	Con_FPrintf(CBLF_CYAN, "World light Intensity %.2f\n", set_intensity ? set_intensity : 1.0f);
	return true;
}

int CCmdFly(int argc, char **argv)
{
	flying ^= 1;
	Con_FPrintf(CBLF_CYAN, "Fly mode %s\n", flying ? "ON" : "0FF");
	return true;
}

int CCmdEdcam(int argc, char **argv)
{
	if (R3D_edit) compat[c_camera] ^= 1;
	return true;
}

int CCmdActiveMobjs(int argc, char **argv)
{
	Con_FPrintf(CBLF_CYAN, "Number of enemies: Total %i\n", stats.global_mobj_count);

	return true;
}

// v220-4 Map testing mainly
int CCmdKillSelf(int argc, char **argv)
{
	if (  !R3D_edit
		&& ppt.viewplrmo
		&& ppt.viewplrmo->player
		&& ppt.viewplrmo->health > 0)
	{
		Con_FPrintf(CBLF_CYAN, "Commit suicide\n");
		ppt.viewplrmo->player->cheats &= ~(CF_GODMODE|CF_NOCLIP);	// Annul godmode/noclip in case set
		plhealth.armorpoints = 0;						// null armour
		P_DamageMobj(ppt.viewplrmo, NULL, ppt.viewplrmo, ppt.viewplrmo->health, 8);
	}

	return true;
}

#pragma optimize("", on)
