//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 id Software, Inc.,
//  Chi Hoang, Lee Killough, Jim Flynn, Rand Phares, Ty Halderman,
//  Jaakko Kernen, G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_doors.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  BOOM and R3D extensions added, reformatted, moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll code and BOOM v2.01
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// DESCRIPTION: Door animation code (opening/closing)
//
//-----------------------------------------------------------------------------

#include "doom/p_local.h"
#include "doom/p_spec.h"
#include "doom/d_englsh.h"
#include "s_main.h"
#include "m_winmem.h"
#include "r3d_defs.h"


//
// T_VerticalDoor
//
void T_VerticalDoor (vldoor_t* door)
{
    result_e	res;
	
    switch(door->direction)
    {
	case 0:
		// WAITING
		if (!--door->topcountdown)
		{
			switch(door->type)
			{
			case blazeRaise:
			case genBlazeRaise:
				door->direction = -1; // time to go back down
				S_SectorSound(door->sector, sfx_bdcls);
				break;
				
			case normal:
			case genRaise:
				door->direction = -1; // time to go back down
				S_SectorSound(door->sector, sfx_dorcls);
				break;
				
			case close30ThenOpen:
			case genCdO:
				door->direction = 1;
				S_SectorSound(door->sector, sfx_doropn);
				break;
				
			case genBlazeCdO:
				door->direction = 1;  // time to go back up
				S_SectorSound(door->sector, sfx_bdopn);
				break;

			default:
				break;
			}
		}
		break;
		
	case 2:
		//  INITIAL WAIT
		if (!--door->topcountdown)
		{
			switch(door->type)
			{
			case raiseIn5Mins:
				door->direction = 1;
				door->type = normal;
				S_SectorSound(door->sector, sfx_doropn);
				break;
				
			default:
				break;
			}
		}
		break;
		
	case -1:
		// DOWN
		res = T_MovePlane ( door->sector,
							door->speed,
							door->sector->floorheight,
							0,
							1,
							door->direction);

		// killough 10/98: implement gradual lighting effects
		if (door->line && door->lighttag && door->topheight - door->sector->floorheight)
		{
			fixed_t a = door->sector->ceilingheight - door->sector->floorheight;
			fixed_t b = door->topheight - door->sector->floorheight;

			EV_LightTurnOnPartway(door->line, FixedDiv(a, b));
		}

		if (res == pastdest)
		{
			switch(door->type)
			{
			case blazeRaise:
			case blazeClose:
			case genBlazeRaise:
			case genBlazeClose:
				door->sector->ceilingdata = NULL;
				P_RemoveThinker (&door->thinker);  // unlink and free
				// Fixes blaze close sounding twice
				// S_StartSectorSound(door->sector, sfx_bdcls);
				break;
				
			case normal:
			case closeNormal:
			case genRaise:
			case genClose:
				door->sector->ceilingdata = NULL;
				P_RemoveThinker (&door->thinker);  // unlink and free
				break;
				
			case close30ThenOpen:
				door->direction = 0;
				door->topcountdown = 35*30;
				break;
				
			case genCdO:
			case genBlazeCdO:
			door->direction = 0;
			door->topcountdown = door->topwait; // jff 5/8/98 insert delay
				break;

			default:
				break;
			}
		}
		else if (res == crushed)
		{
			switch(door->type)
			{
            case genClose:
            case genBlazeClose:
			case blazeClose:
			case closeNormal:		// DO NOT GO BACK UP!
				break;
				
			default:
				door->direction = 1;
				S_SectorSound(door->sector, door->speed == 8*FRACUNIT ? sfx_bdopn : sfx_doropn);	//v220-35
				break;
			}
		}
		break;
		
	case 1:
		// UP
		res = T_MovePlane ( door->sector,
							door->speed,
							door->topheight,
							0,
							1,
							door->direction);

		// killough 10/98: implement gradual lighting effects
		if (door->line && door->lighttag && door->topheight - door->sector->floorheight)
		{
			fixed_t a = door->sector->ceilingheight - door->sector->floorheight;
			fixed_t b = door->topheight - door->sector->floorheight;

			EV_LightTurnOnPartway(door->line, FixedDiv(a, b));
		}

		// handle door reaching the top
		if (res == pastdest)
		{
			switch(door->type)
			{
			case blazeRaise:
			case normal:
			case genRaise:
			case genBlazeRaise:
				door->direction = 0; // wait at top
				door->topcountdown = door->topwait;
				break;
				
			case close30ThenOpen:
			case blazeOpen:
			case openNormal:
			case genBlazeOpen:
			case genOpen:
			case genCdO:
			case genBlazeCdO:
				door->sector->ceilingdata = NULL;
				P_RemoveThinker (&door->thinker);  // unlink and free
				break;
				
			default:
				break;
			}
		}
		break;
    }
}


//
// EV_DoLockedDoor
// Move a locked door up/down
//

extern int	locked_door;	// v220-7

int
EV_DoLockedDoor
( line_t*	line,
 vldoor_e	type,
 mobj_t*	thing )
{
    mbf_player_t*	p = thing->player;
	
    if (!p) return 0;
	
    switch(line->special)
    {
	case 99:	// Blue Lock
	case 133:
		if (!p->cards[it_bluecard] && !p->cards[it_blueskull])
		{
			locked_door = true;
			P_SetMessage(p, PD_BLUEO);
			S_StartSound(sfx_oof, p->plr->mo);
			return 0;
		}
		break;
		
	case 134: // Red Lock
	case 135:
		if (!p->cards[it_redcard] && !p->cards[it_redskull])
		{
			locked_door = true;
			P_SetMessage(p, PD_REDO);
			S_StartSound(sfx_oof, p->plr->mo);
			return 0;
		}
		break;
		
	case 136:	// Yellow Lock
	case 137:
		if (!p->cards[it_yellowcard] &&
			!p->cards[it_yellowskull])
		{
			locked_door = true;
			P_SetMessage(p, PD_YELLOWO);
			S_StartSound(sfx_oof, p->plr->mo);
			return 0;
		}
		break;	
    }
	
    return EV_DoDoor(line,type);
}

int EV_DoDoor(line_t* line, vldoor_e type)
{
    int			secnum, rtn;
    sector_t*	sec;
    vldoor_t*	door;
	
    secnum = -1;
    rtn = 0;

	// Modified door handling
	if (line->lnflags2 & LNF2_DOORTYPE)
	{
		sec = line->backsector;
		goto door_direct;
	}
    
    while ((secnum = P_FindSectorFromTag(line->tag,secnum)) >= 0)
    {
		sec = &sectors[secnum];

door_direct:

      // if the ceiling already moving, don't start the door action
		if (P_SectorActive(ceiling_special,sec)) //jff 2/22/98
		{
			if (line->lnflags2 & LNF2_DOORTYPE)
			{
				return rtn;
			}
			continue;
		}
		
		// new door thinker
		rtn = 1;
		door = (vldoor_t*)W_Calloc(sizeof(*door), THINKID, 0);
		sec->ceilingdata = door;
		
		door->thinker.function = (void (__cdecl *)(void))T_VerticalDoor;
		door->sector = sec;
		door->type = type;
		door->topwait = VDOORWAIT;
		door->speed = VDOORSPEED;
		door->line = line;  // jff 1/31/98 remember line that triggered us
		door->lighttag = 0; // killough 10/98: no light effects with tagged doors

		P_AddFNThinker (&door->thinker);
		
		switch(type)
		{
		case blazeClose:
			door->topheight = P_FindLowestCeilingSurrounding(sec);
			door->topheight -= 4*FRACUNIT;
			door->direction = -1;
			door->speed = VDOORSPEED * 4;
			S_SectorSound(door->sector, sfx_bdcls);
			break;
			
		case closeNormal:
			door->topheight = P_FindLowestCeilingSurrounding(sec);
			door->topheight -= 4*FRACUNIT;
			door->direction = -1;
			S_SectorSound(door->sector, sfx_dorcls);
			break;
			
		case close30ThenOpen:
			door->topheight = sec->ceilingheight;
			door->direction = -1;
			S_SectorSound(door->sector, sfx_dorcls);
			break;
			
		case blazeRaise:
		case blazeOpen:
			door->direction = 1;
			door->topheight = P_FindLowestCeilingSurrounding(sec);
			door->topheight -= 4*FRACUNIT;
			door->speed = VDOORSPEED * 4;
			if (door->topheight != sec->ceilingheight)
				S_SectorSound(door->sector, sfx_bdopn);
			break;
			
		case normal:
		case openNormal:
			door->direction = 1;
			door->topheight = P_FindLowestCeilingSurrounding(sec);
			door->topheight -= 4*FRACUNIT;
			if (door->topheight != sec->ceilingheight)
				S_SectorSound(door->sector, sfx_doropn);
			break;
			
		default:
			break;
		}

		if (line->lnflags2 & LNF2_DOORTYPE)
		{
			return rtn;
		}
	}
    return rtn;
}

void Con_Printfo(char *format, ...);

//
// EV_VerticalDoor : open a door manually, no tag value
//  v220-5 fixed bug when door opening sound was played
//         when it should have been the closing sound
//
void EV_VerticalDoor(line_t* line, mobj_t* thing)
{
extern RDBOOL	R3D_ScriptedMap;
	mbf_player_t*	player;
	sector_t*		sec;
	vldoor_t*		door;
	int				dir = 1;	// default is up

	//	Check for locks
	player = thing->player;

	switch(line->special)
	{
	case 26: // Blue Lock
	case 32:
		if (!player) return;
		
		if (!player->cards[it_bluecard] && !player->cards[it_blueskull])
		{
			P_SetMessage(player, PD_BLUEK);
			S_StartSound(sfx_oof, player->plr->mo);
			return;
		}
		break;
		
	case 27: // Yellow Lock
	case 34:
		if (!player) return;
		
		if (!player->cards[it_yellowcard] &&
			!player->cards[it_yellowskull])
		{
			P_SetMessage(player, PD_YELLOWK);
			S_StartSound(sfx_oof, player->plr->mo);
			return;
		}
		break;
		
	case 28: // Red Lock
	case 33:
		if (!player) return;
		
		if (!player->cards[it_redcard] && !player->cards[it_redskull])
		{
			P_SetMessage(player, PD_REDK);
			S_StartSound(sfx_oof, player->plr->mo);
			return;
		}
		break;
	}

	// if the wrong side of door is pushed, give oof sound. killough 3/20/98
	if (line->sidenum[1] == -1)
	{
		if (player && player->plr->mo == thing)
			S_StartSound(sfx_oof, thing);
		return;
	}

	sec = NULL;

	// v220-35 Support for same sector sided line trigger
	if (R3D_ScriptedMap && !line->frontsector && !line->backsector)
	{
		void	Con_Error(char *error, ...);
		for (int n = -1; (n = P_FindSectorFromTag(line->tag, n)) >= 0;)
		{
			sec = &sectors[n]; break;
		}
		
		if (!sec) Con_Error("Program error [20]: surrogate line fault with tag %i", line->tag);

		for (n = 1; n < sec->linecount; n++)
		{
			if (sec->lines[n]->lnflags2 & LNF2_BLOCKALL)
				return;
		}
	}

	// OK if that does not work revert to default code.
	if (!sec)
	{
		// get the sector on the second side of activating linedef
		sec = sides[line->sidenum[1]].sector;
	}

	// cph 2001/04/05 - if door already has a thinker, use it
	// b211-0 modded to use retaftersound
	if ( (door = (vldoor_t*)sec->ceilingdata) )
	{
		if (   door->thinker.function != (void (__cdecl *)(void))T_VerticalDoor	// It's not a door
			|| !door->line
			|| !door->line->special)					// It's not retriggerable
		{
			return;
		}
		
		// Check what's doing what
		if (door->direction == -1)
		{
			door->direction = 1;	// go back up
		}
		else if (player)
		{
			door->direction = -1;	// go back down
			dir = 0;
		}
		else
		{
			// We're a monster and don't want to shut it; 
			return;	// exit with no action.
		}
	}

	if (line->special == 117 || line->special == 118)
	{	// BLAZING DOOR TYPES
		S_SectorSound(sec, dir ? sfx_bdopn : sfx_bdcls);
	}
	else
	{	// NORMAL DOOR TYPES
		S_SectorSound(sec, dir ? sfx_doropn : sfx_dorcls);
	}

	if (door) return;

	// new door thinker
	door = (vldoor_t*)W_Calloc(sizeof(*door), THINKID, 0);
	sec->ceilingdata = door;
	door->thinker.function = (void (__cdecl *)(void))T_VerticalDoor;
	door->sector = sec;
	door->direction = 1;
	door->speed = VDOORSPEED;
	door->topwait = VDOORWAIT;
	door->line = line; // jff 1/31/98 remember line that triggered us

	// killough 10/98: use gradual lighting changes if nonzero tag given
	door->lighttag = line->tag; // killough 10/98

	P_AddFNThinker (&door->thinker);

	switch(line->special)
	{
	case 1:
	case 26:
	case 27:
	case 28:
		door->type = normal;
		break;
		
	case 31:
	case 32:
	case 33:
	case 34:
		door->type = openNormal;
		line->special = 0;
		break;
		
	case 117:	// blazing door raise
		door->type = blazeRaise;
		door->speed = VDOORSPEED*4;
		break;
	case 118:	// blazing door open
		door->type = blazeOpen;
		line->special = 0;
		door->speed = VDOORSPEED*4;
		break;
		
	default:
		door->lighttag = 0;   // killough 10/98
		break;
	}

	// find the top and bottom of the movement range
	door->topheight = P_FindLowestCeilingSurrounding(sec);
	door->topheight -= 4*FRACUNIT;
}


//
// Spawn a door that closes after 30 seconds
//
void P_SpawnDoorCloseIn30(sector_t* sec)
{
	vldoor_t*	door;

	door = (vldoor_t*)W_Calloc(sizeof(*door), THINKID, 0);

	sec->ceilingdata = door; //jff 2/22/98
	sec->special = 0;

	door->thinker.function = (void (__cdecl *)(void))T_VerticalDoor;
	door->sector = sec;
	door->direction = 0;
	door->type = normal;
	door->speed = VDOORSPEED;
	door->topcountdown = 30 * 35;
	door->line = NULL; // jff 1/31/98 remember line that triggered us
	door->lighttag = 0;  // killough 10/98: no lighting changes

	P_AddFNThinker (&door->thinker);
}

//
// Spawn a door that opens after 5 minutes
//
void P_SpawnDoorRaiseIn5Mins(sector_t* sec)
{
	vldoor_t*	door;

	door = (vldoor_t*)W_Calloc(sizeof(*door), THINKID, 0);

	sec->ceilingdata = door; //jff 2/22/98
	sec->special = 0;

	door->thinker.function = (void (__cdecl *)(void))T_VerticalDoor;
	door->sector = sec;
	door->direction = 2;
	door->type = raiseIn5Mins;
	door->speed = VDOORSPEED;
	door->topheight = P_FindLowestCeilingSurrounding(sec);
	door->topheight -= 4*FRACUNIT;
	door->topwait = VDOORWAIT;
	door->topcountdown = 5 * 60 * 35;
	door->line = NULL; // jff 1/31/98 remember line that triggered us
	door->lighttag = 0;  // killough 10/98: no lighting changes

	P_AddFNThinker (&door->thinker);
}
