//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 id Software, Inc.,
//  Chi Hoang, Lee Killough, Jim Flynn, Rand Phares, Ty Halderman,
//  G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_genlin.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//
// Based on: Doom, BOOM v2.01
//
// DESCRIPTION:
//  Generalized linedef type handlers
//  Floors, Ceilings, Doors, Locked Doors, Lifts, Stairs, Crushers
//
//-----------------------------------------------------------------------------

#include <stdlib.h>

#include "doom/mbf_data.h"
#include "doom/p_local.h"
#include "doom/p_start.h"
#include "doom/p_spec.h"
#include "doom/d_englsh.h"
#include "doom/g_game.h"
#include "doom/tables.h"
#include "con_console.h"
#include "dd_wad.h"
#include "m_misc.h"
#include "p_maputil.h"
#include "p_sight.h"
#include "r_data.h"
#include "s_main.h"
#include "sys_timer.h"

#include "r3d_MAINspec.h"
#include "r3d_compat.h"

sector_t *P_FindModelFloorSector(fixed_t floordestheight, int secnum); //jff 02/04/98
floormove_t *AssignFloor(sector_t *s, floor_e type, RDBOOL crush);

//====== Utils ======================================================
//
// P_FindNextHighestCeiling()
//
//

fixed_t P_FindNextHighestCeiling(sector_t *sec, int currentheight)
{
	sector_t	*other;
	int			i, height;

	for (i = 0; i < sec->linecount; i++)
	{
		if ((other = getNextSector(sec->lines[i],sec)) &&
			(height = other->ceilingheight) > currentheight)
		{
			while (++i < sec->linecount)
			{
				if ((other = getNextSector(sec->lines[i],sec)) &&
					other->ceilingheight < height &&
					other->ceilingheight > currentheight)
				{
					height = other->ceilingheight;
				}
			}
			return height;
		}
	}
	return currentheight;
}

//
// P_FindModelCeilingSector()
//
// Passed a ceiling height and a sector number, return a pointer to a
// a sector with that ceiling height across the lowest numbered two sided
// line surrounding the sector.
//
// Note: If no sector at that height bounds the sector passed, return NULL
//
sector_t *P_FindModelCeilingSector(fixed_t ceildestheight, int secnum)
{
	sector_t	*sec = NULL;
	int			i;
	
	for (i = 0; i < sectors[secnum].linecount; i++)
	{
		if ( twoSided(secnum, i) )
		{
			if (getSide(secnum, i, 0)->sector - sectors == secnum)
				sec = getSector(secnum, i, 1);
			else
				sec = getSector(secnum, i, 0);
			
			if (sec->ceilingheight == ceildestheight)
				break;
		}
	}
	return sec;
}

//
// P_FindNextLowestCeiling()
//
fixed_t P_FindNextLowestCeiling(sector_t *sec, int currentheight)
{
	sector_t	*other;
	int			i, height;

	for (i = 0; i < sec->linecount; i++)
	{
		if ((other = getNextSector(sec->lines[i],sec)) &&
			(height = other->ceilingheight) < currentheight)
		{
			while (++i < sec->linecount)
			{
				if ((other = getNextSector(sec->lines[i],sec)) &&
					other->ceilingheight > height &&
					other->ceilingheight < currentheight)
				{
					height = other->ceilingheight;
				}
			}
			return height;
		}
	}
	return currentheight;
}

//
// P_FindShortestTextureAround()
//
// Passed a sector number, returns the shortest lower texture on a
// linedef bounding the sector.
//
int P_FindShortestTextureAround(int secnum)
{
	int i, res, minsize = 32000; //prevent ovf

	for (i = 0; i < sectors[secnum].linecount; i++)
	{
		if (twoSided(secnum, i))
		{
			side_t *side = getSide(secnum, i, 0);

			if (side->bottomtexture > 0 ||
				(compat[c_model] && !side->bottomtexture))
			{
				res = texArray[side->bottomtexture]->height;
				if (res < minsize) minsize = res;
			}

			side = getSide(secnum, i, 1);

			if (side->bottomtexture > 0 ||
				(compat[c_model] && !side->bottomtexture))
			{
				res = texArray[side->bottomtexture]->height;
				if (res < minsize) minsize = res;
			}
		}
	}
	return minsize;
}

//
// P_FindShortestUpperAround()
//
// Passed a sector number, returns the shortest upper texture on a
// linedef bounding the sector.
//
int P_FindShortestUpperAround(int secnum)
{
	int	i, res;
	int	minsize = 32000; //prevent ovf

	for (i = 0; i < sectors[secnum].linecount; i++)
	{
		if (twoSided(secnum, i))
		{
			side_t *side = getSide(secnum, i, 0);

			if (side->toptexture > 0)
			{
				res = texArray[side->toptexture]->height;
				if(res < minsize) minsize = res;
			}

			side = getSide(secnum, i, 1);

			if (side->toptexture > 0)
			{
				res = texArray[side->toptexture]->height;
				if(res < minsize) minsize = res;
			}
		}
	}
	return minsize;
}

// Used to prevent height overflow
fixed_t targetMAX(fixed_t startval, int texval)
{
	int res = (startval >> FRACBITS) + texval;

	if (res >  32000) res =  32000;
	if (res < -32000) res = -32000;

	return res <<= FRACBITS;
}


//////////////////////////////////////////////////////////
//
// Generalized Linedef Type handlers
//
//////////////////////////////////////////////////////////

//
// EV_DoGenFloor()
//
// Handle generalized floor types
//
// Passed the line activating the generalized floor function
// Returns true if a thinker is created
//
// jff 02/04/98 Added this routine (and file) to handle generalized
// floor movers using bit fields in the line special type.
//
int EV_DoGenFloor(line_t* line)
{
	floormove_t	*floor;
	sector_t	*sec;
	int			secnum = -1;
	int			rtn = 0;
	RDBOOL		manual = false;
	uint		value = (uint)line->special - GenFloorBase;
	
	// parse the bit fields in the line's special type
	
	int Crsh = (value & FloorCrush    ) >> FloorCrushShift;
	int ChgT = (value & FloorChange   ) >> FloorChangeShift;
	int Targ = (value & FloorTarget   ) >> FloorTargetShift;
	int Dirn = (value & FloorDirection) >> FloorDirectionShift;
	int ChgM = (value & FloorModel    ) >> FloorModelShift;
	int Sped = (value & FloorSpeed    ) >> FloorSpeedShift;
	int Trig = (value & TriggerType   ) >> TriggerTypeShift;
	
	//-----------------------------------------------------

	// check if a manual trigger, if so only do the backside sector.
	if (Trig == PushOnce || Trig == PushMany)
	{
		if ( !(sec = line->backsector) ) return rtn;

		secnum = sec - sectors;
		manual = true;
		goto manual_floor;
	}
	
	// if not manual do all sectors tagged the same as the line
	while ((secnum = P_FindSectorFromTag(line->tag,secnum)) >= 0)
	{
		sec = &sectors[secnum];
		
manual_floor:;			 
		// Do not start another function if floor already moving
		if (P_SectorActive(floor_special, sec))
		{
			if (!manual) continue;

			return rtn;
		}
		
		// new floor thinker
		floor = AssignFloor(sec, genFloor, Crsh != 0);
		floor->direction = Dirn ? 1 : -1;
		floor->texture = sec->floorpic;
		floor->newspecial = sec->special;
		floor->oldspecial = sec->oldspecial;	//jff 3/14/98 transfer old special field too

		rtn = 1;
		
		// set the speed of motion
		// R3D added type check to decrease speed with fake_emitter feature
		switch (Sped)
		{
		case SpeedSlow  : floor->speed = FLOORSPEED  ; break;
		case SpeedNormal: floor->speed = FLOORSPEED*2; break;
		case SpeedFast  : floor->speed = FLOORSPEED*4; break;
		case SpeedTurbo : floor->speed = FLOORSPEED*8; break;
		default:
			break;
		}

		if (sec->specFlags2 & SF2_FAKEMOVSECTOR)	// R3D
			floor->speed >>= 2;	// Floors 4*slower

		// set the destination height
		switch(Targ)
		{
		case FtoHnF:
			floor->floordestheight = P_FindHighestFloorSurrounding(sec);
			break;
		case FtoLnF:
			floor->floordestheight = P_FindLowestFloorSurrounding(sec);
			break;
		case FtoNnF:
			floor->floordestheight = Dirn ?
					P_FindNextHighestFloor(sec,sec->floorheight) :
					P_FindNextLowestFloor (sec,sec->floorheight);
			break;
		case FtoLnC:
			floor->floordestheight = P_FindLowestCeilingSurrounding(sec);
			break;
		case FtoC:
			floor->floordestheight = sec->ceilingheight;
			break;
		case FbyST:
			floor->floordestheight = targetMAX(sec->floorheight,
				floor->direction * P_FindShortestTextureAround(secnum));
			break;
		case Fby24:
			floor->floordestheight = floor->sector->floorheight +
				floor->direction * 24*FRACUNIT;
			break;
		case Fby32:
			floor->floordestheight = floor->sector->floorheight +
				floor->direction * 32*FRACUNIT;
			break;
		default:
			break;
		}
		
		// set texture/type change properties
		if (ChgT)	// if a texture change is indicated
		{
			if (ChgM) // if a numeric model change
			{
				//jff 5/23/98 find model with ceiling at target height if target
				//is a ceiling type
				sec = ( Targ == FtoLnC || Targ == FtoC ) ?
						P_FindModelCeilingSector(floor->floordestheight, secnum) :
						P_FindModelFloorSector(floor->floordestheight, secnum);

				if (sec)
				{
					floor->texture = sec->floorpic;
					switch(ChgT)
					{
					case FChgZero:	// zero type
						floor->newspecial = 0;
						//jff 3/14/98 change old field too
						floor->oldspecial = 0;
						floor->type = genFloorChg0;
						break;
					case FChgTyp:	// copy type
						floor->newspecial = sec->special;
						//jff 3/14/98 change old field too
						floor->oldspecial = sec->oldspecial;
						floor->type = genFloorChgT;
						break;
					case FChgTxt:	// leave type be
						floor->type = genFloorChg;
					default:
						break;
					}
				}
			}
			else	 // else if a trigger model change
			{
				floor->texture = line->frontsector->floorpic;
				switch (ChgT)
				{
				case FChgZero:	  // zero type
					floor->newspecial = 0;
					//jff 3/14/98 change old field too
					floor->oldspecial = 0;
					floor->type = genFloorChg0;
					break;
				case FChgTyp:	  // copy type
					floor->newspecial = line->frontsector->special;
					//jff 3/14/98 change old field too
					floor->oldspecial = line->frontsector->oldspecial;
					floor->type = genFloorChgT;
					break;
				case FChgTxt:	  // leave type be
					floor->type = genFloorChg;
				default:
					break;
				}
			}
		}

		if (manual) break;
	}

	return rtn;
}


//
// EV_DoGenCeiling()
//
// Handle generalized ceiling types
//
// Passed the linedef activating the ceiling function
// Returns true if a thinker created
//
// jff 02/04/98 Added this routine (and file) to handle generalized
// floor movers using bit fields in the line special type.
//
int EV_DoGenCeiling(line_t* line)
{
	ceiling_t	*ceiling;
	sector_t	*sec;
	fixed_t		targheight;
	RDBOOL		manual = false;
	uint		value = (uint)line->special - GenCeilingBase;
	int			secnum = -1;
	int			rtn = 0;
	
	// parse the bit fields in the line's special type
	
	int Crsh = (value & CeilingCrush    ) >> CeilingCrushShift;
	int ChgT = (value & CeilingChange   ) >> CeilingChangeShift;
	int Targ = (value & CeilingTarget   ) >> CeilingTargetShift;
	int Dirn = (value & CeilingDirection) >> CeilingDirectionShift;
	int ChgM = (value & CeilingModel    ) >> CeilingModelShift;
	int Sped = (value & CeilingSpeed    ) >> CeilingSpeedShift;
	int Trig = (value & TriggerType     ) >> TriggerTypeShift;
	
	//-----------------------------------------------------

	// check if a manual trigger, if so do just the sector on the backside
	if (Trig == PushOnce || Trig == PushMany)
	{
		if (!(sec = line->backsector)) return rtn;

		secnum = sec-sectors;
		manual = true;
		goto manual_ceiling;
	}
	
	// if not manual do all sectors tagged the same as the line
	while ((secnum = P_FindSectorFromTag(line->tag,secnum)) >= 0)
	{
		sec = &sectors[secnum];
		
manual_ceiling:;			   
		// Do not start another function if ceiling already moving
		if (P_SectorActive(ceiling_special, sec)) //jff 2/22/98
		{
			if (!manual)
				continue;
			else
				return rtn;
		}
		
		// new ceiling thinker
		rtn = 1;
		ceiling = (ceiling_t*)W_Calloc(sizeof(*ceiling), THINKID, 0);
		sec->ceilingdata = ceiling; //jff 2/22/98
		ceiling->thinker.function = (void (__cdecl *)(void))T_MoveCeiling;
		ceiling->crush = Crsh;
		ceiling->direction = Dirn? 1 : -1;
		ceiling->sector = sec;
		ceiling->texture = sec->ceilingpic;
		ceiling->newspecial = sec->special;
		//jff 3/14/98 change old field too
		ceiling->oldspecial = sec->oldspecial;
		ceiling->tag = sec->tag;
		ceiling->type = (!Dirn && Crsh) ? lowerAndCrush : genCeiling;
		P_AddFNThinker(&ceiling->thinker);
		
		// set speed of motion
		switch (Sped)
		{
		case SpeedSlow:
			ceiling->speed = CEILSPEED;
			break;
		case SpeedNormal:
			ceiling->speed = CEILSPEED*2;
			break;
		case SpeedFast:
			ceiling->speed = CEILSPEED*4;
			break;
		case SpeedTurbo:
			ceiling->speed = CEILSPEED*8;
			break;
		default:
			break;
		}
		
		// set destination target height
		switch(Targ)
		{
		case CtoHnC:
			targheight = P_FindHighestCeilingSurrounding(sec);
			break;
		case CtoLnC:
			targheight = P_FindLowestCeilingSurrounding(sec);
			break;
		case CtoNnC:
			targheight = Dirn?
				P_FindNextHighestCeiling(sec,sec->ceilingheight) :
				P_FindNextLowestCeiling(sec,sec->ceilingheight);
			break;
		case CtoHnF:
			targheight = P_FindHighestFloorSurrounding(sec);
			break;
		case CtoF:
			targheight = sec->floorheight;
			break;
		case CbyST:
			targheight = targetMAX(ceiling->sector->ceilingheight,
				ceiling->direction * P_FindShortestUpperAround(secnum));
			break;
		case Cby24:
			targheight = ceiling->sector->ceilingheight +
				ceiling->direction * 24*FRACUNIT;
			break;
		case Cby32:
			targheight = ceiling->sector->ceilingheight +
				ceiling->direction * 32*FRACUNIT;
			break;
		default:
			break;
		}

		if (Dirn)
			ceiling->topheight = targheight;
		else
			ceiling->bottomheight = targheight;
		
		// set texture/type change properties
		if (ChgT)	  // if a texture change is indicated
		{
			if (ChgM)	// if a numeric model change
			{
				sector_t *sec;
				
				//jff 5/23/98 find model with floor at target height if target
				//is a floor type
				sec = (Targ==CtoHnF || Targ==CtoF)? 		
					P_FindModelFloorSector(targheight,secnum) :
				P_FindModelCeilingSector(targheight,secnum);
				if (sec)
				{
					ceiling->texture = sec->ceilingpic;
					switch (ChgT)
					{
					case CChgZero:	// type is zeroed
						ceiling->newspecial = 0;
						//jff 3/14/98 change old field too
						ceiling->oldspecial = 0;
						ceiling->type = genCeilingChg0;
						break;
					case CChgTyp:	// type is copied
						ceiling->newspecial = sec->special;
						//jff 3/14/98 change old field too
						ceiling->oldspecial = sec->oldspecial;
						ceiling->type = genCeilingChgT;
						break;
					case CChgTxt:	// type is left alone
						ceiling->type = genCeilingChg;
						break;
					default:
						break;
					}
				}
			}
			else		// else if a trigger model change
			{
				ceiling->texture = line->frontsector->ceilingpic;
				switch (ChgT)
				{
				case CChgZero:	  // type is zeroed
					ceiling->newspecial = 0;
					//jff 3/14/98 change old field too
					ceiling->oldspecial = 0;
					ceiling->type = genCeilingChg0;
					break;
				case CChgTyp:	  // type is copied
					ceiling->newspecial = line->frontsector->special;
					//jff 3/14/98 change old field too
					ceiling->oldspecial = line->frontsector->oldspecial;
					ceiling->type = genCeilingChgT;
					break;
				case CChgTxt:	  // type is left alone
					ceiling->type = genCeilingChg;
					break;
				default:
					break;
				}
			}
		}

		P_AddActiveCeiling(ceiling);  // add this ceiling to the active list

		if (manual) return rtn;
	}

	return rtn;
}

//
// EV_DoGenLift()
//
// Handle generalized lift types
//
// Passed the linedef activating the lift
// Returns true if a thinker is created
//
int EV_DoGenLift(line_t *line)
{
	sector_t	*sec;
	plat_t		*plat;

	RDBOOL		manual = false;
	RDBOOL		remote = false;
	uint		value = (uint)line->special - GenLockedBase;
	int			secnum = -1;
	int			rtn = 0;
	
	// parse the bit fields in the line's special type
	
	int Targ = (value & LiftTarget ) >> LiftTargetShift;
	int Dely = (value & LiftDelay  ) >> LiftDelayShift;
	int Sped = (value & LiftSpeed  ) >> LiftSpeedShift;
	int Trig = (value & TriggerType) >> TriggerTypeShift;

	//-----------------------------------------------------

	// Activate all <type> plats that are in_stasis
	if (Targ == LnF2HnF) P_ActivateInStasis(line->tag);
	
	// check if a manual trigger, if so do just the sector on the backside
	if (Trig == PushOnce || Trig == PushMany)
	{
		if (!(sec = line->backsector))
			return rtn;
		secnum = sec-sectors;
		manual = true;
		goto manual_lift;
	}
	
	// if not manual do all sectors tagged the same as the line
	while ((secnum = P_FindSectorFromTag(line->tag,secnum)) >= 0)
	{
		sec = &sectors[secnum];
		
manual_lift:;
		// Do not start another function if floor already moving
		if (P_SectorActive(floor_special, sec))
		{
			if (!manual)
				continue;
			else
				return rtn;
		}
		
		// Setup the plat thinker
		rtn = 1;
		plat = (plat_t*)W_Calloc(sizeof(*plat), THINKID, 0);
		
		plat->sector = sec;
		plat->sector->floordata = plat;
		plat->thinker.function = (void (__cdecl *)(void))T_PlatRaise;
		plat->crush = false;
		plat->tag = line->tag;
		
		plat->type = genLift;
		plat->high = sec->floorheight;
		plat->status = down;
		P_AddFNThinker(&plat->thinker);
		
		// setup the target destination height
		switch(Targ)
		{
		case F2LnF:
			remote = line->tag && !(line->backsector && line->backsector->tag == line->tag);
			plat->low = P_FindLowestFloorSurrounding(sec);
			if (plat->low > sec->floorheight)
				plat->low = sec->floorheight;
			break;
		case F2NnF:
			plat->low = P_FindNextLowestFloor(sec,sec->floorheight);
			break;
		case F2LnC:
			plat->low = P_FindLowestCeilingSurrounding(sec);
			if (plat->low > sec->floorheight)
				plat->low = sec->floorheight;
			break;
		case LnF2HnF:
			plat->type = genPerpetual;
			remote = false;
			plat->low = P_FindLowestFloorSurrounding(sec);
			if (plat->low > sec->floorheight)
				plat->low = sec->floorheight;
			plat->high = P_FindHighestFloorSurrounding(sec);
			if (plat->high < sec->floorheight)
				plat->high = sec->floorheight;
			plat->status = (plat_e)(M_RandomD() & 1);
			break;
		default:
			break;
		}
		
		// setup the speed of motion
		switch(Sped)
		{
		default:
		case SpeedSlow:
			plat->speed = PLATSPEED * 2;
			break;
		case SpeedNormal:
			plat->speed = PLATSPEED * 4;
			break;
		case SpeedFast:
			plat->speed = PLATSPEED * 8;
			break;
		case SpeedTurbo:
			plat->speed = PLATSPEED * 16;
			break;
		}
		
		// setup the delay time before the floor returns
		// b220-0 factors were 1, 3, 5, 10
		switch(Dely)
		{
		case 0:
			plat->wait = remote ? 2*35 : 35;
			break;
		case 1:
			plat->wait = remote ? 4*35 : 3*35;
			break;
		case 2:
			plat->wait = remote ? 6*35 : 5*35;
			break;
		case 3:
			plat->wait = 10*35;
			break;
		}
		
		if (!(sec->specFlags2 & SF2_NOMOVSOUND)) S_SectorSound(sec,sfx_pstart);

		P_AddActivePlat(plat); // add this plat to the list of active plats
		
		if (manual) break;
	}

	return rtn;
}

//
// EV_DoGenStairs()
//
// Handle generalized stair building
//
// Passed the linedef activating the stairs
// Returns true if a thinker is created
//
int EV_DoGenStairs(line_t* line)
{	
	floormove_t	*floor;
	sector_t	*sec;
	sector_t	*tsec;

	fixed_t		stairsize;
	fixed_t		speed;
	
	RDBOOL		mov_valid;
	int			osecnum; //jff 3/4/98 preserve loop index
	int			height;
	int			newsecnum;
	int			texture;
	int			i;

	RDBOOL		manual = false;
	uint		value = (uint)line->special - GenStairsBase;
	int			secnum = -1;
	int			rtn = 0;
	
	// parse the bit fields in the line's special type
	
	int Igno = (value & StairIgnore		) >> StairIgnoreShift;
	int Dirn = (value & StairDirection	) >> StairDirectionShift;
	int Step = (value & StairStep		) >> StairStepShift;
	int Sped = (value & StairSpeed		) >> StairSpeedShift;
	int Trig = (value & TriggerType		) >> TriggerTypeShift;
	int Crsh = (value & FloorCrush		) >> FloorCrushShift;
	
	//-----------------------------------------------------

	// check if a manual trigger, if so only do the sector on the backside
	if (Trig == PushOnce || Trig == PushMany)
	{
		if (!(sec = line->backsector)) return rtn;

		secnum = sec - sectors;
		manual = true;
		goto manual_stair;
	}
	
	// if not manual do all sectors tagged the same as the line
	while ((secnum = P_FindSectorFromTag(line->tag,secnum)) >= 0)
	{
		sec = &sectors[secnum];
		
manual_stair:;	   
		//Do not start another function if floor already moving
		//jff 2/26/98 add special lockout condition to wait for entire
		//staircase to build before retriggering
		if (sec->stairlock || P_SectorActive(floor_special, sec))
		{
			if (!manual) continue;

			return rtn;
		}
		
		rtn = 1;
		
		// new floor thinker
		floor = AssignFloor(sec, genBuildStair, Crsh);
		floor->direction = Dirn ? 1 : -1;
		texture = sec->floorpic;

		// setup stepsize for stairs
		switch(Step)
		{
		default:
		case 0: stairsize =  4*FRACUNIT; break;
		case 1: stairsize =  8*FRACUNIT; break;
		case 2: stairsize = 16*FRACUNIT; break;
		case 3: stairsize = 24*FRACUNIT; break;
		}
		
		height = sec->floorheight + floor->direction * stairsize;
		floor->floordestheight = height;

		// setup speed of stair building
		switch(Sped)
		{
		default:
		case SpeedSlow  : floor->speed = FLOORSPEED/4; break;
		case SpeedNormal: floor->speed = FLOORSPEED/2; break;
		case SpeedFast  : floor->speed = FLOORSPEED*2; break;
		case SpeedTurbo : floor->speed = FLOORSPEED*4; break;
		}
		
		speed = floor->speed;
		
		sec->stairlock = -2;		 // jff 2/26/98 set up lock on current sector
		sec->nextsec = -1;
		sec->prevsec = -1;
		
		osecnum = secnum;			 //jff 3/4/98 preserve loop index
		
		// Find next sector to raise
		// 1.	  Find 2-sided line with same sector side[0]
		// 2.	  Other side is the next sector to raise
		do
		{
			mov_valid = false;

			for (i = 0, mov_valid = false; i < sec->linecount; i++)
			{
				if ( !(tsec = sec->lines[i]->backsector) || tsec->stairlock) continue;
				if (!Igno && tsec->floorpic != texture) continue;
				if (sec != sec->lines[i]->frontsector) continue;
				//jff 2/26/98 special lockout condition for retriggering
				if (P_SectorActive(floor_special,tsec)) continue;
						
				height += floor->direction * stairsize;
				newsecnum = tsec - sectors;
				
				// jff 2/26/98
				// link the stair chain in both directions
				// lock the stair sector until building complete
				sec->nextsec = newsecnum;	// link step to next
				sec = tsec;					// change to new sec
				sec->prevsec = secnum;		// link next back
				sec->nextsec = -1;			// set next forward link as end
				sec->stairlock = -2;		// lock the step
				
				secnum = newsecnum;
				
				floor = AssignFloor(sec, genBuildStair, false);
				floor->direction = Dirn ? 1 : -1;
				floor->speed = speed;
				floor->floordestheight = height;
				
				mov_valid = true;
				break;
			}
		}
		while(mov_valid);

		if (manual) return rtn;

		secnum = osecnum; //jff 3/4/98 restore old loop index
	}

	// retriggerable generalized stairs build up or down alternately
	if (rtn)
		line->special ^= StairDirection; // alternate dir on succ activations

	return rtn;
}

//
// EV_DoGenCrusher()
//
// Handle generalized crusher types
//
// Passed the linedef activating the crusher
// Returns true if a thinker created
//
int EV_DoGenCrusher(line_t *line)
{
	int			secnum;
	int			rtn;
	RDBOOL		manual = false;
	sector_t	*sec;
	ceiling_t	*ceiling;
	uint		value = (uint)line->special - GenCrusherBase;
	
	// parse the bit fields in the line's special type
	
	int Slnt = (value & CrusherSilent) >> CrusherSilentShift;
	int Sped = (value & CrusherSpeed ) >> CrusherSpeedShift;
	int Trig = (value & TriggerType  ) >> TriggerTypeShift;

	//-----------------------------------------------------

	//jff 2/22/98  Reactivate in-stasis ceilings...for certain types.
	//jff 4/5/98 return if activated
	rtn = P_ActivateInStasisCeiling(line);
	
	// check if a manual trigger, if so do just the sector on the backside
	if (Trig==PushOnce || Trig==PushMany)
	{
		if (!(sec = line->backsector))
			return rtn;
		secnum = sec-sectors;
		manual = true;
		goto manual_crusher;
	}
	
	secnum = -1;
	// if not manual do all sectors tagged the same as the line
	while ((secnum = P_FindSectorFromTag(line->tag,secnum)) >= 0)
	{
		sec = &sectors[secnum];
		
manual_crusher:;			   
		// Do not start another function if ceiling already moving
		if (P_SectorActive(ceiling_special, sec)) //jff 2/22/98
		{
			if (!manual)
				continue;
			else
				return rtn;
		}
		
		// new ceiling thinker
		rtn = 1;
		ceiling = (ceiling_t*)W_Calloc(sizeof(*ceiling), THINKID, 0);
		sec->ceilingdata = ceiling; //jff 2/22/98
		ceiling->thinker.function = (void (__cdecl *)(void))T_MoveCeiling;
		ceiling->crush = true;
		ceiling->direction = -1;
		ceiling->sector = sec;
		ceiling->texture = sec->ceilingpic;
		ceiling->newspecial = sec->special;
		ceiling->tag = sec->tag;
		ceiling->type = Slnt? genSilentCrusher : genCrusher;
		ceiling->topheight = sec->ceilingheight;
		ceiling->bottomheight = sec->floorheight + (8*FRACUNIT);
		P_AddFNThinker(&ceiling->thinker);
		
		// setup ceiling motion speed
		switch (Sped)
		{
		case SpeedSlow:
			ceiling->speed = CEILSPEED;
			break;
		case SpeedNormal:
			ceiling->speed = CEILSPEED*2;
			break;
		case SpeedFast:
			ceiling->speed = CEILSPEED*4;
			break;
		case SpeedTurbo:
			ceiling->speed = CEILSPEED*8;
			break;
		default:
			break;
		}
		ceiling->oldspeed = ceiling->speed;
		
		P_AddActiveCeiling(ceiling);  // add to list of active ceilings

		if (manual) return rtn;
	}

	return rtn;
}

//
// EV_DoGenLockedDoor()
//
// Handle generalized locked door types
//
// Passed the linedef activating the generalized locked door
// Returns true if a thinker created
//
int EV_DoGenLockedDoor(line_t *line)
{
	vldoor_t	*door;
	sector_t	*sec;
	RDBOOL		manual = false;
	uint		value = (uint)line->special - GenLockedBase;
	int			secnum = -1;
	int			rtn = 0;

	// parse the bit fields in the line's special type
	
	int Kind = (value & LockedKind ) >> LockedKindShift;
	int Sped = (value & LockedSpeed) >> LockedSpeedShift;
	int Trig = (value & TriggerType) >> TriggerTypeShift;
	
	//-----------------------------------------------------

	// check if a manual trigger, if so use the backside sector
	if (Trig == PushOnce || Trig == PushMany)
	{
		if (!(sec = line->backsector)) return rtn;
		secnum = sec - sectors;
		manual = true;
		goto manual_locked;
	}
	
	// if not manual do all sectors tagged the same as the line
	while ((secnum = P_FindSectorFromTag(line->tag,secnum)) >= 0)
	{
		sec = &sectors[secnum];

manual_locked:;
		// Do not start another function if ceiling already moving
		if (P_SectorActive(ceiling_special, sec)) //jff 2/22/98
		{
			if (!manual) continue;

			return rtn;
		}
		
		// new door thinker
		rtn = 1;
		door = (vldoor_t*)W_Calloc(sizeof(*door), THINKID, 0);
		sec->ceilingdata = door; //jff 2/22/98
		
		door->thinker.function = (void (__cdecl *)(void))T_VerticalDoor;
		door->sector = sec;
		door->topwait = VDOORWAIT;
		door->line = line;
		door->topheight = P_FindLowestCeilingSurrounding(sec);
		door->topheight -= 4*FRACUNIT;
		door->direction = 1;

		P_AddFNThinker (&door->thinker);
		
		// killough 10/98: implement gradual lighting
		door->lighttag = (line->special&6) == 6 && 
			line->special > GenLockedBase ? line->tag : 0;
		
		// setup speed of door motion
		switch(Sped)
		{
		default:
		case SpeedSlow:
			door->type = Kind? genOpen : genRaise;
			door->speed = VDOORSPEED;
			break;
		case SpeedNormal:
			door->type = Kind? genOpen : genRaise;
			door->speed = VDOORSPEED*2;
			break;
		case SpeedFast:
			door->type = Kind? genBlazeOpen : genBlazeRaise;
			door->speed = VDOORSPEED*4;
			break;
		case SpeedTurbo:
			door->type = Kind? genBlazeOpen : genBlazeRaise;
			door->speed = VDOORSPEED*8;
			
			break;
		}
		
		// killough 4/15/98: fix generalized door opening sounds
		// (previously they always had the blazing door close sound)
		
		S_SectorSound(door->sector,   // killough 4/15/98
			door->speed >= VDOORSPEED*4 ? sfx_bdopn : sfx_doropn);
		
		if (manual) break;
	}

	return rtn;
}

//
// EV_DoGenDoor()
//
// Handle generalized door types
//
// Passed the linedef activating the generalized door
// Returns true if a thinker's created
//
int EV_DoGenDoor(line_t *line)
{
	vldoor_t	*door;
	sector_t	*sec;
	RDBOOL		fast;
	RDBOOL		manual = false;
	uint		value = (uint)line->special - GenDoorBase;
	int			secnum = -1;
	int			rtn = 0;

	// extract parameters from the line special's bit fields

	int Dely = (value & DoorDelay  ) >> DoorDelayShift;
	int Kind = (value & DoorKind   ) >> DoorKindShift;
	int Sped = (value & DoorSpeed  ) >> DoorSpeedShift;
	int Trig = (value & TriggerType) >> TriggerTypeShift;

	//-----------------------------------------------------

	// check if a manual trigger, if so do just the sector on the backside
	if (Trig == PushOnce || Trig == PushMany)
	{
		if (!(sec = line->backsector)) return rtn;

		secnum = sec-sectors;
		manual = true;
		goto manual_door;
	}

	// if not manual do all sectors tagged the same as the line
	while ( (secnum = P_FindSectorFromTag(line->tag, secnum)) >= 0)
	{
		sec = &sectors[secnum];

manual_door:;

		// Abort if ceiling's active jff 2/22/98
		if (P_SectorActive(ceiling_special, sec))
		{
			if (!manual) continue;

			return rtn;
		}

		// new door thinker
		rtn = 1;
		door = (vldoor_t*)W_Calloc(sizeof(*door), THINKID, 0);
		door->thinker.function = (void (__cdecl *)(void))T_VerticalDoor;
		door->sector = sec;
		door->line = line;			// jff 1/31/98 remember trigger line
		sec->ceilingdata = door;	// jff 2/22/98 

		P_AddFNThinker (&door->thinker);

		// setup delay for door remaining open/closed
		switch(Dely)
		{
		default:
		case 0: door->topwait = 35			; break;
		case 1: door->topwait = VDOORWAIT	; break;
		case 2: door->topwait = 2*VDOORWAIT	; break;
		case 3: door->topwait = 7*VDOORWAIT	; break;
		}

		// setup speed of door motion
		switch(Sped)
		{
		default:
		case SpeedSlow	: door->speed = VDOORSPEED	; break;
		case SpeedNormal: door->speed = VDOORSPEED*2; break;
		case SpeedFast	: door->speed = VDOORSPEED*4; break;
		case SpeedTurbo	: door->speed = VDOORSPEED*8; break;
		}

		// killough 10/98: implement gradual lighting
		door->lighttag = (line->special & 6) == 6
					&& line->special > GenLockedBase ? line->tag : 0;

		// set door kind: whether opens then closes, just opens or closes etc.
		// and assign target heights accordingly.

		// b211-0 reformatted plus door sound set depending on speed
		if (Kind == CdODoor)
		{
			door->topheight = sec->ceilingheight;
		}
		else
		{
			door->topheight = P_FindLowestCeilingSurrounding(sec);
			door->topheight -= 4*FRACUNIT;
		}

		fast = Sped >= SpeedFast;

		if (Kind == OdCDoor || Kind == ODoor)
		{
			door->direction = 1;

			if (Kind == OdCDoor)
				door->type = fast ? genBlazeRaise : genRaise;
			else
				door->type = fast ? genBlazeOpen : genOpen;

			if (door->topheight != sec->ceilingheight)
				S_SectorSound(door->sector, fast ? sfx_bdopn : sfx_doropn);
		}
		else	// CdODoor, CDoor
		{
			door->direction = -1;

			if (Kind == CdODoor)
				door->type = fast ? genBlazeCdO : genCdO;
			else
				door->type = fast ? genBlazeClose : genClose;

			S_SectorSound(door->sector, fast ? sfx_bdcls : sfx_dorcls);
		}

		if (manual) break;
	}

	return rtn;
}
