//-----------------------------------------------------------------------------
//
// Copyright (C) 1998 - 2005 id Software, Inc., Chi Hoang, Lee Killough,
//   Jim Flynn, Rand Phares, Ty Halderman, Jaakko Kernen
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_inter.c
//
//  Revision 2.1.0 build 62 04/Oct/2006 gmj
//  BOOM and R3D extensions added, reformatted, moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 15 (v220-15) 08/Jun/2012 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 23 (v220-23) 10/Jan/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
// Revision 2.2.0 build 34 (v220-34) 12/Nov/2018 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll code and BOOM v2.01
//
// DESCRIPTION:
//	Handling interactions (i.e., collisions).
//
//-----------------------------------------------------------------------------

#include <memory.h>

#include "doom/p_local.h"
#include "doom/am_map.h"
#include "doom/d_englsh.h"
#include "doom/tables.h"

#include "con_console.h"
#include "dd_wad.h"
#include "m_misc.h"
#include "p_particle.h"
#include "net_main.h"
#include "s_main.h"
#include "r_data.h"
#include "r_main.h"
#include "r_things.h"

#include "r3d_compat.h"
#include "r3d_sl3D.h"

#define BONUSADD	6

// a weapon is found with two clip loads,
// a big item has five clip loads
int	maxammo[NUMAMMO] = {200, 50, 300, 50};
int	clipammo[NUMAMMO] = {10, 4, 20, 1};

extern int dropped_lodlink[3];	// v220-9
extern int scpPickupValue[SPPICKUPLIMIT];	// v220-15/v220-21
extern int mapcheat;

float	playerdamagefactor[MAXDAMAGENUM];
short	playerdamagetype[MAXDAMAGENUM];
int		playerdamageidx;	

int enemyheadshottype[MAXDAMAGENUM][3];
int enemyheadshotidx;

#pragma optimize("g", off)

// CODE ------------------------------------------------------------

//
// P_SetTarget Killough 11/98
//
// This function is used to keep track of pointer references to mobj thinkers.
// In Doom, objects such as lost souls could sometimes be removed despite 
// their still being referenced. In Boom, 'target' mobj fields were tested
// during each gametic, and any objects pointed to by them would be prevented
// from being removed. But this was incomplete, and was slow (every mobj was
// checked during every gametic). Now, we keep a count of the number of
// references, and delay removal until the count is 0.
//

void P_SetTarget(mobj_t **mop, mobj_t *targ)
{
	if (*mop)			// If there was a target already, decrease its refcount
		(*mop)->thinker.mobjreferences--;
	if ((*mop = targ))	// Set new target and if non-NULL, increase its counter
		targ->thinker.mobjreferences++;
}

//
// GET STUFF
//

//
// P_GiveAmmo
// Num is the number of clip loads,
// not the individual count (0= 1/2 clip).
// Returns false if the ammo can't be picked up at all
//

RDBOOL
P_GiveAmmo
(mbf_player_t*	player,
 ammotype_t	ammo,
 int		num )
{
    int		oldammo;
	
    if (ammo == am_noammo)
		return false;
	
    if (ammo < 0 || ammo > NUMAMMO)
		Con_Error ("P_GiveAmmo: bad type %i", ammo);
	
    if ( player->ammo[ammo] == player->maxammo[ammo]  )
		return false;
	
    if (num)
		num *= clipammo[ammo];
    else
		num = clipammo[ammo]/2;
    
    if (gameskill == sk_baby
		|| gameskill == sk_nightmare)
    {
		// give double ammo in trainer or nightmare mode
		num <<= 1;
    }
	
    oldammo = player->ammo[ammo];
    player->ammo[ammo] += num;
	
    if (player->ammo[ammo] > player->maxammo[ammo])
		player->ammo[ammo] = player->maxammo[ammo];
	
    // If non zero ammo, 
    // don't change up weapons,
    // player was lower on purpose.
    if (oldammo)
		return true;	
	
    // We were down to zero,
    // so select a new weapon.
    // Preferences are not user selectable.
    switch (ammo)
    {
	case am_clip:
		if (player->readyweapon == wp_fist)
		{
			if (player->weaponowned[wp_chaingun])
				player->pendingweapon = wp_chaingun;
			else
				player->pendingweapon = wp_pistol;
		}
		break;
		
	case am_shell:
		if (player->readyweapon == wp_fist
			|| player->readyweapon == wp_pistol)
		{
			if (player->weaponowned[wp_shotgun])
				player->pendingweapon = wp_shotgun;
		}
		break;
		
	case am_cell:
		if (player->readyweapon == wp_fist
			|| player->readyweapon == wp_pistol)
		{
			if (player->weaponowned[wp_plasma])
				player->pendingweapon = wp_plasma;
		}
		break;
		
	case am_misl:
		if (player->readyweapon == wp_fist)
		{
			if (player->weaponowned[wp_missile])
				player->pendingweapon = wp_missile;
		}
	default:
		break;
    }
	return true;
}


//
// P_GiveWeapon
// The weapon name may have a MF_DROPPED flag ored in.
//
RDBOOL
P_GiveWeapon
(mbf_player_t*	player,
 weapontype_t	weapon,
 RDBOOL	dropped )
{
    RDBOOL	gaveammo;
    RDBOOL	gaveweapon;
	
    if (weaponinfo[weapon].ammo != am_noammo)
    {
		// give one clip with a dropped weapon,
		// two clips with a found weapon
		if (dropped)
			gaveammo = P_GiveAmmo (player, weaponinfo[weapon].ammo, 1);
		else
			gaveammo = P_GiveAmmo (player, weaponinfo[weapon].ammo, 2);
    }
    else
		gaveammo = false;
	
    if (player->weaponowned[weapon])
		gaveweapon = false;
    else
    {
		gaveweapon = true;
		player->weaponowned[weapon] = true;

		if (cfg.weaponAutoSwitch)
			player->pendingweapon = weapon;
    }
	
    return (gaveweapon || gaveammo);
}

 

//
// P_GiveBody
// Returns false if the body isn't needed at all
//
RDBOOL P_GiveBody(mbf_player_t*	player, int num )
{
    if (plhealth.health >= MAXHEALTH)
		return false;
	
    plhealth.health += num;
    if (plhealth.health > MAXHEALTH)
		plhealth.health = MAXHEALTH;
    player->plr->mo->health = plhealth.health;

    return true;
}



//
// P_GiveArmor
// Returns false if the armor is worse
// than the current armor.
//
RDBOOL
P_GiveArmor
(mbf_player_t*	player,
 int		armortype )
{
    int		hits, i;
	
    //hits = armortype*100;
	i = armortype - 1;
	if(i < 0) i = 0; 
	if(i > 1) i = 1;
	hits = armorpoints[i];
    if (plhealth.armorpoints >= hits)
		return false;	// don't pick up
	
    plhealth.armortype = armortype;
    plhealth.armorpoints = hits;
	
    return true;
}

//
// P_GiveCard
//	v220-4 modified for scripted map
static void P_GiveCard (mbf_player_t *player, card_t card, RDBOOL *scplink)
{
	if (player->cards[card]) 
	{
		*scplink = false;
		return;
	}

	// v220-4 add scripted link
	if (*scplink)
	{
		TS_CheckLinkKey(card);
		*scplink = false;
	}

	player->bonuscount = BONUSADD;
	player->cards[card] = 1;
}

//===========================================================================
// P_GivePower
//   Code revised by Lee Killough
//===========================================================================
RDBOOL P_GivePower(mbf_player_t* player, int power)
{
	static const int tics[NUMPOWERS] = {
		INVULNTICS,	// pw_invulnerability
		1,			// pw_strength
		INVISTICS,	// pw_invisibility
		IRONTICS,	// pw_ironfeet
		1,			// pw_allmap
		INFRATICS	// pw_infrared
	};
	
	switch (power)
	{
	case pw_invisibility:
		player->plr->mo->flags |= MF_SHADOW;
		break;
	case pw_allmap:
		player->powers[pw_allmap] = mapcheat = 1;
		return true;
	case pw_strength:
		P_GiveBody(player, maxhealth);
		break;
	case pw_infrared:
		if (ArgExists("-gmax"))
		{
			player->powers[power] = 0x7fffffff;
			return true;
		}
		break;
	}
	
	// Unless player has infinite duration cheat, set duration (killough)
	if (player->powers[power] >= 0)
		player->powers[power] = tics[power];

	return true;
}

//-------------------------------------------------------------------------

// v220-35	DEHACKED keys can prevent the correct lump rom being
//			displayed in the minimised hud
static spritelump_t DEHkeysplumps[6];
static byte keyisset[6];

static void subfn_SetKeyLump(int num, int sprnum, char *name)
{
	int lump;

	keyisset[num] = 0;

	if ( (lump = W_CheckIWADNumForName(name)) >= 0)
	{
		spritedef_t		*sprdef   = &sprites[sprnum];
		spriteframe_t	*sprframe = &sprdef->spriteframes[0];
		spritelump_t	*sprlump;
		lumpinfo_t		*l = lumpinfo + lump;
		patch_t			hdrcache;

		lseek(l->handle, l->position, SEEK_SET);
		read(l->handle, &hdrcache, 8);
		
		sprlump = spritelumps + sprframe->lump[0];
		memcpy(&DEHkeysplumps[num], sprlump, sizeof(spritelump_t));
		memset(DEHkeysplumps[num].color.rgb, 0xff, 4);

		DEHkeysplumps[num].lump   = lump;
		DEHkeysplumps[num].width  = hdrcache.width;
		DEHkeysplumps[num].height = hdrcache.height;
	}
}

void P_PresetIWADkeyLumps()
{
	subfn_SetKeyLump(0, SPR_BKEY, "BKEYA0");
	subfn_SetKeyLump(1, SPR_YKEY, "YKEYA0");
	subfn_SetKeyLump(2, SPR_RKEY, "RKEYA0");
	subfn_SetKeyLump(3, SPR_BSKU, "BSKUA0");
	subfn_SetKeyLump(4, SPR_YSKU, "YSKUA0");
	subfn_SetKeyLump(5, SPR_RSKU, "RSKUA0");
}

//-------------------------------------------------------------------------

// v220-35 Deal with dehacked wandering keys where dehacking
//         with old wads, has led to anomalous behaviour. In
//		   other ports keys are not displayed in a minimised
//		   hud so they do not have this problem.
void P_TouchDEHKeyThing(mobj_t *mo)
{
    mbf_player_t	*player = &mbf_players[0];
	card_t			card;
	char			*msg;
	int				sprnum = mo->sprite;
	int				num, scplink;

	switch(sprnum)
	{
		case SPR_BKEY: num = 0; card = it_bluecard   ; msg = GOTBLUECARD; break;
		case SPR_YKEY: num = 1; card = it_yellowcard ; msg = GOTYELWCARD; break;
		case SPR_RKEY: num = 2; card = it_redcard    ; msg = GOTREDCARD ; break;
		case SPR_BSKU: num = 3; card = it_blueskull  ; msg = GOTBLUESKUL; break;
		case SPR_YSKU: num = 4; card = it_yellowskull; msg = GOTYELWSKUL; break;
		case SPR_RSKU: num = 5; card = it_redskull   ; msg = GOTREDSKULL; break;
		default: return;
	}

	if (!player->cards[card])
	{
		P_SetMessage(player, msg);
		P_GiveCard(player, card, &scplink);

		// Display something sensible in the minimised HUD
		if (!keyisset[num])
		{
			spritedef_t		*sprdef   = &sprites[sprnum];
			spriteframe_t	*sprframe = &sprdef->spriteframes[0];

			if (sprframe)
			{
				spritelump_t	*sprlump = spritelumps + sprframe->lump[0];
				memcpy(sprlump, &DEHkeysplumps[num], sizeof(spritelump_t));
			}
			
			keyisset[num] = true;
		}
	}

	P_RemoveMobj(mo, 32);
}

//
// P_TouchSpecialThing
//  b211-0 blocked status field added
//  v220-16 modified to use new function P_CheckUnblocked
//

void P_TouchSpecialThing(mobj_t* special)
{
    mbf_player_t	*player = gdt.tmthing->player;
	mobj_t			*toucher = gdt.tmthing;

    int		i;
    int		sound;
	int		oldhealth, addhealth = 0;
	RDBOOL	removePU = true;
	RDBOOL	scplink = R3D_ScriptedMap;
	RDBOOL	scripted_pickup = false;

	//-----------------------------------

	if (!player || R3D_edit) return;

	//  v220-16  Check for lines that fully occlude a pickup.
	if (   R3D_ScriptedMap
		&& gdt.tmthing->subsector != special->subsector
		&& !P_CheckUnblocked(special, toucher))
	{
		return;
	}

	sound = sfx_itemup;

	// v220-15
	if (R3D_ScriptedMap)
	{
		int num = special->info->doomednum - 7690;	// v220-20 base value changed from 7692 to 7690

		if (num >= 0 && num < SPPICKUPLIMIT)
		{
			int add_val = scpPickupValue[num];

			switch(num)
			{
			case 0: P_SetMessage(player, GOTSCPARM1); break;
			case 1: P_SetMessage(player, GOTSCPARM2); break;
			case 2: P_SetMessage(player, GOTSCPARM3); break;
			case 3: P_SetMessage(player, GOTSCPARM4); break;
			case 4: P_SetMessage(player, GOTSCPHLT1); break;
			case 5: P_SetMessage(player, GOTSCPHLT2); break;
			case 6: P_SetMessage(player, GOTSCPHLT3); break;
			case 7: P_SetMessage(player, GOTSCPHLT4); break;
			case 8: P_SetMessage(player, GOTSCPHLT5); break;
			case 9: P_SetMessage(player, GOTSCPHLT6); break;
			}

			if (num < 4)	// armour
			{
				plhealth.armorpoints += add_val;
				if (plhealth.armorpoints > 200) plhealth.armorpoints = 200;

				if (plhealth.armorpoints <= 0)
				{
					plhealth.armorpoints = 0;
					plhealth.armortype = 0;
				}
				else
				{
					plhealth.armortype = plhealth.armorpoints > 100 ? 2 : 1;
				}
			}
			else	// health
			{
				plhealth.health += add_val;
				scripted_pickup = plhealth.health <= 0;
				if (plhealth.health > healthlimit) plhealth.health = healthlimit;
				if (plhealth.health > 0) player->plr->mo->health = plhealth.health;
			}

			goto remove_pickup;
		}
	}

    // Identify by sprite.
	switch (special->sprite)
    {
		// armor
	case SPR_ARM1:
		if (!P_GiveArmor (player, 1))
			removePU = false;
		else
			P_SetMessage(player, GOTARMOR);
		break;
		
	case SPR_ARM2:
		if (!P_GiveArmor (player, 2))
			removePU = false;
		else
			P_SetMessage(player, GOTMEGA);
		break;
		
		// bonus items
	case SPR_BON1:
		plhealth.health++;		// can go over 100%
		if (plhealth.health > healthlimit)
			plhealth.health = healthlimit;
		player->plr->mo->health = plhealth.health;
		P_SetMessage(player, GOTHTHBONUS);
		break;
		
	case SPR_BON2:
		plhealth.armorpoints++;		// can go over 100%
		if (plhealth.armorpoints > armorpoints[1]) // 200
			plhealth.armorpoints = armorpoints[1];
		if (!plhealth.armortype)
			plhealth.armortype = 1;
		P_SetMessage(player, GOTARMBONUS);
		break;
		
	case SPR_SOUL:
		plhealth.health += miscDEHval[SoulsphereHealth];
		if (plhealth.health > healthlimit)
			plhealth.health = healthlimit;
		player->plr->mo->health = plhealth.health;
		P_SetMessage(player, GOTSUPER);
		sound = sfx_getpow;
		break;
		
	case SPR_MEGA:
		if (!commercial_game) return;
		plhealth.health = miscDEHval[MegasphereHealth];
		player->plr->mo->health = plhealth.health;
		P_GiveArmor (player,2);
		P_SetMessage(player, GOTMSPHERE);
		sound = sfx_getpow;
		break;
		
		// cards
		// leave cards for everyone
	case SPR_BKEY:
		if (R3D_ScriptedMap && player->cards[it_bluecard]) return;	// v220-23 don't pickup if already in player's possession
		if (!player->cards[it_bluecard]) P_SetMessage(player, GOTBLUECARD);
		P_GiveCard (player, it_bluecard, &scplink);
		break;
		
	case SPR_YKEY:
		if (R3D_ScriptedMap && player->cards[it_yellowcard]) return;	// v220-23
		if (!player->cards[it_yellowcard]) P_SetMessage(player, GOTYELWCARD);
		P_GiveCard (player, it_yellowcard, &scplink);
		break;
		
	case SPR_RKEY:
		if (R3D_ScriptedMap && player->cards[it_redcard]) return;	// v220-23
		if (!player->cards[it_redcard]) P_SetMessage(player, GOTREDCARD);
		P_GiveCard (player, it_redcard, &scplink);
		break;
		
	case SPR_BSKU:
		if (R3D_ScriptedMap && player->cards[it_blueskull]) return;	// v220-23
		if (!player->cards[it_blueskull]) P_SetMessage(player, GOTBLUESKUL);
		P_GiveCard (player, it_blueskull, &scplink);
		break;
		
	case SPR_YSKU:
		if (R3D_ScriptedMap && player->cards[it_yellowskull]) return;	// v220-23
		if (!player->cards[it_yellowskull]) P_SetMessage(player, GOTYELWSKUL);
		P_GiveCard (player, it_yellowskull, &scplink);
		break;
		
	case SPR_RSKU:
		if (R3D_ScriptedMap && player->cards[it_redskull]) return;	// v220-23
		if (!player->cards[it_redskull]) P_SetMessage(player, GOTREDSKULL);
		P_GiveCard (player, it_redskull, &scplink);
		break;
		
		// medikits, heals (b220-0 fixed doom bug)
	case SPR_MEDI:
		addhealth = 25;
	case SPR_STIM:
		if (!addhealth) addhealth = 10;
		oldhealth = plhealth.health;
		
		if (!P_GiveBody (player, addhealth))
			removePU = false;
		else
			P_SetMessage(player, oldhealth < 25 ? GOTMEDINEED :
						addhealth == 10 ? GOTSTIM : GOTMEDIKIT);
		break;
		
		// power ups
	case SPR_PINV:
		if (!P_GivePower (player, pw_invulnerability))
			removePU = false;
		else
		{
			P_SetMessage(player, GOTINVUL);
			sound = sfx_getpow;
		}
		break;
		
	case SPR_PSTR:
		if (!P_GivePower (player, pw_strength))
			removePU = false;
		else
		{
			P_SetMessage(player, GOTBERSERK);
			if (player->readyweapon != wp_fist)
				player->pendingweapon = wp_fist;
			sound = sfx_getpow;
		}
		break;
		
	case SPR_PINS:
		if (!P_GivePower (player, pw_invisibility))
			removePU = false;
		else
		{
			P_SetMessage(player, GOTINVIS);
			sound = sfx_getpow;
		}
		break;
		
	case SPR_SUIT:
		if (!P_GivePower (player, pw_ironfeet))
			removePU = false;
		else
		{
			P_SetMessage(player, GOTSUIT);
			sound = sfx_getpow;

			if (ma_inf.radtime)	// v220-35
				player->powers[pw_ironfeet] = (int)((float)(IRONTICS)*ma_inf.radtime);
		}
		break;
		
	case SPR_PMAP:
		if (!P_GivePower (player, pw_allmap))
			removePU = false;
		else
		{
			P_SetMessage(player, GOTMAP);
			sound = sfx_getpow;
		}
		break;
		
	case SPR_PVIS:
		if (!P_GivePower (player, pw_infrared))
			removePU = false;
		else
		{
			P_SetMessage(player, GOTVISOR);
			sound = sfx_getpow;
		}
		break;
		
		// ammo
	case SPR_CLIP:
		if (special->flags & MF_DROPPED)
		{
			if (!P_GiveAmmo (player,am_clip,0))
				removePU = false;
		}
		else
		{
			if (!P_GiveAmmo (player,am_clip,1))
				removePU = false;
		}
		if (removePU)
			P_SetMessage(player, GOTCLIP);
		break;
		
	case SPR_AMMO:
		if (!P_GiveAmmo (player, am_clip,5))
			removePU = false;
		else
			P_SetMessage(player, GOTCLIPBOX);
		break;
		
	case SPR_ROCK:
		if (!P_GiveAmmo (player, am_misl,1))
			removePU = false;
		else
			P_SetMessage(player, GOTROCKET);
		break;
		
	case SPR_BROK:
		if (!P_GiveAmmo (player, am_misl,5))
			removePU = false;
		else
			P_SetMessage(player, GOTROCKBOX);
		break;
		
	case SPR_CELL:
		if (!P_GiveAmmo (player, am_cell,1))
			removePU = false;
		else
			P_SetMessage(player, GOTCELL);
		break;
		
	case SPR_CELP:
		if (!P_GiveAmmo (player, am_cell,5))
			removePU = false;
		else
			P_SetMessage(player, GOTCELLBOX);
		break;
		
	case SPR_SHEL:
		if (!P_GiveAmmo (player, am_shell,1))
			removePU = false;
		else
			P_SetMessage(player, GOTSHELLS);
		break;
		
	case SPR_SBOX:
		if (!P_GiveAmmo (player, am_shell,5))
			removePU = false;
		else
			P_SetMessage(player, GOTSHELLBOX);
		break;
		
	case SPR_BPAK:
		if (!player->backpack)
		{
			for (i=0 ; i<NUMAMMO ; i++)
				player->maxammo[i] *= 2;
			player->backpack = true;
		}
		for (i=0 ; i<NUMAMMO ; i++)
			P_GiveAmmo (player, (ammotype_t)i, 1);
		P_SetMessage(player, GOTBACKPACK);
		break;
		
		// weapons
	case SPR_BFUG:
		if (!P_GiveWeapon (player, wp_bfg, false) )
			removePU = false;
		else
		{
			P_SetMessage(player, GOTBFG9000);
			sound = sfx_wpnup;
		}
		break;
		
	case SPR_MGUN:
		if (!P_GiveWeapon (player, wp_chaingun, (RDBOOL)((special->flags & MF_DROPPED) > 0)) )
			removePU = false;
		else
		{
			P_SetMessage(player, GOTCHAINGUN);
			sound = sfx_wpnup;
		}
		break;
		
	case SPR_CSAW:
		if (!P_GiveWeapon (player, wp_chainsaw, false) )
			removePU = false;
		else
		{
			P_SetMessage(player, GOTCHAINSAW);
			sound = sfx_wpnup;
		}
		break;
		
	case SPR_LAUN:
		if (!P_GiveWeapon (player, wp_missile, false) )
			removePU = false;
		else
		{
			P_SetMessage(player, GOTLAUNCHER);
			sound = sfx_wpnup;
		}
		break;
		
	case SPR_PLAS:
		if (!P_GiveWeapon (player, wp_plasma, false) )
			removePU = false;
		else
		{
			P_SetMessage(player, GOTPLASMA);
			sound = sfx_wpnup;
		}
		break;
		
	case SPR_SHOT:
		if (!P_GiveWeapon (player, wp_shotgun, (RDBOOL)((special->flags & MF_DROPPED) > 0)) )
			removePU = false;
		else
		{
			P_SetMessage(player, GOTSHOTGUN);
			sound = sfx_wpnup;
		}
		break;
		
	case SPR_SGN2:
		if (!P_GiveWeapon (player, wp_supershotgun, (RDBOOL)((special->flags & MF_DROPPED) > 0)) )
			removePU = false;
		else
		{
			P_SetMessage(player, GOTSHOTGUN2);
			sound = sfx_wpnup;
		}
		break;
		
	default:
		if (special->flags3 & MF3_SPECIAL)
		{
			extern int customPunch;

			player->readyweapon = player->pendingweapon = wp_fist; 
			player->weaponowned[wp_fist] = true;

			// TODO: to fully implement requires wp_xx slots and script
			//       and, if more than one def, extended CUSTOMSPECIAL defs
			//       Nothing requires it at the current time, however, other
			//       than the COD Holy Cross
			scpMessage.scriptMessage = CUSTOMSPECIAL1;
			scpMessage.delay_to_next_text = 0;
			scpMessage.delay_to_fade = 100;
			scpMessage.alpha = 1.0f;
			customPunch = ge_inf.customhits;
			sound = sfx_getpow;
			break;
		}

		if (special->state)
			Con_Error("P_TouchSpecialThing: Unknown gettable thing%i type %i sprite %i state %i sprite %i flags %.8X misc %.8X\n",
				MAPNUM(special), special->type,
				special->sprite,
				special->state - states,
				special->state->sprite,
				special->flags, special->misc);
		else
			Con_Error("P_TouchSpecialThing: Unknown gettable thing");
    }

remove_pickup:;

	if (removePU)
	{
		// v220-3 secret pickups in scripted map 
		if (special->flags3 & MF3_SECRET)
		{
			player->secretcount++;

			if(cfg.secretMsg)
			{
				P_SetMessage(player, "You've found a secret pickup!");
				S_ConsoleSound(sfx_getpow, NULL, player - mbf_players);
			}
		}

		if (special->flags & MF_COUNTITEM)
			player->itemcount++;

		// v220-4 add scripted link
		if (scplink && special->misc & MJF_SCP_PICKUP)
			TS_CheckLinkSpecial(special);

		P_RemoveMobj(special, 32);

		player->bonuscount += BONUSADD;

		if ( !(special->flags2 & MF2_NORENDER) )
			S_ConsoleSound(sound, NULL, player - mbf_players);

		if (scripted_pickup && player->plr->mo->health > 0)
		{
			if (!ge_inf.devmap) player->cheats &= ~CF_GODMODE;
			player->powers[pw_invulnerability] = 0;
			plhealth.armorpoints = 0;
			plhealth.armortype = 0;
			P_DamageMobj(player->plr->mo, NULL, NULL, player->plr->mo->health, 19);
		}
	}
}


//===========================================================================
// checkFloorFlats3D
//===========================================================================
static int checkFloorFlats3D(mobj_t *mo)
{
	int *flatslist = mo->subsector->flats3DArray;

	if (flatslist)
	{
		fixed_t	floor, z = mo->floorz + mo->height;
		int		count = *(int*)flatslist++;

		flatslist += (count - 1)*FLIST_ENTRIES;

		// Remember - floor's higher than ceiling
		while (--count >= 0)
		{
			floor = sectors[flatslist[1]].floorheight +
							flatslist[2] +	// offset
							flatslist[3];	// depth

			if (z >= floor) return floor;

			flatslist -= FLIST_ENTRIES;
		}
	}
	return ONFLOORZ;
}

// target must be set
void P_KillMobj(mobj_t *source, mobj_t *target)
{
	mobjtype_t	item;
	mobj_t		*mo;
	int			x, y, z = ONFLOORZ;
	int			droplod = 0;	// v220-9
	RDBOOL		scriptedtype = false;
	
	if (target->flags3 & MF3_RESURRECT)
	{
		P_SetMobjState (target, target->info->raisestate);
		target->height = target->info->height;
		target->health = target->info->spawnhealth;
		
		if ( !(source && source->player) )
		{
			P_SetTarget(&target->target, NULL);
			P_SetTarget(&target->lastenemy, NULL);
			target->flags &= ~MF_JUSTHIT;
		}
		
		P_UpdateThinker(&target->thinker);
		target->flags3 &= ~MF3_RESURRECT;	// One time only
		return;
	}
	
	target->flags &= ~(MF_SHOOTABLE|MF_FLOAT|MF_SKULLFLY);
	target->height >>= 2;	// v220-33 this wrongly got confined to just the player
	
	if (target->type != MT_SKULL)
		target->flags &= ~MF_NOGRAVITY;
	
	// R3D: bossbrains don't die - they tick to the exit
	if (target->type != MT_BOSSBRAIN)
		target->flags |= (MF_CORPSE|MF_DROPOFF);
		
	P_UpdateThinker(&target->thinker);	// killough 8/29/98
	
	// Set counts for intermission
	// v220-4 add collateral damage count
	
	if (   source
		&& source->flags3 & MF3_FRIEND
		&& target->flags3 & MF3_FRIEND
		&& target->flags & MF_COUNTKILL)
	{
		ma_inf.collateraldamage_count++;
	}
	else if (source && source->player)
	{
		if (target->flags & MF_COUNTKILL)
			source->player->killcount++;
		
		if (target->player)
			source->player->frags[target->player - mbf_players]++;
	}
	else if (target->flags & MF_COUNTKILL
		&& !(target->flags3 & MF3_FRIEND))
	{
		// enemy kills
		mbf_players[0].killcount++;
	}
	
	if (target->player)
	{
		// count environment kills against you
		if (!source)	
			target->player->frags[target->player - mbf_players]++;
		
		target->flags &= ~MF_SOLID;
		target->player->playerstate = PST_DEAD;
		target->player->plr->flags |= DDPF_DEAD;
		
		if (target->player == &mbf_players[0])
		{
			P_DropWeapon (target->player);
			gdt.onladder = false;
			automap_active = false;
			ma_inf.playerdead = R3D_ScriptedMap;
		}
	}

	if (   target->info->xdeathstate
		&& target->health < -target->info->spawnhealth)
	{
		P_SetMobjState (target, target->info->xdeathstate);
	}
	else
	{
		P_SetMobjState (target, target->info->deathstate);
	}
	
	target->health = 0;
	target->tics -= (M_RandomD() & 3);
	
	if (target->tics < 1 || (ge_inf.grenade_id && ge_inf.grenade_id == target->type))
	{
		target->tics = 1;
	}
	
	// R3D cod engine support ----------------
	if (ge_inf.addon_flags & AF_COD)
	{
		if (target->flags4 & MF4_BADASSZOMBIE)
		{
			int t;
			
			if ( (t = P_GetTypeFromDoomednum(2010)) < stats.num_mobjs)
			{
				int x = target->x + ((M_RandomD() - M_RandomD()) << 12);
				int y = target->y + ((M_RandomD() - M_RandomD()) << 12);

				// drop a rocket
				mo = P_SpawnMobj(target->x, target->y, target->z + target->height/2, (mobjtype_t)t, 0);

				P_SetMpos(mo);
			}
			return;
		}
		else
		if (target->flags4 & MF4_BOSSSPID)
		{
			int t;
			
			if ( (t = P_GetTypeFromDoomednum(68)) < stats.num_mobjs)
			{
				int x = target->x + ((M_RandomD() - M_RandomD()) << 12);
				int y = target->y + ((M_RandomD() - M_RandomD()) << 12);

				// spawn an arachnotron
				mo = P_SpawnMobj(target->x, target->y, target->z, (mobjtype_t)t, 0);

				P_SetMpos(mo);
			}
			return;
		}
		else
		if (target->flags4 & MF4_LOSTSOUL)
		{
			int t;
			
			if ( (t = P_GetTypeFromDoomednum(3006)) < stats.num_mobjs)
			{
				// spawn 3006 lost soul
				mo = P_SpawnMobj(target->x, target->y, target->z + target->height, (mobjtype_t)t, 0);

				P_SetMpos(mo);
			}
			return;
		}
	
		if (target->flags4 & MF4_CODENDTYPE)
		{
			// Scan all thinkers to see if this was the last codtype one
			// If so lower a sector with a tag related to the mobj's doomednum
			thinker_t	*th;
			int			dmn = target->info->doomednum;
			RDBOOL		found = false;
			
			for (th = thinkercap.next; th != &thinkercap ;th = th->next)
			{
				if (th->function == (void (__cdecl *)(void))P_MobjThinker
					&& (mo = (mobj_t*)th)->flags4 & MF4_CODENDTYPE
					&& mo->info->doomednum == dmn
					&& mo->health > 0
					&& mo != target)
				{
					found = true; break;
				}
			}
			
			if (!found)
			{
				line_t ld;
				
				ld.tag = 0;
				
				switch (dmn)
				{
				case 3004: ld.tag = 666; break;	// zombie
				case    9: ld.tag = 667; break;	// shotgun guy
				case   58: ld.tag = 668; break;	// spectre
				case 3001: ld.tag = 669; break;	// imp
				case 3002: ld.tag = 670; break;	// demon
				case 3006: ld.tag = 671; break;	// lost soul
				case 3005: ld.tag = 672; break;	// cacodemon
				case 3003: ld.tag = 673; break;	// baron
				case   16: ld.tag = 674; break;	// cyberdemon
				case    7: ld.tag = 675; break;	// spiderboss
				case   65: ld.tag = 676; break;	// chaingun bob
				case   69: ld.tag = 677; break;	// hell knight
				case   68: ld.tag = 678; break;	// arachnotron
				case   71: ld.tag = 679; break;	// pain elemental
				case   66: ld.tag = 680; break;	// revenant
				case   67: ld.tag = 681; break;	// mancubus
				default: break;
				}
				
				if (ld.tag)
					EV_DoFloor(&ld, lowerFloorToLowest);
			}
		}
	}

	// ---------------------------------------
	
	if (ma_inf.noItemDrop) return;	// v220-31 don't drop items with massacre cheat

	// Drop stuff. b211-0 - some code tidying from this point
	
	// b211-0 Check for scripted dropitem
	if ( (item = (mobjtype_t)FetchDropItem(target->info->doomednum)) <= 0)
	{
		if (!item) return;	// Nothing was dropped by Thing
		
		// Standard types
		if (target->type == MT_WOLFSS || target->type == MT_POSSESSED)
		{
			item = MT_CLIP;
			droplod = dropped_lodlink[0];
		}
		else if (target->type == MT_SHOTGUY)
		{
			item = MT_SHOTGUN;
			droplod = dropped_lodlink[1];
		}
		else if (target->type == MT_CHAINGUY)
		{
			item = MT_CHAINGUN;
			droplod = dropped_lodlink[2];
		}
		else
			return;
		
		// Drop item slightly off position to prevent sprite Z flickering 
		x = target->x + ((M_RandomD() - M_RandomD()) << 12);
		y = target->y + ((M_RandomD() - M_RandomD()) << 12);
		z = ONFLOORZ;
	}
	else	// item > 0
	{ 
		mobjinfo_t	*info = &mobjinfo[item];
		
		x = target->x, y = target->y;
		
		validcount++;
		
		if (   info->flags & MF_COUNTKILL
			&& spotIsOccupied(x, y, 0, info->radius, NULL, 0))
		{
			return;	// Can't be done
		}
		
		scriptedtype = true;
		
		if (mobjinfo[item].flags & MF_SPAWNCEILING)
			z = ONCEILINGZ;
		else
			z = ONFLOORZ;
	}
	
	mo = P_SpawnMobj(x, y, z, item, 0);	// spawn item
	mo->lodstate = droplod;	// v220-9
	
	if (z == ONFLOORZ && (z = checkFloorFlats3D(target)) > ONFLOORZ)
	{
		mo->z = mo->floorz = mo->dropoffz = z;
	}
	
	P_SetMpos(mo);

	if (scriptedtype)
	{
		if (mo->flags & MF_COUNTKILL)
		{
			mobj_t *smo = P_SpawnMobj (mo->x, mo->y, mo->z, MT_TFOG, 0);

			P_SetMpos(smo);
			S_StartSound(sfx_telept, smo);

			mo->iden_nums = target->iden_nums;	// inherit target's idens
			mo->angle = target->angle;	// v220-21 inherit target's angle
		}
		else if ( !(mo->flags & MF_SPECIAL) )
		{
			mo->flags |= MF_DROPPED;	// special versions of items
			mo->pickupRadius = mobjinfo[mo->type].radius;
		}
	}
	else
	{
		mo->flags |= MF_DROPPED;	// special versions of items
	}
}

//===========================================================================
// P_DamageMobj
//	Damages both enemies and players
//	"inflictor" is the thing that caused the damage
//	 creature or missile, can be NULL (slime, etc)
//	"source" is the thing to target after taking damage
//	 creature or NULL
//	Source and inflictor are the same for melee attacks.
//	Source can be NULL for slime, barrel explosions
//	and other environmental stuff.
//===========================================================================
void P_DamageMobj(mobj_t* target, mobj_t* inflictor, mobj_t* source, int damage, int from)
{
	extern int		radx_gren;
	extern int		radx_rckt;
	extern RDBOOL	killthing;
	mbf_player_t	*player;
	RDBOOL			justhit;
	RDBOOL			amodel = false;

	// Clients can't harm anybody - or edit mode.
	if (R3D_edit) return;

	if (target->health <= 0) return;	// why bother?
	if (target->ddflags & DDF_HOLDSTATE) return;	// v220-33

	// v220-35 decoration model where shootable is set to allow blocking
	if (target->sprite == SPR_TFOG && !target->info->deathstate) return;

	// killough 8/31/98: allow bouncers to take damage
    if (   !(target->flags & MF_SHOOTABLE)	// shouldn't happen...
		&& !(target->flags3 & MF3_BOUNCES))	// but bouncers can take damage
	{
		return;
	}

	// Special cases
	if (   R3D_ScriptedMap
		&& (amodel = (target->flags3 & MF3_AMODEL) != 0)
		&& !target->info->deathstate)
	{
		return;
	}

    if (target->flags & MF_SKULLFLY)
    {
		target->momx = target->momy = target->momz = 0;
    }

	player = target->player;

	if (player && player->cheats & CF_NOCLIP) return;

	// v220-33 damage modifier for scripted maps
	if (playerdamageidx)
	{
		int olddamage = damage;

		if (player && inflictor && damage > 0 && damage < 100)
		{
			for (int i = 0; i < playerdamageidx; i++)
			{
				if (playerdamagetype[i] == inflictor->type)
				{
					damage = (int)(float(damage) * playerdamagefactor[i]);
					if (!damage) damage = olddamage;
					break;
				}
			}
		}

		if (ge_inf.reportdamage)
		{
			if (target->player && inflictor)
#ifdef _DEBUG
				Con_Printf("from %i type %i ID %s olddamage %i new damage %i\n",
				from, inflictor->type, Def_IDfromType(inflictor->type), olddamage, damage);
#else
				Con_Printf("ID %s olddamage %i new damage %i\n",
					Def_IDfromType(inflictor->type), olddamage, damage);
#endif
		}
	}

    if (   player
		&& gameskill == sk_baby
		&& damage > 1	// v220-33 - previously made player with health 1 immortal!
		&& !(target->flags2 & MF2_NOAIRSUPPLY)
		// v220-31 If self inflicted then normal damage. Some maps require suicide to exit and
		//         in the case of scythe2 map25 the player can't exit in easy mode.
		//         Since a user could be led to believe that the engine is at fault if no mention
		//         is made by the map author that it will not work in easy mode then this prevents
		//         confusion all round. (It isn't mentioned in the scythe2 txt for example).
		&& !(player && inflictor && inflictor->target == source))
	{
		damage >>= 1; 	// take half damage in trainer mode (unless self-inflicted)
	}

	// v220-34
	if (!ma_inf.NOINFIGHTALL && !player && source && inflictor && target->ddflags & DDMF_NOINFIGHT)
	{
		if ((from == 11 || from == 26) && source->ddflags & DDMF_NOINFIGHT)
			return;

		if (from == 27 && !sentient(target)) return;	// v220-35 add sentient test
	}

	// v220-35
	if (   source
		&& source->ddflags & DDF_SCPINFIGHT
		&& target->ddflags & DDF_SCPINFIGHT
		&& sentient(target))
	{
		return;
	}

	// Impart exploding or impact weapon thrust
    if (   inflictor
		&& !(target->flags & MF_NOCLIP)
		// The chainsaw should not inflict thrust
		&& (!source	// R3D check it's valid
		|| !source->player
		|| source->player->readyweapon != wp_chainsaw) )
    {
		angle_t	ang;
		fixed_t thrust;

		// v220-13 check if same type things firing bullets should not damage each other
		if (   from == 26	// $bullets or everything
			&& source == inflictor
			&& inflictor->ddflags & DDMF_NOINFIGHT
			&& inflictor->type == target->type
			&& !((inflictor->flags3 ^ target->flags3) & MF3_FRIEND))
		{
			return;
		}

		ang = R_PointToAngle2(inflictor->x, inflictor->y,
								target->x, target->y);
		
		thrust = damage*(FRACUNIT>>3)*100/target->info->mass;

		// If the inflictor is below the target then make
		// the target fall forwards 50% of the time.
		if (   damage < 40
			&& damage > target->health
			&& target->z - inflictor->z > 64*FRACUNIT
			&& M_RandomD() & 1)
		{
			ang += ANG180;
			thrust *= 4;
		}
		
		ang >>= ANGLETOFINESHIFT;

		target->momx += FixedMul (thrust, finecosine[ang]);
		target->momy += FixedMul (thrust, finesine[ang]);
		if(target->dplayer) target->dplayer->flags |= DDPF_FIXPOS;
		
		// killough $dropoff_fix: thrust objects hanging off ledges
		if (target->flags3 & MF3_FALLING && target->gear >= MAXGEAR)
			target->gear = 0;
		
		// v220-16 If just teleported.
		// This deals with map hack where the player is supposed to be blown
		// across an exit line and be killed which creates a race hazard.
		// e.g. cchest4 map31
		if (player)
		{
			if (ma_inf.playerteleport && inflictor->type == MT_BARREL)
			{
				ma_inf.playerdamage += damage;
				target->reactiontime = 0;	// Must move
				return;
			}

			ma_inf.playerteleport = 0;
		}
    }
	
    // player specific
    if (player)
    {
		if (   inflictor
			&& inflictor->type == cfg.jumpkill
			&& (player->jumptics || target->old_z > target->z))
		{
			return;
		}

		// Below certain threshold,
		// ignore damage in GOD mode, or with INVUL power.
		// or v220-16 if just teleported.
		if (   damage < 1000
			&& (player->cheats & CF_GODMODE
			|| player->powers[pw_invulnerability])
			&& target->subsector->sector->special != 11)	// end game hell hack
		{
			return;
		}
		
		if (plhealth.armortype)
		{
			int saved = plhealth.armortype == 1 ? damage/3 : damage/2;
			
			if (plhealth.armorpoints <= saved)
			{
				// armor is used up
				saved = plhealth.armorpoints;
				plhealth.armortype = 0;
				plhealth.armorpoints = 0;
			}
			else
			{
				plhealth.armorpoints -= saved;
			}

			damage -= saved;
		}

		player->attacker = source;
		plhealth.health -= (damage + ma_inf.playerdamage);
		ma_inf.playerdamage = 0;

		if (plhealth.health < 0) plhealth.health = 0;

		// If not being attacked then don't gib if drowning
		if (source || !(target->flags2 & MF2_NOAIRSUPPLY) )
			player->damagecount += damage;	// add damage after armor / invuln
		
		if (player->damagecount > 100)
			player->damagecount = 100;	// teleport stomp does 10k points...
    }
    
	// How about some particles, yes?
	// Only works when both target and inflictor are real mobjs.
	if (inflictor != source)
		P_SpawnDamageParticleGen(target, inflictor, damage);

	if (killthing) damage = target->health;	// v220-33 headshot

    // do the damage
	if (   !player
		&& R3D_ScriptedMap
		&& from == 27)	// radius attack
	{
		int mp = 1;

		if (radx_rckt && inflictor->type == MT_ROCKET)
		{
			mp = radx_rckt;
		}
		else if (radx_gren && (inflictor->type == MT_GRENADEX || inflictor->type == MT_PIPEBOMB))
		{
			mp = radx_gren;
		}

		target->health -= mp*damage;	// v220-35 damage adjustment
	}
	else
	{
		target->health -= damage;
	}

	// v220-7 special case of player in super damage sector
	//        player is not killed to ensure exit occurs
	if (   player
		&& !plhealth.health
		&& ppt.view_subsector->sector->special == 11)
	{
		plhealth.health = 1;
		target->health  = 1;
	}

	// v220-13 ensure if a voodoo doll is killed that the player is killed
	if (   plhealth.health <= 0
		&& !target->type
		&& ppt.viewplrmo
		&& ppt.viewplrmo->health > 0
		&& target != ppt.viewplrmo)
	{
		ppt.viewplrmo->health = 0;
		P_KillMobj (ppt.viewplrmo, ppt.viewplrmo);
	}

	// v220-19 add 608x types
    if (target->health <= 0)
    {
void Chk608xDeath(mobj_t *mo);
		if (!amodel) target->flags2 |= MF2_CANSLOPE;

		if (   R3D_ScriptedMap
			&& target->info->doomednum >= 6085
			&& target->info->doomednum <= 6089)
		{
			Chk608xDeath(target);
		}

		if (from == 26) target->health = -1;	// V220-35 do not gib with bullet

		P_KillMobj (source, target);
		return;
    }
	
	// If target is a player, set player's target to source,
	// so that a friend can tell who's hurting a player
	if (player)
	{
		P_SetTarget(&target->target, source);
	}
	
	// killough 9/8/98:
	// If target's health is less than 50%, move it to the front of its list.
	// This will slightly increase the chances that enemies will choose to
	// "finish it off", but its main purpose is to alert friends of danger.
	if (target->health*2 < target->info->spawnhealth)
	{
		int	thclass = (target->flags3 & MF3_FRIEND) ? th_friends : th_enemies;
		thinker_t *cap = &thinkerclasscap[thclass];

		//---------------------------

		(target->thinker.cprev->cnext = target->thinker.cnext)->cprev =
			target->thinker.cprev;
		(target->thinker.cnext = cap->cnext)->cprev = &target->thinker;
		(target->thinker.cprev = cap)->cnext = &target->thinker;
	}

    if (   M_RandomD() < target->info->painchance
		&& !(target->flags & MF_SKULLFLY) )
    {
		justhit = true;
		P_SetMobjState(target, target->info->painstate);
    }
	
    target->reactiontime = 0;		// we're awake now...	
	
	if (   source
		&& source != target
		&& source->type != MT_VILE
		&& (!target->threshold || target->type == MT_VILE)
		&& (cfg.inFight || (source->flags3 ^ target->flags3) & MF3_FRIEND))
	{
		// if not intent on another player, chase after this one
		//
		// killough 2/15/98: remember last enemy, to prevent
		// sleeping early; 2/21/98: Place priority on players
		// killough 9/9/98: cleaned up, made more consistent:
		
		if (  !target->lastenemy
			|| target->lastenemy->health <= 0
			||
			( !((target->flags3 ^ target->lastenemy->flags3) & MF3_FRIEND)
			&& target->target != source))		// remember last enemy - killough
		{
			P_SetTarget(&target->lastenemy, target->target);
		}
		
		P_SetTarget(&target->target, source);	// killough 11/98

		target->threshold = BASETHRESHOLD;

		if (target->state == &states[target->info->spawnstate]
			&& target->info->seestate != S_NULL)
		{
			P_SetMobjState (target, target->info->seestate);
		}
	}

	if (!justhit) return;
	
	// killough 11/98: Don't attack a friend, unless hit by that friend.
	if (   target->target == source
		|| !target->target
		|| !(target->flags3 & target->target->flags3 & MF3_FRIEND) )
	{
		target->flags |= MF_JUSTHIT;    // fight back!
	}
}

#pragma optimize("", on)

