//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 id Software, Inc.,
//  Chi Hoang, Lee Killough, Jim Flynn, Rand Phares, Ty Halderman,
//  Jaakko Kernen, G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_plats.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  BOOM and R3D extensions added, reformatted, moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll code and BOOM v2.01
//
// DESCRIPTION:
//	Plats (i.e. elevator platforms) code, raising/lowering.
//
//-----------------------------------------------------------------------------

#include "doom/p_local.h"

#include "con_console.h"
#include "s_main.h"
#include "m_winmem.h"
#include "m_misc.h"
#include "r3d_compat.h"

platlist_t *activeplats;       // killough 2/14/98: made global again

void T_PlatInStasis(void *data)
{
	return;
}

//
// Move a plat up and down
//
void T_PlatRaise(plat_t* plat)
{
	RDBOOL		stopplat = false;
	RDBOOL		hassound = !(plat->sector->specFlags2 & SF2_NOMOVSOUND);
    result_e	res;
	
    switch(plat->status)
    {
	case up:
		res = T_MovePlane(plat->sector,
				  plat->speed,
				  plat->high,
				  plat->crush,
				  0,
				  1);

		if (res == crushed && !plat->crush)
		{
			plat->count = plat->wait;
			plat->status = down;
			if (hassound) S_SectorSound(plat->sector, sfx_pstart);
		}
		else
		{
			if (res == pastdest)
			{
				// if not an instant toggle type, wait, make plat stop sound
				if (plat->type!=toggleUpDn)
				{
					plat->count = plat->wait;
					plat->status = waiting;
					if (hassound) S_SectorSound(plat->sector, sfx_pstop);
				}
				else // else go into stasis awaiting next toggle activation
				{
					plat->oldstatus = plat->status;//jff 3/14/98 after action wait  
					plat->status = in_stasis;      //for reactivation of toggle
				}
				
				// lift types and pure raise types are done at end of up stroke
				// only the perpetual type waits then goes back up
				switch(plat->type)
				{
				case blazeDWUS:
				case downWaitUpStay:
				case raiseAndChange:
				case raiseToNearestAndChange:
				case genLift:
					stopplat = true;
				default:
					break;
				}
			}
		}
		break;
	
	case	down:
		res = T_MovePlane(
				plat->sector,
				plat->speed,
				plat->low,
				0,
				0,
				-1);

		if (res == pastdest)
		{
			// if not an instant toggle, start waiting, make plat stop sound
			if (plat->type != toggleUpDn) //jff 3/14/98 toggle up down
			{
				plat->count = plat->wait;
				plat->status = waiting;
				if (hassound) S_SectorSound(plat->sector, sfx_pstop);
			}
			else // instant toggles go into stasis awaiting next activation
			{
			  plat->oldstatus = plat->status;//jff 3/14/98 after action wait  
			  plat->status = in_stasis;      //for reactivation of toggle
			}
			// remove the plat if its a pure raise type
			switch(plat->type)
			{
			case raiseAndChange:
			case raiseToNearestAndChange:
				stopplat = true;
			default:
			  break;
			}
		}
		break;

	case	waiting:
		if (!--plat->count)
		{
			if (plat->sector->floorheight == plat->low)
				plat->status = up;
			else
				plat->status = down;
			if (hassound) S_SectorSound(plat->sector, sfx_pstart);
		}
		break; //jff 1/27/98 don't pickup code added later to in_stasis

	case	in_stasis:
		return;
    }

	if (stopplat)
	{
		P_RemoveActivePlat(plat);
	}				
	else if (plat->sector->specFlags2 & SF2_STNMOVREP)
	{
		S_SectorSoundRepeat(r3ds.stnmov_id, plat->sector);
	}
}


//
// Do Platforms
//  "amount" is only used for SOME platforms.
//
int EV_DoPlat(line_t *line, plattype_e type, int amount)
{
	int		secnum = -1;
	int		rtn = 0;
	RDBOOL	hassound;
	RDBOOL	remote;
	
	//	Activate all <type> plats that are in_stasis
	if (type == perpetualRaise || (rtn = type == toggleUpDn) )
	{
		P_ActivateInStasis(line->tag);
	}
	
	while ((secnum = P_FindSectorFromTag(line->tag,secnum)) >= 0)
	{
		sector_t	*sec = &sectors[secnum];
		plat_t		*plat;

		// don't start a second floor function if already moving
		if (P_SectorActive(floor_special,sec)) //jff 2/23/98 multiple thinkers
			continue;
		
		hassound = !(sec->specFlags2 & SF2_NOMOVSOUND);
	
		// Find lowest & highest floors around sector
		rtn = 1;
		plat = (plat_t*)W_Calloc(sizeof(*plat), THINKID, 0);
		plat->type = type;
		plat->sector = sec;
		//	plat->sector->specialdata = plat;
		plat->sector->floordata = plat;
		plat->thinker.function = (void (__cdecl *)(void))T_PlatRaise;
		plat->crush = false;
		plat->tag = line->tag;
		P_AddFNThinker(&plat->thinker);
		
		//jff 1/26/98 Avoid raise plat bouncing a head off a ceiling and then
		//going down forever -- default low to plat height when triggered
		plat->low = sec->floorheight;
		
		// set up plat according to type  
		switch(type)
		{
		case raiseToNearestAndChange:
			plat->speed = PLATSPEED/2;
			sec->floorpic = sides[line->sidenum[0]].sector->floorpic;
			plat->high = P_FindNextHighestFloor(sec,sec->floorheight);
			plat->wait = 0;
			plat->status = up;
			// NO MORE DAMAGE, IF APPLICABLE
			sec->special = 0;
			if (hassound) sec->specFlags2 |= SF2_STNMOVREP;	 // R3D flag rep stnmov sound
			break;
			
		case raiseAndChange:
			plat->speed = PLATSPEED/2;
			sec->floorpic = sides[line->sidenum[0]].sector->floorpic;
			plat->high = sec->floorheight + amount*FRACUNIT;
			plat->wait = 0;
			plat->status = up;
			if (hassound) sec->specFlags2 |= SF2_STNMOVREP;	 // R3D flag rep stnmov sound
			break;
			
		case downWaitUpStay:
			plat->speed = PLATSPEED * 4;
			plat->low = P_FindLowestFloorSurrounding(sec);

			if (plat->low > sec->floorheight) plat->low = sec->floorheight;

			plat->high = sec->floorheight;
			remote = line->tag && !(line->backsector && line->backsector->tag == line->tag);
			plat->wait = remote ? 35*PLATWAITTAGGED : 35*PLATWAIT;
			plat->status = down;
			if (hassound) S_SectorSound(sec, sfx_pstart);
			break;
			
		case blazeDWUS:
			plat->speed = PLATSPEED * 8;
			plat->low = P_FindLowestFloorSurrounding(sec);
			
			if (plat->low > sec->floorheight)
				plat->low = sec->floorheight;
			
			plat->high = sec->floorheight;
			plat->wait = 35*PLATWAIT;
			plat->status = down;
			if (hassound) S_SectorSound(sec, sfx_pstart);
			break;
			
		case perpetualRaise:
			plat->speed = PLATSPEED;
			plat->low = P_FindLowestFloorSurrounding(sec);
			
			if (plat->low > sec->floorheight)
				plat->low = sec->floorheight;
			
			plat->high = P_FindHighestFloorSurrounding(sec);
			
			if (plat->high < sec->floorheight)
				plat->high = sec->floorheight;
			
			plat->wait = 35*PLATWAIT;
			plat->status = (plat_e)(M_RandomD() & 1);
			
			if (hassound) S_SectorSound(sec, sfx_pstart);
			break;
			
		case toggleUpDn: //jff 3/14/98 add new type to support instant toggle
			plat->speed = PLATSPEED;  //not used
			plat->wait = 35*PLATWAIT; //not used
			plat->crush = true; //jff 3/14/98 crush anything in the way
			
			// set up toggling between ceiling, floor inclusive
			plat->low = sec->ceilingheight;
			plat->high = sec->floorheight;
			plat->status =  down;
			break;
			
		default:
			break;
		}

		P_AddActivePlat(plat);
	}
	return rtn;
}

// The following were all rewritten by Lee Killough
// to use the new structure which places no limits
// on active plats. It also avoids spending as much
// time searching for active plats. Previously a 
// fixed-size array was used, with NULL indicating
// empty entries, while now a doubly-linked list
// is used.

//
// P_ActivateInStasis()
//
// Activate a plat that has been put in stasis 
// (stopped perpetual floor, instant floor/ceil toggle)
//
// Passed the tag of the plat that should be reactivated
//
void P_ActivateInStasis(int tag)
{
	platlist_t *pl;

	// search the active plats by tag for one in stasis
	for (pl=activeplats; pl; pl=pl->next)
	{
		plat_t *plat = pl->plat;

		if (plat->tag == tag)
		{
			if (plat->status == in_stasis) 
			{
				if (plat->type == toggleUpDn)	//jff 3/14/98 reactivate toggle type
					plat->status = plat->oldstatus == up ? down : up;
				else
					plat->status = plat->oldstatus;

				plat->thinker.function = (void (__cdecl *)(void))T_PlatRaise;
			}
		}
	}
}

//
// EV_StopPlat()
//
// Handler for "stop perpetual floor" linedef type
//
// Passed the linedef that stopped the plat
//
// jff 2/12/98 added int return value, fixed return
//
void EV_StopPlat(line_t* line)
{
	platlist_t *pl;

	for (pl=activeplats; pl; pl=pl->next)  // search the active plats
	{
		plat_t *plat = pl->plat;			 // for one with the tag not in stasis

		if (plat->status != in_stasis && plat->tag == line->tag)
		{
			plat->oldstatus = plat->status;    // put it in stasis
			plat->status = in_stasis;
			plat->thinker.function = (void (__cdecl *)(void))T_PlatInStasis;

			if (plat->sector->specFlags2 & SF2_STNMOVREP)	// R3D
				S_StopSound(r3ds.stnmov_id ,(mobj_t*)&plat->sector->soundorg);
		}
	}
}

//
// P_AddActivePlat()
//
// Add a plat to the head of the active plat list
//
void P_AddActivePlat(plat_t* plat)
{
	platlist_t *list = (platlist_t*)W_Calloc(sizeof(platlist_t), LEVELID, 69);

	list->plat = plat;
	plat->list = list;

	plat->count = plat->wait;

	if ((list->next = activeplats))
		list->next->prev = &list->next;

	list->prev = &activeplats;
	activeplats = list;
}

//
// P_RemoveActivePlat()
//
// Remove plat ptr from the active plat list
// Stops any r3ds.stnmov_id sound at the origin
//
//
void P_RemoveActivePlat(plat_t* plat)
{
	platlist_t *list = plat->list;

	if (plat->sector->specFlags2 & SF2_STNMOVREP)
	{
		plat->sector->specFlags2 &= ~SF2_STNMOVREP;
		S_StopSound(r3ds.stnmov_id, (mobj_t*)&plat->sector->soundorg);
	}

	plat->sector->floordata = NULL; //jff 2/23/98 multiple thinkers
	P_RemoveThinker(&plat->thinker);

	if ((*list->prev = list->next))
		list->next->prev = list->prev;

	W_Free(list, LEVELID, 70);
}

// called after reclaiming all LEVELID mem
void P_NullActivePlats(void)
{
	activeplats = NULL;
}