//-----------------------------------------------------------------------------
//
// Copyright (C) 1998 - 2005 id Software, Inc., Chi Hoang, Lee Killough,
//   Jim Flynn, Rand Phares, Ty Halderman, Jaakko Kernen
// Copyright (C) 2006 - 2018 G Jackson
//
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_saveg.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  BOOM and R3D extensions added, reformatted, moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 14 (v220-14) 01/May/2012 gmj
// Revision 2.2.0 build 15 (v220-15) 08/Jun/2012 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 17 (v220-17) 27/Oct/2012 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
// Revision 2.2.0 build 25 (v220-25) 13/May/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll code and BOOM v2.01
//           Note: Utilizes LZSS compression.
//
// DESCRIPTION:
//      Archiving: SaveGame I/O.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdio.h>
#include <direct.h>
#include <string.h>
#include "lzss.h"
#include <errno.h>

#include "sys_system.h"
#include "sys_stwin.h"

#include "r3d_compat.h"
#include "r3d_sfx.h"

#include "doom/p_local.h"
#include "doom/f_infine.h"
#include "doom/p_svtexarc.h"
#include "doom/g_game.h"
#include "doom/d_englsh.h"

#include "dd_share.h"
#include "dd_wad.h"
#include "con_console.h"
#include "p_data.h"
#include "r_model.h"
#include "r_main.h"
#include "p_particle.h"
#include "r_world.h"
#include "r_things.h"
#include "sys_timer.h"
#include "m_misc.h"
#include "s_main.h"
#include "rend_sky.h"
#include "r3d_xline.h"

#include "r3d_mapinfo.h"
#include "r3d_modelbox.h"
#include "r3d_setupFns.h"

// MACROS ------------------------------------------------------------------

#define MY_SAVE_MAGIC			0x1DEAD666
#define MY_XLINE_SAVE_MAGIC		0x9DEAD666	// v220-25
#define MY_XLINE_CHECK			0x80000000	// v220-25
#define MY_CLIENT_SAVE_MAGIC	0x2DEAD666
#define CONSISTENCY				0x2c
#define SAVEGAMENAME			"Risen3DSav"
#define CLIENTSAVEGAMENAME		"Risen3DCl"

#define MAXARCHIVESIZE			16384

// flags bits can be or'ed with SAVEGAME_VER Backward savegame compatibility flags;
#define FLAG_COMPAT1			0x00000001	// unused
#define FLAG_COMPAT2			0x00000002	// unused
#define FLAG_COMPAT3			0x00000004	// unused
#define FLAG_COMPAT4			0x00000008	// unused

#define	SAVEGAME_VER			0xF3D35300	// v220-35 sub v3
#define	ENDGAME_CHK				0xFA0B0707

// NOTE: clear this if changing SAVEGAME_VER ('unflag' flagged functions)
#define	SAVEGAME_FLAGS			0x00000000
#define	FLAGS_MASK				0x00000000
#define	VERSION_MASK			0xFFFFFFF0

// The version excludes any compatibility flags
// which are always stored to the header.
// Flags, if set, are stored to the LSB of the SAVEGAME_VER
// any flags and xxxx is the release number.
// See FLAG_'s above.

// TYPES -------------------------------------------------------------------

#pragma pack(1)

typedef struct
{
	int		magic;
	int		gamemode;		
	int		leveltime;
	uint	cmdvars;						// v220-35
	char	description[SAVESTRINGSIZE];

	// v220-5 fixed at 16 for backward compatibility
	//        Only the first location is used
	byte	players[SAVEMAXPLR];
	
	byte	skill;
	byte	episode;
	byte	map;

	// v220-5 kept for backward compatibility
	byte	deathmatch;

	byte	nomonsters;
	byte	respawn;
	int		switch_flags;	// b211-0
	int		sprites_count;	// b211-0
	int		states_count;	// b211-0
	int		unused;			// spare

} saveheader_t;

#pragma pack()

char pwadname[MAXPWADNAME];

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void	S_RestartPerpetualCeilingSound(void);
void	setCHECKsize(int *chksecs, int *chklines, int *chkthings, int lump);
byte	*G_CopyAddstats(int *size);
byte	*G_SetAddstats(int *size);
long	getFileTime(char *path);
void	LS_InitTagLists(void);
void	H_SetTHINGsLODstate(void);				// v220-9
void	alterFlags(mobj_t *mo);					// v220-9
void	P_SetNoInfight(mobj_t *mo);				// v220-13
void	R3D_SetParms(char *mapid);				// v220-16
byte	*AM_GetMarkPoints(int *sz, int *nm);	// v220-17
void	AM_SetMarkPoints(byte *mp, uint nm);	// v220-17
int		NGL_getFlags(void);						// v220-31
void	NGL_setFlags(int flags);				// v220-31
void	subfn_ResetHeight(mobj_t *iter);		// v220-31

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern char		base_save_path[MAXPATHNAMESIZE];
extern char		savedescription[SAVESTRINGSIZE]; 
extern char		savefileDeletestr[1024];

extern RDBOOL	wasnomonsters;
extern int		torch_count;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

mobj_t	*braintargets[MAXBRAINTARGETS];
int		numbraintargets;
int		braintargeton;
int		braineasy;

uint	pwadchecksum[MAX_PWADS + 1];	//iwad checksum stored in 0

// v220-31
secmodummy_t	*secdmo;
int				maxsecdnum;
char			savepathfname[512];

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static LZFILE *savefile = NULL;

static char *delmsg = "\nDelete savegame before exiting YES-NO ?\n";

static saveheader_t hdr;
static char savefname[32];

static int file_version;
static int file_flags;
static uint	savegametime;		// time savegame was created

// CODE --------------------------------------------------------------------

void CloseSavefile()
{
	if (savefile)
	{
		lzClose(savefile);
		savefile = NULL;
	}
}

// v220-33
static void Con_ErrorDelete(char *msg)
{
	CloseSavefile();
	sprintf(savefileDeletestr, "LOADGAME ERROR: %s%s", msg, delmsg);
	Sys_Quit(false);
}

static RDBOOL OpenSavefile(char *mode)
{
	if (savefile)
		Con_Error("Program error: SAVEFILE not closed\n");

	savefile = lzOpen(savepathfname, mode);

	return savefile != NULL;
}

// b211-0 Added
static RDBOOL ChkAppendMessage(char *msg, int swf, int svf, int flag)
{
	if ( !((swf & flag) ^ (svf & flag)) ) return false;

	msg[0] = 0;

	switch(flag)
	{
	case SWF_DEH:
		strcat(msg, svf & flag ? "UNTICK " : "TICK    ");
		break;
	case SWF_XA:
	case SWF_NOAMWDS:
		strcpy(msg, svf & flag ? "TICK    " : "UNTICK  ");
		break;
	case SWF_GND:
	case SWF_PBM:
	case SWF_NC:
	case SWF_NM:
	default:
		strcat(msg, svf & flag ? "ENABLE  " : "DISABLE ");
		break;
	}

	switch(flag)
	{
	case SWF_NC:
		strcat(msg, "Basic doom mode or BDMFX\n"); break;
	case SWF_NM:
		strcat(msg, "no monsters\n"); break;
	case SWF_XA:
		strcat(msg, "Core FX\n"); break;
	case SWF_SPZ:
		strcat(msg, "Sitters PP zombies\n"); break;
	case SWF_AM:
		strcat(msg, "Sitters&Abbs Models\n"); break;
	case SWF_FP:
		strcat(msg, "Abbs replacement weapons\n"); break;
	case SWF_AP:
		strcat(msg, "Abbs alternative weapons\n"); break;
	case SWF_FX:
		strcat(msg, "Extra FX (particles)\n"); break;
	case SWF_GND:
		strcat(msg, "Grenade Launcher\n"); break;
	case SWF_PBM:
		strcat(msg, "Pipebomb Launcher\n"); break;
	case SWF_DEH:
		strcat(msg, "No Dehs\n"); break;
	case SWF_NOAMWDS:
		strcat(msg, "No forced models\n"); break;
	default:
		strcat(msg, "Unknown: incompatible savegame\n"); break;
	}
	return true;
}

// V220-33
int global_return;

static void subfn_HandleMessage(int del_option)
{

	char	buff[1512];
	int		contnue, del = strstr(savefileDeletestr, delmsg) != NULL;

	CloseSavefile();

	if (!gl_inited)
	{
		sprintf(buff, "%s\n", savefileDeletestr);
		Sys_MessageBox(buff, true);
		Sys_Quit(false);
	}

	sprintf(buff, "%s\n%s\n", savefileDeletestr, savepathfname);
	savefileDeletestr[0] = 0;

	global_return = true;
	contnue = Sys_CriticalMessage(buff, true);

	if (!contnue && !del_option)
		Sys_Quit(false);
	else if (contnue && !(del || del_option))
		return;

	if (contnue)
	{
		sprintf(buff, "CONFIRM DELETE %s\n", savepathfname);

		if (Sys_CriticalMessage(buff, true) && remove(savepathfname))
		{
			sprintf(buff, "ERROR: FAILED to delete;\n  %s\n",savepathfname);
			Sys_MessageBox(buff, true);
		}
	}
	global_return = false;
}

// b211-0, v220-3 Revised
static int checkSwitchStatus(int svf)
{
	int		swf = ge_inf.switch_flags;
	int		i;
	char	msg[128], buf[512];

	buf[0] = msg[0] = 0;

	for (i = 0; i < SWF_NUM; i++)
	{
		if (ChkAppendMessage(msg, swf, svf, 1 << i))
		{
			if (strlen(buf) + strlen(msg) < 512)
				strcat(buf, msg);
		}
	}

	if (buf[0])
	{
		strcat(savefileDeletestr, "SAVEGAME COMPATIBILITY CHECK\n\n"
			"Please make the following changes prior to restarting your savegame;\n\n");
		strcat(savefileDeletestr, buf);
		subfn_HandleMessage(false);
		return false;
	}

	return true;
}

void SV_Write(void *data, int len)
{
	lzWrite(data, len, savefile);
}

void SV_WriteByte(byte val)
{
	lzPutC(val, savefile);
}

void SV_WriteShort(short val)
{
	lzPutW(val, savefile);	
}

void SV_WriteLong(long val)
{
	lzPutL(val, savefile);
}

void SV_WriteFloat(float val)
{
	lzPutL(*(long*) &val, savefile);
}

void SV_Read(void *data, int len)
{
	lzRead(data, len, savefile);
}

byte SV_ReadByte()
{
	return lzGetC(savefile);
}

short SV_ReadShort()
{
	return lzGetW(savefile);
}

long SV_ReadLong()
{
	return lzGetL(savefile);
}

float SV_ReadFloat()
{
	long val = lzGetL(savefile);
	return *(float*) &val;
}

mobj_t *checkThingRef(int num)
{
	thid_t id = num & 0xFFFF;
	return SetReferencedMobj(id);
}

static void SV_WriteButtonList()
{
	int i, count = 0;

    for (i = 0; i < MAXBUTTONS; i++)
    {
		if (buttonlist[i].btimer > 0) count++;
	}

	SV_WriteShort((short)count);

    for (i = 0; i < MAXBUTTONS; i++)
    {
		if (buttonlist[i].btimer > 0)
		{
			SV_WriteShort((short)i);
			SV_WriteLong(buttonlist[i].line - lines);
			SV_WriteByte((byte)buttonlist[i].where);
			SV_WriteShort((short)buttonlist[i].btexture);
			SV_WriteLong(buttonlist[i].btimer);
		}
    }
}

static void SV_WriteSecmoVars()
{
	int i, count = trig5505.count;

	SV_WriteShort((short)count);

	if (!count) return;

    for (i = 0; i < count; i++) SV_WriteShort((short)secmo[i].timer);
}

static void SV_ReadButtonList(int count)
{
	int i;

	memset(buttonlist, 0, MAXBUTTONS * sizeof(button_t));

    for (i = 0; i < count; i++)
    {
		int bnum = SV_ReadShort();

		buttonlist[bnum].line		= &lines[SV_ReadLong()];
		buttonlist[bnum].where		= (bwhere_e)SV_ReadByte();
		buttonlist[bnum].btexture	= SV_ReadShort();
		buttonlist[bnum].btimer		= SV_ReadLong();
    }
}

static void SV_ReadSecmoVars(int count)
{
	if (count <= 0) return;

	if (count > maxsecdnum)
	{
		maxsecdnum = count;
		secdmo = (secmodummy_t*)W_ReallocCLR(secdmo, count*sizeof(secmodummy_t), STATICID, 0);
	}

    for (int i = 0; i < count; i++) secdmo[i].timer = (int)SV_ReadShort();
}

static void SV_WritePlayer(int playernum)
{
	mbf_player_t temp, *pl = &temp;
	ddplayer_t *dpl;
	int i;

	memcpy(pl, mbf_players + playernum, sizeof(temp));		
	dpl = pl->plr;
	for(i=0; i<NUMPSPRITES; i++)
	{
		if(temp.psprites[i].state)
		{
			temp.psprites[i].state 
				= (state_t*) (temp.psprites[i].state - states);
		}
	}

	SV_WriteByte(1);		// Write a version byte.
	SV_WriteLong(pl->playerstate);
	SV_WriteLong(dpl->viewz);
	SV_WriteLong(dpl->viewheight);
	SV_WriteLong(dpl->deltaviewheight);
	SV_WriteFloat(dpl->lookdir);
	SV_WriteLong(pl->bob);
	
	SV_WriteLong(plhealth.health);
	SV_WriteLong(plhealth.armorpoints);
	SV_WriteLong(plhealth.armortype);

	SV_Write(pl->powers, NUMPOWERS * 4);
	SV_Write(pl->cards, NUMCARDS * 4);
	SV_WriteLong(pl->backpack);	

	SV_Write(pl->frags, 4 * 4);
	SV_WriteLong(pl->readyweapon);
	SV_WriteLong(pl->pendingweapon);

	SV_Write(pl->weaponowned, NUMWEAPONS * 4);
	SV_Write(pl->ammo, NUMAMMO * 4);
	SV_Write(pl->maxammo, NUMAMMO * 4);

	SV_WriteLong(pl->attackdown);
	SV_WriteLong(pl->usedown);

	SV_WriteLong(pl->cheats);

	SV_WriteLong(pl->refire);

	SV_WriteLong(pl->killcount);
	SV_WriteLong(pl->itemcount);
	SV_WriteLong(pl->secretcount);

	SV_WriteLong(pl->damagecount);
	SV_WriteLong(pl->bonuscount);
	
	SV_WriteLong(dpl->extralight);
	SV_WriteLong(dpl->fixedcolormap);

	SV_WriteLong(pl->colormap);
	SV_Write(pl->psprites, NUMPSPRITES * sizeof(pspdef_t));
	
	SV_WriteLong(pl->didsecret);
}

static void SV_ReadPlayer(mbf_player_t *pl)
{
	ddplayer_t	*dpl = pl->plr;
	int			i;

	SV_ReadByte();	// The version (not used yet).

	pl->playerstate = (playerstate_t)SV_ReadLong();
	dpl->viewz = SV_ReadLong();
	dpl->viewheight = SV_ReadLong();
	dpl->deltaviewheight = SV_ReadLong();
	dpl->lookdir = SV_ReadFloat();
	pl->bob = SV_ReadLong();
	
	if (pl == mbf_players)
	{
		plhealth.health = SV_ReadLong();
		plhealth.armorpoints = SV_ReadLong();
		plhealth.armortype = SV_ReadLong();
	}
	else
	{
		SV_ReadLong();
		SV_ReadLong();
		SV_ReadLong();
	}

	SV_Read(pl->powers, NUMPOWERS * 4);
	{
		extern int	mapcheat;
		mapcheat = pl->powers[pw_allmap];
	}

	SV_Read(pl->cards, NUMCARDS * 4);
	pl->backpack = SV_ReadLong();

	SV_Read(pl->frags, 4 * 4);
	pl->readyweapon = (weapontype_t)SV_ReadLong();
	pl->pendingweapon = (weapontype_t)SV_ReadLong();

	SV_Read(pl->weaponowned, NUMWEAPONS * 4);
	SV_Read(pl->ammo, NUMAMMO * 4);
	SV_Read(pl->maxammo, NUMAMMO * 4);

	pl->attackdown = SV_ReadLong();
	pl->usedown = SV_ReadLong();

	pl->cheats = SV_ReadLong();

	pl->refire = SV_ReadLong();

	pl->killcount = SV_ReadLong();
	pl->itemcount = SV_ReadLong();
	pl->secretcount = SV_ReadLong();

	pl->damagecount = SV_ReadLong();
	pl->bonuscount = SV_ReadLong();
	
	dpl->extralight = SV_ReadLong();
	dpl->fixedcolormap = SV_ReadLong();

	pl->colormap = SV_ReadLong();
	SV_Read(pl->psprites, NUMPSPRITES * sizeof(pspdef_t));
	
	pl->didsecret = SV_ReadLong();

	for (i = 0; i < NUMPSPRITES; i++)
	{
		if (pl->psprites[i].state)
		{
			pl->psprites[i].state 
				= &states[ (int) pl->psprites[i].state ];
		}
	}

	// Mark the player for fixpos and fixangles.
	dpl->flags |= DDPF_FIXPOS | DDPF_FIXANGLES;
}

static void SV_WriteMobj(mobj_t *mobj)
{
	mobj_t	mo;
	thid_t	tmp;

	memcpy(&mo, mobj, sizeof(mo));	// Take a copy

	mo.state = (state_t*) (mo.state - states);
	if(mo.player)
		mo.player = (mbf_player_t *)((mo.player - mbf_players) + 1);
	
	SV_WriteShort(mo.thinker.id);

	tmp = mo.target ? mo.target->thinker.id : 0;
	SV_WriteShort(tmp);
	tmp = mo.tracer ? mo.tracer->thinker.id : 0;
	SV_WriteShort(tmp);
	tmp = mo.onmobj ? mo.onmobj->thinker.id : 0;
	SV_WriteShort(tmp);
	tmp = mo.lastenemy ? mo.lastenemy->thinker.id : 0;
	SV_WriteShort(tmp);

	SV_WriteLong(mo.iden_nums);
	
	// Info for drawing: position.
	SV_WriteLong(mo.x);
	SV_WriteLong(mo.y);
	SV_WriteLong(mo.z);
	
	//More drawing info: to determine current sprite.
	SV_WriteLong(mo.angle);		// orientation
	SV_WriteLong(mo.sprite);	// used to find patch_t and flip value
	SV_WriteLong(mo.frame);		// might be ORed with FF_FULLBRIGHT
	
	// The closest interval over all contacted Sectors.
	SV_WriteLong(mo.floorz);
	SV_WriteLong(mo.ceilingz);
	SV_WriteLong(mo.dropoffz);		// killough $dropoff_fix
	SV_WriteShort(mo.gear);			// killough used in torque simulation
	
	// For movement checking.
	SV_WriteLong(mo.radius);
	SV_WriteLong(mo.height);	
	
	// Momentums, used to update position.
	SV_WriteLong(mo.momx);
	SV_WriteLong(mo.momy);
	SV_WriteLong(mo.momz);
	
	SV_WriteLong(mo.type);
	SV_WriteLong(mo.tics);	// state tic counter
	SV_WriteLong( (int) mo.state);
	SV_WriteLong(mo.ddflags);
	SV_WriteLong(mo.flags);
	SV_WriteLong(mo.flags2);
	SV_WriteLong(mo.flags3);	// v220-35 for backward compatibility
	SV_WriteShort(mo.flags4);	// v220-35
	SV_WriteLong(mo.health);
	
	// v220-31 next three were WriteLong liftoffset 
	SV_WriteShort(mo.spawnz);
	SV_WriteByte(mo.flat3Dlevel[0]);
	SV_WriteByte(mo.flat3Dlevel[1]);

	SV_WriteLong(mo.misc);
	SV_WriteFloat(mo.zcenter);	// v220-3

	// Movement direction, movement generation (zig-zagging).
	SV_WriteShort(mo.movedir);		// 0-8
	SV_WriteShort(mo.movecount);	// when 0, select a new dir
	
	// Reaction time: if non 0, don't attack yet.
	// Used by player to freeze a bit after teleporting.
	SV_WriteLong(mo.reactiontime);   
	
	// If >0, the target will be chased
	// no matter what (even if shot)
	SV_WriteLong(mo.threshold);
	
	// Additional info record for player avatars only.
	// Only valid if type == MT_PLAYER
	SV_WriteLong( (int) mo.player);
	
	// Player number last looked for.
	SV_WriteLong(mo.lastlook);	
	
	// For nightmare respawn.
	SV_Write(&mo.spawnpoint, 10);	
	
	SV_WriteShort(mo.t79resurrectid);
	SV_WriteLong(mo.lineocclude);
	SV_WriteShort(mo.scplightlevel);
	SV_WriteShort(mo.scpzoff_ptr);
	SV_WriteShort(mo.startangle);
	SV_WriteFloat(mo.rotateperiod);
	SV_WriteShort(mo.torchtic);		// v220-31
	SV_WriteLong(mo.inverseAlpha);	// v220-31
}


static void SV_ReadMobj(mobj_t *mo)
{
	// The reference will be updated after all mobjs are loaded.
	mo->target = (mobj_t*) (ushort)SV_ReadShort();
	mo->tracer = (mobj_t*) (ushort)SV_ReadShort();
	mo->onmobj = (mobj_t*) (ushort)SV_ReadShort();
	mo->lastenemy = (mobj_t*) (ushort)SV_ReadShort();

	mo->iden_nums = SV_ReadLong();
	
	// Info for drawing: position.
	mo->x = SV_ReadLong();
	mo->y = SV_ReadLong();
	mo->z = SV_ReadLong();
	
	//More drawing info: to determine current sprite.
	mo->angle  = SV_ReadLong();	// orientation
	mo->sprite = (spritenum_t)SV_ReadLong();	// used to find patch_t and flip value
	mo->frame  = SV_ReadLong();	// might be ORed with FF_FULLBRIGHT

	// The closest interval over all contacted Sectors.
	mo->floorz   = SV_ReadLong();
	mo->ceilingz = SV_ReadLong();
	mo->dropoffz = SV_ReadLong();		// killough $dropoff_fix
	mo->gear     = SV_ReadShort();		// killough used in torque simulation

	// For movement checking.
	mo->radius	  = SV_ReadLong();
	mo->height	  = SV_ReadLong();	

	// Momentums, used to update position.
	mo->momx	= SV_ReadLong();
	mo->momy	= SV_ReadLong();
	mo->momz	= SV_ReadLong();

	mo->type	= SV_ReadLong();
	mo->tics	= SV_ReadLong();	// state tic counter
	mo->state	= (state_t*) SV_ReadLong();
	mo->ddflags	= SV_ReadLong();
	mo->flags	= SV_ReadLong();
	mo->flags2  = SV_ReadLong();
	mo->flags2 &= ~MF2_NOTLIT;	// These have to be reset
	mo->flags3	= SV_ReadLong();
	mo->flags4	= SV_ReadShort();
	mo->health  = SV_ReadLong();
	mo->spawnz  = SV_ReadShort();
	mo->flat3Dlevel[0] = (char)SV_ReadByte();
	mo->flat3Dlevel[1] = (char)SV_ReadByte();

	mo->misc = SV_ReadLong() & 0xffffff03;

	mo->zcenter = SV_ReadFloat();	// v220-3

	// Movement direction, movement generation (zig-zagging).
	mo->movedir   = SV_ReadShort();
	mo->movecount = SV_ReadShort();
	
	// Reaction time: if not 0, don't attack yet.
	// Used by player to freeze a bit after teleporting.
	mo->reactiontime = SV_ReadLong();   
	
	// If >0, the target will be chased
	// no matter what (even if shot)
	mo->threshold = SV_ReadLong();
	
	// Additional info record for player avatars only.
	// Only valid if type == MT_PLAYER
	mo->player = (mbf_player_t*) SV_ReadLong();
	
	// Player number last looked for.
	mo->lastlook = SV_ReadLong();	
	
	// For nightmare respawn.
	SV_Read(&mo->spawnpoint, 10);	
	
	mo->t79resurrectid = SV_ReadShort();
	mo->lineocclude = SV_ReadLong();
	mo->scplightlevel = SV_ReadShort();
	mo->scpzoff_ptr = SV_ReadShort();
	mo->startangle = SV_ReadShort();
	mo->rotateperiod = SV_ReadFloat();
	mo->rotateprev = Sys_GetRealTime();
	mo->torchtic = SV_ReadShort();
	mo->inverseAlpha = SV_ReadLong();

	//-----------------------------

	// Restore! (unmangle)
	mo->visangle = mo->angle >> 16;

	//  b220 R3D proprietory - amodel door / interactive amodel support
	//  zero the pointers - these are re-instated later
	mo->modelboxnum = 0;

	mo->state = &states[(int)mo->state];

	if (mo->player)
	{
		// Saved players might be in a different order than at the moment.
		int pnum = (int) mo->player - 1;

		if (pnum < 0 || pnum >= MAXPLAYERS)
			Con_ErrorD("Invalid save game: player error");

		mo->player = &mbf_players[pnum];

		if (mo->player->plr)
		{
			mo->dplayer = mo->player->plr;
			mo->dplayer->mo = mo;
			mo->dplayer->clAngle = mo->angle;
			mo->dplayer->clLookDir = 0;
		}
	}

	P_SetMpos(mo);
	alterFlags(mo);

	if (mo->flags & (MF_SPECIAL|MF_DROPPED))
		mo->pickupRadius = mobjinfo[mo->type].radius;

	P_SetNoInfight(mo);
}

static void SV_WriteSector(sector_t *sec)
{
	int type = 128;	// v220-16 was 0, 128 indicates SIF_WASSEEN is used

	if (sec->siflags & SIF_FLOOR  ) type |= 1;
	if (sec->siflags & SIF_CEILING) type |= 2;
	if (sec->siflags & SIF_WASSEEN) type |= 4;

	// Type byte.
	SV_WriteByte((byte)type);	

	if (type & 1)
	{
		SV_WriteLong(sec->flooroffx);
		SV_WriteLong(sec->flooroffy);
	}

	if (type & 2)
	{
		SV_WriteLong(sec->ceiloffx);
		SV_WriteLong(sec->ceiloffy);
	}

	SV_WriteShort((short)(sec->floorheight >> FRACBITS));
	SV_WriteShort((short)(sec->ceilingheight >> FRACBITS));
	SV_WriteShort(SV_FlatArchiveNum(sec->floorpic));
	SV_WriteShort(SV_FlatArchiveNum(sec->ceilingpic));
	SV_WriteByte((byte)sec->lightlevel);
	SV_WriteShort(sec->special);
	SV_WriteShort(sec->tag);
}

static void SV_ReadSector(sector_t *sec)
{
	int		type;
	// v220-15 secret checks added
	RDBOOL	secret1 = sec->special == 9;
	RDBOOL	secret2 = (sec->special & SECRET_MASK) > 0;

	// Read the type.
	type = SV_ReadByte();

	if (type & 1)
	{
		sec->flooroffx = SV_ReadLong();
		sec->flooroffy = SV_ReadLong();
	}

	if (type & 2)
	{
		sec->ceiloffx = SV_ReadLong();
		sec->ceiloffy = SV_ReadLong();
	}

	if (type & 4 || !(type & 128))
	{
		sec->siflags |= SIF_WASSEEN;
	}

	sec->floorheight   = SV_ReadShort() << FRACBITS;
	sec->ceilingheight = SV_ReadShort() << FRACBITS;
	sec->floorpic      = SV_ReadShort();
	sec->ceilingpic    = SV_ReadShort();
	
	sec->ceilingdata  = NULL;
	sec->floordata    = NULL;
	sec->lightingdata = NULL;
	sec->soundtarget  = NULL;
	sec->thinglist    = NULL;

	// The flat numbers are actually archive numbers.
	sec->floorpic   = SV_GetArchiveFlat(sec->floorpic);
	sec->ceilingpic = SV_GetArchiveFlat(sec->ceilingpic);

	sec->lightlevel = SV_ReadByte();
	sec->special = SV_ReadShort();
	sec->tag = SV_ReadShort();

	if ((secret1 && !sec->special) || (secret2 && !(sec->special & SECRET_MASK)))
		sec->siflags |= SIF_WASSECRET;	// v220-15
}

static void SV_WriteLine(line_t *li)
{
	int i;
	side_t *si;

	SV_WriteShort(li->ml_flags);
	SV_WriteShort(li->special);
	SV_WriteShort(li->tag);
	SV_WriteShort(li->xtraflags);	// v220-31
	SV_WriteLong(li->lnflags2);
	SV_WriteByte(li->scpFlags);

	for (i = 0; i < 2; i++)
	{
		if(li->sidenum[i] < 0) continue;

		si = &sides[li->sidenum[i]];

		SV_WriteShort((short)(si->textureoffset >> FRACBITS));
		SV_WriteShort((short)(si->rowoffset >> FRACBITS));
		SV_WriteShort(SV_TextureArchiveNum(si->toptexture));
		SV_WriteShort(SV_TextureArchiveNum(si->bottomtexture));
		SV_WriteShort(SV_TextureArchiveNum(si->midtexture));
	}
}

static void SV_ReadLine(line_t *li)
{
	side_t	*si;
	short	f;
	int		i;
	int		ln;

	// code for backward compatibility
	f = (SV_ReadShort() & ~ML_EXTRAFLAGS) | (li->ml_flags & ML_EXTRAFLAGS);
	ln = li->lnflags2 & LNF2_NOSAVE;	// v220-16
	li->ml_flags = f;

	li->special = SV_ReadShort();
	li->tag = SV_ReadShort();
	li->xtraflags = SV_ReadShort();

	li->lnflags2 = SV_ReadLong() & ~LNF2_NOSAVE;
	
	// v220-16 restore preset types
	if (ln) li->lnflags2 |= ln;

	li->scpFlags &= SCPF_HIMASK;
	li->scpFlags |= (SV_ReadByte() & SCPF_LOWMASK);

	for (i = 0; i < 2; i++)
	{
		if(li->sidenum[i] < 0) continue;

		si = &sides[li->sidenum[i]];

		si->textureoffset = SV_ReadShort() << FRACBITS;
		si->rowoffset     = SV_ReadShort() << FRACBITS;

		// Read in the archive numbers
		si->toptexture    = SV_ReadShort();
		si->bottomtexture = SV_ReadShort();
		si->midtexture    = SV_ReadShort();

		// Get the true textures using the archive numbers.
		si->toptexture    = SV_GetArchiveTexture(si->toptexture);
		si->bottomtexture = SV_GetArchiveTexture(si->bottomtexture);
		si->midtexture    = SV_GetArchiveTexture(si->midtexture);
	}
}

//
// P_ArchivePlayers
//
// Note: the id is taken from the Client.ID file
//
static void P_ArchivePlayers(void)
{
    int i;
	
    for (i = 0; i < MAXPLAYERS; i++)
    {
		if (mbf_players[i].plr->ingame)
		{
			// Kept to allow current savegames to be loaded
			// but the id is now just the count.
			SV_WriteLong(i);
			SV_WritePlayer(i);
		}
    }
}

//
// P_UnArchivePlayers
//
static void P_UnArchivePlayers(RDBOOL *infile, RDBOOL *loaded)
{
	mbf_player_t	dummy_player;
	ddplayer_t		dummy_ddplayer;
	mbf_player_t	*player;
	
    int		i;

	// Setup the dummy.
	dummy_player.plr = &dummy_ddplayer;
	
    for(i=0; i<MAXPLAYERS; i++)
    {
		if(!infile[i]) continue;
		
		// Was the ID. Now not needed. Left to allow
		// loading of previous savegames
		SV_ReadLong();

		// We are only interested in player 0
		if (!i)
		{
			loaded[i] = true;
			player = mbf_players;
		}
		// kept for backward savegame compatibility
		else
		{
			// Use a dummy to load the data.
			player = &dummy_player;
			loaded[i] = false;
		}

		SV_ReadPlayer(player);

		player->plr->mo = NULL;	// Set in P_UnArchiveThinkers()
		player->hint_msg = NULL;
		player->attacker = NULL;
    }
}

//
// P_ArchiveWorld
//
static void P_ArchiveWorld (void)
{
    int			i;
    sector_t*	sec;
    line_t*		li;

	// Write the texture archives.
	SV_WriteTextureArchive();

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
		SV_WriteSector(sec);

	for (i = 0, li = lines; i < numlines; i++, li++)
		SV_WriteLine(li);

	SV_WriteFloat(skyAngle);
	SV_WriteLong(customPunch);
	SV_WriteLong(currentMapMusLumpnum);
	SV_WriteLong(waterdynamics);
	SV_WriteLong(flying);
	SV_WriteByte( (byte)(cfg.berserkBug != 0) );

	// v220-17 although written these are not restored.
	//   it is left to the user to set these in all cases other
	//   than scripted maps which can set these independantly.
	//
	SV_WriteByte( (byte)(cfg.fallingDamage != 0) );
	SV_WriteByte( (byte)(cfg.underwaterDamage != 0) );
}

//
// P_UnArchiveWorld
//
static void P_UnArchiveWorld (void)
{
    int	 i, mlump;

	// Read the texture archives.
	SV_ReadTextureArchive();

	// Load sectors.
	for (i = 0; i < numsectors; i++) SV_ReadSector(&sectors[i]);
	
	// Load lines.
	for (i = 0; i < numlines; i++) SV_ReadLine(&lines[i]);

	skyAngle = SV_ReadFloat();
	customPunch = SV_ReadLong();
	mlump = SV_ReadLong();
	i = SV_ReadLong();
	if (!(ge_inf.mapset & GEF_WATERDYNAMICS)) waterdynamics = i;	// prevent old save game from turning wd off if mapset
	flying = SV_ReadLong();

	// v220-31 reset lumpnum if out of range
	currentMapMusLumpnum = -1;

	if (mlump >= 0 && mlump < numlumps) currentMapMusLumpnum = mlump;

	cfg.berserkBug = SV_ReadByte() != 0;

	// v220-17 the next two are ignored
	SV_ReadByte();	// was used to set cfg.fallingDamage
	SV_ReadByte();	// was used to set cfg.underwaterDamage
}

//
// Thinkers
//
typedef enum
{
    tc_end,
    tc_mobj,
} 
thinkerclass_t;

//
// P_ArchiveThinkers
//
static void P_ArchiveThinkers (void)
{
	th7900_t	*t79;
	th7700_t	*t77;
	thinker_t	*th;
	int			i, k;

	if (R3D_ScriptedMap)
	{
		SV_Write(&msgchangedarray, NUMSPOTS);	// v220-35
		SV_Write(&T78_disable, sizeof(T78_disable));

		// Clear any t79resurrectids that may have been set with a previous savegame
		for(th = thinkercap.next ; th != &thinkercap ; th=th->next)
		{
			if (th->function == (void (__cdecl *)(void))P_MobjThinker)
			{
				((mobj_t*) th)->t79resurrectid = 0;
			}
		}

		// Set the t79resurrectids
		for (i = 0, t79 = th7900; i < NUMSPOTS; i++, t79++)
		{
			if (t79->mo && t79->mo->thinker.function == (void (__cdecl *)(void))P_MobjThinker)
				t79->mo->t79resurrectid = i + 1;
		}
	}
	
	// save off the current mobjs
	for(th = thinkercap.next ; th != &thinkercap ; th=th->next)
	{
		if (th->function == (void (__cdecl *)(void))P_MobjThinker)
		{
			SV_WriteByte(tc_mobj);
			SV_WriteMobj((mobj_t*) th);
		}
	}

	// add a terminating marker
	SV_WriteByte(tc_end);	
	
	if (R3D_ScriptedMap)
	{
		for (i = 0, t79 = th7900; i < NUMSPOTS; i++, t79++)
		{
			SV_WriteLong(t79->t79flags);
			SV_WriteLong(t79->activesec);
		}
		for (i = 0, t77 = th7700; i < NUMSPOTS; i++, t77++)
		{
			SV_WriteShort(t77->blocked);

			for (k = 0; k < t77->spawn_count; k++)
				SV_WriteShort(t77->info[k].t77special);
		}
		// v220-14
		SV_Write(&t79list, sizeof(t79list));

		SV_WriteLong(ENDGAME_CHK);
	}

	SV_Write(&thinker_uwater, sizeof(uwater_t));
}

//
// P_UnArchiveThinkers
//
static void P_UnArchiveThinkers (void)
{
	th7900_t	*t79;
	th7700_t	*t77;
    thinker_t	*th;
    mobj_t		*mobj;
	int			i, k, n;
    byte		tclass;

	if (R3D_ScriptedMap)
	{
		extern char* linetext[];
		SV_Read(&msgchangedarray, NUMSPOTS);	// v220-35
		SV_Read(&T78_disable, sizeof(T78_disable));

		for (i = 0, t79 = th7900; i < NUMSPOTS; i++, t79++)
		{
			if (msgchangedarray[i])
				th7900[th7900[i].witht79num].textmsg = linetext[th7900[i].changemsgnum - 1];

			t79->mo = NULL, t79->ss = NULL;
		}
	}

    // read in saved mobjs
    while (1)
    {
		if ( (tclass = SV_ReadByte()) == tc_end) break;

		if (tclass == tc_mobj)
		{
			thid_t idnum = SV_ReadShort();

			if ( !(mobj = NewReferencedMobj(idnum)) ) 	Con_ErrorD("Bad savegame id\n");

			SV_ReadMobj(mobj);

			if (   mobj->type == MT_TELEPORTMAN	// v220-4 compatibility (see tpthings)
				|| (mobj->dplayer && !mobj->dplayer->ingame))
			{
				// This player has gone
				P_SetMobjID(mobj->thinker.id, false);
				memset(mobj, 0, sizeof(mobj_t));
				continue;
			}

			// v220-22 Ensure ppt.viewplrmo is set.
			// Where voodoo dolls are present then the final
			// mobj with dplayer set will be the player
			if (mobj->dplayer)
			{
				ppt.viewplayer = mobj->dplayer;
				ppt.viewplrmo  = mobj;
			}

			mobj->info = &mobjinfo[mobj->type];
			P_SetThingPosition (mobj);
			P_AddThinker(&mobj->thinker);

			if (R3D_ScriptedMap)
			{
				if ( (n = mobj->t79resurrectid) > 0)
					th7900[n - 1].mo = mobj;
			}
		}
		else
		{
			Con_ErrorD("Error: invalid mobj class");
		}
    }

	// Update references to things.
	for (th = thinkercap.next; th != &thinkercap; th = th->next)
	{
		if (th->function == (void (__cdecl *)(void))P_MobjThinker)
		{
			mobj = (mobj_t*) th;

			mobj->target = checkThingRef((int)mobj->target) ;
			if (mobj->target) mobj->target->thinker.mobjreferences++;
			mobj->tracer = checkThingRef((int)mobj->tracer) ;
			if (mobj->tracer) mobj->tracer->thinker.mobjreferences++;
			mobj->onmobj = checkThingRef((int)mobj->onmobj) ;
			if (mobj->onmobj) mobj->onmobj->thinker.mobjreferences++;
			mobj->lastenemy = checkThingRef((int)mobj->lastenemy) ;
			if (mobj->lastenemy) mobj->lastenemy->thinker.mobjreferences++;
		}
	}

	// v220-9 set lodstates
	H_SetTHINGsLODstate();

	if (R3D_ScriptedMap)
	{
		for (i = 0, t79 = th7900; i < NUMSPOTS; i++, t79++)
		{
			t79->t79flags  = SV_ReadLong();
			t79->activesec = SV_ReadLong();

			if (t79->killambient >= 0 && t79->t79flags & T79_NOAMBIENT)
				S_KillAmbient(t79->killambient);

			if (t79->t79flags & T79_NOHUDKEY)
				GL_SetIgnoreFlag(t79->t79keys);
		}

		for (i = 0, t77 = th7700; i < NUMSPOTS; i++, t77++)
		{
			t77->blocked = SV_ReadShort();

			for (k = 0; k < t77->spawn_count; k++)
				t77->info[k].t77special = SV_ReadShort();
		}

		SV_Read(&t79list, sizeof(t79list));

		if (SV_ReadLong() != ENDGAME_CHK)
		{
			CloseSavefile();
			Con_ErrorD("Invalid save game: EOF check error");
		}
	}

	SV_Read(&thinker_uwater, sizeof(uwater_t));
}


//
// P_ArchiveSpecials
//
enum
{
    tc_ceiling,
    tc_door,
    tc_floor,
    tc_plat,
    tc_flash,
    tc_strobe,
    tc_glow,
	tc_flicker,     // killough	10/ 4/98
	tc_elevator,    // jff		 2/22/98 new elevator type thinker
	tc_scroll,      // killough	 3/ 7/98 new scroll effect thinker
	tc_pusher,      // phares	 3/22/98 new push/pull effect thinker
	tc_phased,		// R3D
	// R3D - scripted thinkers;
	tc_modelkill,
	tc_modelspawn,
	tc_modelreset,
	tc_switchreset,
	tc_soundactivate,
	tc_delayunblock,
	tc_delayresurrect,
	tc_statemonitor,
	tc_endlevel,
	tc_killdelayed,
	tc_lineocclude,
	tc_adjustplat,
	tc_listsound,
	// v220-31 scripted thinker
	tc_activatedelay,
	// v220-31 dynlight types
	tc_dynglow,
	tc_dynfireflicker,
    tc_endspecials
} specials_e;	

//
// Special Things to handle:
//
// T_MoveCeiling	(ceiling_t)
// T_VerticalDoor	 (vldoor_t)
// T_MoveFloor	  (floormove_t)
// T_LightFlash	 (lightflash_t)
// T_StrobeFlash	 (strobe_t)
// T_Glow			   (glow_t)
// T_PlatRaise		   (plat_t)
// T_MoveElevator	   (plat_t)		// jff		 2/22/98
// T_Scroll			 (scroll_t)		// killough	 3/ 7/98
// T_Pusher			 (pusher_t)		// phares	 3/22/98
// T_FireFlicker(fireflicker_t)		// killough	10/ 4/98
// T_SinglePhased	 (phased_t)		// R3D
// T_Phased			 (phased_t)		// R3D

// T_ModelKill		    (modelkill_t)		// R3D Script thinkers
// T_ModelSpawn		   (modelspawn_t)
// T_ModelReset		   (modelreset_t)
// T_SwitchReset	   (modelreset_t)
// T_SoundActivate	(scriptedsound_t)
// T_DelayUnblock	    (delayline_t)
// T_DelayResurrect	   (modelreset_t)
// T_StateMonitor	   (modelreset_t)
// T_Endlevel		    (delayline_t)
// T_KillDelayed	    (delayline_t)
// T_LineOcclude	  (lineocclude_t)
// T_AdjustCeil		   (adjustplat_t)
// T_PlayListsound	    (delayline_t)
// added v330-31
// T_ActivateDelayed   (modelreset_t)
// T_DynGlow              (dynglow_t)		// R3D dynlight 55xx thinkers
// T_DynFireFlicker(dynfireflicker_t)

static void P_ArchiveSpecials (void)
{
    thinker_t		*th;
    ceiling_t		ceiling;
    vldoor_t		door;
    floormove_t		floor;
    plat_t			plat;
    lightflash_t	flash;
    strobe_t		strobe;
    glow_t			glow;
	fireflicker_t	flicker;
	elevator_t		elevator;
	scroll_t		scroll;
	pusher_t		pusher;
	phased_t		phased;

	// Scripted thinkers
	modelkill_t		modelkill;
	modelspawn_t	modelspawn;
	modelreset_t	modelreset;
	modelreset_t	switchreset;
	scriptedsound_t soundactivate;
	delayline_t		delayunblock;
	modelreset_t	delayresurrect;
	modelreset_t	statemonitor;
	delayline_t		endlevel;
	delayline_t		killdelayed;
	lineocclude_t	lineocclude;
	adjustplat_t	adjustplat;
	delayline_t		listsound;
	modelreset_t	activatedelayed;	// v220-31

	// v220-31 type 5505 thinkers
	dynglow_t		dynglow;
	dynfireflicker_t  dynfireflicker;

    // save off the current thinkers
    for (th = fnthinkcap.next ; th != &fnthinkcap ; th=th->next)
    {
		// Changed scheme gets rid of NULL thinkers
		if (th->function == (void (__cdecl *)(void))T_CeilInStasis)
		{
			ceilinglist_t *cl = activeceilings;

			for ( ; cl; cl=cl->next)	// search for activeceiling
				if (cl->ceiling == (ceiling_t *) th)	//jff 2/22/98
					goto ceiling;

			continue;
		}

		if (th->function == (void (__cdecl *)(void))T_PlatInStasis)
		{
			platlist_t *pl = activeplats;

			for ( ; pl; pl=pl->next)		// search for activeplats
				if (pl->plat == (plat_t *) th)			// killough 2/14/98
					goto plat;

			continue;
		}
		
		if (th->function == (void (__cdecl *)(void))T_MoveCeiling)
		{
ceiling:
			SV_WriteByte(tc_ceiling);
			memcpy(&ceiling, th, sizeof(ceiling));
			if (th->function == (void (__cdecl *)(void))T_CeilInStasis)
				ceiling.thinker.function = NULL;
			ceiling.sector = (sector_t *)(ceiling.sector - sectors);
			SV_Write(&ceiling, sizeof(ceiling));
			continue;
		}
		
		if (th->function == (void (__cdecl *)(void))T_VerticalDoor)
		{
			int n = -1;

			SV_WriteByte(tc_door);
			memcpy(&door, th, sizeof(door));
			door.sector = (sector_t *)(door.sector - sectors);

			if (door.line) n = door.line - lines;
			door.line = (line_t*)n;
			SV_Write(&door, sizeof(door));
			continue;
		}
		
		if (th->function == (void (__cdecl *)(void))T_MoveFloor)
		{
			SV_WriteByte(tc_floor);
			memcpy(&floor, th, sizeof(floor));
			floor.sector = (sector_t *)(floor.sector - sectors);
			SV_Write(&floor, sizeof(floor));
			continue;
		}
		
		if (th->function == (void (__cdecl *)(void))T_PlatRaise)
		{
plat:
			SV_WriteByte(tc_plat);
			memcpy(&plat, th, sizeof(plat));
			if (th->function == (void (__cdecl *)(void))T_PlatInStasis)
				plat.thinker.function = NULL;
			plat.sector = (sector_t *)(plat.sector - sectors);
			SV_Write(&plat, sizeof(plat));
			continue;
		}
		
		if (th->function == (void (__cdecl *)(void))T_LightFlash)
		{
			SV_WriteByte(tc_flash);
			memcpy(&flash, th, sizeof(flash));
			flash.sector = (sector_t *)(flash.sector - sectors);
			SV_Write(&flash, sizeof(flash));
			continue;
		}
		
		if (th->function == (void (__cdecl *)(void))T_StrobeFlash)
		{
			SV_WriteByte(tc_strobe);
			memcpy(&strobe, th, sizeof(strobe));
			strobe.sector = (sector_t *)(strobe.sector - sectors);
			SV_Write(&strobe, sizeof(strobe));
			continue;
		}
		
		if (th->function == (void (__cdecl *)(void))T_Glow)
		{
			SV_WriteByte(tc_glow);
			memcpy(&glow, th, sizeof(glow));
			glow.sector = (sector_t *)(glow.sector - sectors);
			SV_Write(&glow, sizeof(glow));
			continue;
		}

		if (th->function == (void (__cdecl *)(void))T_FireFlicker)
		{
			SV_WriteByte(tc_flicker);
			memcpy(&flicker, th, sizeof(flicker));
			flicker.sector = (sector_t *)(flicker.sector - sectors);
			SV_Write(&flicker, sizeof(flicker));
			continue;
		}
		
		if (th->function == (void (__cdecl *)(void))T_MoveElevator)
		{
			SV_WriteByte(tc_elevator);
			memcpy(&elevator, th, sizeof(elevator));
			elevator.sector = (sector_t *)(elevator.sector - sectors);
			SV_Write(&elevator, sizeof(elevator));
			continue;
		}
		
		if (th->function == (void (__cdecl *)(void))T_Scroll)
		{
			SV_WriteByte(tc_scroll);
			memcpy(&scroll, th, sizeof(scroll));
			SV_Write(&scroll, sizeof(scroll));
			continue;
		}
		
		if (th->function == (void (__cdecl *)(void))T_Pusher)
		{
			SV_WriteByte(tc_pusher);
			memcpy(&pusher, th, sizeof(pusher));
			SV_Write(&pusher, sizeof(pusher));
			continue;
		}

		if (th->function == (void (__cdecl *)(void))T_Phased ||
			th->function == (void (__cdecl *)(void))T_SinglePhased)
		{
			int i;
			SV_WriteByte(tc_phased);
			memcpy(&phased, th, sizeof(phased));
			SV_Write(&phased, sizeof(phased));
			for (i = 0; i < 2*phased.count; i++)
				SV_WriteLong(phased.slist[i]);
			continue;
		}

		// SCRIPTED thinkers

		if (th->function == (void (__cdecl *)(void))T_ModelKill)
		{
			SV_WriteByte(tc_modelkill);
			memcpy(&modelkill, th, sizeof(modelkill));
			SV_Write(&modelkill, sizeof(modelkill));
			continue;
		}

		if (th->function == (void (__cdecl *)(void))T_ModelSpawn)
		{
			SV_WriteByte(tc_modelspawn);
			memcpy(&modelspawn, th, sizeof(modelspawn));
			SV_Write(&modelspawn, sizeof(modelspawn));
			continue;
		}

		if (th->function == (void (__cdecl *)(void))T_ModelReset)
		{
			SV_WriteByte(tc_modelreset);
			memcpy(&modelreset, th, sizeof(modelreset));
			SV_Write(&modelreset, sizeof(modelreset));
			continue;
		}

		if (th->function == (void (__cdecl *)(void))T_SwitchReset)
		{
			SV_WriteByte(tc_switchreset);
			memcpy(&switchreset, th, sizeof(switchreset));
			SV_Write(&switchreset, sizeof(switchreset));
			continue;
		}

		if (th->function == (void (__cdecl *)(void))T_SoundActivate)
		{
			SV_WriteByte(tc_soundactivate);
			memcpy(&soundactivate, th, sizeof(soundactivate));
			SV_Write(&soundactivate, sizeof(soundactivate));
			continue;
		}

		if (th->function == (void (__cdecl *)(void))T_DelayUnblock)
		{
			SV_WriteByte(tc_delayunblock);
			memcpy(&delayunblock, th, sizeof(delayunblock));
			SV_Write(&delayunblock, sizeof(delayunblock));
			continue;
		}

		if (th->function == (void (__cdecl *)(void))T_DelayResurrect)
		{
			SV_WriteByte(tc_delayresurrect);
			memcpy(&delayresurrect, th, sizeof(delayresurrect));
			SV_Write(&delayresurrect, sizeof(delayresurrect));
			continue;
		}

		if (th->function == (void (__cdecl *)(void))T_StateMonitor)
		{
			SV_WriteByte(tc_statemonitor);
			memcpy(&statemonitor, th, sizeof(statemonitor));
			SV_Write(&statemonitor, sizeof(statemonitor));
			continue;
		}

		if (th->function == (void (__cdecl *)(void))T_Endlevel)
		{
			SV_WriteByte(tc_endlevel);
			memcpy(&endlevel, th, sizeof(endlevel));
			SV_Write(&endlevel, sizeof(endlevel));
			continue;
		}

		if (th->function == (void (__cdecl *)(void))T_KillDelayed)
		{
			SV_WriteByte(tc_killdelayed);
			memcpy(&killdelayed, th, sizeof(killdelayed));
			SV_Write(&killdelayed, sizeof(killdelayed));
			continue;
		}

		if (th->function == (void (__cdecl *)(void))T_LineOcclude)
		{
			SV_WriteByte(tc_lineocclude);
			memcpy(&lineocclude, th, sizeof(lineocclude));
			SV_Write(&lineocclude, sizeof(lineocclude));
			continue;
		}

		if (th->function == (void (__cdecl *)(void))T_AdjustCeil)
		{
			SV_WriteByte(tc_adjustplat);
			memcpy(&adjustplat, th, sizeof(adjustplat));
			SV_Write(&adjustplat, sizeof(adjustplat));
			continue;
		}

		if (th->function == (void (__cdecl *)(void))T_PlayListsound)
		{
			SV_WriteByte(tc_listsound);
			memcpy(&listsound, th, sizeof(listsound));
			SV_Write(&listsound, sizeof(listsound));
		}

		if (th->function == (void (__cdecl *)(void))T_ActivateDelayed)
		{
			SV_WriteByte(tc_activatedelay);
			memcpy(&activatedelayed, th, sizeof(activatedelayed));
			SV_Write(&activatedelayed, sizeof(activatedelayed));
		}

		// DYNLIGHT THINKERS 5505

		// v220-31 dynlight thinkers
		if (th->function == (void (__cdecl *)(void))T_DynGlow)
		{
			SV_WriteByte(tc_dynglow);
			memcpy(&dynglow, th, sizeof(dynglow_t));
			SV_Write(&dynglow, sizeof(dynglow_t));
		}

		if (th->function == (void (__cdecl *)(void))T_DynFireFlicker)
		{
			SV_WriteByte(tc_dynfireflicker);
			memcpy(&dynfireflicker, th, sizeof(dynfireflicker_t));
			SV_Write(&dynfireflicker, sizeof(dynfireflicker_t));
		}
	}
    // add a terminating marker
    SV_WriteByte(tc_endspecials);	
}


//
// P_UnArchiveSpecials
//
static int GetSector(int secnum, char *type)
{
	if(secnum >= numsectors || secnum < 0)
		Con_Error("Incompatible savegame\n%s: bad sector number", type);
	return secnum;
}

static void P_UnArchiveSpecials (void)
{
    ceiling_t		*ceiling;
    vldoor_t		*door;
    floormove_t		*floor;
    plat_t			*plat;
    lightflash_t	*flash;
    strobe_t		*strobe;
    glow_t			*glow;
	fireflicker_t	*flicker;
	elevator_t		*elevator;
	scroll_t		*scroll;
	pusher_t		*pusher;
	phased_t		*phased;
	// Script thinkers
	modelkill_t		*modelkill;
	modelspawn_t	*modelspawn;
	modelreset_t	*modelreset;
	modelreset_t	*switchreset;
	scriptedsound_t *soundactivate;
	delayline_t		*delayunblock;
	modelreset_t	*delayresurrect;
	modelreset_t	*statemonitor;
	delayline_t		*endlevel;
	delayline_t		*killdelayed;
	lineocclude_t	*lineocclude;
	adjustplat_t	*adjustplat;
	delayline_t		*listsound;
	modelreset_t	*activatedelayed;	// v220-31
	// v220-31 dynlight thinkers
	dynglow_t		*dynglow;
	dynfireflicker_t*dynfireflicker;

	int	i;
	
    // read in saved thinkers
    while (1)
    {
		byte tclass = SV_ReadByte();
		switch (tclass)
		{
		case tc_ceiling:
			ceiling = (ceiling_t*)W_Calloc(sizeof(ceiling_t), THINKID, 0);
			SV_Read(ceiling, sizeof(ceiling_t));

			ceiling->sector = &sectors[GetSector((int)ceiling->sector, "tc_ceiling")];
			ceiling->sector->ceilingdata = ceiling;

			if (ceiling->thinker.function)
				ceiling->thinker.function = (void (__cdecl *)(void))T_MoveCeiling;
			else
				ceiling->thinker.function = (void (__cdecl *)(void))T_CeilInStasis;

			P_AddFNThinker (&ceiling->thinker);
			P_AddActiveCeiling(ceiling);
			break;
			
		case tc_door:
			{
				int n;

				door = (vldoor_t*)W_Calloc(sizeof(vldoor_t), THINKID, 0);
				SV_Read(door, sizeof(vldoor_t));
				door->sector = &sectors[GetSector((int)door->sector, "tc_door")];
				n = (int)door->line;
				door->line = (n >= 0 && n < numlines ? &lines[n] : NULL);
				door->sector->ceilingdata = door;
				door->thinker.function = (void (__cdecl *)(void))T_VerticalDoor;
				P_AddFNThinker (&door->thinker);
			}
			break;
			
		case tc_floor:
			floor = (floormove_t*)W_Calloc(sizeof(floormove_t), THINKID, 0);
			SV_Read(floor, sizeof(floormove_t));
			floor->sector = &sectors[GetSector((int)floor->sector, "tc_floor")];
			floor->sector->floordata = floor;
			floor->thinker.function = (void (__cdecl *)(void))T_MoveFloor;
			P_AddFNThinker (&floor->thinker);
			break;
			
		case tc_plat:
			plat = (plat_t*)W_Calloc(sizeof(plat_t), THINKID, 0);
			SV_Read(plat, sizeof(plat_t));
			plat->sector = &sectors[GetSector((int)plat->sector, "tc_plat")];
			plat->sector->floordata = plat;
			
			if (plat->thinker.function)
				plat->thinker.function = (void (__cdecl *)(void))T_PlatRaise;
			else
				plat->thinker.function = (void (__cdecl *)(void))T_PlatInStasis;

			P_AddFNThinker (&plat->thinker);
			P_AddActivePlat(plat);
			break;
			
		case tc_flash:
			flash = (lightflash_t*)W_Calloc(sizeof(lightflash_t), THINKID, 0);
			SV_Read(flash, sizeof(lightflash_t));
			flash->sector = &sectors[GetSector((int)flash->sector, "tc_flash")];
			flash->thinker.function = (void (__cdecl *)(void))T_LightFlash;
			P_AddFNThinker (&flash->thinker);
			break;
			
		case tc_strobe:
			strobe = (strobe_t*)W_Calloc(sizeof(strobe_t), THINKID, 0);
			SV_Read(strobe, sizeof(strobe_t));
			strobe->sector = &sectors[GetSector((int)strobe->sector, "tc_strobe")];
			strobe->thinker.function = (void (__cdecl *)(void))T_StrobeFlash;
			P_AddFNThinker (&strobe->thinker);
			break;
			
		case tc_glow:
			glow = (glow_t*)W_Calloc(sizeof(glow_t), THINKID, 0);
			SV_Read(glow, sizeof(glow_t));
			glow->sector = &sectors[GetSector((int)glow->sector, "tc_glow")];
			glow->thinker.function = (void (__cdecl *)(void))T_Glow;
			P_AddFNThinker (&glow->thinker);
			break;
			
		case tc_flicker:
			flicker = (fireflicker_t*)W_Calloc(sizeof(fireflicker_t), THINKID, 0);
			SV_Read(flicker, sizeof(fireflicker_t));
			flicker->sector = &sectors[GetSector((int)flicker->sector, "tc_flicker")];
			flicker->thinker.function = (void (__cdecl *)(void))T_FireFlicker;
			P_AddFNThinker (&flicker->thinker);
			break;
		
		case tc_elevator:
			elevator = (elevator_t*)W_Calloc(sizeof(elevator_t), THINKID, 0);
			SV_Read(elevator, sizeof(elevator_t));
			elevator->sector = &sectors[GetSector((int)elevator->sector, "tc_elevator")];
			elevator->sector->floordata = elevator;
			elevator->sector->ceilingdata = elevator;
			elevator->thinker.function = (void (__cdecl *)(void))T_MoveElevator;
			P_AddFNThinker (&elevator->thinker);
			break;
		
		case tc_scroll:
			scroll = (scroll_t*)W_Calloc(sizeof(scroll_t), THINKID, 0);
			SV_Read(scroll, sizeof(scroll_t));
			scroll->thinker.function = (void (__cdecl *)(void))T_Scroll;
			P_AddFNThinker (&scroll->thinker);
			break;
		
		case tc_pusher:
			pusher = (pusher_t*)W_Calloc(sizeof(pusher_t), THINKID, 0);
			SV_Read(pusher, sizeof(pusher_t));
			pusher->thinker.function = (void (__cdecl *)(void))T_Pusher;
			pusher->source = P_GetPushThing(GetSector(pusher->affectee, "tc_pusher"));
			P_AddFNThinker (&pusher->thinker);
			break;

		case tc_phased:
			phased = (phased_t*)W_Calloc(sizeof(phased_t), THINKID, 0);
			SV_Read(phased, sizeof(phased_t));

			if (phased->count < 1 || phased->count > 256)	// 256 arbitrarily chosen
				Con_Error("tc_phased: Bad count %i", phased->count);

			phased->slist = (int*)W_Calloc(2*phased->count*sizeof(int), THINKID, 0);

			for (i = 0; i < 2*phased->count; i++)
				phased->slist[i] = SV_ReadLong();
			if (phased->count == 1)
				phased->thinker.function = (void (__cdecl *)(void))T_SinglePhased;
			else
				phased->thinker.function = (void (__cdecl *)(void))T_Phased;

			P_AddFNThinker (&phased->thinker);
			break;

		// SCRIPTED thinkers

		case tc_modelkill:

			modelkill = (modelkill_t*)W_Calloc(sizeof(modelkill_t), THINKID, 0);
			SV_Read(modelkill, sizeof(modelkill_t));
			modelkill->thinker.function = (void (__cdecl *)(void))T_ModelKill;
			P_AddFNThinker (&modelkill->thinker);
			break;

		case tc_modelspawn:

			modelspawn = (modelspawn_t*)W_Calloc(sizeof(modelspawn_t), THINKID, 0);
			SV_Read(modelspawn, sizeof(modelspawn_t));
			modelspawn->thinker.function = (void (__cdecl *)(void))T_ModelSpawn;
			P_AddFNThinker (&modelspawn->thinker);
			break;

		case tc_modelreset:

			modelreset = (modelreset_t*)W_Calloc(sizeof(modelreset_t), THINKID, 0);
			SV_Read(modelreset, sizeof(modelreset_t));
			modelreset->thinker.function = (void (__cdecl *)(void))T_ModelReset;
			P_AddFNThinker (&modelreset->thinker);
			break;

		case tc_switchreset:

			switchreset = (modelreset_t*)W_Calloc(sizeof(modelreset_t), THINKID, 0);
			SV_Read(switchreset, sizeof(modelreset_t));
			switchreset->thinker.function = (void (__cdecl *)(void))T_SwitchReset;
			P_AddFNThinker (&switchreset->thinker);
			break;

		case tc_soundactivate:

			soundactivate = (scriptedsound_t*)W_Calloc(sizeof(scriptedsound_t), THINKID, 0);
			SV_Read(soundactivate, sizeof(scriptedsound_t));
			soundactivate->thinker.function = (void (__cdecl *)(void))T_SoundActivate;
			P_AddFNThinker (&soundactivate->thinker);
			if ( (i = soundactivate->t79id) > 0)
				th7900[i - 1].ss = soundactivate;
			soundactivate->delay = 0;
			break;

		case tc_delayunblock:

			delayunblock = (delayline_t*)W_Calloc(sizeof(delayline_t), THINKID, 0);
			SV_Read(delayunblock, sizeof(delayline_t));
			delayunblock->thinker.function = (void (__cdecl *)(void))T_DelayUnblock;
			P_AddFNThinker (&delayunblock->thinker);
			break;

		case tc_delayresurrect:

			delayresurrect = (modelreset_t*)W_Calloc(sizeof(modelreset_t), THINKID, 0);
			SV_Read(delayresurrect, sizeof(modelreset_t));
			delayresurrect->thinker.function = (void (__cdecl *)(void))T_DelayResurrect;
			P_AddFNThinker (&delayresurrect->thinker);
			break;

		case tc_statemonitor:

			statemonitor = (modelreset_t*)W_Calloc(sizeof(modelreset_t), THINKID, 0);
			SV_Read(statemonitor, sizeof(modelreset_t));
			statemonitor->thinker.function = (void (__cdecl *)(void))T_StateMonitor;
			P_AddFNThinker (&statemonitor->thinker);
			break;

		case tc_endlevel:

			endlevel = (delayline_t*)W_Calloc(sizeof(delayline_t), THINKID, 0);
			SV_Read(endlevel, sizeof(delayline_t));
			endlevel->thinker.function = (void (__cdecl *)(void))T_Endlevel;
			P_AddFNThinker (&endlevel->thinker);
			break;

		case tc_killdelayed:

			killdelayed = (delayline_t*)W_Calloc(sizeof(delayline_t), THINKID, 0);
			SV_Read(killdelayed, sizeof(delayline_t));
			killdelayed->thinker.function = (void (__cdecl *)(void))T_KillDelayed;
			P_AddFNThinker (&killdelayed->thinker);
			break;

		case tc_lineocclude:

			lineocclude = (lineocclude_t*)W_Calloc(sizeof(lineocclude_t), THINKID, 0);
			SV_Read(lineocclude, sizeof(lineocclude_t));
			lineocclude->thinker.function = (void (__cdecl *)(void))T_LineOcclude;
			P_AddFNThinker (&lineocclude->thinker);
			break;

		case tc_adjustplat:

			adjustplat = (adjustplat_t*)W_Calloc(sizeof(adjustplat_t), THINKID, 0);
			SV_Read(adjustplat, sizeof(adjustplat_t));
			adjustplat->thinker.function = (void (__cdecl *)(void))T_AdjustCeil;
			P_AddFNThinker (&adjustplat->thinker);
			break;

		case tc_listsound:

			listsound = (delayline_t*)W_Calloc(sizeof(delayline_t), THINKID, 0);
			SV_Read(listsound, sizeof(delayline_t));
			listsound->thinker.function = (void (__cdecl *)(void))T_PlayListsound;
			P_AddFNThinker (&listsound->thinker);
			break;

			// v220-31 New Thinkers 
		case tc_activatedelay:

			activatedelayed = (modelreset_t*)W_Calloc(sizeof(modelreset_t), THINKID, 0);
			SV_Read(activatedelayed, sizeof(modelreset_t));
			activatedelayed->thinker.function = (void (__cdecl *)(void))T_ActivateDelayed;
			P_AddFNThinker (&activatedelayed->thinker);
			break;

		// DYNLIGHT THINKERS 5505

		case tc_dynglow:

			dynglow = (dynglow_t*)W_Calloc(sizeof(dynglow_t), THINKID, 0);
			SV_Read(dynglow, sizeof(dynglow_t));
			dynglow->thinker.function = (void (__cdecl *)(void))T_DynGlow;
			P_AddFNThinker (&dynglow->thinker);
			break;

		case tc_dynfireflicker:
			dynfireflicker = (dynfireflicker_t*)W_Calloc(sizeof(dynfireflicker_t), THINKID, 0);
			SV_Read(dynfireflicker, sizeof(dynfireflicker_t));
			dynfireflicker->thinker.function = (void (__cdecl *)(void))T_DynFireFlicker;
			P_AddFNThinker (&dynfireflicker->thinker);
			break;

			// v220-31 End of new Thinkers 
		case tc_endspecials:
				return;	// end of list
			
		default:
			Con_ErrorDelete("invalid thinker class");
		}
    }
}

static void P_ArchiveBrain(void)
{
	int i = numbraintargets;

	if (braineasy) i |= 0x80;
	SV_WriteByte((byte)i);
	SV_WriteByte((byte)braintargeton);

	// Write the mobj id references
	for (i = 0; i < numbraintargets; i++)
	{
		mobj_t *mo = braintargets[i];
		thid_t id = mo ? mo->thinker.id : 0;

		SV_WriteShort(id);
	}
}

static void P_UnArchiveBrain(void)
{
	int i;

	numbraintargets = SV_ReadByte();

	if (numbraintargets & 0x80)
	{
		braineasy = 1;
		numbraintargets &= 0x7f;
	}

	braintargeton = SV_ReadByte();

	for (i = 0; i < numbraintargets; i++)
	{
		braintargets[i] = SetReferencedMobj((thid_t)SV_ReadShort());
	}
}

static void th7800flags(RDBOOL restore)
{
	short	flagsarray[NUMSPOTS];
	int		i;

	if (restore)
	{
		lzRead(flagsarray, NUMSPOTS * sizeof(short), savefile);
		for (i = 0; i < NUMSPOTS; i++) th7800[i].setflags = flagsarray[i];
	}
	else
	{
		for (i = 0; i < NUMSPOTS; i++) flagsarray[i] = th7800[i].setflags;
		lzWrite(flagsarray, NUMSPOTS * sizeof(short), savefile);
	}
}

static void SV_ArchiveCRCs()
{
	int chksecs, chklines, chkthings, i;

	for (i = 0; i <= MAX_PWADS; i++)
	{
		if (pwadchecksum[i]) SV_WriteLong(pwadchecksum[i]);
	}

	setCHECKsize(&chksecs, &chklines, &chkthings, -2);

	SV_WriteLong(chklines);
	SV_WriteLong(chksecs);
	SV_WriteLong(chkthings);
}

static int SV_UnarchiveCRCs()
{
	int i;

	for (i = 0; i <= MAX_PWADS; i++)
	{
		if (pwadchecksum[i] && (uint)SV_ReadLong() != pwadchecksum[i])
		{
			return false;
		}
	}

	if (file_version == SAVEGAME_VER)
	{
		char lumpname[8];
		int chksecs, chklines, chkthings, lump, mapnum;

		if (hdr.gamemode == commercial)
		{
			sprintf(lumpname, "map%02i", hdr.map);
			mapnum = hdr.map;
		}
		else
		{
			sprintf(lumpname,"E%iM%i", hdr.episode & 0x0F, hdr.map);
			mapnum = (hdr.episode & 0x0F) * 10 + (hdr.map);
		}

		if ((lump = W_CheckNumForName(lumpname)) < 0) return false;

		setCHECKsize(&chksecs, &chklines, &chkthings, lump);

		if (SV_ReadLong() != chklines) return false;
		if (SV_ReadLong() != chksecs ) return false;
		if (SV_ReadLong() != chkthings) return false;
		return true;
	}

	return false;
}

void SV_SaveGameFile(int slot)
{
	// Form the file name
	sprintf(savefname, SAVEGAMENAME"%i.r3dsg", slot);
	sprintf(savepathfname, "%s%s", base_save_path, savefname);
}

int SV_SaveGame(int slot, char *description)
{
	byte	*ptr;
    int		i, n;

	SV_SaveGameFile(slot);

	// Open the file.
	if (!OpenSavefile("wp"))
	{
		Con_Message("P_SaveGame: couldn't open \"%s\" for writing.\n", savepathfname);

		return false;	// No success.
	}

	P_RunThinkers();	// Ensure all dead thinkers are removed

	// Version.
	SV_WriteLong(SAVEGAME_VER|SAVEGAME_FLAGS);	// Version

	// Set the header values.
	hdr.magic      = use_xl_lines ? MY_XLINE_SAVE_MAGIC : MY_SAVE_MAGIC;
	hdr.gamemode   = gamemode;

	strncpy(hdr.description, description, SAVESTRINGSIZE - 2);
	hdr.description[SAVESTRINGSIZE - 1] = 0;

	hdr.skill      = (byte)(gameskill&0x7);
	hdr.episode    = gameepisode;
	hdr.episode   |= 0x80;				// High bit indicates a pwad name is saved (backward compatibility issue).
	hdr.map        = gamemap;
	hdr.deathmatch = false;
	hdr.nomonsters = nomonsters;
	hdr.respawn    = (respawnmonsters ? 1 : 0);
	hdr.leveltime  = leveltime;
	hdr.cmdvars = 0;

	if (useDynLights) hdr.cmdvars |= 0x00000001;
	if (cfg.useTorchLight) hdr.cmdvars |= 0x00000002;
	if (cfg.sineCompress) hdr.cmdvars |= 0x00000004;
	if (cfg.setVtxLight) hdr.cmdvars |= 0x00000008;
	if (ma_inf.addlolight) hdr.cmdvars |= 0x00000010;

	// b211-0 Compatibility checks
	hdr.switch_flags  = ge_inf.switch_flags;		// b211-0
	hdr.sprites_count = numsprites;					// b211-0
	hdr.states_count  = ge_inf.states_count;		// b211-0

	for(i=0; i<MAXPLAYERS; i++)
		hdr.players[i] = mbf_players[i].plr->ingame;

	// Write the header.
	lzWrite(&hdr, sizeof(hdr), savefile);

	strcpy (pwadname, mainPwadname);
	lzWrite(pwadname, MAXPWADNAME, savefile);

	SV_ArchiveCRCs();
	SV_InitTextureArchives();

	ptr = S_FetchASFLAGS();
	lzWrite(ptr, MAXAMBIENT, savefile);
	M_Free(ptr);

	// v220-4 save the addstats values
	ptr = G_CopyAddstats(&i);
	lzWrite(ptr, i, savefile);
	M_Free(ptr);

	SV_Write(&wadani, sizeof(wadani_t));
    P_ArchivePlayers();
    P_ArchiveWorld();
    P_ArchiveThinkers();
    P_ArchiveSpecials();
	P_ArchiveBrain();

	// v220-17 store markpoint
	SV_WriteLong(*(int*)"MARK");	// id used to identify that the
	SV_WriteLong(*(int*)"PNTS");	// savegame has markpoints stored
	ptr = AM_GetMarkPoints(&i, &n);
	lzWrite(ptr, i, savefile);
	SV_WriteByte((byte)n);
	M_Free(ptr);

	SV_WriteLong(ma_inf.massacre_count);
	SV_WriteLong(ma_inf.collateraldamage_count);
	
	// Save button list
	SV_WriteButtonList();

	lzWrite(soundstarts, MAXLINESOUNDS * sizeof(int), savefile);

	// Save secmo timer values
	SV_WriteSecmoVars();

	// v220-31 store hudsprite and 7800 flags
	if (R3D_ScriptedMap)
	{
		SV_WriteLong(NGL_getFlags());
		th7800flags(false);
	}

	// To be absolutely sure...
	SV_WriteByte(CONSISTENCY);
	
	CloseSavefile();

	return true;
}

// Checks validity and opens file.
// Closing file is the caller's responsibility if closefile == false.
// v220-35 rewritten
RDBOOL SV_CheckLoadGame(RDBOOL checkswitch, RDBOOL versiononly, RDBOOL startup)
{
	if ( !(savegametime = (uint)getFileTime(savepathfname)) )
	{
		if (checkswitch)
		{
			sprintf(savefileDeletestr, "LOADGAME ERROR: Savegame not found;\n");
			subfn_HandleMessage(false);
		}
		return false;	// File does not exist
	}

	pwadname[0] = 0;

	if (!OpenSavefile("rp"))
	{
		if (checkswitch)
		{
			sprintf(savefileDeletestr, "LOADGAME ERROR: Savegame could not be opened;%s", delmsg);
			subfn_HandleMessage(false);
		}
		return false;	// File does not exist
	}

	file_version  = SV_ReadLong();
	file_flags    = (file_version & FLAGS_MASK);
	file_version &= VERSION_MASK;

	if (file_version != SAVEGAME_VER)	// Version
	{
		if (checkswitch)
		{
			pr_OFinsert("STATES CHECK %i %i\n", hdr.states_count, ge_inf.states_count);
			if (hdr.states_count != ge_inf.states_count)
			{
				sprintf(savefileDeletestr, "LOADGAME ERROR: states mismatch.\nCheck launcher selections.\n");
				subfn_HandleMessage(true);
				return false;
			}
			sprintf(savefileDeletestr, "LOADGAME ERROR: Savegame is an old version which is no longer supported;%s", delmsg);
			subfn_HandleMessage(true);
		}
		return false;
	}
	else
	{
		lzRead(&hdr, sizeof(hdr), savefile);		// Read the header.

		if (versiononly) return true;

		if ((hdr.magic & ~MY_XLINE_CHECK) != MY_SAVE_MAGIC) 
		{
			if (checkswitch)
			{
				sprintf(savefileDeletestr, "LOADGAME ERROR: Savegame's header is invalid;%s", delmsg);
				subfn_HandleMessage(true);
			}
			return false;
		}
		else
		{
			if (hdr.episode & 0x80)
			{
				lzRead(pwadname, MAXPWADNAME, savefile);
			}

			if (hdr.gamemode != gamemode)
			{
				if (pwadname[0])
				{
					sprintf(savefileDeletestr, "LOADGAME ERROR: Savegame is for %s;\n", pwadname);
					subfn_HandleMessage(false);
				}
				else
				{
					if (checkswitch)
					{
						sprintf(savefileDeletestr, "LOADGAME ERROR: Savegame's gamemode is invalid;\n%s", delmsg);
						subfn_HandleMessage(true);
					}
				}
				return false;
			}

			if (!SV_UnarchiveCRCs())
			{
				char buff[512];

				buff[0] = 0;

				if (pwadname[0])
				{
					if (!_stricmp(pwadname, mainPwadname))
					{
						int r3dUserlen = (int)strlen(r3dUserDir.path);
						int is_pwads = r3dUserlen > 4 && !stricmp("pwads\\", r3dUserDir.path + r3dUserlen - 6);

						sprintf(buff,"\nCRC ERROR : %s\n\n"
									 "Possible causes;\n\n"
									 "1) Wrong IWAD selected\n"
									 "2) Game wad has been changed\\updated\n", mainPwadname);

						if (!startup && is_pwads)
							strcat(buff, "3) Nodes rebuilt using different type\n");

					}
					else
					{
						sprintf(buff,"\nSavegame for : %s\n"
									 "Pwad loaded  : %s\n",
									  pwadname, mainPwadname);
					}
				}
				else
					sprintf(buff, " savegame probably not for this iwad.\n");

				if (startup)
				{
					Sys_MessageBox(buff, false);
					return false;
				}

				if (checkswitch)
				{
					sprintf(savefileDeletestr, "LOADGAME ERROR:%s\n", buff);
					subfn_HandleMessage(true);
				}
				return false;
			}
		}

		// v220-7 check if the save date is ahead of any wad dates
		if (checkswitch && MapFileDateGreaterThan(savegametime))
		{
			sprintf(savefileDeletestr, "LOADGAME ERROR: Savegame's date pre-dates the wad's%s", delmsg);
			subfn_HandleMessage(true);
			return false;
		}
	}
	
	if (checkswitch)	// b211-0 Check status of switches - modified v220-3
	{
		int flagmatch = ge_inf.switch_flags == hdr.switch_flags;

		if (   hdr.sprites_count != numsprites
			|| hdr.states_count  != ge_inf.states_count)
		{
			if (flagmatch)
			{
				sprintf(savefileDeletestr,
					"LOADGAME ERROR: states, deds or models have changed since saving.\n%s", delmsg);
				subfn_HandleMessage(true);
			}
			else
			{
				checkSwitchStatus(hdr.switch_flags);
			}

			return false;
		}

		if (!flagmatch && !checkSwitchStatus(hdr.switch_flags))
			return false;
	}

	return true;
}

// For internal calls (menu function etc.)
int SV_GetSaveDescription(char *str)
{
	int		fileok = false;
	RDBOOL	valid = SV_CheckLoadGame(false, true, false);
	
	savefileDeletestr[0] = 0;	// only checking
	strcpy(str, "EMPTY SLOT");	// Assume empty until checked

	if (valid)
	{
		if (file_version == SAVEGAME_VER)
		{
			if (valid)
			{
				strncpy(str, hdr.description, SAVESTRINGSIZE - 2);
				str[SAVESTRINGSIZE - 1] = 0;
				fileok = true;
			}
			else if (hdr.episode & 0x80)
			{
				// Insert details of pwad file is meant for
				if (_stricmp(pwadname, mainPwadname)) // Check it's not this one
				{
					char	*src = pwadname, *dest;
					int		i, len = strlen(src);

					if (len > SAVESTRINGSIZE - 10) len = SAVESTRINGSIZE - 10;

					if (len)
					{
						strcpy(str, "FILE FOR ");
						dest = str + 9;

						for (i = 0; i < len; i++)
						{
							*dest++ = toupper((int)*src++);
						}
						*dest = 0;
					}
					else
					{
						strcpy(str, "FILE FOR UNKNOWN IWAD");
					}
				}
			}
		}
	}
	
	CloseSavefile();
	return fileok;
}

//===========================================================================
// SV_LoadGame
//===========================================================================
int SV_LoadGame()
{
	const char *skillstr[] = { "1-easy", "2-medium", "3-normal", "4-hard", "5-nightmare" };
	RDBOOL	infile[MAXPLAYERS], loaded[MAXPLAYERS];
	byte	endbyte;
	int		i, skill;
	byte	*ptr;

	if (!SV_CheckLoadGame(true, false, false))	// b211-0
	{
		CloseSavefile();
		return false;
	}

	skill = hdr.skill & 0x3;	// 0 - 4
	gameskill &= 0x3;			// Just in case is out of range

	if (gameskill != skill)
		pr_OFinsert("WARNING: The \"%s\" game skill saved will override that of \"%s\" set in the launcher\n", skillstr[skill], skillstr[gameskill]);

	gameepisode	 = gamemode == commercial ? 0 : (hdr.episode & 0x0F);
	gamemap		 = hdr.map;
	// hdr.deathmatch; not used
	nomonsters   = hdr.nomonsters;
	respawnmonsters  = (hdr.respawn & 1) > 0;
		
	if (!nomonsters && wasnomonsters)
	{
		CloseSavefile();
		Con_ErrorD( "Saved game needs monster definitions.\n"
					"Disable nomonsters before restarting");
	}

	// Only current player set. Kept for backward savegame compatibility
	for(i = 0; i < MAXPLAYERS; i++) infile[i] = hdr.players[i];

	// Make sure an opening briefing is not shown.
	brief_disabled = -1;

	useDynLights =  (hdr.cmdvars & 0x00000001) != 0;
	if (useDynLights) ge_inf.flags |= GEF_DYNLIGHTS;

	// Load the level.
	G_InitNew(skill, gameepisode, gamemap);	// gameskill will be set here

	// v220-25 check the savegame is compatible with an xline nodebuild
	if ((num_xlsame && !(hdr.magic & MY_XLINE_CHECK)) || (!num_xlsame && (hdr.magic & MY_XLINE_CHECK)))
	{
		CloseSavefile();
		
		if (Sys_CriticalMessage("LOADGAME ERROR: file is incompatible with current nodes and\n"
								"    continuing will delete it and restart the game from new.", true))
		{
			if (remove(savepathfname))
				Con_Error("ERROR: Could not delete %s : %s\n", savepathfname, strerror(errno));

			return false;
		}

		Sys_Quit(false);
	}

	cfg.useTorchLight = (hdr.cmdvars & 0x00000002) != 0;
	cfg.sineCompress = (hdr.cmdvars & 0x00000004) != 0;
	cfg.setVtxLight = (hdr.cmdvars & 0x00000008) != 0;
	ma_inf.addlolight = (hdr.cmdvars & 0x00000010) != 0;

	// Set the time.
	leveltime = hdr.leveltime;

	// Restore the ASFLAGS
	ptr = (byte*)M_Calloc(MAXAMBIENT, 0);
	lzRead(ptr, MAXAMBIENT, savefile);
	S_RestoreASFLAGS(ptr);
	M_Free(ptr);

	// v220-4 restore the addstats values
	ptr = G_SetAddstats(&i);
	lzRead(ptr, i, savefile);
	
	G_GetFriendStats();

	// init the mobj array, thinkers and blockrings
	ResetArrays(true);
	
	// Dearchive all the data.
	SV_Read(&wadani, sizeof(wadani_t));
	P_UnArchivePlayers(infile, loaded);
	P_UnArchiveWorld();
	P_UnArchiveThinkers();
	P_UnArchiveSpecials();
	P_UnArchiveBrain();

	// v220-35 MARK & PNTS retained as a double check
	if (   SV_ReadLong() == *(int*)"MARK"
		&& SV_ReadLong() == *(int*)"PNTS")
	{
		ptr = AM_GetMarkPoints(&i, NULL);
		lzRead(ptr, i, savefile);
		i = SV_ReadByte();
		AM_SetMarkPoints(ptr, (uint)i);
		M_Free(ptr);
	}

	ma_inf.massacre_count = SV_ReadLong();
	ma_inf.collateraldamage_count = SV_ReadLong();

	// Regenerate button list
	SV_ReadButtonList(SV_ReadShort());

	lzRead(soundstarts, MAXLINESOUNDS * sizeof(int), savefile);
	SV_ReadSecmoVars((int)SV_ReadShort());

	// v220-31 restore hud sprite ignore flags and 7800 flags
	if (R3D_ScriptedMap)
	{
		NGL_setFlags(SV_ReadLong());
		th7800flags(true);
	}

	endbyte = SV_ReadByte();

	CloseSavefile();		// We're done.

	// Check consistency.
	if (endbyte != CONSISTENCY)
		Con_ErrorDelete("consistency test failed");

	LS_InitTagLists();				// Update hash table

	// Notify the players that weren't in the savegame.
	for(i=0; i<MAXPLAYERS; i++)
	{
		if (!loaded[i] && mbf_players[i].plr->ingame)
		{
			if (!i) P_SetMessage(mbf_players, GET_TXT(TXT_LOADMISSING));

			if (i != 0)
			{
				Con_Message("Error: player %i set to ingame cleared", i);
				mbf_players[i].plr->ingame = false;
			}
		}
	}

	// v220-33 check if MF2_NOTLIT needs setting
	{
		thinker_t	*th;
		mobj_t		*mo;

		for (th = thinkercap.next; th != &thinkercap; th = th->next)
		{
			if ((void (__cdecl *)(void))th->function == (void (__cdecl *)(void))P_MobjThinker)
			{
				mo = (mobj_t*) th;

				if (mo->type == MT_PUFF || mo->flags & MF_NOSECTOR)
				{
					mo->flags2 |= MF2_NOTLIT;
				}

				// Needed to reset MF2_NOTLIT after loading a savegame
				if (   hasDEHactors
					&& mo->sprite > 0
					&& mo->sprite < numsprites
					&& mo->type < DH_NUMOBJECTS
					&& sprites[mo->sprite].sprflags & SPR_NOMODEL)
				{
					mo->flags2 |= MF2_NOTLIT;
				}

				if (wadani.mapnummorespawn >= 0 && wadani.mapnummorespawn == MAPNUM(mo))
				{
					wadani.mobj = mo;
				}
			}
		}
	}

	P_CreateDynLightStructures(true);	// v220-31
	PH_CheckTorchGlow();				// v220-33
	P_SetupDynlightList();				// v220-31

	MB_Init(true);				// Regenerate all amodel doors/boxes and set to saved position
	TS_CheckLinkedModels();		// Regenerate linked flame models


	// Update the scsm values
	for (i = 0; i < scsm.fnum; i++)		// floor
	{
		scsm.f[i].oldx = scsm.f[i].dx = scsm.f[i].sdx = sectors[scsm.f[i].s].flooroffx;
		scsm.f[i].oldy = scsm.f[i].dy = scsm.f[i].sdy = sectors[scsm.f[i].s].flooroffy;
	}

	for (i = 0; i < scsm.cnum; i++)		// ceiling
	{
		scsm.c[i].oldx = scsm.c[i].dx = scsm.c[i].sdx = sectors[scsm.c[i].s].ceiloffx;
		scsm.c[i].oldy = scsm.c[i].dy = scsm.c[i].sdy = sectors[scsm.c[i].s].ceiloffy;
	}

	for (i = 0; i < scsm.snum; i++)		// side
	{
		scsm.s[i].oldx = scsm.s[i].dx = scsm.s[i].sdx = sides[scsm.s[i].s].textureoffset;
		scsm.s[i].oldy = scsm.s[i].dy = scsm.s[i].sdy = sides[scsm.s[i].s].rowoffset;
	}

	R3D_SetParms(P_GetCurrentMapid());
	Con_Message("LOADGAME episode %i map %i: success\n\n", gameepisode, gamemap);
	ma_inf.loadgame = true;

	for (i = 0; i < MAXLINESOUNDS; i++)
	{
		if (soundstarts[i])
		{
			S_SoundRepeat(soundstarts[i]|(HSF_REPEAT|HSF_NO_ATTENUATION), NULL, 1.0f, 0);
		}
	}

	/*
	Con_Message(">>>> Restored massacre count = %i\n", ma_inf.massacre_count);
	Con_Message(">>>> Restored collateral damage count = %i\n", ma_inf.collateraldamage_count);
	*/

	return true;
}
