//-----------------------------------------------------------------------------
//
// Copyright (C) 1998 - 2005 id Software, Inc., Jaakko Kernen, G Jackson
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_start.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified for R3D, moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll code
//
// DESCRIPTION: player starts (note: netgame not supported)
//
// HEADER FILES ------------------------------------------------------------

#include <string.h>

#include "doom/p_local.h"
#include "doom/p_start.h"
#include "doom/g_game.h"

#include "con_console.h"
#include "net_main.h"
#include "m_misc.h"
#include "r_data.h"
#include "r_main.h"
#include "p_maputil.h"
#include "sys_system.h"

#include "r3d_compat.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// Maintain single and multi player starting spots.
typedef struct {
	mapthing_t	playerstart;
	int			mapnum;
} mapplr_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int		numplayerstarts;
mobj_t	*teleport_flash;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static mapplr_t	*playerstarts = NULL;

// CODE --------------------------------------------------------------------

//==========================================================================
// P_RegisterPlayerStart
//==========================================================================
void P_RegisterPlayerStart(mapthing_t *mthing, int mapnum)
{
	static int maxplayerstarts = 0;

	// Enough room?
	if (numplayerstarts >= maxplayerstarts)
	{
		maxplayerstarts += 16;
		playerstarts = (mapplr_t*)W_Realloc(playerstarts,
			maxplayerstarts * sizeof(mapplr_t), STATICID, 0);
	}

	memcpy(&playerstarts[numplayerstarts].playerstart, mthing, sizeof(mapthing_t));
	playerstarts[numplayerstarts].mapnum = mapnum;
	numplayerstarts++;
}

//==========================================================================
// P_DealPlayerStarts  v220-4 modified
//==========================================================================
void P_DealPlayerStarts()
{
	int	i, n = numplayerstarts;

	// Find the highest numbered start with doomednum 1
	// Failing this sequentially check for spots having a doomednum 2 to 4
	for (i = 1; i <= MAXPLAYERS; i++)
	{
		while (--n >= 0)
		{
			if (playerstarts[n].playerstart.doomednum == i)
			{
				mbf_players[0].startspot = n;
				return;
			}
		}
	}

	Con_Error("No playerstarts!\n");
}

//==========================================================================
// P_SpawnPlayers
//	Spawns all players, using the method appropriate for current game mode.
//	Called during level setup.
//==========================================================================

static line_t *llist[16];
static int llistcnt;

static RDBOOL PIT_LineCheck(line_t* ld, int *data)
{
	if (ld->ml_flags & ML_TWOSIDED) return true;

	if (   data[BOXRIGHT] <= ld->bbox[BOXLEFT ]
		|| data[BOXLEFT ] >= ld->bbox[BOXRIGHT]
		|| data[BOXTOP  ] <= ld->bbox[BOXBASE ]
		|| data[BOXBASE ] >= ld->bbox[BOXTOP  ]
		|| P_BoxOnLineSide(data, ld) != -1)
	{
		return true;
	}

	if (llistcnt < 16)
	{
		llist[llistcnt++] = ld;
	}
	return true;
}

static void setLineList(fixed_t x, fixed_t y, fixed_t r)
{
	int		xl, xh;
	int		yl, yh;
	int		bx, by;
	int		box[4];
	//-------------------

	llistcnt = 0;

	// v220-4 Check the position is valid
	M_setBoxOffsets(box, x, y, r);
	M_getBlockOffsets(box, 0, &xl, &xh, &yl, &yh);

	validcount++;

	for (bx = xl ; bx <= xh ; bx++)
	{
		for (by = yl ; by <= yh ; by++)
			P_BlockLinesIterator(bx, by, (int (__cdecl *)(struct line_s *,void *))PIT_LineCheck, box);
	}
}

// v220-4 check if position/radius would intersect a 1's line
RDBOOL checkLineSpawnedTP(fixed_t x, fixed_t y, fixed_t r)
{
	// For a scripted map spawning stuff in 1's lines is not
	// allowed. If it occurs in a standard map then this has
	// always happened and no judgement can be made. If
	// prevented all kinds of weird stuff might happen.
	if (R3D_ScriptedMap)
		setLineList(x, y, r);
	else
		llistcnt = 0;

	return !llistcnt;
}

static int PIT_PlayerZpos(sector_t *sec, mobj_t *mo)
{
	fixed_t diff = sec->floorheight - mo->z;

	if (diff > 0 && diff <= MAXSTEP)
	{
		RDBOOL x = (sec->sibounds[BOXRIGHT] + sec->sibounds[BOXLEFT])/2.0f > FIX2FLT(mo->x);
		RDBOOL y = (sec->sibounds[BOXTOP  ] + sec->sibounds[BOXBASE])/2.0f > FIX2FLT(mo->y);

		mo->z = sec->floorheight;	// adjust to step height
		mo->momx = x ? 1 : -1;		// set the minimum momentum to get P_Move to kick in
		mo->momy = y ? 1 : -1;
		return false;
	}

	return true;
}

// Players have not been spawned at this point as map analysis must first be
// completed to allow for other map issues to be checked for at spawn time.
int playerGNUM(int num)
{
	mapplr_t	*mt;
	int			i;
	// Note: the startspot has been set in P_DealPlayerStarts()
	for (i = 0, mt = playerstarts; i < numplayerstarts; i++, mt++)
	{
		if (mbf_players[num].startspot == i)
			return R_PointInSubsector(mt->playerstart.x<<16, mt->playerstart.y<<16)->viewgroup;
	}

	Con_ErrorD("Bad map error - the map is unusable");
	return 0;	// happy compiler
}

// v220-33 revised
void P_SpawnPlayers()
{
	mapplr_t	*mt;
	mapthing_t	*mp;
	mobj_t		*mo = NULL;
	int			box[4];
	int			i, n, x, y;
	int			r;

	//------------------

	for (i = 1; i <= MAXPLAYERS; i++)
	{
		n = -1;	
		mt = playerstarts;

		while (++n < numplayerstarts)
		{
			if (mt[n].playerstart.doomednum == i)
			{
				mo = P_SpawnPlayer(&mt[n].playerstart, i - 1, mt[n].mapnum<<16);	// Spawn player
				// v220-35 In some maps voodoo dolls get stuck. Setting
				//         their radius to 14 map units solves the problem
				mo->radius = (fixed_t)(14.f * FRACUNIT);
				ma_inf.voodoodolls++;
				mp = &mt[n].playerstart;
			}
		}

		if (mo)
		{
			mo->radius = (fixed_t)mo->info->radius<<16;
			ppt.view_subsector = mo->subsector;
			plr_viewgroup  = mo->subsector->viewgroup;
			ppt.viewplayer = mo->dplayer;
			ppt.viewplrmo  = mo;
			ppt.viewer_fake_sector = -1;
			ma_inf.voodoodolls--;
			break;	// don't spawn players with higher doomednums
		}
	}

	if (!mo) Con_ErrorD("ERROR: Map has no player");

	mo->dplayer->flags |= DDPF_LOCAL;
	mo->ddflags |= DDMF_DONTDRAW;

	P_SetupPsprites(&mbf_players[0]);		// setup pistol psprite

	// v220-3 check player's starting height
	P_ThingSectorsIterator(mo, (int (__cdecl *)(struct sector_s *,void *))PIT_PlayerZpos, mo);

	// Initialise ppt values, these are used by the renderer - R3D

	// The player's radius could be changed using a ded but this
	// is not viable. As such the radius is fixed here at 16<<FRACBITS
	mo->radius = mo->info->radius = 0x100000;

	// v220-4 In scythe2 map28 the author, in his infinite wisdom, has decided to spawn
	//        the player inside vertical bars. Since this is incompatible with changes
	//        made to the move code some adjustment has to be made.
	setLineList(mo->x, mo->y, mo->radius);

	if (!llistcnt) return;

	Con_Message("MAP ERROR: Console player is spawned in 1's lines;");

	r = mo->radius, x = mo->x, y = mo->y;
	M_ClearBox(box);

	for (i = 0; i < llistcnt; i++)
	{
		M_AddLineToBox(box, llist[i]);
		Con_Message(" %i", llist[i] - lines);
	}

	Con_Message("\n");

	if (mo->x > box[BOXLEFT] && mo->x < box[BOXRIGHT])
	{
		if (mo->y > box[BOXTOP] && mo->y - r < box[BOXTOP])
		{
			mo->y = box[BOXTOP] + r;
		}
		else
		if (mo->y < box[BOXBASE] && mo->y + r > box[BOXBASE])
		{
			mo->y = box[BOXBASE] - r;
		}
	}

	if (mo->y > box[BOXBASE] && mo->y < box[BOXTOP])
	{
		if (mo->x > box[BOXRIGHT] && mo->x - r < box[BOXRIGHT])
		{
			mo->x = box[BOXRIGHT] + r;
		}
		else
		if (mo->x < box[BOXLEFT] && mo->x + r > box[BOXLEFT])
		{
			mo->x = box[BOXLEFT] - r;
		}
	}

	setLineList(mo->x, mo->y, r);

	if (!llistcnt)
	{
		mp->x = mo->x>>16, mp->y = mo->y>>16;

		Con_Message("  player start has been moved from x %i y %i to x %i y %i\n",
			x>>16, y>>16, mo->x>>16, mo->y>>16);
	}
	else
		Sys_MessageBox("Player's start position is invalid and cannot be changed\nIf stuck use noclip\n", false);
}

static mobj_t **remArray;
static int	remArrayCount;
static int	remArrayMaxCount;

// v220-5 check if spawned in a decoration
// v220-33 add MF_SKULLFLY - may be used for an effect e.g. scrchhet map01
static RDBOOL PIT_CheckOverlap(mobj_t* thing, mobj_t *mo)
{
	fixed_t		blockdist;

	if (   thing->info->seestate
		|| !(thing->flags & MF_SOLID)
		|| thing->flags  & (MF_SPECIAL|MF_DROPPED|MF_NOCLIP|MF_CORPSE|MF_SKULLFLY)
		|| thing->flags2 & MF2_VDOOR
		|| thing->flags3 & (MF3_TOUCHY|MF3_AMODEL)
		|| thing->modelboxnum > 0)
	{
		return true;
	}

	blockdist = mo->radius + thing->radius;
	
	if (   abs(thing->x - mo->x) >= blockdist
		|| abs(thing->y - mo->y) >= blockdist )
	{
		return true;	// didn't hit it
	}

	if (   thing->z + thing->height >= mo->z
		&& mo->z + mo->height >= thing->z)
	{
		if (thing->flags & MF_COUNTKILL)
		{
			if (remArrayCount >= remArrayMaxCount)
			{
				remArrayMaxCount +=16;
				remArray = (mobj_t**)realloc(remArray, remArrayMaxCount);
			}

			remArray[remArrayCount++] = mo;
		}
		else
		{
			thing->flags |= MF_NOCLIP;	// Set the item to no clip

			if (thing->info->doomednum == 90)	// v220-9 recognised as a teleport flash
			{
				thing->flags2 |= MF2_NORENDER;
				teleport_flash = thing;
			}
#ifdef _DEBUG
			Con_Message("MAP ERROR: The player has been spawned in T%i Doomednum %i Type %i Flags %.8X\n", MAPNUM(thing), thing->info->doomednum, thing->type, thing->flags);
#else
			Con_Message("MAP ERROR: The player has been spawned in T%i\n", MAPNUM(thing));
#endif
		}
	}

	return true;
}

void P_KillMobj(mobj_t *source, mobj_t *target);

// v220-5 Rare but has been seen where a home brewed map has spawned the
//        player in a decoration. Unfortunately some earlier ports prior
//        to this have tolerated such nonsense which just helps lead to
//        these kind of amateur map authoring errors being made.
// v220-9 It appears that prBoom has used a dehacked Thing 90 (MT_STALAG).
//        a type not recognised by DeeP for example, to cause a TP flash.
void P_CheckPlayerCanMove(mobj_t *mo)
{
	int			box[4];
	int			xl, xh;
	int			yl, yh;
	int			bx, by;
	int			i;

	teleport_flash = NULL;

	M_setBoxOffsets(box, mo->x, mo->y, mo->radius);
	M_getBlockOffsets(box, MAXRADIUS, &xl, &xh, &yl, &yh);

	validcount++;
	mo->validcount = validcount;	// don't check self

	remArray = NULL;
	remArrayCount = 0;
	remArrayMaxCount = 0;

	for (bx = xl ; bx <= xh ; bx++)
	{
		for (by = yl ; by <= yh ; by++)
			P_BlockThingsIterator(bx, by, (int (__cdecl *)(struct mobj_s *,void *))PIT_CheckOverlap, mo);
	}

	for (i = 0; i < remArrayCount; i++)
	{
		if (remArray[i]->thinker.function == (void (__cdecl *)(void))P_MobjThinker)
			P_KillMobj(remArray[i], remArray[i]);	// Kill, don't remove
	}

	if (remArray) free(remArray);
}
