//-----------------------------------------------------------------------------
//
// Copyright (C) 1998 - 2005 id Software, Inc., Chi Hoang, Lee Killough,
//   Jim Flynn, Rand Phares, Ty Halderman, Jaakko Kernen
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_teleport.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  BOOM and R3D extensions added, reformatted, moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll code and BOOM v2.01
//
//
// DESCRIPTION:
//      Teleportation.
//
//-----------------------------------------------------------------------------

#include <stdio.h>

#include "doom/p_local.h"
#include "doom/p_spec.h"
#include "doom/p_view.h"
#include "doom/tables.h"
#include "m_misc.h"
#include "s_main.h"
#include "r_data.h"
#include "p_maputil.h"
#include "con_console.h"
#include "sys_system.h"

#include "r3d_defs.h"
#include "r3d_compat.h"
#include "r3d_edit.h"

void	P_ClearThingSRVO(mobj_t *mo, RDBOOL clrmpos);
int		FindIfThingIsOnLine(mobj_t *th, RDBOOL diffsec);

extern byte walkCheck[];

typedef struct {
	int		used;	// Used for node check
	int		mnum;	// Used for debugging
	int		tag;	// Used for debugging

	int		secn;
	int		exit;
	fixed_t x, y;
	angle_t	an;
} tpthing_t;

// v220-5 support for tpthings
static struct settp_t {
	fixed_t		posx;
	fixed_t		posy;
	angle_t		an;
	int			exit;

	tpthing_t	*tpthings;
	int			num_tpthings;
} settp;

//======================== TP_CheckTPSpawnedSectors
//
// Check for gl nodes error and/or map errors with MT_TELEPORTMAN
//

static RDBOOL tpSpecial(int num)
{
	switch(num)
	{
		// Thing teleporters
		// All are expected to link to a sector with an MT_TELEPORTMAN thing
		case 39:	// teleport trigger
		case 97:	// teleport retrigger
		case 174:
		case 195:
		case 125:	// teleport monsteronly trigger
		case 126:	// as 125
		case 207:	// silent thing teleporters
		case 208:	// as 207
		case 209:
		case 210:
		case 268:	// add monster-only silent
		case 269:	// as 268
			return true;
		default:
			return false;
	}
}

// v220-22 If checks are OK reposition the TPMAN just off line
static RDBOOL repositionTPMAN(mobj_t *m, line_t *ld, int type)
{
#include "m_bams.h"
	binangle_t angle = bamsAtan2(
		ld->dy >> FRACBITS, 
		ld->dx >> FRACBITS) >> (ANGLETOFINESHIFT - FRACBITS);
	fixed_t dx = FixedMul(FRACUNIT, finesine  [angle]);
	fixed_t dy = FixedMul(FRACUNIT, finecosine[angle]);
	fixed_t	x = m->x, y = m->y;
	subsector_t	*ss;

	// --------------------------

	if (type) { x += dx; y += dy; } else { x -= dx; y -= dy; } 

	ss = R_PointInSubsector(x, y);

	if (   ( type && ld->backsector  == ss->sector)
		|| (!type && ld->frontsector == ss->sector) )
	{
		m->subsector = ss;
		m->x = x, m->y = y;
		return true;
	}

	return false;
}

// v220-5 rationalise MT_TELEPORTMAN thinkers. They only provide position
//        and angle info, don't think and waste resources. They are dealt
//        with following a call to LS_CheckSpawnedbySkill in case some weird
//        map has not set them to easy, medium and hard. This must also be
//        called following analysis and tpthings are also used to check that
//        TP.s are correctly linked 

// v220-22 revised to cope with bad map design
static void Set_TPThings(RDBOOL removeThing)
{
	thinker_t	*thinker;
	mobj_t		*m;

	if (settp.tpthings) W_Free(settp.tpthings, STATICID, 0);

	memset(&settp, 0, sizeof(settp));

	for (thinker = thinkercap.next; thinker!= &thinkercap; thinker = thinker->next)
	{
		if (thinker->function == (void (__cdecl *)(void))P_MobjThinker &&
			((mobj_t *) thinker)->type == MT_TELEPORTMAN)
		{
			settp.num_tpthings++;
		}
	}

	if (settp.num_tpthings)
	{
		mobj_t **rem = (mobj_t**)malloc(settp.num_tpthings * sizeof(mobj_t*));
		int    i, lnum;

		settp.tpthings = (tpthing_t*)W_Malloc(settp.num_tpthings * sizeof(tpthing_t), STATICID, 57);
		settp.num_tpthings = 0;

		for (thinker = thinkercap.next; thinker != &thinkercap; thinker = thinker->next)
		{
			if (   thinker->function == (void (__cdecl *)(void))P_MobjThinker
				&& (m = (mobj_t *) thinker)->type == MT_TELEPORTMAN)
			{
				subsector_t	*ss = m->subsector;
				line_t		*ld;
				seg_t		*sg;
				RDBOOL		valid = false;
				line_t		line;
				int			tag = ss->sector->tag, secn = ss->sector - sectors;

				// check that the tag is valid
				if (tag)
				{
					line.tag = tag;

					for (i = -1; (i = P_FindLineFromLineTag(&line, i)) >= 0;)
					{
						if (tpSpecial(lines[i].special))
						{
							valid = true;
							break;
						}
					}

					if (!valid) tag = 0;
				}

				if (   !tag
					&& (lnum = FindIfThingIsOnLine(m, false)) >= 0
					&& lines[lnum].backsector
					&& lines[lnum].frontsector != lines[lnum].backsector)
				{
					// Yes folks! Daft map making actually happens. A TELEPORTMAN placed
					// bang on a line means we cannot trust that the  correct  subsector
					// and hence sector has been returned by P_PointInSubsector(). Since
					// it may not be tagged (this does happen) checks have to be run.

					int		tag1 = lines[lnum].frontsector->tag;
					int		tag2 = lines[lnum].backsector->tag;

					valid = false;

					// If the line has a tag then check that another line tagged to the
					// sector has a tp special. Note that only one tag should be set else
					// something else has gone wrong as the initial tag should either be
					// set or was deemed invalid

					if (tag1 && tag1 != m->subsector->sector->tag)
					{
						line.tag = tag1;

						for (i = -1; (i = P_FindLineFromLineTag(&line, i)) >= 0;)
						{
							if (tpSpecial(lines[i].special))
							{
								if (repositionTPMAN(m, &lines[lnum], 0))
								{
									valid = true;
									tag = tag1;
									secn = lines[lnum].frontsector - sectors;
									ss = m->subsector;
								}
								break;
							}
						}
					}
					else if (tag2 && tag2 != m->subsector->sector->tag)
					{
						line.tag = tag2;

						for (i = -1; (i = P_FindLineFromLineTag(&line, i)) >= 0;)
						{
							if (tpSpecial(lines[i].special))
							{
								if (repositionTPMAN(m, &lines[lnum], 1))
								{
									valid = true;
									tag = tag2;
									secn = lines[lnum].backsector - sectors;
									ss = m->subsector;
								}
								break;
							}
						}
					}

					if (valid)
					{
						pr_OFinsert("Map error with badly positioned tpman T%i fixed\n", MAPNUM(m));
						ma_inf.maperror++;
					}
				}

				// v220-22 check the tp subsec for exit lines
				settp.tpthings[settp.num_tpthings].exit = 0;
				sg = &segs[ss->firstline];

				for (i = 0; i < ss->linecount; i++, sg++)
				{
					if (   (ld = sg->linedef)
						&& (ld->special == 52 || ld->special == 124))
					{
						settp.tpthings[settp.num_tpthings].exit = 18;	// delay factor
						break;
					}
				}

				settp.tpthings[settp.num_tpthings].used = false;
				settp.tpthings[settp.num_tpthings].mnum = MAPNUM(m);
				settp.tpthings[settp.num_tpthings].x = m->x;
				settp.tpthings[settp.num_tpthings].y = m->y;
				settp.tpthings[settp.num_tpthings].an = m->angle;
				settp.tpthings[settp.num_tpthings].tag  = tag;
				settp.tpthings[settp.num_tpthings].secn = secn;

				rem[settp.num_tpthings++] = m;
			}
		}

		if (removeThing)
		{
			for (i = 0; i < settp.num_tpthings; i++)
				P_RemoveMobj(rem[i], 34);
		}

		M_Free(rem);
	}
}

static void setSecLimits(int idx, fixed_t *box)
{
	sector_t *s = &sectors[idx];
	int		 i;

	for (i = 0; i < s->linecount; i++)
	{
		if (!(s->lines[i]->spflags & SPF_CONTROLLINE))
			M_AddLineToBox(box, s->lines[i]);
	}
}

// v220-5 revised
void TP_CheckTPSpawnedSectors(RDBOOL removeThing)
{
	line_t		*ld;
	int			*lnchk = NULL;
	int			lncount = 0;
	int			maxlncount = 0;
	int			*lnerr_check = NULL;
	int			lnerr_count = 0;
	int			maxlnerr_count = 0;
	int			*err_check;
	int			err_count = 0;
	int			i, k, s;
	RDBOOL		changed = false;

	//------------------------------

	// PASS 1
	Set_TPThings(removeThing);

	if (!settp.num_tpthings) return;

	// PASS 2
	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->special && tpSpecial(ld->special))
		{
			RDBOOL	found = false;
			int		srefcount = 0;

			for (s = -1; (s = P_FindSectorFromTag(ld->tag, s)) >= 0;)
			{
				srefcount++;

				for (k = 0; k < settp.num_tpthings; k++)
				{
					if (settp.tpthings[k].secn == s)
					{
						settp.tpthings[k].used = true;
						found = true;
					}
				}
			}

			if (ld->ml_flags & ML_TWOSIDED)
			{
				if (!found)
				{
					if (srefcount)
					{
						if (lncount >= maxlncount)
						{
							maxlncount += 16;
							lnchk = (int*)M_Realloc(lnchk, maxlncount * sizeof(int), 46);
						}

						lnchk[lncount++] = i;
					}
					else
					{
						if (lnerr_count >= maxlnerr_count)
						{
							maxlnerr_count += 16;
							lnerr_check = (int*)M_Realloc(lnerr_check, maxlnerr_count * sizeof(int), 47);
						}

						lnerr_check[lnerr_count++] = i;
					}
				}
			}
		}
	}

	if (!lncount && !lnerr_check)
		return;	// All lines tag a sector with a teleport platform

	err_check = (int*)M_Calloc(settp.num_tpthings * sizeof(int), 0);

	// PASS 3
	for (i = 0; i < lncount; i++)
	{
		ld = &lines[lnchk[i]];

		// Check things to see if they are contained in a referenced sector
		for (s = -1; (s = P_FindSectorFromTag(ld->tag, s)) >= 0;)
		{
			for (k = 0; k < settp.num_tpthings; k++)
			{
				if (!settp.tpthings[k].used)
				{
					fixed_t box[4];
					fixed_t x = settp.tpthings[k].x;
					fixed_t y = settp.tpthings[k].y;

					M_ClearBox(box);
					setSecLimits(s, box);

					// See if it fits in s
					if (   x > box[BOXLEFT ]
						&& x < box[BOXRIGHT]
						&& y > box[BOXBASE ]
						&& y < box[BOXTOP  ])
					{
						settp.tpthings[k].secn = s;
						settp.tpthings[k].used = true;
						settp.tpthings[k].tag = ld->tag;
						err_check[k] = 0x80000000;
						changed = true;
						break;
					}
				}
			}
		}
	}

	for (i = 0; i < settp.num_tpthings; i++)
	{
		if (!settp.tpthings[i].used)
		{
			int t = settp.tpthings[i].tag;

			if (t)
			{
				for (k = 0; k < numlines; k++)
				{
					if (lines[k].tag == t && tpSpecial(lines[k].special))
						goto settp_ok;
				}

				err_check[err_count++] |= i;
			}
settp_ok:;
		}
	}

	if (removeThing)
	{
		if (lnerr_count || err_count || changed)
		{
			Con_Message("TELEPORTMAN CHECKS ====\n");

			if (changed)
			{
				for (i = 0; i < settp.num_tpthings; i++)
				{
					if (err_check[i] & 0x80000000)
						Con_Message("  NODES FIX: unlinked TP T%i linked to tag %i\n",
							settp.tpthings[i].mnum, settp.tpthings[i].tag);
				}
			}

			for (i = 0; i < lnerr_count; i++)
			{
				ma_inf.maperror++;
				if (R3D_edit || ge_inf.showtptagerrors)
					Con_Message("  MAP ERROR  L%i's tag %i is unreferenced\n", lnerr_check[i], lines[lnerr_check[i]].tag);
			}

			for (i = 0; i < err_count; i++)
			{
				int n = err_check[i] & ~0x80000000;

				Con_Message("  MAP ERROR  T%i S%i references a missing tag %i\n",
						settp.tpthings[n].mnum,
						settp.tpthings[n].secn,
						settp.tpthings[n].tag);

				ma_inf.maperror++;
			}

			Con_Message("END TELEPORTMAN ====\n");
		}
	}

	M_Free(err_check);
	M_Free(lnerr_check);
	M_Free(lnchk);
}

//================== END TP_CheckTPSpawnedSectors

// v220-4 add ability to allow player only with scripted maps
static RDBOOL TP_Verify(line_t *line, mobj_t *thing)
{
	if (thing->flags & MF_MISSILE) return false;

	if (   line->scpFlags & SCPF_PLAYERONLY
		&& (!thing->dplayer
		||  thing->dplayer->mo != thing))
	{
		return false;
	}

	return true;
}

// v220-5
static RDBOOL setTPLanding(int secidx)
{
	int i;

	for (i = 0; i < settp.num_tpthings; i++)
	{
		if (settp.tpthings[i].secn == secidx)
		{ 
			settp.posx = settp.tpthings[i].x;
			settp.posy = settp.tpthings[i].y,
			settp.an = settp.tpthings[i].an;
			settp.exit = settp.tpthings[i].exit;
			return true;
		}
	}

	return false;
}

int getTPgnum(int secidx)
{
	int i;

	for (i = 0; i < settp.num_tpthings; i++)
	{
		if (settp.tpthings[i].secn == secidx)
		{
			return R_PointInSubsector(settp.tpthings[i].x, settp.tpthings[i].y)->viewgroup;
		}
	}
	return -1;
}

//
// TELEPORTATION
//
// killough 5/3/98: reformatted, cleaned up

int EV_Teleport(line_t *line, mobj_t *thing)
{
	int       n;
	
	// don't teleport missiles
	// Don't teleport if hit back of line,
	//  so you can get out of teleporter.
	if (!TP_Verify(line, thing)) return 0;
	
	// killough 1/31/98: improve performance by using
	// P_FindSectorFromTag instead of simple linear search.
	
	for (n = -1; (n = P_FindSectorFromTag(line->tag, n)) >= 0;)
	{
		fixed_t	oldx = thing->x, oldy = thing->y, oldz = thing->z;

		if (!setTPLanding(n)) continue;	// v220-5 use tpthings list.

		if (!P_TeleportMove(thing, settp.posx, settp.posy, 0, false, false))
			continue; // if blocked see if there's another tagged sector
		
		thing->z = thing->dropoffz = thing->floorz;

		P_SetMpos(thing);	// v220-5

		if (!R3D_edit)
		{
			// spawn teleport fog at source
			mobj_t *fog = P_SpawnMobj(oldx, oldy, oldz, MT_TFOG, 0);
			P_SetMpos(fog);

			// emit sound at source (if recording the sound will not get saved)
			S_LocalSound(sfx_telept, fog);

			
			// spawn teleport fog and emit sound at destination
			fog = P_SpawnMobj(thing->x + 20*finecosine[settp.an >> ANGLETOFINESHIFT],
							  thing->y + 20*finesine[settp.an >> ANGLETOFINESHIFT],
							  thing->z, MT_TFOG, 0);

			P_SetMpos(fog);
			S_StartSound(sfx_telept, fog);
		}
		
		thing->angle = settp.an;
		thing->momx = thing->momy = thing->momz = 0;
		
		if (thing->dplayer && thing->dplayer->mo == thing)
		{
			thing->dplayer->viewz = thing->z + thing->dplayer->viewheight;
			thing->dplayer->clAngle = settp.an;
			thing->dplayer->clLookDir = 0;
			thing->dplayer->lookdir = 0;
			thing->dplayer->flags |= DDPF_FIXANGLES;
			ma_inf.playerteleport = settp.exit;
			// if player then don't move for a bit
			thing->reactiontime = 18;
		}

		return 1;
	}

	return 0;
}

//
// Silent TELEPORTATION, by Lee Killough
// Primarily for rooms-over-rooms etc.
//

int EV_SilentTeleport(line_t *line, mobj_t *thing)
{
	if (!TP_Verify(line, thing)) return 0;

	for (int n = -1; (n = P_FindSectorFromTag(line->tag, n)) >= 0;)
	{
		ddplayer_t	*player;
		angle_t		angle;
		fixed_t		c, s, z, momx, momy;

		if (!setTPLanding(n)) continue;	// v220-5 use tpthings list.

		// Height of thing above ground, in case of mid-air teleports:
		z = thing->z - thing->floorz;
		
		// Get the angle between the exit thing and source linedef.
		// Rotate 90 degrees, so that walking perpendicularly across
		// teleporter linedef causes thing to exit in the direction
		// indicated by the exit thing.
		angle =
			R_PointToAngle2(0, 0, line->dx, line->dy) - settp.an + ANG90;
		
		// Sine, cosine of angle adjustment
		s = finesine[angle>>ANGLETOFINESHIFT];
		c = finecosine[angle>>ANGLETOFINESHIFT];
		
		// Whether this is a player, and if so, its pointer to ddplayer_t
		player = thing->dplayer && thing->dplayer->mo == thing ?
			thing->dplayer : NULL;
		
		// Momentum of thing crossing teleporter linedef
		momx = thing->momx;
		momy = thing->momy;
		
		// Attempt to teleport
		if (!P_TeleportMove(thing, settp.posx, settp.posy, 0, false, false))
			continue; // if blocked see if there's another tagged sector
		
		// v220-13 support dehacked spawnshot
		if (thing->type == MT_SPAWNSHOT && thing->flags & MF_SPAWNCEILING)
		{
			sector_t *sc = thing->subsector->sector;

			// ceiling spawned landing
			thing->z = thing->floorz = thing->dropoffz = sc->ceilingheight - thing->height - 32*FRACUNIT;
			thing->momz = 0;
		}
		else
		{
			// Adjust z position to be same height above ground as before
			thing->z = z + (thing->dropoffz = thing->floorz);
			gdt.pt_oldz = thing->z;
		}
		
		// Rotate thing according to difference in angles
		thing->angle += angle;
		
		P_ClearThingSRVO(thing, true);	// v220-5
		P_SetMpos(thing);	// v220-5

		if (thing->flags2 & MF2_TELEPCORPSE)
		{
			thing->momx = thing->momy = 0;
			thing->flags &= ~MF_DROPOFF;
		}
		else
		{
			// Rotate thing's momentum to come out of exit just like it entered
			thing->momx = FixedMul(momx, c) - FixedMul(momy, s);
			thing->momy = FixedMul(momy, c) + FixedMul(momx, s);
		}
		
		// Adjust player's view, in case there has been a height change
		// Voodoo dolls are excluded by making sure dplayer->mo == thing.
		if (player)
		{
			// Save the current deltaviewheight, used in stepping
			fixed_t deltaviewheight = player->deltaviewheight;

			// Copy angle
			player->clAngle = thing->angle;

			// Clear deltaviewheight, since we don't want any changes
			player->deltaviewheight = 0;
			
			// Set player's view according to the newly set parameters
			P_CalcHeight(thing->player);
			
			// Reset the delta to have the same dynamics as before
			player->deltaviewheight = deltaviewheight;
			ma_inf.playerteleport = settp.exit;
		}

		return 1;
	}
	return 0;
}

//
// Silent linedef-based TELEPORTATION, by Lee Killough
// Primarily for rooms-over-rooms etc.
// This is the complete player-preserving kind of teleporter.
// It has advantages over the teleporter with thing exits.
//

// Also used for 3D line teleport although not silent
// v220-35 revised

int EV_SilentLineTeleport(line_t *line, mobj_t *thing,
						  RDBOOL reverse)
{
	static line_t **t = NULL;
	static maxn = 0;
	// -----------------------------------
	
	if (!TP_Verify(line, thing)) return 0;

	// v220-31 Unfortunately there are cases that hang by a thread and
	//         work by sheer chance. These can be dealt with by sifting
	//         through various cases.

	line_t	*L;
	
	// First check for maps where the tags match and there is only one
	// which is two sided with no special. This is what we would expect.
	for (int n = 0, i = -1; (i = P_FindLineFromLineTag(line, i)) >= 0;)
	{
		if ( (L = lines + i) != line && L->backsector)
		{
			if (!L->special) goto linefound;

			if (maxn <= n)
			{
				maxn += 32;
				t = (line_t**)W_Realloc(t, maxn * sizeof(line_t*), STATICID, 0);
			}

			t[n++] = L;
		}
	}

	// The other expected case is where there are only two lines with the same tag
	if (n == 1) { L = t[0]; goto linefound; }

	if (!n) return 0;	// No matching tagged lines found

	// v220-31 We are now mired in difficulties.
	//         With sinedie map02 then various lines with the same tag come into
	//         play and have specials set when keys are supposed to tp to L5508
	//         (which should not have a special set) and it's a wonder it works
	//         at all (even with this fix to narrow the field down). This has been
	//         made redundant in the case of sinedie through having an entry in
	//         R3D_WADanalyse to clear the special on L5508 but has been left just
	//         in case some other errant map pops up.

	for (i = 0; i < n; i++)
	{
		L = t[i];

		if (    L->special == 243
			||  L->special == 244
			||  (L->special >= 262 && L->special <= 267) )
		{
			goto linefound;
		}
	}

	L = t[0];	// If all else fails use the first line found

linefound:;
	
	// Save current position for fog spawn at source
	// if a 3D line teleport (these are not silent)
	fixed_t oldx = thing->x, oldy = thing->y, oldz = thing->z;
	fixed_t pos = 0, nposx = 0, nposy = 0;	// position and offsets from line

	// v220-35 by using int64 for greater accuracy it was found that
	//         the previous correction code was no longer needed.
	__int64 res = (__int64)line->dx*line->dx + (__int64)line->dy*line->dy;

	if (res)
	{
		__int64 num = (__int64)(thing->x-line->v1->x)*line->dx + 
					  (__int64)(thing->y-line->v1->y)*line->dy;

		// convert back to fixed_t
		if (num <= 0)        { /* the high F's will make no difference */ }
		else if (num >= res) { pos = 1<<30; }
		else                 { pos = (fixed_t)(num / (res>>30)); }

		nposx = thing->x - line->v1->x - fixed_t(((__int64)line->dx*pos)>>30);
		nposy = thing->y - line->v1->y - fixed_t(((__int64)line->dy*pos)>>30);
	}
	
	// Get the angle between the two linedefs, for rotating
	// orientation and momentum. Rotate 180 degrees, and flip
	// the position across the exit linedef, if reversed.
	angle_t angle =
		R_PointToAngle2(0, 0, L->dx, L->dy) -
		R_PointToAngle2(0, 0, line->dx, line->dy);

	if (!reverse)
	{
		angle += ANG180;
		pos = (1<<30) - pos;
	}

	// Sine, cosine of angle adjustment
	fixed_t s = finesine[angle>>ANGLETOFINESHIFT];
	fixed_t c = finecosine[angle>>ANGLETOFINESHIFT];
	
	// Rotate position along normal to match exit linedef
	fixed_t x = (fixed_t)(((__int64)nposx*c - (__int64)nposy*s)>>16);
	fixed_t y = (fixed_t)(((__int64)nposy*c - (__int64)nposx*s)>>16);

	// Interpolate position across the exit linedef
	x += L->v1->x + int(((__int64)pos*L->dx)>>30);
	y += L->v1->y + int(((__int64)pos*L->dy)>>30);

	// Whether this is a player, and if so, a pointer to its player_t.
	// Voodoo dolls are excluded by making sure thing->dplayer->mo==thing.
	ddplayer_t *player = thing->dplayer && thing->dplayer->mo == thing ?
		thing->dplayer : NULL;

	if (L->specType >= R3D_40)
	{
		// 3D lines treated differently - also sound played and fog spawned
		mobj_t	 *smo;
		unsigned an = thing->angle>>ANGLETOFINESHIFT;
		fixed_t	z = L->backsector->floorheight +
				L->d3d->offset3D + L->d3d->totalheight;

		// Attempt to teleport, aborting if blocked
		if (!P_TeleportMove(thing, x, y, z, true, false))
			return 0;

		smo = P_SpawnMobj(oldx, oldy, oldz, MT_TFOG, 0);
		P_SetMpos(smo);

		S_StartSound(sfx_telept,smo);

		thing->floorz = thing->dropoffz = thing->z = z;
		thing->momx = thing->momy = 0;
		thing->reactiontime = 18;

		smo = P_SpawnMobj(x + 20*finecosine[an],
						  y + 20*finesine[an], thing->z, MT_TFOG, 0);
		P_SetMpos(smo);

		S_StartSound(sfx_telept, smo);
	}
	else
	{		
		// Whether walking towards first side of exit linedef steps down
		int stepdown =
			L->frontsector->floorheight < L->backsector->floorheight;
	
		// Height of thing above ground
		fixed_t z = thing->z - thing->floorz;
		
		// Attempt to teleport, aborting if blocked
		if (!P_TeleportMove(thing, x, y, 0, false, false))
			return 0;

		// Adjust z position to be same height above ground as before.
		// Ground level at the exit is measured as the higher of the
		// two floor heights at the exit linedef.
		thing->z = z + (thing->floorz = sides[L->sidenum[stepdown]].sector->floorheight);
		thing->dropoffz = thing->floorz;
	}

	// Rotate thing's orientation according to difference in linedef angles
	thing->angle += angle;
	gdt.pt_oldz = thing->z;

	if (thing->flags2 & MF2_TELEPCORPSE)
	{
		thing->momx = thing->momy = 0;
		thing->flags &= ~MF_DROPOFF;
	}
	else
	{
		// Momentum of thing crossing teleporter linedef
		x = thing->momx;
		y = thing->momy;
		
		// Rotate thing's momentum to come out of exit just like it entered
		thing->momx = (fixed_t)(((__int64)x*c - (__int64)y*s)>>16);
		thing->momy = (fixed_t)(((__int64)y*c + (__int64)x*s)>>16);
	}
	
	P_SetMpos(thing);	// v220-5

	// Adjust a player's view, in case there has been a height change
	if (player)
	{
		// Save the current deltaviewheight, used in stepping
		fixed_t deltaviewheight = player->deltaviewheight;
		
		// Copy angle
		player->clAngle = thing->angle;

		// Clear deltaviewheight, since we don't want any changes now
		player->deltaviewheight = 0;
		
		// Set player's view according to the newly set parameters
		P_CalcHeight(thing->player);
		
		// Reset the delta to have the same dynamics as before
		player->deltaviewheight = deltaviewheight;
		ma_inf.playerteleport = settp.exit;
	}

	return 1;
}
