//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: x_hair.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//
// Based on: Doomsday v1.7.8 JDoom.dll code
//
// DESCRIPTION: draws and deals with selection of crosshair
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdlib.h>

#include "r3d_compat.h"

#include "doom/x_hair.h"

#include "con_console.h"
#include "con_functions.h"
#include "gl_main.h"
#include "r_main.h"

// OGL
#include "r3d_OGLfunc.h"

// MACROS ------------------------------------------------------------------

#define MAX_XLINES	16

// TYPES -------------------------------------------------------------------

typedef struct
{
	int				x, y;		// In window coordinates (*not* 320x200!).
} crosspoint_t;

typedef struct
{
	crosspoint_t	a, b;
} crossline_t;

typedef struct
{
	int				numLines;
	crossline_t		lines[MAX_XLINES];
} cross_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

#define XL(x1,y1,x2,y2) {{x1,y1},{x2,y2}}

cross_t crosshairs[NUM_XHAIRS] = 
{
	// cross +
	{ 4, {	XL(-5, 0,-2, 0), XL( 0,-5, 0,-2),
			XL( 5, 0, 2, 0), XL( 0, 5, 0, 2) }
	},
	// ><
	{ 4, {	XL(-7,-5,-2, 0), XL(-7, 5,-2, 0),
			XL( 7,-5, 2, 0), XL( 7, 5, 2, 0) }
	},
	// square (filled)
	{ 4, {	XL(-3,-3,-3, 3), XL(-3, 3, 3, 3),
			XL( 3, 3, 3,-3), XL( 3,-3,-3,-3) }
	},
	// square (unfilled)
	{ 8, {	XL(-4,-4,-4,-2), XL(-4, 2,-4, 4),
			XL(-4, 4,-2, 4), XL( 2, 4, 4, 4),
			XL( 4, 4, 4, 2), XL( 4,-2, 4,-4),
			XL( 4,-4, 2,-4), XL(-2,-4,-4,-4) }
	},
	// diamond
	{ 4, {	XL( 0,-3, 3, 0), XL( 3, 0, 0, 3),
			XL( 0, 3,-3, 0), XL(-3, 0, 0,-3) }
	},
	// caret ^
	{ 2, {	XL(-4,-4, 0, 0), XL( 0, 0, 4,-4) }
	}
};

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

// Draw a crosshair in the centre of the client area.
// NOTE: cfg.xhairType should be set prior to calling
void X_Drawer(byte *xhairColor)
{
	cross_t	*cross;
	float	size    = (float)(cfg.xhairSize + 1) * 0.5f;
	float	centerX = (float) current_w  * 0.5f;
	float	centerY = (float) current_h * 0.5f;
	int		i, num, diff = SCREENHEIGHT - viewSCRheight;

	if (ge_inf.noxhair) return;

	if (diff)
	{
		centerY -= (float)(diff/2)*((float)current_h / (float)SCREENHEIGHT);
	}

	if ( (num = cfg.xhairType - 1) < 0) num = 0;
	if (num >= NUM_XHAIRS) num = NUM_XHAIRS - 1;
	cross = &crosshairs[num];

	glDisable(GL_TEXTURE_2D);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	// We need the 1:1 coordinates.
	glOrtho(0, current_w, current_h, 0, -1, 1);

	glColor3ubv(xhairColor);
	glBegin(GL_LINES);

	for( i = 0; i < cross->numLines; i++)
	{
		crossline_t *xline = cross->lines + i;
		glVertex2f(size*xline->a.x + centerX, size*xline->a.y + centerY);
		glVertex2f(size*xline->b.x + centerX, size*xline->b.y + centerY);
	}

	glEnd();

	// Restore the matrices
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glEnable(GL_TEXTURE_2D);
}

int CCmdCrosshair(int argc, char **argv)
{
	if (inZoomMode)
	{
		Con_Printf( "Unselectable when in zoom mode\n");
		return true;
	}
	if(argc == 1)
	{
		Con_Printf( "Usage:\n");
		Con_Printf( " crosshair (num)             : 0 no crosshair else 1 to %d\n", NUM_XHAIRS);
		Con_Printf( " crosshair size (size)       : 1 for normal size\n");
		Con_Printf( " crosshair color (r) (g) (b) : rgb range 0 to 255\n\n");
		Con_Printf( "Current values: num %d, size %d, color %d, %d, %d\n",
			cfg.xhairType, cfg.xhairSize, cfg.xhairColor[0], cfg.xhairColor[1], cfg.xhairColor[2]);
		Con_Printf("NOTE: This does not set the ZOOM xhair, this type is fixed\n");
		return true;
	}
	else if(argc == 2) // Type
	{
		cfg.xhairType = strtol(argv[1], NULL, 0);

		if (cfg.xhairType > NUM_XHAIRS) cfg.xhairType = NUM_XHAIRS;
		if (cfg.xhairType < 0) cfg.xhairType = 0;

		Con_Printf( "Crosshair %d selected.\n", cfg.xhairType);
	}
	else if(argc == 3) // Size.
	{
		if(_stricmp(argv[1], "size")) return false;
		cfg.xhairSize = strtol(argv[2], NULL, 0);
		if(cfg.xhairSize < 0) cfg.xhairSize = 0;
		Con_Printf( "Crosshair size set to %d.\n", cfg.xhairSize);
	}
	else if(argc == 5) // Color
	{
		int i;

		if(_stricmp(argv[1], "color")) return false;

		for(i=0; i<3; i++) 
		{
			cfg.xhairColor[i] = (byte)(strtol(argv[2+i], NULL, 0));
			// Clamp.
			if(cfg.xhairColor[i] < 0) cfg.xhairColor[i] = 0;
			if(cfg.xhairColor[i] > 255) cfg.xhairColor[i] = 255;
		}

		Con_Printf( "Crosshair color set to (%d, %d, %d).\n", 
			cfg.xhairColor[0], cfg.xhairColor[1], cfg.xhairColor[2]);
	}
	else
		return false;

	return true;
}