//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2012 G Jackson, Jaakko Kernen, Jack Vermeulen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: con_bar.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//  Extensively modified for Risen3D
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
//
// Based on: Doomsday v1.7.8
//
// DESCRIPTION:  console loading bar
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include "de_base.h"
#include "de_console.h"
#include "de_graphics.h"
#include "de_ui.h"

#include "sys_stwin.h"
#include "sys_timer.h"
#include "sys_system.h"
#include "r3d_compat.h"

#include "r3d_OGLfunc.h"


// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

#define PROGRESS_TITLE_LENGTH  63         // v220-13 max message length

static char		progress_title[PROGRESS_TITLE_LENGTH + 1];  // v220-13
static int		progress_max;
static int		progress, progress_shown;

static int      oldpercent;		// progress compare  to see if display should be done
static float	stepstarttime;	// automatic timing per step   for displays that have an szid
static float	steptotaltime;	// automatic timing cumulative for displays that have an szid 
static float	print_time;		// for deducting time taken to draw the con bar

static float	timestart;		// start time

// CODE --------------------------------------------------------------------

// I just named these Con for now

//==========================================================================
// Con_TimingStart - generic timer
//                   sets the start time of a section to be timed
//==========================================================================
void Con_TimingStart(void)
{
    timestart =  Sys_GetSeconds();	// set the start time
	print_time = 0;					// zero the print_time accumulator
}

//==========================================================================
// Con_TimingEnd - generic timer
//                 return the elapsed time in double seconds up to now
//                 shows the data on the console if szId is set
//==========================================================================
float Con_TimingEnd(char* szid)
{
    float elapsed = Sys_GetSeconds() - timestart; // elapsed time till now
    
    return elapsed - print_time;	// deduct time taken to print the con bar
}


//==========================================================================
// Con_InitProgress
//
// resets the comparisons for Con_Progress
//==========================================================================

void Con_InitProgress(const char *title, int full)
{
	if (ge_inf.loadlevel || !gl_inited) return;	// v220-4
	
    oldpercent = 0;       // reset comparison percent

	strncpy(progress_title, title, PROGRESS_TITLE_LENGTH);
	progress_title[PROGRESS_TITLE_LENGTH] = 0;	// v220-13
	
	if (full >= 0)				// some safety
	{
		progress       = 0;
        progress_shown = 0;
		if(!full) full = 1;		// need at least 1

		progress_max = full;	// limit to compare to for a percent bar
	}

	stepstarttime = Sys_GetSeconds(); // set starting time
    steptotaltime = 0;                // clear total elapsed step time

	Con_Progress(0, 0);
}

//===========================================================================
// Con_HideProgress
//===========================================================================
void Con_HideProgress(void)
{
	if (!gl_inited) return;
	
	Con_InitProgress("WAIT...", 1);
}

//==========================================================================
// 
//  Con_Progress(int count, int flags)
//
//	Draws a progress bar. Flags control what's to be done. 
//  All "count" data is relative to a percent of a preset maximum set by init
//
//   count : the current value to control the percent displayed
//
//     1. For a percent bar it represents the current item of a known maximum number of items
//     2. For a cumulative count its the amount to increment the total 
//    
//   flags: -- docs only defined in con_bar.h
//
//#define PBARF_PERCENT	1	// jav do progress using count as a percent of max (default is incremental counting)
//
//  ** GENERAL FLOW **
//
// 1. Con_InitProgress is called first.  
//    This resets the time and also sets the maximum expected
//
// 2. Con_Progress is called next
//
//    a. For displays that do not know exactly where they are, use the incremental "count" version (as before)
//       It best if the maximum is 100 and then just split the various Con_Progess(countvalues ...) up so they
//       more or less accurately reflect the percent done.  I use the timing giving to give me a fairly accurate
//       approximation.
//  
//    b. For displays that know exactly how much they have to do, use the percent flag.  Then call
//       Con_Progress(count...) in the loop where count is the current indexed value being processed.
//       It skips 5 percent at a time so this does not to overload the vid display.
//
// 3. Con_HideProgress is called at the end
//
// ***********************
// * Percent example: 
// ***********************
//
//	Con_InitProgress("R_InitSprites..." , numspritelumps);     // set maximum 
//
//	for(i = 0, sl = spritelumps; i < numspritelumps; i++, sl++)
//	{
// 		Con_Progress(i, PBARF_PERCENT); // percent display
//        ... code..
//	}
//
//	Con_HideProgress();
//
//
// ***********************
// * Count example:  
// ***********************
//	Con_InitProgress("Analyse..." , 100);     // set maximum at 100
//
//  .. some code...
// 	Con_Progress(10, 0);	// percent is now at 10
//  .. some code...
// 	Con_Progress( 5, 0);	// percent is now at 15
//  .. some code...
// 	Con_Progress(50, 0);	// percent is now at 65
//  .. some code...
// 	Con_Progress(45, 0);	// percent is now at 100
//
// 	Con_Progress( 0, 0);	// just to get the total time since each time is for PREVIOUS step
//							// you can also use this to time parts without changing the bar
//	Con_HideProgress();
//
//
//==========================================================================

void Con_Progress(int count, int flags)
{
	static char	szProgress[8];

	float	drawtime;
	int		percent;
	int		x, y, w, h, bor = 2, bar = 10;
	int		fonthgt;

	// --------------------------------------

	if (ge_inf.loadlevel || !gl_inited) return;		// v220-4 don't draw loading bar
	
	if(flags & PBARF_PERCENT)						// this is a percent display
	{
		percent = (count*100)/progress_max;
		if (percent <= oldpercent) return;			// don't do anything
		
		progress = count;							// this will be percent bar below
		oldpercent = percent + 5;					// force 5 percent jumps otherwise it gets slow	
	}
	else											// this is a incremental count display
	{ 
		progress += count;							// increment progress by count 
	}

	sprintf(szProgress," %d",progress);
	
	// jav this is a cumulative elapsed time and an actual elapsed time
	// for now verbose only
	
	// use Con_TimingStart() and Con_TimingEnd() to get elapsed time for a block of stuff - gmj
	
	// Make sure it's in range.
	
	if(progress < 0) progress = 0;
	if(progress > progress_max) progress = progress_max;
	
	progress_shown = progress;

	drawtime = Sys_GetSeconds();

	// This'll redraw the startup screen to this page 
	// (necessary if page flipping is used by the display adapter).
	GL_DrawPagePatch(titlepic_num);

	FR_SetFont(0);
	fonthgt = FR_TextHeight("A");	// jav this is actual percent bar

	// Go into screen projection mode.
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	// 1-to-1 mapping for the whole window.
	glOrtho(0, current_w, current_h, 0, -1, 1);
		
	// Calculate the size and dimensions of the progress window.
	w = current_w - 30;
	if(w < 50 ) w = 50;			// An unusual occurrance...
	if(w > 610) w = 610;		// Restrict width to the case of 640x480.
	
	x = (current_w - w) / 2;	// Center on screen.
	h = 2*bor + fonthgt + 15 + bar;
	y = current_h - 15 - h;
	
	// Draw the (opaque black) shadow.
	
	UI_Gradient(x + 4, y + 4, w, h, UI_COL(UIC_SHADOW), 0, 1, 1);
	
	// Background
	
	UI_DrawRect(x, y, w, h, bor, UI_COL(UIC_BRD_HI), UI_COL(UIC_BRD_MED), UI_COL(UIC_BRD_LOW));
	
	x += bor;
	y += bor;
	w -= 2*bor;
	h -= 2*bor;
	UI_Gradient(x, y, w, h, UI_COL(UIC_BG_MEDIUM), UI_COL(UIC_BG_LIGHT), 1, 1);
	
	// Title.
	x += 5;
	y += 5;
	w -= 10;
	
	glColor3f(.0, 0, 0.3f);	//R3Dcolours - shadow, draw first
	FR_TextOut(progress_title, x+2, y+2);
	
	glColor3f(.9f, .9f, .0f);		//followed by text
	FR_TextOut(progress_title, x, y);
	
	y += fonthgt + 5;
	
	// Bar
	
	UI_Gradient(x, y, w, bar, UI_COL(UIC_SHADOW), 0, .5f, .5f);
	UI_DrawRect(x+1, y+1, 
		8 + (w-8) *progress/progress_max - 2, bar - 2, 4,
		UI_COL(UIC_TEXT), UI_COL(UIC_BG_DARK), UI_COL(UIC_BG_LIGHT));
		
	// Restore old projection matrix.
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	GL_DoUpdate();
	
	print_time += Sys_GetSeconds() - drawtime;	// R3D	
}

//========================== END ==============================================