//-----------------------------------------------------------------------------
//
// Copyright (C) 2005 Jack Vermeulen, Jaakko Kernen
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: con_console.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  modified for use with R3D
//  jav add statics
//  jav rearrange order
//  jav some equates for console buffer size
//  gmj remove dedicated
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
//  vid-arrays added
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 24/Apr/2016 gmj
//  "berserk-bug" and "game-corpsetime" removed
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:  Doom console
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <math.h>
#include <process.h>
#include <direct.h>
#include <errno.h>

#include "de_base.h"
#include "de_console.h"
#include "de_system.h"
#include "de_network.h"
#include "de_graphics.h"
#include "de_refresh.h"
#include "de_render.h"
#include "de_play.h"
#include "de_audio.h"
#include "de_ui.h"
#include "de_misc.h"

#include "r3d_compat.h"
#include "r3d_mapinfo.h"

#include "r3d_OGLfunc.h"

// MACROS ------------------------------------------------------------------

#define SC_EMPTY_QUOTE		-1

#define MAXBUFFERLINES 512
#define BUFFSIZE 4095
#define SUBCSIZE  255			// b211-0 explicitly limit subcmd size

// TYPES -------------------------------------------------------------------

typedef struct
{
	char *name;
	char *command;
} calias_t;

typedef struct
{
	uint	marker;		// The tic on which the cmd should be executed.
	uint	used;		// Is this in use?
	char	*subcmd;	// A single command w/args.
} execbuff_t;

typedef struct
{
	char word[64];			// 64 chars MAX for words.
} knownword_t;

// Console command.

typedef struct
{
	char		*name;
	int			(*func)(int argc, char **argv);
	char		*help;		// pointer to help text.
} ccmd_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void Sfx_Doppler(void);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

static void Con_AddCommandList(ccmd_t *cmdlist);
static void Con_ClearBuffer(void);
static void conPrintf(int flags);
static void SplitIntoSubCommands(char *command, int markerOffset);
static void ClearExecBuffer(void);
static int  executeSubCmd(char *subcmd);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern cvar_t inputCVars[];

extern int		bufferLines;
extern float	mus_compress;
extern int		panel_32;
extern int		playerdrag;
extern int		zdistptr;

extern int		rendSharp;
extern int		mididev;
extern uint		numMidiDev;
extern int		swappartvis;		// v220-31

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int		vidShowFPS = false;
int		vidVsync   = 1;
int		con_error;

// v220-3
RDBOOL	water_trans;

int		water_opacity = 170;	// default
int		set_light_defaults = 0;

float	ConsoleY; // Where the console bottom is currently?
float	maxLightDistance = 3072.0f;	// default
int		console_flat;
int	    consoleDump = true;

char	configFileName[256];

RDBOOL	con_freezegame;			// v220-28
RDBOOL	con_active;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static ddfont_t Cfont;				     // The console font.

static float	CcolYellow[3] = { 1, .85f, .3f };

static float	ConsoleOpenY;			// Where the console bottom is when open.
static int		consoleTurn;			// The rotation variable.

static byte	    conSilentCVars   = 1;		
static int		consoleLight     = 50;  // amount to show texture
static int      consoleAlpha     = 75;  // transparency
static int		conCompMode      = 0;	// Completion mode.
static int		consoleActiveKey = 96;	// '`' translated from 0x29
static int	    consoleShowKeys  = false;
static RDBOOL	ConsoleSilent    = false;

//static byte	    consoleShadowText= true;  // R3D now also used in con_start.c 

static cvar_t	*cvars = NULL;
static int		numCVars = 0;

static ccmd_t	*ccmds = NULL;			// The list of console commands.
static int		numCCmds = 0;			// Number of console commands.

static calias_t *caliases = NULL;
static int		numCAliases = 0;

static knownword_t *knownWords = 0;	// The list of known words (for completion).
static int numKnownWords = 0;

static char     trimmed_float_buffer[32];	// Returned by TrimmedFloat().

static float ConsoleDestY;		// Where the console bottom should be?
static int   ConsoleTime;		// How many ticks has the console been open?

static float funnyAng;
static RDBOOL openingOrClosing = true;

static float fontFx, fontSy;	// Font x factor and y size.

static cbline_t *cbuffer;		// This is the buffer.

int bufferLines;				// How many lines are there in the buffer?

//static int maxBufferLines;	// jav Maximum number of lines in the buffer.
static int maxLineLen = 100;	// Maximum length of a line (default - adjusted later). jav was 70
static int bPos;				// Where the write cursor is? (which line)
static int bFirst;				// The first visible line.
static int bLineOff;			// How many lines from bPos? (+vislines)

static char cmdLine[2048];		// The command line.
static cbline_t *oldCmds;		// The old commands buffer.
static int numOldCmds;
static int ocPos;				// Old commands buffer position.
static int complPos;			// Where is the completion cursor?
static int lastCompletion;		// The index of the last completion (in knownWords).

static execbuff_t *exBuff;
static int        exBuffSize;

static execbuff_t *curExec;

static char prbuff[BUFFSIZE + 1];


// CODE --------------------------------------------------------------------

#pragma optimize("g", off)

#define DEFCC(name)		int name(int argc, char **argv)

// List of all (ccmds) commands linked to text aliases.
// This can only be checked manually - the compiler will
// not help if anything's wrong.
//
DEFCC(CCmdActiveMobjs);
DEFCC(CCmdAddSub);
DEFCC(CCmdAlias);
DEFCC(CCmdAutoMap);
DEFCC(CCmdBackgroundTurn);
DEFCC(CCmdBind);
DEFCC(CCmdCheat);
DEFCC(CCmdCheatGive);
DEFCC(CCmdCheatGod);
DEFCC(CCmdCheatMassacre);
DEFCC(CCmdCheatKill666);

DEFCC(CCmdCheatNoClip);
DEFCC(CCmdCheatReveal);
DEFCC(CCmdCheatWarp);
DEFCC(CCmdClearBindings);
DEFCC(CCmdConsole);
DEFCC(CCmdCrosshair);
DEFCC(CCmdDeleteBind);
DEFCC(CCmdDump);
DEFCC(CCmdDynIntensity);
DEFCC(CCmdEcho);

DEFCC(CCmdEdcam);
DEFCC(CCmdEditWarc);
DEFCC(CCmdEditWarl);
DEFCC(CCmdEditWars);
DEFCC(CCmdEditWarss);
DEFCC(CCmdFly);
DEFCC(CCmdFog);
DEFCC(CCmdHudCycle);
DEFCC(CCmdKillSelf);
DEFCC(CCmdListActs);

DEFCC(CCmdListAliases);
DEFCC(CCmdListBindings);
DEFCC(CCmdListCmds);
DEFCC(CCmdListFiles);
DEFCC(CCmdListVars);
DEFCC(CCmdMenuAction);
DEFCC(CCmdMsgRefresh);
DEFCC(CCmdOpenPanel);
DEFCC(CCmdParse);

DEFCC(CCmdPause);
DEFCC(CCmdPlayMusic);
DEFCC(CCmdPlaySound);
DEFCC(CCmdPrintPlayerCoords);
DEFCC(CCmdQuit);
DEFCC(CCmdRebuildMidi);
DEFCC(CCmdRepeat);
DEFCC(CCmdResetTextures);
DEFCC(CCmdScreenShot);

DEFCC(CCmdSetCamera);
DEFCC(CCmdSkyBox);
DEFCC(CCmdStopInFine);
DEFCC(CCmdStopMusic);
DEFCC(CCmdToggleList);
DEFCC(CCmdTorch);
DEFCC(CCmdWalkRun);
DEFCC(CCmdVersion);

DEFCC(CCmdWait);
DEFCC(CCmdZoom);
DEFCC(CCmdSwapRenderingOrder);

// Console commands. Names in LOWER CASE (yeah, that's cOnsISteNt).
static ccmd_t engineCCmds[] =
{
	"automap",		CCmdAutoMap,		"Show automap.",
	"actions",		CCmdListActs,		"List all action commands.",
	"after",		CCmdWait,			"Execute the specified command after a delay.",
	"alias",		CCmdAlias,			"Create aliases for a (set of) console commands.",
	"bind",			CCmdBind,			"Bind a console command to an event.",
	"bindr",		CCmdBind,			"Bind a console command to an event (keys with repeat).",
	"bgturn",		CCmdBackgroundTurn, "Set console background rotation speed.",
	"clear",		CCmdConsole,		"Clear the console buffer.",
	"clearbinds",	CCmdClearBindings,	"Deletes all existing bindings.",
	"crosshair",	CCmdCrosshair,		"Crosshair setup.",
	"dump",			CCmdDump,			"Dump a data lump currently loaded in memory.",
	"fog",			CCmdFog,			"Modify fog settings.",
	"help",			CCmdConsole,		"Show information about the console.",
	"listaliases",	CCmdListAliases,	"List all aliases and their expanded forms.",
	"listbindings",	CCmdListBindings,	"List all event bindings.",
	"listcmds",		CCmdListCmds,		"List all console commands.",
	"listfiles",	CCmdListFiles,		"List all the loaded data files and show information about them.",
	"listvars",		CCmdListVars,		"List all console variables and their values.",
	"msgrefresh",	CCmdMsgRefresh,		"Show last HUD message.",
	"panel",		CCmdOpenPanel,		"Open the Risen3D Control Panel.",
	"pause",		CCmdPause,			"Pause the game.",
	"playmusic",	CCmdPlayMusic,		"Set/reset current music or play external music file",
	"pm",			CCmdPlayMusic,		"Set/reset current music or play external music file",
	"playsound",	CCmdPlaySound,		"Play a lumpname sound.",
	"quit!",		CCmdQuit,			"Exit the game immediately.",
	"texreset",		CCmdResetTextures,	"Force a texture reload.",
	"midirebuild",	CCmdRebuildMidi,	"Rebuild current midifile.",
	"togglelist",	CCmdToggleList,		"Toggle Play List.",
	"version",		CCmdVersion,		"Show detailed version information.",

	"safebind",		CCmdBind,			"Bind a command to an event, unless the event is already bound.",
	"safebindr",	CCmdBind,			"Bind a command to an event, unless the event is already bound.",
	"sbcycle",		CCmdHudCycle,		"Status bar type.",
	"screenshot",	CCmdScreenShot,		"Takes a screenshot. Saved to DOOMnn.TGA.",
	"skybox",		CCmdSkyBox,			"Set a skybox by name. jpg or tga types are searched - jpg takes precedence",
	"stopmusic",	CCmdStopMusic,		"Stop any currently playing music.",

	"add",			CCmdAddSub,			"Add something to a cvar.",
	"dec",			CCmdAddSub,			"Subtract 1 from a cvar.",
	"inc",			CCmdAddSub,			"Add 1 to a cvar.",
	"sub",			CCmdAddSub,			"Subtract something from a cvar.",
	"echo",			CCmdEcho,			"Echo the parameters on separate lines.",

	// Cheats
	"cheat",		CCmdCheat,			"Issue a cheat code using the original Doom cheats.",
	"give",			CCmdCheatGive,		"Gives you weapons, ammo, power-ups, etc.",
	"god",			CCmdCheatGod,		"God mode.",
	"kill",			CCmdCheatMassacre,	"Kill all the monsters on the level.",
	"noclip",		CCmdCheatNoClip,	"No movement clipping (walk through walls).",
	"reveal",		CCmdCheatReveal,	"Map cheat.",
	"warp",			CCmdCheatWarp,		"Warp to another map.",

	// Edit functions
	"warss",		CCmdEditWarss,		"Warp to subsector - Edit mode.",
	"wars",			CCmdEditWars,		"Warp to sector - Edit mode.",
	"warl",			CCmdEditWarl,		"Warp to line - Edit mode.",
	"warc",			CCmdEditWarc,		"Warp to co-ords - Edit mode.",

	// Menu Actions
	"helpscreen",	CCmdMenuAction,		"Show the Help screens.",
	"savegame",		CCmdMenuAction,		"Open the save game menu.",
	"loadgame",		CCmdMenuAction,		"Open the load game menu.",
	"soundmenu",	CCmdMenuAction,		"Open the sound settings menu.",
	"quicksave",	CCmdMenuAction,		"Quicksave the game.",
	"endgame",		CCmdMenuAction,		"End the game.",
	"togglemsgs",	CCmdMenuAction,		"Messages on/off.",
	"quickload",	CCmdMenuAction,		"Load the quicksaved game.",
	"quit",			CCmdMenuAction,		"Quit the game",
	"togglegamma",	CCmdMenuAction,		"Switch to console gamma screen",
	"zoom",			CCmdZoom,			"Toggle zoom mode",

	"stopfinale",	CCmdStopInFine,		"Stop the currently playing interlude/finale.",

	"coord",		CCmdPrintPlayerCoords, "Print the coordinates of the player.",

	// $democam: console commands
	"makecam",		CCmdSetCamera,		"Toggle camera mode.",

	// R3D
	"edcam",		CCmdEdcam,			"Set edit camera screen ON-OFF.",
	"fly",			CCmdFly,			"Set flying ON-OFF.",
	"dintensity",	CCmdDynIntensity,	"Cycles through 0.05 to 1: 0.05 = min 0, 1.0 = max",
	"mj",			CCmdActiveMobjs,	"Show number of active shootable mobjs.",
	"killself",		CCmdKillSelf,		"Commit suicide",
	"kill666",		CCmdCheatKill666,	"Activate 666/667 stuff when using nomonsters",
	"svp",			CCmdSwapRenderingOrder, "Swap rendering order of vissprites and particles",
	"torch",		CCmdTorch,			"Toggle using torch light (if allowed)",
	"walkrun",		CCmdWalkRun,		"Toggle walk-run",
	NULL
};

// Cvars - provide access to global variables via the console.
static cvar_t engineCVars[] = 
{
//===========================================================================
// The new, consistent names
//===========================================================================
	// Scripted map
	"scp-txtdelay",			0,			CVT_FLOAT,	&cfg.text_delay,
																	1.0f, 3.0f,	"Delay to text fade with scripted maps [1.0 to 3.0]",	// v220-4
	"scp-usemapmusic",		0,			CVT_INT,	&cfg.usemapmusic,	0, 1,	"1=Always start scripted map with own music",		// v220-4

	// Console
	"con-alpha",			0,			CVT_INT,	&consoleAlpha,	0, 100,		"Console background opacity",
	"con-light",			0,			CVT_INT,	&consoleLight,	0, 100,		"Console background light level",
	"con-completion",		0,			CVT_INT,	&conCompMode,	0, 1,		"How to complete words when pressing Tab:\n0=Show completions, 1=Cycle through them",
	"con-dump",				0,			CVT_INT,	&consoleDump,	0, 1,		"1=Dump all console messages to Risen3D.out",
	"con-key-activate",		0,			CVT_INT,	&consoleActiveKey,
																	0, 255,		"Key to activate the console (default is the key above TAB - shows as 96)",
	"con-key-show",			CVF_NO_ARCHIVE,CVT_INT,	&consoleShowKeys,
																	0, 1,		"1=Show ASCII codes of pressed keys in the console",
	"con-var-silent",		0,			CVT_BYTE,	&conSilentCVars,0, 1,		"1=Don't show the value of a cvar when setting it",

	// Video
	// v220-3 bpp must be CVT_INT
	"vid-bpp",				CVF_NO_ARCHIVE|CVF_HIDE,CVT_INT,&panel_32,0,1,		"bpp 0=16 1=32",

	"vid-fps",				0,			CVT_INT,	&vidShowFPS,	0,  1,		"1=Show FPS counter on screen",
	"vid-gamma",			0,			CVT_FLOAT,	&vid_gamma,		0.5f, 1.5f,	"Display gamma (1 - optimum)",
	"vid-contrast",			0,			CVT_FLOAT,	&vid_contrast,	0.5f, 1.5f,	"Display contrast (1 - optimum)",
	"vid-bright",			0,			CVT_FLOAT,	&vid_bright,	-.5f, .7f,	"Display brightness (0 - optimum)",
	"vid-cap",				CVF_INT_PANEL, CVT_INT,	&cfg.vid_cap,	4, 11,		"Video cap 4 to 11 (4 = 100fps to 11 = 330fps)",
	"vid-compress",			0,			CVT_INT,	&cfg.lightCompress,
																	0, 140,		"Compress lighting range None=0 Max=140",
    "vid-vsync",            CVF_CPBUTTON, CVT_INT,	&vidVsync,		0,  1,		"sync frame rate to: 0 vid-cap, 1 monitor",	// vid_cap
    "vid-aspect",           CVF_NO_ARCHIVE|CVF_HIDE,
										CVT_INT,	&cfg.use43,		0,  1,		"",

	// v220-3 set distance for darkening
	"vid-distance-darken",	0,			CVT_FLOAT,	&maxLightDistance,
																	1024, 8192,	"Darken with distance min 1024 max 8192",
	// v220-5 flash light levels
	"flash-damage",			0,			CVT_FLOAT,&cfg.flashdamage,	0,  1.0f,	"flash damage intensity 0.0 to 1.0",
	"flash-pickup",			0,			CVT_FLOAT,&cfg.flashpickup,	0,  1.0f,	"flash pickup intensity 0.0 to 1.0",

	// Render
	"rend-splash",			0,			CVT_INT, &cfg.splashsprite,	0,	1,		"1=Enable splash sprites (footsteps must be enabled)",
	"rend-light-ambient",	CVF_NO_ARCHIVE,	CVT_INT,&r_ambient,		0, 255,		"Ambient light level: ded map entry defined - can only be changed in the console in edit mode",
	"rend-light",			0,			CVT_INT,	&useDynLights,	0, 1,		"1=Render dynamic lights",
	"rend-light-torch",		0,			CVT_INT,	&cfg.useTorchLight,0, 1,	"0=Use vertex lighting 1=Use torch light",
	"rend-light-sine",		0,			CVT_INT,	&cfg.sineCompress, 0, 1,	"0=Use linear lighting 1=Use non-linear lighting",
	"rend-light-vtx",		0,			CVT_INT,	&cfg.setVtxLight,  0, 1,	"0=Even vertex lighting 1=Distance vertex lighting",
	"rend-light-default",	CVF_NO_ARCHIVE|CVF_HIDE,
										CVT_INT,	&set_light_defaults,  0, 1,	"",
	"rend-light-bright",	0,			CVT_FLOAT,	&dlFactor,		0.5f, 0.9f,	"Intensity factor for dynamic lights. min 0.5, max 0.9, default 0.7",
	"rend-addlolight",		CVF_NO_ARCHIVE,	CVT_INT,&ma_inf.addlolight,0, 1,	"1=Increase low light",

	// * Render-Glow
	"rend-glow",			0,			CVT_INT,	&r_texglow,		0, 1,		"1=Enable glowing textures",
	"rend-bloodslime",		0,			CVT_INT,	&r_bloodglow,	0, 1,		"1=Enable glowing blood and slime (enable glow must be on)",
	"rend-glow-wall",		0,			CVT_INT,	&useWallGlow,	0, 1,		"1=Render glow on walls",

	// * Render-fov
	"rend-base-fov",		0,			CVT_INT,	&init_FOV,		1, 179,		"Field of view - this should only be adjusted in the console.",

	// * Render-Texture
	// $translucent_water
	"rend-water-trans",		0,			CVT_INT,	&water_trans,	0, 1,		"Translucent water 0-OFF 1-ON",
	"rend-opacity",			0,			CVT_INT,	&water_opacity,	150, 255,	"Degree of water opacity (default) 170 min=150 max=255",

	"rend-sharp",			0,			CVT_INT,	&rendSharp,		0, 1,		"Sprites draw: 0-Blurred edge as in Doomsday 1=Sharp edge",
	"rend-tex-detail",		CVF_NO_ARCHIVE,	CVT_INT,&r_detail,		0, 1,		"1=Render with detail textures",
	"rend-tex-detail-scale", CVF_NO_MIN|CVF_NO_MAX, CVT_FLOAT, &detailScale,
																	0, 0,		"Global detail texture factor",
	"rend-tex-detail-strength", 0,		CVT_FLOAT,	&detailFactor, 0, 10,		"Global detail texture strength factor",
	"rend-tex-detail-far",	CVF_NO_MAX, CVT_FLOAT,	&detailMaxDist,	1, 0,		"Maximum distance where detail textures are visible",

	// * Render-Sky
	"rend-skybox",			CVF_NO_ARCHIVE,	CVT_INT,&r_skybox,		0, 1,		"1=Render skybox",

	// * Render-Sprite
	"rend-sprite-align-angle", 0,		CVT_FLOAT,	&maxSpriteAngle, 0, 90,		"Maximum angle for slanted sprites (spralign 2)",

	// * Render-Model
	"rend-anisotropic" , CVF_PROTECTED,	CVT_INT,	&anisotropic,	0, 16, "Level of anisotropic filtering (1 to 16)",
	"rend-model-lights", CVF_SCP_FIXED,	CVT_INT,	&modelLights,	0, MAX_MODEL_LIGHTS, "Maximum number of light sources on models",
	"rend-model-aspect", CVF_NO_MAX|CVF_NO_MIN, CVT_FLOAT, &rModelAspectMod,
																	0, 0,		"Scale for MD2 z-axis when model is loaded",
	"rend-model-distance",	0,	CVT_INT,	&zdistptr,				0, MAXZLISTPTR,		"Distance when a model is swapped for a sprite; 0 smallest to 4 Never",

	// * Rend-Particle
	// Note: "rend-particle-max" and "rend-particle-visible-near" removed v220-4
	"rend-particle-rate",	0,			CVT_FLOAT,	&r_particle_spawn_rate,
																	0.5f, 2,	"Particle spawn rate multiplier (default: 1)",
	"rend-particle-diffuse", CVF_NO_MAX, CVT_FLOAT,	&rend_particle_diffuse,
																	1, 4,		"Diffuse factor for particles near the camera",
	"rend-smoke",			0,			CVT_INT,	&cfg.weapsmoke,0, 1,		"1=Render weapon firing smoke",

	// * Rend-Shadow
	"rend-shadow",			0,			CVT_INT,	&cfg.useShadows,0, 1,		"1=Render shadows under objects",
	"rend-shadow-darkness",	0,			CVT_FLOAT,	&shadowFactor,	0, 1,		"Darkness factor for object shadows",
	"rend-shadow-far",		CVF_NO_MAX,	CVT_INT,	&shadowMaxDist,	0, 0,		"Maximum distance where shadows are visible",
	"rend-shadow-radius-max", CVF_NO_MAX, CVT_INT,	&shadowMaxRad,	0, 0,		"Maximum radius of object shadows",
	"aggressive-flush",		0,			CVT_INT,	&cfg.glflush,	0, 1,		"0=Normal flush (fastest performance) 1=Use aggressive flush (e.g. if using an NVidia GTX1080)",

	// Sound
	"sound-info",			0,			CVT_BYTE,	&sound_info,	0, 1,		"1=Show sound debug information",
	"sound-reverb-volume",	0,			CVT_FLOAT,	&sfx_reverb_strength,
																	0, 1,		"Reverb effects general volume (0=disable)",
	"audio-ricoh",			0,			CVT_INT,	&cfg.ricochet,	0, 9,		"0- No richochet 1 to 9 degree of randomness",
	"audio-pitch",			0,			CVT_INT,	&cfg.pitch,		0, 1,		"0- disable 1- enable: random pitch change with flagged sounds",
	"audio-doppler",		CVF_DOPPLER,CVT_FLOAT,	&cfg.doppler,	0, 0.2f,	"0- none to 0.2f - default: doppler pitch change with 3D sounds",	// v220-35 upper limit was 1.0
	"sound-footsteps",		0,			CVT_INT,	&cfg.footsteps,	0, 1,		"1=Footstep sounds (has no effect in BDM mode)",
	"sound-weapon",			0,			CVT_INT,	&cfg.weaponsounds,	0, 1,	"1=Enable weapon ready sounds (has no effect in BDM mode)",

	// Music 
	"mus-compress"	,		0,			CVT_FLOAT,	&mus_compress,	0.5f, 2.0f, "Compress MUS (0.5 to 2) 1=no change, 2=double volume",
	"midi-dev",				0,			CVT_INT,	&mididev,		0, 10,		"Midi dev number (use  Control panel/Audio)",

	"vol-sfx"	,			0,			CVT_INT,	&sfx_volume,	1, 255,		"Sound effects volume (1-255)",
	"vol-ext",				0,			CVT_INT,	&ext_volume,	1, 255,		"Music volume (0-255)",
	"vol-step",				0,		  CVT_FLOAT,	&cfg.footstepvol,0.1f, 1.f,	"Footstep volume (0.1-1.0)",
	"vol-splash",			0,		  CVT_FLOAT,	&cfg.splashvol,	0.1f, 1.f,	"Splash volume (0.1-1.0)",
	"vol-midi",				0,			CVT_INT,	&mid_volume,	1, 255,		"Midi volume (0-255)",

	"pos_slider",			CVF_NO_ARCHIVE,	CVT_INT,&pos_slider,	1, 255,		"",

	// Input
	// * Input-Key
	"player-drag",			0,			CVT_INT,	&playerdrag,	0, 10,		"Player drag when stopping - 0 (default) to 10 (max)",
	"input-key-wait1",		0,			CVT_INT,	&repWait1,		5, 35,		"The number of 35 Hz ticks to wait before first key repeat (5 - 35)",
	"input-key-wait2",		0,			CVT_INT,	&repWait2,		2, 10,		"The number of 35 Hz ticks to wait between key repeats (1 - 10)",
	"input-key-show-scancodes", CVF_NO_ARCHIVE,	CVT_INT,&showScanCodes,
																	0, 1,		"1=Show scancodes of all pressed keys in the console",

	// * Input-Joy
	"input-joy-sensi",		0,		CVT_INT,	&cfg.joySensitivity,0,  9,		"Joystick sensitivity [0 - 9]",
	"input-joy-deadzone",	0,		CVT_INT,	&cfg.joyDeadzone,	0, 80,		"Joystick dead zone as a percentage [0 - 80]", 

	// * Input-Mouse
	"input-mouse-wheel-sensi",	CVF_NO_MAX,	CVT_INT,	&mouseWheelSensi,	0,  0, "Mouse wheel sensitivity",
	"input-mouse-x-disable",	0,			CVT_INT,	&mouseDisableX,		0,  1, "1=Disable mouse X axis",
	"input-mouse-y-disable",	0,			CVT_INT,	&mouseDisableY,		0,  1, "1=Disable mouse Y axis",
	"input-mouse-y-inverse",	0,			CVT_INT,	&mouseInverseY,		0,  1, "1=Inversed mouse Y axis",
	"input-mouse-x-sensi",		CVF_NO_MAX,	CVT_INT,	&cfg.mouseSensiX,	0, 25, "Mouse X axis sensitivity",
	"input-mouse-y-sensi",		CVF_NO_MAX,	CVT_INT,	&cfg.mouseSensiY,	0, 25, "Mouse Y axis sensitivity",

	"ctl-look-joy-inverse",	0,			CVT_INT,	&cfg.jlookInverseY, 0, 1,	"1=Inverse joystick look Y axis",
	"ctl-look-mouse-inverse", 0,		CVT_INT,	&cfg.mlookInverseY, 0, 1,	"1=Inverse mouse look Y axis",
	"ctl-look-joy-delta",	0,			CVT_INT,	&cfg.jlookDeltaMode, 0, 1, "1=Joystick values => look angle delta",

	"input-joy-enable",	0,			CVT_INT,	&cfg.joyEnable,		0, 1,	"Enable joystick 0=No 1=Yes",
	"input-joy-x",		0,			CVT_INT,	&cfg.joyaxis[0],	0, 4,	"X axis control: 0=None, 1=Move, 2=Turn, 3=Strafe, 4=Look",
	"input-joy-y",		0,			CVT_INT,	&cfg.joyaxis[1],	0, 4,	"Y axis control",
	"input-joy-z",		0,			CVT_INT,	&cfg.joyaxis[2],	0, 4,	"Z axis control",
	"input-joy-rx",		0,			CVT_INT,	&cfg.joyaxis[3],	0, 4,	"X rotational axis control",
	"input-joy-ry",		0,			CVT_INT,	&cfg.joyaxis[4],	0, 4,	"Y rotational axis control",
	"input-joy-rz",		0,			CVT_INT,	&cfg.joyaxis[5],	0, 4,	"Z rotational axis control",
	"input-joy-slider1", 0,			CVT_INT,	&cfg.joyaxis[6],	0, 4,	"First slider control",
	"input-joy-slider2", 0,			CVT_INT,	&cfg.joyaxis[7],	0, 4,	"Second slider control",
	"msg-echo",			0,			CVT_INT,	&cfg.echoMsg,		0, 1,	"1=Echo all messages to the console",
	"ctl-look-speed",	0,			CVT_INT,	&cfg.lookSpeed,		1, 5,	"The speed of looking up/down",
	"ctl-use-dclick",	0,			CVT_INT,	&cfg.dclickuse,		0, 1,	"1=Doubleclick forward/strafe equals use key",
	"ctl-look-pov",		0,			CVT_INT,	&cfg.povLookAround,	0, 1,	"1=Look around using the POV hat",
	"ctl-look-mouse",	0,			CVT_INT,	&cfg.usemlook,		0, 1,	"1=Mouse look active",
	"ctl-look-joy",		0,			CVT_INT,	&cfg.usejlook,		0, 1,	"1=Joystick look active",
	"ctl-run",			0,			CVT_INT,	&cfg.alwaysRun,		0, 1,	"1=Always run",
	"ctl-look-spring",	0,			CVT_INT,	&cfg.lookSpring,	0, 1,	"1=Lookspring active",

	// v220-4 Fixed if a scripted map
	"damage-falling",	CVF_SCP_FIXED,	CVT_INT,	&cfg.fallingDamage,	0, 1,	"1=Falling damage enabled",
	"damage-underwater",CVF_SCP_FIXED,	CVT_INT,	&cfg.underwaterDamage,0,1,	"1=Underwater damage enabled",
	"hud-size",			CVF_SCP_FIXED,	CVT_INT,	&cfg.sbarscale,		1, 5,	"Status bar options (1-5)",

	"map-alpha",		0,			CVT_FLOAT,	&cfg.automapAlpha,	-0.1f, 1.f,	"Alpha level of the automap background -0.1 to 1.0",
	"map-rotate",		0,			CVT_INT,	&cfg.automapRotate, 0, 1,	"1=Automap turns with player",
	"map-showdead",		0,			CVT_INT,	&cfg.show_dead	  , 0, 1,	"1=Show dead things in the automap",
	"map-cheat-counter",0,			CVT_INT,	&cfg.counterCheat,	0, 1,	"1=show kills/items/secret in the automap",
	"map-line-width",	0,			CVT_FLOAT,	&cfg.linesz,		1.f, 3.f,	"automap line width 1.0 to 3.0",
#if 0		// $bulletrange
	"map-bullet-range" ,CVF_HIDE,	CVT_INT,	&cfg.bulletRange,	0, 2,	"Bullet range 0 = Doom, 1 = Doom*2, 2 = Doom*4",
#endif

	"menu-flash-r",		0,			CVT_FLOAT,	&cfg.flashcolor[0],			0, 1,	"Menu selection flash color, red component",
	"menu-flash-g",		0,			CVT_FLOAT,	&cfg.flashcolor[1],			0, 1,	"Menu selection flash color, green component",
	"menu-flash-b",		0,			CVT_FLOAT,	&cfg.flashcolor[2],			0, 1,	"Menu selection flash color, blue component",
	"menu-flash-speed",	0,			CVT_INT,	&cfg.flashspeed,			0, 50,	"Menu selection flash speed",
	"menu-turningskull",0,			CVT_INT,	&cfg.turningSkull,			0, 1,	"1=Menu skull turns at slider items",
	"hud-health",		0,			CVT_INT,	&cfg.hudShown[HUD_HEALTH],	0, 1,	"1=Show health in HUD",
	"hud-armor",		0,			CVT_INT,	&cfg.hudShown[HUD_ARMOR],	0, 1,	"1=Show armor in HUD",
	"hud-ammo",			0,			CVT_INT,	&cfg.hudShown[HUD_AMMO],	0, 1,	"1=Show ammo in HUD",
	"hud-keys",			0,			CVT_INT,	&cfg.hudShown[HUD_KEYS],	0, 1,	"1=Show keys in HUD",
	"hud-frags",		0,			CVT_INT,	&cfg.hudShown[HUD_FRAGS],	0, 1,	"1=Show deathmatch frags in HUD",
	"hud-scale",		0,			CVT_FLOAT,	&cfg.hudScale,				0.3f, 0.8f, "Scaling for HUD info",
	"hud-color-r",		0,			CVT_FLOAT,	&cfg.hudColor[0],			0, 1,	"HUD info color",
	"hud-color-g",		0,			CVT_FLOAT,	&cfg.hudColor[1],			0, 1,	"HUD info color",
	"hud-color-b",		0,			CVT_FLOAT,	&cfg.hudColor[2],			0, 1,	"HUD info color",
	"hud-frags-all",	0,			CVT_BYTE,	&hu_showallfrags,			0, 1,	"Debug: HUD shows all frags of all players",

	"view-cross-type", CVF_NO_MAX|CVF_PROTECTED, CVT_INT, &cfg.xhairType,	0, 0,	"The current crosshair type",
	"view-cross-r",		0,			CVT_BYTE,	&cfg.xhairColor[0],			0, 255,	"Red crosshair color component",
	"view-cross-g",		0,			CVT_BYTE,	&cfg.xhairColor[1],			0, 255, "Green crosshair color component",
	"view-cross-b",		0,			CVT_BYTE,	&cfg.xhairColor[2],			0, 255, "Blue crosshair color component",
	"view-cross-size", CVF_NO_MAX,	CVT_INT,	&cfg.xhairSize,				0, 1,	"Crosshair size: 1=Normal",
	"view-bob-height",	0,			CVT_FLOAT,	&cfg.bobView,				0, 1,	"Scale for viewheight bobbing",
	"view-bob-weapon",	0,			CVT_FLOAT,	&cfg.bobWeapon,				0, 1,	"Scale for player weapon bobbing",

	"msg-show",			0,			CVT_INT,	&showMessages,   0, 1,  "1=Show messages",

	// Player data.
	"player-jump",			CVF_SCP_FIXED,		CVT_INT,	&cfg.jumpEnabled,	0, 1,	"1=Allow jumping",
	"player-jump-power",	CVF_SCP_FIXED,		CVT_FLOAT,	&cfg.jumpPower,		0, 10,	"Jump power",
	"player-autoswitch",	0,		CVT_INT,	&cfg.weaponAutoSwitch, 0, 1, "1=Change weapon automatically when picking one up",
	"player-camera-noclip", 0,		CVT_INT,	&cfg.cameraNoClip,	0, 1,	"1=Camera players have no movement clipping",
	"player-waterdynamics", 0,		CVT_INT,	&waterdynamics,		0, 1,	"1=water dynamics on  0-off. Inhibited in edit mode",
	
	"msg-secret",		0,			CVT_INT,	&cfg.secretMsg,		0, 1,	"1=Announce the discovery of secret areas",

	"menu-scale",		0,			CVT_FLOAT,	&cfg.menuScale,		.1f, 1,	"Scaling for menus.",
	"menu-effect",		0,			CVT_INT,	&cfg.menuEffects,	0, 2,	"Disable menu effects: 1=type-in, 2=all",
	"hud-title",		0,			CVT_INT,	&cfg.levelTitle,	0, 1,	"1=Show level title and author in the beginning",
	"menu-color-r", 	0,			CVT_FLOAT,	&cfg.menuColor[0],	0, 1,	"Menu color red component",
	"menu-color-g",		0,			CVT_FLOAT,	&cfg.menuColor[1],	0, 1,	"Menu color green component",
	"menu-color-b",		0,			CVT_FLOAT,	&cfg.menuColor[2],	0, 1,	"Menu color blue component",
	"menu-fog",			0,			CVT_INT,	&cfg.menuFog,		0, 1,	"Menu fog mode: 0=blue vertical, 1=black smoke",

	"msg-count",		0,			CVT_INT,	&cfg.msgCount,		0, 8,	"Number of HUD messages displayed at the same time",
	"msg-scale",		CVF_NO_MAX,	CVT_FLOAT,	&cfg.msgScale,		0, 0,	"Scaling factor for HUD messages",
	"msg-uptime",		CVF_NO_MAX,	CVT_INT,	&cfg.msgUptime,		35, 0,	"Number of tics to keep HUD messages on screen",
	"msg-blink",		0,			CVT_INT,	&cfg.msgBlink,		0, 1,	"1=HUD messages blink when they're printed",
	
	"lc",				0,			CVT_INT,	&cfg.lightCompress,	0, 140,	"Set light compression 0 - 140",
	"fakeglow",			0,			CVT_INT,	&cfg.fakeglow,		0, 1,	"Enable underwater glow. 1 = ON, 0 = OFF",
	"inFight",			0,			CVT_INT,	&cfg.inFight,		0, 1,	"Enable infighting. 1 = ON, 0 = OFF",
	// v220-33
	"set-zbias",		0,			CVT_FLOAT,	&cfg.zbias,			0.0001f, 0.0022f, "Set zbias between 0.0001 and 0.0022",
	"amdevwidth",		0,			CVT_INT,	&cfg.amDevboxWidth,	120, 180,	"Devmap am box width: 120 to 180",
	"ameditwidth",		0,			CVT_INT,	&cfg.amEditboxWidth,	160, 220,	"Edit am box width: 160 to 220",
	"odpm",				0,			CVT_INT,	&cfg.olddoommessage,	0, 1,	"1 = unless scripted use original doom player message display",
	NULL
};

//===========================================================================
// Con_Init
//===========================================================================
void Con_Init()
{
	con_active = false;
	con_freezegame = false;
	ConsoleY     = 0;
	ConsoleOpenY = 90;
	ConsoleDestY = 0;
	ConsoleTime  = 0;

	funnyAng = 0;

	// Font size in VGA coordinates. (Everything is in VGA coords.)
	fontFx = 1;
	fontSy = 9;

	// The buffer.
	cbuffer        = NULL;
	bufferLines    = 0;	
	//maxBufferLines = 1024;     // jav was 256; made into a define

	// The old commands buffer.
	oldCmds    = 0;
	numOldCmds = 0;
	ocPos      = 0;				// No commands yet.

	bPos     = 0;
	bFirst   = 0;
	bLineOff = 0;

	complPos = 0;
	lastCompletion = -1;

	exBuff     = NULL;
	exBuffSize = 0;

	// Register the engine commands and variables.

	Con_AddCommandList(engineCCmds);
	Con_AddVariableList(engineCVars);
	Con_AddVariableList(inputCVars);
}

//=============================================================================
//
//=============================================================================
void Con_Open(int yes)
{
	// Clear all action keys, keyup events won't go 
	// to bindings processing when the console is open.
	Con_ClearActions();

	openingOrClosing = true;

	if (yes)
	{
		con_active = true;
		con_freezegame = !shiftDown;
		ConsoleDestY = ConsoleOpenY;
		ConsoleTime = 0;
	}
	else
	{
		strcpy(cmdLine, "");
		con_active = false;
		con_freezegame = false;
		ConsoleDestY = 0;
	}
}

//==========================================================================
// Con_Message
//	Print to Risen3D.out
//  Print to the console only if appStatus == 2
//==========================================================================
void Con_Message(char *message, ...)
{
	if(message[0])
	{
		va_list argptr;

		va_start(argptr, message);
		_vsnprintf(prbuff, BUFFSIZE, message, argptr);
		va_end(argptr);
			
		prbuff[BUFFSIZE] = 0;  
		
#ifdef _DEBUG
		pr_OFdirect(prbuff);
		return;
#endif

		// dump messages to RISEN3D.OUT file if consoleDump is set
		// or appStatus indicates the engine is initialising
        if (appStatus < 2)
		{
			// Nothing is printed to the console at this time
			pr_OFdirect(prbuff);
			return;
		}

		// dump messages to RISEN3D.OUT file since Con_Printf will not if consoleDump is set
        if(!consoleDump) pr_OFdirect(prbuff);

		// Also print in the console.
		Con_Printf("%s", prbuff);
	}
}

//=============================================================================
// Print into the buffer during initialisation
// Information printed to the console is kept to the bare minimum.
//=============================================================================

void prDisplayMsg(char *msg, char *response)
{
	int isq = response && *response == 'Q';
	// V220-33 added
	if (!isq && gl_inited)
	{
		if (titlepic_num >= 0 && appStatus == 2)
		{
			glMatrixMode(GL_PROJECTION);
			GL_DrawPagePatch(titlepic_num);
			SwapBuffers(mainHdc);
			GL_DrawPagePatch(titlepic_num);
			SwapBuffers(mainHdc);
			glMatrixMode(GL_TEXTURE);
		}
	}

	if (msg && msg[0])
	{
		int i;
		if (!response || isq) Sys_MessageBox(msg, true);
		pr_OFdirect("\n\n");
		for (i = 0; i < 64; i++) pr_OFdirect("!");
		pr_OFdirect("\n");
		pr_OFdirect(msg);
		if (response)
		{
			if (response[0] == 'Y')
				pr_OFdirect("\nUser chose to continue\n");
			else if (!isq)
				pr_OFdirect("\nUser chose to exit");
		}
		pr_OFdirect("\n");
		for (i = 0; i < 64; i++) pr_OFdirect("!");
		pr_OFdirect("\n\n");
	}
}

//=============================================================================
// Con_Error
//=============================================================================
void Con_Error (char *error, ...)
{
	if (appStatus == 3) exit(-1);

	con_error = true;

	if (error && error[0])
	{
		va_list argptr;

		va_start(argptr, error);
		_vsnprintf(prbuff, BUFFSIZE, error, argptr);
		va_end(argptr);

		prbuff[BUFFSIZE] = 0;  	
		prDisplayMsg(prbuff, NULL);
	}

	Sys_Quit(false);
}

//=============================================================================
// Con_ErrorD
//=============================================================================
void Con_ErrorD (char *error)
{
	if (appStatus == 3) exit(-1);

	con_error = true;
	prDisplayMsg(error, "Q");

	Sys_Quit(false);
}

//=============================================================================
// Con_ErrorDev	v220-31
//=============================================================================
void Con_ErrorDev (char *error, ...)
{
	if (appStatus == 3) exit(-1);

	if (error && error[0])
	{
		va_list argptr;

		va_start(argptr, error);
		_vsnprintf(prbuff, BUFFSIZE, error, argptr);
		va_end(argptr);

		prbuff[BUFFSIZE] = 0;
		
		if (ge_inf.devmap)
			prDisplayMsg(prbuff, NULL);
		else
			pr_OFdirect(prbuff);
	}

	if (ge_inf.devmap)
	{
		con_error = true;
		Sys_Quit(false);
	}
}

//=============================================================================
// Shut down console buffer and command 
//=============================================================================
void Con_Shutdown()
{	
	int		i, k;
	char	*ptr;

	Con_ClearBuffer();				// Free the buffer.

	for(i = 0; i < numOldCmds; i++)	// Free the old commands.
    {
        M_Free(oldCmds[i].text);
    }

	M_Free(oldCmds);
	oldCmds    = NULL;
	numOldCmds = 0;

	M_Free(knownWords);
	knownWords    = NULL;
	numKnownWords = 0;

	for(i = 0; i < numCVars; i++)	// Free the data of the data cvars.
    {
		if(cvars[i].flags & CVF_CAN_FREE && cvars[i].type == CVT_CHARPTR)
		{
			ptr = *(char**) cvars[i].ptr;

			// Multiple vars could be using the same pointer,
			// make sure it gets freed only once.

			for(k = i; k < numCVars; k++)
            {
				if(cvars[k].type == CVT_CHARPTR
					&& ptr == *(char**)cvars[k].ptr)
				{
					cvars[k].flags &= ~CVF_CAN_FREE;
				}
            }

			M_Free(ptr);
		}
    }

	M_Free(cvars);
	cvars    = NULL;
	numCVars = 0;

	M_Free(ccmds);
	ccmds = NULL;
	numCCmds = 0;

	for(i = 0; i < numCAliases; i++)	// Free the alias data.
	{
		M_Free(caliases[i].command);
		M_Free(caliases[i].name);
	}

 	M_Free(caliases);
	caliases = NULL;
	numCAliases = 0;

	ClearExecBuffer();
}

//=============================================================================
// Print into the buffer.and to outfile
//=============================================================================
void Con_Printfo(char *format, ...)
{
	va_list	args;
	
	va_start(args, format);
	_vsnprintf(prbuff, BUFFSIZE, format, args);
	va_end(args);
	
	prbuff[BUFFSIZE] = 0;

	if (!consoleDump) pr_OFinsert("%s", prbuff);

	conPrintf(CBLF_WHITE);
}

//=============================================================================
// Print into the buffer.
//=============================================================================
void Con_Printf(char *format, ...)
{
	va_list	args;
	
	va_start(args, format);
	_vsnprintf(prbuff, BUFFSIZE, format, args);
	va_end(args);
	
	prbuff[BUFFSIZE] = 0;

	conPrintf(CBLF_WHITE);
}

//=============================================================================
// jav - only print if verbose is true
//=============================================================================

void Con_FPrintf(int flags, char *format, ...)
{
	if ((verbose || con_active) && !ConsoleSilent)
	{
		va_list		args;

		va_start(args, format);
		_vsnprintf(prbuff, BUFFSIZE, format, args);
		va_end(args);

		prbuff[BUFFSIZE] = 0;  
		conPrintf(flags);
	}
}


//=============================================================================
//  Flagged to only print if edit mode is false
//=============================================================================

void Con_playFPrintf(RDBOOL always, int flags, char *format, ...)
{
	if (!(R3D_edit && !always) && !ConsoleSilent)
	{
		va_list		args;

		va_start(args, format);
		_vsnprintf(prbuff, BUFFSIZE, format, args);
		va_end(args);

		prbuff[BUFFSIZE] = 0;
		conPrintf(flags);
	}
}


//--------------------------------------------------------------------------
//
// Console Variable Handling
//
//--------------------------------------------------------------------------

char *plural(int val)
{
	if (val != 1) return "s";
	return "";
}

static char *TrimmedFloat(float val)
{
	char *ptr = trimmed_float_buffer;

	sprintf(ptr, "%f", val);

	// Get rid of the extra zeros.

	for(ptr += strlen(ptr)-1; ptr >= trimmed_float_buffer; ptr--)
	{
		if(*ptr == '0') 
        { 
           *ptr = 0;
        }
		else 
        if(*ptr == '.') 
		{
			*ptr = 0;
			break;
		}
		else break;
	}
	return trimmed_float_buffer;
}

//===========================================================================
// Con_SetString
//===========================================================================
void Con_SetString(char *name, char *text)
{
	cvar_t *cvar = Con_GetVariable(name);

	if(!cvar) return;
	if(cvar->type == CVT_CHARPTR)
	{
		// Free the old string, if one exists.
		if(cvar->flags & CVF_CAN_FREE && *(char**) cvar->ptr) 
		{	M_Free( *(char**) cvar->ptr);  // jav
        }

		// Allocate a new string.

		cvar->flags |= CVF_CAN_FREE;
		*(char**) cvar->ptr = (char*)M_Malloc(strlen(text)+1, 0);
		strcpy( *(char**) cvar->ptr, text);
	}
	else 
	{	Con_ErrorD("Con_SetString: cvar is not of type char*");
    }
}

//===========================================================================
// used here and ui_panel.c m_multi.c
//===========================================================================
cvar_t *Con_GetVariable(char *name)
{
	int		i;

	for(i = 0; i < numCVars; i++)
	{
		if (!_stricmp(name, cvars[i].name))
		{	return cvars + i;
        }
    }

	return NULL;	// No match
}

//===========================================================================
// Con_SetInteger
//	Also works with bytes.
//===========================================================================
void Con_SetInteger(char *name, int value)
{
	cvar_t *var = Con_GetVariable(name);

	if(!var) return;

	if(var->type == CVT_BYTE ) *(byte* ) var->ptr = (byte )value;
	if(var->type == CVT_INT  ) *(int*  ) var->ptr = value;
	if(var->type == CVT_FLOAT) *(float*) var->ptr = (float)value;
}

//===========================================================================
// Con_SetFloat
//===========================================================================
void Con_SetFloat(char *name, float value)
{
	cvar_t *var = Con_GetVariable(name);

	if(!var) return;
	if(var->type == CVT_BYTE ) *(byte* ) var->ptr = (byte) value;
	if(var->type == CVT_INT  ) *(int*  ) var->ptr = (int ) value;
	if(var->type == CVT_FLOAT) *(float*) var->ptr = value;
}

//===========================================================================
// Con_GetInteger
//===========================================================================
int Con_GetInteger(char *name)
{
	cvar_t *var = Con_GetVariable(name);

	if(!var) return 0;

	if(var->type == CVT_INT ) return *(int* ) var->ptr;
	if(var->type == CVT_BYTE) return *(byte*) var->ptr; //What's this doing here?

	return *(int*) var->ptr;
}

//===========================================================================
// Con_GetFloat
//===========================================================================
float Con_GetFloat(char *name)
{
	cvar_t *var = Con_GetVariable(name);

	if(!var || var->type != CVT_FLOAT) return 0;
	
    return *(float*) var->ptr;
}

//===========================================================================
// Con_GetByte
//===========================================================================
byte Con_GetByte(char *name)
{
	cvar_t *var = Con_GetVariable(name);

	if(!var) return 0;
	if(var->type == CVT_BYTE) return *(byte*) var->ptr;
	if(var->type == CVT_INT ) return *(int *) var->ptr; //What's this doing here?
	return 0;
}

//===========================================================================
// Con_GetString
//===========================================================================
char *Con_GetString(char *name)
{
	cvar_t *var = Con_GetVariable(name);

	if(!var || var->type != CVT_CHARPTR) return "";
	return *(char**) var->ptr;
}

//=============================================================================
//
//  Con_SaveDefaults
// 	Saves all bindings, aliases and archiveable console variables.
// 	The output file is a collection of console commands.
// 
//=============================================================================
void Con_SaveDefaults(void)
{
	cvar_t			*var;
	FILE			*file;
	int				i;

	maxLightDistance = ge_inf.maxLightDistance;

	if ( (file = fopen(configFileName, "wt")) == NULL)
	{
		pr_OFinsert("M_SaveDefaults: Can't open %s for writing.\n", configFileName);
		return;
	}

	fprintf(file, "# R I S E N 3 D %s config file\n", R3D_edit ? "edit" : "play");

	fprintf(file, "#\n# CONSOLE VARIABLES\n#\n\n");
	
	if (cfg.oldbarscale >= 0) cfg.sbarscale = cfg.oldbarscale;

	// We'll write all the console variables that are flagged for archiving.		
	for (i = 0; i < numCVars; i++)
	{
		var = cvars + i;
		if(var->type == CVT_NULL || var->flags & CVF_NO_ARCHIVE) continue;

		// First print the comment (help text).
		M_WriteCommented(file, var->help);

		fprintf(file, "%s ", var->name);

		if(var->flags & CVF_PROTECTED) fprintf(file, "force ");
		if(var->type == CVT_BYTE     ) fprintf(file, "%d", *(byte*)var->ptr);
		if(var->type == CVT_INT      ) fprintf(file, "%d", *(int*)var->ptr);
		if(var->type == CVT_FLOAT    ) fprintf(file, "%s", TrimmedFloat(*(float*)var->ptr));
		if(var->type == CVT_CHARPTR  ) 
		{
			fprintf(file, "\"");
			M_WriteTextEsc(file, *(char**) var->ptr);
			fprintf(file, "\"");
		}
		fprintf(file, "\n\n");
	}
	
	fprintf(file, "#\n# BINDINGS\n#\n\n");
	B_WriteToFile(file);	

	fclose(file);
}
//=============================================================================
//
//=============================================================================
static void PrepareCmdArgs(cmdargs_t *cargs, char *lpCmdLine)
{
	int		i, len = strlen(lpCmdLine);

	if (len > 1023)
	{
		pr_OFdirect("command line length of 1023 exceeded\n");
		Sys_Quit(false);
	}

	// Prepare the data.
	memset(cargs, 0, sizeof(cmdargs_t));
	// This is half the available size to allow space for
	// expansion/mem moves
	strncpy(cargs->cmdLine, lpCmdLine, 1024);
	cargs->cmdLine[1025] = 0;

	// Prepare.

	for(i=0; i<len; i++)
	{
		// Whitespaces are separators.
		if (FORMAT_CHAR(cargs->cmdLine[i])) cargs->cmdLine[i] = 0;

		if(cargs->cmdLine[i] == '\\' 
        && (cargs->cmdLine[i+1] == '"' ||cargs->cmdLine[i+1] == '\\')) // Escape sequence?
		{
			memmove(cargs->cmdLine+i, cargs->cmdLine+i+1, sizeof(cargs->cmdLine)-i-1);
			len--;
			continue;
		}

		if(cargs->cmdLine[i] == '"') // Find the end.
		{
			int start = i;
			cargs->cmdLine[i] = 0;
			for(++i; i<len && cargs->cmdLine[i] != '"'; i++)
			{
				if(cargs->cmdLine[i] == '\\' 
                && (cargs->cmdLine[i+1] == '"' ||cargs->cmdLine[i+1] == '\\')) // Escape sequence?
				{
					memmove(cargs->cmdLine+i, cargs->cmdLine+i+1, sizeof(cargs->cmdLine)-i-1);
					len--;
					continue;
				}
			}
			// Unterminated quotes?

			if(i == len) break;

			// An empty set of quotes?

			if(i == start+1) 
				cargs->cmdLine[i] = SC_EMPTY_QUOTE;
			else
				cargs->cmdLine[i] = 0;
		}
	}

	// Scan through the cmdLine and get the beginning of each token.
	cargs->argc = 0;
	for(i=0; i<len; i++)
	{
		if(!cargs->cmdLine[i]) continue;
		// Is this an empty quote?
		if(cargs->cmdLine[i] == SC_EMPTY_QUOTE)
			cargs->cmdLine[i] = 0;	// Just an empty string.
		cargs->argv[cargs->argc++] = cargs->cmdLine + i;
		i += strlen(cargs->cmdLine+i);
	}
}

//=============================================================================
//
//=============================================================================

static int __cdecl wordListSorter(const void *e1, const void *e2)
{
	return _stricmp(*(char**)e1, *(char**)e2);
}

static int __cdecl knownWordListSorter(const void *e1, const void *e2)
{
	return _stricmp(( (knownword_t*) e1)->word, ( (knownword_t*) e2)->word);
}


//===========================================================================
// Con_MaxLineLength
//===========================================================================
void Con_MaxLineLength(void)
{
	int cw = FR_TextWidth("A");

	if(!cw)  // ???
	{
		maxLineLen = 100;  // jav was 70 ,need at least 100
		return;
	}
	maxLineLen = current_w/cw - 2;
	if(maxLineLen > 250) maxLineLen = 250;
}

//===========================================================================
// Con_UpdateKnownWords
//	Variables with CVF_HIDE are not considered known words.
//===========================================================================
void Con_UpdateKnownWords()
{
	int		i, c, known_vars;
	int		len;

	// Count the number of visible console variables.
	for(i = 0, known_vars = 0; i < numCVars; i++)
	{	
        if(!(cvars[i].flags & CVF_HIDE)) known_vars++;
    }

	// Fill the known words table.

    numKnownWords = numCCmds + known_vars + numCAliases;
	len = sizeof(knownword_t) * numKnownWords;

	M_Free(knownWords);
	knownWords = (knownword_t*)M_Calloc(len, 200);	//Ensure enough space

	// Commands, variables and aliases are known words.	
    for(i = 0, c = 0; i < numCCmds; i++, c++)
	{
		strncpy(knownWords[c].word, ccmds[i].name, 63);
	}
	for(i = 0; i < numCVars; i++) 
	{
		if(!(cvars[i].flags & CVF_HIDE))
			strncpy(knownWords[c++].word, cvars[i].name, 63);
	}
	for(i = 0; i < numCAliases; i++, c++)
	{
		strncpy(knownWords[c].word, caliases[i].name, 63);
	}

	// Sort into alphabetical order.
	qsort(knownWords, numKnownWords, sizeof(knownword_t), knownWordListSorter);
}

static void subfn_AddCommand(ccmd_t *cmd)
{
	numCCmds++;
	ccmds = (ccmd_t*)M_Realloc(ccmds, sizeof(ccmd_t) * numCCmds, 4);
	memcpy(ccmds + numCCmds-1, cmd, sizeof(ccmd_t));
}

//=============================================================================
//
//=============================================================================
static void Con_AddCommandList(ccmd_t *cmdlist)
{
    for(; cmdlist->name; cmdlist++) 
    {   
        subfn_AddCommand(cmdlist);
    }

	// Sort them.
	qsort(ccmds, numCCmds, sizeof(ccmd_t), wordListSorter);
}
//=============================================================================
//
//=============================================================================
static void subfn_AddVariable(cvar_t *var)
{
	cvars = (cvar_t*)M_Realloc(cvars, sizeof(cvar_t) * (numCVars + 1), 5);

	memcpy(cvars + numCVars, var, sizeof(cvar_t));	// Add the new cvar inf

	numCVars++;
}

//=============================================================================
//
//=============================================================================
void Con_AddVariableList(cvar_t *varlist)
{
	for(; varlist->name; varlist++) 
    {  
        subfn_AddVariable(varlist);
    }

	// Sort them.
	qsort(cvars, numCVars, sizeof(cvar_t), wordListSorter);
}

//=============================================================================
//
// Returns NULL if the specified alias can't be found.
//=============================================================================
calias_t *Con_GetAlias(char *name)
{
	int			i;

	// Try to find the alias.

	for(i = 0; i < numCAliases; i++)
	{
		if(!_stricmp(caliases[i].name, name))
			return caliases + i;
    }
	return NULL;
}

//===========================================================================
// Con_Alias
//	Create an alias.
//===========================================================================
void Con_Alias(char *aName, char *command)
{
	calias_t	*cal = Con_GetAlias(aName);
	RDBOOL		ignore = false;

	// Should we ignore this alias?
	if(command == NULL || !command[0]) 
		ignore = true;
	
	if(cal && ignore)
	{
		// This alias will be removed.
		int i, idx = cal - caliases;

		if (idx < 0 || idx >= numCAliases)
			Con_Error("Con_Alias: bad index number %i", idx);

		M_Free(cal->name   ); cal->name    = NULL;
		M_Free(cal->command); cal->command = NULL;

		numCAliases -= 1;

		if (numCAliases <= 0)
		{
			numCAliases = 0;
			M_Free(caliases); caliases = NULL;
			return;
		}

		for (i = idx; i < numCAliases; i++)
		{
			caliases[i] = caliases[i + 1];
		}

		caliases = (calias_t*)M_Realloc(caliases, sizeof(calias_t) * numCAliases, 6);

		return;	// We're done.
	}

	// Does the alias already exist?
	if(cal)
	{
		cal->command = (char*)M_Realloc(cal->command, strlen(command)+1, 7); // jav
		strcpy(cal->command, command);
		return;
	}

	// We need to create a new alias.
	caliases = (calias_t*)M_Realloc(caliases, sizeof(calias_t) * (++numCAliases), 8);// jav
	cal = caliases + numCAliases-1;
	// Allocate memory for them.
	cal->name    = (char*)M_Malloc(strlen(aName)+1, 1);   // jav
	cal->command = (char*)M_Malloc(strlen(command)+1, 2); // jav
	strcpy(cal->name, aName);
	strcpy(cal->command, command);
//	cal->refcount = 0;

	// Sort them.
	qsort(caliases, numCAliases, sizeof(calias_t), wordListSorter);

	Con_UpdateKnownWords();
}

//=============================================================================
//
//=============================================================================
static void Con_ClearBuffer(void)
{
	int i, count = 0;

	// Free the buffer.
	for(i = 0; i < bufferLines; i++) 
    {  
        M_Free(cbuffer[i].text); // jav
    }

	M_Free(cbuffer);

	cbuffer     = NULL;
	bufferLines = 0;
	bPos        = 0;
	bFirst      = 0;
	bLineOff    = 0;
}

//=============================================================================
//
//=============================================================================
static void ClearExecBuffer(void)
{
	if (exBuffSize)
	{
		int i;

		for (i = 0; i < exBuffSize; i++)
			M_Free(exBuff[i].subcmd);

		M_Free(exBuff);
		exBuff     = NULL;
		exBuffSize = 0;
	}
}

void subcmdError()
{
	Con_ErrorD( "ERROR Excessively long subcommand encountered.\n\n"
				"Try deleting the Risen3D.cfg before re-starting\n");
}

//=============================================================================
//
//=============================================================================
static void QueueCmd(char *singleCmd, int at_tic)
{
	execbuff_t	*ptr = NULL;
	int			i;

	if (!singleCmd[0]) return;			// Check there's something to store
	
	if (strlen(singleCmd) >= SUBCSIZE)	// and that it's within bounds
		 subcmdError();

	// Look for an empty spot.
	for (i = 0; i < exBuffSize; i++)
	{
		if (!exBuff[i].used)
		{
			ptr = exBuff + i;
			break;
		}
	}

	if (ptr == NULL)
	{
		// No empty places, allocate a new one.
		exBuff = (execbuff_t*)M_Realloc(exBuff, sizeof(execbuff_t) * ++exBuffSize, 9);
		ptr    = exBuff + exBuffSize-1;
		ptr->subcmd = NULL;
	}

	ptr->subcmd = (char*)M_Realloc(ptr->subcmd, strlen(singleCmd) + 1, 10);
	strcpy(ptr->subcmd, singleCmd);
	ptr->used = true;
	ptr->marker = at_tic;
}

//============================================================================
// Returns false if an executed command fails.
//============================================================================
static RDBOOL CheckExecBuffer()
{
	RDBOOL  repeat = true;
	RDBOOL	ret  = true;

	int		i;
	char	storage[SUBCSIZE + 1];

	while (repeat) // We'll keep checking until all is done.
	{
		repeat = false;

		// Execute the commands marked for this or a previous tic.
		for (i = 0; i < exBuffSize; i++)
		{	
			execbuff_t *ptr = exBuff + i;

			if (!ptr->used || ptr->marker > (unsigned) systics) continue;

			// Execute command.
			curExec   = ptr;
			ptr->used = false;

			if (strlen(ptr->subcmd) >= SUBCSIZE)
				subcmdError();

			strcpy(storage, ptr->subcmd);

			if (!executeSubCmd(storage)) ret = false;

			repeat = true;
		}
	}

	return ret;
}

//============================================================================
//
// dd_loop uses
//============================================================================
void Con_Ticker(void)
{
	CheckExecBuffer();

	if(ConsoleY == 0) openingOrClosing = true;

	// Move the console to the destination Y.

	if(ConsoleDestY > ConsoleY)
	{
		float diff = (ConsoleDestY - ConsoleY)/4;
		if(diff < 1) diff = 1;
		ConsoleY += diff;		
		if(ConsoleY > ConsoleDestY) ConsoleY = ConsoleDestY;
	}
	else 
    if(ConsoleDestY < ConsoleY)
	{
		float diff = (ConsoleY - ConsoleDestY)/4;
		if(diff < 1) diff = 1;
		ConsoleY -= diff;		
		if(ConsoleY < ConsoleDestY) ConsoleY = ConsoleDestY;
	}

	if(ConsoleY == ConsoleOpenY) openingOrClosing = false;

	funnyAng += consoleTurn/10000.0f;//.002f;

	if(!con_active) return;	// We have nothing further to do here.

	ConsoleTime++;	// Increase the ticker.
}

//=============================================================================
//
// con_start uses
//=============================================================================

cbline_t *Con_GetBufferLine(int num)
{
	int i, newLines;

	if (num < 0) return 0;	// This is unacceptable!

	// See if we already have that line.

	if (num < bufferLines) return cbuffer + num;

	// Then we'll have to allocate more lines. Usually just one, though.

	newLines     = num+1 - bufferLines;
	bufferLines += newLines;
	cbuffer      = (cbline_t*)M_Realloc(cbuffer, sizeof(cbline_t)*bufferLines, 11); // jav

	for (i = 0; i < newLines; i++)
	{
		cbline_t *line = cbuffer + i + bufferLines - newLines;
		memset(line, 0, sizeof(cbline_t));
	}

	return cbuffer + num;
}
//=============================================================================
//
//=============================================================================
static void addLineText(cbline_t *line, char *txt)
{
	int newLen = line->len + strlen(txt);

	if(newLen > maxLineLen)
	{	
		Con_Error("addLineText: Console line %i exceeds max allowed %i",
			newLen, maxLineLen);
    }

	// Allocate more memory.
	line->text = (char*)M_Realloc(line->text, newLen +1, 12 ); // jav

	// Copy the new text to the appropriate location.
	strcpy(line->text+line->len, txt);
	
    // Update the length of the line.
	line->len = newLen;
}

//=============================================================================
//
//=============================================================================
static void addOldCmd(const char *txt)
{
	cbline_t *line;

	if(!strcmp(txt, "")) return; // Don't add empty commands.

	// Allocate more memory.

	oldCmds = (cbline_t*)M_Realloc(oldCmds, sizeof(cbline_t) * ++numOldCmds, 13); // jav
	line    = oldCmds + numOldCmds-1;

	// Clear the line.

	memset(line, 0, sizeof(cbline_t));
	line->len  = strlen(txt);
	line->text = (char*)M_Malloc(line->len + 1, 3); // +1 for terminating 0
	strcpy(line->text, txt);
}

//=============================================================================
//
//=============================================================================
static void printcvar(cvar_t *var, char *prefix)
{
	char equals = '=';
	
	if(var->flags & CVF_PROTECTED) equals = ':';

	Con_Printf(prefix);

	switch(var->type)
	{
	case CVT_NULL:
		Con_Printf("%s", var->name);
		break;
	case CVT_BYTE:
		Con_Printf("%s %c %d", var->name, equals, *(byte*)var->ptr);
		break;
	case CVT_INT:
		Con_Printf("%s %c %d", var->name, equals, *(int*)var->ptr);
		break;
	case CVT_FLOAT:
		Con_Printf("%s %c %s", var->name, equals,TrimmedFloat(*(float*)var->ptr));
		break;
	case CVT_CHARPTR:
		Con_Printf("%s %c %s", var->name, equals, *(char**)var->ptr);
		break;
	default:
		Con_Printf("%s (bad type!)", var->name);
		break;
	}
	Con_Printf("\n");
}

// expcommand gets reallocated in the expansion process.
//=============================================================================
//
//=============================================================================

static void expandWithArguments(char **expcommand, cmdargs_t *args)
{
	char *text = *expcommand;
	int size = strlen(text)+1;
	int	i, p, off;

	for(i=0; text[i]; i++)
	{
		if(text[i] == '%' && (text[i+1] >= '1' && text[i+1] <= '9'))
		{
			char	*substr;
			int		aidx = text[i+1]-'1'+1;
			
			// Expand! (or delete)
			if (aidx >= args->argc)	// Handle null pointers
				substr = "";
			else
				substr = args->argv[aidx];

			// First get rid of the %n.
			memmove(text+i, text+i+2, size-i-2);
			// Reallocate.
			off = strlen(substr);
			text = *expcommand = (char*)M_Realloc(*expcommand, size+=off-2, 14);
			if(off)
			{
				// Make room for the insert.
				memmove(text+i+off, text+i, size-i-off);												
				memcpy(text+i, substr, off);
			}
			i += off-1;
		}
		else if(text[i] == '%' && text[i+1] == '0')
		{
			// First get rid of the %0.
			memmove(text+i, text+i+2, size-i-2);
			text = *expcommand = (char*)M_Realloc(*expcommand, size-=2, 15);
			for(p=args->argc-1; p>0; p--)
			{
				off = strlen(args->argv[p])+1;
				text = *expcommand = (char*)M_Realloc(*expcommand, size+=off, 16);
				memmove(text+i+off, text+i, size-i-off);
				text[i] = ' ';
				memcpy(text+i+1, args->argv[p], off-1);
			}
		}
	}
}

//=============================================================================
// Execute sub command
//=============================================================================
static int executeSubCmd(char *subcmd)
{
	int			i;
	char		prefix;
	cmdargs_t	args;


	PrepareCmdArgs(&args, subcmd);

	if (!args.argc) return true;

	if (args.argc == 1)	// An action?
	{
		prefix = args.argv[0][0];

		if (prefix == '+' || prefix == '-')
		{
			return Con_ActionCommand(args.argv[0], true);
		}
		
        // What about a prefix-less action?
		
        if (   strlen(args.argv[0]) <= 8 
			&& Con_ActionCommand(args.argv[0], false)) 
		{	
           return true; // There was one!
        }
	}

	// Try to find a matching command.
	for(i=0; i<numCCmds; i++)
	{
		if(!_stricmp(ccmds[i].name, args.argv[0]))
		{
			int cret;

			if (appStatus == 0)
			{
				return true;	// commands can't be run as yet
			}
			else if (appStatus == 1)
			{
				// Disallow warp on startup
				if (!_stricmp(ccmds[i].name, "warp"))
					return true;
			}

			if (!ccmds[i].func)
			{
				Con_Message( "CMD program error: '%s' has no matching function\n", ccmds[i].name);
				return true;
			}

			cret = ccmds[i].func(args.argc, args.argv);

			if (cret == false)
			{
				if (_stricmp(ccmds[i].name, "safebind")) // i.e. not safebind
				{
					Con_Message( "CMD error: '%s' failed.\n", ccmds[i].name);
				}
			}
			// We're quite done.
			return cret;
		}
	}

	// Then try the cvars?

	for(i = 0; i < numCVars; i++)
	{
		if(!_stricmp(cvars[i].name, args.argv[0]))
		{
			cvar_t *var = cvars + i;
			RDBOOL out_of_range = false;
			RDBOOL setting = false;
			RDBOOL cannot_set_int = false;
			RDBOOL cannot_set_float = false;
			RDBOOL panel_only = false;

			if( args.argc == 2 
            || (args.argc == 3 && !_stricmp(args.argv[1], "force")) )
			{
				char *argptr = args.argv[ args.argc-1 ];
				RDBOOL forced = args.argc==3;

				setting = true;

				if (var->flags & CVF_CPBUTTON && appStatus == 2)
				{
					Con_Printf ("%s   This can only be set using a Control Panel button\n",
								var->name);
				}
				else if (var->flags & CVF_PROTECTED && !forced)
				{
					Con_Printf ("%s is protected. You shouldn't change its value.\n"
								"If you must then use: \"%s force %s\"\n",
								var->name, var->name, argptr);
				}
				else if(var->type == CVT_BYTE)
				{
					byte val = (byte) strtol(argptr, NULL, 0);
					if (!forced
						&&
						((!(var->flags & CVF_NO_MIN) && val < var->min) 
						|| (!(var->flags & CVF_NO_MAX) && val > var->max))
						)
					{
						out_of_range = true;
                    }
					else
					{
						*(byte*)var->ptr = val;
                    }
				}
				else if(var->type == CVT_INT)
				{
					int val = strtol(argptr, NULL, 0);

					if (!strcmp(var->name, "midi-dev") && var->max > (int)numMidiDev) var->max = (float)numMidiDev;

					if (R3D_ScriptedMap && var->flags & CVF_SCP_FIXED)
					{
						cannot_set_int = true;
					}
					else if (!ConsoleSilent && var->flags & CVF_INT_PANEL)
					{
						panel_only = true;
					}
					else if (!forced
						&& ((!(var->flags & CVF_NO_MIN) && val < var->min)
						||  (!(var->flags & CVF_NO_MAX) && val > var->max)) )
					{
						out_of_range = true;
					}
					else
					{
						*(int*)var->ptr = val;
					}
				}
				else if(var->type == CVT_FLOAT)
				{
					float val = (float)(strtod(argptr, NULL));

					if (R3D_ScriptedMap && var->flags & CVF_SCP_FIXED)
					{
						cannot_set_float = true;
					}
					else if (!forced
						&&
						((!(var->flags & CVF_NO_MIN) && val < var->min)
						|| (!(var->flags & CVF_NO_MAX) && val > var->max))
						)
					{
						out_of_range = true;
					}
					else
					{
						*(float*)var->ptr = val;
					}
				}
				else if(var->type == CVT_CHARPTR)
				{
					Con_SetString(var->name, argptr);
				}
			}

			if (out_of_range)
			{
				if (!(var->flags & (CVF_NO_MIN|CVF_NO_MAX)))
				{
					char temp[20];
					strcpy(temp, TrimmedFloat(var->min));
					Con_Printf( "Error: %s >= %s <= %s\n", 
						        temp, var->name, TrimmedFloat(var->max));
				}
				else if (var->flags & CVF_NO_MAX)
				{
					Con_Printf( "Error: %s >= %s\n", var->name, 
				        		TrimmedFloat(var->min));
				}
				else 
				{
					Con_Printf( "Error: %s <= %s\n", var->name, 
						        TrimmedFloat(var->max));
				}
			}
			
			if (panel_only)
			{
				Con_Printf("Value for \"%s\" can only be set in Control Panel\n", var->name);
				return true;
			}
			
			if (cannot_set_int)
			{
				Con_Printf("Scripted map: %s value fixed at %i\n", var->name, *(int*)var->ptr);
				return true;
			}
			
			if (cannot_set_float)
			{
				Con_Printf("Scripted map: %s value fixed at %.2f\n", var->name, *(float*)var->ptr);
				return true;
			}

			if (var->flags & CVF_DOPPLER && !out_of_range)
			{
				Sfx_Doppler();
			}

            if (!setting 
				|| (setting && !conSilentCVars) )	// Show the value.
			{	
               printcvar(var, "");
            }
			return true;
		}
	}

	// How about an alias then?
	for(i = 0; i < numCAliases; i++)
	{
		if(!_stricmp(args.argv[0], caliases[i].name))
		{
			calias_t *cal = caliases + i;
			char *expcommand;

			// Expand the command with arguments.
			expcommand = (char*)M_Malloc(strlen(cal->command)+1, 4);
			strcpy(expcommand, cal->command);
			
            expandWithArguments(&expcommand, &args);
			// Do it, man!
			SplitIntoSubCommands(expcommand, 0);
			
            M_Free(expcommand); // jav
			// Decrement the refcount.
			//cal->refcount--;
			//return ret;
			return true;
		}
	}

	// What *is* that?
	Con_Printf("%s: no such command or variable.\n", args.argv[0]);

	return false;
}

//=============================================================================
// Splits the command into subcommands and queues them into the execution 
// buffer.  The command may contain many commands, separated with semicolons.
// b211-0 revised
//=============================================================================
static void SplitIntoSubCommands(char *command, int markerOffset)
{
	int		gpos  = 0;
	int     scpos = 0;
	
	int		inquotes = false;
	int     escape   = false;
	
	char	subcmd[SUBCSIZE + 1];              // largest allowed
	RDBOOL	end;
	
	// Is there a command to execute?	
	if (!command || command[0] == 0) return;
	
	// Jump over initial spaces/semicolons.
	while (command[gpos] == ';' || command[gpos] == ' ') gpos++;
		
	while (scpos < SUBCSIZE)
	{
		escape = false;
		
		if (inquotes && command[gpos] == '\\') // Escape sequence?
		{
			subcmd[scpos++] = command[gpos++];
			escape = true;
		}
		
		if (command[gpos] == '"' && !escape) 
			inquotes = !inquotes;

		end = command[gpos] == 0;
		
		// Collect characters.
		if (!end) subcmd[scpos++] = command[gpos++];
		
		if (end || (command[gpos] == ';' && !inquotes))
		{
			subcmd[scpos] = 0;
			QueueCmd(subcmd, systics + markerOffset);
			if (end) break;
			
			// Get next command
			scpos = 0;
			while(command[gpos] == ';' || command[gpos] == ' ') gpos++;
		}
	}

	if (command[gpos]) subcmdError();
}


//===========================================================================
// Con_Execute
//	Returns false if a command fails.
//===========================================================================
int Con_Execute(char *command, int silent)
{
	int ret;
    
	if (silent) ConsoleSilent = true;

    SplitIntoSubCommands(command, 0);

	ret = CheckExecBuffer();

	if (silent) ConsoleSilent = false;
	return ret;
}

//===========================================================================
// Con_Executef
//===========================================================================
int Con_Executef(int silent, char *command, ...)
{
	va_list argptr;

	va_start(argptr, command);
	vsprintf(prbuff, command, argptr);
	va_end(argptr);

	return Con_Execute(prbuff, silent);
}

//=============================================================================
//
//=============================================================================

static void processCmd()
{
	DD_ClearKeyRepeaters();

	// Add the command line to the oldCmds buffer.

	addOldCmd(cmdLine);
	ocPos = numOldCmds;

	Con_Execute(cmdLine, false);
}

//=============================================================================
//
//=============================================================================

static void updateCmdLine()
{
	if(ocPos == numOldCmds)
		strcpy(cmdLine, "");
	else
		strcpy(cmdLine, oldCmds[ocPos].text);

	complPos = strlen(cmdLine);
	lastCompletion = -1;
}

//===========================================================================
// stramb
//	Ambiguous string check. 'amb' is cut at the first character that 
//	differs when compared to 'str' (case ignored).
//===========================================================================
void stramb(char *amb, const char *str)
{
	while(*str 
		&& tolower( (unsigned) *amb) == tolower( (unsigned) *str)) 
	{
		amb++;
		str++;
	}
	*amb = 0;
}

//===========================================================================
// completeWord
//	Look at the last word and try to complete it. If there are
//	several possibilities, print them.
//===========================================================================
static void completeWord()
{
	int		pass, i, c, cp = strlen(cmdLine)-1;
	int		numcomp = 0;
	char	word[100], *wordBegin;
	char	unambiguous[256];
	char	*completion=0;	// Pointer to the completed word.
	cvar_t	*cvar;

	if(conCompMode == 1) cp = complPos-1;
	memset(unambiguous, 0, sizeof(unambiguous));

	if(cp < 0) return;

	// Skip over any whitespace behind the cursor.

	while(cp > 0 && cmdLine[cp] == ' ') cp--;
	
	// Rewind the word pointer until space or a semicolon is found.

	while(cp > 0 
		&& cmdLine[cp-1] != ' ' 
		&& cmdLine[cp-1] != ';' 
		&& cmdLine[cp-1] != '"') cp--;

	// Now cp is at the beginning of the word that needs completing.

	strcpy(word, wordBegin = cmdLine + cp);

	if(conCompMode == 1)
	   word[complPos - cp] = 0;	// Only check a partial word.

	// The completions we know are the cvars and ccmds.

	for(pass = 1; pass <= 2; pass++)
	{
		if(pass == 2)	// Print the possible completions.
		{	
           Con_Printf("Completions:\n");
        }

		// Look through the known words.
		for(c=0, i = (conCompMode==0? 0 : (lastCompletion+1)); 	c < numKnownWords; c++, i++)
		{
			if(i > numKnownWords-1) i = 0;

			if(!strnicmp(knownWords[i].word, word, strlen(word)))
			{
				// This matches!
				if(!unambiguous[0])
					strcpy(unambiguous, knownWords[i].word);
				else
					stramb(unambiguous, knownWords[i].word);
				// Pass #1: count completions, update completion.
				if(pass == 1) 
				{
					numcomp++;
					completion = knownWords[i].word;
					if(conCompMode == 1)
					{
						lastCompletion = i;
						break;
					}
				}
				else // Pass #2: print it.
				{
					// Print the value of all cvars.
					if((cvar = Con_GetVariable(knownWords[i].word)) != NULL)
						printcvar(cvar, "  ");
					else
						Con_Printf("  %s\n", knownWords[i].word);
				}
			}
		}
		if(numcomp <= 1 || conCompMode == 1) break;				
	}

	// Was a single completion found?

	if(numcomp == 1)
	{
		strcpy(wordBegin, completion);
		strcat(wordBegin, " ");
	}
	else 
    if(numcomp > 1)
	{
		// More than one completion; only complete the unambiguous part.
		strcpy(wordBegin, unambiguous);
	}
}

//=============================================================================
//  Con_Responder (dd_input and con_console)
// 	Returns true if the event is eaten.
// 
//=============================================================================
RDBOOL Con_Responder(event_t *event)
{
	int pos;
	byte ch;

	if (consoleShowKeys && event->type == ev_keydown)
	{
		Con_Printf("Keydown: ASCII %i (0x%x)\n", event->data1, event->data1);
	}

	// Special console key: Shift-Escape opens the Control Panel.

	if (   shiftDown 
		&& event->type == ev_keydown 
		&& event->data1 == DDKEY_ESCAPE)
	{
		Con_Execute("panel", true);  // direct access into/outof menu
		return true;
	}

	if (   event->type == ev_keydown 
		&& event->data1 == DDKEY_F4)
	{
		if (!menu_active)
			Con_Execute("SoundMenu", true);  // direct access to main control panel
		return true;
	}

	if (!con_active)
	{
		// In this case we are only interested in the activation key.

		if(event->type == ev_keydown 
        && event->data1 == consoleActiveKey) // && !MenuActive)
		{
			Con_Open(!menu_active);
			return true;
		}
		return false;
	}

	// All keyups are eaten by the console.

	if(event->type == ev_keyup) return true; 

	// We only want keydown events.
	if(event->type != ev_keydown && event->type != ev_keyrepeat) 
		return false;

	// In this case the console is active and operational.
	// Check the shutdown key.

	if(event->data1 == consoleActiveKey)
	{
		if(shiftDown) // Shift-Tilde to fullscreen and halfscreen.
        {
			ConsoleDestY = ConsoleOpenY = (ConsoleDestY == 200? 100.0f : 200.0f);
        }
		else
        {
			Con_Open(false);
        }
		return true;
	}

	// Hitting Escape in the console closes it.

	if(event->data1 == DDKEY_ESCAPE)
	{
		Con_Open(false);
		return false;	// Let the menu know about this.
	}

	switch(event->data1)
	{
    // jav this these work like old UP arrow
	case DDKEY_LEFTARROW:
		if(--ocPos < 0) ocPos = 0;
		// Update the command line.
		updateCmdLine();
		break;
    // jav this work like old DOWN arrow
	case DDKEY_RIGHTARROW:
		if(++ocPos > numOldCmds) ocPos = numOldCmds;
		updateCmdLine();
		break;

    // jav UP a line (old code shown)
	case DDKEY_UPARROW:
		//if(--ocPos < 0) ocPos = 0; jav old code
		// Update the command line.
		//updateCmdLine();
        // jav now this moves up 1 line
		bLineOff += 1;
		if(bLineOff > bPos-1) bLineOff = bPos-1;

		break;
    // jav DOWN a line (old code shown)
	case DDKEY_DOWNARROW:
		//if(++ocPos > numOldCmds) ocPos = numOldCmds;
		//updateCmdLine();
        // jav now moves down 1 line
		bLineOff -= 1;
		if(bLineOff < 0) bLineOff = 0;
		break;

		// jav UP a page was up 2 lines (new uparrow see above)
	case DDKEY_PGUP:
		bLineOff += -1 + (int)( ( (float)ConsoleOpenY + fontSy/2 )/fontSy );       // jav
		if(bLineOff > bPos-1) bLineOff = bPos-1;
		break;
		
		// jav DOWN a page was down 2 lines (new downarrow see above)
	case DDKEY_PGDN:
		bLineOff -= -1 + (int)( ( (float)ConsoleOpenY + fontSy/2 )/fontSy );       // jav
		if(bLineOff < 0) bLineOff = 0;
		break;

	// change size of console
	case DDKEY_INS:
		ConsoleOpenY -= fontSy * (shiftDown? 3 : 1);
		if(ConsoleOpenY < fontSy) ConsoleOpenY = fontSy;
		ConsoleDestY = ConsoleOpenY;
		break;

	// change the size of the console
	case DDKEY_DEL:
		ConsoleOpenY += fontSy * (shiftDown? 3 : 1);
		if(ConsoleOpenY > 200) ConsoleOpenY = 200;
		ConsoleDestY = ConsoleOpenY;
		break;

	// last line
	case DDKEY_END:
		bLineOff = 0;
		break;
	// first line

	case DDKEY_HOME:
		bLineOff = bPos-1;
		break;

	case DDKEY_ENTER:
		// Print the command line with yellow text.
		Con_FPrintf(CBLF_YELLOW, ">%s\n", cmdLine);
		// Process the command line.
		processCmd();		
		// Clear it.
		cmdLine[0] = 0;
		complPos = 0;
		lastCompletion = -1;
		break;

	case DDKEY_BACKSPACE:
		pos = strlen(cmdLine);
		if(pos) cmdLine[pos-1] = 0;
		complPos = strlen(cmdLine);
		lastCompletion = -1;
		break;

	case DDKEY_TAB:
		completeWord();
		break;
    // Check for a character
	default:	
		ch = (byte)(event->data1);
		if(ch < 32 || ch > 127 && ch < DD_HIGHEST_KEYCODE) return true;
		ch = DD_ModKey(ch);

		pos = strlen(cmdLine);
		if(pos > 79) pos = 79;
		cmdLine[pos] = ch;
		cmdLine[pos+1] = 0;
		complPos = strlen(cmdLine);
		lastCompletion = -1;
		break;
	}
	// The console is very hungry for keys...
	return true;
}

//=============================================================================
//
//=============================================================================
static void consoleSetColor(int fl, float alpha)
{
	float	r=0, g=0, b=0;
	int		count=0;

	// Calculate the average of the given colors.
	if(fl & CBLF_BLACK)
	{
		count++;
	}
	if(fl & CBLF_BLUE)
	{
		b += 1;
		count++;
	}
	if(fl & CBLF_GREEN)
	{
		g += 1;
		count++;
	}
	if(fl & CBLF_CYAN)
	{
		g += 1;
		b += 1;
		count++;
	}
	if(fl & CBLF_RED)
	{
		r += 1;
		count++;
	}
	if(fl & CBLF_MAGENTA)
	{
		r += 1;
		b += 1;
		count++;
	}
	if(fl & CBLF_YELLOW)
	{
		r += CcolYellow[0];
		g += CcolYellow[1];
		b += CcolYellow[2];
		count++;
	}
	if(fl & CBLF_WHITE)
	{
		r += 1;
		g += 1;
		b += 1;
		count++;
	}
	// Calculate the average.
	if(count)
	{
		r /= count;
		g /= count;
		b /= count;
	}
	if(fl & CBLF_LIGHT)
	{
		r += (1-r)/2;
		g += (1-g)/2;
		b += (1-b)/2;
	}
	glColor4f(r, g, b, alpha);
}

void Con_SetFont(ddfont_t *cfont)
{
	Cfont = *cfont;
}

//===========================================================================
// Con_DrawRuler2
//===========================================================================
void Con_DrawRuler2(int y, int lineHeight, float alpha, int scrWidth)
{
	int xoff = 5;
	int rh   = 6;

	UI_DrawRectEx((float)xoff, (float)(y + (lineHeight - rh)/2 + 1), 
		scrWidth - 2*xoff, rh, (float)(rh/2), 
		UI_COL(UIC_TEXT), UI_COL(UIC_BG_DARK), UI_COL(UIC_BG_LIGHT),
		alpha);
}

//===========================================================================
// Con_DrawRuler
//===========================================================================
void Con_DrawRuler(int y, int lineHeight, float alpha)
{
	Con_DrawRuler2(y, lineHeight, alpha, current_w);
}

//===========================================================================
// Con_Drawer for when user presses console key (not startup)
//	Slightly messy...
//===========================================================================
void Con_Drawer(void)
{
	int		i;	// Line count and buffer cursor.
	float	x, y;
	float	closeFade = 1;
	float	gtosMulY = (float)current_h / 200.0f;
	char	buff[256];
	float	fontScaledY = Cfont.height * Cfont.sizeY;

	if( ConsoleY == 0) return;	// We have nothing to do here.

	fontSy = fontScaledY/gtosMulY;

	// Do we have a font?
	FR_SetFont(0);

	if(Cfont.uiTextOut == NULL)
	{
		Cfont.flags		= DDFONT_WHITE;
		Cfont.height	= FR_TextHeight("Con");
		Cfont.sizeX		= 1;
		Cfont.sizeY		= 1;
		Cfont.uiTextOut	= (void (__cdecl *)(char *,int,int))FR_TextOut;
		Cfont.Width		= FR_TextWidth;
		Cfont.Filter	= NULL;
	}

	// Go into screen projection mode.

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, current_w, current_h, 0, -1, 1);

	if(openingOrClosing)
	{
		closeFade = ConsoleY / (float) ConsoleOpenY;
	}

	// The console is composed of 
    // 1. the main area background
    // 2. the border

	GL_SetFlat(console_flat);

	// Let's make it a bit more interesting.

	glMatrixMode(GL_TEXTURE);
	glPushMatrix();
	
    glLoadIdentity();
	glTranslatef((float)(2*sin(funnyAng/4)), (float)(2*cos(funnyAng/4)), 0);
	glRotatef(funnyAng*3, 0, 0, 1);

	glColor4f (consoleLight/100.0f, consoleLight/100.0f, 
        	   consoleLight/100.0f, closeFade * consoleAlpha/100);
	GL_DrawRectTiled(0, ConsoleY*gtosMulY + 4.0f, (float)current_w, (float)(-current_h-4), 64.0f, 64.0f);

	glMatrixMode(GL_TEXTURE);
	glPopMatrix();

	// Switch to the projection matrix
	glMatrixMode(GL_PROJECTION);

	// The border.
	GL_DrawRect(0, ConsoleY*gtosMulY + 3, (float)current_w, 2,0,0,0, closeFade);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glScalef(Cfont.sizeX, Cfont.sizeY, 1.0f);

	// show the version text on the bottom right

	y = ConsoleY*gtosMulY - fontScaledY*2;

	if(y > -fontScaledY)
	{
		int ssw = (int)(current_w/Cfont.sizeX);	 // scaled screen width
		// Print the version.
		if (Cfont.Filter) 
		{   
           Cfont.Filter(VERSIONTEXT);
        }
	/*	else// if(consoleShadowText)
		{
			// Draw a shadow.
			glColor3f(0, 0, 0);
			Cfont.uiTextOut(VERSIONTEXT,ssw - Cfont.Width(VERSIONTEXT)-20,
				(int)(y/Cfont.sizeY + 1));
		}*/
		glColor4f(CcolYellow[0], CcolYellow[1], CcolYellow[2], closeFade);
		Cfont.uiTextOut(VERSIONTEXT, ssw - Cfont.Width(VERSIONTEXT)-20,
			(int)(y/Cfont.sizeY));
	}

	glColor4f(1, 1, 1, closeFade);

	// The text in the console buffer will be drawn from the bottom up (!).

	for(i = bPos-bLineOff-1, y = ConsoleY*gtosMulY-fontScaledY*2; 
		i >= 0 && i < bufferLines && y > -fontScaledY; i--)
	{
		cbline_t *line = cbuffer + i;
		if(line->flags & CBLF_RULER)
		{
			// Draw a ruler here, and nothing else.
			Con_DrawRuler2((int)(y/Cfont.sizeY), Cfont.height, closeFade,
				(int)(current_w/Cfont.sizeX));
		}
		else
		{
			memset(buff, 0, sizeof(buff));
			strncpy(buff, line->text, 255);

			x = line->flags & CBLF_CENTER? (current_w/Cfont.sizeX 
				- Cfont.Width(buff))/2 : 2;

			if(Cfont.Filter) 
			{
				Cfont.Filter(buff);
            }
			else 
           // if(consoleShadowText)
			{
				// Draw a shadow.
				glColor3f(0, 0, 0);
				Cfont.uiTextOut(buff, (int)(x+2), (int)(y/Cfont.sizeY+2));
			}
			
			// Set the color.

			if(Cfont.flags & DDFONT_WHITE) // Can it be colored?
			{	
               consoleSetColor(line->flags, closeFade);
            }

            //consoleSetColor(line->flags,CBLF_WHITE);
			Cfont.uiTextOut(buff, (int)x, (int)(y/Cfont.sizeY));
		}		
		// Move up.
		y -= fontScaledY;
	}
	
	// The command line.
	strcpy(buff, ">");
	strcat(buff, cmdLine);
	if(ConsoleTime & 0x10) strcat(buff, "_");

	if(Cfont.Filter) 
	{
		Cfont.Filter(buff);
    }
	else 
//    if(consoleShadowText)
	{
		// Draw a shadow.
		glColor3f(0, 0, 0);
		Cfont.uiTextOut(buff, 4, (int)(2+(ConsoleY*gtosMulY-fontScaledY)/Cfont.sizeY));
	}

	if(Cfont.flags & DDFONT_WHITE)
		glColor4f(CcolYellow[0], CcolYellow[1], CcolYellow[2], closeFade);
	else
		glColor4f(1, 1, 1, closeFade);

	Cfont.uiTextOut(buff, 2, (int)((ConsoleY*gtosMulY-fontScaledY)/Cfont.sizeY));

	// Restore the original matrices.
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
}
/*
static char* szRulerLine ="----------------------------------------------------------------------\n";

//===========================================================================
// Con_AddRuler
//	A ruler line will be added into the console. bPos is moved down by 1.
//===========================================================================
void Con_AddRuler(void)
{
	cbline_t *line = Con_GetBufferLine(bPos++);

	line->flags |= CBLF_RULER;  // mark as ruler line
	
   if (consoleDump) pr_OFdirect(szRulerLine);
}
*/
//=============================================================================
// Message must have been placed in prbuff before calling
//=============================================================================

static void conPrintf(int flags)
{
	static char lbuf[256];	// This is as big as it gets jav from 251 to 256

	uint		i;
	int			lbc;         // line buffer cursor
	cbline_t	*line;

    if (appStatus < 2)
	{
		// Nothing is printed to the console at this time
		pr_OFdirect(prbuff);
		return;
	}

	if (consoleDump) pr_OFdirect(prbuff); // write to RISEN3D.out

	if (flags & CBLF_RULER)
	{
		if (consoleDump) pr_OFrepeat(NULL, "-", 64, 1);
		
		line = Con_GetBufferLine(bPos++);
		line->flags |= CBLF_RULER;  // mark as ruler line
		flags &= ~CBLF_RULER;
	}

	// We have the text we want to add in the buffer in prbuff.
	line = Con_GetBufferLine(bPos);	// Get a pointer to the current line.

	line->flags = flags;
	
    memset(lbuf, 0, maxLineLen+1);	

	for(i=0, lbc=0; i < strlen(prbuff); i++)
	{
		if(prbuff[i] == '\n' || lbc+line->len >= maxLineLen)	// A new line?
		{
			// Set the line text.
			addLineText(line, lbuf);
			
            // Clear the line write buffer.
			memset(lbuf, 0, maxLineLen+1);
			lbc = 0;
			
            // Get the next line.
			line = Con_GetBufferLine(++bPos);
			line->flags = flags;

			// Newlines won't get in the buffer at all.
			if(prbuff[i] == '\n') continue;
		}
		lbuf[lbc++] = prbuff[i];
	}
	// Something still in the write buffer?

	if(lbc) addLineText(line, lbuf);

	// Now that something new has been printed, it will be shown.

	bLineOff = 0;

	// Check if there are too many lines.

	if(bufferLines > MAXBUFFERLINES)
	{
		int rev = bufferLines - MAXBUFFERLINES;

		// The first 'rev' lines get removed.

		for(i = 0; (int) i < rev; i++ ) cbuffer[i].text = 0;

		memmove(cbuffer, cbuffer + rev, sizeof(cbline_t)*(bufferLines-rev));

    	cbuffer = (cbline_t*)M_Realloc(cbuffer, sizeof(cbline_t)*(bufferLines-=rev), 17); // jav
	
    	// Move the current position.
		bPos -= rev;
	}
}


//=============================================================================
// As you can see, several commands can be handled inside one command function.
//=============================================================================
int CCmdConsole(int argc, char **argv)
{
	if(!_stricmp(argv[0], "help"))
	{
		if(argc == 2)
		{
			int	i;
			if(!_stricmp(argv[1], "(what)"))
			{
				Con_Printf( "You've got to be kidding!\n");
				return true;
			}
			// We need to look through the cvars and ccmds to see if there's a match.
			for(i=0; i<numCCmds; i++)
			{	
                if(!_stricmp(argv[1], ccmds[i].name))
				{
					Con_Printf("%s\n", ccmds[i].help);
					return true;
				}
            }

			for(i=0; i<numCVars; i++)
			{	
                if(!_stricmp(argv[1], cvars[i].name))
				{
					Con_Printf("%s\n", cvars[i].help);
					return true;
				}
            }

			Con_Printf("There's no help about '%s'.\n", argv[1]);
		}
		else
		{
			Con_FPrintf(CBLF_RULER | CBLF_YELLOW | CBLF_CENTER, 
				"-=-  R I S E N 3 D  Console -=-\n");	// R3DVER - Risen3D
			Con_Printf("Keys:\n");
			Con_Printf("Tilde            Open/close the console\n");
			Con_Printf("Shift-Tilde      Switch between half and full screen mode\n");
			Con_Printf("PageUp/Down      Page up/down a page lines\n");   // jav was Scroll up/down two lines
			Con_Printf("CursorUp/Down    Scroll up/down 1 line\n");       // jav
			Con_Printf("CursorLeft/Right Buffer clear/restore\n");     // jav
			Con_Printf("Ins/Del          Move console window up/down one line\n");
			Con_Printf("Shift-Ins/Del    Move console window three lines at a time\n");
			Con_Printf("Home             Jump to the beginning of the buffer\n");
			Con_Printf("End              Jump to the end of the buffer\n");
			Con_Printf("\n");
			Con_Printf("Type \"listcmds\" to see a list of available commands\n");
			Con_Printf("Type \"help (what)\" to see information about (what)\n");
			Con_FPrintf(CBLF_RULER, "\n");
		}
	}
	else if(!_stricmp(argv[0], "clear"))
	{
		Con_ClearBuffer();
	}
	return true;	
}

//=============================================================================
//
//=============================================================================
int CCmdListCmds(int argc, char **argv)
{
	int		i;

	Con_Printf("Console commands:\n");

	for(i = 0; i < numCCmds; i++)
	{
		if(argc > 1) // Is there a filter?
		{  
           if(strnicmp(ccmds[i].name, argv[1], strlen(argv[1])))
		   {	
              continue;
           }
        }

		Con_Printf( "  %s\n", ccmds[i].name);
	}
	return true;
}

//=============================================================================
//
//=============================================================================
int CCmdListVars(int argc, char **argv)
{
	int		i;

	Con_Printf("Console variables:\n");
	for(i = 0; i < numCVars; i++)
	{
		if(cvars[i].flags & CVF_HIDE) continue;
		if(argc > 1) // Is there a filter?
			if(strnicmp(cvars[i].name, argv[1], strlen(argv[1])))
				continue;
		printcvar(cvars+i, "  ");
	}
	return true;
}

//=============================================================================
//
//=============================================================================
int CCmdListAliases(int argc, char **argv)
{
	int		i;

	Con_Printf("Aliases:\n");
	for(i=0; i<numCAliases; i++)
	{
		if(argc > 1) // Is there a filter?
		{ 
           if(strnicmp(caliases[i].name, argv[1], strlen(argv[1])))
		   {  continue;
           }
        }

		Con_Printf( "  %s == %s\n", caliases[i].name, caliases[i].command);
	}
	return true;
}

//=============================================================================
//
//=============================================================================
int CCmdVersion(int argc, char **argv)
{
	Con_Printf( "  "VERSIONTEXT"  ("__DATE__")\n"); //R3DVER
	Con_Printf( "  Risen3D is based on Doomsday V1.7.8 dated 26th Feb 2003\n");
	Con_Printf( "  Compiler: Microsoft Visual C/C++ 6.0 (SP3)\n");
	return true;
}

//=============================================================================
//
//=============================================================================
int CCmdQuit(int argc, char **argv)
{
	// No questions asked.
	Sys_Quit(false);
	return true;
}

//=============================================================================
//
//=============================================================================
int CCmdListFiles(int argc, char **argv)
{
	extern int           numfilenameRecords;
	extern filerecord_t *filenameRecords;
	int		i;
	
	for(i = 0; i < numfilenameRecords; i++)
	{
		Con_Printf("%s (%d lump%s)",
    			filenameRecords[i].filename,
	    		filenameRecords[i].numlumps,
		    	plural(filenameRecords[i].numlumps));

		if(filenameRecords[i].handle == IWADhandle)	
			Con_Printf(" [%08x]", pwadchecksum[0]);

		Con_Printf("\n");
	}

	Con_Printf("Total: %d lumps in %d files.\n", numlumps, numfilenameRecords);
	return true;
}

//=============================================================================
//
//=============================================================================
int CCmdBackgroundTurn(int argc, char **argv)
{
	if(argc != 2)
	{
		Con_Printf( "Usage: bgturn (speed)\n");
		Con_Printf( "Negative speeds are allowed. Default: 20.\n");
		Con_Printf( "Current bgturn = %d.\n", consoleTurn);
		return true;
	}
	consoleTurn = atoi(argv[1]);
	if(!consoleTurn) funnyAng = 0;
	return true;
}

//=============================================================================
//
//=============================================================================
int CCmdDump(int argc, char **argv)
{
	char fname[100];
	FILE *file;
	int lump;
	byte *lumpPtr;

	if(argc != 2)
	{
		Con_Printf( "Usage: dump (name)\n");
		Con_Printf( "Writes out the specified lump to (name).dum.\n");
		return true;		
	}
	
    if( ( lump = W_CheckNumForName(argv[1])) == -1)
	{
		Con_Printf( "No such lump.\n");
		return false;
	}
	
    if ( (lumpPtr = (byte*)M_CacheLumpNum(lump, NULL, false)) == NULL)
		return false;

	sprintf(fname, "%s.dum", argv[1]);

	if( (file = fopen(fname, "wb")) == NULL) 
	{
		Con_Printf("Couldn't open %s for writing. %s\n", fname, strerror(errno));
		M_Free(lumpPtr);
		return false;
	}
	
    fwrite(lumpPtr, 1, lumpinfo[lump].size, file);
	fclose(file);
	M_Free(lumpPtr);

	Con_Printf( "%s dumped to %s.\n", argv[1], fname);
	return true;
}

//===========================================================================
//	Aliases will be saved to the config file.
//===========================================================================
int CCmdAlias(int argc, char **argv)
{
	if(argc != 3 && argc != 2)
	{
		Con_Printf("Usage: %s (alias) (cmd)\n", argv[0]);
		Con_Printf("Example: alias bigfont \"font size 3\".\n");
		Con_Printf("Use %%1-%%9 to pass the alias arguments to the command.\n");
		return true;
	}
	Con_Alias(argv[1], argc==3? argv[2] : NULL);
	if(argc != 3)
		Con_Printf("Alias '%s' deleted.\n", argv[1]);
	return true;
}

//=============================================================================
//
//=============================================================================
int CCmdParse(int argc, char **argv)
{
	int		i;

	if(argc == 1)
	{
		Con_Printf("Usage: %s (file) ...\n", argv[0]);
		return true;
	}
	for(i = 1; i < argc; i++)
	{
		Con_Printf("Parsing %s.\n", argv[i]);
		Con_ParseCommands(argv[i], false);
	}
	return true;
}

//=============================================================================
//
//=============================================================================
int CCmdDeleteBind(int argc, char **argv)
{
	int		i;

	if(argc == 1)
	{
		Con_Printf( "Usage: %s (cmd) ...\n", argv[0]);
		return true;
	}
	for(i=1; i<argc; i++) B_ClearBinding(argv[i]);
	return true;
}	

//=============================================================================
//
//=============================================================================
int CCmdWait(int argc, char **argv)
{
	int off;

	if(argc != 3)
	{
		Con_Printf( "Usage: %s (tics) (cmd)\n", argv[0]);
		Con_Printf( "For example, '%s 35 \"echo End\"'.\n", argv[0]);
		return true;
	}
	off = atoi(argv[1]);
	if(off < 0) off = 0;
	SplitIntoSubCommands(argv[2], off);
	return true;
}

//=============================================================================
//
//=============================================================================
int CCmdRepeat(int argc, char **argv)
{
	int count, off, interval;

	if(argc != 4)
	{
		Con_Printf("Usage: %s (count) (interval) (cmd)\n", argv[0]);
		Con_Printf("For example, '%s 10 35 \"screenshot\".\n", argv[0]);
		return true;
	}
	count = atoi(argv[1]);
	interval = atoi(argv[2]);
	off = 0;
	while(count-- > 0)
	{
		off += interval;
		SplitIntoSubCommands(argv[3], off);
	}
	return true;
}

//=============================================================================
//
//=============================================================================
int CCmdEcho(int argc, char **argv)
{
	int		i;

	for(i=1; i<argc; i++) Con_Printf( "%s\n", argv[i]);
	return true;
}

//=============================================================================
// Rather messy, wouldn't you say?
//=============================================================================
int CCmdAddSub(int argc, char **argv)
{
	RDBOOL	force = false, incdec;
	float	val, mod = 0;
	cvar_t	*cvar;

	incdec = !_stricmp(argv[0], "inc") || !_stricmp(argv[0], "dec");

	if (argc == 1)
	{
		Con_Printf( "Usage: %s (cvar) %s(force)\n", argv[0], incdec ? "" : "(val) ");
		Con_Printf( "Use force to make cvars go off limits.\n");
		return true;
	}

	if ( !(cvar = Con_GetVariable(argv[1])) ) return false;

	if ((incdec && argc >= 3) || (!incdec && argc >= 4))
	{
		force = !_stricmp(argv[incdec? 2 : 3], "force");
	}

	if     (cvar->type == CVT_BYTE ) val = *(byte*) cvar->ptr;
	else if(cvar->type == CVT_INT  ) val = (float)(*(int*) cvar->ptr);
	else if(cvar->type == CVT_FLOAT) val = *(float*) cvar->ptr;
	else
		return false;	// not recognised

	if      (!_stricmp(argv[0], "inc")) mod =  1;
	else if (!_stricmp(argv[0], "dec")) mod = -1;
	else
		mod = (float)(strtod(argv[2], NULL));

	if (!_stricmp(argv[0], "sub")) mod = -mod;

	val += mod;

	if(!force)
	{
		if(!(cvar->flags & CVF_NO_MAX) && val > cvar->max) val = cvar->max;
		if(!(cvar->flags & CVF_NO_MIN) && val < cvar->min) val = cvar->min;
	}

	if      (cvar->type == CVT_BYTE ) *(byte* ) cvar->ptr = (byte)val;
	else if (cvar->type == CVT_INT  ) *(int*  ) cvar->ptr = (int)val;
	else if (cvar->type == CVT_FLOAT) *(float*) cvar->ptr = val;

	return true;
}

int Con_ParseCommands(char *fileName, RDBOOL silent)
{
	static char	buff[512];
	DFILE	*file;
	int		line = 0;

	_chdir(ddRuntimeDir.path);

	// Open the file.
	if((file = F_Open(fileName, "ft")) == NULL) return false;

	// Each line in this file should be a command (or a comment).
	for(;;)
	{
		M_ReadLine(buff, 512, file); line++;

		if (strlen(buff) >= 511) Con_Error("%s is corrupt - delete and restart", fileName);

		if (!M_IsComment(buff))
		{
			// If a start up config then execute the commands silently.
			if(!Con_Execute(buff, silent))
				Con_Message( "%s(%d): error executing command\n  \"%s\"\n", fileName, line, buff);
		}
	
		if(deof(file)) break;
	}

	F_Close(file);	
	return true;
}


// Fog settings v220-3 revised
int CCmdFog(int argc, char **argv)
{
	int		i;

	if (R3D_edit || (R3D_ScriptedMap && !ge_inf.devmap) )
	{
		Con_Printf("Fog values cannot be set in edit or scripted mode\n");
		return false;
	}

	if (gamestate != GS_LEVEL)
	{
		Con_Printf("Fog values can only be set in game\n");
		return false;
	}

	if (R3D_ScriptedMap && fogval.isset)
	{
		Con_Printf("Fog values cannot be changed in a scripted map whilst fog is in use\n");
		return false;
	}

	if (r_texglow) Con_Printf( "NOTE: when fog is active glow and detail textures are disabled\n");

	if (argc == 1)
	{
		Con_Printf( "Usage: %s (cmd) (args)\n", argv[0]);
		Con_Printf( "(cmd)    off, reset, set, mode, color, start, end, density.\n");
		Con_Printf( "reset    resets to using mapinfo values\n");
		Con_Printf( "set      (arg) MAP or OUTSIDE (default values are set for arg sent)\n");
		Con_Printf( "mode     (arg) LINEAR or EXP or EXP2\n");
		Con_Printf( "color    (args) R G B (where R, G and B are values in the range 0-255)\n");
		Con_Printf( "Example: fog set outside (will set default outside fog)");
		Con_Printf( "NOTE: Density has no effect with LINEAR mode\n");
		Con_Printf( "      Start and end has no effect with EXP modes\n");
		return true;		
	}

	if (!_stricmp(argv[1], "off"))
	{
		GL_UseFog(false);
		underwaterfog = outsidefog = false;
		Con_Printf("Fog disabled\n");
		shortListQuads = R3D_slopetypes;
	}
	else if(!_stricmp(argv[1], "reset"))
	{
		GL_UseFog(false);
		MI_ActionMapInfo(P_GetCurrentMapid());

		shortListQuads = outsidefog || R3D_slopetypes;
	}
	else if (!_stricmp(argv[1], "set") && argc == 3)
	{
		byte col[4] = { 191, 191, 191, 255 };

		if (!_stricmp(argv[2], "map") || (underwaterfog = !_stricmp(argv[2], "underwater")) )
		{
			GL_FogvColor(col);
			GL_FogvDensity(0.0005f);
			GL_FogvMode(GL_EXP);
			outsidefog = false;
			shortListQuads = R3D_slopetypes;

			if (!underwaterfog)
				GL_UseFog(true);
		}
		else if (!_stricmp(argv[2], "outside"))
		{
			GL_FogvColor(col);
			GL_FogvStart(200.0f);
			GL_FogvEnd(2800.0f);
			GL_FogvMode(GL_LINEAR);
			underwaterfog = false;
			outsidefog = shortListQuads = true;
			GL_UseFog(true);
		}
		else return false;
	}
	else if(!_stricmp(argv[1], "mode") && argc == 3)
	{
		if (outsidefog)
		{
			Con_Printf( "Map has outside fog. Mode cannot be changed from linear.\n");
		}
		else if(!_stricmp(argv[2], "linear"))	
		{
			GL_FogvMode(GL_LINEAR);
		}
		else if(!_stricmp(argv[2], "exp")) 
		{
			GL_FogvMode(GL_EXP);
		}
		else if(!_stricmp(argv[2], "exp2")) 
		{
			GL_FogvMode(GL_EXP2);
		}
		else return false;
	}
	else if(!_stricmp(argv[1], "color") && argc == 5)
	{
		byte col[4];

		for(i = 0; i < 3; i++)
			col[i] = (byte)(strtol(argv[2 + i], NULL, 0));

		GL_FogvColor(col);
		Con_Printf( "Fog color set.\n");
	}	
	else if(!_stricmp(argv[1], "start") && argc == 3)
	{
		GL_FogvStart((float)(strtod(argv[2], NULL)));
		Con_Printf( "Fog start distance set.\n");
	}
	else if(!_stricmp(argv[1], "end") && argc == 3)
	{
		GL_FogvEnd((float)(strtod(argv[2], NULL)));
		Con_Printf( "Fog end distance set.\n");
	}
	else if(!_stricmp(argv[1], "density") && argc == 3)
	{
		GL_FogvDensity((float)(strtod(argv[2], NULL)));
		Con_Printf( "Fog density set.\n");
	}
	else
		return false;

	return true;
}

#pragma optimize("", on)
