//-----------------------------------------------------------------------------
//
// Copyright (C) 2005 Jaakko Kernen
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: dd_input.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified for R3D, CCmdDumpKeyMap() and CCmdKeyMap() removed
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:	Primary game loop
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <ctype.h>
#include <string.h>

#include "de_base.h"
#include "de_console.h"
#include "de_system.h"
#include "de_misc.h"
#include "de_ui.h"

#include "net_main.h"
#include "r3d_compat.h"

// MACROS ------------------------------------------------------------------

#define MAX_DOWNKEYS	16		// Most keyboards support 6 or 7. Using 16 is
								// to make sure the key repeater gets flushed

#define SC_RSHIFT       0x36
#define SC_LSHIFT       0x2a
#define SC_UPARROW      0x48
#define SC_DOWNARROW    0x50
#define SC_LEFTARROW    0x4b
#define SC_RIGHTARROW   0x4d

#define CLAMP(x) DD_JoyAxisClamp(&x) //x = (x < -100? -100 : x > 100? 100 : x)

// TYPES -------------------------------------------------------------------

typedef struct repeater_s
{
	int key;			// The H2 key code (0 if not in use).
	int	timer;			// How's the time?
	int count;			// How many times has been repeated?
} repeater_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

int D_PrivilegedResponder(event_t *event);
RDBOOL M_Responder (event_t *ev);
RDBOOL G_Responder (event_t* ev);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

static void DD_ReadKeyboard(void);
static void DD_ReadMouse(void);
static void DD_ReadJoystick(void);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern RDBOOL waspaused;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

event_t		events[MAXEVENTS];

int			mouseInverseY = false;
int			mouseWheelSensi = 10;	// I'm shooting in the dark here.

// The initial and secondary repeater delays (tics).
int			repWait1 = 15, repWait2 = 4; 
int			mouseDisableX = false, mouseDisableY = false;
int			showScanCodes = false;
RDBOOL		shiftDown = false, altDown = false;

// A customizable mapping of the scantokey array.
char		keyMapPath[256] = ">Data\\KeyMaps\\";
byte		keyMappings[256];
byte		shiftKeyMappings[256], altKeyMappings[256];

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// See DIK table in DINPUT.H (headed with DIK_ESCAPE - use goto definition)
static byte scantokey[256] =	
{
//	0				1			2				3				4			5					6				7
//	8				9			A				B				C			D					E				F
// 0 ---------------------------------------------------------------------------------------------------------------------------
	0  ,			DDKEY_ESCAPE,'1',			'2',			'3',		'4',				'5',			'6',
	'7',			'8',		'9',			'0',			'-',		'=',				DDKEY_BACKSPACE,DDKEY_TAB,	// 0
// 1 ---------------------------------------------------------------------------------------------------------------------------
	'q',			'w',		'e',			'r',			't',		'y',				'u',			'i',
	'o',			'p',		'[',			']',	DDKEY_ENTER,		DDKEY_RCTRL,		'a',			's',		// 1
// 2 ---------------------------------------------------------------------------------------------------------------------------
	'd',			'f',		'g',			'h',			'j',		'k',				'l',			';',
	0  ,			DDKEY_GRAVE	,DDKEY_RSHIFT,	0  ,			'z',		'x',				'c',			'v',		// 2
// 3 ---------------------------------------------------------------------------------------------------------------------------
	'b',			'n',		'm',			',',			'.',		'/',				DDKEY_RSHIFT,	'*',
	DDKEY_RALT,		' ',		DDKEY_CAPSLOCK,	DDKEY_F1,		DDKEY_F2,	DDKEY_F3,			DDKEY_F4,		DDKEY_F5,   // 3
// 4 ---------------------------------------------------------------------------------------------------------------------------
	DDKEY_F6,		DDKEY_F7,	DDKEY_F8,		DDKEY_F9,		DDKEY_F10,	DDKEY_NUMLOCK,		DDKEY_SCROLL,	DDKEY_NUMPAD7,
	DDKEY_NUMPAD8,	DDKEY_NUMPAD9, DDKEY_SUBTRACT,	DDKEY_NUMPAD4,DDKEY_NUMPAD5, DDKEY_NUMPAD6,	DDKEY_ADD	,	DDKEY_NUMPAD1, // 4
// 5 ---------------------------------------------------------------------------------------------------------------------------
	DDKEY_NUMPAD2,	DDKEY_NUMPAD3, DDKEY_NUMPAD0, DDKEY_DECIMAL,0  ,		0  , 				0  ,			DDKEY_F11,
	DDKEY_F12,		0  ,		0  ,			0  ,			0  ,		0  ,				0  ,			0,			// 5
// 6 ---------------------------------------------------------------------------------------------------------------------------
	0  ,			0  ,		0  ,			0  ,			0  ,		0  ,				0  ,			0,
	0  ,			0  ,		0  ,			0  ,			0,			0  ,				0  ,			0,			// 6
// 7 ---------------------------------------------------------------------------------------------------------------------------
	0  ,			0  ,		0  ,			0  ,			0,			0  ,				0  ,			0,
	0  ,			0  ,		0  ,			0,				0  ,		0  ,				0  ,			0,			// 7
// 8 ---------------------------------------------------------------------------------------------------------------------------
	0  ,			0  ,		0  ,			0  ,			0,			0  ,				0  ,			0,
	0,				0  ,		0  ,			0,				0  ,		0  ,				0  ,			0,			// 8
// 9 ---------------------------------------------------------------------------------------------------------------------------
	0  ,			0  ,		0  ,			0  ,			0,			0  ,				0  ,			0,
	0  ,			0  ,		0  ,			0,				DDKEY_ENTER, DDKEY_RCTRL,		0  ,			0,			// 9
// A ---------------------------------------------------------------------------------------------------------------------------
	0  ,			0  ,		0  ,			0  ,			0,			0  ,				0  ,			0,
	0  ,			0  ,		0  ,			0,				0  ,		0  ,				0  ,			0,			// A
// B ---------------------------------------------------------------------------------------------------------------------------
	0  ,			0  ,		0  ,			0  ,			0,DDKEY_DIVIDE ,				0  ,			0,
	DDKEY_RALT,		0  ,		0  ,			0,				0  ,		0  ,				0  ,			0,			// B
// C ---------------------------------------------------------------------------------------------------------------------------
	0  ,			0  ,		0  ,			0  ,			0,			0,					0  ,			DDKEY_HOME,
	DDKEY_UPARROW,	DDKEY_PGUP,	0  ,			DDKEY_LEFTARROW,0  ,		DDKEY_RIGHTARROW,	0  ,			DDKEY_END,	// C
// D
	DDKEY_DOWNARROW,DDKEY_PGDN, DDKEY_INS,		DDKEY_DEL,		0,			0  ,				0  ,			0,
	0  ,			0  ,		0  ,			0,				0  ,		0  ,				0  ,			0,			// D
// E ---------------------------------------------------------------------------------------------------------------------------	
	0  ,			0  ,		0  ,			0  ,			0,			0  ,				0  ,			0,
	0  ,			0  ,		0  ,			0,				0  ,		0  ,				0  ,			0,			// E
// F ---------------------------------------------------------------------------------------------------------------------------
	0  ,			0  ,		0  ,			0  ,			0,			0  ,				0  ,			0,
	0  ,			0  ,		0  ,			0,				0  ,		0  ,				0  ,			0			// F
};

static char defaultShiftTable[96] =	// Contains characters 32 to 127.
{
/* 32 */	' ', 0, 0, 0, 0, 0, 0, '"',
/* 40 */	0, 0, 0, 0, '<', '_', '>', '?', ')', '!',
/* 50 */	'@', '#', '$', '%', '^', '&', '*', '(', 0, ':',
/* 60 */	0, '+', 0, 0, 0, 'a', 'b', 'c', 'd', 'e',
/* 70 */	'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o',
/* 80 */	'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y',
/* 90 */	'z', '{', '|', '}', 0, 0, 0, 'A', 'B', 'C',
/* 100 */	'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
/* 110 */	'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W',
/* 120 */	'X', 'Y', 'Z', 0, 0, 0, 0, 0
};

static repeater_t keyReps[MAX_DOWNKEYS];
static int		oldMouseButtons = 0;
static int		oldJoyBState = 0;
static float	oldPOV = IJOY_POV_CENTER;

static int		eventhead;
static int		eventtail;

// CODE --------------------------------------------------------------------

//===========================================================================
// DD_DefaultKeyMapping
//===========================================================================
void DD_DefaultKeyMapping(void)
{
	int i;

	for(i = 0; i < 256; i++)
	{
		keyMappings[i] = scantokey[i];
		shiftKeyMappings[i] = (byte)(i >= 32 && i <= 127 
			&& defaultShiftTable[i - 32]? defaultShiftTable[i - 32] : i);
		altKeyMappings[i] = (byte)i;
	}
}

//===========================================================================
// DD_InitInput
//	Initializes the key mappings to the default values.
//===========================================================================
void DD_InitInput(void)
{
	DD_DefaultKeyMapping();
}

//===========================================================================
// DD_KeyOrCode
//===========================================================================
int DD_KeyOrCode(char *token)
{
	char *end = M_FindWhite(token);

	if(end - token > 1)
	{
		// Longer than one character, it must be a number.
		// Rule out base 8
		return strtol(token, 0, !strnicmp(token, "0x", 2) ? 16 : 10);
	}
	// Direct mapping.
	return (unsigned char) *token;
}

//==========================================================================
// DD_ProcessEvents
//	Send all the events of the given timestamp down the responder chain.
//==========================================================================
void DD_ProcessEvents(void)
{
	static int	altActive = false;
	event_t		*ev = NULL;

	if (paused) return;

	DD_ReadMouse();
	DD_ReadJoystick();
	DD_ReadKeyboard();

	for(; eventtail != eventhead; eventtail = (++eventtail)&(MAXEVENTS-1))
	{
		ev = &events[eventtail];

		// Track the state of Shift and Alt.
		if(ev->data1 == DDKEY_RSHIFT)
		{
			if(ev->type == ev_keydown) shiftDown = true;
			else if(ev->type == ev_keyup) shiftDown = false;
		}

		if (ev->data1 == DDKEY_RALT)
		{
			if(ev->type == ev_keydown)
			{
				altDown = true;
				altActive = true;
			}
			else if(ev->type == ev_keyup)
			{
				altDown = false;
			}
		}

		if (altActive)
		{
			if (   ev->data1 == DDKEY_ESCAPE
				|| ev->data1 == DDKEY_ENTER)
			{
				continue;
			}

			if (ev->data1 == DDKEY_TAB)
			{
				if (ui_active && !paused) paused = true;
				if (automap_active && !paused) paused = true;	// v220-33
				continue;
			}

			// v220-31 deal with exit from ui etc.
			if (ev->data1 == DDKEY_F4) appTerminate = true;
		}

		// v220-5 discard all events
		if (waspaused) continue;

		// Does the special responder use this event?
		if (D_PrivilegedResponder(ev)) continue;
		if (UI_Responder (ev)) continue;		// Control Panel
		if (Con_Responder(ev)) continue;		// Console
		if (M_Responder  (ev)) continue;		// Doom Menus
		if (G_Responder  (ev)) continue;		// Game responder
		// if G_Responder returned true then bindings can't be used
		// (e.g. when chatting).
		B_Responder(ev);						// Bindings
	}

	// v220-5 if app was swapped clear the keys, events and actions
	if (waspaused)
	{
		memset(keyReps, 0, sizeof(keyReps));
		memset(events , 0, sizeof(events));
		Con_ClearActions();
		altActive = altDown = false;
		waspaused = false;
	}

	if (altActive && ev && ev->type == ev_keyup) altActive = false;
}

//==========================================================================
// DD_PostEvent
//	Called by the I/O functions when input is detected.
//==========================================================================
void DD_PostEvent(event_t *ev)
{
	if (paused || waspaused) return;

	events[eventhead] = *ev;
	eventhead = (++eventhead)&(MAXEVENTS-1);
}

//===========================================================================
// DD_ScanToKey
//===========================================================================
byte DD_ScanToKey(byte scan)
{
	return keyMappings[scan];
}

//===========================================================================
// DD_ModKey
//	Apply all active modifiers to the key.
//===========================================================================
byte DD_ModKey(byte key)
{
	if(shiftDown) key = shiftKeyMappings[key];
	if(altDown  ) key = altKeyMappings[key];
	return key;
}

//===========================================================================
// DD_ClearKeyRepeaters
//===========================================================================
void DD_ClearKeyRepeaters(void)
{
	memset(keyReps, 0, sizeof(keyReps));
}

//===========================================================================
// DD_ReadKeyboard
//===========================================================================
static void DD_ReadKeyboard(void)
{
	repeater_t	*rep;
	event_t		ev;
	keyevent_t	keyevs[KEYBUFSIZE];
	int			i, k, numkeyevs;

	// Check the repeaters.
	ev.type = ev_keyrepeat;

	k = systics;	// The current time.

	for (i = 0; i < MAX_DOWNKEYS; i++)
	{
		if ( !(rep = &keyReps[i])->key ) continue;

		ev.data1 = rep->key;

		if(!rep->count)
		{
			// The first time.
			if (k - rep->timer >= repWait1)
			{
				rep->count++;
				rep->timer += repWait1;
				DD_PostEvent(&ev);
			}
		}
		else
		{
			while(k - rep->timer >= repWait2)
			{
				rep->count++;
				rep->timer += repWait2;
				DD_PostEvent(&ev);				
			}
		}
	}

	// Read the keyboard events.
	numkeyevs = I_GetKeyEvents(keyevs);

	// Translate to game keys.
	for(i = 0; i < numkeyevs; i++)
	{
		keyevent_t *ke = keyevs + i;
		
		// Check the type of the event.
		if (ke->event == IKE_KEY_DOWN)		// Key pressed?
			ev.type = ev_keydown;
		else if (ke->event == IKE_KEY_UP)	// Key released?
			ev.type = ev_keyup;
		
		// Use the table to translate the scancode to a ddkey.		
		ev.data1 = DD_ScanToKey(ke->code);

		// Should we print a message in the console?
		if (showScanCodes && ev.type == ev_keydown)
			Con_Printf("Scancode: %i (0x%x)\n", ev.data1, ev.data1);

		// Maintain the repeater table.
		if(ev.type == ev_keydown)
		{
			// Find an empty repeater.
			for(k = 0; k < MAX_DOWNKEYS; k++)
			{
				if(!keyReps[k].key)
				{
					keyReps[k].key = ev.data1;
					keyReps[k].timer = systics;
					keyReps[k].count = 0;
					break;
				}
			}
		}
		else if(ev.type == ev_keyup)
		{
			// Clear any repeaters with this key.
			for(k = 0; k < MAX_DOWNKEYS; k++)
			{
				if(keyReps[k].key == ev.data1)
					keyReps[k].key = 0;
			}
		}

		// Post the event.
		DD_PostEvent(&ev);
	}
}

void flushMouse()
{
	mousestate_t	mouse;

	if (!mouseInited) return;

	I_GetMouseState(&mouse, true);
}

//===========================================================================
// DD_ReadMouse
//	Mouse events.
//===========================================================================
static void DD_ReadMouse(void)
{
	event_t			ev;
	mousestate_t	mouse;
	int				change;

	if (!mouseInited) return;

	// Get the mouse state.
	I_GetMouseState(&mouse, false);

	ev.type = ev_mouse;
	ev.data1 = mouse.x;
	ev.data2 = mouse.y;
	ev.data3 = mouse.z;
	if(!ui_active)
	{
		if(mouseDisableX) ev.data1 = 0;
		if(mouseDisableY) ev.data2 = 0;
		if(!mouseInverseY) ev.data2 = -ev.data2;
	}
	DD_PostEvent (&ev);

	// Insert the possible mouse Z axis into the button flags.
	if(abs(ev.data3) >= mouseWheelSensi)
	{
		mouse.buttons |= ev.data3 > 0? DDMB_MWHEELUP : DDMB_MWHEELDOWN;
	}

	// Check the buttons and send the appropriate events.
	change = oldMouseButtons ^ mouse.buttons; // The change mask.
	// Send the relevant events.
	if((ev.data1 = mouse.buttons & change))
	{
		ev.type = ev_mousebdown;
		DD_PostEvent(&ev);
	}
	if((ev.data1 = oldMouseButtons & change))
	{
		ev.type = ev_mousebup;
		DD_PostEvent(&ev);
	}
	oldMouseButtons = mouse.buttons;
}

//===========================================================================
// DD_JoyAxisClamp
//	Applies the dead zone and clamps the value to -100...100.
//===========================================================================
void DD_JoyAxisClamp(int *val)
{
	if(abs(*val) < cfg.joyDeadzone)
	{
		// In the dead zone, just go to zero.
		*val = 0;
		return;
	}
	// Remove the dead zone.
	*val += *val > 0? -cfg.joyDeadzone : cfg.joyDeadzone;
	// Normalize.
	*val = (int)((float)(*val) * 100.0f/(100 - cfg.joyDeadzone));	
	// Clamp.
	if(*val > 100) *val = 100;
	if(*val < -100) *val = -100;
}

//===========================================================================
// DD_ReadJoystick
//===========================================================================
static void DD_ReadJoystick(void)
{
	event_t		ev;
	joystate_t	state;
	int			i, bstate;
	int			div;

	if (!joyInited) return;

	I_GetJoystickState(&state);

	bstate = 0;
	// Check the buttons.
	for(i = 0; i < IJOY_MAXBUTTONS; i++) 
		if(state.buttons[i]) bstate |= 1<<i; // Set the bits.

	// Check for button state changes. 
	i = oldJoyBState ^ bstate; // The change mask.
	// Send the relevant events.
	if((ev.data1 = bstate & i))
	{
		ev.type = ev_joybdown;
		DD_PostEvent(&ev);
	}
	if((ev.data1 = oldJoyBState & i))
	{
		ev.type = ev_joybup;
		DD_PostEvent(&ev);
	}
	oldJoyBState = bstate;

	// Check for a POV change.
	if(state.povAngle != oldPOV)
	{
		if(oldPOV != IJOY_POV_CENTER)
		{
			// Send a notification that the existing POV angle is no
			// longer active.
			ev.type = ev_povup;
			ev.data1 = (int) (oldPOV/45 + .5);	// Round off correctly w/.5.
			DD_PostEvent(&ev);
		}
		if(state.povAngle != IJOY_POV_CENTER)
		{
			// The new angle becomes active.
			ev.type = ev_povdown;
			ev.data1 = (int) (state.povAngle/45 + .5);
			DD_PostEvent(&ev);
		}
		oldPOV = state.povAngle;
	}

	// Send the joystick movement event (XYZ and rotation-XYZ).
	ev.type = ev_joystick;

	div = 100 - cfg.joySensitivity*10;

	// The input code returns the axis positions in the range -10000..10000.
	// The output axis data must be in range -100..100.
	// Increased sensitivity causes the axes to max out earlier.
	// Check that the divisor is valid.
	if(div < 10) div = 10;
	if(div > 100) div = 100;

	ev.data1 = state.axis[0] / div;
	ev.data2 = state.axis[1] / div;
	ev.data3 = state.axis[2] / div;
	ev.data4 = state.rotAxis[0] / div;
	ev.data5 = state.rotAxis[1] / div;
	ev.data6 = state.rotAxis[2] / div;
	CLAMP(ev.data1);
	CLAMP(ev.data2);
	CLAMP(ev.data3);
	CLAMP(ev.data4);
	CLAMP(ev.data5);
	CLAMP(ev.data6);

	DD_PostEvent(&ev);

	// The sliders.
	memset(&ev, 0, sizeof(ev));
	ev.type = ev_joyslider;

	ev.data1 = state.slider[0] / div;
	ev.data2 = state.slider[1] / div;
	CLAMP(ev.data1);
	CLAMP(ev.data2);

	DD_PostEvent(&ev);
}
