//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: dedfile.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  XG stuff removed
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//  Minor editing changes
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:  Engine Definition File Management
//
//-----------------------------------------------------------------------------

#include "dedfile.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "m_winmem.h"

// Helper Routines -------------------------------------------------------

// Returns a pointer to the new block of memory.
void *DED_NewEntries(void **ptr, ded_count_t * cnt, int elem_size, int count)
{
	void *np;

	cnt->num += count;

	if (cnt->num > cnt->max)
	{
		// v220-35 Increase the size of the array.
		// 32 is a trade off between speed (fastest with array doubling) and
		// optimum memory usage (very slow if realloc is needed for each call).
		cnt->max = cnt->num + 32;

		*ptr = M_Realloc(*ptr, elem_size * cnt->max, 22);
	}

	np = (char *) *ptr + (cnt->num - count) * elem_size;
	memset(np, 0, elem_size * count);	// Clear the new entries.
	return np;
}

// Returns a pointer to the new block of memory.
void *DED_NewEntry(void **ptr, ded_count_t * cnt, int elem_size)
{
	return DED_NewEntries(ptr, cnt, elem_size, 1);
}

void DED_DelEntry(int index, void **ptr, ded_count_t * cnt, int elem_size)
{
	if(index < 0 || index >= cnt->num)
		return;					// Can't do that!

	memmove((char *) *ptr + elem_size * index,
			(char *) *ptr + elem_size * (index + 1),
			elem_size * (cnt->num - index - 1));

	if(--cnt->num < cnt->max / 2)
	{
		cnt->max /= 2;
		*ptr = M_Realloc(*ptr, elem_size * cnt->max, 23);
	}
}

void DED_DelArray(void **ptr, ded_count_t * cnt)
{
	M_Free(*ptr);
	*ptr = 0;
	cnt->num = cnt->max = 0;
}

void DED_ZCount(ded_count_t * c)
{
	c->num = c->max = 0;
}

// DED Code --------------------------------------------------------------

void DED_Init(ded_t *ded)
{
	memset(ded, 0, sizeof(*ded));
	ded->version = DED_VERSION;
}

void DED_Destroy(ded_t *ded)
{
	int i;

	M_Free(ded->includes);
	M_Free(ded->flags);
	M_Free(ded->mobjs);
	M_Free(ded->states);
	M_Free(ded->sprites);
	M_Free(ded->lights);
	M_Free(ded->models);
	M_Free(ded->sounds);
	M_Free(ded->music);
	M_Free(ded->mapinfo);
	for(i=0; i<ded->count.text.num; i++) M_Free(ded->text[i].text);
	M_Free(ded->text);
	for(i=0; i<ded->count.tenviron.num; i++) M_Free(ded->tenviron[i].textures);
	M_Free(ded->tenviron);
	for(i=0; i<ded->count.values.num; i++)
	{
		M_Free(ded->values[i].id);
		M_Free(ded->values[i].text);
	}
	M_Free(ded->values);
}

int DED_AddInclude(ded_t *ded, char *inc)
{
	ded_path_t *pth = (ded_path_t*)DED_NewEntry( (void**) &ded->includes, 
		&ded->count.includes, sizeof(ded_path_t));
	strcpy(pth->path, inc);
	return pth - ded->includes;
}

int DED_AddMobj(ded_t *ded, char *idstr)
{
	ded_mobj_t *mo = (ded_mobj_t*)DED_NewEntry( (void**) &ded->mobjs,
		&ded->count.mobjs, sizeof(ded_mobj_t));
	strcpy(mo->id, idstr);
	return mo - ded->mobjs;
}

int DED_AddFlag(ded_t *ded, char *name, int value)
{
	ded_flag_t *fl = (ded_flag_t*)DED_NewEntry( (void**) &ded->flags,
		&ded->count.flags, sizeof(ded_flag_t));
	strcpy(fl->id, name);
	fl->value = value;
	return fl - ded->flags;
}

int DED_AddModel(ded_t *ded, char *spr)
{
	int i;
	ded_model_t *md = (ded_model_t*)DED_NewEntry( (void**) &ded->models,
		&ded->count.models, sizeof(ded_model_t));

	strcpy(md->sprite.id, spr);
	md->interrange[1] = 1;
	md->scale[0] = md->scale[1] = md->scale[2] = 1;
	for(i = 0; i < 4; i++) // Init submodels, too.
	{
		md->sub[i].shinycolor[0] 
			= md->sub[i].shinycolor[1] 
			= md->sub[i].shinycolor[2] = 1;
	}
	return md - ded->models;
}

int DED_AddState(ded_t *ded, char *id)
{
	ded_state_t *st = (ded_state_t*)DED_NewEntry( (void**) &ded->states,
		&ded->count.states, sizeof(ded_state_t));
	strcpy(st->id, id);
	return st - ded->states;
}

int DED_AddSprite(ded_t *ded, const char *name)
{
	ded_sprid_t *sp = (ded_sprid_t*)DED_NewEntry( (void**) &ded->sprites,
		&ded->count.sprites, sizeof(ded_sprid_t));
	strcpy(sp->id, name);
	return sp - ded->sprites;
}

int DED_AddLight(ded_t *ded, const char *stateid)
{
	ded_light_t *light = (ded_light_t*)DED_NewEntry( (void**) &ded->lights,
		&ded->count.lights, sizeof(ded_light_t));
	strcpy(light->state, stateid);
	return light - ded->lights;
}

int DED_AddSound(ded_t *ded, char *id)
{
	ded_sound_t *snd = (ded_sound_t*)DED_NewEntry( (void**) &ded->sounds,
		&ded->count.sounds, sizeof(ded_sound_t));
	strcpy(snd->id, id);
	return snd - ded->sounds;
}

int DED_AddMusic(ded_t *ded, char *id)
{
	ded_music_t *mus = (ded_music_t*)DED_NewEntry( (void**) &ded->music,
		&ded->count.music, sizeof(ded_music_t));
	strcpy(mus->id, id);
	return mus - ded->music;
}

int DED_AddMapInfo(ded_t *ded, char *str)
{
	ded_mapinfo_t *inf = (ded_mapinfo_t*)DED_NewEntry( (void**) &ded->mapinfo,
		&ded->count.mapinfo, sizeof(ded_mapinfo_t));
	strcpy(inf->id, str);
	// Set defaults - will be changed if defined
	inf->gravity = 1.0f;
	inf->sky_height = 0.49f;
	inf->horizon_offset = -0.105f;
	return inf - ded->mapinfo;
}

int DED_AddText(ded_t *ded, char *id)
{
	ded_text_t *txt = (ded_text_t*)DED_NewEntry( (void**) &ded->text,
		&ded->count.text, sizeof(ded_text_t));
	strcpy(txt->id, id);
	return txt - ded->text;
}

int DED_AddTexEnviron(ded_t *ded, char *id)
{
	ded_tenviron_t *env = (ded_tenviron_t*)DED_NewEntry( (void**) &ded->tenviron,
		&ded->count.tenviron, sizeof(ded_tenviron_t));
	strcpy(env->id, id);
	return env - ded->tenviron;
}

int DED_AddValue(ded_t *ded, const char *id)
{
	ded_value_t *val = (ded_value_t*)DED_NewEntry( (void**) &ded->values,
		&ded->count.values, sizeof(ded_value_t));
	if(id)
	{
		val->id = (char*)M_Malloc(strlen(id)+1, 9);
		strcpy(val->id, id);
	}
	return val - ded->values;
}

int DED_AddDetail(ded_t *ded)
{
	ded_detailtexture_t *dtl = (ded_detailtexture_t*)DED_NewEntry( (void**) &ded->dettex,
		&ded->count.details, sizeof(ded_detailtexture_t));
	dtl->det_lumpname[0] = 0;
	dtl->scale = 1;
	dtl->strength = 1;
	return dtl - ded->dettex;
}

int DED_AddPtcGen(ded_t *ded, const char *state)
{
	ded_ptcgen_t *gen = (ded_ptcgen_t*)DED_NewEntry( (void**) &ded->ptcgens,
		&ded->count.ptcgens, sizeof(ded_ptcgen_t));
	strcpy(gen->state, state);
	return gen - ded->ptcgens;
}

int DED_AddFinale(ded_t *ded)
{
	ded_finale_t *fin = (ded_finale_t*)DED_NewEntry( (void**) &ded->finales,
		&ded->count.finales, sizeof(ded_finale_t));
	return fin - ded->finales;
}
