//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2012 G Jackson, Jaakko Kernen, Jack Vermeulen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: dedread.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  modified for use with R3D - script error checking added
//  changes added for fast model loading
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:  Engine Definition file reader
//
//-----------------------------------------------------------------------------

//** At the moment the idea is that a lot of macros are used to read
//** a more or less fixed structure of definitions, and if an error
//** occurs then "goto out_of_here;". It leads to a lot more code 
//** than an elegant parser would require.
//**
//** The current implementation of the reader is a "structural"
//** approach: the definition file is parsed based on the structure 
//** implied by the read tokens. A true parser would have syntax 
//** definitions for a bunch of tokens, and the same parsing rules
//** would be applied for everything.


// HEADER FILES ------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "de_base.h"
#include "de_console.h"
#include "de_system.h"
#include "de_misc.h"
#include "de_refresh.h"
#include "m_winmem.h"
#include "r3d_mapinfo.h"
#include "r3d_compat.h"

// MACROS ------------------------------------------------------------------

#ifndef true
#define true 1
#endif

#ifndef false
#define false 0
#endif

// Some macros.
#define STOPCHAR(x)	(isspace(x) || x == ';' || x == '#' || x == '{' \
					|| x == '}' || x == '=' || x == '"' || x == '*')

#define ISTOKEN(X)	(!_stricmp(token, X))

#define READSTR(X)	if(!ReadString(X, sizeof(X))) { \
					SetError("Syntax error in string value."); \
					retval = false; goto ded_end_read; }

#define MISSING_SC_ERROR	SetError("Missing semicolon."); \
							retval = false; goto ded_end_read; 

#define CHECKSC		ReadToken(); if(!ISTOKEN(";")) { MISSING_SC_ERROR; }

#define FINDBEGIN	while(!ISTOKEN("{") && !dedend) ReadToken(); 
#define FINDEND		while(!ISTOKEN("}") && !dedend) ReadToken(); 

#define ISLABEL(X)	(!_stricmp(label, X))

#define READLABEL	if(!ReadLabel(label)) { retval = false; goto ded_end_read; } \
					if(ISLABEL("}")) break;

#define READLABEL_NOBREAK	if(!ReadLabel(label)) { retval = false; goto ded_end_read; }

#define READBYTE(X)		if(!ReadByte(&X)) { retval = false; goto ded_end_read; }
#define READSHORT(X)	if(!ReadShort(&X)) { retval = false; goto ded_end_read; }
#define READINT(X)		if(!ReadInt(&X, false)) { retval = false; goto ded_end_read; }
#define READUINT(X)		if(!ReadInt(&X, true)) { retval = false; goto ded_end_read; }
#define READFLT(X)		if(!ReadFloat(&X)) { retval = false; goto ded_end_read; }
#define READNBYTEVEC(X,N) if(!ReadNByteVector(X, N)) { retval = false; goto ded_end_read; }

#define RV_BYTE(lab, X)	if(ISLABEL(lab)) { READBYTE(X);  } else
#define RV_SHRT(lab, X)	if(ISLABEL(lab)) { READSHORT(X); } else
#define RV_INT(lab, X)	if(ISLABEL(lab)) { READINT(X);   } else
#define RV_UINT(lab, X)	if(ISLABEL(lab)) { READUINT(X);  } else
#define RV_FLT(lab, X)	if(ISLABEL(lab)) { READFLT(X);   } else
#define RV_VEC(lab, X, N)	if(ISLABEL(lab)) { int b; FINDBEGIN; \
						for(b=0; b<N; b++) {READFLT(X[b])} ReadToken(); } else
#define RV_NBVEC(lab, X, N)	if(ISLABEL(lab)) { READNBYTEVEC(X,N); } else
#define RV_STR(lab, X)	if(ISLABEL(lab)) { READSTR(X); } else
#define RV_STR_INT(lab, S, I)	if(ISLABEL(lab)) { if(!ReadString(S,sizeof(S))) \
								I = strtol(token,0,0); } else
#define RV_END			{ SetError("Unknown label."); retval = false; goto ded_end_read; }

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern RDBOOL		cache_ded_model;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

char	token[128];
char	dedReadError[256];

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static char *sourceBuffer, *sourcePos;
static RDBOOL dedend = true;

// v6 compatbility
static RDBOOL needs_semicolon;
static RDBOOL return_semicolon;
static RDBOOL brace_semicolon;

// CODE --------------------------------------------------------------------

//==========================================================================
// SetError
//==========================================================================
static void SetError(char *str)
{
	sprintf(dedReadError, "%s", str);
}

//==========================================================================
// FGetC
//	Reads a single character from the input file. Increments the line
//	number counter if necessary.
//==========================================================================
static int FGetC()
{
	int ch = (unsigned char) *sourcePos; 

	if(ch) sourcePos++; else dedend = true;	
	if(ch == '\r') return FGetC();
	
    return ch;
}

//==========================================================================
// FUngetC
//	Undoes an FGetC.
//==========================================================================
static int FUngetC(int ch)
{
	if(dedend) return 0;
	if(sourcePos > sourceBuffer) sourcePos--;
	return ch;
}

//==========================================================================
// SkipComment
//	Reads stuff until a newline is found.
//==========================================================================
static void SkipComment()
{
	int ch = FGetC();
	RDBOOL seq = false;
	
	if(ch == '\n') return; // Comment ends right away.
	if(ch != '>') // Single-line comment?
	{
		while(FGetC() != '\n' && !dedend);
	}
	else // Multiline comment?
	{
		while(!dedend)
		{
			ch = FGetC();
			if(seq) 
			{
				if(ch == '#') break;
				seq = false;
			}
			if(ch == '<') seq = true;
		}
	}
}

//==========================================================================
// ReadToken
//==========================================================================
static int ReadToken()
{
	char	*out = token;
	int		ch = FGetC();
	//-------------------

	if(dedend) return false;

	if (return_semicolon)
	{
		FUngetC(ch);
		token[0] = ';'; token[1] = 0;
		return_semicolon = false;
		return true;
	}

	if (brace_semicolon && ch == '}')
	{
		int nch;

		brace_semicolon = false;

		while (!dedend && ((nch = FGetC()) == ' ' || nch == 0x09));

		if (nch == ';')
			FUngetC(nch);
		else
			return_semicolon = true;	
	}

	// Skip whitespace and comments in the beginning.
	while((ch == '#' || isspace(ch))) 
	{
		if(ch == '#') SkipComment();
		ch = FGetC();
		if(dedend) return false;
	}

	// Always store the first character.

	*out++ = ch;
	if(STOPCHAR(ch))
	{
		// Stop here.
		*out = 0;
		return true;
	}

	while(!STOPCHAR(ch) && !dedend)
	{
		// Store the character in the buffer.
		ch = FGetC();
		*out++ = ch;
	}
	*(out-1) = 0;	// End token.
	// Put the last read character back in the stream.
	FUngetC(ch);
	return true;
}

//==========================================================================
// ReadStringEx
//	Current pos in the file is at the first ".
//	Does not expand escape sequences, only checks for \".
//	Returns true if successful.
//  v220-7 script error handling added
//==========================================================================
static int ReadStringEx(char *dest, int maxlen, RDBOOL inside, RDBOOL doubleq)
{
	char *ptr = dest;
	int ch, esc = false, newl = false;

	if (!inside)
	{
		ReadToken();

		if (!stricmp(token, "enable"))
		{
			if (needs_semicolon)
			{
				while (!dedend && ((ch = FGetC()) == ' ' || ch == 0x09));

				if (ch != ';') return_semicolon = true;
				FUngetC(ch);
			}

			strcpy(dest, "slf_enable");
			return true;
		}
		else if (!stricmp(token, "?") && needs_semicolon)
		{
			while (!dedend && ((ch = FGetC()) == ' ' || ch == 0x09));

			if (ch != ';') return_semicolon = true;
			FUngetC(ch);
		}
	}

	// Start reading the characters.
	ch = FGetC();

	while(esc || ch != '"')	// The string-end-character.
	{
		if (dedend) return false;

		if (newl)	// If a newline is found, skip all whitespace that follows
		{
			if(isspace(ch))
			{
				ch = FGetC();
				continue;
			}
			else 
			{
				// End the skip.
				newl = false;
			}
		}

		// An escape character?

		if (!esc && ch == '\\') 
        {
			esc = true;
        }
		else
		{
			// In case it's something other than \" or \\, just insert
			// the whole sequence as-is.
			if(esc && ch != '"' && ch != '\\') *ptr++ = '\\';
			esc = false;
		}
		
        if (ch == '\n') newl = true;
		
        if (!esc && !newl) 
		{
			// Store the character in the buffer.
			*ptr++ = ch;

			if (doubleq && ch == '"') *ptr++ = '"';
			
			if (ptr - dest >= maxlen)
				Con_Error("DED entry %s exceeds max allowed length of %i\n", dest, maxlen - 1);
		}
		
        ch = FGetC();   // Get the next character
	}

    *ptr = 0;	// End the string in a null.

	if (needs_semicolon && ch != ';')
	{
		if (ch == '"') ch = FGetC();
		
		while ( (ch = FGetC()) && !dedend)
		{
			if (ch == ';') { FUngetC(ch); break; }
			
			if (ch <= ' ') continue;
			
			FUngetC(ch);
			
			return_semicolon = true;
			break;
		}
	}
	return true;
}

//===========================================================================
// ReadString
//===========================================================================
static int ReadString(char *dest, int maxlen)
{
	return ReadStringEx(dest, maxlen, false, false);
}

//===========================================================================
// ReadNByteVector
//===========================================================================
static int ReadNByteVector(unsigned char *dest, int max)
{
	int i;

	FINDBEGIN;
	for(i = 0; i < max; i++)
	{
		ReadToken();
		if(ISTOKEN("}")) return true;
		dest[i] = (byte)strtoul(token, 0, 0);
	}
	FINDEND;	
	return true;
}

//==========================================================================
// ReadByte
//	Returns true if successful.
//==========================================================================
static int ReadByte(unsigned char *dest)
{
	ReadToken();
	if(ISTOKEN(";"))
	{
		SetError("Missing integer value.");
		return false;
	}
	*dest = (byte)strtoul(token, 0, 0);
	return true;
}

//==========================================================================
// ReadInt
//	Returns true if successful.
//==========================================================================
static int ReadInt(int *dest, int unsign)
{
	ReadToken();

	if(ISTOKEN(";"))
	{
		SetError("Missing integer value.");
		return false;
	}

	*dest = unsign ? strtoul(token, NULL, 0) : strtol(token, NULL, 0);
	return true;
}

//==========================================================================
// ReadShort
//	Returns true if successful.
//==========================================================================
static short ReadShort(short *dest)
{
	ReadToken();

	if(ISTOKEN(";"))
	{
		SetError("Missing short value.");
		return false;
	}

	*dest = (short)strtol(token, NULL, 0);
	return true;
}

//==========================================================================
// ReadFloat
//	Returns true if successful.
//==========================================================================
static int ReadFloat(float *dest)
{
	ReadToken();

	if(ISTOKEN(";"))
	{
		SetError("Missing float value.");
		return false;
	}

	*dest = (float) strtod(token, 0);
	return true;
}

//==========================================================================
// ReadLabel
//	Returns true if successful.
//==========================================================================
static int ReadLabel(char *label)
{
	strcpy(label, "");
	for(;;)
	{
		ReadToken();
		if(dedend)
		{
			SetError("Unexpected end of file.");
			return false;
		}
		if(ISTOKEN("}"))	// End block.
		{
			strcpy(label, token);
			return true;
		}
		if(ISTOKEN(";"))
		{
			SetError("Label without value.");
			return false;
		}
		if(ISTOKEN("=") || ISTOKEN("{")) break;
		if(label[0]) strcat(label, " ");
		strcat(label, token);
	}
	return true;
}					

//===========================================================================
// DED_InitReader
//===========================================================================
static void DED_InitReader(char *buffer)
{
	sourcePos = sourceBuffer = buffer;
	dedend = false;
}

static ded_mapinfo_t *checkMINF(char *str)
{
	extern ded_t defs;		// The main definitions database.

	int i;

	for (i = 0; i < defs.count.mapinfo.num - 1; i++)
	{
		ded_mapinfo_t	*mt = &defs.mapinfo[i];
		if (!_stricmp(str, mt->id))
			return mt;
	}

	return NULL;
}

//==========================================================================
// DED_ReadData
//	Reads definitions from the given buffer.
//	The definition is being loaded from 'sourcefile' (DED or WAD).
//	The buffer must be null-terminated.
//==========================================================================

// b211-0 Added extended function to work with Defs directory structure.
//     If an Include references a file in a different directory from the
//     calling ded's then, rather than having to specify a fullpath to the
//     file, the include statement needs only to specify the path in the
//     defs tree.
//     e.g. Include "Model\AddOn\fname.ded";

static void addDedInclude(char *tmp, char *source, ded_t *ded)
{
	int idx = DED_AddInclude(ded, "");
	int st = tmp[0] == '\\' || tmp[0] == '/';
	
	M_TranslatePath(tmp, tmp);

	if (st || !Dir_IsAbsolute(tmp))
	{
		// establish ded directory
		if (strchr(tmp, (int)'\\') || strchr(tmp, (int)'/'))
		{
			sprintf(ded->includes[idx].path, "%sDefs\\%s", ddRuntimeDir.path, tmp + st);
		}
		else
		{
			sprintf(ded->includes[idx].path, "%s%s", source, tmp);
		}
	}
	else
		strcpy(ded->includes[idx].path, tmp);
}

static int DED_ReadData(ded_t *ded, 
                        int    bDestroyOld, 
                        char  *buffer,
				        char  *sourceFile)
{
    char lastmodelpath[DED_PATH_LEN]={0};

	int idx, retval = true;
	char label[128], tmp[256];

	ded_mobj_t  *mo;
	ded_state_t *st;
	
    ded_light_t *lig;
	int prev_ligdef_idx = -1; // For "Copy".
	
    ded_model_t *mdl;
    ded_model_t *prevmdl = 0;
	int prev_modef_idx = -1; // For "Copy".
	
    ded_sound_t   *snd;
	ded_mapinfo_t *mi;
	ded_str_t     *tn;
	ded_value_t   *val;
	
    ded_detailtexture_t *dtl;
    int prev_dtldef_idx = -1; // For "Copy".

	ded_ptcgen_t    *gen;
	int prev_gendef_idx = -1; // For "Copy".

	ded_finale_t *fin;

	int  sub;
	int  depth;
	char *rootstr = 0, *ptr;
	int bCopyNext = false;

	directory_t fileDir;

	RDBOOL	corpseFile;

	// ----------------------------------------------------

	// For including other files -- we must know where we are.
	Dir_FileDir(sourceFile, &fileDir);

	corpseFile = strstr(sourceFile, "Monsters.ded") != NULL;

	if(bDestroyOld)
	{
		// Clear all existing definitions.
		DED_Destroy(ded);
		DED_Init(ded);
	}

	// Start parsing the data...

	DED_InitReader(buffer);

	while(ReadToken())
	{
		if(ISTOKEN("Copy") || ISTOKEN("*")) 
		{
			bCopyNext = true;
			continue; // Read the next token.
		}

		// v220-35 add gnd and pbm skip commands
		if(ISTOKEN("Skipifpbm") && ge_inf.switch_flags & SWF_PBM)
		{
			goto ded_end_read;
		}

		if(ISTOKEN("Skipifgnd") && ge_inf.switch_flags & SWF_GND)
		{
			goto ded_end_read;
		}

		if(ISTOKEN("Skipifnotpbm") && !(ge_inf.switch_flags & SWF_PBM))
		{
			goto ded_end_read;
		}

		if(ISTOKEN("Skipifnotgnd") && !(ge_inf.switch_flags & SWF_GND))
		{
			goto ded_end_read;
		}

		if(ISTOKEN("SkipIf"))
		{
			sub = 1;
			ReadToken();
			if(ISTOKEN("Not"))
			{
				sub = 0;
				ReadToken();
			}
			else if (corpseFile && ge_inf.allowCorpse)
			{
				if (!strcmp(token, "-nm"))
				{
					CHECKSC;
					continue;
				}
			}

			if (   sub == (ArgCheck(token) != 0)
				&& !(ge_inf.holycrosswad && ISTOKEN("-hc")))	// v220-35 addition
			{
				// Skip encountered. Get out of here.
				goto ded_end_read; 
			}
			CHECKSC;
		}

		if(ISTOKEN("Include"))
		{
			READSTR(tmp);

			if (tmp) addDedInclude(tmp, fileDir.path, ded);	// b211-0

			CHECKSC;
		}

		if(ISTOKEN("IncludeIf")) // An optional include.
		{
			sub = 1;
			ReadToken();
			if(ISTOKEN("Not")) 
			{
				sub = 0; 
				ReadToken();
			}

			if(sub == (ArgCheck(token) != 0))
			{
				READSTR(tmp);
				if (tmp) addDedInclude(tmp, fileDir.path, ded);	// b211-0
			}
			else
			{
				// Arg not found; just skip it.
				READSTR(tmp);
			}
			CHECKSC;
		}

// v220-7 compatibility with defn written for DD 1.8.6 added		
        if (ISTOKEN("Decoration")) // Compatibility (values ignored)
		{
			char	temp_str[128];
			float	temp_float;
			float	temp_vec[3];
			int		temp_int;

			FINDBEGIN;
			for(;;)
			{
				READLABEL;
				RV_STR("Flat", temp_str)
				RV_STR("Texture", temp_str)
				if(ISLABEL("Light"))
				{
					FINDBEGIN;
					for(;;)
					{
						needs_semicolon = true;
						READLABEL;
						RV_FLT("Halo radius",temp_float)
						RV_INT("Flare texture",temp_int)
						RV_FLT("Radius",temp_float)
						RV_VEC("Color" , temp_vec, 3)
						RV_VEC("Offset", temp_vec, 2)
						RV_VEC("Levels", temp_vec, 2)
						RV_END
						CHECKSC;
						needs_semicolon = false;
					}
				}
				else
					RV_END

				brace_semicolon = true;
				CHECKSC;
			}
		}
		
        if (ISTOKEN("Group")) // Texture blending compatibility (values ignored)
		{
			char temp_str[128];
			int  temp_int;

			FINDBEGIN;
			for(;;)
			{
				READLABEL;
				RV_STR("Flags", temp_str)
				if(ISLABEL("Texture"))
				{
					FINDBEGIN;
					for(;;)
					{
						needs_semicolon = true;
						READLABEL;
						RV_STR("ID",temp_str)
						RV_INT("Tics",temp_int)
						RV_END
						CHECKSC;
						needs_semicolon = false;
					}
				}
				else
					RV_END

				brace_semicolon = true;
				CHECKSC;
			}
		}

        if(ISTOKEN("Map")) // Info
		{
			float	fogcol[3] = {0, 0, 0 };
			int		sklayer_vals = 0;

			idx = DED_AddMapInfo(ded, "");
			mi = ded->mapinfo + idx;

			FINDBEGIN;
			for(;;)
			{
				READLABEL;
				RV_STR("ID",            mi->id)
				RV_STR("Name",          mi->mapname)
				RV_STR("Author",        mi->author)
				RV_STR("Flags",         mi->flags)
				RV_STR("Music",         mi->music.path)
				RV_INT("Par time",      mi->partime)	// v220-4 changed from float to int
				RV_FLT("Fog color R",	fogcol[0])		// b211-0
				RV_FLT("Fog color G",	fogcol[1])		// b211-0
				RV_FLT("Fog color B",	fogcol[2])		// b211-0
				RV_FLT("Fog start",     mi->fog_start)
				RV_FLT("Fog end",       mi->fog_end)
				RV_FLT("Fog density",   mi->fog_density)
				RV_FLT("Ambient light", mi->ambient)
				RV_FLT("Gravity",       mi->gravity)
				RV_FLT("Sky height",    mi->sky_height)
				RV_FLT("Horizon offset",mi->horizon_offset)

				if (ISLABEL("Sky Layer 1") || ISLABEL("Sky Layer 2"))
				{
					int	offset = (int)((byte)label[10] - '0');
					ded_skylayer_t *sl = mi->sky_layers + offset - 1;
					
					sklayer_vals |= offset;

					FINDBEGIN;
					for(;;)
					{
						needs_semicolon = true;
						READLABEL;
						RV_STR("Flags",       sl->flags)
						RV_STR("Texture",     sl->texture)
						RV_FLT("Offset",      sl->offset)
						RV_FLT("Color limit", sl->color_limit)
						RV_END
						CHECKSC;
						needs_semicolon = false;
					}
				}
				else
					RV_END

				brace_semicolon = true;
				CHECKSC;
			}
// end compatibility

			// b211-0 changes for way mapinfo is now designed
			if (!M_ValidateID(mi->id))
			{
				DED_DelEntry(idx, (void**) &mi, &defs.count.mapinfo, sizeof(ded_mapinfo_t));
			}
			else
			{
				ded_mapinfo_t *inf = checkMINF(mi->id);

				if (inf)	// Was there an earlier version of this def?
				{
					inf->R3Dflags |= R3Dcustomded;
					strcpy(inf->mapname, mi->mapname);
					strcpy(inf->author , mi->author);
					strcpy(inf->flags  , mi->flags);
					strcpy(inf->music.path  , mi->music.path);

					// v220-9 don't copy to first skylayer if unset
					//        This is prevent a problem if an author thinks that not
					//        setting them means the original values will be used.
					if (sklayer_vals & 1)
						memcpy(&inf->sky_layers[0], &mi->sky_layers[0], sizeof(ded_skylayer_t));

					if (sklayer_vals & 2)	// v220-9
						memcpy(&inf->sky_layers[1], &mi->sky_layers[1], sizeof(ded_skylayer_t));
					else
						memset(&inf->sky_layers[1], 0, sizeof(ded_skylayer_t));

					inf->partime = mi->partime;
					inf->ambient = mi->ambient;
					inf->gravity = mi->gravity;
					inf->sky_height = mi->sky_height;
					inf->horizon_offset = mi->horizon_offset;

					if (Def_EvalFlags(mi->flags) & (MIF_FOG|MIF_FOGEXP|MIF_FOGEXP2|MIF_FOGOUTSIDE|MIF_FOGUNDERWATER))	// v220-3 bug fix
					{
						inf->fog_color[0] = (byte)(fogcol[0]*255.0f);
						inf->fog_color[1] = (byte)(fogcol[1]*255.0f);
						inf->fog_color[2] = (byte)(fogcol[2]*255.0f);
						inf->fog_color[3] = 0xFF;
						inf->fog_start   = mi->fog_start;
						inf->fog_end     = mi->fog_end;
						inf->fog_density = mi->fog_density;
					}
					else
					{
						inf->fog_color[0] = 0;
						inf->fog_color[1] = 0;
						inf->fog_color[2] = 0;
						inf->fog_start   = 0;
						inf->fog_end     = 0;
						inf->fog_density = 0;
					}

					DED_DelEntry(idx, (void**) &mi, &defs.count.mapinfo, sizeof(ded_mapinfo_t));
				}
			}
		}

		if(ISTOKEN("ModelPath")
		|| ISTOKEN("Model path"))  // 12-20-2005 typo? - jav
		{
			// A new model path. append to the list

			READSTR(label);
			CHECKSC;

			// Mods to speed up model loading - jav
			R_AddModelPath(label, true);
            strcpy(lastmodelpath,label);    // 12-23-2005 remember this path
            Dir_ValidDir(lastmodelpath);    // 12-23-2005 convert slashes, remove trailing blanks make sure it ends with a slash
		}

		if(ISTOKEN("Header"))
		{
			FINDBEGIN;
			// If we're appending an existing definition, skip the header.
			if(!bDestroyOld)
			{
				while(!ISTOKEN("}") && !dedend) ReadToken();
			}
			else 
            for(;;)
			{
				READLABEL;
				RV_INT("Version",      ded->version)
				RV_STR("Thing prefix", ded->mobj_prefix)
				RV_STR("State prefix", ded->state_prefix)
				RV_STR("Sprite prefix",ded->sprite_prefix)
				RV_STR("Sfx prefix",   ded->sfx_prefix)
				RV_STR("Mus prefix",   ded->mus_prefix)
				RV_STR("Text prefix",  ded->text_prefix)
				RV_STR("Model path",   ded->model_path)
				RV_STR("ModelPath",    ded->model_path)   //12-20-2005 typo
				RV_STR("Common model flags",   ded->model_flags)
				RV_FLT("Default model scale",  ded->model_scale)
				RV_FLT("Default model offset", ded->model_offset)
				RV_END
				CHECKSC;
			}
		}

		if(ISTOKEN("Flag"))
		{
			// A new flag.
			idx = DED_AddFlag(ded, "", 0);
			FINDBEGIN;
			for(;;)
			{
				READLABEL;
				RV_STR("ID", ded->flags[idx].id)
				RV_UINT("Value", ded->flags[idx].value)
				RV_END
				CHECKSC;
			}
		}

		if(ISTOKEN("Thing"))
		{
			// A new thing.
			idx = DED_AddMobj(ded, "");
			mo = ded->mobjs + idx;
			mo->cache = cache_ded_model;	// R3D b211-0
			
			FINDBEGIN;
	
    		for(;;)
			{
				READLABEL;
				RV_STR("ID", mo->id)
				RV_INT("DoomEd number", mo->doomednum)
				RV_STR("Name",          mo->name)
				RV_STR("Spawn state",   mo->spawnstate)
				RV_STR("See state",     mo->seestate)
				RV_STR("Pain state",    mo->painstate)
				RV_STR("Melee state",   mo->meleestate)
				RV_STR("Missile state", mo->missilestate)
				RV_STR("Death state",   mo->deathstate)
				RV_STR("Xdeath state",  mo->xdeathstate)
				RV_STR("Raise state",   mo->raisestate)
				RV_STR("See sound",     mo->seesound)
				RV_STR("Attack sound",  mo->attacksound)
				RV_STR("Pain sound",    mo->painsound)
				RV_STR("Death sound",   mo->deathsound)
				RV_STR("Active sound",  mo->activesound)
				RV_INT("Reaction time", mo->reactiontime)
				RV_INT("Pain chance",   mo->painchance)
				RV_INT("Spawn health",  mo->spawnhealth)
				RV_FLT("Speed",         mo->speed)
				RV_FLT("Radius",        mo->radius)
				RV_FLT("Height",        mo->height)
				RV_INT("Mass",          mo->mass)
				RV_INT("Damage",        mo->damage)
				RV_INT("Weapon offset", mo->weapon_offset)	// v220-4
				RV_SHRT("Light",		mo->scplightlevel)	// v220-13
				RV_STR("Flags",         mo->flags[0])
				RV_STR("Flags2",        mo->flags[1])
				RV_STR("Flags3",        mo->flags[2])
				RV_END
				CHECKSC;
			}
		}

		if(ISTOKEN("State"))
		{
			// A new state.
			idx = DED_AddState(ded, "");
			st  = ded->states + idx;
		
        	FINDBEGIN;
		
        	for(;;)
			{
				READLABEL;
				RV_STR("ID",    st->id)
				RV_STR("Flags", st->flags)
				RV_STR("Sprite",st->sprite.id)
				RV_INT("Frame", st->frame)
				RV_INT("Tics",  st->tics)
				RV_STR("Action",st->action)
				RV_STR("Next state", st->nextstate)
				RV_END
				CHECKSC;
			}
		}

		if(ISTOKEN("Sprite"))
		{
			// A new sprite

			idx = DED_AddSprite(ded, "");
			FINDBEGIN;
			for(;;)
			{
				READLABEL;
				RV_STR("ID", ded->sprites[idx].id)
				RV_END
				CHECKSC;
			}
		}

		if(ISTOKEN("Light"))
		{
			// A new light.
			idx = DED_AddLight(ded, "");
			lig = ded->lights + idx;
			if(prev_ligdef_idx >= 0 && bCopyNext) 
			{
				// Should we copy the previous definition?
				memcpy(lig, ded->lights + prev_ligdef_idx, sizeof(*lig));
			}
			FINDBEGIN;
			for(;;)
			{
				READLABEL;
				RV_STR("State",    lig->state)
				RV_FLT("X Offset", lig->xoffset)
				RV_FLT("Y Offset", lig->yoffset)
				RV_FLT("Size",     lig->size)
				RV_FLT("Red",      lig->color[0])
				RV_FLT("Green",    lig->color[1])
				RV_FLT("Blue",     lig->color[2])
				RV_VEC("Color",    lig->color, 3)
				RV_STR("Flags",    lig->flags_string)
				RV_END
				CHECKSC;
			}
			prev_ligdef_idx = idx;
		}

		if(ISTOKEN("Model"))
		{
			// A new model.
			idx = DED_AddModel(ded, "");
			mdl = ded->models + idx;
			sub = 0;
            strcpy(mdl->model_path, lastmodelpath); // 12-12-3005 store with model

            if(prev_modef_idx >= 0) 
			{
				prevmdl = ded->models + prev_modef_idx;
				// Should we copy the previous definition?
				if(bCopyNext) memcpy(mdl, prevmdl, sizeof(*mdl));
			}

            FINDBEGIN;
			for(;;)
			{
				READLABEL;
				RV_STR("State",     mdl->state)
				RV_INT("Off",       mdl->off)
				RV_STR("Sprite",    mdl->sprite.id)
				RV_INT("Sprite frame", mdl->spriteframe)
				RV_STR("Group",     mdl->group)
				RV_INT("Selector",  mdl->selector)
				RV_STR("Flags",     mdl->flags)
				RV_FLT("Inter",     mdl->intermark)
				RV_INT("Skin tics", mdl->skintics)
				RV_FLT("Resize",    mdl->resize)
				if(ISLABEL("Scale"))
				{
					READFLT(mdl->scale[1]);
					mdl->scale[0] = mdl->scale[2] = mdl->scale[1];
				} 
				else RV_VEC("Scale XYZ", mdl->scale, 3)
				RV_FLT("Offset",         mdl->offset[1])
				RV_VEC("Offset XYZ",     mdl->offset, 3)
				RV_VEC("Interpolate",    mdl->interrange, 2)
				RV_FLT("Shadow radius",  mdl->shadowradius)
				if(ISLABEL("Md2") || ISLABEL("Sub"))
				{
					if(sub >= MAX_FRAME_MODELS)
						Con_Error("DED_ReadData: Too many submodels (%s)",
						mdl->state);

					FINDBEGIN;
					for(;;)
					{
						READLABEL;
						RV_STR("File",        mdl->sub[sub].filename.path)
						RV_STR("Frame",       mdl->sub[sub].frame)
						RV_INT("Frame range", mdl->sub[sub].framerange)
						RV_INT("Skin",        mdl->sub[sub].skin)
						RV_INT("Skin range",  mdl->sub[sub].skinrange)
						RV_VEC("Offset XYZ",  mdl->sub[sub].offset, 3)
						RV_STR("Flags",       mdl->sub[sub].flags)
						RV_FLT("Transparent", mdl->sub[sub].alpha)
						RV_FLT("Parm",        mdl->sub[sub].parm)
						RV_BYTE("Selskin mask", mdl->sub[sub].selskinbits[0])
						RV_BYTE("Selskin shift",mdl->sub[sub].selskinbits[1])
						RV_NBVEC("Selskins",   mdl->sub[sub].selskins, 8)
						RV_STR("Shiny skin",   mdl->sub[sub].shinyskin)
						RV_FLT("Shiny",        mdl->sub[sub].shiny)
						RV_VEC("Shiny color",  mdl->sub[sub].shinycolor, 3)
						RV_END
						CHECKSC;
					}
					sub++;
				}
				else 
                {
                   RV_END
                }

				CHECKSC;
			}

			// Some post-processing. No point in doing this in a fancy way,
			// the whole reader will be rewritten sooner or later...
			if(prevmdl)
			{
				int i;
				if(!strcmp(mdl->state, "-"))	 strcpy(mdl->state,		prevmdl->state);
				if(!strcmp(mdl->sprite.id, "-")) strcpy(mdl->sprite.id,	prevmdl->sprite.id);
				if(!strcmp(mdl->group, "-"))	 strcpy(mdl->group,		prevmdl->group);
				if(!strcmp(mdl->flags, "-"))	 strcpy(mdl->flags,		prevmdl->flags);

				for(i=0; i< MAX_FRAME_MODELS; i++)
				{
					if(!strcmp(mdl->sub[i].filename.path, "-"))	strcpy(mdl->sub[i].filename.path,	prevmdl->sub[i].filename.path);
					if(!strcmp(mdl->sub[i].frame, "-"))			strcpy(mdl->sub[i].frame,			prevmdl->sub[i].frame);
					if(!strcmp(mdl->sub[i].flags, "-"))			strcpy(mdl->sub[i].flags,			prevmdl->sub[i].flags);
				}
			}
			prev_modef_idx = idx;			
		}
		
        if(ISTOKEN("Sound"))
		{
			// A new sound.
			idx = DED_AddSound(ded, "");
			snd = ded->sounds + idx;
			FINDBEGIN;
			for(;;)
			{
				READLABEL;
				RV_STR("ID",   snd->id)
				RV_STR("Lump", snd->lumpname)
				RV_STR("Name", snd->name)
				RV_STR("Link", snd->link)
				RV_INT("Link pitch",  snd->link_pitch)
				RV_INT("Link volume", snd->link_volume)
				RV_INT("Priority",    snd->priority)
				RV_INT("Max channels",snd->dedmaxchan)
				RV_INT("Group", snd->group)
				RV_STR("Flags", snd->flags)
				RV_STR("Ext",   snd->ext.path)
				RV_STR("File",  snd->ext.path)
				RV_STR("File name", snd->ext.path)
				RV_END
				CHECKSC;
			}
		}
		
        if(ISTOKEN("Music"))
		{
			// A new music.
			// b211-0 CD TRACK removed
			idx = DED_AddMusic(ded, "");
			FINDBEGIN;
			for(;;)
			{
				READLABEL;
				RV_STR("ID",       ded->music[idx].id)
				RV_STR("Lump",     ded->music[idx].lumpname)
				RV_STR("File name",ded->music[idx].path.path)
				RV_STR("File",     ded->music[idx].path.path)
				RV_STR("Ext",      ded->music[idx].path.path)	// Both work.
				RV_END
				CHECKSC;
			}
		}

		if(ISTOKEN("Text"))
		{
			// A new text.
			idx = DED_AddText(ded, "");
			FINDBEGIN;
			for(;;)
			{
				READLABEL;
				RV_STR("ID", ded->text[idx].id)
				if(ISLABEL("Text"))
				{
					// Allocate a 'huge' buffer

					char *temp = (char*)M_Malloc(0x10000, 10); // 65k (must be consistent I think)
					if(ReadString(temp, 0xffff))
					{
						int len = strlen(temp) + 1;
						ded->text[idx].text = (char*)M_Realloc(temp, len, 24);
						//memcpy(ded->text[idx].text, temp, len);
					}
					else
					{
						M_Free(temp);
						SetError("Syntax error in text value.");
						retval = false;
						goto ded_end_read;
					}
				}
				else RV_END
				CHECKSC;
			}
		}
		
        if(ISTOKEN("Texture")) // Environment
		{
			// A new texenv.
			idx = DED_AddTexEnviron(ded, "");
			FINDBEGIN;
			for(;;)
			{
				READLABEL;
				RV_STR("ID", ded->tenviron[idx].id)
				if(ISLABEL("Texture"))
				{
					// A new texture name.
					tn = (ded_str_t*)DED_NewEntry((void**)&ded->tenviron[idx].textures,
						&ded->tenviron[idx].count, sizeof(*tn));
					FINDBEGIN;
					for(;;)
					{
						READLABEL;
						RV_STR("ID", tn->str)
						RV_END
						CHECKSC;
					}
				}
				else RV_END
				CHECKSC;
			}
		}
		
        if(ISTOKEN("Values")) // String Values
		{
			depth = 0;
			rootstr = (char*)M_Calloc(1, 202);//, 1);	// A null string.
			FINDBEGIN;	
			for(;;)
			{
				// Get the next label but don't stop on }.
				READLABEL_NOBREAK;
				if(strchr(label, '|'))
				{
					SetError("Value labels can't include | characters (ASCII 124).");
					retval = false;
					goto ded_end_read;
				}
				
                if(ISTOKEN("="))
				{
					// Define a new string.
					char *temp = (char*)M_Malloc(0x10000, 11);	// A 'huge' buffer.
					if(ReadString(temp, 0xffff))
					{
						// Reallocate the buffer down to actual string length.
						temp = (char*)M_Realloc(temp, strlen(temp) + 1, 25);				
						// Get a new value entry.
						idx = DED_AddValue(ded, 0);
						val = ded->values + idx;
						val->text = temp;
						// Compose the identifier.
						val->id = (char*)M_Malloc(strlen(rootstr) + strlen(label) + 1, 12);
						strcpy(val->id, rootstr);
						strcat(val->id, label);
					}
					else
					{
						M_Free(temp);
						SetError("Syntax error in string value.");
						retval = false;
						goto ded_end_read;
					}
				}
				else if(ISTOKEN("{"))
				{
					// Begin a new group.
					rootstr = (char*)M_Realloc(rootstr, strlen(rootstr)	+ strlen(label) + 2, 26);
					strcat(rootstr, label);
					strcat(rootstr, "|");	// The separator.
					// Increase group level.
					depth++;
					continue;
				}
				else if(ISTOKEN("}"))
				{
					// End group.
					if(!depth) break;	// End root depth.
					// Decrease level and modify roostr.
					depth--;
					sub = strlen(rootstr);
					rootstr[sub-1] = 0;	// Remove last |.
					ptr = strrchr(rootstr, '|');
					if(ptr)
					{
						ptr[1] = 0;
						rootstr = (char*)M_Realloc(rootstr, strlen(rootstr) + 1, 27);
					}
					else
					{
						// Back to level zero.
						rootstr = (char*)M_Realloc(rootstr, 1, 28);
						*rootstr = 0;
					}
				}
				else
				{
					// Only the above characters are allowed.
					SetError("Illegal token."); 
					retval = false; 
					goto ded_end_read;
				}
				CHECKSC;
			}
			M_Free(rootstr);
			rootstr = 0;
		}

		if(ISTOKEN("Detail")) // Detail Texture
		{
			idx = DED_AddDetail(ded/*, ""*/);
			dtl = ded->dettex + idx;
			if(prev_dtldef_idx >= 0 && bCopyNext) 
			{
				// Should we copy the previous definition?
				memcpy(dtl, ded->dettex + prev_dtldef_idx, sizeof(*dtl));
			}
			FINDBEGIN;
			for(;;)
			{
				READLABEL;
				RV_STR("Wall",     dtl->wall)
				RV_STR("Flat",     dtl->flat)
				RV_STR("Lump",     dtl->det_lumpname)
				RV_FLT("Scale",    dtl->scale)
				RV_FLT("Strength", dtl->strength)
				RV_FLT("Distance", dtl->maxdist)
				RV_END
				CHECKSC;
			}
			prev_dtldef_idx = idx;
		}

		if(ISTOKEN("Generator")) // Particle Generator
		{
			idx = DED_AddPtcGen(ded, "");
			gen = ded->ptcgens + idx;
			sub = 0;
			if(prev_gendef_idx >= 0 && bCopyNext) 
			{
				// Should we copy the previous definition?
				memcpy(gen, ded->ptcgens + prev_gendef_idx, sizeof(*gen));
			}
			FINDBEGIN;
			for(;;)
			{
				READLABEL;
				RV_STR("State",       gen->state)
				RV_STR("Flat",        gen->flat)
				RV_STR("Mobj",        gen->type)
				RV_STR("Alt mobj",    gen->type2)
				RV_STR("Damage mobj", gen->damage)
				RV_STR("Map",         gen->map)
				RV_STR("Flags",       gen->flags_string)
				RV_FLT("Speed",       gen->speed)
				RV_FLT("Speed Rnd",   gen->spd_variance)
				RV_VEC("Vector",      gen->vector, 3)
				RV_FLT("Vector Rnd",  gen->vec_variance)
				RV_VEC("Center",      gen->center, 3)
				RV_FLT("Spawn radius",gen->spawn_radius)
				RV_FLT("Min spawn radius", gen->min_spawn_radius)
				RV_FLT("Distance",    gen->maxdist)
				RV_INT("Spawn age",   gen->spawn_age)
				RV_INT("Max age",     gen->max_age)
				RV_INT("Particles",   gen->particles)
				RV_FLT("Spawn rate",  gen->spawn_rate)
				RV_FLT("Spawn Rnd",   gen->spawn_variance)
				RV_INT("Presim",      gen->presim)
				RV_INT("Alt start",   gen->alt_start)
				RV_FLT("Alt Rnd",     gen->alt_variance)
				RV_VEC("Force axis",  gen->force_axis, 3)
				RV_FLT("Force radius",gen->force_radius)
				RV_FLT("Force",       gen->force)
				RV_VEC("Force origin",gen->force_origin, 3)
				if(ISLABEL("Stage"))
				{
					ded_ptcstage_t *st = gen->stages + sub;
					if(sub >= DED_PTC_STAGES) // 12-10-2005 >=
					{
						SetError("Too many generator stages."); 
						retval = false; 
						goto ded_end_read;
					}
					FINDBEGIN;
					for(;;)
					{
						READLABEL;
						RV_STR("Type",       st->type)
						RV_INT("Tics",       st->tics)
						RV_FLT("Rnd",        st->variance)
						RV_VEC("Color",      st->color, 4)
						RV_FLT("Radius",     st->radius)
						RV_FLT("Radius rnd", st->radius_variance)
						RV_STR("Flags",      st->flags)
						RV_FLT("Bounce",     st->bounce)
						RV_FLT("Gravity",    st->gravity)
						RV_FLT("Resistance", st->resistance)
						RV_END
						CHECKSC;
					}
					sub++;
				} 
				else RV_END
				CHECKSC;
			}
			prev_gendef_idx = idx;
		}

		if(ISTOKEN("Finale"))
		{
			idx = DED_AddFinale(ded);
			fin = ded->finales + idx;
			FINDBEGIN;
			for(;;)
			{
				READLABEL;

				RV_STR("Brief", fin->brief)
				RV_STR("Debrief", fin->debrief)

				// Backward compatibility
				RV_STR("Before", fin->brief)
				RV_STR("After", fin->debrief)

				RV_INT("Game", fin->game)
				if(ISLABEL("Script"))
				{
					// Allocate an "enormous" 64K buffer.
					char *temp = (char*)M_Calloc(0x10000, 203);//, 1);
                    char *ptr;
					if(fin->script) M_Free(fin->script);

					FINDBEGIN;
					ptr = temp;
					ReadToken();
					
                    while(!ISTOKEN("}") && !dedend)
					{
						if(ptr != temp) *ptr++ = ' ';
						strcpy(ptr, token);
						ptr += strlen(token);
						if(ISTOKEN("\""))
						{
							ReadStringEx(ptr, 0x10000 - (ptr-temp), true, true);
							ptr += strlen(ptr); // Continue from the null.
							*ptr++ = '"';
						}
						ReadToken();
						// Skip all semicolons.
						while(ISTOKEN(";") && !dedend) ReadToken();
					}
					fin->script = (char*)M_Realloc(temp, strlen(temp) + 1, 29);
				}
				else RV_END
				CHECKSC;
			}
		}
		bCopyNext = false;
	}
	
ded_end_read:
	M_Free(rootstr);
//	fclose(file);
	return retval;
}

//===========================================================================
// DED_Read
//	Returns true if the file was successfully loaded.
//===========================================================================

int DED_Read(ded_t *ded, char *sPathName, int bDestroyOld) // jav 12-23-2005 path fix
{
	DFILE *file;
	char *defData;
	int len;
	int result;
	char translated[256];

	M_TranslatePath(sPathName, translated);

	if ((file = F_Open(translated, "rb")) == NULL)
	{
		SetError("File can't be opened for reading.");
		return false;
	}

	// Allocate a large enough buffer and read the file.

	F_Seek(file, 0, SEEK_END);

	len = F_Tell(file);
	F_Rewind(file);
	
    defData = (char*)M_Calloc(len + 1, 204);//, 1);
	
    F_Read(defData, len, file);
	F_Close(file);
	
    // Parse the definitions

	result = DED_ReadData(ded, bDestroyOld, defData, translated);
	
    // Now we can free the buffer.
	
    M_Free(defData);
	
    return result;
}


//===========================================================================
// DED_ReadLump
//	Reads definitions from the given lump.
//===========================================================================
int DED_ReadLump(ded_t *ded, int lump, int bDestroyOld)
{
	int  result;
    char *deddata, *ptr;	// ptr for free in case deddata gets changed

	if(lump < 0 || lump >= numlumps) 
	{
		SetError("Bad lump number.");
		return false;
	}

    deddata = ptr = (char*)M_CacheLumpNum(lump, NULL, true);

	result = DED_ReadData(ded, bDestroyOld, deddata, W_LumpSourceFile(lump));

	M_Free(ptr);

	return result;
}

