//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2009 Graham Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: gl_font.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
//  Revised and simplified. TT font support added
// Revision 2.2.0 build 17 (v220-17) 27/Oct/2012 gmj
// Revision 2.2.0 build 18 (v220-18) 05/Jan/2013 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:   DGL Font Renderer
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>

#include "de_base.h"
#include "de_console.h"
#include "de_system.h"
#include "de_graphics.h"
#include "de_misc.h"

#include "r3d_OGLfunc.h"
#include "r3d_compat.h"

// MACROS ------------------------------------------------------------------

#define	NUMBER_FONTS 4

// TYPES -------------------------------------------------------------------

// Data for a character.
typedef struct
{
	int	x, y;				// The upper left corner of the character.
	int	w, h;				// The width and height.
} jfrchar_t;

// Data for a font.
typedef struct
{
	DGLuint		texture;	// The name of the texture for this font.
	int			texWidth, texHeight;
	jfrchar_t	chars[MAX_CHARS];	
} jfrfont_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int		yesno_exit;
int		currentFont = 0;	// Index of the current/default font

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// Four fonts are required;
//  one used in the console the others for Control Panel
static jfrfont_t fonts[NUMBER_FONTS];		// The list of fonts.

// CODE --------------------------------------------------------------------

//===========================================================================
// findPow2
//===========================================================================
static int findPow2(int num)
{
	int cumul;
	for(cumul = 1; num > cumul; cumul *= 2);
	return cumul;
}

//===========================================================================
//	Prepare a GDI font. Select it as the current font. v220-5
//===========================================================================
static void FR_PrepareGDIFont(HFONT hfont)
{
	HBITMAP		hbmp;
	RECT		rect;
	HDC			f_hdc;

	jfrfont_t	*font;
	int			i, x, y, maxh;
	int			bmpWidth = 256, bmpHeight = 0;
	int			imgWidth, imgHeight;
	uint		*image;

	//--------------------------

	// Create a new font.
	font = fonts + currentFont;
		
	// Now we'll create the actual data.
	f_hdc = CreateCompatibleDC(NULL);
	SetMapMode(f_hdc, MM_TEXT);
	SelectObject(f_hdc, hfont);

	// Let's first find out the sizes of all the characters.
	// Then we can decide how large a texture we need.
	for (i = 0, x = 0, y = 0, maxh = 0; i < 256; i++)
	{
		jfrchar_t *fc = font->chars + i;
		SIZE size;
		const char  ch[2] = { (const char)i, 0 };

		GetTextExtentPoint32(f_hdc, ch, 1, &size);
		fc->w = size.cx;
		fc->h = size.cy;
		maxh = max(maxh, fc->h);
		x += fc->w + 1;

		if (x >= bmpWidth)	
		{
			x = 0;
			y += maxh + 1;
			maxh = 0;
		}
	}

	bmpHeight = y + maxh;
	hbmp = CreateCompatibleBitmap(f_hdc, bmpWidth, bmpHeight);
	SelectObject(f_hdc, hbmp);
	SetBkMode(f_hdc, OPAQUE);
	SetBkColor(f_hdc, 0);
	SetTextColor(f_hdc, 0xffffff);
	rect.left = 0;
	rect.top = 0;
	rect.right = bmpWidth;
	rect.bottom = bmpHeight;
	FillRect(f_hdc, &rect, (struct HBRUSH__ *)GetStockObject(BLACK_BRUSH));

	// Print all the characters.
	for (i = x = y = maxh = 0; i < 256; i++)
	{
		jfrchar_t	*fc = font->chars + i;
		const char	ch[2] = { (const char)i, 0 };

		if (x + fc->w + 1 >= bmpWidth)
		{
			y += maxh + 1;
			x = maxh = 0;
		}

		if (i) TextOut(f_hdc, x+1, y+1, ch, 1);

		fc->x = x+1;
		fc->y = y+1;
		maxh = max(maxh, fc->h);
		x += fc->w + 1;
	}

	// Now we can make a version that GL can read.	
	imgWidth = findPow2(bmpWidth);
	imgHeight = findPow2(bmpHeight);

	image = (uint*)M_Calloc(4*imgWidth*imgHeight, 209);

	for (y = 0; y < bmpHeight; y++)
	{
		for(x = 0; x < bmpWidth; x++)
		{
			if(GetPixel(f_hdc, x, y))
				image[x + y*imgWidth] = 0xffffffff;
		}
	}

	font->texWidth = imgWidth;
	font->texHeight = imgHeight;

	GL_NewTexture(&font->texture);
	GL_TexImage(GL_RGBA, imgWidth, imgHeight, 0, image);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	// We no longer need these.
	M_Free(image);
	DeleteObject(hbmp);
	DeleteDC(f_hdc);
}

//===========================================================================
// FR_SetFont	v220-5
//	Select the required font.
//===========================================================================
void FR_SetFont(int idx)
{
	currentFont = idx;

	if (idx)
	{
		// Fonts other than 0 need to be set according to the screen height.
		// One of three possible fonts will be created
		idx = cMode.height < 600 ? 3 : cMode.height < 800? 1 : 2;
	}

	if (idx >= NUMBER_FONTS)
	{
		pr_OFinsert("Program error: invalid font %i requested", idx);
		Sys_Quit(true);
	}

	if (!fonts[currentFont].texture)
	{
		HFONT	hfont;
		char	*winfontname;
		int		pointsize;

		//-------------------------

		if (fonts[currentFont].texture) return;

		switch(idx)
		{
		case 0: winfontname = "Fixedsys"     ; pointsize = 12; break;
		case 1: winfontname = "System"       ; pointsize = 12; break;
		case 2: winfontname = "MS Sans Serif"; pointsize = 18; break;
		case 3:	FR_PrepareGDIFont((struct HFONT__ *)GetStockObject(SYSTEM_FONT));
			return;	
		default: Con_Error("Error: bad font number %i", idx);
		}
		
		hfont = CreateFont(
			-MulDiv(pointsize, GetDeviceCaps(mainHdc, LOGPIXELSY), 72), 
			0, 0, 0, 0, FALSE,
			FALSE, FALSE, DEFAULT_CHARSET,
			OUT_DEFAULT_PRECIS,
			CLIP_DEFAULT_PRECIS,
			DEFAULT_QUALITY,
			VARIABLE_PITCH | FF_SWISS,
			winfontname);

		if (!hfont)
		{
			pr_OFinsert("Failed to create \"%s\" font", winfontname);
			Sys_Quit(true);
		}

		FR_PrepareGDIFont(hfont);
		DeleteObject(hfont);
	}
}

//===========================================================================
// FR_CharWidth
//===========================================================================
int FR_CharWidth(int ch)
{
	return fonts[currentFont].chars[ch].w;
}

//===========================================================================
// FR_TextWidth
//===========================================================================
int FR_TextWidth(char *text)
{
	jfrfont_t	*cf = fonts + currentFont;
	int			i, w, len;

	if (!text || !(len = strlen(text))) return 0;

	// Just add them together.
	for (i = w = 0; i < len; i++) w += cf->chars[(byte)text[i]].w;
	
	return w;
}

//===========================================================================
// FR_TextHeight
//===========================================================================
int FR_TextHeight(char *text)
{
	jfrfont_t	*cf = fonts + currentFont;
	int			i, h, len;

	if (!text || !(len = strlen(text))) return 0;

	// Find the greatest height.
	for (i = h = 0; i < len; i++) h = max(h, cf->chars[(byte)text[i]].h);

	return h;
}

//===========================================================================
// FR_TextOutOffset
//	(x,y) is the upper left corner. Returns the length.
//===========================================================================
int FR_TextOutOffset(char *text, float x, float y)
{
	jfrfont_t	*cf = fonts + currentFont;
	int			i, len, width = 0;

	if (!text || !(len = strlen(text))) return 0;

	GL_Bind(cf->texture);	// Set the texture.

	glBegin(GL_QUADS);

	// Print the chars
	for (i = 0; i < len; i++)
	{
		jfrchar_t *ct = cf->chars + (byte)text[i];
		float cx   = (float)ct->x       , cy   = (float)ct->y ;
		float cw   = (float)ct->w       , ch   = (float)ct->h;
		float texw = (float)cf->texWidth, texh = (float)cf->texHeight;

		glTexCoord2f(cx/texw, cy/texh);
		glVertex2f(x, y);
		// Upper right.
		glTexCoord2f((cx + cw)/texw, cy/texh);
		glVertex2f(x + ct->w, y);
		// Lower right.
		glTexCoord2f((cx + cw)/texw, (cy + ch)/texh);
		glVertex2f(x + ct->w, y + ct->h);
		// Lower left.
		glTexCoord2f(cx/texw, (cy + ch)/texh);
		glVertex2f(x, y + ct->h);

		// Move on.
		width += ct->w;
		x += ct->w;
	}

	glEnd();

	return width;
}

int FR_TextOut(char *text, int x, int y)
{
	return FR_TextOutOffset(text, (float)x, (float)y);
}

//===========================================================================
// FR_TextOutXOffCentre - yoff lines from centre
//===========================================================================
void FR_TextOutXOffCentre(char *msg, int yoff)
{
	FR_TextOut(msg,
		(640 - FR_TextWidth (msg))/2,
		(400 - FR_TextHeight(msg))/2 + yoff );
}

//===========================================================================
// FR_FreeFonts - called if changing screen size
//===========================================================================
void FR_FreeFonts()
{
	int	i;

	for (i = 0; i < NUMBER_FONTS; i++)
	{
		GL_DeleteTexIndex(&fonts[i].texture, "font");
		fonts[i].texture = 0;
	}
}

// =============================================================

RDBOOL GL_CheckWindow(int *width, int *height);
RDBOOL critical_inuse = false;

// NOTE: only simple strings handled (since that's all that's required)
//       long lines of 79+ chars (or whatever fits in a width of 560) with no spaces will be truncated
//       Messages must be formatted to stay within a window height of 360.
static void subfn_CMstrings(const char *msg, int *wd, int *ht, int x1, int y1, int exit_type, int centre, int fh, RDBOOL set, int prendmsg)
{
extern int global_return;

	static char msg0[] = {"Press Y to continue"};
	static char msg1[] = {"Press Y to close window"};
	static char msg2[] = {"Press N to exit"};
	static char msg3[] = {"Press N to return"};

	jfrfont_t	*cf = fonts + currentFont;
	char		buffer[1024];
	uint		len = strlen(msg);
	uint		a, c, i, k;
	int			lf;
	int			w;

	if (set) { *wd = 0; *ht = 0; }

	for (a = c = i = w = 0; i <= len; i++)
	{
		if (msg[i] == 0xd) continue;

		if ( (lf = msg[i] == 0xa) || !msg[i] || w > 560)
		{
			if (c)
			{
				buffer[c] = 0;

				if (w > 560)
				{
					// scan backwards until finding a space
					for (k = c - 1; k > 0; --k)
					{
						w -= cf->chars[(byte)msg[i--]].w;

						if (buffer[k] == ' ')
						{
							buffer[k] = 0;
						
							if (w <= 560) break;
						}
					}
				}

				if (strlen(buffer) >= 79) buffer[79] = 0;

				if (set)
				{
					w = FR_TextWidth(buffer);
					if (w > *wd) *wd = w;
				}
				else
				{
					FR_TextOutOffset(buffer, (float)x1, (float)(y1 + a));
				}
				a += fh;
				c = w = 0; 
			}
			else if (lf)
			{
				a += fh;
			}

			if (!msg[i]) break;
			continue;
		}

		if (c >= 1023) break;	// insurance;

		buffer[c++] = msg[i];
		w += cf->chars[(byte)msg[i]].w;
	}

	if (prendmsg)
	{
		a += fh * 2;

		if (set)
		{
			if (centre)
			{
				if ( (w = FR_TextWidth(msg0)) > *wd) *wd = w;
			}
			else
			{
				if ( (w = FR_TextWidth(msg1)) > *wd) *wd = w;
			}
		}
		else
		{
			if (centre)
			{
				FR_TextOutOffset(msg0, (float)x1, (float)(y1 + a));
			}
			else
			{
				FR_TextOutOffset(msg1, (float)x1, (float)(y1 + a));
			}
		}

		a += fh;

		if (!exit_type)
		{
			if (set)
			{
				if ( (w = FR_TextWidth(global_return ? msg3 : msg2)) > *wd) *wd = w;
			}
			else
			{
				FR_TextOutOffset(global_return ? msg3 : msg2, (float)x1, (float)(y1 + a));
			}

			a += fh;
		}
	}
	else
	{
		a += fh;
	}

	if (set) *ht = a;
}

// v220-22 If called direct then can be called from dd_loop.
//         Currently used to warn about clipping mode.
RDBOOL glMessageBox(const char *msg, int centre, int exit_type, int prendmsg)
{
	static const float		hcol[] = { 1.0f, 1.0f, 1.0f };
	static const float		mcol[] = { 0.6f, 0.6f, 0.6f };
	static const float		lcol[] = { 0.1f, 0.1f, 0.1f };

	int			matrixmode;
	int			x, y, x1, y1;
	int			x_w, y_h, xbrd, ybrd;
	int			xbrd_h, ybrd_h;
	int			wd, ht, fh;
	int			brd;
	int			val;

	// Set the message strings
	FR_SetFont(0);
	fh = FR_TextHeight("W");
	subfn_CMstrings(msg, &wd, &ht, 0, 0, exit_type, centre, fh, true, prendmsg);

	brd = 8;

	if (ht & 1) ht++;
	if (wd & 1) wd++;

	ht += 2 * brd + 20;
	wd += 2 * brd + 20;

	// check that message fits (it should)
	if (centre)
	{
		if ( (x = (current_w - wd)/2) < 0) return false;
		if ( (y = (current_h - ht)/2) < 0) return false;
	}
	else
	{
		if ((x = 10) + wd >= current_w) return false;
		if ((y = 40) + ht >= current_h) return false;
	}

	x1 = x + 20;		y1 = y + 20;
	x_w = x + wd;		y_h = y + ht;
	xbrd = x + brd;		ybrd = y + brd;
	xbrd_h = x_w - brd;	ybrd_h = y_h - brd;

	glGetIntegerv(GL_MATRIX_MODE, &matrixmode);

	glGetIntegerv(GL_CULL_FACE_MODE, &val);
	if (val != GL_FALSE) glDisable(GL_CULL_FACE);
	glGetIntegerv(GL_FRONT_FACE, &val);
	if (val != GL_CW) glFrontFace(GL_CW);

	glClear(GL_DEPTH_BUFFER_BIT);
	// Go into screen projection mode.
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();

	// Some older cards may not support enough proj. matrices.
	if (GL_CheckError("OGL msg window") != GL_NO_ERROR)
	{
		Sys_Quit(true);
	}

	glLoadIdentity();
	glOrtho(0, current_w, current_h, 0, -1, 1);
	glDisable(GL_TEXTURE_2D);
	glBegin(GL_QUADS);

	// Top.
	glColor3fv(hcol);
	glVertex2i(xbrd_h, ybrd); glVertex2i(xbrd, ybrd); glVertex2i(x, y);
	glColor3fv(lcol);
	glVertex2i(x_w, y);

	// Right.
	glVertex2i(xbrd_h, ybrd); glVertex2i(x_w, y);
	glColor3fv(mcol);
	glVertex2i(x_w, y_h); glVertex2i(xbrd_h, ybrd_h);

	// Bottom.
	glVertex2i(xbrd, ybrd_h); glVertex2i(xbrd_h, ybrd_h); glVertex2i(x_w, y_h);
	glColor3fv(lcol);
	glVertex2i(x, y_h);

	// Left.
	glColor3fv(hcol);
	glVertex2i(x, y); glVertex2i(xbrd, ybrd);
	glColor3fv(lcol);
	glVertex2i(xbrd, ybrd_h); glVertex2i(x, y_h);

	// Centre
	glColor3fv(hcol);
	glVertex2i(xbrd_h , ybrd_h); glVertex2i(xbrd, ybrd_h); glVertex2i(xbrd, ybrd); glVertex2i(xbrd_h, ybrd);

	glEnd();
	glClear(GL_DEPTH_BUFFER_BIT);
	glEnable(GL_TEXTURE_2D);

	// draw the message strings
	glColor3f(0.f, 0.f, 0.f);

	subfn_CMstrings(msg, NULL, NULL, x1, y1, exit_type, centre, fh, false, prendmsg);

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	if (matrixmode != GL_PROJECTION) glMatrixMode(matrixmode);

	return true;
}

// v220-17 Introduced because, unlike XP, Vista/Win7/Win10 fails to
//         draw the MessageBox() on top with a fullscreen OGL window.
//         NOTE: this should only be called during game load.
// v220-18 modified
int FR_CriticalMessage(const char *msg, int exit_type)
{
	int			ret;
	int			firsttime;

	// ------------------------------------

	if (cMode.w_checkwin < 0) return true;

	if (critical_inuse || !glMessageBox(msg, 1, exit_type, 1))
		return exit_type;

	glFinish();
	SwapBuffers(mainHdc);

	ret = 0;
	critical_inuse = true;
	firsttime = true;

	while(1)
	{
		Sleep(5);

		if (yesno_exit < 0)
		{
			ret = exit_type || yesno_exit == -1;
			break;
		}

		yesno_exit = exit_type + 1;

		MessagePump();
		if (firsttime)
		{
			firsttime = false;
			glMessageBox(msg, 1, exit_type, 1);
			glFinish();
			SwapBuffers(mainHdc);
		}
	}

	yesno_exit = 0;
	critical_inuse = false;

	return ret;
}

// Added v220-35
int FR_StandardMessage(const char *msg)
{
	int			firsttime = true;
	int			ret;

	// ------------------------------------

	glFinish();
	SwapBuffers(mainHdc);

	yesno_exit = 1;
	glMessageBox(msg, 1, 0, 0);

	while(1)
	{
		Sleep(5);

		if (yesno_exit < 0)
		{
			ret = yesno_exit < 0;
			break;
		}

		MessagePump();

		if (firsttime)
		{
			firsttime = false;
			glMessageBox(msg, 1, 0, 0);
			glFinish();
			SwapBuffers(mainHdc);
		}
	}

	glFinish();
	SwapBuffers(mainHdc);

	ret = yesno_exit;
	yesno_exit = 0;

	return (ret == -1);
}
