//-----------------------------------------------------------------------------
//
// Copyright (C) 2009 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: jpeg.c First written v220-6
//
// DESCRIPTION: These functions create a basic interface to the
//              The Independent JPEG Group's JPEG software
//              available from http://www.ijg.org/	The jpeg files them-
//			    selves have been stripped of unwanted features by Randy
//              Height for the ZDoom port.       
//
//-----------------------------------------------------------------------------

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "con_console.h"
#include "m_winmem.h"
#include "de_misc.h"
#include "dd_wad.h"

// This is the main lib header
#include "..\jpegV6b\jpeglib.h"

// The following headers are needed to create
// an independant source manager.
#include "..\jpegV6b\jinclude.h"
#include "..\jpegV6b\jerror.h"

#define INPUT_BUF_SIZE  4096

// Structure for independant source manager
typedef struct {
	struct jpeg_source_mgr pub;	/* public fields */

	DFILE	*Infile;
	JOCTET	Buffer[INPUT_BUF_SIZE];
	boolean	Start_Of_File;
} source_mgr;

typedef source_mgr *src_ptr;

struct	jpeg_decompress_struct cinfo;
struct	jpeg_error_mgr jerr;

static JOCTET Buffer[INPUT_BUF_SIZE * SIZEOF(JOCTET)];
static void	*mgrmem;

// redirected functions. These are based on those
// contained in jdatasrc.c as the same method is used
// to provide a proprietory source manager. See
// 'Caveat programmer' in jdatasrc.c
static void Init_Source(j_decompress_ptr cinfo)
{
	src_ptr src = (src_ptr) cinfo->src;

	src->Start_Of_File = TRUE;
}

static boolean Fill_Input_Buffer(j_decompress_ptr cinfo)
{
	src_ptr src = (src_ptr) cinfo->src;
	size_t nbytes;

	nbytes = F_Read(src->Buffer, INPUT_BUF_SIZE, src->Infile);

	if (nbytes <= 0) {
		if (src->Start_Of_File)	/* Treat empty input file as fatal error */
			ERREXIT(cinfo, JERR_INPUT_EMPTY);
		WARNMS(cinfo, JWRN_JPEG_EOF);
		/* Insert a fake EOI marker */
		src->Buffer[0] = (JOCTET) 0xFF;
		src->Buffer[1] = (JOCTET) JPEG_EOI;
		nbytes = 2;
	}

	src->pub.next_input_byte = src->Buffer;
	src->pub.bytes_in_buffer = nbytes;
	src->Start_Of_File = FALSE;

	return TRUE;
}

static void Skip_Input_Data(j_decompress_ptr cinfo, long num_bytes)
{
	src_ptr src = (src_ptr) cinfo->src;

	if (num_bytes > 0)
	{
		while (num_bytes > (long) src->pub.bytes_in_buffer) {
			num_bytes -= (long) src->pub.bytes_in_buffer;
			(void) Fill_Input_Buffer(cinfo);
		}
		src->pub.next_input_byte += (size_t) num_bytes;
		src->pub.bytes_in_buffer -= (size_t) num_bytes;
	}
}

static void Term_Source(j_decompress_ptr cinfo)
{
	/* no work necessary here */
}

void init_src(j_decompress_ptr cinfo, DFILE * infile)
{
	src_ptr src = (src_ptr) cinfo->src;

	src = (src_ptr) cinfo->src;
	src->Infile = infile;
	src->Start_Of_File = true;

	// set pointers to redirected functions
	src->pub.init_source = Init_Source;
	src->pub.fill_input_buffer = Fill_Input_Buffer;
	src->pub.skip_input_data = Skip_Input_Data;
	src->pub.resync_to_restart = jpeg_resync_to_restart; // use default method
	src->pub.term_source = Term_Source;
	src->pub.bytes_in_buffer = 0;		// forces Fill_Input_Buffer on first read
	src->pub.next_input_byte = NULL;	// until buffer loaded
}

// Reverse the byte order for a short
static ushort SWAPSHORT (ushort x)
{
	return  ((x & 0x00ff) << 8) | ((x & 0xff00) >> 8);
}

#define M_SOF0  0xc0
#define M_SOF3  0xc3

static RDBOOL jpg_Size(DFILE *infile, int *width, int *height)
{
	ushort	w[2];		// fetch up to 2 words
	byte	b[4];		// fetch up to 4 bytes
	int		count = 48;	// don't loop too many times

	F_Seek(infile, 0, SEEK_SET);

	if (F_Read(b, 4, infile) != 4) return false;
	
	if (b[0] != 0xff || b[1] != 0xd8 || b[2] != 0xff)
		return false;

	b[1] = b[3];

	// Process a SOFn marker in range M_SOF0 to M_SOF2
	while (1)
	{
		if (!count--) return false;

		// if it's not a valid jpg marker, bail out
		if (b[1] < (byte)M_SOF0) return false;

		// have we reached a valid marker?
		if (b[1] < (byte)M_SOF3) break;

		if (F_Read(w, 2, infile) != 2) return false;

		w[0] = SWAPSHORT(w[0]);	// convert to big endian

		// get offset to next marker - should be at least 5
		if ((ushort)(w[0]) < 5) return false;

		// deduct two bytes as the offset we want to apply
		// is from the start of w[0] not the end
		F_Seek(infile, (long)((ushort)w[0] - 2) , SEEK_CUR);

		if (F_Read(b, 2, infile) != 2 || (byte)*b != 0xFF)
			return false;

		if (deof(infile)) return false;
	}

	// move on 3 bytes and with any luck...
	F_Seek(infile, 3 , SEEK_CUR);
	// ...we've landed on the sizes
	if (F_Read(w, 4, infile) != 4) return false;

	*height = (int)((ushort)SWAPSHORT(w[0]));
	*width  = (int)((ushort)SWAPSHORT(w[1]));

	return true;
}

// We have to fart about because we don't want to start the state
// machine just to validate the header. Unfortunately the jpeg
// code does not have have a function that will validate the file
// without either doing more work than is wanted or bombing out.
RDBOOL JPG_Data(char *filename, int lump, int *width, int *height)
{
	DFILE *infile = filename ? F_Open(filename, "fb" ) : F_OpenLump(lump);
	RDBOOL	ret = false;

	if (infile != NULL)
	{
		ret = jpg_Size(infile, width, height);
		F_Close(infile);
	}

	return ret;
}

void JPEG_Error(j_common_ptr cinfo)
{
	char buffer[200];	// recommended size

	(*cinfo->err->format_message) (cinfo, buffer);
	Con_Error("JPEG failure: %s\n", buffer);
}

byte *JPG_Load(char *filename, int lump, int *width, int *height, int *pixsz, int from)
{
	DFILE	*infile = filename ? F_Open(filename, "fb" ) : F_OpenLump(lump);
	JSAMPLE *data = NULL;
	char	*name = filename ? filename : lumpinfo[lump].name;

	if (infile)
	{
		uint	rows;
		byte	*grey_convert = NULL;
		
		memset(&cinfo, 0, sizeof(cinfo));

		jpeg_create_decompress(&cinfo);	// do this first

		cinfo.err = jpeg_std_error(&jerr);
		cinfo.err->output_message = JPEG_Error;
		cinfo.err->error_exit = JPEG_Error;

		if (!mgrmem) mgrmem = W_Malloc(SIZEOF(source_mgr), STATICID, 1);
		cinfo.src = (struct jpeg_source_mgr*)mgrmem;

		init_src(&cinfo, infile);
		jpeg_read_header(&cinfo, TRUE);

		*width  = (int)cinfo.image_width;
		*height = (int)cinfo.image_height;

		if ( !((cinfo.out_color_space == JCS_GRAYSCALE  && cinfo.num_components == 1)
			|| (cinfo.out_color_space == JCS_RGB  && cinfo.num_components == 3)))
		{
			Con_Message("jpeg %s\nUnsupported color format type %i px %i\n",
				name, cinfo.out_color_space, cinfo.num_components);
			goto jpg_exit;
		}

		jpeg_start_decompress(&cinfo);

		*pixsz = cinfo.num_components;	// Not interested in alpha

		if (cinfo.num_components == 1)
		{
			*pixsz = 3;	// grey is converted to RGB
			grey_convert = (byte*)M_Malloc(cinfo.output_width, 0);
		}

		// set safe buffer size. Since it's destroyed oversize does not matter
		data = (byte*)M_Malloc(cinfo.output_width * cinfo.output_height * 4, 1001);
		rows = 0;

		// Input the data row by row - this ensures interrupted streaming will work
		while (cinfo.output_scanline < cinfo.output_height)
		{
			if (grey_convert)
			{
				byte	*ptr = data + cinfo.output_width * 3 * rows;
				uint	i;

				jpeg_read_scanlines(&cinfo, &grey_convert, 1);

				for (i = 0; i < cinfo.output_width; i++)
				{
					*ptr++ = grey_convert[i];
					*ptr++ = grey_convert[i];
					*ptr++ = grey_convert[i];
				}
			}
			else
			{
				byte *ptr = data + cinfo.output_width * cinfo.output_components * rows;
				jpeg_read_scanlines(&cinfo, &ptr, 1);
			}
			rows++;
		}

		jpeg_finish_decompress(&cinfo);

jpg_exit:
		jpeg_destroy_decompress(&cinfo);
		M_Free(grey_convert);
		F_Close(infile);
	}

	if (!data)
		Con_Error("JPG %s \"%s\" is bad",  filename ? "file" : "lump", name);

	return data;
}
