//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2011 G Jackson, Jaakko Kernen, Jack Vermeulen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: gl_main.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 12 (v220-12) 01/Oct/2011 gmj
// Revision 2.2.0 build 23 (v220-23) 10/Jan/2014 gmj
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:  Main OpenGL routines
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include "r3d_OGLfunc.h"

#include "de_base.h"
#include "de_console.h"
#include "de_system.h"
#include "de_graphics.h"
#include "de_refresh.h"
#include "de_render.h"
#include "de_misc.h"

#include "r3d_compat.h"
#include "r3d_edit.h"
#include "r3d_mapinfo.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void	getDesktopParms(int *width, int *height, int *freq, int *bpp);
uint	checkEnumerated(void);
RDBOOL	GL_DrawPagePatch(int lnum);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

RDBOOL	gl_was_inited;
char	aspect_str[10];
int		default_FOV;
int		init_FOV;
int		defaultNearClip = 5;	// v220-33
uint	current_win[4][2];

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern DGLuint		menuFogTexture;	
extern DGLubyte		*screen_buffer;
extern RDBOOL		ScreenError;
extern int			setup_active;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// Default gamma etc.
float		vid_bright   = 0.0f;
float		vid_gamma    = 1.0f;
float		vid_contrast = 1.0f;
float		screen_ratio;

RDBOOL		gl_inited = false;
RDBOOL		update_gamma = false;

// v220-3
dispmode_t		*modeNlist;
uint			scr_max_width  = MIN_SCR_WIDTH;
uint			scr_max_height = MIN_SCR_HEIGHT;
int				modeNcount;

byte			fogcolour[4];

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static uint		monitor_w;
static uint		monitor_h;
static int		clear_count;	// v220-10 used to clear screen at transition

// CODE --------------------------------------------------------------------

//==========================================================================
// GL_DoUpdate
//	Swaps buffers / blits the back buffer to the front.
//  Called once each frame from dd_loop
//==========================================================================
void GL_DoUpdate(void)
{
	if (ge_inf.clear_screen & 2)
	{
		glViewport(0, 0, cMode.width, cMode.height);
		glClearColor(0.f, 0.f, 0.f, 0.f);
		glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);	
			
		SwapBuffers(mainHdc);

		glClearColor(0.f, 0.f, 0.f, 0.f);
		glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

		glViewport(aspect_offset_x, aspect_offset_y, current_w, current_h);

		ge_inf.clear_screen = (ge_inf.clear_screen & 1);
		clear_count = (aspect_offset_x || aspect_offset_y) ? 10 : 0;
	}

	// The Windows toolbar can stay in view even in fullscreem mode.
	// To overcome this the following clearing function is run.
	if (clear_count)
	{
		clear_count--;

		// v220-10 to clear undrawn area just clear the whole buffer
		glViewport(0, 0, cMode.width, cMode.height);
		glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
		glClear(GL_COLOR_BUFFER_BIT);	
		glViewport(aspect_offset_x, aspect_offset_y, current_w, current_h);
	}

	glFlush();

	SwapBuffers(mainHdc);
}

//===========================================================================
// GL_InitFont
//===========================================================================
void GL_InitFont(void)
{
	FR_SetFont(1);	// 1) init the variable font
	FR_SetFont(0);	// 2) init the fixed font (this is the one needed next)
	Con_MaxLineLength();
}

// v220-3 screen handling stuff
static int __cdecl widthSorter(const void *e1, const void *e2)
{
	uint w1 = ((dispmode_t*) e1)->width, w2 = ((dispmode_t*) e2)->width;

	if (w1 > w2) return -1;
	if (w1 < w2) return  1;
	return 0;
}

// v220-3 Get a sorted width list of display resolutions
static void GL_SetDisplayModes()
{
	dispmode_t *tmp;
	DEVMODE	testMode;
	int		i, k, n, lc;

	// v220-12	In some rare circumstances the monitor inf will have a setting
	//			of 1920 x 1080 for full HD capability  although the monitor's
	//			true max resolution is lower.
	//			E.g. The HP w2207h has a max res of 1680 x 1050 but where fitted
	//			to some systems has been frigged to support HD (albeit with poor
	//			results as would be expected). This will throw all the values
	//			set to calculate the required FOV etc.
	//			The following has been added to allow the user to specify the
	//			true max values using -user_ds w h on the command line

	RDBOOL		user_ds = false;
	RDBOOL		user_dsvalid = false;
	RDBOOL		dt_valid = !ArgExists("-nodt");
	uint		user_maxw = 0, user_maxh = 0;

	if (ArgCheckWith("-user_ds", 2))
	{
		dt_valid = false;	// user overrides desktop
		user_ds = true;
		user_maxw = atoi(ArgNext());
		user_maxh = atoi(ArgNext());
	}
	else if (dt_valid)
	{
		user_ds = true;
		user_maxw = cMode.max_width;
		user_maxh = cMode.max_height;
	}

	// Count number of entries
	for (i = n = 0; ; i++)
	{
		memset(&testMode, 0, sizeof(testMode));
		testMode.dmSize = sizeof(DEVMODE);

		if (!EnumDisplaySettings(NULL, i, &testMode)) break;

		if (   testMode.dmBitsPerPel >= 16
			&& testMode.dmPelsWidth  >= MIN_SCR_WIDTH
			&& testMode.dmPelsHeight >= MIN_SCR_HEIGHT)
		{
			// Check any user_ds settings
			if (   user_ds
				&& testMode.dmPelsWidth  == user_maxw
				&& testMode.dmPelsHeight == user_maxh)
			{
				user_dsvalid = true;
			}
			n++;
		}
	}

	if (!n)
		Con_Error("HARDWARE PROBLEM: No 16/32 bpp screen modes available\n");

	if (user_ds)
		Con_Message("%s: MAXW %i MAXH %i have been %s\n", dt_valid ? "DESKTOP" : "USER_DS", user_maxw, user_maxh, user_dsvalid ? "accepted" : "rejected");
	
	tmp = (dispmode_t*)M_Calloc(n * sizeof(dispmode_t), 0);

	if (ArgExists("-desktop"))
	{
		scr_max_width = user_maxw;
		scr_max_height = user_maxh;
		lc = 0;
	}
	else
	{
		// Populate the dispmode array
		for (i = lc = 0; lc < n; i++)
		{
			memset(&testMode, 0, sizeof(testMode));
			testMode.dmSize = sizeof(DEVMODE);
			if (!EnumDisplaySettings(NULL, i, &testMode)) break;

			if (   testMode.dmBitsPerPel >= 16
				&& testMode.dmPelsWidth  >= MIN_SCR_WIDTH
				&& testMode.dmPelsHeight >= MIN_SCR_HEIGHT)
			{
				RDBOOL found = false;

				for (k = 0; k < lc; k++)
				{
					if (   tmp[k].width  == testMode.dmPelsWidth
						&& tmp[k].height == testMode.dmPelsHeight
						&& tmp[k].freq   == testMode.dmDisplayFrequency)
					{
						found = true; break;
					}
				}
				
				if (!found)
				{
					if (user_dsvalid)
					{
						if (   testMode.dmPelsWidth  > user_maxw
							|| testMode.dmPelsHeight > user_maxh)
						{
							continue;
						}
					}

					if (testMode.dmPelsWidth > scr_max_width)
						scr_max_width = testMode.dmPelsWidth;

					if (testMode.dmPelsHeight > scr_max_height)
						scr_max_height = testMode.dmPelsHeight;

					tmp[lc].width  = testMode.dmPelsWidth;
					tmp[lc].height = testMode.dmPelsHeight;
					tmp[lc++].freq = testMode.dmDisplayFrequency;
				}
			}
		}
	}

	screen_ratio = (float)scr_max_width/(float)scr_max_height;
	modeNcount = lc;

	if (modeNcount > 1)	// Order the modelist by width
		qsort(tmp, modeNcount, sizeof(dispmode_t), widthSorter);

	modeNcount++;

	if (modeNcount > 1)
	{
		modeNlist = (dispmode_t*)W_Calloc(modeNcount*sizeof(dispmode_t), STATICID, 0);
		memcpy(&modeNlist[1], tmp, (modeNcount - 1)*sizeof(dispmode_t));
	}
	else	// v220-23	cover case where no matches were found
	{
		modeNcount++;
		modeNcount++;
		modeNlist = (dispmode_t*)W_Calloc(modeNcount*sizeof(dispmode_t), STATICID, 0);
		modeNlist[1].freq = cMode.freq;
		modeNlist[1].width = user_maxw;
		modeNlist[1].height = user_maxh;
		modeNlist[1].xpos = 0;
		modeNlist[1].ypos = 0;
		modeNlist[2].freq = cMode.freq;
		modeNlist[2].width = user_maxw - 100;
		modeNlist[2].height = user_maxh - 100;
		modeNlist[2].xpos = 10;
		modeNlist[2].ypos = 10;
	}

	M_Free(tmp);
}

void getDEVMODEparms()
{
	DEVMODE getMode;
	memset(&getMode, 0, sizeof(getMode));	
	getMode.dmSize = sizeof(getMode);
	EnumDisplaySettings(NULL, ENUM_CURRENT_SETTINGS, &getMode);

	current_win[0][0] = current_win[0][1] = getMode.dmPelsWidth;
	current_win[1][0] = current_win[1][1] = getMode.dmPelsHeight;
	current_win[2][0] = current_win[2][1] = getMode.dmDisplayFrequency;
	current_win[3][0] = current_win[3][1] = getMode.dmBitsPerPel;
}

//===========================================================================
// OGL init entry point.  Only called once
//===========================================================================
void GL_Init(void)
{
	RDBOOL	screenfile_err = false;
	RDBOOL	noscreenfile = true;
	uint	f;

	// Reset if out of range.
	if (vid_bright < -.5f || vid_bright > .7f)
		vid_bright = 0;
	if (vid_gamma <  .5f || vid_gamma > 1.5f)
		vid_gamma = 1.0f;
	if (vid_contrast <  .5f || vid_contrast > 1.5f)
		vid_contrast = 1.0f;

	// Read in the screenfile (if it exists)and set the cMode array
		// get the values on entry
	getDesktopParms(&cMode.max_width, &cMode.max_height, &cMode.w_initialfreq, &cMode.w_initialbpp);
	
	// If there's a screen file get the values
	if (SR_LoadFile(screenfile, "SCREEN", &screenfile_err) && !screenfile_err)
	{
		char	*nextstr;
		int		i, version = 0, n[10];
		RDBOOL	found_screen = false;

		memset(n, 0, sizeof(n));
		noscreenfile = false;

		while ( (nextstr = SR_NextSTR(false)) )
		{
			if (!_stricmp(nextstr, "[SCREEN]"))
			{
				found_screen = true;
				continue;
			}

			if (found_screen)
			{
				if (!_stricmp(nextstr, "width"     )) { cMode.width  = SR_NextNum(); n[0] = true; }
				else if (!_stricmp(nextstr, "height"    )) { cMode.height = SR_NextNum(); n[1] = true; }
				else if (!_stricmp(nextstr, "freq"      )) { cMode.freq = SR_NextNum(); n[2] = true; }
				else if (!_stricmp(nextstr, "bpp"       )) { cMode.bpp  = SR_NextNum(); n[3] = true; }
				else if (!_stricmp(nextstr, "xpos"      )) { cMode.xpos = SR_NextNum(); n[4] = true; }
				else if (!_stricmp(nextstr, "ypos"      )) { cMode.ypos = SR_NextNum(); n[5] = true; }
				else if (!_stricmp(nextstr, "fullscreen")) { cMode.fullscreen = SR_NextNum() != 0; n[6] = true; }
				else if (!_stricmp(nextstr, "monitor_w" )) { monitor_w = SR_NextNum(); n[7] = true; }
				else if (!_stricmp(nextstr, "monitor_h" )) { monitor_h = SR_NextNum(); n[8] = true; }
				else if (!_stricmp(nextstr, "use43"     )) { cfg.use43 = SR_NextNum(); n[9] = true; }
				else
					noscreenfile = true;

				if (noscreenfile) break;
			}
		}

		if (cMode.width < MIN_SCR_WIDTH || cMode.height < MIN_SCR_HEIGHT)
			noscreenfile = true;

		SR_EndScript();

		if (!noscreenfile)
		{
			for (i = 0; i < 10; i++)
			{
				if (!n[i]) {noscreenfile = true; break; }
			}
		}
	}
	
	if (noscreenfile)
	{
		if ( (_access(screenfile, 4)) != -1)
			remove(screenfile);

		cMode.width  = cMode.max_width;
		cMode.height = cMode.max_height;
		cMode.bpp  = cMode.w_initialbpp;
		cMode.freq = 0;
		cMode.xpos = 0;
		cMode.ypos = 0;
		cMode.fullscreen = true;
		monitor_w = 0;
		monitor_h = 0;

		init_FOV = 0;	// Clear any value change by cfg
	}

	if ( !(f = checkEnumerated()) )
	{
		ScreenError = true;
		init_FOV = 0;
		Con_ErrorD(NULL);
	}

	GL_SetDisplayModes();	// Set the display modes array
	cMode.w_setdefaults = true;	// Ensure the start-up settings are stored

	getDEVMODEparms();
	GL_MainInit();
	GL_PostInit(cMode.width, cMode.height, true, false);

	if (   (monitor_w && monitor_w != scr_max_width)
		|| (monitor_h && monitor_h != scr_max_height) )
	{
		ScreenError = true;
		init_FOV = 0;
		Con_ErrorD(NULL);
	}

	// v220-5 set the aspect ratio string
	{
		int i, p[] = { 2, 3, 5, 0 };
		int	x = scr_max_width, y = scr_max_height;

		// Use ascending primes
		for (i = 0; p[i]; i++)
			while (!(x % p[i]) && !(y % p[i])) x /= p[i], y /= p[i];

		sprintf(aspect_str, "%i:%i", x, y);
	}

	update_gamma = true;

	default_FOV = (int)( ((float)scr_max_height/(float)scr_max_width) * 40.0f + 65.0f );

	if (!init_FOV) init_FOV = default_FOV;
	
	GL_SetRatioAndFov();

	Con_Printf("Desktop resolution    : %d x %d\n", cMode.max_width, cMode.max_height);
	Con_Printf("Fullscreen resolution : %d x %d\n", scr_max_width, scr_max_height);

	// Clear the front and back buffers and draw titlepic to both buffers
	if (aspect_offset_x || aspect_offset_y)
		glViewport(0, 0, cMode.width, cMode.height);

	glClearColor(0.f, 0.f, 0.f, 0.f);
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);	
		
	SwapBuffers(mainHdc);

	glClearColor(0.f, 0.f, 0.f, 0.f);
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	if (aspect_offset_x || aspect_offset_y)
		glViewport(aspect_offset_x, aspect_offset_y, current_w, current_h);

	glMatrixMode(GL_PROJECTION);
	GL_DrawPagePatch(titlepic_num);
	SwapBuffers(mainHdc);
	GL_DrawPagePatch(titlepic_num);
	SwapBuffers(mainHdc);
	glMatrixMode(GL_TEXTURE);
}

//===========================================================================
// GL_InitRefresh
//	Initializes the graphics library for refresh. Also called at update.
//===========================================================================
void GL_InitRefresh(void)
{
	GL_InitFont();	// Initialize font renderer.
	GL_LoadSystemTextures();
}

// v220-3 Frees screen buffer used in the ui
void GL_FreeScreenBuffer()
{
	if (screen_buffer)
	{
		M_Free(screen_buffer); screen_buffer = NULL;
	}
}

//===========================================================================
// GL_Shutdown - called once on exit
//===========================================================================
void GL_Shutdown(void)
{
	HGLRC hglrc;

	if(!gl_inited) return;	// Not yet initialized fully.

	GL_FreeScreenBuffer();
	GL_DeleteAllTextures();
	
	hglrc = wglGetCurrentContext();

	if (hglrc)
	{
		// Delete the rendering context.
		wglMakeCurrent(NULL, NULL);
		wglDeleteContext(hglrc);

		gl_was_inited = true;
	}

	gl_inited = false;
}

float	yfov, aspect;	// v220-4 aspect made global
RDBOOL	reduceNearClip;	// v220-10
RDBOOL	nearestClip;

static void switchTo3DState(RDBOOL pspr_model)
{
	yfov = pspr_model ? cMode.weaponFOV*fieldOfView/(setFOV*aspect) : fieldOfView/aspect;

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(yfov, aspect, nearestClip ? 0.1 : reduceNearClip ? 1.0 : defaultNearClip, 32768.0f);
	glFrontFace(GL_CW);
	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);

	// Note the projection matrix uses a left-handed coordinate system.
	glScalef(1.0f, pspr_model ? 1.2f : 1.0f, -1.0f);
}


//===========================================================================
// GL_RenderFrame() v220-3
//  v220-5 4:3 aspect ratio support added
//===========================================================================
void GL_RenderFrame()
{
extern int		pspr_xoffset;

	int		viewpw = current_w;
	int		viewph = current_h;
	int		psptype;
	int		offx = ge_inf.noHudSpriteScale ? 0 : pspr_xoffset;

	// Push the 2D matrices on the stack.
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();

	if (viewSCRheight != SCREENHEIGHT)
	{
		// change the viewport
		viewph = (int)(viewSCRheight * viewph/200.0f) + 1;
		glViewport(aspect_offset_x, current_h - viewph + aspect_offset_y,
			viewpw, viewph);
	}
	else
	{
		glViewport(aspect_offset_x, aspect_offset_y, viewpw, viewph);
	}

	// Scaling to the monitor is done by altering the aspect ratio
	if (cMode.fullscreen)
	{
		viewpw = (int)((float)viewpw * cMode.scalex);
		viewph = (int)((float)viewph * cMode.scaley);
	}

	// The 3D projection matrix.
	aspect = (float)viewpw/(float)viewph; 	// Assume square pixels
	switchTo3DState(false);					// Change to 3D projection

	if (ppt.listset) RL_RenderAllLists();	// We can now, at last, render the frame
	
	// Clear the matrices
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(offx, 320 - offx, viewSCRheight, 0, -1, 1);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glDisable(GL_DEPTH_TEST);

	// Render the player's weapon
	// psptype flags are set in R_ProjectPlayerSprites,
	R_ProjectPlayerSprites(&psptype);		// Only if 3D models exist for them.
	
	if (psptype & REND_PSPRITE)				// If the 2D versions are needed - R3D
		Rend_DrawPlayerSprites();

	if (psptype & REND_PMODEL)
	{
		switchTo3DState(true);
		Rend_Draw3DPlayerSprites();
	}

	// Restore the 2D matrices
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glDisable(GL_CULL_FACE);
	glDisable(GL_DEPTH_TEST);

	if (viewSCRheight != SCREENHEIGHT)		// Restore viewport. 
		glViewport(aspect_offset_x, aspect_offset_y, current_w, current_h);
}

//===========================================================================
// GL_UseFog
//===========================================================================
void GL_UseFog(int yes)
{
	if (!whitefog && yes)
	{
		// Fog is turned on.
		whitefog = true;
		glEnable(GL_FOG);
	}
	else if (whitefog && !yes)
	{
		// Fog must be turned off.
		whitefog = false;
		glDisable(GL_FOG);
	}
}

//===========================================================================
// G_LoadMenuFog
//===========================================================================
DGLuint G_LoadMenuFog(void)
{
	int lump;

	if (   !menuFogTexture
		&& (lump = W_CheckNumForName("MENUFOG")) >= 0)
	{
		uint size;
		byte *data = (byte*)M_CacheLumpNum(lump, &size, true);

		if (size != 64*64) Con_ErrorD("Invalid menuFogTexture");

		GL_NewTexture(&menuFogTexture);
		GL_TexImage(GL_LUMINANCE, 64, 64, 0, data);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		
		M_Free(data);
	}

	return menuFogTexture;
}

void GL_DeleteFontLists()
{
	if (ttFontList)
	{
		glDeleteLists(ttFontList, 1);
		ttFontList = 0;
	}
}

//===========================================================================
// GL_ChangeResolution
//	Changes the resolution to the specified one. 
//	The change is carried out by shutting down OGL rendering
//	and then restarting it with the new screen dimensions.
//===========================================================================
int GL_ChangeResolution(int w, int h, RDBOOL centre)
{
	GL_FreeScreenBuffer();
	GL_PostInit(w, h, centre, true);
	FR_FreeFonts();
	GL_InitFont();	// Re-initialize fonts to suit new screen size.

	return true;	
}

RDBOOL G_SetScriptedFog(sector_t *s)
{
	if (!s->wvar || !s->wvar->fog_density) return false;
	if (fogval.isset) return true;

	fogval.isset = true;
	fogval.outsidefog    = outsidefog   ; outsidefog = false;
	fogval.underwaterfog = underwaterfog; underwaterfog = true;

	glFogfv(GL_FOG_COLOR , s->wvar->fogrgba);
	glFogf(GL_FOG_DENSITY, s->wvar->fog_density);
	glFogi(GL_FOG_MODE   , GL_EXP);

	shortListQuads = R3D_slopetypes;
	return true;
}

void G_ResetScriptedFog()
{
	if (!fogval.isset) return;

	fogval.isset = false;

	outsidefog    = fogval.outsidefog;
	underwaterfog = fogval.underwaterfog;

	glFogfv(GL_FOG_COLOR , fogval.rgba);
	glFogf(GL_FOG_DENSITY, fogval.density);
	glFogi(GL_FOG_MODE   , fogval.mode);

	shortListQuads = R3D_slopetypes || outsidefog;
	if (outsidefog) GL_UseFog(true);
}

//===========================================================================
// Fog settings - called indirectly to allow values to be printed out
//===========================================================================
void GL_FogvColor(byte *color)
{
	int		i;

	memcpy(fogcolour, color, 3);
	fogcolour[3] = 0xFF;

	for (i = 0; i < 4; i++)
	{
		fogval.rgba[i] = fogcolour[i]/255.0f;
	}

	glFogfv(GL_FOG_COLOR, fogval.rgba);

	Con_Message("fog_colour RGB %i %i %i\n",
		fogcolour[0], fogcolour[1], fogcolour[2]);
}

void GL_FogvMode(uint param)
{
	glFogi(GL_FOG_MODE, param);
	fogval.mode = param;

	Con_Message("fog_mode %s\n",param == GL_LINEAR ?
			"GL_LINEAR" : param == GL_EXP ?
			"GL_EXP" : "GL_EXP2");
}

void GL_FogvDensity(float param)
{
	glFogf(GL_FOG_DENSITY, param);
	fogval.density = param;

	Con_Message("fog_density %.5f\n",param);
}

void GL_FogvStart(float param)
{
	fogval.fogstart = param >= 0 ? param : -param;
	glFogf(GL_FOG_START, fogval.fogstart);

	Con_Message("fog_start %.2f\n",fogval.fogstart);
}

void GL_FogvEnd(float param)
{
	fogval.fogend = param >= 0 ? param : -param;
	if (fogval.fogend < 10.0f) fogval.fogend = 10.0f;
	glFogf(GL_FOG_END, fogval.fogend);

	Con_Message("fog_end %.2f\n", fogval.fogend);
}

//===========================================================================
// GL_GrabScreen
//	Copies the current contents of the frame buffer and returns a pointer
//	to data containing 24-bit RGB triplets. The caller must free the 
//	returned buffer using M_Free()
//  v220-3 rewritten
//===========================================================================
DGLubyte *GL_GrabScreen()
{
	GLubyte	*buffer;
	GLint   bufmem;

	// Multiplier of 3 should be enough but 4 used for safety.
	buffer = (GLubyte*)M_Malloc(current_w * current_h * 4, 13);

	glPixelStorei(GL_PACK_ALIGNMENT  , 1);
	glPixelStorei(GL_PACK_ROW_LENGTH , 0);
	glPixelStorei(GL_PACK_SKIP_ROWS  , 0);
	glPixelStorei(GL_PACK_SKIP_PIXELS, 0);
	glGetIntegerv(GL_READ_BUFFER, &bufmem);
	glReadBuffer(GL_FRONT);
	glReadPixels(aspect_offset_x, aspect_offset_y,
			current_w, current_h, GL_BGR, GL_UNSIGNED_BYTE, buffer);
	glReadBuffer(bufmem);

	return buffer;
}
