//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2009 G. Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: gl_tga.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
//  tga loader rewritten
// Revision 2.2.0 build 6 (v220-8) 14/Aug/2010 gmj
//  simplified. readpixels stored as BGR to eliminate byte reordering
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:  tga file load/save
//
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "de_system.h"
#include "de_graphics.h"
#include "m_args.h"
#include "m_winmem.h"
#include "con_console.h"

void	pr_OFinsert(char *message, ...);

// Save the rgb888 buffer as Targa 24.
int TGA_Save24_rgb888(char *filename, int w, int h, byte *buffer)
{
	FILE 			*file;
    TARGA_HEADER	header;

    if((file=fopen(filename, "wb")) == NULL) return 0; // Huh?

    // Now we have the file. Let's setup the Targa header.
	memset(&header, 0, sizeof(header));

	/* info only - already cleared
	header.idFieldSize = 0;			// No identification field.
    header.colorMapType = 0;		// No color map.
    header.xOrigin = 0;
    header.yOrigin = 0;
	header.imageDescriptor.attributeBits = 0;
    header.imageDescriptor.reserved = 0;		// Must be zero.
	*/

    header.imageType = 2;			// Targa type 2 (unmapped RGB).
    header.imageWidth = w;
    header.imageHeight = h;
	header.imagePixelSize = 24;		// 8 bits per channel.
    header.imageDescriptor.screenOrigin = TGA_SCREEN_ORIGIN_LOWER;	// Must be.
    header.imageDescriptor.dataInterleave = TGA_INTERLEAVE_NONE;

    // Write the header.
    fwrite(&header, sizeof(header), 1, file);
	fwrite(buffer, w*h*3, 1, file);
	fclose(file);
	// A success!
	return 1;
}

int subfn_checkHeader(TARGA_HEADER	*header)
{
	if (   (header->colorMapType != 0 && header->colorMapType != 1)
		|| (header->imageType != 2 && header->imageType != 10)
		|| (header->imagePixelSize != 24 && header->imagePixelSize != 32)
		|| (header->imageDescriptor.attributeBits != 8 && header->imageDescriptor.attributeBits != 0)
		|| header->imageDescriptor.screenOrigin != TGA_SCREEN_ORIGIN_LOWER)
	{
		pr_OFinsert("TGA_LoadFile: unsupported format\n");
		pr_OFinsert("  (datatype=%i cmtype=%ipxsize=%i abits=%i orig=%i)\n",
			header->imageType,
			header->colorMapType,
			header->imagePixelSize,
			header->imageDescriptor.attributeBits,
			header->imageDescriptor.screenOrigin);
		return false;
	}
	return true;
}

//===========================================================================
// TGA_LoadFile888
//	Loads a 24 or a 32-bit (24-bit color + 8-bit alpha) TGA image.
//	Returns a buffer with the image if successful
//
//  v220-4
//   Extended to read type 10 (compressed) files 
//
//	WARNING: This is a limited TGA loader (see subfn_checkHeader)
//           Only type 2 or 10, 24/32 pixel size,
//	         attrbits 0/8 and lower left origin supported.
//===========================================================================

static struct tgadata_t
{
	int		xstep, ystep;
	int		psz, xwidth;
	int		size;
	byte	*pixbuf, *pbptr;

} dat;

// store the pixels in TGA_SCREEN_ORIGIN_LOWER order
static RDBOOL setPixArray(byte *p)
{
	byte	*ptr;
	int		offset;

	if (dat.xstep >= dat.xwidth) { dat.xstep = 0; dat.ystep--; };

	offset = dat.ystep * dat.xwidth + dat.xstep++;

	if (offset < 0 || offset >= dat.size) 
	{
		Con_Message("TGA_LOAD: offset error size %i offset %i", dat.size, offset);
		return false;
	}

	offset *= dat.psz;
	ptr = dat.pixbuf + offset;
	
	// tga's are in bgr or bgra order
	ptr[2] = p[0];
	ptr[1] = p[1];
	ptr[0] = p[2];

	if (dat.psz == 4) ptr[3] = p[3];

	dat.pbptr += dat.psz;

	return true;
}

// NOTE: can only read an actual file - not a lump
//       could be easily modified but R3D does not need this
byte *TGA_LoadFile888(char *filename, int *w, int *h, int *pixSize)
{
    TARGA_HEADER	header;
	DFILE			*file;
	RDBOOL			success = false;
	byte			*buf = NULL, *bufptr;
	byte			*end;
	byte			data[6];				// 5 max bytes to be read +1 to align
	int				read_numbytes;
    int				i, datasize;

	if ( !(file = F_OpenFile(filename, "fb")) ) return NULL;
	
	// Clear any previous pointer (it should have been reclaimed by now)
	dat.pixbuf = NULL,

	// Read and check the header.
	F_Read(&header, sizeof(header), file);

	if (!subfn_checkHeader(&header)) goto abort;
	if ((dat.size = header.imageWidth * header.imageHeight) <= 0) goto abort;

	F_Seek(file, 0, SEEK_END);

	// if there's a colormap then ignore it
	i = header.idFieldSize + header.colorMapType * header.colorMapLength;

	if ( (datasize = F_Tell(file) - sizeof(header) - i) <= 0) goto abort;

	F_Rewind(file);
	F_Seek(file, sizeof(header) + i, SEEK_CUR);

	// Preload the file - this speeds things up considerably
	bufptr = buf = (byte*)M_Malloc(datasize, 0);
	if (F_Read(buf, datasize, file) != datasize) goto abort;

	// Determine format.
	dat.psz = read_numbytes = header.imagePixelSize == 24 ? 3 : 4;

	// set the image buffer
	dat.pixbuf = dat.pbptr = (byte*)M_Calloc(dat.size * 4, 0);
	end = dat.pixbuf + dat.size * dat.psz;

	dat.xwidth = header.imageWidth;
	dat.xstep = 0;
	dat.ystep = header.imageHeight - 1;

	if (header.imageType == 10) read_numbytes++;	// +1 for leading byte

	while (dat.pbptr < end)
	{
		if (bufptr - buf + read_numbytes >= datasize) goto abort;
		memcpy(data, bufptr, read_numbytes);
		bufptr += read_numbytes;

		if (header.imageType == 2)					// Uncompressed
		{
			if (!setPixArray(data)) goto abort;
		}
		else // must be header.imageType == 10		// Compressed
		{
			int bn = data[0] & 0x7f;				// Leading byte identifies what follows
			int d = (data[0] & 0x80) ? 1 : 0;		// 1 means chain of identical pixels

			if (!setPixArray(data + 1)) goto abort;

			for (i = 0; i < bn; i++)
			{
				if (!d)
				{
					if (bufptr - buf + dat.psz >= datasize) goto abort;
					memcpy(data, bufptr, dat.psz);
					bufptr += dat.psz;
				}

				if (!setPixArray(data + d)) goto abort;
			}
		}
	}

	success = true;

abort:;

	F_Close(file);
	if (buf) M_Free(buf);

	if (success)
	{
		*pixSize = dat.psz;
		*w = header.imageWidth;
		*h = header.imageHeight;

		return dat.pixbuf;
	}
	else
	{
		if (dat.pixbuf) M_Free(dat.pixbuf);

		*pixSize = 0; *w = 0; *h =0;

		return NULL;
	}
}

//===========================================================================
// TGA_GetSize
//	Returns true if the file was found and successfully read.
//===========================================================================
int TGA_GetSize(char *filename, int *w, int *h)
{
    TARGA_HEADER header;
	DFILE	*file = F_Open(filename, "fb");
	int		ret;

	if(!file) return 0;

	F_Read(&header, sizeof(header), file);
	ret = subfn_checkHeader(&header);
	F_Close(file);

	if (w) *w = header.imageWidth;
	if (h) *h = header.imageHeight;
	return ret;
}