//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2018 Randy Heit, G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: m_bams.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//  Modified for R3D
// NOTE: from v220-6 these functions are only used where accuracy
//       is unimportant. They have been retained as they confer a
//       small speed advantage.
// Revision 2.2.0 build 12 (v220-12) 01/Oct/2011 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//
// Based on: Doomsday v1.7.8
//           ZDOOM v1.22 December 12, 1999 (dis. marked v1.23beta33)
//
// DESCRIPTION: trigonometric functions. Pre-calculated tables are
//              set up here and can be used where speed is required
//              over accuracy.
//
//-----------------------------------------------------------------------------

#include <memory.h>
#include <math.h>

#include "dd_share.h"
#include "m_bams.h"
#include "doom/tables.h"

#define BAMS_TABLE_ACCURACY_SHIFT	13
#define BAMS_TABLE_ACCURACY			(1 << BAMS_TABLE_ACCURACY_SHIFT)


// 8192 entries covers 0 to 360deg the first quadrant is appended
// to complete the range for the cosine lookup.  This is more
// memory efficient than using two tables
fixed_t finesine[10240];
fixed_t	*finecosine = &finesine[2048];	// sine + 90deg

static binangle_t atantable[BAMS_TABLE_ACCURACY];

//==========================================================================
// bamsInit
//  Based on ZDoom's method for eliminating fixed tables in the source.
//==========================================================================
void bamsInit()
{
	const double	pimul = DPI*2/FINEANGLES;
	int				i;
	
	for (i = 0; i < 10240; i++)
		finesine[i] = (angle_t)(FRACUNIT * sin(i*pimul));

	// v220-33 rounding error frig. Without it sprites' rendering (not models)
	//    is affected if the player's angle is exactly at one of these values.
	//    This becomes more obvious if the player fires a weapon that spawns a
	//    momentary sprite (such are the mysteries of cause and effect).
	finesine[   0] |= 1;
	finesine[2048] |= 1;
	finesine[4096] |= 1;
	finesine[6144] |= 1;
	finesine[8192] |= 1;

	// v220-13 method for preventing fallthrough changed (search fallthrough)

	for (i = 0; i < BAMS_TABLE_ACCURACY; i++)
		atantable[i] = RAD2BANG(atan((double)i/(double)BAMS_TABLE_ACCURACY));
}

binangle_t bamsAtan2(int y, int x)
{
	binangle_t	bang;
	__int64		absy = y, absx = x;	// << TABLE_ACCURACY needs space.

	if(!x && !y) return BANG_0;	// Indeterminate.

	// Make sure the absolute values are absolute.
	if(absy < 0) absy = -absy;
	if(absx < 0) absx = -absx;
	
	// We'll first determine what the angle is in the first quadrant.
	// That's what the tables are for.
	if(!absy) 
		bang = BANG_0;
	else if(absy == absx) 
		bang = BANG_45;	
	else if(!absx) 
		bang = BANG_90;
	else
	{
		// The special cases didn't help. Use the tables.
		// absx and absy can't be zero here.
		if(absy > absx)
			bang = BANG_90-atantable[ (absx << BAMS_TABLE_ACCURACY_SHIFT) / absy ];
		else
			bang = atantable[ (absy << BAMS_TABLE_ACCURACY_SHIFT) / absx ];
	}
		
	// Now we know the angle in the first quadrant.
	// Let's look at the signs and choose the right quadrant.

	if (x < 0) bang = BANG_180 - bang; // Flip horizontally
	if (y < 0) bang = BANG_180 + (BANG_180 - bang); // Flip vertically

	// This is the final angle.
	return bang;
}
