//-----------------------------------------------------------------------------
//
// Copyright (C) 2005 id Software, Inc., Jaakko Kernen
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
//
// $Log: m_misc.c
// 
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Moved some routines here, deleted unused ones.
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
//
// Based on: Doom, Doomsday v1.7.8
//
//
// DESCRIPTION:
//	Random and pseudo random numbers, vector, box, file utilities,
//	TGA Screenshots etc. (i.e. miscellany).
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdlib.h>
#include <io.h>
#include <direct.h>
#include <conio.h>
#include <math.h>

#include "de_base.h"
#include "de_console.h"
#include "de_system.h"
#include "de_graphics.h"
#include "de_refresh.h"
#include "de_misc.h"

#include "r3d_compat.h"
#include "r3d_edit.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

ushort rndindex;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

//==========================================================================
// M_AccurateDistance
//==========================================================================
float M_AccurateDistance(fixed_t dx, fixed_t dy)
{
	if (!dx)
		return FIX2FLT(abs(dy));
	if (!dy)
		return FIX2FLT(abs(dx));
	{
		double fx = (double)FIX2FLT(dx), fy = (double)FIX2FLT(dy);
		return (float)sqrt(fx*fx + fy*fy);
	}
}

//==========================================================================
// R3D M_AccurateDistanceFP used with xlinemem
//==========================================================================
float M_AccurateDistanceFP(float dx, float dy)
{
	if (!dx)
		return FABS(dy);
	if (!dy)
		return FABS(dx);
	{
		double fx = (double)dx, fy = (double)dy;
		return (float)sqrt(fx*fx + fy*fy);
	}
}

//==========================================================================
// M_AccurateLength
//==========================================================================
float M_AccurateLength(line_t *ld)
{
	return M_AccurateDistance(ld->v2->x - ld->v1->x, ld->v2->y - ld->v1->y);
}

//===========================================================================
// M_LineUnitVector
//	Returns a two-component float unit vector parallel to the line.
//===========================================================================
void M_LineUnitVector(line_t *line, float *unitvec)
{
	float dx = FIX2FLT(line->dx), dy = FIX2FLT(line->dy);
	float len = M_ApproxDistancef(dx, dy);

	if(len)
	{
		unitvec[VX] = dx/len;
		unitvec[VY] = dy/len;
	}
	else
	{
		unitvec[VX] = unitvec[VY] = 0;
	}
}

//===========================================================================
// M_ProjectPointOnLinef
//  Input is fixed_t, output is floating point. Gap is the 
//  distance left between the line and the projected point.
//===========================================================================
void M_ProjectPointOnLinef (fixed_t *point, fixed_t *linepoint,
							fixed_t *delta, float gap, float *result)
{
	float line[2] = { FIX2FLT(delta[VX]), FIX2FLT(delta[VY]) };
	float div = DOTPROD2(line, line);

	//---------------------------------

	if (div)
	{
		float pointvec[2] = { 
			FIX2FLT( point[VX] - linepoint[VX] ), 
			FIX2FLT( point[VY] - linepoint[VY] )
		};

		div = DOTPROD2(pointvec, line)/div;

		result[VX] = FIX2FLT(linepoint[VX]) + line[VX] * div;
		result[VY] = FIX2FLT(linepoint[VY]) + line[VY] * div;

		// Is a gap is needed?
		if (gap)
		{
			float diff[2], dist;

			diff[VX] = result[VX] - FIX2FLT(point[VX]);
			diff[VY] = result[VY] - FIX2FLT(point[VY]);

			if ( (dist = M_ApproxDistancef(diff[VX], diff[VY])) != 0)
			{
				result[0] -= diff[0]/dist * gap;
				result[1] -= diff[1]/dist * gap;
			}
		}
	}
	else
	{
		result[0] = FIX2FLT(point[VX]);
		result[1] = FIX2FLT(point[VY]);
	}
}

//===========================================================================
// M_SkipWhite
//===========================================================================
char *M_SkipWhite(char *str)
{
	while(*str && FORMAT_CHAR(*str)) str++;
	return str;
}

//===========================================================================
// M_FindWhite
//===========================================================================
char *M_FindWhite(char *str)
{
	while(*str && !FORMAT_CHAR(*str)) str++;
	return str;
}

//===========================================================================
// M_SkipPlus
//===========================================================================
char *M_SkipPlus(char *str)
{
	while(*str && *str == '+') str++;
	return str;
}

//===========================================================================
// M_FindPlus
//===========================================================================
char *M_FindPlus(char *str)
{
	while(*str && !(*str == '+')) str++;
	return str;
}

//==========================================================================
// M_ExtractFileBase
//==========================================================================
void M_ExtractFileBase(char *path, char *dest)
{
	char *src;
	int length;

	src = path+strlen(path)-1;

	// Back up until a \ or the start
	while(src != path && *(src-1) != '\\' && *(src-1) != '/')
	{
		src--;
	}

	// Copy up to eight characters
	memset(dest, 0, 8);
	length = 0;
	while(*src && *src != '.')
	{
		if(++length == 9)
		{
			Con_Error("Filename base of %s > 8 chars", path);
		}
		*dest++ = toupper((int)*src++);
	}
}

//===========================================================================
// M_ReadLine
//===========================================================================
void M_ReadLine(char *buffer, int len, DFILE *file)
{
	int		p;
	char	ch;

	memset(buffer, 0, len);
	for(p = 0; p < len-1;)	// Make the last null stay there.
	{
		ch = F_GetC(file);
		if(ch == '\r') continue;
		if(deof(file) || ch == '\n') break;
		buffer[p++] = ch;
	}
}

//===========================================================================
// M_IsComment
//===========================================================================
RDBOOL M_IsComment(char *buffer)
{
	int           i = 0;
	
	// skip spaces up to first character not the end
	while(isspace(buffer[i]) && buffer[i]) i++;
	
	if (buffer[i] == '#'  // comment
		|| !buffer[i] )   // jav no data - mainly to get con_exe tracing cleaner from blank lines in cfg file
	{
		return true;     
	}
	
	return false;
}

//===========================================================================
// b211-0 M_ValidateID
//  Check that a map id is valid
//===========================================================================
RDBOOL M_ValidateID(char *id)
{
	int num;

	if (commercial_game)
	{
		return strlen(id) == 5
			&& !strnicmp(id, "map", 3)
			&& (num = atoi(&id[3])) > 0
			&& num <= 99;
	}
	return strlen(id) == 4
		&& (id[0] == 'E' || id[0] == 'e')
		&& (id[2] == 'M' || id[2] == 'm')
		&& (num = id[1] - '0') > 0 && num <= 9
		&& (num = id[3] - '0') > 0 && num <= 9;
}

//===========================================================================
// v220-3 M_SetGlobalYaw
//  Yaw value for things with MFF_SPIN
//  v220-5 Used with df_ flags to make Thing turn through
//         360deg in either 24, 12, 6 or 3 seconds.
//===========================================================================
void M_SetGlobalYaw()
{
	static uint prevtime;

	// v220-5 set to 24 seconds.
	ma_inf.global_yaw = 360.0f * (Sys_GetRealTime() - prevtime)/24000.0f;

	if (ma_inf.global_yaw >= 360.0f) prevtime = Sys_GetRealTime();
}

//===========================================================================
// M_Normalize
//	Normalize a vector. Returns the former length.
//===========================================================================
float M_Normalize(float *a)
{
	float len = (float)(sqrt(a[VX]*a[VX] + a[VY]*a[VY] + a[VZ]*a[VZ]));

	if(len)
	{
		a[VX] /= len;
		a[VY] /= len;
		a[VZ] /= len;
	}
	return len;
}

//===========================================================================
// M_DotProduct
//===========================================================================
float M_DotProduct(float *a, float *b)
{
	return a[VX]*b[VX] + a[VY]*b[VY] + a[VZ]*b[VZ];
}

//===========================================================================
// M_CrossProduct
//	Cross product of two vectors.
//===========================================================================
void M_CrossProduct(float *a, float *b, float *out)
{
	out[VX] = a[VY]*b[VZ] - a[VZ]*b[VY];
	out[VY] = a[VZ]*b[VX] - a[VX]*b[VZ];
	out[VZ] = a[VX]*b[VY] - a[VY]*b[VX];
}

//===========================================================================
// M_PointCrossProduct
//	Cross product of two vectors composed of three points.
//===========================================================================
void M_PointCrossProduct(float *v1, float *v2, float *v3, float *out)
{
	float a[3], b[3];
	int i;

	for(i = 0; i < 3; i++)
	{
		a[i] = v2[i] - v1[i];
		b[i] = v3[i] - v1[i];
	}
	M_CrossProduct(a, b, out);
}

//===========================================================================
// M_RotateVector
//	First yaw, then pitch. Two consecutive 2D rotations.
//	Probably could be done a bit more efficiently.
//===========================================================================
void M_RotateVector(float vec[3], float degYaw, float degPitch)
{
	float	radYaw   = degYaw  /180.0f * PI;
	float	radPitch = degPitch/180.0f * PI;
	float	Cos, Sin, res[3];

	// Yaw.
	if(radYaw != 0)
	{
		Cos = (float)cos(radYaw); 
		Sin = (float)sin(radYaw);
		res[VX] = vec[VX] *  Cos + vec[VY] * Sin;
		res[VY] = vec[VX] * -Sin + vec[VY] * Cos;
		vec[VX] = res[VX];
		vec[VY] = res[VY];
	}
	// Pitch.
	if(radPitch != 0)
	{
		Cos = (float)cos(radPitch);
		Sin = (float)sin(radPitch);
		res[VZ] = vec[VZ] *  Cos + vec[VX] * Sin;
		res[VX] = vec[VZ] * -Sin + vec[VX] * Cos;
		vec[VZ] = res[VZ];
		vec[VX] = res[VX];
	}
}

//===========================================================================
// M_PointUnitLineDistance
//	Line a -> b, point c. The line must be exactly one unit long!
//===========================================================================
float M_PointUnitLineDistance(float *a, float *b, float *c)
{
	return FABS(((a[VY]-c[VY])*(b[VX]-a[VX]) - (a[VX]-c[VX])*(b[VY]-a[VY])));
}

//===========================================================================
// M_PointLineDistance
//	Line a -> b, point c.
//===========================================================================
float M_PointLineDistance(float *a, float *b, float *c)
{
	float d[2], len;

	d[VX] = b[VX] - a[VX];
	d[VY] = b[VY] - a[VY];
	len = (float)(sqrt(d[VX]*d[VX] + d[VY]*d[VY])); // Accurate.
	if(!len) return 0;
	return FABS(((a[VY]-c[VY])*(b[VX]-a[VX]) - (a[VX]-c[VX])*(b[VY]-a[VY]))/len);
}

//===========================================================================
// M_GroupVis v220-4
//===========================================================================
RDBOOL M_GroupVis(subsector_t *ss)
{
	if (R3D_edit && automap_active)
		return !(ppt.view_subsector->ssflags & SSF_NORENDER);
	
	return (ss->viewgroup == plr_viewgroup
		|| !ss->viewgroup
		|| !plr_viewgroup);
}

//===========================================================================
// M_ClearBox
//===========================================================================
void M_ClearBox(fixed_t *box)
{
	box[BOXTOP ] = box[BOXRIGHT] = DDMININT;
	box[BOXBASE] = box[BOXLEFT ] = DDMAXINT;
}

//===========================================================================
// M_AddToBox
//===========================================================================
void M_AddToBox(fixed_t *box, fixed_t x, fixed_t y)
{
	if (x < box[BOXLEFT ]) box[BOXLEFT ] = x;
	else
	if (x > box[BOXRIGHT]) box[BOXRIGHT] = x;

	if (y < box[BOXBASE ]) box[BOXBASE ] = y;
	else
	if (y > box[BOXTOP  ]) box[BOXTOP  ] = y;
}

//===========================================================================
// M_AddLineToBox - R3D
//===========================================================================
void M_AddLineToBox(fixed_t *box, line_t *ld)
{
	M_AddToBox(box, ld->v1->x, ld->v1->y);
	M_AddToBox(box, ld->v2->x, ld->v2->y);
}

//===========================================================================
// M_CompareBox
//===========================================================================
RDBOOL M_CompareBox(fixed_t *inner, fixed_t *outer)
{
	return	   inner[BOXRIGHT] < outer[BOXRIGHT]
			&& inner[BOXLEFT ] > outer[BOXLEFT ]
			&& inner[BOXBASE ] > outer[BOXBASE ]
			&& inner[BOXTOP  ] < outer[BOXTOP  ];
}

//===========================================================================
// M_CompareFPBox
//===========================================================================
RDBOOL M_CompareFPBox(float *inner, float *outer)
{
	return	   inner[BOXRIGHT] < outer[BOXRIGHT]
			&& inner[BOXLEFT ] > outer[BOXLEFT ]
			&& inner[BOXBASE ] > outer[BOXBASE ]
			&& inner[BOXTOP  ] < outer[BOXTOP  ];
}

//===========================================================================
// M_SameBox
//===========================================================================
RDBOOL M_SameBox(fixed_t *inner, fixed_t *outer)
{
	return	   inner[BOXRIGHT] == outer[BOXRIGHT]
			&& inner[BOXLEFT ] == outer[BOXLEFT ]
			&& inner[BOXBASE ] == outer[BOXBASE ]
			&& inner[BOXTOP  ] == outer[BOXTOP  ];
}

//===========================================================================
// M_FitBox
//===========================================================================
RDBOOL M_FitBox(fixed_t *inner, fixed_t *outer)
{
	return	   inner[BOXRIGHT] <= outer[BOXRIGHT]
			&& inner[BOXLEFT ] >= outer[BOXLEFT ]
			&& inner[BOXBASE ] >= outer[BOXBASE ]
			&& inner[BOXTOP  ] <= outer[BOXTOP  ];
}

//===========================================================================
// M_FitFPBox
//===========================================================================
RDBOOL M_FitFPBox(float *inner, float *outer)
{
	return	   inner[BOXRIGHT] <= outer[BOXRIGHT]
			&& inner[BOXLEFT ] >= outer[BOXLEFT ]
			&& inner[BOXBASE ] >= outer[BOXBASE ]
			&& inner[BOXTOP  ] <= outer[BOXTOP  ];
}

//===========================================================================
// M_ComparexyBox (x,y inside the box)
//===========================================================================
RDBOOL M_ComparexyBox(fixed_t x, fixed_t y, fixed_t *outer)
{
	return	   x < outer[BOXRIGHT]
			&& x > outer[BOXLEFT ]
			&& y > outer[BOXBASE ]
			&& y < outer[BOXTOP  ];
}

//===========================================================================
// M_CompareFPxyBox
//===========================================================================
RDBOOL M_CompareFPxyBox(float x, float y, float *outer)
{
	return	   x < outer[BOXRIGHT]
			&& x > outer[BOXLEFT ]
			&& y > outer[BOXBASE ]
			&& y < outer[BOXTOP  ];
}

//===========================================================================
// M_SetFPBoxFromIntBox
//===========================================================================
void M_SetFPBoxFromIntBox(float *fpbox, fixed_t *intbox)
{
	fpbox[BOXRIGHT] = FIX2FLT(intbox[BOXRIGHT]);
	fpbox[BOXLEFT ] = FIX2FLT(intbox[BOXLEFT ]);
	fpbox[BOXBASE ] = FIX2FLT(intbox[BOXBASE ]);
	fpbox[BOXTOP  ] = FIX2FLT(intbox[BOXTOP  ]);
}

//===========================================================================
// M_AreaBox
//===========================================================================
int M_AreaBox(fixed_t *box)
{
	return	((box[BOXRIGHT]>>16) - (box[BOXLEFT]>>16)) *
			((box[BOXTOP  ]>>16) - (box[BOXBASE]>>16));
}

//===========================================================================
// M_AreaFPBox
//===========================================================================
float M_AreaFPBox(float *box)
{
	return	(float)	((int)box[BOXRIGHT] - (int)box[BOXLEFT]) *
					((int)box[BOXTOP  ] - (int)box[BOXBASE]);
}

//===========================================================================
// M_FPAreaBox
//===========================================================================
float M_FPAreaBox(fixed_t *box)
{
	return	(float)(box[BOXRIGHT] - box[BOXLEFT]) *
			(float)(box[BOXTOP  ] - box[BOXBASE]);
}

//===========================================================================
// Offset routines - limit extension changes R3D
//===========================================================================
RDBOOL M_FitSSECBox(fixed_t *box, fixed_t fx, fixed_t fy, fixed_t fr)
{
	if (box[BOXLEFT ] > fx - fr) return false;
	if (box[BOXRIGHT] < fx + fr) return false;
	if (box[BOXBASE ] > fy - fr) return false;
	if (box[BOXTOP  ] < fy + fr) return false;
	return true;
}

void M_setBoxOffsets(int *box, int x, int y, int radius)
{
	box[BOXRIGHT] = M_AddLIMIT(x,  radius);
	box[BOXLEFT ] = M_AddLIMIT(x, -radius);
	box[BOXTOP  ] = M_AddLIMIT(y,  radius);
	box[BOXBASE ] = M_AddLIMIT(y, -radius);
}

void M_getBlockOffsets(int *data, int adjust, int *xl, int *xh, int *yl, int *yh)
{
	*xl = MAPBS3(data[BOXLEFT ], bmaporgx, -adjust);
	*xh = MAPBS3(data[BOXRIGHT], bmaporgx,  adjust);
	*yl = MAPBS3(data[BOXBASE ], bmaporgy, -adjust);
	*yh = MAPBS3(data[BOXTOP  ], bmaporgy,  adjust);

	if (*xl < 0) *xl = 0;
	if (*yl < 0) *yl = 0;
	if (*xh >= bmapwidth ) *xh = bmapwidth  - 1;
	if (*yh >= bmapheight) *yh = bmapheight - 1;
}

int M_AddLIMIT(__int64 a, __int64 b)
{
	__int64 res = a + b;
	
	if (res > DDMAXINT) return DDMAXINT;
	if (res < DDMININT) return DDMININT;

	return (int)res;
}

//===========================================================================
// M_ApproxDistancef
//===========================================================================
float M_ApproxDistancef(float dx, float dy)
{
	dx = FABS(dx); 
	dy = FABS(dy);
	if(dx < dy) return dx + dy - dx/2;
	return dx + dy - dy/2;
}

//===========================================================================
// v220-13 DeEscapeString
//===========================================================================
#if 0	// not currently used
void DeEscapeString(char *str)
{
	char		*buf, *out, *in;
	unsigned	len;

	if (!str)
		Con_ErrorD("ReplaceInString: sent unset string");

	len = strlen(str) + 1;
	buf = (char*)M_Calloc(len, 205);

	for (out = buf, in = str; *in; out++, in++)
	{
		if (*in == '\\') 
		{
			in++;
			if      (*in == 'n') *out = '\n';				// Newline.
			else if (*in == 'r') *out = '\r';				// Carriage return.
			else if (*in == 't') *out = '\t';				// Tab.
			else if (*in == '_' || *in == 's') *out = ' ';	// Space.
			else *out = *in;
			continue;
		}
		*out = *in;
	}

	strcpy(str, buf);

	M_Free(buf);
}
#endif

//===========================================================================
//  v220-9 case insensitive version of strstr
//===========================================================================
RDBOOL strstrCI(char *str, char *charset)
{
	size_t	i, len, ret = 0;
	char	*src = str;
	char	*in, *cset;

	//-----------------------

	if (!str) return false;

	len = strlen(str);

	if (!len || !charset || len < strlen(charset)) return false;

	in = (char*)calloc(len + 1, 1);

	for (i = 0; i < len; i++)
	{	in[i] = toupper((int)*src++);
	}

	len = strlen(charset);

	cset = (char*)calloc(len + 1, 1);

	for (i = 0; i < len; i++)
	{	cset[i] = toupper((int)*charset++);
	}

	ret = strstr(in, cset) != NULL;

	M_Free(cset);
	M_Free(in);

	return ret;
}

//===========================================================================
// M_WriteFile
//===========================================================================

#ifndef O_BINARY
#define O_BINARY 0
#endif

RDBOOL M_WriteFile (char *name, void *source, int length)
{
	int handle, count;

	handle = open (name, O_WRONLY | O_CREAT | O_TRUNC | O_BINARY, 0666);
	if (handle == -1)
		return false;
	count = write (handle, source, length);
	close (handle);

	if (count < length)
		return false;

	return true;
}

void M_WriteCommented(FILE *file, char *text)
{
	char *buff = (char*)M_Malloc(strlen(text)+1, 24), *line;

	strcpy(buff, text);
	line = strtok(buff, "\n");
	while(line)
	{
		fprintf(file, "# %s\n", line);
		line = strtok(NULL, "\n");
	}
	M_Free(buff);
}

//===========================================================================
// M_WriteTextEsc
//	The caller must provide the opening and closing quotes.
//===========================================================================
void M_WriteTextEsc(FILE *file, char *text)
{
	int		i;

	for(i = 0; text[i]; i++)
	{
		if(text[i] == '"' || text[i] == '\\') fprintf(file, "\\");
		fprintf(file, "%c", text[i]);
	}
}

//===========================================================================
//	Writes a Targa file of the specified depth.
//   Currently only called with 24bits
//===========================================================================
static RDBOOL subfn_ScreenShot(char *filename, int bits)
{
	DGLubyte	*screen;

	if (bits != 24) return false;

	screen = GL_GrabScreen();
	TGA_Save24_rgb888(filename, current_w, current_h, screen);	
	M_Free(screen);

	return true;
}

//===========================================================================
// M_DoScreenShot - Rewritten (was G_DoScreenShot()) R3D
//===========================================================================
int M_DoScreenShot()
{
	int		i, len;
	char	fname[256];

	sprintf(fname, "%sDOOM00.TGA", ddRuntimeDir.path);	// Find a file name
	len = strlen(fname);

	for (i = 0; i < 100; i++)
	{
		fname[len-6] = i/10 + '0';
		fname[len-5] = i%10 + '0';

		if ( (_access(fname, 0)) == -1)
			break;  // found an unused filename
	}

	if (i < 100 && subfn_ScreenShot(fname, 24)) return i;

	return -1;
}

//===========================================================================
// M_strcatQuoted
//===========================================================================
void M_strcatQuoted(char *dest, char *src)
{
	int	k = strlen(dest)+1, i;

	strcat(dest, "\"");
	for(i=0; src[i]; i++)
	{
		if(src[i] == '"') 
		{
			strcat(dest, "\\\"");
			k += 2;
		}
		else
		{
			dest[k++] = src[i];
			dest[k] = 0;
		}
	}
	strcat(dest, "\"");
}

//===========================================================================
// v220-6 Copy file. If infile == outfile the infile dates are not preserved
//===========================================================================
void M_CopyFile(char *infile, char *outfile)
{
	FILE	*file;
	uint	num;

	if ( (file = fopen(infile, "rb")) != NULL)
	{
		char *filebuff;
		uint size;

		fseek(file, 0, SEEK_END);
		size = ftell(file);
		rewind(file);

		if (!size)
		{
			fclose(file);
			Con_Error("File has no data: %s", infile);
		}

		if ( !(filebuff = (char*)malloc(size)) )
		{
			fclose(file);
			Con_ErrorD("ERROR: Insufficient memory to create backup");
		}

		num = fread(filebuff, 1, size, file);
		fclose(file);
		if (num != size) Con_Error("ERROR: read error with %s", infile);

		if (!_stricmp(infile, outfile))
			remove(infile);

		if ( (file = fopen(outfile, "wb")) != NULL)
		{
			num = fwrite(filebuff, 1, size, file);
			fclose(file);

			if (num != size)
				Con_Error("ERROR: failed to write to %s", outfile);
		}

		M_Free(filebuff);
	}
	else
		Con_Error("ERROR: could not open %s", infile);
}

//===========================================================================
// M_PrependBasePath
//  b211-0 revised
//  NOTE: caller must ensure that newpath is at least size MAXPATH
//===========================================================================
void M_PrependBasePath(char *path, char *newpath)
{
	if (Dir_IsAbsolute(path))		// Can't prepend to absolute paths
	{
		strcpy(newpath, path);
	}
	else if (strlen(ddRuntimeDir.path) + strlen(path) < MAXPATH)
	{
		char *ptr = path;

		while (*ptr == '\\' || *ptr == '/') ptr++;

		sprintf(newpath, "%s%s", ddRuntimeDir.path, ptr);
	}
	else
		Con_Error("M_PrependBasePath: path length %s%s too long",
												ddRuntimeDir.path, path);

    Dir_ConvertSlashes(newpath);

}

//===========================================================================
// M_PrependBaseToFile b211-0
//  Adds base path if necessary, converts slashes
//  and verifies fpath has a filename
//  NOTE: caller must ensure that fpath is at least size MAXPATH
// (should be MAXFNAMEX but this is the original ded limit and to
//  save wasting space I have kept it as is.  The user will have to
//  avoid using excessively long paths + filenames).
//===========================================================================
RDBOOL M_PrependBaseToFile(char *fpath)
{
	char	szdrive[MAXDRIVE];
	char	szfilename[MAXFNAME];
	char	temp[MAXPATH];

    Dir_ConvertSlashes(fpath);

	strcpy(temp, fpath);

	_splitpath(fpath, szdrive, NULL, szfilename, NULL);

	if (!szfilename[0]) return false;

	if (!szdrive[0])
	{
		char *ptr = fpath;

		while (*ptr == '\\') ptr++;

		if (strlen(ddRuntimeDir.path) + strlen(ptr) >= MAXPATH)
			Con_Error("M_PrependBasePath: path length %s%s too long",
													ddRuntimeDir.path, ptr);

		sprintf(temp, "%s%s", ddRuntimeDir.path, ptr);
	}

	strcpy(fpath, temp);
	return true;
}

//===========================================================================
// M_TranslatePath
//	Expands >.
//===========================================================================
void M_TranslatePath(char *path, char *translated)
{
	char buf[MAXPATH];	// b211-0
	
	if (path[0] == '>' || path[0] == '}')
	{
		path++;

		if(!Dir_IsAbsolute(path))
			M_PrependBasePath(path, buf);
		else
			strcpy(buf, path);
	}
	else
	{
		strcpy(buf, path);
	}

    Dir_ConvertSlashes(buf);    // 12-23-2005
	strcpy(translated, buf);
}

//===========================================================================
// M_FileExists
//	Also checks for '>'.
//	The file must be a *real* file!
//===========================================================================
int M_FileExists(char *file)
{
	char buf[MAXPATH];

	M_TranslatePath(file, buf);
	return !access(buf, 4); // Read permission?
}

//===========================================================================
// M_CheckPath
//	Check that the given directory exists. If it doesn't, create it.
//	The path must be relative!
//	Return true if the directory already exists.
//===========================================================================
RDBOOL M_CheckPath(char *path)
{
	char full[256];
	char buf[256], *ptr, *endptr;

	if(!access(path, 0)) return true; // Quick test.

	// Convert all backslashes to normal slashes.
	strcpy(full, path);
	while((ptr = strchr(full, '\\')) != 0) *ptr = '/';

	// Check and create the path in segments.
	ptr = full;
	memset(buf, 0, sizeof(buf));
	for(;;)
	{
		endptr = strchr(ptr, '/');
		if(!endptr) strcat(buf, ptr); else strncat(buf, ptr, endptr-ptr);
		if(access(buf, 0))
		{
			// Path doesn't exist, create it.
			mkdir(buf);		
		}
		strcat(buf, "\\");
		ptr = endptr + 1;
		if(!endptr) break;
	}
	return false;
}

const byte rndtable[256] = {
	  0,   8, 109, 220, 222, 241, 149, 107,  75, 248, 254, 140,  16,  66,  74,  21,
	211,  47,  80, 242, 154,  27, 205, 128, 161,  89,  77,  36,  95, 110,  85,  48,
	212, 140, 211, 249,  22,  79, 200,  50,  28, 188,  52, 140, 202, 120,  68, 145,
	 62,  70, 184, 190,  91, 197, 152, 224, 149, 104,  25, 178, 252, 182, 202, 182,
	141, 197,   4,  81, 181, 242, 145,  42,  39, 227, 156, 198, 225, 193, 219,  93,
	122, 175, 249,   0, 175, 143,  70, 239,  46, 246, 163,  53, 163, 109, 168, 135,
	  2, 235,  25,  92,  20, 145, 138,  77,  69, 166,  78, 176, 173, 212, 166, 113,
	 94, 161,  41,  50, 239,  49, 111, 164,  70,  60,   2,  37, 171,  75, 136, 156,
	 11,  56,  42, 146, 138, 229,  73, 146,  77,  61,  98, 196, 135, 106,  63, 197,
	195,  86,  96, 203, 113, 101, 170, 247, 181, 113,  80, 250, 108,   7, 255, 237,
	129, 226,  79, 107, 112, 166, 103, 241,  24, 223, 239, 120, 198,  58,  60,  82,
	128,   3, 184,  66, 143, 224, 145, 224,  81, 206, 163,  45,  63,  90, 168, 114,
	 59,  33, 159,  95,  28, 139, 123,  98, 125, 196,  15,  70, 194, 253,  54,  14,
	109, 226,  71,  17, 161,  93, 186,  87, 244, 138,  20,  52, 123, 251,  26,  36,
     17,  46,  52, 231, 232,  76,  31, 221,  84,  37, 216, 165, 212, 106, 197, 242,
	 98,  43,  39, 175, 254, 145, 190,  84, 118, 222, 187, 136, 120, 163, 236, 249
};

//==========================================================================
// M_RandomD Returns a 0-255 number
//  A deterministic pseudo random table to preserve
//  demo sync (not actually required in Risen3D)
//==========================================================================
int M_RandomD()
{
	if (++rndindex > 255) rndindex = 0;
	return (int)rndtable[rndindex];
}
/*
//returns a 0-255
int M_Random()
{
	return rand() % 256;
}
*/
// return fp value from  -1.f to 1.f
// Used only for particles and sounds
float M_FRandom()
{
	return (float)((rand() % 256) | ((rand() % 256) << 8)) / 65535.0f;
}
