//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
//
// $Log: m_string.c
// 
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//  Modified for R3D, unused routines edited out
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:
//  Simple dynamic string management (like MFC's CString, but not 
//  nearly as complex). Uses standard win memory routines
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <string.h>

#include "de_base.h"
#include "de_misc.h"
#include "m_winmem.h"

// MACROS ------------------------------------------------------------------

//#define REALLOC_AMOUNT	64		

// CODE --------------------------------------------------------------------

//===========================================================================
// Str_Init
//	Call this for uninitialized strings. Global variables are 
//	automatically cleared, so they don't need initialization.
//===========================================================================
void Str_Init(ddstring_t *ds)
{
	memset(ds, 0, sizeof(*ds));	
}

//===========================================================================
// Str_Free
//	Empty an existing string. Totally destroys it. The string is 
//	considered re-initialized.
//===========================================================================
void Str_Free(ddstring_t *ds)
{
	if(ds->size) 
	{
		// The string has memory allocated, free it.
		M_Free(ds->str);
	}
	memset(ds, 0, sizeof(*ds));
}

//===========================================================================
// Str_Alloc
//  for_length must be the current length + new bytes needed
//===========================================================================
static void Str_Alloc(ddstring_t *ds, int for_length)
{
	char *ptr;

	if (ds->size > for_length)
		return; // Enough memory already allocated

	// Add additional amount to possibly accomodate
	// further strings being appended later.
	// Must at least be +1 for the terminating NULL
	for_length += 512;

	// Allocate memory
	ptr = (char*)M_Calloc(for_length, 234);

	if (ds->length)
	{
		memcpy(ptr, ds->str, ds->length);
		M_Free(ds->str);
	}

	ds->str = ptr;
	ds->size = for_length;						// Record allocation size
}

//===========================================================================
// Str_AppendComma
//===========================================================================
void Str_AppendComma(ddstring_t *ds, const char *append_text)
{
	int	len = ds->length + strlen(append_text) + 1;

	Str_Alloc(ds, len + 1);
	strcat(ds->str, append_text);
	strcat(ds->str, ",");
	ds->length = len;
}

//===========================================================================
// Str_TokenMatch - R3D
//  case insensitive search for a string in a ',' tokenised list
//===========================================================================
RDBOOL Str_TokenMatch(ddstring_t *ds, const char *new_text)
{
	uint len;

	if (ds->str && (len = strlen(ds->str)) > 0 && strlen(new_text))
	{
		char *token;
		char *str = (char*)M_Calloc(len + 1, 235);	// Must be calloc

		memcpy(str, ds->str, len);
		token = strtok(str, ",");
		
		while(token)
		{
			if (!_stricmp(new_text, token))
			{
				M_Free(str);
				return true;
			}
			token = strtok(NULL, ",");
		}
		M_Free(str);
	}

	return false;
}

//===========================================================================
// Str_Text
//	This is safe for all strings.
//===========================================================================
char *Str_Text(ddstring_t *ds)
{
	return ds->str? ds->str : "";
}
