//-----------------------------------------------------------------------------
//
// Copyright (C) 1998 Lee Killough
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_blockmap.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Lee Killough's blockmap code imported to this file and then
//  adapted (but not altered functionally) for Risen3D.
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
// 
//
// DESCRIPTION: builds blockmap
//
// Lee Killough's solution - adapted for use with Risen3D.exe

// --------- Lee Killough's comments -------------
//
// killough 10/98:
//
// Rewritten to use faster algorithm.
//
// New procedure uses Bresenham-like algorithm on the linedefs, adding the
// linedef to each block visited from the beginning to the end of the linedef.
//
// ------------------------------------------------


// HEADER FILES ------------------------------------------------------------

#include <string.h>

#include "de_base.h"
#include "de_console.h"
#include "de_play.h"
#include "de_refresh.h"
#include "sys_timer.h"
#include "m_args.h"
#include "m_misc.h"
#include "r3d_compat.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// b211 - might as well go here as anywhere else
int			numvertexes;
vertex_t	*vertexes;

int			numsegs;
seg_t		*segs;

int			numsectors;
sector_t	*sectors;
skyline_t	*skl;			// v220-35 placed here as related to sectors

int			numsubsectors;
subsector_t	*subsectors;

int			numnodes;
node_t		*nodes;

int			numlines;
line_t		*lines;

int			numsides;
side_t		*sides;

long		*blockmaplump;			// offsets in blockmap are from here
long		*blockmap;
int			bmapwidth, bmapheight;	// in mapblocks
fixed_t		bmaporgx, bmaporgy;		// origin of block map
linkmobj_t	*blockrings;			// for thing rings


int			maplimits[4];

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

// v220-31 R3D bug fixed where greatest max/min limits may not have been set
//         blockrings now use static memory resized as needed
void P_CreateBlockmap()
{
	static		int maxblockmapmem = 0;

	vertex_t	*v = vertexes;
	int			minx, miny, maxx, maxy;
	int			i, x, y;

	// First find limits of map
	minx = maxx = v->x >> FRACBITS;
	miny = maxy = v->y >> FRACBITS;
	
	for (i = 1; i < numvertexes; i++, v++)
	{
		x = v->x >> FRACBITS, y = v->y >> FRACBITS;

		if (x < minx) minx = x; else if (x > maxx) maxx = x;
		if (y < miny) miny = y; else if (y > maxy) maxy = y;
	}
	
	// Save blockmap parameters
	bmaporgx = minx << FRACBITS;
	bmaporgy = miny << FRACBITS;

	bmapwidth  = ((maxx-minx) >> MAPBTOFRAC) + 1;
	bmapheight = ((maxy-miny) >> MAPBTOFRAC) + 1;

	maplimits[BOXLEFT ] = minx;
	maplimits[BOXRIGHT] = maxx;
	maplimits[BOXBASE ] = miny;
	maplimits[BOXTOP  ] = maxy;
	
	pr_OFdirect("\nMAP DIMENSIONS:\n");
	pr_OFinsert("  minx %i maxx %i miny %i maxy %i\n", minx, maxx, miny, maxy);
	pr_OFinsert("  block: width %i height %i\n", bmapwidth, bmapheight);

	// Compute blockmap, which is stored as a 2d array of variable-sized lists.
	//
	// Pseudocode:
	//
	// For each linedef:
	//
	//   Map the starting and ending vertices to blocks.
	//
	//   Starting in the starting vertex's block, do:
	//
	//     Add linedef to current block's list, dynamically resizing it.
	//
	//     If current block is the same as the ending vertex's block, exit loop.
	//
	//     Move to an adjacent block by moving towards the ending block in 
	//     either the x or y direction, to the block which contains the linedef.
	
	{
		typedef struct { int n, nalloc, *list; } bmap_t;	// blocklist structure

		int		tot = bmapwidth * bmapheight;				// size of blockmap
		bmap_t	*bmap = (bmap_t*)calloc(tot, sizeof(*bmap));			// array of blocklists
		line_t	*ld = lines;
		
		for (i = 0; i < numlines; i++, ld++)
		{
			// starting coordinates
			int x = (ld->v1->x >> FRACBITS) - minx;
			int y = (ld->v1->y >> FRACBITS) - miny;
			
			// x-y deltas
			int adx = ld->dx >> FRACBITS, dx = adx < 0 ? -1 : 1;
			int ady = ld->dy >> FRACBITS, dy = ady < 0 ? -1 : 1; 
			
			// difference in preferring to move across y (>0) instead of x (<0)
			int diff = !adx ? 1 : !ady ? -1 :
				(((x >> MAPBTOFRAC) << MAPBTOFRAC) + 
				(dx > 0 ? MAPBLOCKUNITS-1 : 0) - x) * (ady = abs(ady)) * dx -
				(((y >> MAPBTOFRAC) << MAPBTOFRAC) + 
				(dy > 0 ? MAPBLOCKUNITS-1 : 0) - y) * (adx = abs(adx)) * dy;
			
			// starting block, and pointer to its blocklist structure
			int b = (y >> MAPBTOFRAC)*bmapwidth + (x >> MAPBTOFRAC);
			
			// ending block
			int bend = (((ld->v2->y >> FRACBITS) - miny) >> MAPBTOFRAC) *
				bmapwidth + (((ld->v2->x >> FRACBITS) - minx) >> MAPBTOFRAC);
			
			// delta for pointer when moving across y
			dy *= bmapwidth;
			
			// deltas for diff inside the loop
			adx <<= MAPBTOFRAC;
			ady <<= MAPBTOFRAC;
			
			// Now we simply iterate block-by-block until we reach the end block.
			while (b >= 0 && b < tot)    // failsafe -- should ALWAYS be true
			{
				// Increase size of allocated list if necessary
				if (bmap[b].n >= bmap[b].nalloc)
				{
					bmap[b].nalloc = bmap[b].nalloc ? bmap[b].nalloc*2 : 8;
					bmap[b].list = (int*)realloc(bmap[b].list, bmap[b].nalloc * sizeof(*bmap->list));
				}
				
				// Add linedef to end of list
				bmap[b].list[bmap[b].n++] = i;
				
				// If we have reached the last block, exit
				if (b == bend) break;
				
				// Move in either the x or y direction to the next block
				if (diff < 0) 
					diff += ady, b += dx;
				else
					diff -= adx, b += dy;
			}
		}
		
		// Compute the total size of the blockmap.
		//
		// Compression of empty blocks is performed by reserving two offset words
		// at tot and tot+1.
		//
		// 4 words, unused if this routine is called, are reserved at the start.
		
		{
			int i, count = tot+6;  // we need at least 1 word per block, plus reserved's
			
			for (i = 0; i < tot; i++)
			{
				if (bmap[i].n)
					count += bmap[i].n + 2; // 1 header word + 1 trailer word + blocklist
			}
				
			// Allocate blockmap lump with computed count
			blockmaplump = (long*)W_Calloc(sizeof(*blockmaplump) * count, LEVELID, 3);
		}									 
		
		// Now compress the blockmap.
		{
			int		i, ndx = tot += 4;	// Advance index to start of linedef lists
			bmap_t	*bp = bmap;			// Start of uncompressed blockmap
			
			blockmaplump[ndx++] = 0;	// Store an empty blockmap list at start
			blockmaplump[ndx++] = -1;	// (used for compression)
			
			for (i = 4; i < tot; i++, bp++)
			{
				if (bp->n)						// Non-empty blocklist
				{
					blockmaplump[ndx] = 0;		// Store index & header
					blockmaplump[i] = ndx++;

					do							// Copy linedef list
					{
						blockmaplump[ndx++] = bp->list[--bp->n];
					}
					while (bp->n);

					blockmaplump[ndx++] = -1;	// Store trailer
					free(bp->list);				// Free linedef list
				}
				else            // Empty blocklist: point to reserved empty blocklist
					blockmaplump[i] = tot;
			}
				
		}
		
		free(bmap);    // Free uncompressed blockmap
	}

	blockmap = blockmaplump + 4;
	
	// Initialise linkmobj blockrings.
	int maxblockmemneeded = sizeof(linkmobj_t) * bmapwidth * bmapheight;

	if (maxblockmemneeded > maxblockmapmem)
	{
		maxblockmapmem = maxblockmemneeded;
		blockrings = (linkmobj_t*)W_Realloc(blockrings, maxblockmemneeded, STATICID, 4);
	}

	for (i = 0; i < bmapwidth * bmapheight; i++)
		blockrings[i].next = blockrings[i].prev = (mobj_t*) &blockrings[i];
}
