//-----------------------------------------------------------------------------
//
// Copyright (C) 1998 - 2005 id Software Inc., Chi Hoang, Lee Killough,
//   Jim Flynn, Rand Phares, Ty Halderman, Jaakko Kernen
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_maputil.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//  Modified/extended for Risen3D
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
//  Code unchanged only comments changed
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//
// Based on: Doom, Doomsday v1.7.8 and BOOM v2.01
//
//
// DESCRIPTION:
//	Various functions needed for move positioning.
//
// Note: elegant line and block linking code was introduced by Jaakko
//       Kernen into DoomsDay to replace method used in MBF/Boom.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <math.h>

#include "de_base.h"
#include "de_play.h"
#include "de_refresh.h"

#include "doom\p_local.h"

#include "con_console.h"
#include "m_bams.h"
#include "m_misc.h"
#include "r3d_edit.h"
#include "r3d_MAINspec.h"
#include "r3d_slopes.h"
#include "s_main.h"

// MACROS ------------------------------------------------------------------

// Linkstore is list of pointers gathered when iterating stuff.
// This is pretty much the only way to avoid *all* potential problems
// caused by callback routines behaving badly (moving or destroying 
// things). The idea is to get a snapshot of all the objects being
// iterated before any callbacks are called. The hardcoded limit is
// a drag, but it would be highly unlikely to exceed 2048 things/lines
// in one block or sector.

// R3D code added to prevent linkstore overflow

#define NP_DISMISS(pile, node) (pile.nodes[node].object_ptr = 0)
#define MAXLINKED		2048

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

linklist_t	link_list;
divline_t	trace;
mobj_t		*shootthing;
mobj_t		*slidemo;

// b211 used with 3D lines - difference between sliding and blocking
RDBOOL		is_sliding;

const double rad2ang = ((double)ANG180)/DPI;

// v220-22
int			use_asmnode;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

//===========================================================================
// P_ApproxDistance
//	Gives an estimation of distance (not exact).
//===========================================================================
fixed_t P_ApproxDistance (fixed_t dx, fixed_t dy)
{
	dx = abs(dx), dy = abs(dy);

	return (dx < dy ? dx + dy - (dx>>1) : dx + dy - (dy>>1));
}

//===========================================================================
// P_ApproxDistance3
//	Gives an estimation of 3D distance (not exact).
//	The Z axis aspect ratio is corrected.
//===========================================================================
fixed_t P_ApproxDistance3(fixed_t dx, fixed_t dy, fixed_t dz)
{
	return P_ApproxDistance(P_ApproxDistance(dx, dy), 
		FixedMul(dz, (int)1.2f*FRACUNIT));
}

//===========================================================================
// b220-0 P_FPAccurateDistance3 - returns accurate 3D distance as a float
//===========================================================================
static float P_FPAccurateDistance3(mobj_t *start,
						  double dx, double dy, double dz)
{
	// If this is a mobj, center the Z pos.
//	if (start && start->thinker.function == (void (__cdecl *)(void))P_MobjThinker)
//		dz += (double)(start->height >> 17);

	return (float)sqrt(dx*dx + dy*dy + dz*dz);
}

//===========================================================================
// P_MobjSoundDistancef v220-3
//	Either end or fixpoint must be specified.
//	The distance is measured in 3D.
//  This is only used for sound distances from the player
//===========================================================================
float P_MobjSoundDistancef(mobj_t *end, float *fixpoint, float attenuation)
{
	mobj_t *start = ppt.viewplrmo;

	if (!attenuation) return 0;

	if (start)
	{
		if (end)		// Start -> end
		{
			if (end == start) return 0;	// player sound

			return P_FPAccurateDistance3(start,
				FIX2FLT(end->x - start->x) * attenuation,
				FIX2FLT(end->y - start->y) * attenuation,
				FIX2FLT(end->z - start->z) * attenuation);
		}

		if (fixpoint)	// Start -> fixpoint
		{
			return P_FPAccurateDistance3(NULL,
				(fixpoint[VX] - FIX2FLT(start->x)) * attenuation,
				(fixpoint[VY] - FIX2FLT(start->y)) * attenuation,
				(fixpoint[VZ] - FIX2FLT(start->z)) * attenuation);
		}
	}

	return 0;
}

//
// P_PointOnNodeSide
//  Returns 0 or 1
//
int P_PointOnNodeSide (fixed_t x, fixed_t y, const node_t *node)
{
	if (use_asmnode)
		return asmPointOnSide(y - node->y, node->dx, node->x - x, node->dy) > 0;

	if (!node->dx)
		return x <= node->x ? node->dy > 0 : node->dy < 0;
	
	if (!node->dy)
		return y <= node->y ? node->dx < 0 : node->dx > 0;
	
	x -= node->x;
	y -= node->y;
	
	// Try to quickly decide by looking at sign bits.
	if ((node->dy ^ node->dx ^ x ^ y) < 0)
		return (node->dy ^ x) < 0;  // (left is negative)

	return FixedMul(y, node->dx>>FRACBITS) >= FixedMul(node->dy>>FRACBITS, x);
}

//
// P_PointOnLineSide - Returns 0 (front or on line) or 1 (back)
//
int P_PointOnLineSide (fixed_t x, fixed_t y, const line_t *line)
{
	return asmPointOnSide(y - line->v1->y, line->dx, line->v1->x - x, line->dy) > 0;
}

//
// P_PointOnViewSide - Returns 0 (front or on line) or 1 (back)
//
int P_PointOnViewSide (float x, float y, float dx, float dy)
{
	return asmPointOnSide(viewy - (int)(y)*FRACUNIT, (int)(dx)*FRACUNIT, (int)(x)*FRACUNIT - viewx, (int)(dy)*FRACUNIT) > 0;
}

//
// P_CheckOnLineSide
//   Returns 0 or 1 if either side of the line or 2 if on the line
//
int P_CheckOnLineSide (fixed_t x, fixed_t y, const line_t *line)
{
	int ret = asmPointOnSide(y - line->v1->y, line->dx, line->v1->x - x, line->dy);

	return (!ret ? 2 : ret > 0);
}

// v220-3 moved from r_util
//===========================================================================
//  b211-0 replacement functions
//  The pre-calculated tantoangle table has been removed. Testing with recent
//  CPU's revealed that the old code made no discernable speed difference.

// Get angle to x, y from the viewer. This is only called after
//        dbViewx and dbViewy have been set (in R_SetViewPos())
angle_t R_PointToAngle(fixed_t x, fixed_t y)
{
extern double	dbViewx;
extern double	dbViewy;

	return (angle_t)(atan2(double(y) - dbViewy, double(x) - dbViewx) * rad2ang);
}

// Get angle to x, y from px, py
angle_t R_PointToAngle2(fixed_t px, fixed_t py, fixed_t x, fixed_t y)
{
	return (angle_t)(atan2 (double(y) - double(py),
							double(x) - double(px)) * rad2ang);
}

//===========================================================================
// R_PointInSubsector
//===========================================================================
subsector_t *R_PointInSubsector(fixed_t x, fixed_t y)
{
	int	nodenum = numnodes - 1;

	if (!numnodes)	// single subsector is a special case
		return (subsector_t*) subsectors;

	while (!(nodenum & NFx_SUBSECTOR))
	{
		nodenum = nodes[nodenum].children[P_PointOnNodeSide(x, y, nodes+nodenum)];
	}

	return &subsectors[nodenum & ~NFx_SUBSECTOR];
}

//===========================================================================

//
// P_BoxOnLineSide
//
// Considers the line to be infinite
// Returns side 0 or 1, -1 if box crosses the line
//
// killough 5/3/98 (and v220-6): reformatted, cleaned up

int P_BoxOnLineSide (fixed_t *tmbox, line_t *ld)
{
	int p;

	if ((slopetype_e)ld->slopetype == ST_HORIZONTAL)
		return
			(tmbox[BOXBASE] > ld->v1->y) == (p = tmbox[BOXTOP] > ld->v1->y) ?
			p ^ (ld->dx < 0) : -1;

	if ((slopetype_e)ld->slopetype == ST_VERTICAL)
		return
			(tmbox[BOXLEFT] < ld->v1->x) == (p = tmbox[BOXRIGHT] < ld->v1->x) ?
			p ^ (ld->dy < 0) : -1;

	if ((slopetype_e)ld->slopetype == ST_POSITIVE)
		return
			P_PointOnLineSide(tmbox[BOXRIGHT], tmbox[BOXBASE], ld) ==
			(p = P_PointOnLineSide(tmbox[BOXLEFT], tmbox[BOXTOP], ld)) ? p : -1;

	return	// ST_NEGATIVE
		P_PointOnLineSide(tmbox[BOXLEFT], tmbox[BOXBASE], ld) ==
		(p = P_PointOnLineSide(tmbox[BOXRIGHT], tmbox[BOXTOP], ld)) ? p : -1;
}

//
// P_PointOnDivlineSide - Returns 0 or 1
//
static int P_PointOnDivlineSide (fixed_t x, fixed_t y, const divline_t *line)
{
	return asmPointOnSide(y - line->y, line->dx, line->x - x, line->dy) >= 0;
}

//
//P_MakeDivline
//
static void P_MakeDivline (const line_t *li, divline_t *dl)
{
	dl->x = li->v1->x;
	dl->y = li->v1->y;
	dl->dx = li->dx;
	dl->dy = li->dy;
}

//
// P_InterceptVector
//
// Returns the fractional intercept point along the first divline
// This is only called by the addthings and addlines traversers
//
// b211-0 changed to use __int64 to prevent ovf with the now doubled
//  traverse limit.

fixed_t P_InterceptVector (const divline_t *v2, const divline_t *v1)
{
	__int64 num, den = ((__int64)v1->dy * (__int64)v2->dx) - ((__int64)v1->dx * (__int64)v2->dy);

    if ( !(den >>= FRACBITS) ) return 0;

	num = (__int64)(v1->x - v2->x) * (__int64)v1->dy - (__int64)(v1->y - v2->y) * (__int64)v1->dx;

	return (fixed_t)(num / den);
}

//
// P_LineOpening
//
// Sets opentop and openbottom to the window through a two sided line
// OPTIMIZE: keep this precalculated
//
// Risen3D build_62 - 3D line types, sloped flats and blocking
// based on texture height calculated for rendering added.

fixed_t opentop, openbottom, openrange, lowfloor;

extern float vpitch;

// v220-31 slope code revised
fixed_t lineProject(fixed_t fht, sector_t *sec, line_t *ld, fixed_t x, fixed_t y)
{
	float	div, line[2], result[2];

	line[0] = FIX2FLT(ld->dx), line[1] = FIX2FLT(ld->dy);

	if ( (div = DOTPROD2(line, line)) )
	{
		float pointvec[2] = { 
			FIX2FLT(x - ld->v1->x), 
			FIX2FLT(y - ld->v1->y)
		};
		float diff[2], dist;

		div = DOTPROD2(pointvec, line)/div;

		result[VX] = FIX2FLT(ld->v1->x) + line[VX] * div;
		result[VY] = FIX2FLT(ld->v1->y) + line[VY] * div;

		diff[VX] = result[VX] - FIX2FLT(x);
		diff[VY] = result[VY] - FIX2FLT(y);

		if ( (dist = M_ApproxDistancef(diff[VX], diff[VY])) != 0)
		{
			result[0] -= diff[0]/dist * -0.5f;
			result[1] -= diff[1]/dist * -0.5f;
		}

		x = (int)(result[0]*FRACUNIT), y = (int)(result[1]*FRACUNIT);

		if (R_PointInSubsector(x, y)->sector == sec)
			return SL_PointHeight(sec, x, y, LOSLOPE);
	}

	return fht;

}

void P_LineOpening (line_t *ld, int moheight, int x, int y)
{
	if (   ld->backsector == NULL	// single sided line
		|| ld->lnflags2 & LNF2_BLOCKALL)
	{
		openrange = 0;
		return;
	}

	sector_t *frnt = ld->frontsector;
	sector_t *back = ld->backsector;

	if (ld->specType >= R3D_40)
	{
		lowfloor = openbottom = back->floorheight;
		opentop  = back->ceilingheight;

		if (!(ld->spflags & SPF_NOTEX && !is_sliding) && moheight > DDMININT)
		{
			int n = frnt->liftsec3D;
			int bottomStep = ld->d3d->offset3D +
				(n >= 0 ? sectors[n].floorheight : openbottom);

			if (moheight <= bottomStep)
			{
				opentop = bottomStep;
			}
			else
			{
				openbottom = bottomStep + ld->d3d->totalheight;
			}
		}
	}
	else
	{
		fixed_t	fhtf, fhtb;

		// Set the defaults
		fhtf = frnt->floorheight;
		fhtb = back->floorheight;

		if (R3D_maptype)
		{
			int		slp1 = frnt->specFlags2 & SF2_LOSLOPE;
			int		slp2 = back->specFlags2 & SF2_LOSLOPE;
			RDBOOL	isset = false;

			if (R3D_slopetypes)
			{
				int		sd = P_PointOnLineSide(x, y, ld);	// new values being tested

				if (slp1 && slp2 && !(origTEX[ld - lines] & 0x11) )
				{
					// Line goes from slope to slope and no mid or low tex indicates
					// the slopes are contiguous with no dropoff. As such just use
					// the things new (to be tested) x,y pos to determine the floor height.
					fhtf = fhtb = sd ? SL_PointHeight(back, x, y, LOSLOPE) : SL_PointHeight(frnt, x, y, LOSLOPE);
					isset = true;
				}
				else
				if (moheight == DDMININT)	// use the gdt array
				{
					int		sd2 = P_PointOnLineSide(gdt.tmthing->x, gdt.tmthing->y, ld);	// current position

					if (sd != sd2)
					{
						fhtf = SL_PointHeight(sd ? back : frnt, x, y, LOSLOPE);
						fhtb = SL_PointHeight(sd2 ? back : frnt, gdt.tmthing->x, gdt.tmthing->y, LOSLOPE);
						isset = true;
					}
				}

				if (!isset)
				{
					// See if we need to project the things x,y pos to predict the slope height
					if (slp1)
					{
						fhtf = sd ? lineProject(fhtf, frnt, ld, x, y) : SL_PointHeight(frnt, x, y, LOSLOPE);
					}

					if (slp2)
					{
						fhtb = sd ? SL_PointHeight(back, x, y, LOSLOPE) : lineProject(fhtb, back, ld, x, y);
					}
				}
			}

			openbottom = MAX_OF(fhtf, fhtb);
			lowfloor   = MIN_OF(fhtf, fhtb);
			opentop = MIN_OF(SL_PointHeight(frnt, x, y, HISLOPE), SL_PointHeight(back, x, y, HISLOPE) );

			// Find next highest and lowest floor heights for stair calculations
			// Note: for normal doom behaviour these won't be changed from the defaults
			if (!slp1 && !slp2 && abs(fhtf - fhtb) <= MAXSTEP)
			{
				if (lowfloor > gdt.nextlowest && lowfloor < gdt.tmz) gdt.nextlowest = lowfloor;
				if (openbottom < gdt.nexthighest && openbottom > gdt.tmz) gdt.nexthighest = openbottom;
			}
		}
		else
		{
			openbottom = MAX_OF(fhtf, fhtb);
			lowfloor   = MIN_OF(fhtf, fhtb);
			opentop = MIN_OF(frnt->ceilingheight, back->ceilingheight);
			if (abs(fhtf - fhtb) <= MAXSTEP)
			{
				if (lowfloor > gdt.nextlowest && lowfloor < gdt.tmz) gdt.nextlowest = lowfloor;
				if (openbottom < gdt.nexthighest && openbottom > gdt.tmz) gdt.nexthighest = openbottom;
			}
		}
	}
	openrange = opentop - openbottom;
}

// For lines >= R3D_40 with a mid tex
static void SetStepLineOpenings(line_t *ld)
{
	sector_t *front = ld->frontsector,
			 *back  = ld->backsector;

	lowfloor = openbottom = back->floorheight;
	opentop  = back->ceilingheight;

	if ( !(ld->spflags & SPF_NOTEX) )
	{
		int bottomStep = ld->d3d->offset3D +
			(front->liftsec3D >= 0 ?
			 sectors[front->liftsec3D].floorheight : openbottom);

		if (lop.shootz <= bottomStep)
		{
			opentop = bottomStep;
		}
		else
		{
			openbottom = bottomStep + ld->d3d->totalheight;
		}
	}
	else if (ld->xtraflags & XT_MBLINE && !R3D_edit)
	{
		int bottomStep = (ld->frontsector->liftsec3D >= 0 ?
			 sectors[front->liftsec3D].floorheight : openbottom);

		if (lop.shootz <= bottomStep)
		{
			opentop = bottomStep;
		}
		else
		{
			openbottom = bottomStep + ld->d3d->totalheight;
		}
	}
}

void Set2DLineOpenings(line_t *ld)
{
	opentop = MIN_OF(lop.front->ceilingheight, lop.back->ceilingheight);
	openbottom = MAX_OF(lop.front->floorheight, lop.back->floorheight);

	// v220-33 added
	if (ld->lnflags2 & (LNF2_LOWTRIWALL|LNF2_HITRIWALL) && opentop > openbottom)
	{
		fixed_t		x = trace.x + FixedMul (trace.dx, lop.frac);
		fixed_t		y = trace.y + FixedMul (trace.dy, lop.frac);

		if (ld->lnflags2 & LNF2_LOWTRIWALL)
		{
			sector_t	*s = ld->frontsector->stptr >= 0 ? ld->frontsector : ld->backsector;
			openbottom = MAX_OF((SL_PointHeightZ(s, x, y, LOSLOPE)>>FRACBITS)<<FRACBITS, openbottom);
		}
		if (ld->lnflags2 & LNF2_HITRIWALL)
		{
			sector_t	*s = ld->frontsector->stptr >= 0 ? ld->frontsector : ld->backsector;
			opentop = MIN_OF((SL_PointHeightZ(s, x, y, HISLOPE)>>FRACBITS)<<FRACBITS, opentop);
		}
	}

	lowfloor   = MIN_OF(lop.front->floorheight, lop.back->floorheight);
}

static sector_t *checkOffset(line_t *ld)
{
	sector_t *sf, *sb = ld->backsector;

	if (!sb) return NULL;

	sf = ld->frontsector;

	if (sb->ceiloffset < DDMAXINT)
	{
		opentop = MIN_OF(sf->ceilingheight, sb->floorheight + sb->ceiloffset);
		return sb;
	}

	if ( sf->ceiloffset < DDMAXINT)
	{
		opentop = MIN_OF(sb->ceilingheight, sf->floorheight + sf->ceiloffset);
		return sf;
	}
	return NULL;
}

// v220-16 renamed from CheckLineOpenings2D
//         avloslope/avhislope removed as were not being used
RDBOOL CheckLineOpenings(line_t *ld)
{
	// NOTE: returning false means line is to be ignored

	fixed_t slope;

	if (   ld->lnflags2 & LNF2_UNBLOCKED
		&& ld->frontsector->floorheight == ld->backsector->floorheight)
	{
		return false;	// v220-16 emulate doom bug
	}

	if (ld->specType >= R3D_40)
		SetStepLineOpenings(ld);
	else
		Set2DLineOpenings(ld);	

	if (openbottom >= opentop) return true;

	lop.dist = FixedMul (lop.attackrange, lop.frac);

	// Which is correct? FIXME
#if 1
	slope = FixedDiv (openbottom - lop.shootz , lop.dist);

	if (slope > lop.aimslope)
	{
		if (lop.autoaim)
		{
			// Auto adjust upwards - R3D
			lop.aimslope = slope;
			return false;
		}
		return true;
	}
	
	slope = FixedDiv (opentop - lop.shootz , lop.dist);

	if (slope < lop.aimslope)
	{
		sector_t *s;

		if (R3D_ScriptedMap && (s = checkOffset(ld)) )
		{
			slope = FixedDiv (opentop - lop.shootz , lop.dist);

			if (   slope >= lop.aimslope
				&& s->scatter && !((rand()%100) > s->scatter))
			{
				return false;
			}
		}

		if (lop.autoaim)
		{
			// Auto adjust downwards - R3D
			lop.aimslope = slope;
			return false;
		}
		return true;
	}
#else
	if (lop.front->floorheight != lop.back->floorheight)
	{
		slope = FixedDiv (openbottom - lop.shootz , lop.dist);

		if (slope > lop.aimslope)
		{
			if (lop.autoaim)
			{
				// Auto adjust upwards - R3D
				lop.aimslope = slope;
				return false;
			}
			return true;
		}
	}
	
	if (lop.front->ceilingheight != lop.back->ceilingheight)
	{
		slope = FixedDiv (opentop - lop.shootz , lop.dist);

		if (slope < lop.aimslope)
		{
			sector_t *s;

			if (R3D_ScriptedMap && (s = checkOffset(ld)) )
			{
				slope = FixedDiv (opentop - lop.shootz , lop.dist);

				if (   slope >= lop.aimslope
					&& s->scatter && !((rand()%100) > s->scatter))
				{
					return false;
				}
			}

			if (lop.autoaim)
			{
				// Auto adjust downwards - R3D
				lop.aimslope = slope;
				return false;
			}
			return true;
		}
	}
#endif

	return false;

}

// Used with bullets. Returns true if blocked
RDBOOL P_BulletLineOpening (line_t *ld)
{
	// NOTE: returning false means line is to be ignored

	if ( !(ld->ml_flags & ML_TWOSIDED) || ld->lnflags2 & LNF2_OCCLUDE)
	{
		return true;
	}

	if (ld->lnflags2 & LNF2_BLOCKALL)
	{
		// check if bullets are allowed to pass
		if (ld->bv)
		{
			if ( (byte)(rand() % 100) >= ld->bv)
				return true;
			// if false then line openings will be checked
		}
		else
			return true;
	}

	if (ld->specType >= R3D_40)
	{
		if (ld->lnflags2 & LNF2_3DNOBLOCK) return false;

		if (!(!R3D_edit && ld->lnflags2 & LNF2_LOWTEXMASKED))
		{
			int height = ld->frontsector->liftsec3D >= 0 ?
				sectors[ld->frontsector->liftsec3D].floorheight :
				ld->backsector->floorheight;

			int bottomStep = height + ld->d3d->offset3D;

			int offset = (!R3D_edit && ld->specType == R3D_4A) ?
				0 : ld->lnflags2 & LNF2_MIDTEXMASKED ?
					ld->d3d->lineheightlow : ld->d3d->totalheight;

			int topStep = bottomStep + offset;

			return	FixedDiv (topStep - lop.shootz , lop.dist) > lop.aimslope &&
					FixedDiv (bottomStep - lop.shootz , lop.dist) < lop.aimslope;
		}

		return false;
	}

	return CheckLineOpenings(ld);
}

/* =========================================================================

						THING POSITION SETTING

========================================================================= */

//==========================================================================
// P_GetBlockRoot
//	Returns a pointer to the root linkmobj_t of the given mobj. If such
//	a block does not exist, NULL is returned.
//==========================================================================
static mobj_t *P_GetBlockRoot(int blockx, int blocky)
{
	// We must be in the block map range.
	if (   blockx <  0         || blocky < 0 
		|| blockx >= bmapwidth || blocky >= bmapheight) 
	{
		return NULL;
	}
	return (mobj_t*) (blockrings + (blocky*bmapwidth + blockx));
}

//===========================================================================
// P_UnlinkFromSector
//	Only call if it is certain the thing is linked to a sector!
//===========================================================================
static void P_UnlinkFromSector(mobj_t *thing)
{
	// Two links to update: 
	// 1) The link to us from the previous node (sprev, always set) will
	//    be modified to point to the node following us.
	// 2) If there is a node following us, set its sprev pointer to point
	//    to the pointer that points back to it (our sprev, just modified).

	if ( (*thing->sprev = thing->snext) ) 
		thing->snext->sprev = thing->sprev;

	thing->snext = NULL, thing->sprev = NULL;	// Not linked any more.
}

//===========================================================================
// P_UnlinkFromBlock
//	Only call if it is certain that the thing is linked to a block!
//===========================================================================
static void P_UnlinkFromBlock(mobj_t *thing)
{
	(thing->bnext->bprev = thing->bprev)->bnext = thing->bnext;

	thing->bnext = thing->bprev = NULL;	// Not linked any more.
}

//===========================================================================
// P_UnlinkFromLines
//	Unlinks the thing from all the lines it's been linked to. Can be called
//	without checking that the list does indeed contain lines.
//===========================================================================
static void P_UnlinkFromLines(mobj_t *thing)
{
	linknode_t *tn = thingnodes.nodes;
	nodeindex_t nix;

	// Try unlinking from lines.
	if(!thing->lineroot) return; // A zero index means it's not linked.
	
	// Unlink from each line.
	for(nix = tn[thing->lineroot].next; nix != thing->lineroot; 
		nix = tn[nix].next)
	{
		// Data is the linenode index that corresponds to this thing.
		NP_Unlink(&linenodes, tn[nix].data);
		// We don't need these nodes any more, mark them as unused.
		// Dismissing is a macro.
		NP_DISMISS(linenodes, tn[nix].data);
		NP_DISMISS(thingnodes, nix);
	}

	// The thing no longer has a line ring.
	NP_DISMISS(thingnodes, thing->lineroot);
	thing->lineroot = 0;
}

//==========================================================================
// P_UnlinkThing
//	Unlinks a thing from block map and sectors where appropriate.
//==========================================================================
void P_UnlinkThing(mobj_t *thing)
{
	if (thing->sprev) P_UnlinkFromSector(thing);
	if (thing->bnext) P_UnlinkFromBlock(thing);
	P_UnlinkFromLines(thing);
}

static int	thbox[4];

//===========================================================================
// PIT_LinkToLines
//	The given line might cross the thing. If necessary, link the mobj 
//	into the line's ring.  To prevent duplicate nodes being added ensure
//  validcount's been set by the caller.
//===========================================================================
static RDBOOL PIT_LinkToLines(line_t *ld, mobj_t *thing)
{
	nodeindex_t nix;

	// One sided lines not linked as a mobj can't legally cross one.
	// Lines marked bad are also ignored - R3D
	if ( !(ld->ml_flags & ML_TWOSIDED)
		|| ld->ml_flags & ML_BADLINE )
	{
		return true;
	}

	if (   thbox[BOXRIGHT] <= ld->bbox[BOXLEFT ]
		|| thbox[BOXLEFT ] >= ld->bbox[BOXRIGHT]
		|| thbox[BOXTOP  ] <= ld->bbox[BOXBASE ]
		|| thbox[BOXBASE ] >= ld->bbox[BOXTOP  ]
		|| P_BoxOnLineSide(thbox, ld) != -1)
	{
		// Line does not cross the thing's bounding box.
		return true;
	}

	// Add a node to the thing's ring.
	
	nix = NP_New(&thingnodes, ld);
	
	NP_Link(&thingnodes, nix, thing->lineroot);

	// Add a node to the line's ring. Also store the linenode's index
	// into the thingring's node, so unlinking is easy.
	
	thingnodes.nodes[nix].data = NP_New(&linenodes, thing);
	
	NP_Link(&linenodes, thingnodes.nodes[nix].data, linelinks[ld - lines]);
	
	return true;
}

//===========================================================================
// P_LinkToLines
//	The thing must be currently unlinked.
//===========================================================================
static void P_LinkToLines(mobj_t *thing, fixed_t radius)
{
	int		bx, by, xl, yl, xh, yh;

	// Get a new root node.
	thing->lineroot = NP_New(&thingnodes, NP_ROOT_NODE);

	// Set up a line iterator for doing the linking.
	M_setBoxOffsets  (thbox, thing->x, thing->y, radius);
	M_getBlockOffsets(thbox, 0, &xl, &xh, &yl, &yh);

	validcount++;

	for(bx = xl; bx <= xh; bx++)
	{
		for(by = yl; by <= yh; by++)
		{
			P_BlockLinesIterator(bx, by, (int (__cdecl *)(struct line_s *,void *))PIT_LinkToLines, thing);
		}
	}
}

//==========================================================================
// P_LinkThing
//	Links a thing into both a block and a subsector based on it's (x,y).
//	Sets thing->subsector properly. Calling with flags==0 only updates
//	the subsector pointer. Can be called without unlinking first.
//==========================================================================
void P_LinkThing(mobj_t *thing, int flags)
{
	// Link into the subsector.
	thing->subsector = R_PointInSubsector(thing->x, thing->y);

	// v220-5 check if this is the player and, if so,
	//        change the plr_viewgroup etc.
	if (   thing->dplayer
		&& thing->dplayer->mo == thing)
	{
		int os;

		thing->old_z = thing->z;	// used for jumpkill
		ppt.view_subsector = thing->subsector;
		plr_viewgroup = thing->subsector->viewgroup;

		ppt.viewer_fake_sector = (ppt.view_subsector->ssflags & SSF_IGNOREFAKE) > 0 ?
			-1 : ppt.view_subsector->sector->heightsec;

		if (   ppt.viewer_fake_sector < 0
			&& ppt.view_subsector->floorsec >= 0
			&& (os = sectors[ppt.view_subsector->floorsec].heightsec) >= 0)
		{
			ppt.viewer_fake_sector = os;
		}

		if (SHarray && !(thing->player->cheats & CF_NOCLIP))
		{
			int n;

			os = thing->subsector->sector - sectors;

			if (last_player_sector != os && (n = (byte)SHarray[os].cp))
			{
				if (SHcontainer[n] && !S_IsPlaying(SHcontainer[n], thing))
					S_StartSound(SHcontainer[n], thing);
			}

			if (SHarray[os].fp & SH_STOPMUSIC) Music_Stop();

			last_player_sector = os;
		}
	}

	if(flags & DDLINK_SECTOR)
	{
		sector_t *sec = thing->subsector->sector;

		// Unlink from the current sector, if any.
		if (thing->sprev) P_UnlinkFromSector(thing);

		// Link the new thing to the head of the list.
		// Prev pointers point to the pointer that points back to us.
		// (Which practically disallows traversing the list backwards.)

		if ( (thing->snext = sec->thinglist) )
			thing->snext->sprev = &thing->snext;

		*(thing->sprev = &sec->thinglist) = thing;
	}
	
	// Link into blockmap.
	if(flags & DDLINK_BLOCKMAP)
	{
		mobj_t *root;

		// Unlink from the old block, if any.
		if(thing->bnext) P_UnlinkFromBlock(thing);

		// Link into the block we're currently in.
		root = P_GetBlockRoot(MAPBS(thing->x,bmaporgx), MAPBS(thing->y,bmaporgy));

		if (root)	// Only link if we're inside the blockmap.		
		{
			thing->bprev = root;
			thing->bnext = root->bnext;
			thing->bnext->bprev = root->bnext = thing;
		}
	}

	// Link into lines. Radius checks first added v220-8
	if ( !(flags & DDLINK_NOLINE) )
	{
		fixed_t	radius;

		if (thing->flags3 & MF3_USETRUERADIUS)
		{
			// v220-4 get the things true (visible) radius
			radius = R_MaxRadius(thing, false, 0);
			if (thing->radius > radius) radius = thing->radius;
		}
		else
			radius = thing->radius;

		// Unlink from any existing lines.
		P_UnlinkFromLines(thing);

		// Link to all contacted lines.
		P_LinkToLines(thing, radius);

		// v220-8 check if fits in subsec and if not set/reset flag
		//        amodels are always unclipped as cannot always
		//        use the true radius as a large asymetric model
		//        may end up being rendered when it should be clipped.
		if (   thing->flags3 & MF3_AMODEL
			|| !M_FitSSECBox(	thing->subsector->ssbox,
								thing->x,
								thing->y,
								radius))
		{
			thing->flags2 |= MF2_UNCLIPPED;
		}
		else
		{
			thing->flags2 &= ~MF2_UNCLIPPED;
		}
	}
}


/*
===============================================================================

						BLOCK MAP ITERATORS

 For each line/thing in the given mapblock, call the passed function.
 If the function returns false, exit with false without checking anything
 else. Linkstore is collected before any callbacks are called (not done
 for line iterators since those can't change).

 Some defensive coding added - R3D

===============================================================================
*/

//===========================================================================
// P_BlockLinesIterator
//	The validcount flags are used to avoid checking lines that are marked 
//	in multiple mapblocks, so increment validcount before the first call 
//	to P_BlockLinesIterator, then make one or more calls to it.
//===========================================================================
RDBOOL P_BlockLinesIterator(int x, int y, RDBOOL(*func)(line_t*,void*),
																void *data)
{
	if (x >= 0 && y >= 0 && x < bmapwidth && y < bmapheight)
	{
		int			offset = *(blockmap + y*bmapwidth + x);
		const long	*list = blockmaplump + offset;
		line_t		*ld;

		// skip starting delimiter
		while( *(++list) >= 0)
		{
			ld = &lines[*list];

			if (ld->validcount != validcount)
			{
				ld->validcount = validcount;
				
				if (!func(ld, data)) return false;
			}
		}
	}
	
	return true;
}

//===========================================================================
// P_BlockThingsIterator
//	'func' can do whatever it pleases to the mobjs.
//===========================================================================
RDBOOL P_BlockThingsIterator(int x, int y, RDBOOL(*func)(mobj_t*,void*),
																void *data)
{
	mobj_t	*mobj, *root = P_GetBlockRoot(x, y);
	
	if (root)
	{
		void	*linkstore[MAXLINKED];
		int		i, c = 0;

		// Gather all the things in the block into the list.
		for (mobj = root->bnext; mobj != root; mobj = mobj->bnext)
		{
			if (mobj->validcount != validcount)
			{
				mobj->validcount = validcount;
				linkstore[c] = mobj;
				if (c + 1 >= MAXLINKED) break;
				c++;
			}
		}

		for (i = 0; i < c; i++)
		{
			if (!func((mobj_t*)linkstore[i], data)) return false;
		}
	}
	return true;
}

//===========================================================================
// P_BlockStandardThingsIterator
//	'func' can do whatever it pleases to the mobjs.
//===========================================================================
RDBOOL P_BlockStandardThingsIterator(int x, int y, RDBOOL(*func)(mobj_t*,void*),
																void *data)
{
	mobj_t	*mobj, *root = P_GetBlockRoot(x, y);
	
	if (root)
	{
		void	*linkstore[MAXLINKED];
		int		i, c = 0;

		// Gather all required things in the block into the list.
		for (mobj = root->bnext; mobj != root; mobj = mobj->bnext)
		{
			if (mobj->validcount != validcount)
			{
				mobj->validcount = validcount;

				if (mobj->flags & MF_SHOOTABLE)	// v220-14 these are the only ones needed
				{
					linkstore[c] = mobj;
					if (c + 1 >= MAXLINKED) break;
					c++;
				}
			}
		}

		for (i = 0; i < c; i++)
		{
			if (!func((mobj_t*)linkstore[i], data)) return false;
		}
	}
	return true;
}

//===========================================================================
// P_BlockSpecialIterator - R3D
//===========================================================================
RDBOOL P_BlockSpecialIterator(int x, int y, RDBOOL(*func)(mobj_t*,void*),
																void *data)
{
extern int		modelboxcnt;

	for (int i = 0; i < modelboxcnt; i++)
	{
		mobj_t *mo = MB_GetMBmo(i);

		for (int k = 0; k < 4; k++)
		{
			if (!func(mo, data)) return false;
		}
	}
	return true;
}

//===========================================================================
// P_ThingLinesIterator
//	The callback function will be called once for each line that crosses
//	through the object. This means all the lines will be two-sided.
//===========================================================================
RDBOOL P_ThingLinesIterator
	(mobj_t *thing, RDBOOL (*func)(line_t*,void*), void *data)
{
	if (thing->lineroot) // Check there are lines to process.
	{
		nodeindex_t	nix;
		linknode_t	*tn = thingnodes.nodes;
		void		*linkstore[MAXLINKED];
		int			i, c = 0;

		for (nix = tn[thing->lineroot].next;
			 nix != thing->lineroot;
			 nix = tn[nix].next)
		{
			linkstore[c] = tn[nix].object_ptr;
			if (c + 1 >= MAXLINKED) break;
			c++;
		}

		for (i = 0; i < c; i++)
		{
			if (!func((line_t*)linkstore[i], data)) return false;
		}
	}
	return true;
}

//===========================================================================
// P_ThingSectorsIterator
//	The callback function will be called once for each sector the thing
//  is touching (totally or partly inside). This is not a 3D check;
//  the thing may actually reside above or below the sector.
//===========================================================================
RDBOOL P_ThingSectorsIterator
	(mobj_t *thing, RDBOOL (*func)(sector_t*,void*), void *data)
{
	static int	scvalidcount = 0;
	void		*linkstore[MAXLINKED];
	sector_t	*sec;
	int			i, c = 1;

	// Always process the thing's own sector first.
	*linkstore = sec = thing->subsector->sector;
	sec->scvalidcount = ++scvalidcount;

	if (thing->lineroot) // Check there are lines to process.
	{
		nodeindex_t	nix;
		linknode_t	*tn = thingnodes.nodes;
		line_t		*ld;

		for (nix = tn[thing->lineroot].next;
			 nix != thing->lineroot; 
			 nix = tn[nix].next)
		{
			// All these lines are two-sided.
			ld = (line_t*) tn[nix].object_ptr;

			if ( (sec = ld->frontsector)->scvalidcount != scvalidcount)
			{
				linkstore[c] = sec;
				sec->scvalidcount = scvalidcount;
				if (c + 1 >= MAXLINKED) break;
				c++;
			}

			if ( (sec = ld->backsector)->scvalidcount != scvalidcount)
			{
				linkstore[c] = sec;
				sec->scvalidcount = scvalidcount;
				if (c + 1 >= MAXLINKED) break;
				c++;
			}
		}
	}

	for (i  = 0; i < c; i++)
	{
		if (!func((sector_t*)linkstore[i], data)) return false;
	}

	return true;
}

//===========================================================================
// b211-0 P_SectorTouchingThingsList
//	Populates list sent with mobj's touching the sector
//  If add_amodel is false then no amodels will be included
//  if true then only amodels with their centre x,y point in the sector
//  are included.
//===========================================================================
static void addLinkList(mobj_t *th)
{
	if (link_list.listcnt >= link_list.maxlistcnt)
	{
		link_list.maxlistcnt += 64;

		link_list.mobj_list = (mobj_t**)W_Realloc(link_list.mobj_list,
			link_list.maxlistcnt*sizeof(mobj_t*), STATICID, 0);
	}

	link_list.mobj_list[link_list.listcnt++] = th;
}

void P_SectorTouchingThingsList(sector_t *sector, RDBOOL add_amodel)
{
	nodeindex_t	root, nix;
	linknode_t	*ln = linenodes.nodes;
	mobj_t		*mo;
	line_t		*li;
	int			i, c = 0;

	link_list.listcnt = 0;

	validcount++;

	// First process the things that obviously are in the sector.
	for (mo = sector->thinglist; mo; mo = mo->snext)
	{
		if (mo->validcount != validcount)
		{
			mo->validcount = validcount;
			
			if (mo->thinker.function == (void (__cdecl *)(void))P_MobjThinker)
			{
				// Don't include amodels unless told
				// NOTE: amodels are only included if inside the sector
				if (!add_amodel && mo->flags3 & MF3_AMODEL) continue;

				// These are always skipped
				if (mo->flags2 & MF2_VDOOR) continue;

				addLinkList(mo);
			}
		}
	}

	// Then check the sector's lines.
	for (i = 0; i < sector->linecount; i++)
	{
		li = sector->lines[i];

		if ( (root = linelinks[li - lines]) )
		{
			for (nix = ln[root].next; nix != root; nix = ln[nix].next)
			{
				mo = (mobj_t*) ln[nix].object_ptr;

				if (mo->validcount != validcount)
				{
					// Removed ones can still have a line link
					if (mo->thinker.function != (void (__cdecl *)(void))P_MobjThinker) continue;

					mo->validcount = validcount;

					// Only add sector linked things
					if (mo->flags & MF_NOSECTOR) continue;
					// These are adjusted independently of the move code
					if (mo->flags2 & MF2_VDOOR) continue;
					// Don't include amodels
					if (mo->flags3 & MF3_AMODEL) continue;

					addLinkList(mo);
				}
			}
		}
	}
}

//===========================================================================
// b211-0 P_MoveSectorThingsIterator
//  This is intended for use with height changing sectors only.
//	'func' is called for each mobj that is (even partly) inside the sector.
//  This is not a 3D test, the mobjs may be above or under the sector.
//  Amodels are only processed if they are in the sector sent.
//  Any found are all are tied to the sector with func being ignored.
//===========================================================================
RDBOOL P_MoveSectorThingsIterator
	(sector_t *sector, RDBOOL (*func)(mobj_t*,void*), void *data)
{
	mobj_t		*mo;
	int			i, flags;

	// Must be called first (also initialises link_list)
	P_SectorTouchingThingsList(sector, true);

	// Link amodels to a plat's floor/ceiling
	// These must all be processed hence dealing with amodels first
	if ( (flags = sector->specFlags & SF_MOVSECTOR) )
	{
		// Link amodels to a plat's floor/ceiling
		for (i = 0; i < link_list.listcnt; i++)
		{
			mo = link_list.mobj_list[i];

			if (mo->flags3 & MF3_AMODEL)
			{
				if (!(mo->flags2 & MF2_VDOOR))
				{
					if (mo->flags & MF_SPAWNCEILING)	// These never have gravity
					{
						if (flags & SF_MOVCEIL)
						{
							mo->ceilingz = sector->ceilingheight;
 							mo->z = mo->ceilingz - mo->height;
						}
					}
					else if (flags & SF_MOVFLOOR)
					{
 						mo->z = sector->floorheight;

						if (mo->flags3 & MF3_CHECKPOS)
							func(mo, data);
					}
				}
			}
			else
			{
				if (   flags & SF_MOVCEIL
					&& mo->flags & MF_SPAWNCEILING
					&& mo->flags & MF_NOGRAVITY
					&& mo->subsector->sector == sector)
				{
					mo->ceilingz = sector->ceilingheight;
 					mo->z = mo->ceilingz - mo->height;
				}
			}
		}
	}

	// Link all things partially in the sector except for amodels
	for (i = 0; i < link_list.listcnt; i++)
	{
		if ( !((mo = link_list.mobj_list[i])->flags3 & MF3_AMODEL) )
		{
			func(mo, data);
		}
	}

	link_list.listcnt = 0;
	return true;
}

/*
===============================================================================

					         INTERCEPT ROUTINES

===============================================================================
*/

// 1/11/98 killough: Intercept limit removed
static intercept_t *intercepts = NULL, *intercept_p = NULL;

// Check for limit and increment if necessary
static void check_intercept(void)
{
	static size_t num_intercepts = 0;
	size_t offset = intercept_p - intercepts;

	if (offset >= num_intercepts)
	{
		num_intercepts = num_intercepts ? num_intercepts*2 : 128;
		intercepts = (intercept_t*)W_Realloc(intercepts, sizeof(intercept_t) * num_intercepts, STATICID, 208);
		intercept_p = intercepts + offset;
	}
}

//===========================================================================
//  PIT_AddLineIntercepts
//
// Looks for lines in the given block that intercept the given trace
// to add to the intercepts list
// A line is crossed if its endpoints are on opposite sides of the trace
//===========================================================================
RDBOOL PIT_AddLineIntercepts (line_t *ld, mobj_t *mo)
{
	int			s1, s2;
	fixed_t		frac;
	divline_t	dl;

	// R3D - If mo is set then it's an Amodel Door line
	//       Check that mo is facing the front side
	if (mo)
	{
		if (   mo->x + mo->radius <= ld->bbox[BOXLEFT ]
			|| mo->x - mo->radius >= ld->bbox[BOXRIGHT]
			|| mo->y + mo->radius <= ld->bbox[BOXBASE ]
			|| mo->y - mo->radius >= ld->bbox[BOXTOP  ])
		{
			return true;
		}
	}

	if (   !is_sliding
		&& ld->specType == R3D_40
		&& !(ld->xtraflags & XT_MBLINE && !R3D_edit)
		&& ld->spflags & SPF_NOTEX
		&& ld->ml_flags & ML_SECRET)
	{
		return true;	// ignore this line type
	}

	// avoid precision problems with two routines
	if (abs(trace.dx) > FRACUNIT*16 || abs(trace.dy) > FRACUNIT*16)
	{
		s1 = P_PointOnDivlineSide (ld->v1->x, ld->v1->y, &trace);
		s2 = P_PointOnDivlineSide (ld->v2->x, ld->v2->y, &trace);
	}
	else
	{
		s1 = P_PointOnLineSide (trace.x, trace.y, ld);
		s2 = P_PointOnLineSide (trace.x+trace.dx, trace.y+trace.dy, ld);
	}

	if (s1 == s2) return true;	// line isn't crossed

	P_MakeDivline(ld, &dl);

	frac = P_InterceptVector (&trace, &dl);

	if (frac < 0) return true;	// behind source

	if (ld->backsector && ld->backsector->sector3D >= 0) lop.has3Dflats = true;
	
	check_intercept();			// killough
	
	intercept_p->frac = frac;
	intercept_p->isaline = true;
	intercept_p->d.line = ld;
	intercept_p++;

	return true;				// continue
}

//===========================================================================
//  PIT_AddSpecialIntercepts
//   b220 R3D proprietory - interactive amodel / model door support 
//===========================================================================
RDBOOL PIT_AddSpecialIntercepts (mobj_t *thing, void *data)
{
	int i, num;

	if ( (num = thing->modelboxnum - 1) >= 0)
	{
		for (i = 0; i < 4; i++)
		{
			PIT_AddLineIntercepts(MB_GetLine(num, i, NULL), NULL);
		}
	}

	return true;
}

//===========================================================================
//  PIT_AddThingIntercepts
//===========================================================================
RDBOOL PIT_AddThingIntercepts(mobj_t *thing, void *data)
{
	fixed_t		x1,y1, x2,y2, t;
	int			s1, s2;
	divline_t	dl;
	fixed_t		frac;
	
	if (thing == shootthing || !(thing->flags & MF_SHOOTABLE) )
	{
		return true;
	}

	// check a corner to corner crossection for hit
	x1 = thing->x - thing->radius;
	x2 = thing->x + thing->radius;
	y1 = thing->y - thing->radius;
	y2 = thing->y + thing->radius;			

	if ((trace.dx ^ trace.dy) > 0)
	{
		t = y1; y1 = y2; y2 = t;	// swap y1 y2
	}

	s1 = P_PointOnDivlineSide (x1, y1, &trace);
	s2 = P_PointOnDivlineSide (x2, y2, &trace);

	if (s1 == s2) return true;		// line isn't crossed
	
	dl.x = x1;
	dl.y = y1;
	dl.dx = x2 - x1;
	dl.dy = y2 - y1;

	frac = P_InterceptVector (&trace, &dl);

	if (frac < 0) return true;	// behind source

	check_intercept();			// killough

	intercept_p->frac = frac;
	intercept_p->isaline = false;
	intercept_p->d.thing = thing;
	intercept_p++;

	return true;				// keep going
}

static int __cdecl fracSORT(const void *e1, const void *e2)
{
	intercept_t *m1 = (intercept_t*) e1;
	intercept_t *m2 = (intercept_t*) e2;

	return m1->frac < m2->frac ? -1 :m1->frac > m2->frac ? 1 : 0;
}

//===========================================================================
//  P_TraverseIntercepts
//    Returns true if the traverser function returns true for all lines
//    R3D: reformatted
//===========================================================================
static RDBOOL P_TraverseIntercepts(traverser_t func, int flags)
{
	intercept_t	*in;
	int			count = intercept_p - intercepts;

	if (count > 1)
		qsort((void*)intercepts, count, sizeof(intercept_t), fracSORT);

	for (int i = 0; i < count; i++)
    {
		RDBOOL	sw3D = false;
		
		in = &intercepts[i];

		if (in->frac > FRACUNIT) return true;

		if (is_sliding)
		{
			func(in);
		}
		else
		{
			if (in->isaline)
			{
				sw3D = flags & PT_3DSWITCH && in->d.line->ovflags & OVF_3DSWITCH;
			}

			if (!func(in))
			{
				// Don't bail out if a 3D switch is not activated because
				// the activator's height doesn't match. There may be one
				// stacked higher or lower that does - b220-0.
				if (!sw3D || R3D_edit) return false;	// ignore: func deals with this
			}
			else if (sw3D)
			{
				make_noway_sound = false;
			}
		}
    }

	return true;	// everything was traversed
}

//===========================================================================
//  P_PathTraverse
//
// Traces a line from x1,y1 to x2,y2, calling the traverser function for each
// Returns true if the traverser function returns true for all lines
//
// b220-0 reworked
// v220-16 revised to work properly with greater bullet distance
//===========================================================================
RDBOOL P_PathTraverse (fixed_t x1, fixed_t y1, fixed_t x2, fixed_t y2, RDBOOL usedist, int flags, RDBOOL (*trav) (intercept_t *))
{
	fixed_t		xintercept, yintercept;
	fixed_t		xstep, ystep;
	fixed_t		xb1, yb1;
	fixed_t		partial;
	int			xt1, yt1, xt2, yt2;
	int			mapxstep, mapystep;
	int			count;
	RDBOOL		add_specials = R3D_amodeldoors && !R3D_edit && flags & (PT_ADDLINES|PT_ADDSPEC);

	intercept_p = intercepts;
	
	// don't side exactly on a line
	if ( ((x1 - bmaporgx) & (MAPBLOCKSIZE - 1)) == 0)
		x1 += FRACUNIT;

	if ( ((y1 - bmaporgy) & (MAPBLOCKSIZE - 1)) == 0)
		y1 += FRACUNIT;

	trace.x = x1;
	trace.y = y1;

	xt1 = MAPBS(x1, bmaporgx);
	yt1 = MAPBS(y1, bmaporgy);

	if (usedist)
	{
		trace.dx = x2;
		trace.dy = y2;
		xt2 = MAPBS3(x1, bmaporgx, x2);
		yt2 = MAPBS3(y1, bmaporgy, y2);
	}
	else
	{
		trace.dx = x2 - x1;
		trace.dy = y2 - y1;
		xt2 = MAPBS(x2, bmaporgx);
		yt2 = MAPBS(y2, bmaporgy);
	}

	xb1 = MAPBSF(x1, bmaporgx);
	yb1 = MAPBSF(y1, bmaporgy);

	if (xt2 > xt1)
	{
		mapxstep = 1;
		partial = FRACUNIT - (xb1 & (FRACUNIT - 1));
		ystep = FixedDiv(trace.dy, abs(trace.dx));
	}
	else if (xt2 < xt1)
	{
		mapxstep = -1;
		partial = xb1 & (FRACUNIT - 1);
		ystep = FixedDiv(trace.dy, abs(trace.dx));
	}
	else
	{
		mapxstep = 0;
		partial = FRACUNIT;
		ystep = 256*FRACUNIT;
	}	

	yintercept = yb1 + FixedMul(partial, ystep);

	if (yt2 > yt1)
	{
		mapystep = 1;
		partial = FRACUNIT - (yb1 & (FRACUNIT - 1));
		xstep = FixedDiv(trace.dx, abs(trace.dy));
	}
	else if (yt2 < yt1)
	{
		mapystep = -1;
		partial = yb1 & (FRACUNIT - 1);
		xstep = FixedDiv(trace.dx, abs(trace.dy));
	}
	else
	{
		mapystep = 0;
		partial = FRACUNIT;
		xstep = 256*FRACUNIT;
	}

	xintercept = xb1 + FixedMul(partial, xstep);

	// Cover starting position in case no lines are contacted
	lop.has3Dflats = R3D_linemap && R_PointInSubsector(x1, y1)->flats3DArray;

	validcount++;

	// add_specials: done first as is not map block dependent
	if (add_specials)
		P_BlockSpecialIterator(xt1, yt1, (int (__cdecl *)(struct mobj_s *,void *))PIT_AddSpecialIntercepts, &flags);

	// step through map blocks
	// Count is present to prevent a round off error from skipping the break
	for (count = 0 ; count < 64 ; count++)
	{
		// in range ?
		if (xt1 < 0 || yt1 < 0 || xt1 > bmapwidth || yt1 > bmapheight) break;

		if (flags & (PT_ADDLINES|PT_ADDLINESPTC))
			P_BlockLinesIterator (xt1, yt1, (int (__cdecl *)(struct line_s *,void *))PIT_AddLineIntercepts, NULL);

		if (flags & PT_ADDTHINGS)	// This is only called when shooting at a thing
			P_BlockStandardThingsIterator(xt1, yt1, (int (__cdecl *)(struct mobj_s *,void *))PIT_AddThingIntercepts, NULL);
		
		if (xt1 == xt2 && yt1 == yt2) break;
		
		if ( (yintercept >> FRACBITS) == yt1)
		{
			yintercept += ystep;
			xt1 += mapxstep;
		}
		else if ( (xintercept >> FRACBITS) == xt1)
		{
			xintercept += xstep;
			yt1 += mapystep;
		}
	}

	// go through the sorted list
	return P_TraverseIntercepts (trav, flags);
}
