//-----------------------------------------------------------------------------
//
// Copyright (C) 2005 Jaakko Kernen
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_particle.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Replaced zone memory with win mem [PARTCLID]
//  Added [Risen3D] 3D line type collision
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
//  Particle rendering on 3D flats
//  Particle rendering lists generated to allow correct
//  rendering order with translucent flats
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:   Particle generation.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include "de_base.h"
#include "de_console.h"
#include "de_network.h"
#include "de_play.h"
#include "de_refresh.h"
#include "de_misc.h"

#include "r3d_MAINspec.h"

#include <math.h>
#include <string.h>

// MACROS ------------------------------------------------------------------

#define DOT2F(a,b)			( FIX2FLT(a[VX])*FIX2FLT(b[VX]) + FIX2FLT(a[VY])*FIX2FLT(b[VY]) )
#define VECMUL(a,scalar)	( a[VX]=FixedMul(a[VX],scalar), a[VY]=FixedMul(a[VY],scalar) )
#define VECADD(a,b)			( a[VX]+=b[VX], a[VY]+=b[VY] )
#define VECMULADD(a,scal,b)	( a[VX]+=FixedMul(scal, b[VX]), a[VY]+=FixedMul(scal,b[VY]) )
#define VECSUB(a,b)			( a[VX]-=b[VX], a[VY]-=b[VY] )
#define VECCPY(a,b)			( a[VX]=b[VX], a[VY]=b[VY] )
#define PRAND				( rand() % 256 )

// TYPES -------------------------------------------------------------------

typedef struct {
	ded_ptcgen_t *def[2];
	sector_t	 *sec;
} defsector_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

void PG_InitForLevel(void);

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int		r_use_particles;
float	r_particle_spawn_rate = 1;	// Unmodified.
ptcgen_t *active_ptcgens[MAX_ACTIVE_PTCGENS];


// PRIVATE DATA DEFINITIONS ------------------------------------------------

static defsector_t *defsectors;
static int defseccount;

static struct pvar_t {
	line_t	*ptc_hitline;
	fixed_t	mbox[4];
	fixed_t	tmpz;
	fixed_t	tmprad;
	fixed_t	tmcross;
	fixed_t	tmpx1;
	fixed_t	tmpx2;
	fixed_t	tmpy1;
	fixed_t	tmpy2;
	int		side;

} pvar;


// CODE --------------------------------------------------------------------

//===========================================================================
// P_FreePtcGen
//===========================================================================
void P_FreePtcGen(ptcgen_t *gen)
{
	int i;

	for (i = 0; i < MAX_ACTIVE_PTCGENS; i++)
	{
		if (active_ptcgens[i] == gen)
		{
			active_ptcgens[i] = NULL;

			// Destroy it.
			if (gen->thinker.function == (void (__cdecl *)(void))P_PtcGenThinker)
			{
				W_Free(gen->ptcs, PARTCLID, 508);
				gen->ptcs = NULL;
				gen->count = 0;
				gen->thinker.function = (think_t) -1;
			}
			break;
		}
	}
}

//===========================================================================
// P_NewPtcGen
//	Allocates a new active ptcgen and adds it to the list of active ptcgens.
//===========================================================================
ptcgen_t *P_NewPtcGen(void)
{
	ptcgen_t	*gen;
	int			i, oldest = -1, index = -1, maxage;

	// Find a suitable spot in the active ptcgens list.
	for (i = 0; i < MAX_ACTIVE_PTCGENS; i++)
	{
		if (!active_ptcgens[i])
		{
			index = i; break;
		}
		else if ( !(active_ptcgens[i]->flags & PGF_STATIC)
			&& (oldest < 0 || active_ptcgens[i]->age > maxage))
		{
			oldest = i;
			maxage = active_ptcgens[i]->age;
		}
	}

	if (index < 0) 
	{
		if (oldest < 0) return NULL; // No free slots

		// Destroy the oldest generator.
		P_FreePtcGen(active_ptcgens[oldest]);
		index = oldest;
	}

	// R3D - We generate a new thinker rather than using the inactive one.
	//       The inactive one, which is now only known to the thinker list,
	//       will be entirely removed when the thinker list is run as its
	//		 function should have been set to (think_t) -1;
	gen = (ptcgen_t*)W_Calloc(sizeof(ptcgen_t), PARTCLID, 0);
	active_ptcgens[index] = gen;

	// Link the thinker into the list
	gen->thinker.function = (void (__cdecl *)(void))P_PtcGenThinker;
	P_AddParticleThinker(&gen->thinker);
	return gen;
}

//===========================================================================
// P_InitParticleGen
//	Set gen->count prior to calling this function.
//===========================================================================
static void P_InitParticleGen(ptcgen_t *gen, ded_ptcgen_t *def)
{
	int i;
	
	if(gen->count <= 0) gen->count = 1;

	// Make sure no generator is type-triggered by default.
	gen->type = gen->type2 = -1; 

	gen->def   = def;
	gen->flags = def->flags; 
	gen->ptcs  = (particle_t*)W_Malloc(sizeof(particle_t) * gen->count, PARTCLID, 2);

	for(i = 0; i < MAX_PTC_STAGES; i++)
	{
		gen->stages[i].bounce     = (int)(FRACUNIT * def->stages[i].bounce);
		gen->stages[i].resistance = (int)(FRACUNIT * (1 - def->stages[i].resistance));
		gen->stages[i].radius     = (int)(FRACUNIT * def->stages[i].radius);
		gen->stages[i].gravity    = (int)(FRACUNIT * def->stages[i].gravity);

		gen->stages[i].type  = Def_EvalFlags(def->stages[i].type);
		gen->stages[i].flags = Def_EvalFlags(def->stages[i].flags); 
	}

	// Init some data.
	for(i = 0; i < 3; i++)
	{
		gen->center[i] = (int)(FRACUNIT * def->center[i]);
		gen->vector[i] = (int)(FRACUNIT * def->vector[i]);
	}

	// Mark everything unused.
	memset(gen->ptcs, -1, sizeof(particle_t) * gen->count);

	// Clear the contact pointers.
	for(i = 0; i < gen->count; i++) gen->ptcs[i].contact = NULL;
}

//===========================================================================
// P_PresimParticleGen
//===========================================================================
static void P_PresimParticleGen(ptcgen_t *gen, int tics)
{
	for(; tics > 0; tics--) P_PtcGenThinker(gen);

	// Reset age so presim doesn't affect it.
	gen->age = 0;
}

//===========================================================================
// P_SpawnParticleGen
//	Creates a new mobj-triggered particle generator based on the given 
//	definition. The generator is added to the list of active ptcgens.
//===========================================================================
void P_SpawnParticleGen(ded_ptcgen_t *def, mobj_t *source)
{
	ptcgen_t *gen = P_NewPtcGen();

	if (!gen) return;

	// Initialize the particle generator.
	gen->count = def->particles;
	P_InitParticleGen(gen, def);
	gen->source = source;
	gen->srcid = source->thinker.id;

	// Is there a need to pre-simulate?
	P_PresimParticleGen(gen, def->presim);
}

//===========================================================================
// P_SpawnPlaneParticleGen
//	Creates a new flat-triggered particle generator based on the given 
//	definition. The generator is added to the list of active ptcgens.
//===========================================================================
void P_SpawnPlaneParticleGen(ded_ptcgen_t *def, sector_t *sec,
								   int is_ceiling)
{
	ptcgen_t *gen;
	float *box, width, height;

	if (!r_use_particles || !(gen = P_NewPtcGen())) return;

	// Size of source sector might determine count.
	if(def->flags & PGF_PARTS_PER_128)
	{
		// This is rather a rough estimate of sector area.
		box = sec->sibounds;
		width  = (box[BOXRIGHT] - box[BOXLEFT]) / 128.0f;
		height = (box[BOXBASE ] - box[BOXTOP ]) / 128.0f;
		gen->area = width * height;
		gen->count = (int)((float)def->particles * gen->area);
	}
	else
		gen->count = def->particles;

	// Initialize the particle generator.
	P_InitParticleGen(gen, def);
	gen->sector = sec;
	gen->plane = is_ceiling;

	// Is there a need to pre-simulate?
	P_PresimParticleGen(gen, def->presim);
}

//===========================================================================
// P_Uncertain
//	The offset is spherical and random.
//	Low and High should be positive.
//===========================================================================
void P_Uncertain(fixed_t *pos, fixed_t low, fixed_t high)
{
	int i;
	
	if(!low)
	{
		// The simple, cubic algorithm.
		for(i = 0; i < 3; i++) 
			pos[i] += (high * (PRAND - PRAND)) / 255;
	}
	else
	{
		// The more complicated, spherical algorithm.
		angle_t theta = PRAND << (24 - ANGLETOFINESHIFT);
		angle_t	phi   = (angle_t)(acos(2*(PRAND/255.0f) - 1)/PI * (ANG180 >> ANGLETOFINESHIFT));
		fixed_t off   = ((high - low) * (PRAND - PRAND)) / 255;

		if (off < 0) off -= low; else off += low;

		pos[VX] += FixedMul(FixedMul(finecosine[theta], finesine[phi]), off);
		pos[VY] += FixedMul(FixedMul(finesine  [theta], finesine[phi]), off);
		pos[VZ] += FixedMul(FixedMul(finecosine[phi  ], (int)(FRACUNIT * 0.8333)), off);
	}
}

// Risen3D 3D floor support modified v220-31
static float getLowLimit(subsector_t *ss, mobj_t *src)
{
	if (src)
	{
		int	*flatslist = ss->flats3DArray;

		if (flatslist)
		{
			int i, count = *(int*)flatslist++;

			// Remember - floor's higher than ceiling
			for (i = 0; i < count; i++)
			{
			//	sector_t *lsec = &sectors[flatslist[0]];
				sector_t *rsec = &sectors[flatslist[1]];
				int ceil  = rsec->smoothfloor + flatslist[2];	// offset
				int floor = ceil + flatslist[3];				// depth

				if (src->z >= ceil) return FIX2FLT(floor + FRACUNIT);

				flatslist += FLIST_ENTRIES;
			}
		}
	}

	return -32768.0f;
}

//===========================================================================
// P_NewParticle
//	Spawns a new particle.
//===========================================================================
static void P_NewParticle(ptcgen_t *gen)
{
	ded_ptcgen_t	*def = gen->def;
	modeldef_t		*mf = NULL, *nextmf = NULL;
	subsector_t		*subsec = NULL;
	particle_t		*pt;
	angle_t			ang, ang2;
	int				uncertain, len, i;
	float			*box, inter = -1;

	// Check for a model-only generators.
	if (gen->source)
	{
		inter = R_CheckModelFor(gen->source, &mf, &nextmf);

		if ( mf && mf->flags & MFF_NO_PARTICLES) return;
		if (!mf && def->flags & PGF_MODEL_ONLY ) return;

		// v220-3 scripted underwater player bubbles
		if (   mf
			&& gen->flags & PGF_BUBBLES
			&& !(thinker_uwater.tuwflags & TUWF_HASBUBBLES))
		{
			return;
		}
	}

	// Keep the spawn cursor in the valid range.
	if(++gen->spawncp >= gen->count) gen->spawncp -= gen->count;

	// Set the particle's data.
	pt = gen->ptcs + gen->spawncp;
	pt->stage  = 0;
	pt->subsector = NULL;

	if (M_FRandom() < def->alt_variance) pt->stage = def->alt_start;

	pt->tics = (short)(def->stages[pt->stage].tics 
		* (1 - def->stages[pt->stage].variance * M_FRandom()));

	// Launch vector.
	pt->mov[VX] = (int)(gen->vector[VX] +
		FRACUNIT * (def->vec_variance * (M_FRandom()-M_FRandom()) ));

	pt->mov[VY] = (int)(gen->vector[VY] +
		FRACUNIT * (def->vec_variance * (M_FRandom()-M_FRandom()) ));

	pt->mov[VZ] = (int)(gen->vector[VZ] +
		FRACUNIT * (def->vec_variance * (M_FRandom()-M_FRandom()) ));

	// Apply some aspect ratio scaling to the momentum vector.
	// This counters the 200/240 difference nearly completely.
	pt->mov[VX] = FixedMul(pt->mov[VX], (int)(FRACUNIT * 1.1));
	pt->mov[VZ] = FixedMul(pt->mov[VZ], (int)(FRACUNIT * 1.1));
	pt->mov[VY] = FixedMul(pt->mov[VY], (int)(FRACUNIT * 0.95));

	// Set proper speed.
	uncertain = (int)(def->speed * (1 - def->spd_variance * M_FRandom()) * FRACUNIT);
	len = P_ApproxDistance(P_ApproxDistance(pt->mov[VX], pt->mov[VY]), pt->mov[VZ]);
	if(!len) len = FRACUNIT;
	len = FixedDiv(uncertain, len);

	for (i = 0; i < 3; i++)
		pt->mov[i] = FixedMul(pt->mov[i], len);

	// The source is a mobj?
	if (gen->source) 
	{
		if (gen->flags & PGF_RELATIVE_VELOCITY)
		{
			pt->mov[VX] += gen->source->momx;
			pt->mov[VY] += gen->source->momy;
			pt->mov[VZ] += gen->source->momz;
		}

		// Position.
		pt->pos[VX] = gen->source->x;
		pt->pos[VY] = gen->source->y;
		pt->pos[VZ] = gen->source->z;

		// v220-5 For rockets move the gen smoke back
		if (gen->source->type == MT_ROCKET)
		{
			pt->pos[VX] -= gen->source->momx;
			pt->pos[VY] -= gen->source->momy;
			pt->pos[VZ] -= gen->source->momz;
		}

		P_Uncertain(pt->pos, (int)(FRACUNIT * def->min_spawn_radius), 
			(int)(FRACUNIT * def->spawn_radius));
		
		if (!(gen->flags & PGF_SMOKE))
		{
			// Offset to the real center.
			pt->pos[VZ] += gen->center[VZ];

			// Calculate XY center with mobj angle.
			ang = gen->source->visangle << 16;
			ang2 = (ang + ANG90) >> ANGLETOFINESHIFT;
			ang >>= ANGLETOFINESHIFT;
			pt->pos[VX] += FixedMul(finecosine[ang], gen->center[VX]);
			pt->pos[VY] += FixedMul(finesine  [ang], gen->center[VX]);

			// There might be an offset from the model of the mobj.
			if (mf)
			{
				float off[3];

				memcpy(off, mf->ptcoffset, sizeof(off));
				
				// Interpolate the offset.
				if (inter > 0 && nextmf)
				{
					for(i = 0; i < 3; i++)
						off[i] += (nextmf->ptcoffset[i] - mf->ptcoffset[i]) * inter;	
				}
				
				// Apply it to the particle coords.
				pt->pos[VX] += FixedMul(finecosine[ang] , (int)(FRACUNIT * off[VX]));
				pt->pos[VX] += FixedMul(finecosine[ang2], (int)(FRACUNIT * off[VZ]));
				pt->pos[VY] += FixedMul(finesine[ang]   , (int)(FRACUNIT * off[VX]));
				pt->pos[VY] += FixedMul(finesine[ang2]  , (int)(FRACUNIT * off[VZ]));
				pt->pos[VZ] += (int)(FRACUNIT * off[VY]);

				// v220-3 project the source forward of the viewpoint by 10
				if (gen->flags & PGF_BUBBLES)
				{
					pt->pos[VX] += FixedMul(finecosine[ang], 10*FRACUNIT);
					pt->pos[VY] += FixedMul(finesine[ang]  , 10*FRACUNIT);
					pt->pos[VZ] = viewz + (int)(gen->center[VZ]*FRACUNIT)
						+ (int)(dplookdir*FRACUNIT / 8.0f);	// v220-4 move the z point when looking up/down
				}
			}
		}
	}
	else if (gen->sector) // The source is a plane?
	{
		int		x, y;
		float	dx, dy;
		
		if (!gen->sector->linecount) goto spawn_failed;

		i = gen->stages[pt->stage].radius;

		// Choose a random spot inside the sector, on the spawn plane.
		if (gen->flags & PGF_SPACE_SPAWN)
		{
			pt->pos[VZ] = gen->sector->floorheight + i
				+ FixedMul(PRAND << 8, gen->sector->ceilingheight
				- gen->sector->floorheight - 2*i);
		}
		else
		{
			if (gen->plane && gen->flags & PGF_CEILING_SPAWN)	// if specifically ceiling flagged
			{
				pt->pos[VZ] = gen->sector->ceilingheight - i;	// Spawn on the ceiling
			}
			else 
			{
				pt->pos[VZ] = gen->sector->floorheight + i;		// Spawn on the floor
			}
		}

		// Choosing the XY spot is a bit more difficult.
		// But we must be fast and only sufficiently accurate.
		
		// FIXME: Nothing prevents spawning on the wrong side (or inside)
		// of one-sided walls (large diagonal subsectors!).

		box = gen->sector->sibounds;

		dx = M_FRandom()*(box[BOXRIGHT] - box[BOXLEFT]);
		dy = M_FRandom()*(box[BOXTOP  ] - box[BOXBASE]);

		if (dx > 64.0f) dx = 64.0f; else if (dx < -64.0f) dx = -64.0f;
		if (dy > 64.0f) dy = 64.0f; else if (dy < -64.0f) dy = -64.0f;

		x = (int)(FRACUNIT * (box[BOXLEFT] + dx));
		y = (int)(FRACUNIT * (box[BOXTOP ] + dy));

		subsec = R_PointInSubsector(x, y);

		if (subsec->sector != gen->sector)
			goto spawn_failed;

		// Try a couple of times to get a good random spot.
		for (i = 0; i < 2; i++)
		{
			subsector_t *ss;

			pt->pos[VX] = subsec->ssbox[BOXLEFT] + PRAND * (subsec->ssbox[BOXRIGHT] - subsec->ssbox[BOXLEFT]);
			pt->pos[VY] = subsec->ssbox[BOXBASE] + PRAND * (subsec->ssbox[BOXTOP  ] - subsec->ssbox[BOXBASE]);

			ss = R_PointInSubsector(pt->pos[VX], pt->pos[VY]);

			if (ss->sector == subsec->sector)
			{
				pt->subsector = subsec = ss;
				break; // This is a good place.
			}
		}

		if (i == 2) goto spawn_failed;	// No good place found
	}
	else if(gen->flags & PGF_UNTRIGGERED)
	{
		// The center position is the spawn origin.
		pt->pos[VX] = gen->center[VX];
		pt->pos[VY] = gen->center[VY];
		pt->pos[VZ] = gen->center[VZ];
		P_Uncertain(pt->pos, (int)(FRACUNIT * def->min_spawn_radius),
			(int)(FRACUNIT * def->spawn_radius));
	}

	// The other place where this gets updated is after moving over 
	// a two-sided line.
	if (!(pt->subsector && pt->subsector->sector == gen->sector))
	{
		pt->subsector = R_PointInSubsector(pt->pos[VX], pt->pos[VY]);
	}

	pt->lowlimit = getLowLimit(pt->subsector, gen->source);

	if (   gen->flags & PGF_BUBBLES
		&& !(thinker_uwater.tuwflags & TUWF_HASBUBBLES))
	{
spawn_failed:;
		pt->stage = -1; return;			
	}
}

//===========================================================================
// P_ManyNewParticles
//	Spawn multiple new particles using all applicable sources.
//===========================================================================
void P_ManyNewParticles(ptcgen_t *gen)
{
	thinker_t	*it;
	mobj_t		*mo;

	// Scan all thinkers. 
	for (it = thinkercap.next; it != &thinkercap; it = it->next)
	{
		if (it->function != (void (__cdecl *)(void))P_MobjThinker) continue;

		mo = (mobj_t*)it;
		
		// Type match?
		if (mo->type == gen->type || mo->type == gen->type2)
		{
			// Someone might think this is a slight hack...
			gen->source = mo;
			P_NewParticle(gen);
		}
	}

	gen->source = NULL;	// The generator has no real source.
}

//===========================================================================
// PIT_CheckLinePtc
//===========================================================================
RDBOOL PIT_CheckLinePtc(intercept_t* in)
{
	line_t	*ld = in->d.line;

	if (   pvar.mbox[BOXRIGHT] <= ld->bbox[BOXLEFT ]
		|| pvar.mbox[BOXLEFT ] >= ld->bbox[BOXRIGHT]
		|| pvar.mbox[BOXTOP  ] <= ld->bbox[BOXBASE ]
		|| pvar.mbox[BOXBASE ] >= ld->bbox[BOXTOP  ])
	{
		return true; // Bounding box misses the line completely.
	}

	// Movement must cross the line.
	if ((pvar.side = P_PointOnLineSide(pvar.tmpx1, pvar.tmpy1, ld))
		          == P_PointOnLineSide(pvar.tmpx2, pvar.tmpy2, ld))
	{
		return true;
	}

	if (!(ld->ml_flags & ML_TWOSIDED))
	{
		pvar.ptc_hitline = ld; 
		return false;
	}

	if (ld->specType >= R3D_40)	// R3D lines
	{
		int	 texn = pvar.side ? 0x30 : 0x03;

		if (origTEX[ld - lines] & texn && !(ld->lnflags2 & (LNF2_LOWTEXMASKED|LNF2_3DNOBLOCK)) )
		{
			int		sec3d = ld->frontsector->liftsec3D;
			fixed_t	lo = ld->d3d->offset3D + (sec3d >= 0 ?
								sectors[sec3d].floorheight : ld->backsector->floorheight);
			fixed_t	hi = lo + (ld->lnflags2 & LNF2_MIDTEXMASKED ?
								ld->d3d->lineheightlow : ld->d3d->totalheight);

			if (hi > lo && pvar.tmpz + pvar.tmprad > lo && pvar.tmpz - pvar.tmprad < hi)
			{
				pvar.ptc_hitline = ld;
				return false;
			}
		}
	}
	else
	{
		// Determine the opening we have here.
		fixed_t ceil  = MIN_OF(ld->frontsector->ceilingheight, ld->backsector->ceilingheight);
		fixed_t floor = MAX_OF(ld->frontsector->floorheight  , ld->backsector->floorheight);

		// There is a backsector. We possibly might hit something.
		if (pvar.tmpz - pvar.tmprad < floor || pvar.tmpz + pvar.tmprad > ceil)
		{
			pvar.ptc_hitline = ld;
			return false;
		}
	}

	// There is a possibility that the new position is in a new sector.
	pvar.tmcross = true; // Afterwards, update the sector pointer.

	// False alarm, continue checking.
	return true;
}

//===========================================================================
// P_TouchParticle
//	Particle touches something solid. Returns false if the particle dies.
//===========================================================================
int P_TouchParticle(particle_t *pt, ptcstage_t *stage, RDBOOL touchWall)
{
	if (stage->flags & PTCF_DIE_TOUCH) 
	{
		// Particle dies from touch.
		pt->stage = -1;
		return false;
	}

	if (   stage->flags & PTCF_STAGE_TOUCH
		|| ( touchWall && stage->flags & PTCF_STAGE_WALL_TOUCH)
		|| (!touchWall && stage->flags & PTCF_STAGE_FLAT_TOUCH))
	{
		// Particle advances to the next stage.
		pt->tics = 0;
	}
	// Particle survives the touch.
	return true;
}

//===========================================================================
// P_FixedCrossProduct
//	Semi-fixed, actually.
//===========================================================================
void P_FixedCrossProduct(float *fa, fixed_t *b, fixed_t *result)
{
	result[VX] = FixedMul((int)(FRACUNIT * fa[VY]),
		(int)(b[VZ]) - FixedMul((int)(FRACUNIT * fa[VZ]), b[VY]));

	result[VY] = FixedMul((int)(FRACUNIT * fa[VZ]),
		(int)(b[VX]) - FixedMul((int)(FRACUNIT * fa[VX]), b[VZ]));

	result[VZ] = FixedMul((int)(FRACUNIT * fa[VX]),
		(int)(b[VY]) - FixedMul((int)(FRACUNIT * fa[VY]), b[VX]));
}

//===========================================================================
// P_GetParticleRadius
//	Takes care of consistent variance.
//	Currently only used visually, collisions use the constant radius.
//	The variance can be negative (results will be larger).
//===========================================================================
float P_GetParticleRadius(ded_ptcstage_t *stage_def, int ptc_index)
{
	// R3D made static
	static float rnd[16] = {
		0.875f	, 0.125f	, 0.3125f	, 0.75f,
		0.5f	, 0.375f	, 0.5625f	, 0.0625f,
		1.0f	, 0.6875f	, 0.625f	, 0.4375f,
		0.8125f	, 0.1875f	, 0.9375f	, 0.25f
	};

	if (!stage_def->radius_variance) return stage_def->radius;

	return (rnd[ptc_index & 0xf] * stage_def->radius_variance
		+ (1 - stage_def->radius_variance)) * stage_def->radius;
}

//===========================================================================
// P_GetParticleZ
//	A particle may be attached to the floor or ceiling of the sector.
//===========================================================================
fixed_t P_GetParticleZ(particle_t *pt)
{
	if(pt->pos[VZ] == DDMAXINT)
		return pt->subsector->sector->ceilingheight - 0x200;

	if(pt->pos[VZ] == DDMININT)
		return pt->subsector->sector->floorheight + 0x200;

	return pt->pos[VZ];
}

//===========================================================================
// P_MoveParticle
//	The movement is done in two steps: 
//	Z movement is done first. Skyflat kills the particle.
//	XY movement checks for hits with solid walls (no backsector).
//	This is supposed to be fast and simple (but not too simple).
//===========================================================================
void P_MoveParticle(ptcgen_t *gen, particle_t *pt)
{
	ptcstage_t	*st = gen->stages + pt->stage;
	fixed_t		hardRadius = st->radius/2;
	RDBOOL		zbounce = false;
	RDBOOL		hitFloor;
	int			i, set_x, set_y, set_z;

	//-------------------------------

	// Changes to momentum.
	pt->mov[VZ] -= FixedMul(mapgravity, st->gravity);

	// Sphere force pull and turn.
	// Only applicable to sourced or untriggered generators. For other
	// types it's difficult to define the center coordinates.
	if (  st->flags & PTCF_SPHERE_FORCE
		&& (gen->source || gen->flags & PGF_UNTRIGGERED))
	{
		fixed_t delta[3], cross[3], dist;

		if (gen->source)
		{
			delta[VX] = pt->pos[VX] - gen->source->x;
			delta[VY] = pt->pos[VY] - gen->source->y;
			delta[VZ] = P_GetParticleZ(pt) - (gen->source->z + gen->center[VZ]);
		}
		else
		{
			for(i = 0; i < 3; i++) delta[i] = pt->pos[i] - gen->center[i];
		}

		// Apply the offset (to source coords).
		for (i = 0; i < 3; i++) 
			delta[i] -= (int)(FRACUNIT * gen->def->force_origin[i]);

		// Counter Doom's aspect ratio
		delta[VZ] = FixedMul(delta[VZ], 0x13333); // (int)(1.2f * FRACUNIT)

		dist = P_ApproxDistance(P_ApproxDistance(delta[VX], delta[VY]), delta[VZ]);

		if(dist)
		{
			// Radial force pushes the particles on the surface of a sphere.
			if (gen->def->force)
			{
				int a = (int)(dist - FRACUNIT * gen->def->force_radius);
				int b = (int)(FRACUNIT * gen->def->force);

				// Normalize delta vector, multiply with (dist - forceRadius),
				// multiply with radial force strength.
				for (i = 0; i < 3; i++)
					pt->mov[i] -= FixedMul( FixedMul(FixedDiv(delta[i], dist), a ), b);
			}

			// Rotate!
			if (   gen->def->force_axis[VX]
				|| gen->def->force_axis[VY]
				|| gen->def->force_axis[VZ])
			{
				P_FixedCrossProduct(gen->def->force_axis, delta, cross);

				for(i = 0; i < 3; i++)
					pt->mov[i] += cross[i] >> 8;
			}
		}
	}

	if (st->resistance != FRACUNIT)
	{
		for(i = 0; i < 3; i++)
			pt->mov[i] = FixedMul(pt->mov[i], st->resistance);
	}

	// The particle is 'soft': half of radius is ignored.
	// The exception is plane flat particles, which are rendered flat 
	// against planes. They are almost entirely soft when it comes to plane
	// collisions.
	if (st->flags & PTCF_PLANE_FLAT) hardRadius = FRACUNIT;

	// Check the new Z position only if not stuck to a plane.
	if (pt->pos[VZ] != DDMININT && pt->pos[VZ] != DDMAXINT)
	{
		set_z = pt->pos[VZ] + pt->mov[VZ];

		// R3D underwater bubble support
		if (   gen->flags & PGF_BUBBLES
			&& (pt->subsector->sector->heightsec < 0
			||  set_z > sectors[pt->subsector->sector->heightsec].floorheight) )
		{
			// Kill the particle; z is above the fake ceiling or
			// the particle is not in a sector with a fake sector
			pt->stage = -1;
			return;
		}

		if (set_z > pt->subsector->sector->ceilingheight - hardRadius)
		{
			// The Z is through the roof! 
			if (pt->subsector->sector->ceilingpic == skyflatnum) 
			{
				// Special case: particle gets lost in the sky.
				pt->stage = -1;
				return;
			}

			if (!P_TouchParticle(pt, st, false)) return;

			set_z = pt->subsector->sector->ceilingheight - hardRadius;
			zbounce = true;
			hitFloor = false;
		}

		// Also check the floor.
		if (set_z < pt->subsector->sector->floorheight + hardRadius)
		{
			if(pt->subsector->sector->floorpic == skyflatnum) 
			{
				pt->stage = -1;
				return;
			}

			if(!P_TouchParticle(pt, st, false)) return;

			set_z = pt->subsector->sector->floorheight + hardRadius;
			zbounce = true;
			hitFloor = true;
		}

		if(zbounce) 
		{
			pt->mov[VZ] = FixedMul(-pt->mov[VZ], st->bounce);

			if (!pt->mov[VZ])
			{
				// The particle has stopped moving. This means its Z-movement 
				// has ceased because of the collision with a plane. Plane-flat
				// particles will stick to the plane.
				if(st->flags & PTCF_PLANE_FLAT)
				{
					set_z = hitFloor? DDMININT : DDMAXINT;
				}
			}
		}
		
		// Move to the new Z coordinate.
		pt->pos[VZ] = set_z;
	}

	// Now check the XY direction. 
	// - Check if the movement crosses any solid lines.
	// - If it does, quit when first one contacted and apply appropriate
	//   bounce (result depends on the angle of the contacted wall).
	set_x = pt->pos[VX] + pt->mov[VX];
	set_y = pt->pos[VY] + pt->mov[VY];

	pvar.tmcross = false; // Has crossed potential sector boundary?

	// XY movement can be skipped if the particle is not moving on the
	// XY plane.
	if (!pt->mov[VX] && !pt->mov[VY]) 
	{
		// If the particle is contacting a line, there is a chance that the
		// particle should be killed (if it's moving slowly at max).
		if (pt->contact)
		{
			sector_t *front = pt->contact->frontsector;
			sector_t *back  = pt->contact->backsector;

			if (front && back && abs(pt->mov[VZ]) < FRACUNIT/2)
			{
				// If the particle is in the opening of a 2-sided line, it's
				// quite likely that it shouldn't be here...
				if (   P_GetParticleZ(pt) > MAX_OF(front->floorheight, back->floorheight) 
					&& P_GetParticleZ(pt) < MIN_OF(front->ceilingheight , back->ceilingheight ) )
				{
					// Kill the particle.
					pt->stage = -1;
					return;
				}
			}
		}

		goto end_xy_iteration;
	}

	if (gen->flags & PGF_SUSPEND)
		return;	// v220-4 These types never hit a wall

	// We're moving in XY, so if we don't hit anything there can't be any 
	// line contact.
	pt->contact = NULL;

	// Bounding box of the movement line.
	pvar.tmpz = set_z;
	pvar.tmprad = st->radius;
	pvar.tmpx1 = pt->pos[VX];
	pvar.tmpx2 = set_x;
	pvar.tmpy1 = pt->pos[VY];
	pvar.tmpy2 = set_y;

	pvar.mbox[BOXTOP  ] = M_AddLIMIT(MAX_OF(set_y, pt->pos[VY]),  st->radius);
	pvar.mbox[BOXBASE ] = M_AddLIMIT(MIN_OF(set_y, pt->pos[VY]), -st->radius);
	pvar.mbox[BOXRIGHT] = M_AddLIMIT(MAX_OF(set_x, pt->pos[VX]),  st->radius);
	pvar.mbox[BOXLEFT ] = M_AddLIMIT(MIN_OF(set_x, pt->pos[VX]), -st->radius);

	pvar.ptc_hitline = NULL;

	// v220-4 trace lines in correct order to avoid wrong line being processed
	P_PathTraverse(pvar.tmpx1, pvar.tmpy1, pvar.tmpx2, pvar.tmpy2, false, PT_ADDLINESPTC, (int (__cdecl *)(struct intercept_s *))PIT_CheckLinePtc);

	if (pvar.ptc_hitline)
	{
		float	dotpf;
		int		normal[2], dotp;
		
		// Must survive the touch.
		if (!P_TouchParticle(pt, st, true)) return;

		// Correct line side ?
		if (pvar.side && !pvar.ptc_hitline->backsector)
		{
			pt->stage = -1;	// Kill the particle.
			return;
		}

		// There was a hit! Calculate bounce vector.
		// - Project movement vector on the normal of hitline.
		// - Calculate the difference to the point on the normal.
		// - Add the difference to movement vector, negate movement.
		// - Multiply with bounce.
		// Calculate the normal. 
		normal[VX] = -pvar.ptc_hitline->dx;
		normal[VY] = -pvar.ptc_hitline->dy;

		if ( !(dotpf = DOT2F(normal, normal)) ) return;	// Unlikely

		// Calculate as floating point so we don't overflow.
		dotp = (int)(FRACUNIT * (DOT2F(pt->mov, normal) / dotpf));
		VECMUL(normal, dotp);
		VECSUB(normal, pt->mov);
		VECMULADD(pt->mov, 2*FRACUNIT, normal);
		VECMUL(pt->mov, st->bounce);

		// Continue from the old position - dont update pt->pos.
		// This line is the latest contacted line.
		pt->contact = pvar.ptc_hitline;
		return;
	}

end_xy_iteration:

	// The move is OK.
	pt->pos[VX] = set_x;
	pt->pos[VY] = set_y;
	
	// Should we update the subsector pointer?
	if (pvar.tmcross)
	{
		pt->subsector = R_PointInSubsector(set_x, set_y);
		pt->lowlimit = getLowLimit(pt->subsector, gen->source);
	}
}

//===========================================================================
// P_PtcGenThinker
//	Spawn and move particles.
//===========================================================================
void P_PtcGenThinker(ptcgen_t *gen)
{
	ded_ptcgen_t	*def = gen->def;
	particle_t		*pt;
	float			newparts;
	int				i;

	// Check if source should be destroyed
	if (   !(gen->flags & PGF_UNTRIGGERED) 
		&& !P_IsUsedMobjID((thid_t)gen->srcid))
	{
		gen->source = NULL;		// Spawning new particles is not possible
	}
	else
	{
		// v220-4 Don't waste time calculating positions for generator particles
		//        that are not visible. Note: only for gens that are attached to
		//        a mobj that does not move.  The gnf_suspend flag has to be set
		//        for any such type in the ded. R3D does not check for compliance.

		if (gen->source && gen->flags & PGF_SUSPENDED) return;
	}
		
	// Time to die?
	if (++gen->age > def->max_age && def->max_age >= 0) 
	{
		P_FreePtcGen(gen);
		return;
	}

	// Spawn new particles?
	if (
		(gen->age <= def->spawn_age || def->spawn_age < 0)
		&&
		(  gen->source
		|| gen->sector 
		|| gen->type >= 0 
		|| gen->flags & PGF_UNTRIGGERED)
		)
	{
		if (gen->flags & (PGF_PARTS_PER_128|PGF_SCALED_RATE))
			newparts = def->spawn_rate * gen->area;	// Density spawning
		else
			newparts = def->spawn_rate;				// Normal spawning

		newparts *= r_particle_spawn_rate 
			* (1 - def->spawn_variance * M_FRandom());

		gen->spawncount += newparts;

		while(gen->spawncount >= 1)
		{
			// Spawn a new particle.
			if (gen->type >= 0) // Type-triggered?
				P_ManyNewParticles(gen);
			else
				P_NewParticle(gen);

			gen->spawncount--;
		}
	}

	// Move particles.
	for (i = 0, pt = gen->ptcs; i < gen->count; i++, pt++)
	{
		if (pt->stage < 0) continue; // Not in use.

		if (pt->tics-- <= 0)
		{
			// Advance to next stage.
			if (   ++pt->stage == MAX_PTC_STAGES
				|| gen->stages[pt->stage].type == PTC_NONE)
			{
				pt->stage = -1;	// Kill the particle
				continue;
			}

			pt->tics = (short)(def->stages[pt->stage].tics
				* (1 - def->stages[pt->stage].variance * M_FRandom()));
		}

		P_MoveParticle(gen, pt);	// Try to move.
	}
}

//===========================================================================
// P_GetPtcGenForFlat
//	Returns the ptcgen definition for the given flat.
//===========================================================================
ded_ptcgen_t *P_GetPtcGenForFlat(int flatpic)
{
	int	i;

	for(i = 0; i < defs.count.ptcgens.num; i++)
	{
		if(defs.ptcgens[i].flat_num == flatpic)
			return defs.ptcgens + i;
	}

	return NULL;
}

//===========================================================================
// P_HasActivePtcGen
//	Returns true if there is an active ptcgen for the given plane.
//  0 floor 1 ceiling
//===========================================================================
RDBOOL P_HasActivePtcGen(sector_t *sector, int plane)
{
	int i;

	for(i = 0; i < MAX_ACTIVE_PTCGENS; i++)
	{
		if (   active_ptcgens[i] 
			&& active_ptcgens[i]->sector == sector
			&& active_ptcgens[i]->plane == plane)
		{
			return true;
		}
	}
	return false;
}

//===========================================================================
// P_CheckPtcPlanes
//	Spawns new ptcgens for planes, if necessary.
//===========================================================================
void P_CheckPtcPlanes(void)
{
	ded_ptcgen_t *def;
	int			 i, p, plane;

	// There is no need to do this on every tic.
	if (!defsectors || gametic % 4) return; 

	for (i = 0; i < defseccount; i++)
	{
		sector_t *sec = defsectors[i].sec;

		// R3D ignore empty or unseen sectors
		if (sec->siflags & SIF_VISIBLE)
		{
			for(p = 0; p < 2; p++)
			{
				if ( !(def = defsectors[i].def[p]) ) continue;

				plane = p;

				if (def->flags & PGF_FLOOR_SPAWN)
				{
					plane = 0;
				}
				else if(def->flags & PGF_CEILING_SPAWN)	// Floor takes precedence
				{
					plane = 1;
				}

				if (!P_HasActivePtcGen(sec, plane))
				{
					// Spawn it!
					P_SpawnPlaneParticleGen(def, sec, plane);
				}
			}
		}
	}
}

//===========================================================================
// P_SpawnTypeParticleGens
//	Spawns all type-triggered particle generators, regardless of whether
//	the type of thing exists in the level or not (things might be 
//	dynamically created).
//===========================================================================
void P_SpawnTypeParticleGens(void)
{
	int i;
	ded_ptcgen_t *def;
	ptcgen_t *gen;

	PG_InitForLevel();

	for(i = 0, def = defs.ptcgens; i < defs.count.ptcgens.num; i++, def++)
	{
		if (def->type_num < 0) continue;

		if (!(gen = P_NewPtcGen())) return; // No more generators.

		// Initialize the particle generator.
		gen->count = def->particles;
		P_InitParticleGen(gen, def);
		gen->type = def->type_num;
		gen->type2 = def->type2_num;

		// Is there a need to pre-simulate?
		P_PresimParticleGen(gen, def->presim);
	}
}

//===========================================================================
// P_SpawnMapParticleGens
//===========================================================================
void P_SpawnMapParticleGens()
{
	int i;
	ded_ptcgen_t *def;
	ptcgen_t *gen;

	if(!r_use_particles) return;

	for(i = 0, def = defs.ptcgens; i < defs.count.ptcgens.num; i++, def++)
	{
		if (!def->map[0] || _stricmp(def->map, mapinfo->id)) continue;

		if (!(gen = P_NewPtcGen())) return; // No more generators.

		// Initialize the particle generator.
		gen->count = def->particles;
		P_InitParticleGen(gen, def);
		gen->flags |= PGF_UNTRIGGERED;

		// Is there a need to pre-simulate?
		P_PresimParticleGen(gen, def->presim);
	}
}

//===========================================================================
// P_SpawnDamageParticleGen
//	A public function (games can call this directly).
//===========================================================================
void P_SpawnDamageParticleGen(mobj_t *mo, mobj_t *inflictor, int amount)
{
	ptcgen_t *gen;
	ded_ptcgen_t *def;
	int i;
	fixed_t len;

	// Are particles allowed?
	if (   !r_use_particles
		|| !mo
		|| !inflictor
		|| amount <= 0
		|| mo->flags2 & MF2_DONTDRAW) 
		return;
		
	// Search for a suitable definition.
	for(i = 0, def = defs.ptcgens; i < defs.count.ptcgens.num; i++, def++)
	{
		// It must be for this type of mobj.
		if(def->damage_num != mo->type) continue; 

		// Create it.
		if(!(gen = P_NewPtcGen())) return; // No more generators.
		gen->count = def->particles;
		P_InitParticleGen(gen, def);
		gen->flags |= PGF_UNTRIGGERED;
		gen->area = (float)amount;
		if(gen->area < 1) gen->area = 1;

		// Calculate appropriate center coordinates and the vector.
		gen->center[VX] += mo->x;
		gen->center[VY] += mo->y;
		gen->center[VZ] += mo->z + mo->height/2;
		gen->vector[VX] += mo->x - inflictor->x;
		gen->vector[VY] += mo->y - inflictor->y;
		gen->vector[VZ] += mo->z + mo->height/2 - inflictor->z 
			- inflictor->height/2;
		
		// Normalize the vector.
		len = P_ApproxDistance(P_ApproxDistance(gen->vector[VX], 
			gen->vector[VY]), gen->vector[VZ]);
		if(len) 
		{
			int k;
			for(k = 0; k < 3; k++) 
				gen->vector[k] = FixedDiv(gen->vector[k], len);
		}

		// Is there a need to pre-simulate?
		P_PresimParticleGen(gen, def->presim);
	}
}

// P_InitParticledefSectors

void P_InitParticledefSectors()
{
	ded_ptcgen_t	*deff, *defc;
	sector_t		*sec;
	int				c, i;
	
	defsectors = NULL;

	for (defseccount = i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (sec->linecount)
		{
			if (   (deff = P_GetPtcGenForFlat(sec->floorpic))
				|| (defc = P_GetPtcGenForFlat(sec->ceilingpic)) )
			{
				defseccount++;
			}
		}
	}

	if (defseccount)
	{
		defsectors = (defsector_t*)W_Calloc(defseccount * sizeof(defsector_t), LEVELID, 5);

		for (c = i = 0, sec = sectors; i < numsectors; i++, sec++)
		{
			if (sec->linecount)
			{
				deff = P_GetPtcGenForFlat(sec->floorpic);
				defc = P_GetPtcGenForFlat(sec->ceilingpic);

				if (deff || defc)
				{
					defsectors[c].def[0] = deff;
					defsectors[c].def[1] = defc;
					defsectors[c++].sec = sec;
				}
			}
		}
	}
}
