//-----------------------------------------------------------------------------
//
//  Copyright (C) 1998 - 2005 id Software Inc.,
//   Chi Hoang, Lee Killough, Jim Flynn, Rand Phares, Ty Halderman,
//  Copyright (C) 2008 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_think.c
//
// Revision R3D v2.1.0 build 62 04/Oct/2006 gmj
//  modified/extended for R3D
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doom, BOOM v2.01
//
// DESCRIPTION:   Thinker handler.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <string.h>

#include "de_base.h"

#include "m_winmem.h"
#include "p_maputil.h"
#include "con_console.h"
#include "r3d_compat.h"
#include "r3d_sfx.h"
#include "r3d_modelbox.h"
#include "p_particle.h"
#include "doom/p_local.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void P_ClearMobjIDs(void);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// R3D thinkers split into 3 main classes
thinker_t	thinkercap;	// Head and tail of the mobj thinker list
thinker_t	fnthinkcap;	// Head and tail of the functions thinker list
thinker_t	ptthinkcap;	// Head and tail of the particles thinker list

// killough 8/29/98: we maintain several separate threads, each containing
// a special class of thinkers, to allow more efficient searches.
// R3D - used with the mobj thinker list 
thinker_t	thinkerclasscap[NUMTHCLASS];

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static thinker_t *currentthinker;

// CODE --------------------------------------------------------------------

// Running thinkers is too important to allow any funny business
#pragma optimize("g",off)

// PRIVATE DATA DEFINITIONS ------------------------------------------------

//==========================================================================
// P_InitParticleThinkers
//==========================================================================
void P_InitParticleThinkers(void)
{
	W_FreeALL(PARTCLID);
	ptthinkcap.prev = ptthinkcap.next = &ptthinkcap;
}

//==========================================================================
// P_InitThinkers
//==========================================================================
void P_InitThinkers(void)
{
	int i;

	// Initialise function thinkers
	W_FreeALL(THINKID);
	fnthinkcap.prev = fnthinkcap.next = &fnthinkcap;

	// Initialise mobj thinkers
	thinkercap.prev = thinkercap.next = &thinkercap;

	for (i = 0; i < NUMTHCLASS; i++)  // killough 8/29/98: initialize threaded lists
		thinkerclasscap[i].cprev = thinkerclasscap[i].cnext = &thinkerclasscap[i];

	memset(mobjarray, 0, sizeof(mobjarray));

	ma_inf.mobjsCreated = ma_inf.mobjsRemoved = 0;	// For stats on exit

	P_ClearMobjIDs();

	SetSFXnoway(NULL, -1);
}

//==========================================================================
// P_RemoveThinker
//	Deallocation is lazy -- it will not actually
//  be freed until its thinking turn comes up.
//  Players are not freed until the map is restarted
//==========================================================================

//==========================================================================
// 
// Maintain separate threads - killough 8/29/98
//  b211-0 - revised
//
//==========================================================================
void P_UpdateThinker(thinker_t *thinker)
{
	mobj_t		*mo = (mobj_t*)thinker;
	thinker_t	*th = thinker->cnext;
	int			thclass = th_misc;	// default

	if (thinker->function != (void (__cdecl *)(void))P_MobjThinker)
		Con_ErrorD("P_UpdateThinker: wrong type");

	// find the class the thinker belongs to if not th_misc
	if (mo->health > 0 && (mo->flags & MF_COUNTKILL || mo->type == MT_SKULL))
	{
		thclass = mo->flags3 & MF3_FRIEND ? th_friends : th_enemies;
	}

	// Remove from current thread
	(th->cprev = thinker->cprev)->cnext = th;
	
	// Add to appropriate thread
	th = &thinkerclasscap[thclass];
	th->cprev->cnext = thinker;
	thinker->cnext = th;
	thinker->cprev = th->cprev;
	th->cprev = thinker;
}

// P_RemoveThinkerDelayed()
//
// Called automatically as part of the thinker loop in P_RunThinkers(),
// on nodes which are pending deletion.
//
// If this thinker has no more pointers referencing it indirectly,
// remove it, and set currentthinker to one node preceeding it, so
// that the next step in P_RunThinkers() will get its successor.
//
// Only used with mobj's

static void P_RemoveThinkerDelayed(thinker_t *thinker)
{
	if (!thinker->mobjreferences)
	{
		thinker_t *next = thinker->next;

		P_SetMobjID(thinker->id, false);	// Flag the ID as free.
		thinker->id = thinker->spawnid = 0; // b211-0
		(next->prev = currentthinker = thinker->prev)->next = next;
		ma_inf.mobjsRemoved++;	// For stats on exit
	}
}

void P_RemoveThinker(thinker_t *thinker)
{
	if (thinker->function == (void (__cdecl *)(void))P_MobjThinker)
	{
		thinker->function = (void (__cdecl *)(void))P_RemoveThinkerDelayed;

		// killough 8/29/98: remove immediately from threaded list
		(thinker->cnext->cprev = thinker->cprev)->cnext = thinker->cnext;
		
		thinker->spawnid = 0;
	}
	else	// fnthinkers
	{
		thinker->function = (think_t) -1;
	}
}

//==========================================================================
// P_RunThinkers
//==========================================================================
void P_RunThinkers(void)
{
	thinker_t	*fnthinker;
	int			(*thfunc)(void*);

	stats.global_mobj_count = stats.missile_mobj_count = 0;
	ma_inf.isdropped = 0;
	reduceNearClip = false;

	currentthinker = thinkercap.next;	// Run MOBJ thinkers first

	while(currentthinker != &thinkercap)
	{
		thfunc = (int (__cdecl *)(void *))currentthinker->function;

		if (thfunc == (int (__cdecl *)(void*))P_MobjThinker)
		{
			P_MobjThinker((mobj_t*)currentthinker);				

			// v220-10 If still active clear MJMISC_TRANS_TC flag if set.
			// Note this must be done after running the mobj thinker
			// and before running scroll thinkers.
			if (currentthinker->function == (void (__cdecl *)(void))P_MobjThinker)
			{
				((mobj_t*)currentthinker)->misc &= ~MJMISC_TRANS_TC;
			}
		}
		else if (thfunc == (int (__cdecl *)(void*))P_RemoveThinkerDelayed)
		{
			P_RemoveThinkerDelayed((thinker_t*)currentthinker);
		}
		else
		{
			thinker_t *next = currentthinker->next;
			Con_Message("ERROR WARNING: invalid thinker still linked %p\n", currentthinker->function);
			currentthinker->next->prev = currentthinker->prev;
			currentthinker->prev->next = next;
		}

		currentthinker = currentthinker->next;
	}

	// Run the model box thinkers
	MB_RunModelBoxThinkers();
	
	// Run THINKID things
	fnthinker = fnthinkcap.next;
	
	while (fnthinker != &fnthinkcap)
	{
		if (fnthinker->function == (think_t)-1)
		{
			thinker_t *next = fnthinker->next;
			// Time to remove it
			fnthinker->next->prev = fnthinker->prev;
			fnthinker->prev->next = next;

			W_Free(fnthinker, THINKID, 509);
		}
		else if (fnthinker->function)	// should be
		{
			thfunc = (int (__cdecl *)(void *))fnthinker->function;

			if (thfunc == (int (__cdecl *)(void*))P_MobjLinkFlame)
			{
				// do nothing. Only used as an identifier in R_AddSprites()
			}
#ifndef _DEBUG
			else	// v220-35 direct function call
			{
				thfunc((void*)fnthinker);
			}
#else
			// function check
			else if (thfunc == (int (__cdecl *)(void*))T_CeilInStasis)
				T_CeilInStasis((void*)fnthinker);
			else if (thfunc == (int (__cdecl *)(void*))T_PlatInStasis)
				T_PlatInStasis((void*)fnthinker);
			else if (thfunc == (int (__cdecl *)(void*))T_MoveCeiling)
				T_MoveCeiling((ceiling_t*)fnthinker);
			else if (thfunc == (int (__cdecl *)(void*))T_VerticalDoor)
				T_VerticalDoor((vldoor_t*)fnthinker);
			else if (thfunc == (int (__cdecl *)(void*))T_MoveFloor)
				T_MoveFloor((floormove_t*)fnthinker);
			else if (thfunc == (int (__cdecl *)(void*))T_PlatRaise)
				T_PlatRaise((plat_t*)fnthinker);
			else if (thfunc == (int (__cdecl *)(void*))T_LightFlash)
				T_LightFlash((lightflash_t*)fnthinker);
			else if (thfunc == (int (__cdecl *)(void*))T_StrobeFlash)
				T_StrobeFlash((strobe_t*)fnthinker);
			else if (thfunc == (int (__cdecl *)(void*))T_Glow)
				T_Glow((glow_t*)fnthinker);
			else if (thfunc == (int (__cdecl *)(void*))T_FireFlicker)
				T_FireFlicker((fireflicker_t*)fnthinker);
			else if (thfunc == (int (__cdecl *)(void*))T_MoveElevator)
				T_MoveElevator((elevator_t*)fnthinker);
			else if (thfunc == (int (__cdecl *)(void*))T_Scroll)
				T_Scroll((scroll_t*)fnthinker);
			else if (thfunc == (int (__cdecl *)(void*))T_Pusher)
				T_Pusher((pusher_t*)fnthinker);
			else if (thfunc == (int (__cdecl *)(void*))T_Phased)
				T_Phased((phased_t*)fnthinker);
			else if (thfunc == (int (__cdecl *)(void*))T_SinglePhased)
				T_SinglePhased((phased_t*)fnthinker);
			else if (thfunc == (int (__cdecl *)(void*))T_DynLightFlash)
				T_DynLightFlash((dynstrobe_t*)fnthinker);
			else if (thfunc == (int (__cdecl *)(void*))T_DynStrobeFlash)
				T_DynStrobeFlash((dynstrobe_t*)fnthinker);
			else if (thfunc == (int (__cdecl *)(void*))T_DynFlicker)
				T_DynFlicker((dynstrobe_t*)fnthinker);
			else if (thfunc == (int (__cdecl *)(void*))T_DynGlow)
				T_DynGlow((dynglow_t*)fnthinker);
			else if (thfunc == (int (__cdecl *)(void*))T_DynFireFlicker)
				T_DynFireFlicker((dynfireflicker_t*)fnthinker);
			else if (R3D_ScriptedMap)
			{
				if (thfunc == (int (__cdecl *)(void*))T_ModelKill)
					T_ModelKill((modelkill_t*)fnthinker);
				else if (thfunc == (int (__cdecl *)(void*))T_ModelSpawn)
					T_ModelSpawn((modelspawn_t*)fnthinker);
				else if (thfunc == (int (__cdecl *)(void*))T_ModelReset)
					T_ModelReset((modelreset_t*)fnthinker);
				else if (thfunc == (int (__cdecl *)(void*))T_SwitchReset)
					T_SwitchReset((modelreset_t*)fnthinker);
				else if (thfunc == (int (__cdecl *)(void*))T_SoundActivate)
					T_SoundActivate((scriptedsound_t*)fnthinker);
				else if (thfunc == (int (__cdecl *)(void*))T_DelayUnblock)
					T_DelayUnblock((delayline_t*)fnthinker);
				else if (thfunc == (int (__cdecl *)(void*))T_DelayResurrect)
					T_DelayResurrect((modelreset_t*)fnthinker);
				else if (thfunc == (int (__cdecl *)(void*))T_StateMonitor)
					T_StateMonitor((modelreset_t*)fnthinker);
				else if (thfunc == (int (__cdecl *)(void*))T_KillDelayed)
					T_KillDelayed((delayline_t*)fnthinker);
				else if (thfunc == (int (__cdecl *)(void*))T_LineOcclude)
					T_LineOcclude((lineocclude_t*)fnthinker);
				else if (thfunc == (int (__cdecl *)(void*))T_AdjustCeil)
					T_AdjustCeil((adjustplat_t*)fnthinker);
				else if (thfunc == (int (__cdecl *)(void*))T_PlayListsound)
					T_PlayListsound((delayline_t*)fnthinker);
				else if (thfunc == (int (__cdecl *)(void*))T_ActivateDelayed)
					T_ActivateDelayed((modelreset_t*)fnthinker);
				else if (thfunc == (int (__cdecl *)(void*))T_Endlevel)
					T_Endlevel((delayline_t*)fnthinker);
				else
					Con_Error("Unknown THINKER called");
			}
			else
				Con_Error("Unknown THINKER called");
#endif
		}

		fnthinker = fnthinker->next;
	}

	// Run Particle thinkers
	fnthinker = ptthinkcap.next;
	
	while(fnthinker != &ptthinkcap)
	{
		thinker_t *next = fnthinker->next;

		if (fnthinker->function == (think_t) -1)
		{
			// Time to remove it
			fnthinker->next->prev = fnthinker->prev;
			fnthinker->prev->next = next;

			W_Free(fnthinker, PARTCLID, 0);
		}
		else
			P_PtcGenThinker((ptcgen_t*)fnthinker);

		fnthinker = next;
	}

	P_UpdatePhasedLights();	// v220-31
	
	// Dedicated thinkers
	T_Underwater();
	T_SFXNoWay();
	T_MAPMusic();

	Sfx_LineSounds();
}

//==========================================================================
// P_AddThinker
//	Adds a new mobj thinker at the end of the list.
//==========================================================================
void P_AddThinker(thinker_t *thinker)
{
	if (thinker->function != (void (__cdecl *)(void))P_MobjThinker)	
		Con_Error("FN thinker on wrong list");

	// Link the thinker to the thinker list.
	thinkercap.prev->next = thinker;
	thinker->next = &thinkercap;
	thinker->prev = thinkercap.prev;
	thinkercap.prev = thinker;

    // killough 11/98: init reference counter to 0
	thinker->mobjreferences = 0;
	
	// killough 8/29/98: set sentinel pointers, and then add to appropriate list
	thinker->cnext = thinker->cprev = thinker;
	P_UpdateThinker(thinker);
}

//==========================================================================
// P_AddFNThinker
//	Adds a new think thinker at the end of the list.
//==========================================================================
void P_AddFNThinker(thinker_t *thinker)
{
	if (   thinker->function == (void (__cdecl *)(void))P_MobjThinker
		|| thinker->function == (void (__cdecl *)(void))P_PtcGenThinker)
	{
		Con_Error("Mobj thinker on wrong list");
	}

	// Link the thinker to the thinker list.
	fnthinkcap.prev->next = thinker;
	thinker->next = &fnthinkcap;
	thinker->prev = fnthinkcap.prev;
	fnthinkcap.prev = thinker;

	thinker->id = 0;				// Zero indicates it's not a mobj
}

//==========================================================================
// P_AddParticleThinker
//	Adds a new particle thinker at the end of the list.
//==========================================================================
void P_AddParticleThinker(thinker_t *thinker)
{
	if (thinker->function != (void (__cdecl *)(void))P_PtcGenThinker)	
		Con_Error("Mobj thinker on wrong list");

	// Link the thinker to the thinker list.
	ptthinkcap.prev->next = thinker;
	thinker->next = &ptthinkcap;
	thinker->prev = ptthinkcap.prev;
	ptthinkcap.prev = thinker;

	thinker->id = 0;				// Zero indicates it's not a mobj
}

#pragma optimize("", on)
