//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_iterate.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
//
// DESCRIPTION: analyses multiple sector nesting
//
//-----------------------------------------------------------------------------

#include <memory.h>
#include <math.h>

#include "r3d_MAINspec.h"
#include "m_misc.h"

static int		conres;
static line_t	*ref;

//--------------------------------------------------------------------------

static RDBOOL checkProjectedXpoint(line_t *cmp)
{
	vertex_t	*s, *e, *d, test;
	double		ax, ay, bx, by;
	double		cx, cy, dx, dy;
	double		dv, r;

	// Check lines don't form a straight line
	if (subfn_CheckConnectedLineSlope(cmp, ref))
		return true;

	// Quick vertical/horizontal slope reject
	if ((ref->slopetype == ST_HORIZONTAL &&
		 cmp->slopetype == ST_HORIZONTAL) ||
		(ref->slopetype == ST_VERTICAL &&
		 cmp->slopetype == ST_VERTICAL))
	{
		return false;
	}

	// It's only lines that are angled that cause problems
	if ((ref->slopetype == ST_HORIZONTAL &&
		 cmp->slopetype == ST_VERTICAL) ||
		(ref->slopetype == ST_HORIZONTAL &&
		 cmp->slopetype == ST_VERTICAL))
	{
		return false;
	}

	// If vertexes are shared then not intersecting
	if (ref->v1 == cmp->v1
		|| ref->v1 == cmp->v2
		|| ref->v2 == cmp->v1
		|| ref->v2 == cmp->v2)
	{
		return false;
	}

	s = ref->v1, e = ref->v2, d = cmp->v1;
	ax = (double)s->x, ay = (double)s->y, bx = (double)e->x, by = (double)e->y;
	cx = (double)d->x, cy = (double)d->y, dx = cx+(double)cmp->dx, dy = cy+(double)cmp->dy;

	// If dv zero then lines have same slope (and therefore cannot intersect)
	if (!(dv = (bx - ax)*(dy - cy) - (by - ay)*(dx - cx)))
	{
		return false;
	}

	r = ((ay - cy)*(dx - cx) - (ax - cx)*(dy - cy)) / dv;

	// If the intersect is on line ref then reject cmp as a valid line
	test.x = (fixed_t)(ax + r*(bx - ax));
	test.y = (fixed_t)(ay + r*(by - ay));

	return VertexIsOnLine(&test, ref);
}

// Don't use seg fv float values.	rev60
static __int64 getDist(vertex_t *v1, vertex_t *v2)
{
	double d1 = (double)v1->x - (double)v2->x;
	double d2 = (double)v1->y - (double)v2->y;

	return (__int64)sqrt(d1*d1 + d2*d2);
}

static seg_t *subfn_FindNearestSeg(line_t *ld)
{
	byte	*linetest = (byte*)M_Calloc(numlines, 272);
	byte	*segarray = (byte*)M_Calloc(numsegs , 273);
	seg_t	*sg = segs;
	int		i;
    int		xl, xh, yl, yh, x, y;
	int		count = 0;

	__int64 maxdist = 0x7FFFFFFFFFFFFFFF;

	seg_t	*nearestseg = NULL;

	validcount++;
	ld->validcount = validcount;	//Don't test self

	M_getBlockOffsets(ld->bbox, 0, &xl, &xh, &yl, &yh);

	for (x = xl ; x <= xh ; x++)
	{
		for (y = yl ; y <= yh ; y++)
		{
			const long	*list;
			int			offset;

			if (x < 0 || y < 0 || x >= bmapwidth || y >= bmapheight)
				continue;

			offset = y*bmapwidth + x;

			offset = *(blockmap + offset);
			list = blockmaplump + offset;

			// skip starting delimiter
			while( *(++list) >= 0)
			{
				line_t *lt = &lines[*list];

				if (lt->validcount == validcount)
					continue;		// line has already been checked

				lt->validcount = validcount;

				if (!(lt->ml_flags & ML_TWOSIDED)
					|| lt->specType >= R3D_40
					|| lt->frontsector == lt->backsector)
				{
					continue;
				}

				linetest[lt - lines] = 1;
			}
		}
	}

	// first mark all segs referencing ld
	for (i = 0, sg = segs; i < numsegs; i++, sg++)
	{
		if (sg->linedef == ld)
			segarray[i] = 2;
		else if (sg->linedef
			&& sg->backsector
			&& linetest[sg->linedef - lines] == 1
			&& !checkProjectedXpoint(sg->linedef))
		{
			segarray[i] = 4;
			count++;
		}
	}

	if (!count)
	{
		M_Free(segarray);
		M_Free(linetest);
		return NULL;
	}

	for (i = 0, sg = segs; i < numsegs; i++, sg++)
	{
		if (segarray[i] == 2)
		{
			seg_t *csg;
			int k;
			for (k = 0, csg = segs; k < numsegs; k++, csg++)
			{
				if (segarray[k] == 4)
				{
					__int64 dist = 
						(getDist(sg->v1, csg->v1) +
						 getDist(sg->v1, csg->v2) +
						 getDist(sg->v2, csg->v1) +
						 getDist(sg->v2, csg->v2));

					if (dist >= (__int64)(2*FRACUNIT) && dist < maxdist)
					{
						maxdist = dist;
						nearestseg = csg;
					}
				}
			}
		}
	}

	M_Free(segarray);
	M_Free(linetest);

	return nearestseg;
}

static RDBOOL it_SegFacingDir(float v1[2], float v2[2], float midx, float midy)
{
	return ((v1[VY] - v2[VY])*(v1[VX] - midx) + (v2[VX] - v1[VX])*(v1[VY] - midy)) > 0;
}

static int getSecFromNearestLine(line_t *li)
{
	seg_t	*sg = subfn_FindNearestSeg(li);
	int		secn = -1;
	float	midx, midy;

	if (!sg) return -1;

	midx = (FIX2FLT(li->v1->x) + FIX2FLT(li->v2->x)) / 2;
	midy = (FIX2FLT(li->v1->y) + FIX2FLT(li->v2->y)) / 2;

	secn = it_SegFacingDir(sg->fv1, sg->fv2, midx, midy) ?
		sg->frontsector - sectors : sg->backsector ?	// Insurance only
		sg->backsector  - sectors : -1;

	if (secn >= 0
		&& !(li->frontsector->specFlags & SF_MOVSECTOR)
		&& sectors[secn].floorheight >= li->frontsector->floorheight)
	{
		// There is probably a line that would be crossed if projecting
		// a line from the midpoint to the line giving this result.
		// For the sake of speed we'll see if the line is connected and,
		// if so, use the other sector.
		if (li->v1 == sg->linedef->v1
			|| li->v1 == sg->linedef->v2
			|| li->v2 == sg->linedef->v1
			|| li->v2 == sg->linedef->v2)
		{
			if (secn == sg->frontsector - sectors)
				secn = sg->backsector - sectors;
			else
				secn = sg->frontsector - sectors;
			// Did this work?
			if (sectors[secn].floorheight >= li->frontsector->floorheight)
				secn = -1;	// No
		}
		else
			secn = -1;
	}
	return secn;
}

static int CheckSsecForContainingSec(sector_t *sec, byte *markline)
{
	int i;

	CLR_hitline();

	conres = -1;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (!markline || markline[ld - lines] == 2)
		{
			ref = ld;

			if ((conres = getSecFromNearestLine(ld)) >= 0)
			{
				break;
			}
		}
	}

	return conres;
}

static RDBOOL subfn_ChkRingSectors(int secn, byte *markline, char *msg)
{
	int		set_count = 0, ring_count = 0;
	int		valid, secnum = -1;
	int		i, k;

	sector_t *sec = &sectors[secn];

	for (i = 0; i < sec->linecount; i++)
	{
		line_t	*ld = sec->lines[i];
		int count = 0;

		if (markline[ld - lines] == 1
			&& LS_CheckRingInSec(markline, ld, sec))
		{
			ring_count++;

			if (ld->frontsector->heightsec >= 0)
			{
				if (ld->frontsector != ld->backsector)
				{
					secnum = ld->backsector - sectors;
				}
			}
			else
			{
				secnum = CheckSsecForContainingSec(sec, markline);
			}
			
			valid  = (secnum >= 0 && secnum != secn);

			if (valid) set_count++;

			for (k = 0; k < sec->linecount; k++)
			{
				int lnum = sec->lines[k] - lines;

				if (markline[lnum] == 2)
				{
					markline[lnum] = 0; // It's job is done
					if (valid)
					{
						Setspec(sec->lines[k], secnum, 38001);
						Settype(sec->lines[k], R3D_3D, 38001);
						count++;	// for debug
					}
				}
			}
		}
	}
	return set_count && set_count == ring_count;
}

// GLOBAL FUNCTIONS ================================================

static RDBOOL subfn_VtxConnectedToTexLine(line_t *ld, vertex_t *vx)
{
	int		xl, xh, yl, yh, x, y;

	validcount++;

	ld->validcount = validcount;

	M_getBlockOffsets(ld->bbox, 0, &xl, &xh, &yl, &yh);

	for (x = xl ; x <= xh ; x++)
	{
		for (y = yl ; y <= yh ; y++)
		{
			const long	*list;
			int			offset;

			if (x < 0 || y < 0 || x >= bmapwidth || y >= bmapheight)
				continue;

			offset = y*bmapwidth + x;

			offset = *(blockmap + offset);
			list = blockmaplump + offset;

			// skip starting delimiter
			while( *(++list) >= 0)
			{
				line_t *lt = &lines[*list];

				if (lt->validcount == validcount)
					continue;		// line has already been checked

				lt->validcount = validcount;

				if (lt->frontsector == lt->backsector)
					continue;

				if (origTEX[*list]
					&& (lt->v1 == vx || lt->v2 == vx))
				{
					return true;
				}
			}
		}
	}
	return false;
}

static RDBOOL VerifyOpenStatic(sector_t *sec)
{
	int i;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (vtxarray[ld->v1 - vertexes] > 2
			&& !subfn_VtxConnectedToTexLine(ld, ld->v1))
		{
			return false;
		}
		if (vtxarray[ld->v2 - vertexes] > 2
			&& !subfn_VtxConnectedToTexLine(ld, ld->v2))
		{
			return false;
		}
	}
	return true;
}

// v220-3 replacement routine
static int isConnected(sector_t *sec)
{
	vertexowner_t *own;
	int		i, j, k, idx;
	int		secnum = XS_GetBoundarySecnum(sec);

	if (secnum < 0) return -1;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t	*ld = sec->lines[i];
		RDBOOL	valid = false;

		if (   ld->frontsector != ld->backsector
			|| ld->frontsector - sectors == secnum)
		{
			return -1;
		}

		for (j = 0; j < 2; j++)
		{
			idx = j ? (ld->v2 - vertexes) : (ld->v1 - vertexes);
			own = vertexowners + idx;

			if (own->num == 1) continue;

			for (k = 0; k < own->num; k++)
			{
				if (own->list[k] == secnum)
				{
					valid = true;
					break;
				}
			}

			if (!valid) return -1;
		}
	}

	return secnum;
}

void LS_CheckOpenStaticContained()
{
	sector_t *sec;
	int		 i, k, m, secnum;
	int		 mask = NO_LINECROSSING|NO_SINGLEVTX1|NO_COINLINE;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i]
			&& sec->heightsec < 0
			&& sec->linecount > 1
			&& !(sec->specFlags & (SF_MOVSECTOR|SF_SECCLOSED))
			&& !(sec->specFlags & SF_SIDETEXTRUE)
			&& !SecCtrl(i)
			&& LS_SameSecLineCondition(sec, mask)
			&& VerifyOpenStatic(sec))
		{
			// Because of the way subsectors are created there may
			// be those with only one line.  If so, and this has meant
			// no containing subsec could be referenced, then see if an
			// unset line is connected to another line that has been set.

			if ( (secnum = isConnected(sec)) >= 0)
			{
				for (k = 0; k < sec->linecount; k++)
				{
					line_t	 *ld = sec->lines[k];
					ssline_t *sl = &sslines[ld - lines];

					lnsec[ld - lines] = -1;

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->floorsec < 0 && ss->ceilsec < 0)
							SetOverride(ss, secnum, 38011);
					}
				}
			}

			if (secnum >= 0) Setsecdone3(sec, 38012);
		}
	}
}

void LS_CheckSupportLineRings()
{
	sector_t	*sec;
	int			i;
	byte		*markline = (byte*)M_Calloc(numlines, 274);

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i]
			&& sec->specFlags & SF_MOVFLOOR)
		{
			int k, hsec, count = 0;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (!(ld->ml_flags & ML_TWOSIDED)
					|| !Vtxs2(ld)
					|| !(ld->frontsector->specFlags & SF_MOVFLOOR))
					continue;

				if (ld->frontsector == ld->backsector)
				{
					if (!(origTEX[i] & 0x22)
						&& subfn_FindLineCrossing(ld, LC_DIFFSECLINE|LC_VTXISONLINE) == NULL)
					{
						markline[i] = 1;
						count++;
					}
				}
				else if (!origTEX[i]
					&& Vtfl3(ld)
					&& (hsec = ld->frontsector->heightsec) >= 0
					&& ld->backsector->heightsec == hsec
					&& !(ld->backsector->specFlags & SF_MOVSECTOR))
				{
					markline[i] = 1;
					count++;
				}
			}

			if (count > 2)
			{
				if (subfn_ChkRingSectors(i, markline, "SUPPORT"))
					Setsecdone3(sec, 38003);
			}
		}
	}
	M_Free(markline);
}
