//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2011 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_maperror.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// DESCRIPTION: checks for map design errors
//
//-----------------------------------------------------------------------------

#include  <math.h>

#include "m_misc.h"
#include "r3d_MAINspec.h"
#include "r3d_compat.h"

// CODE -----------------------------------------------------------------------

static void subfn_InhibitSSECocclude(sector_t *sec)
{
	subsector_t *ss;
	int			i, k;

	CLR_hitline();

	for (i = 0; i < sec->linecount; i++)
	{
		hitline[sec->lines[i] - lines] = 1;
	}

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		seg_t *sg = &segs[ss->firstline];
		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (sg->linedef
				&& hitline[sg->linedef - lines])
			{
				ss->ssflags |= SSF_NOOCCLUDE;
				break;
			}
		}
	}
}

static void subfn_CheckFloorCeilHeightError(sector_t *sec)
{
	int i, ch = DDMININT, hs = -2, tempch;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (!(ld->ml_flags & ML_TWOSIDED) ||
			ld->frontsector == ld->backsector ||
			!(origTEX[ld - lines] & 0x11) ||
			origTEX[ld - lines] & 0x66)
		{
			return;
		}
		if (ld->backsector != sec)
		{
			if (ld->backsector->ceilingpic != skyflatnum)
				return;

			if (hs == -2)
				hs = ld->backsector->heightsec;
			else if (hs != ld->backsector->heightsec)
				return;

			tempch = ld->backsector->ceilingheight;
		}
		if (ld->frontsector != sec)
		{
			if (ld->frontsector->ceilingpic != skyflatnum)
				return;

			if (hs == -2)
				hs = ld->frontsector->heightsec;
			else if (hs != ld->frontsector->heightsec)
				return;

			tempch = ld->frontsector->ceilingheight;
		}

		if (ch == DDMININT)
			ch = tempch;
		else if (ch != tempch)
			return;
	}

	if (ch > sec->ceilingheight)
	{
		if (sec->specFlags & SF_MOVSECTOR)
		{
			Setr3dch_byH(sec, ch, 33001);
			subfn_InhibitSSECocclude(sec);
			return;
		}

		// v220-10 use hs to catch map hack
		if (hs >= 0)
			sec->ceilingpic = sectors[hs].floorpic;
		else
			Setr3dchangech(sec, ch, 33001);

		SetNOocclude(sec, 33001);
	}
}

void LS_CheckFloorCeilHeightError()
{
	sector_t *sec;
	int		 i;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if ( !(sec->specFlags & SF_MOVSECTOR)
			&& sec->ceilingpic == skyflatnum
			&& sec->floorpic != skyflatnum
			&& sec->ceilingheight == sec->floorheight
			&& !(sec->ceilingheight & FRACUNIT))
		{
			subfn_CheckFloorCeilHeightError(sec);
		}
	}
}

//=============================================================================

// Catch map error
void LS_CheckBacksector()
{
	line_t *ld;
	int i;
	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (!origTEX[ld - lines] &&
			ld->specType < R3D_40 &&
			ld->mtflags &&
			!ChkLptr(ld) &&
			Vtxs2(ld) && Vtfl3(ld) &&
			subfn_FindLineConnected(ld, 9) != NULL)	// Check if connected to non compliant line
		{
			Settype(ld, R3D_18, 33002);
			pr_OFinsert("  MAP SECTOR ERROR: line %i cannot be determined\n", ld - lines);
			ma_inf.maperror++;
		}
	}
}

//=============================================================================

// LS_FixSubsecErrorsWithShortSegs
//
// Find and fix invalid subsectors (if possible) caused by a line
// crossing another but only by 1 FRACUNIT (or less)

// Checks line length and also if line is vertical or horizontal.
static RDBOOL ValidLength(line_t *ld)
{
	return M_AccurateLength(ld) > 2*FRACUNIT;
}

static RDBOOL subfn_CheckIntersect(subsector_t *ss, seg_t *seg)
{
	seg_t	*sc = &segs[ss->firstline];
	line_t	*lc = seg->linedef;
	int		i;

	vertex_t *v = NULL;

	// The seg's vertex should correspond to one of its line vertexes.
	if (seg->linedef->v1->x == seg->v1->x &&
		seg->linedef->v1->y == seg->v1->y)
		v = seg->linedef->v1;
	else if (seg->linedef->v1->x == seg->v2->x &&
		seg->linedef->v1->y == seg->v2->y)
		v = seg->linedef->v1;
	else if (seg->linedef->v2->x == seg->v1->x &&
		seg->linedef->v2->y == seg->v1->y)
		v = seg->linedef->v2;
	else if (seg->linedef->v2->x == seg->v2->x &&
		seg->linedef->v2->y == seg->v2->y)
		v = seg->linedef->v2;

	if (!v) return false;

	for (i = 0; i < ss->linecount; i++, sc++)
	{
		if (   sc->linedef && sc->backsector
			&& sc->frontsector != sc->backsector
			&& ValidLength(sc->linedef)
			&& subfn_CheckLineIntersect(lc, sc->linedef, LC_TESTLINE|LC_NOVTXONLINE))
		{
			if (ss->sector != sc->frontsector)
			{
				if (verbose)
					pr_OFinsert("  fault caused by missplaced vertex %5i corrected\n", v - vertexes);
				ChangeSSECsec(ss, sc->frontsector, 33003);
				SetSSF_BASESSECFIX(ss, 33003);
				return true;
			}
		}
	}
	return false;
}

// Checks for lines crossing caused by a vertex misplacement of <= FRACUNIT.
void LS_FixSubsecErrorsWithShortSegs()
{
	int i, count = 0;

	for (i = 0; i < numsubsectors; i++)
	{
		subsector_t *ss = subsectors + i;
		seg_t		*seg = &segs[ss->firstline];
		int			k;

		for (k = 0; k < ss->linecount; k++, seg++)
		{
			if (  !seg->backsector
				&& seg->linedef
				&& seg->length < 1.0f)
			{
				seg->sgflags |= R3DSEGF_SHORTSEG;

				if (   LSS(seg->linedef)
					&& origTEX[seg->linedef - lines] != 0x02
					&& origTEX[seg->linedef - lines] != 0x20
					&& ValidLength(seg->linedef)
					&& subfn_CheckIntersect(ss, seg))
				{
					count++; break;
				}
			}
		}
	}

	if (count)
	{
		if (!verbose)
			pr_OFinsert("MAP ERROR%s FIXED: %i faults caused by missplaced vertexes.\n", plural(count), count);

		ma_inf.maperror += count;
	}
}

// v220-10 check if a top texture should be added
// v220-8  check for missing midtex with 1's lines
void LS_AddTexDefinitions()
{
	line_t	*ld, *lc;
	int		i, tx;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->spflags & SPF_CONTROLLINE) continue;

		if (ld->ml_flags & ML_TWOSIDED)
		{
			if (   origTEX[i] == 0x01
				&& ld->frontsector->specFlags & SF_MOVCEIL
				&& ld->backsector->specFlags & SF_MOVCEIL
				&& !(ld->frontsector->ceilingpic == skyflatnum
				&&   ld->backsector->ceilingpic == skyflatnum)
				&& ld->backsector->ceilingheight > ld->frontsector->ceilingheight)
			{
				tx = sides[ld->sidenum[0]].bottomtexture;
				sides[ld->sidenum[0]].toptexture = tx;
				origTEX[i] |= 0x04;
			}
		}
		else
		if (   !(origTEX[i] & 0x02)
			&& !(!ld->sidenum[0] && ma_inf.sd_err_check)	// v220-35 In case side 0 has had to be shared
			&& ld->frontsector->ceilingheight > ld->frontsector->floorheight
			&& !(ld->frontsector->floorpic == skyflatnum && ld->frontsector->ceilingpic == skyflatnum)
			&& !(ld->spflags & SPF_CONTROLLINE)	// Doesn't matter with these
			&& (lc = subfn_FindLineConnected(&lines[i], 34)))
		{
			short txn = sides[lc->sidenum[0]].midtexture;

			sides[ld->sidenum[0]].midtexture = txn;
			origTEX[i] |= 0x02;
			ld->scpFlags |= SCPF_REPERROR;
			ma_inf.midtexerror++;
		}
	}
}

