//-----------------------------------------------------------------------------
//
// Copyright (C) 1999 Randy Heit
// Copyright (C) 2006 - 2012 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_mapinfo.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Limited support for ZDOOM MAPINFO lump
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
//  Integrates support for all methods into a mapinfo array
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
//  Code dealing with R3Dinfine and R3Dbypass changed
// Revision 2.2.0 build 24 (v220-24) 15/Feb/2014 gmj
// Revision 2.2.0 build 26 (v220-26) 05/Jul/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
//
// Based on: Doomsday v1.7.8
//           ZDOOM v1.22 December 12, 1999 (dis. marked v1.23beta33)
//
// DESCRIPTION: adds support for mapinfo's using ZDoom parameters
//
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <string.h>
#include <memory.h>

#include "r3d_compat.h"

#include "doom/d_englsh.h"
#include "con_console.h"
#include "dd_wad.h"
#include "gl_main.h"
#include "gl_tex.h"
#include "dd_defns.h"
#include "m_misc.h"
#include "p_data.h"
#include "r_world.h"
#include "s_mus.h"
#include "rend_sky.h"

#include "r3d_mapinfo.h"
#include "r3d_edit.h"

#include "r3d_OGLfunc.h"

// MACROS ------------------------------------------------------------------

#define MAXSTRLEN	32	// Could be 64 see ded_mapinfo_t;
#define MAXEPISODE  10

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void	ReplaceInString(char *str, char *occurance, char *replacewith);
char	*Def_ModifyTextDef(int texn, char *str);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern char	*R3Dinfine;
extern char	*R3Dbypass;
extern RDBOOL EMAPINFO_script;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

RDBOOL	outsidefog, underwaterfog;
int		fimusic_num;
int		fipic_num;
int		episodeinf[9][3];

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static int	fiflat_num = -1;
static int	episodeptr[MAXEPISODE];
static int	numepisodes;
static char	automaplevel[64];
static char nextmapname[64];

// CODE --------------------------------------------------------------------

// Check for a dehacked flat
int MI_GetFlat(char *levelid)
{
	int n = fiflat_num;

	// MAPINF0 takes precedence so MI_GetBriefings must be called first
	if (levelid && n < 0)
	{
		int i;

		for (i = 0; i < defs.count.mapinfo.num; i++)
		{
			if ( !_stricmp(defs.mapinfo[i].id, levelid) )
			{
				if (defs.mapinfo[i].bexbgflat[0])
					n = W_CheckNumForName(defs.mapinfo[i].bexbgflat);

				break;
			}
		}
	}

	fiflat_num = -1;

	return n;
}

// This will only return a value > 0 if MAPINFO is present
int MI_NextMap(RDBOOL secret, int *gameepisode)
{
	char *t = "";

	if (secret)
	{
		if (mapinfo->secretnext[0]) t = mapinfo->secretnext;
	}
	else if (mapinfo->nextmap[0])
		t = mapinfo->nextmap;

	if (t[0] && W_CheckNumForName(t) >= 0)
	{
		if (commercial_game)
			return (t[3] - '0')*10 + (t[4] - '0');

		*gameepisode = t[1] - '0';
		return t[3] - '0';
	}

	return 0;
}

char *NXTEXT[6];

// v220-31 add support for "exittextislump"
static void getClusterVars(char *defn)
{
	char	*nextstr;
	RDBOOL	islump = false;
	int		n, type;

	while( (nextstr = SR_NextSTR(false)) )
	{
		if (!_stricmp(nextstr, "clusterdef") || !_stricmp(nextstr, "map"))
			break;

		if (!_stricmp(nextstr, "exittextislump")) islump = true;

		if (!_stricmp(nextstr, "flat"))
		{
			nextstr = SR_NextSTR(true);
			fiflat_num = W_CheckNumForName(nextstr);
		}

		if (!_stricmp(nextstr, "pic"))
		{
			nextstr = SR_NextSTR(true);
			n = W_CheckNumForName(nextstr);
			type = DoomLumpIdentify(n, NULL, NULL, DLI_GRAPHIC);
			if (type == LMP_PATCH || type == LMP_PNG) fipic_num = n;
		}

		if (!_stricmp(nextstr, "music"))
		{
			nextstr = SR_NextSTR(true);
			fimusic_num = W_CheckNumForName(nextstr);
		}

		if (!_stricmp(nextstr, defn))
		{
			char *ptr = NULL;
			int  n;

			nextstr = SR_NextSTR(true);

			// v220-4 support mapinfo 'lookup' keyword
			if (!_stricmp(nextstr, "lookup"))
			{
				nextstr = SR_NextSTR(true);

				if (   strlen(nextstr) == 6
					&& nextstr[1] >= '1'
					&& nextstr[1] <= '6'
					&& !_stricmp(&nextstr[2], "TEXT"))
				{
					char *c[6] = { "C1TEXT", "C2TEXT", "C3TEXT", "C4TEXT", "C5TEXT", "C6TEXT" };
					char *e[6] = { "E1TEXT", "E2TEXT", "E3TEXT", "E4TEXT", "", "" };
					char *p[6] = { "P1TEXT", "P2TEXT", "P3TEXT", "P4TEXT", "P5TEXT", "P6TEXT" };
					char *t[6] = { "T1TEXT", "T2TEXT", "T3TEXT", "T4TEXT", "T5TEXT", "T6TEXT" };
					int  cv[] = { TXT_C1TEXT, TXT_C2TEXT, TXT_C3TEXT, TXT_C4TEXT, TXT_C5TEXT, TXT_C6TEXT };
					int  ev[] = { TXT_E1TEXT, TXT_E2TEXT, TXT_E3TEXT, TXT_E4TEXT, 0, 0 };
					int  pv[] = { TXT_P1TEXT, TXT_P2TEXT, TXT_P3TEXT, TXT_P4TEXT, TXT_P5TEXT, TXT_P6TEXT };
					int  tv[] = { TXT_T1TEXT, TXT_T2TEXT, TXT_T3TEXT, TXT_T4TEXT, TXT_T5TEXT, TXT_T6TEXT };

					char *p1 = NULL;
					int  v;
					
					n = nextstr[1] - '1';

					//--------------------------------------

					if (*nextstr == 'C' || *nextstr == 'c') p1 = c[n], v = cv[n];	// doom2
					else
					if (*nextstr == 'E' || *nextstr == 'e') p1 = e[n], v = ev[n];	// udoom
					else
					if (*nextstr == 'P' || *nextstr == 'p') p1 = p[n], v = pv[n];	// plutonia
					else
					if (*nextstr == 'T' || *nextstr == 't') p1 = t[n], v = tv[n];	// tnt

					if (p1 && !_stricmp(nextstr, p1)) ptr = (char*)GET_TXT(v);
				}
			}
			else
			{
				ptr = nextstr;

				// v220-24 check if nextstr is a $NxTEST type and, if so,
				//         whether this has been set from a LANGUAGE lump
				if (   strlen(ptr) == 7
					&& ptr[0] == '$'
					&& !_stricmp(&ptr[3], "TEXT")
					&& (n = ptr[2] - '0' - 1) >= 0
					&& n < 6
					&& NXTEXT[n])	// If set assume 2nd char must be C, E, P or T
				{
					ptr = NXTEXT[n];	// redirect ptr
				}
			}

			if (ptr)
			{
				int	lump = islump ? W_CheckPWADNumForName(ptr) : -1;

				if (lump >= 0)
				{
					lumpinfo_t	*linf = &lumpinfo[lump];
					char		*lmptxt = (char*)calloc(linf->size + 1, 1);

					lseek(linf->handle, linf->position, SEEK_SET);
					read(linf->handle, lmptxt, linf->size);
					R3Dinfine = Def_ModifyTextDef(TXT_R3D_INFINE_TEXT, lmptxt);
					free(lmptxt);
				}
				else
				{
					R3Dinfine = Def_ModifyTextDef(TXT_R3D_INFINE_TEXT, ptr);
				}
			}
		}
	}
}

// Set up any briefings
int MI_GetBriefings(RDBOOL after)
{
	int		cluster_num = -1;

	fimusic_num = fiflat_num = fipic_num = -1;
	
	R3Dinfine[0] = 0;

	if (   !(mapinfo->R3Dflags & R3Dhasmapinfo)
		|| !SR_LoadPWADLumpName("MAPINFO"))
	{
		return -1;
	}

	if (  (cluster_num = mapinfo->cluster) >= 0
		&& cluster_num < 100)
	{
		char	*defn = after ? "exittext" : "entertext";
		char	*nextstr;

		R3Dinfine = Def_ModifyTextDef(TXT_R3D_INFINE_TEXT, R3Dbypass);

		// Advance script to the relevant clusterdef point
		while( (nextstr = SR_NextSTR(false)) )
		{
			if (   !_stricmp(nextstr, "clusterdef")
				&& SR_NextNum() == cluster_num)
			{
				getClusterVars(defn);
				break;
			}
		}
	}
	else
	{
		cluster_num = 0;
	}

	SR_EndScript();

	return cluster_num;
}

// This only checks for HUSTR_EPI or NXTEXT where N = E and X is '1' to '6' (where gamemode == retail)
// or HUSTR_ or  NXTEXT where N = C,P or T and X is '1' to '6' (where gamemode == commercial)
static int checkLANGUAGE()
{
	char	*nextstr;
	int		num, lumpnum, lastlump = 0;
	uint	len;

	if (gamemode == shareware) return numepisodes;

	while ((lumpnum = W_FindScriptLump("LANGUAGE", &lastlump)) != -1)
	{
		EMAPINFO_script = true;

		SR_LoadLump(lumpnum, "LANGUAGE");

		// v220-24 Note ' ' splits text inside []
		// For compatibility with other scripts (e.g. R3D_WADanalyse)
		// this is checked in two stages instead of treating [] as quotes
		if (!(nextstr = SR_NextSTR(false)) ) goto language_end;
		if (!_stricmp(nextstr, "[enu ")) goto language_end;
		if (!(nextstr = SR_NextSTR(false)) ) goto language_end;
		if (!_stricmp(nextstr, "default]")) goto language_end;

		while( (nextstr = SR_NextSTR(false)) )
		{
			char copystr[16];
			char id[8];

			if ( (len = strlen(nextstr)) > 10 || len < 6) continue;

			strcpy(copystr, nextstr);
			_strupr(copystr);

			if ( !(nextstr = SR_NextSTR(false)) ) goto language_end;

			if (   len == 6
				&& !strcmp(copystr + 2, "TEXT")
				&& (num = copystr[1] - '0' - 1) >= 0
				&& num < 5
				&& ((copystr[0] == 'C' && gamemode == commercial)
				||  (copystr[0] == 'E' && gamemode != commercial)
				||  (copystr[0] == 'P' && gamemode == commercial)
				||  (copystr[0] == 'T' && gamemode == commercial)) )
			{
				// if NXTEXT gets set then this will override anything set in MAPINFO
				len = strlen(nextstr);
				NXTEXT[num] = (char*)W_Malloc(len + 1, STATICID, 19);
				strcpy(NXTEXT[num], nextstr);
			}
			else if (gamemode == commercial)
			{
				if (   len > 6
					&& len < 9
					&& !strncmp(copystr, "HUSTR_", 6)
					&& (num = atoi(copystr + 6)) > 0
					&& num < 100)
				{
					sprintf(id, "MAP%02i", num);

					for (int i = 0; i < defs.count.mapinfo.num; i++)
					{
						if (!_stricmp(defs.mapinfo[i].id, id))
						{
							// If set in DEHACKED this will overwrite it
							strncpy(defs.mapinfo[i].mapname, nextstr, DED_STRINGID_LEN);
							defs.mapinfo[i].mapname[DED_STRINGID_LEN] = 0;
							break;
						}
					}
				}
			}
			else
			{
				if (   len == 10
					&& !strncmp(copystr, "HUSTR_EPI", 9)
					&& (num = atoi(copystr + 9)) > 0
					&& num < 10)
				{
					if (episodeptr[num] == num)
						Def_ModifyTextDef(TXT_EPISODE1 + num - 1,  nextstr);
					else
						SR_SCPWarn("No pwad map exists with episode %i\n", num);
				}
				else
				if (   len == 10
					&& !strncmp(copystr, "HUSTR_E", 7)
					&& copystr[8] == 'M')
				{
					strncpy (id, &copystr[6], 4);
					id[4] = 0;

					for (int i = 0; i < defs.count.mapinfo.num; i++)
					{
						if (!_stricmp(defs.mapinfo[i].id, id))
						{
							// If set in DEHACKED this will overwrite it
							strncpy(defs.mapinfo[i].mapname, nextstr, DED_STRINGID_LEN);
							defs.mapinfo[i].mapname[DED_STRINGID_LEN] = 0;
							break;
						}
					}
				}
			}
			
language_end:;

			EMAPINFO_script = false;
			SR_EndScript();
		}
	}
	return numepisodes;
}

int MI_SetEpisodes()
{
	char	*nextstr;
	int		i;

	memset(episodeptr, 0, sizeof(episodeptr));
	numepisodes = 0;

	if (haspwads && gamemode != commercial)
	{
		int n;

		for (i = 10; i < 100; i++)
		{
			n = i/10;

			if (!episodeptr[n] && haslevelExMx[i])
			{
				episodeptr[n] = n;
				if (n > numepisodes) numepisodes = n;
			}
		}
	}

	for (i = 0; i < 9; i++) episodeinf[i][1] = -1;

	checkLANGUAGE();
	if (!SR_LoadPWADLumpName("MAPINFO")) return numepisodes;

	while( (nextstr = SR_NextSTR(false)) )
	{
		if (!_stricmp(nextstr, "episode"))
		{
			nextstr = SR_NextSTR(true);

			// v220-31 extended for doom2 episode support
			if (strlen(nextstr) == 5 && !strnicmp(nextstr, "map", 3))
			{
				int num = atoi(nextstr + 3);

				if (num < 0 || num > 99)
					Con_Error("Invalid map %s\n", nextstr);
				else if (!haslevelMapxx[num])	// v220-26 don't abort
					SR_SCPWarn("No pwad map exists with map %s\n", nextstr);
				else
				{
					if (numepisodes > 9) SR_SCPErr("MAPINFO has too many episodes\n");

					nextstr = SR_NextSTR(true);

					if (!_stricmp(nextstr, "{"))
					{
						while (nextstr = SR_NextSTR(false))
						{
							if (!nextstr) break;

							if (!_stricmp(nextstr, "picname"))
							{
								if ( !(nextstr = SR_NextSTR(false)) ) break;

								episodeinf[numepisodes][1] = W_CheckNumForName(nextstr);

								if (episodeinf[numepisodes] < 0)
								{
									SR_SCPWarn("Lump %s not found\n", nextstr); break;
								}
							}

							if (!_stricmp(nextstr, "key"))
							{
								if ( !(nextstr = SR_NextSTR(false)) ) break;

								// detected but currently unused
								episodeinf[numepisodes][2] = atoi(nextstr);
							}

							if (_stricmp(nextstr, "}")) break;
						}

						episodeinf[numepisodes++][0] = num;
					}
					else
					{
						SR_KeepString();
					}
				}
			}
			else if (strlen(nextstr) == 4
				&& !strnicmp(nextstr, "E", 1)
				&& !strnicmp(nextstr + 2, "M", 1))
			{
				int ep = nextstr[1] - '0';

				// NOTE: E0Mx is possible but ep 0 is not currently supported
				if (ep > 0 && ep < 10 && episodeptr[ep] == ep)
				{
					if (!_stricmp(SR_NextSTR(true), "name"))
						Def_ModifyTextDef(TXT_EPISODE1 + ep - 1, SR_NextSTR(true));
				}
				else
				{
					SR_SCPWarn("No pwad map exists with episode %s\n", nextstr);
				}
			}
			else
				SR_SCPWarn("Could not set episode with string %s\n", nextstr);
		}
	}

	SR_EndScript();

	return numepisodes;
}

// Called at start of game
void MI_VerifyMapInfo(char *level_id)
{
	if (!(mapinfo = Def_GetMapInfo(level_id)))
		Con_Error("unset mapinfo %s", level_id);

	GL_UseFog(false);
	underwaterfog = outsidefog = false;

	// Reset sky parms
	skyAngleGlobal = mapinfo->sky_layers->skyAngleIncrement;
	skyAngle = 0;
}

// v220-3 revised
void MI_ActionMapInfo(char *level_id)
{
	int flags = R3D_edit ? 0 : Def_EvalFlags(mapinfo->flags);

	underwaterfog = outsidefog = false;

	if (!flags) return;

	if (flags & (MIF_FOGOUTSIDE|MIF_FOG))
	{
		if (mapinfo->fog_start >= mapinfo->fog_end)
			return;

		GL_FogvMode(GL_LINEAR);
		GL_FogvStart(mapinfo->fog_start);
		GL_FogvEnd(mapinfo->fog_end);

		if (flags & MIF_FOGOUTSIDE) outsidefog = true;
	}
	else
	{
		if (flags & MIF_FOGEXP)
			GL_FogvMode(GL_EXP);
		else	// default to exp2
			GL_FogvMode(GL_EXP2);

		GL_FogvDensity(mapinfo->fog_density);

		if (flags & MIF_FOGUNDERWATER) underwaterfog = true;
	}

	GL_FogvColor(mapinfo->fog_color);
	GL_UseFog(true);
}

int R3D_EpisodeNum(int num)
{
	if (num <= numepisodes) return episodeptr[num];

	return 0;
}

