//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_mus2mid.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 14 (v220-13) 02/Apr/2012 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//
// DESCRIPTION: Mus to Midi convertor
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <windows.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>
#include <direct.h>
#include <process.h>

#include "dd_types.h"
#include "dd_wad.h"
#include "p_data.h"
#include "m_args.h"
#include "con_console.h"
#include "sys_timer.h"
#include "r3d_compat.h"
#include "r3d_mus2mid.h"

// DEFS  -------------------------------------------------------------------

#define MTrk_BUFFER_SIZE		4096
#define MTrk_BUFFER_INCREMENT	4096

#define MUS_EV_NOTEOFF 		((byte)0x00)
#define MUS_EV_NOTEON		((byte)0x10)
#define MUS_EV_PITCHBEND	((byte)0x20)
#define MUS_EV_SYSEVENT		((byte)0x30)
#define MUS_EV_CTRLCHANGE	((byte)0x40)
#define MUS_EV_SCORE_END	((byte)0x60)

enum // MUS controllers.
{
	MUS_CTRL_INSTRUMENT,
	MUS_CTRL_BANK,
	MUS_CTRL_MODULATION,
	MUS_CTRL_VOLUME,
	MUS_CTRL_PAN,
	MUS_CTRL_EXPRESSION,
	MUS_CTRL_REVERB,
	MUS_CTRL_CHORUS,
	MUS_CTRL_SUSTAIN_PEDAL,
	MUS_CTRL_SOFT_PEDAL,

	// The valueless controllers.
	MUS_CTRL_SOUNDS_OFF,
	MUS_CTRL_NOTES_OFF,
	MUS_CTRL_MONO,
	MUS_CTRL_POLY,
	MUS_CTRL_RESET_ALL,
	NUM_MUS_CTRLS
};

// TYPES -------------------------------------------------------------------


#pragma pack(2)
typedef struct{
	int		ID;					// identifier "MUS" 0x1A
    ushort	scoreLen;
    ushort	scoreStart;
	ushort	channels;			// number of primary channels
	ushort	secondaryChannels;	// number of secondary channels
	ushort	instrCnt;
	ushort	dummy;
	// The events list begins here.
} MUSHeader_t;
#pragma pack()

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// EXTERNAL DATA DEFINITIONS -----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

float	mus_compress = 1.f;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static char ctrlMus2Midi[NUM_MUS_CTRLS] =
{
	0,		// Not used.
	0,		// Bank select.
	1,		// Modulation.
	7,		// Volume.
	10,		// Pan.
	11,		// Expression.
	91,		// Reverb.
	93,		// Chorus.
	64,		// Sustain pedal.
	67,		// Soft pedal.

// The valueless controllers:
	120,	// All sounds off.
	123,	// All notes off.
	126,	// Mono.
	127,	// Poly.
	121		// Reset all controllers.
};

static MUSHeader_t musHdr;

static byte		*lumpCur, *startPtr;

typedef struct {
	byte	chanVols;		// Last volume for each channel.
	byte	oldmidiStatus;
	byte	*chbuffer;
	byte	channelMap;
	byte	channelUsed;
	byte	timecount;
	DWORD	deltaTime;		// In ticks.
	UINT	chptr;
	UINT	chlength;		// chsize has been nabbed by MS
} chaninf_t;

static chaninf_t *chinf = NULL;

// Some of the following are not strictly needed but help if loading the file into a midi editor
static byte MIDIheader[] =
{
	'M','T','h','d', 0, 0, 0, 6,
	0, 0, 0, 0,															// tracks (to be set)
	0, 70,																// Default midi tics
	'M','T','r','k',													// track 0 (for following info)
	0, 0, 0, 41,														// length to end of this
	0, 0xFF, 2, 12, 'R','3','D',' ','M','U','S','t','o','M','I','D',	// id
	0, 0xFF, 0x59, 0x02, 0x00, 0x00,									// key signature
	0, 0xFF, 0x51, 0x03, 0x07, 0xA1, 0x20,								// tempo
	0, 0xFF, 0x58, 0x04, 0x04, 0x02, 0x16, 0x04,						// time signature
	0, 0xFF, 0x2F, 0x00													// EOT
};

static UINT		total_time;
static RDBOOL	event_error;

static byte	EOT[4] = {0x00, 0xFF, 0x2F, 0x00};

// CODE --------------------------------------------------------------------

//===========================================================================
// Mus_IsMUSorMIDILump
//	Returns true if the specified lump contains a MUS song.
//===========================================================================
int Mus_IsMUSorMIDILump(int lump)
{
	int type = DoomLumpIdentify(lump, NULL, NULL, DLI_SOUND);

	if (type == LMP_MUS   ) return M2M_MUS;
	if (type == LMP_MIDI  ) return M2M_MIDI;
	if (type == LMP_RMI   ) return M2M_RMI;
	// v220-33 add further check
	if ((type >= LMP_TXT && type <= LMP_PATCH) || type == LMP_SND) return M2M_BADLUMP;

	return M2M_UNKNOWN;
}

static byte getChannel(int mChan)
{
	int i;

	// Map sequentially from 0 upwards.
	// v220-14 This got deleted in v220-13
	if (chinf[mChan].channelMap != 0xFF)
	{
		return chinf[mChan].channelMap;
	}
	
	if (mChan == 15)
	{	// Redirect MUS channel 16 to MIDI channel 10 (percussion).
		chinf[15].channelMap = 9;
		// Could be preset making i != 9 test unnecessary - but for clarity left as is
		chinf[9].channelUsed = true;
		return 9;
	}

	for (i = 0; i < MIDI_MAXNUMCHANS; i++)
	{
		if (i != 9 && chinf[i].channelUsed == false)
		{
			chinf[mChan].channelMap = (byte)i;
			chinf[i].channelUsed = true;
			return i;
		}
	}

	Con_ErrorD("Program fault [4]");
	return 0;
}

static void setVarDelta(chaninf_t *ch)
{
	DWORD val = ch->deltaTime;
	DWORD buffer = val & 0x7F;
	int	  tc = 0;

	while ( (val >>= 7) )
	{
		buffer <<= 8;
		buffer |= ((val & 0x7F) | 0x80);
	}
	
	ch->deltaTime = 0;

	while (true)
	{
		ch->chbuffer[ch->chptr++] = (byte)(buffer & 0xFF);
		if (!(buffer & 0x80)) break;
		buffer >>= 8;
	}

	ch->timecount = tc;
}

static RDBOOL subfn_WriteMidiEvent()
{
	chaninf_t *ch;
	byte	evDesc;
	byte	midiStatus, midiChan, midiParm1, midiParm2;
	int		i, numParms = 2;	//Default: 2 parameters
	int		needed = 1;
	DWORD	dwTd;

	if (lumpCur - startPtr >= musHdr.scoreLen) return false;

	evDesc = *lumpCur++;
	if ( (evDesc & 0x70) == MUS_EV_SCORE_END) return false;
	if (lumpCur - startPtr >= musHdr.scoreLen) return false;

	switch( (evDesc & 0x70) )
	{
	case MUS_EV_CTRLCHANGE: needed++; break;
	case MUS_EV_NOTEON:
		if (*lumpCur & 0x80) needed++;
		break;
	}
		
	if (lumpCur - startPtr + needed >= musHdr.scoreLen) return false;

	midiParm1 = *lumpCur++;
	midiParm2 = 0;
	
	// Use mask to get channel number.
	midiChan = getChannel(evDesc & 0x0F);
	ch = &chinf[midiChan];
	midiStatus = midiChan;

	// Check and adjust, if necessary, buffer allocation
	if (ch->chlength < (ch->chptr + 16))
	{
		ch->chlength += MTrk_BUFFER_INCREMENT;
		ch->chbuffer = (byte*)M_Realloc(ch->chbuffer, ch->chlength, 36);
	}

	// Construct the MIDI event.
	switch( (evDesc & 0x70) )
	{
	case MUS_EV_NOTEOFF:
		midiStatus |= 0x80;
		midiParm2 = 64;	// use a default velocity
		break;
		
	case MUS_EV_NOTEON:
		midiStatus |= 0x90;

		if (midiParm1 & 0x80)	// Check for volume
		{
			byte vol = (*lumpCur++) & 0x7F;

			if (mus_compress > 1.f)
			{
				int newvol = vol;

				newvol += (int)((float)(127 - vol) * mus_compress);

				if (newvol > 127) newvol = 127;

				vol = (byte)newvol;
			}
			else if (mus_compress < 1.f)
				vol = (int)((float)vol * mus_compress);

			ch->chanVols = vol;

			midiParm1 &= 0x7f;
		}

		// Check for instrument select error on percussion channel
		if (midiChan == 9 && (midiParm1 < 35 || midiParm1 > 81))
		{
			midiParm1 = 80;		// Substitute mute triangle
		}

		midiParm2 = ch->chanVols;

		break;				

	case MUS_EV_CTRLCHANGE:
		midiParm2 = *lumpCur++;
		// The instrument control is mapped to another kind of MIDI event.
		// This is identified by 0 followed by the event type
		if(!midiParm1)
		{
			midiStatus |= 0xc0;
			midiParm1 = midiParm2;
			numParms = 1;
		}
		else
		{
			midiStatus |= 0xb0;
			// Use the conversion table.
			midiParm1 = ctrlMus2Midi[midiParm1 & 0xF];
		}
		break;
	
	case MUS_EV_PITCHBEND:
		midiStatus |= 0xe0;
		midiParm2 = (midiParm1 >> 1) & 127;
		midiParm1 = (midiParm1 & 1) << 6;
		break;
		
	case MUS_EV_SYSEVENT:
		midiStatus |= 0xb0;
		midiParm1 = ctrlMus2Midi[midiParm1 & 0xF];
		break;
		
	default:
		event_error = true;
		pr_OFinsert("MUS_SongPlayer: Unknown MUS event %d\n", evDesc);
		return false;
	}

	setVarDelta(ch);

	if (ch->oldmidiStatus != midiStatus)
	{
		ch->oldmidiStatus = midiStatus;
 		ch->chbuffer[ch->chptr++] = midiStatus;
	}

	ch->chbuffer[ch->chptr++] = midiParm1;

	if (numParms == 2)
		ch->chbuffer[ch->chptr++] = midiParm2;

	// Check if this was the last event in a group.
	if (!(evDesc & 0x80)) return true;

	// Read the time delta.
	for (dwTd = 0, i = 0; i < 4; i++)
	{
		byte c = *lumpCur++;
		dwTd = dwTd*128 + (c & 0x7f);
		if(!(c & 0x80)) break;
	}

	// Safety net where MUS_EV_SCORE_END is missing
	if (lumpCur - startPtr > musHdr.scoreLen) return false;

	total_time += dwTd;	// For stats

	// Increment time delta for all channels
	for (i = 0; i < MIDI_MAXNUMCHANS; i++)
	{
		chinf[i].deltaTime += dwTd;
	}

	return true;
}

static RDBOOL BuildMidiFile(int lump, char *fname)
{
	DWORD	szHdr = sizeof(MIDIheader);
	byte	*buffer, *bufptr;
	byte	trackcount = 0;
	UINT	length, i;
	UINT	buffsize = 0;
	int		diff;

	// Count number of tracks set
	// Buffsize is set at least large enough for the chbuffers.
	// timecount[i] is not deducted to give a margin.
	for (i = 0; i < MIDI_MAXNUMCHANS; i++)
	{
		if (chinf[i].chptr)
		{
			trackcount++;
			buffsize += (chinf[i].chptr + 12);	// Max required - some will need slightly less
		}
	}

	if ((diff = musHdr.channels - trackcount + 1))
	{
		char *lumpname = lumpinfo[lump].name;

		if (diff > 0)
			pr_OFinsert("MUS file %s: header states %i channel%s but %i %s unused.\n",
			lumpname, musHdr.channels, plural(musHdr.channels), diff, diff > 1 ? "are" : "is");
		else
			pr_OFinsert("MUS file %s may be corrupt: tracks %i != channels %i\n",
			 lumpname, trackcount, musHdr.channels + 1);
	}

	buffsize += szHdr;
	buffer = (byte*)M_Calloc(buffsize, 281);

	// Store midi header
	MIDIheader[ 9] = trackcount != 0;
	MIDIheader[11] = trackcount + 1;
	memcpy(buffer, MIDIheader, szHdr);

	bufptr = buffer + szHdr;

	for (i = 0; i < MIDI_MAXNUMCHANS; i++)
	{
		chaninf_t *ch = &chinf[i];

		if (ch->chptr)
		{
			// Timecount will only be set if they are the last non-zero bytes
			// added to a track. In this case they must not be included,
			uint clen = ch->chptr - ch->timecount;

			//-------------------------------------

			_memccpy(bufptr, "MTrk", 0, 4); bufptr += 4;

			// Check for track error
			if (clen > 1 && ch->chbuffer[clen-1] & 0x80) clen -= 2;
			
			// Write size
			clen += 4;
			*bufptr++ = (byte)((clen & 0xff000000) >> 24);
			*bufptr++ = (byte)((clen & 0x00ff0000) >> 16);
			*bufptr++ = (byte)((clen & 0x0000ff00) >> 8);
			*bufptr++ = (byte) (clen & 0x000000ff);
			clen -= 4;

			// Write track data
			memcpy(bufptr, ch->chbuffer, clen); bufptr += clen;

			// End of track marker
			memcpy(bufptr, EOT, 4); bufptr += 4;
		}
	}

	length = bufptr - buffer;

	{
		FILE *f;

		if ( (f = fopen(fname , "wb")) == NULL)
		{
			pr_OFinsert("Couldn't create file %s", fname);
			length = 0;
		}
		else
		{
			fwrite(buffer, length, 1, f);
			fclose(f);
		}

		M_Free(buffer);
	}

	return length > 0;
}

// The lump in data->lump must be pre-validated as a mus type before calling
// The output is stored to a file sent in filename
int ConvertMusToMidi(int lump, char *midiFilename)
{
	byte	*lumpPtr, *ptr;
	int		i, ret = false;

	memset(&musHdr, 0, sizeof(MUSHeader_t));
	
	lumpPtr = lumpCur = (byte*)M_CacheLumpNum(lump, NULL, false);

	if (!lumpPtr) return false;

	ptr = lumpPtr;

	// v220-6 Check for wad in wad mus lump 
	if (lumpinfo[lump].liflags & LINF_LMPISWAD)
	{
		wad_header_t wadhdr = *(wad_header_t *) lumpPtr;
		lump_header_t lhdr  = *(lump_header_t*)(lumpPtr + wadhdr.dir_start) ;
		ptr += lhdr.start;
	}

	memcpy(&musHdr, ptr, sizeof(MUSHeader_t));

	// Initialise arrays, chbuffers will be resized as required
	chinf = (chaninf_t*)M_Calloc( MIDI_MAXNUMCHANS * sizeof(chaninf_t), 282);

	for (i = 0; i < MIDI_MAXNUMCHANS; i++)
	{
		chinf[i].chlength	= MTrk_BUFFER_SIZE;
		chinf[i].channelMap	= 0xFF;
		chinf[i].chbuffer	= (byte*)M_Malloc(MTrk_BUFFER_SIZE, 59);
		chinf[i].oldmidiStatus = 0xFF;
	}

	startPtr = lumpCur = ptr + musHdr.scoreStart;

	total_time = 0;
	event_error = false;

	while (subfn_WriteMidiEvent());	// loop ends

	if ( (ret = !event_error) ) ret = BuildMidiFile(lump, midiFilename);

	if (ge_inf.devmap && !event_error && ret)
	{
		uint ms = (total_time * 1000)/140;

		pr_OFinsert("R3D MUS->MID: lump %s, evt %i, play time %02d:%02d [%i:%i]\n",
								lumpinfo[lump].name,
								total_time,
								ms / 1000 / 60,
								ms / 1000 % 60,
								lumpCur-startPtr, musHdr.scoreLen);
	}

	for (i = 0; i < MIDI_MAXNUMCHANS; i++)
		M_Free(chinf[i].chbuffer);

	M_Free(chinf);
	M_Free(lumpPtr);

	return ret;
}
