//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_nodesfix.c
//
// Revision 2.1.0 build 62 06/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 14 (v220-14) 01/May/2012 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//
// DESCRIPTION: looks for subsector assignment errors from nodes
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <string.h>
#include <math.h>

#include "r3d_MAINspec.h"
#include "r3d_compat.h"

#include "doom/mbf_data.h"
#include "doom/p_local.h"
#include "m_misc.h"
#include "r_things.h"
#include "r3d_utils.h"
#include "p_maputil.h"


// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------


//================== LS_CheckThingsHeight - b211-0 revised
//
// Adjusts the height of Things touching 3D support lines

static RDBOOL setHeightWith3DLine(mobj_t *mo, line_t *ld)
{
	if (PointIsOnLine(mo->x, mo->y, ld, 1))
	{
		int lowht  = ld->backsector->floorheight + ld->d3d->offset3D;
		int height = lowht + ld->d3d->totalheight;

		if (mo->flags & MF_SPAWNCEILING)
		{
			if (mo->height <= ld->d3d->offset3D)
			{
				mo->ceilingz = lowht;
				mo->z = lowht - GetSpriteMaxHeight(mo->sprite);
				mo->flags2 |= MF2_HEIGHTSET;
				return true;
			}
		}
		else if (mo->z != height)
		{
			int lift = ld->frontsector->liftsec3D;
			if (lift >= 0) height += sectors[lift].floorheight;

			mo->z = mo->dropoffz = mo->floorz = height;
			mo->ceilingz = ld->frontsector->ceilingheight;
			mo->flags2 |= MF2_HEIGHTSET;
			return true;
		}
	}

	return false;
}

static RDBOOL setHeightWithNormalLine(mobj_t *mo, line_t *ld)
{
	RDBOOL	online;
	int		fht = ld->frontsector->floorheight;

	if (!mo->radius || mo->z >= fht) return false;

	if (mo->flags3 & MF3_AMODEL)
	{
		vertex_t	vtx;

		vtx.x = mo->x;
		vtx.y = mo->y;

		online = VertexIsOnLine(&vtx, ld);	// v220-5 use accurate onLine test with amodels
	}
	else
	{
		// v220-14 if line entirely inside Thing then reject
		if (   ld->bbox[BOXLEFT ] < mo->x - mo->radius
			|| ld->bbox[BOXRIGHT] > mo->x + mo->radius
			|| ld->bbox[BOXBASE ] < mo->y - mo->radius
			|| ld->bbox[BOXTOP  ] > mo->y + mo->radius)
		{
			online = PointIsOnLine(mo->x, mo->y, ld, 4);	// Doom default within 4 mapunits
		}
		else
		{
			online = false;
		}
	}

	if (online)	
	{
		mo->floorz = mo->dropoffz = mo->z = fht;
		mo->flags2 |= MF2_HEIGHTSET;
	}

	return online;
}

// v220-16
static line_t *checkLinesIterator(int x, int y, vertex_t *vtx)
{
	int			offset;
	const long	*list;
	
	if (x < 0 || y < 0 || x >= bmapwidth || y >= bmapheight)
		return NULL;

	offset = y*bmapwidth + x;

	offset = *(blockmap + offset);
	list = blockmaplump + offset;

	// skip starting delimiter
	while( *(++list) >= 0)
	{
		line_t *ld = &lines[*list];

		if (ld->validcount == validcount)
			continue;		// line has already been checked

		ld->validcount = validcount;

		if (   ld->specType < R3D_40
			&& ld->frontsector == ld->backsector
			&& VertexIsOnLine(vtx, ld))	// use accurate onLine test
		{
			return ld;
		}
	}

	return NULL;
}

// v220-16
// NOTE: only the ssline array is set at this point
static RDBOOL subfn_CheckIfOnSameSidedLine(mobj_t *mo)
{
	subsector_t	*ss, *ssec;
	line_t		*ld;

	// In most circumstances a same sector sided line belongs to the
	// logical sector but, regrettably, not always.
	if (   !((ss = mo->subsector)->sector->specFlags & SF_MOVFLOOR)
	// v220-31 eliminate 3D flats ssecs
		&& !ss->sector->liftsec3D	// $liftsec3D
		&& !ss->flats3DArray)
	{
		line_t	*li = NULL;
		seg_t	*sg;
		int		i;

		// find if on line
		for (i = 0, sg = &segs[ss->firstline]; i < ss->linecount; i++, sg++)
		{
			if (  (ld = sg->linedef)
				&& ld->frontsector == ld->backsector
				&& !(ld->frontsector->specFlags & SF_MOVFLOOR)
				&& XYonLine(mo->x, mo->y, ld, 1))
			{
				li = ld; break;
			}
		}

		if (li)
		{
			ssline_t *sl = &sslines[li - lines];

			for (i = 0; i < sl->num; i++)
			{
				ssec = &subsectors[sl->list[i]];

				if (ssec->sector->floorheight > mo->z)
				{
					mo->floorz = mo->dropoffz = mo->z = ssec->sector->floorheight;
					mo->flags2 |= MF2_HEIGHTSET;
					return true;
				}
			}
		}
	}

	{
		vertex_t	vtx;
		int			box[4];
		int			xl, xh, yl, yh, bx, by;

		vtx.x = mo->x;
		vtx.y = mo->y;

		M_setBoxOffsets(box, mo->x, mo->y, mo->radius);
		M_getBlockOffsets(box, 0, &xl, &xh, &yl, &yh);

		validcount++;

		for (bx=xl ; bx<=xh ; bx++)
		{
			for (by=yl ; by<=yh ; by++)
			{
				if ( (ld = checkLinesIterator(bx, by, &vtx)) )
				{
					mo->floorz = mo->dropoffz = mo->z = ld->frontsector->floorheight;
					mo->flags2 |= MF2_HEIGHTSET;
					return true;
				}
			}
		}
	}

	return false;
}

// v220-5/v220-6/v220-31 modified
// NOTE: only the ssline array is set at this point and must follow LG_FakeMissing
int LS_CheckThingsHeight(void)
{
	subsector_t	*ss;
	thinker_t	*it;
	mobj_t		*mo;
	seg_t		*sg;
	line_t		*ld;
	int			*flatslist;
	int			ht, lnum = -1;
	int			i, count = 0;
	int			mask2 = MF2_GLUECEIL|MF2_DONTDRAW|MF2_HEIGHTSET;

	for (it = thinkercap.next; it != &thinkercap; it = it->next)
	{
		if (it->function != (void (__cdecl *)(void))P_MobjThinker) continue;

		mo = (mobj_t*)it;

		if (mo->flags & MF_NOSECTOR)
		{
			if (   mo->type == MT_TELEPORTMAN
				&& (lnum = FindIfThingIsOnLine(mo, false)) >= 0
				&& lines[lnum].spflags & SPF_NOTEX
				&& lines[lnum].specType == R3D_40
				&& lines[lnum].d3d->offset3D)	// no point continuing if 0
			{
				ld->lnflags2 |= LNF2_3DTELEPLINE;
			}
			continue;	// Only case of thing with MF_NOSECTOR set
		}

		// NOTE: dynlights have MF2_DONTDRAW set and are dealt with elsewhere
		if (mo->flags2 & (MF2_HEIGHTSET|MF2_DONTDRAW)) continue;

		if (mo->spawnz == -32000)	// Only time a negative number is used
		{
			mo->flags2 |= MF2_HEIGHTSET;
			mo->z = mo->floorz = mo->dropoffz =
				mo->subsector->sector->floorheight;
		}
		// if 2nd term !true then pointless as will just fall
		else if (mo->spawnz > 0 && mo->flags & MF_NOGRAVITY)
		{
			mo->flags2 |= MF2_HEIGHTSET;
			mo->z = mo->floorz = mo->dropoffz =
				mo->subsector->sector->floorheight + (mo->spawnz * FRACUNIT) ;
		}

		if (mo->spawnz != 0) continue;	// dealt with above
		if (mo->flags & MF_SPAWNCEILING) continue;	// not supported here

		ss = mo->subsector;

		// v220-31 Look for a Thing in a flats3DArray that has been placed on
		//         a 3D offset line. There is a high chance it might have been.
		if ( (flatslist = mo->subsector->flats3DArray) != NULL)
		{
			if ( (lnum = FindIfThingIsOn3DLine(mo)) >= 0)
			{
				int floor = mo->subsector->sector->floorheight;

				ht = lines[lnum].d3d->offset3D + lines[lnum].d3d->totalheight;
				ht = LS_SetFlatPointer(floor, ht, 0, &mo->flat3Dlevel[1], flatslist);
				mo->z = mo->floorz = mo->dropoffz = ht;

				mo->flags2 |= MF2_HEIGHTSET;
			}
			else if (!mo->info->seestate
				&& mo->z == ss->sector->floorheight
				&& ss->sector->liftsec3D >= 0)
			{
				mo->flags2 |= MF2_SPECNOMOVE;
			}
			continue;
		}

		// v220-16 modified for use with zdnodes
		if (mo->radius && subfn_CheckIfOnSameSidedLine(mo)) continue;

		// Compensate for any changed subsecs
		if (mo->z < ss->sector->floorheight)
		{
			mo->z = ss->sector->floorheight;
			mo->floorz = mo->dropoffz = mo->z;
		}

		sg = &segs[ss->firstline];
		validcount++;

		for (i = 0; i < ss->linecount; i++, sg++)
		{
			if (!(ld = sg->linedef) || ld->validcount == validcount)
				continue;

			ld->validcount = validcount;

			if (ld->frontsector == ld->backsector)
			{
				if (ld->specType >= R3D_40)
				{
					if (   ld->specType < R3D_4A
						&& !(ld->ovflags & OVF_3DLADDERLINE)
						&& setHeightWith3DLine(mo, ld))
					{
						break;
					}
				}
				else if (!(mo->flags2 & mask2)
					&& setHeightWithNormalLine(mo, ld))
				{
					count++;
					break;
				}
			}
		}

		// v220-5 normally inanimate objects are treated differently to overcome problems
		//        with maps where they are spawned overlapping adjacent sectors. This is
		//        caused by their having exaggerated radii which, in the case of specials,
		//        is set to allow them to be picked up at a greater distance than their
		//        true radii would allow.
		if (  !mo->info->seestate
			&& mo->type
			&& !(mo->flags & (MF_SHOOTABLE|MF_SPAWNCEILING))
			&& !(mo->subsector->sector->specFlags & SF_MOVFIXED)
			&& thingFitsInSector(mo, mo->subsector->sector)	// fits in at least one sector axis
			&& FindIfThingIsOnLine(mo, true) < 0)			// is not on a line with mo's sector
		{
			mo->flags2 |= MF2_SPECNOMOVE;
		}
	}

	return count;
}

//=========== LS_CheckDualLineSectorChoice

// Check selected sector where there are only two lines and two sectors.
// Nodes builders can get this wrong.
int LS_CheckDualLineSectorChoice()
{
	subsector_t *ss;
	int i, k, count = 0;

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		if (!(ss->ssflags & SSF_BASESSECFIX))
		{
			seg_t *seg = &segs[ss->firstline], *seg1 = NULL, *seg2 = NULL;

			for (k = 0; k < ss->linecount; k++, seg++)
			{
				if (seg->linedef)
				{
					if (!seg1)
						seg1 = seg;
					else
					{
						if (seg2 || seg1->linedef == seg->linedef)
						{
							seg1 = NULL;
							break;
						}
						seg2 = seg;
					}
				}
			}

			if (!seg1 || !seg2 || !seg1->backsector || !seg2->backsector)
				continue;

			if (seg1->frontsector == seg1->backsector &&
				seg1->frontsector == ss->sector &&
				seg2->frontsector != ss->sector &&
				seg2->backsector  == ss->sector &&
				seg1->length < seg2->length)
			{
				ChangeSSECsec(ss, seg2->frontsector, 23004);
				count++;
			}
			else if (seg2->frontsector == seg2->backsector &&
				seg2->frontsector == ss->sector &&
				seg1->frontsector != ss->sector &&
				seg1->backsector  == ss->sector &&
				seg2->length < seg1->length)
			{
				ChangeSSECsec(ss, seg1->frontsector, 23005);
				count++;
			}
		}
	}
	return count;
}

//================================================================
//
// Orphaned sector checking routines; Jan 2003
// This happens when the nodes builder assigns a line to a sector
// with ,say, four lines, three of which are controlling lines
// and only one of which is actually rendered in the playable space.
// If this single line has been made coincident with a line in the
// correct sector then the nodes builder can wrongly discard the
// correct line's sector for the single coincident line's sector
//
RDBOOL subfn_CheckSectorType(sector_t *sec)
{
	int i, count = 0;
	for (i = 0; i < sec->linecount; i++)
		if (!(sec->lines[i]->spflags & SPF_CONTROLLINE))
			count++;
	return count == 1;
}

int subfn_CheckIfOrphaned(sector_t *right, sector_t *wrong)
{
	subsector_t *ssec;
	int i, k, m, count = 0;

	for ( i = 0; i < right->linecount; i++)
	{
		line_t *ld = right->lines[i];
		RDBOOL lineok = false;

		for(ssec = subsectors, k = 0; k < numsubsectors; k++, ssec++)
			if (ssec->sector == right)
			{
				seg_t *seg = &segs[ssec->firstline];

				for(m = 0; m < ssec->linecount; m++, seg++)
					if (seg->linedef && seg->linedef == ld)
					{
						lineok = true;	// Line found in correct sector;
						break;
					}
			}

		if (lineok) continue;

		// If line not found then search in wrong sector and,
		// if found, change ssec->sector ptr to 'right'.
		for(ssec = subsectors, k = 0; k < numsubsectors; k++, ssec++)
		{
			if (ssec->sector == wrong)
			{
				seg_t *seg = &segs[ssec->firstline];

				for(m = 0; m < ssec->linecount; m++, seg++)
				{
					if (seg->linedef && seg->linedef == ld &&
						subfn_CheckSectorType(ssec->sector))
					{
						ChangeSSECsec(ssec, right, 23006);
						count++;
						break;
					}
				}
			}
		}
	}
	return count;
}

int LS_CheckOrphanedSectors()
{
	line_t	*ld = lines, *ldc;
	int		i, k, count;

	for (count = i = 0; i < numlines; i++, ld++)
	{
		if (ld->specType < R3D_40 &&
			ld->backsector == NULL &&
			ld->frontsector->linecount == 4 &&
			Vtxs2(ld) &&
			sides[ld->sidenum[0]].texFlags == T_MID)
		{
			sector_t *s = ld->frontsector;
			for ( k = 0; k < 4; k++)
			{
				if (s->lines[k]->ml_flags & ML_TWOSIDED &&
					s->lines[k]->specType < R3D_40 &&
					(ldc = subfn_FindCoincidentLine(s->lines[k], 0)) != NULL &&
					LSS(ldc) &&
					Vtxs1(ldc))
				{
					count += subfn_CheckIfOrphaned(s, ldc->frontsector);
				}
			}
		}
	}
	return count;
}

// If conditions are met then check line is in correct ssec->sector
void subfn_CheckNodesWithLine(sector_t *sec, sector_t *s, line_t *ldf)
{
	subsector_t *ssec = subsectors;
	int i, k;
	
	for (i = 0; i < numsubsectors; i++, ssec++)
	{
		if (ssec->sector == s)
		{
			seg_t *seg = &segs[ssec->firstline];
			
			for(k = 0; k < ssec->linecount; k++, seg++)
			{
				if (seg->linedef == ldf)
				{
					ChangeSSECsec(ssec, sec, 23003);
					break;
				}
			}
		}
	}
}


//====================================================

// Extract the correct sector
static int subfn_CheckIfSpecSec(subsector_t *ss, int *osec)
{
	seg_t	*sg = &segs[ss->firstline];
	line_t	*ld;
	int		sectid = ss->sector - sectors;
	int		i, tempid, specSec = -1;

	*osec = -1;

	for (i = 0; i < ss->linecount; i++, sg++)
	{
		if (   !(ld = sg->linedef)
			|| !(ld->ml_flags & ML_TWOSIDED)
			|| ld->spflags & SPF_CONTROLLINE)
		{
			continue;
		}

		if (   *osec != -2
			&& ld->frontsector != ss->sector
			&& !(ld->frontsector->specFlags & (SF_NOLOTEX|SF_NOHITEX)))
		{
			if (*osec < 0) *osec = ld->frontsector - sectors;
			else if (*osec != ld->frontsector - sectors) *osec = -2;
		}

		if (   *osec != -2
			&& ld->backsector != ss->sector
			&& !(ld->backsector->specFlags & (SF_NOLOTEX|SF_NOHITEX)))
		{
			if (*osec < 0) *osec = ld->backsector - sectors;
			else if (*osec != ld->backsector - sectors) *osec = -2;
		}

		if (  (tempid = GetLptr(ld)) < 0
			|| tempid == sectid)
		{
			continue;
		}

		// We may have multiple redirectors.  If so take
		// the sector with the lowest ceiling height.
		if (specSec >= 0 && tempid != specSec)
		{
			sector_t *spSec = &sectors[specSec];
			sector_t *tpSec = &sectors[tempid];

			if (spSec->floorheight == tpSec->floorheight &&
				tpSec->ceilingheight < spSec->ceilingheight)
			{
				return tempid;
			}
		}
		else
		{
			specSec = tempid;
		}
	}
	return specSec;
}

// v220-5
static RDBOOL ssecInvalid(subsector_t *ss)
{
	seg_t	*sg = &segs[ss->firstline];
	int		i;

	if (ss->sector->specFlags & SF_NOLOTEX) return false; 

	for (i = 0; i < ss->linecount; i++, sg++)
	{
		if (   sg->frontsector == sg->backsector
			|| (sg->linedef && lxarray[sg->linedef - lines]) )
		{
			return false;
		}
	}
	return true;
}

void LS_CheckSpecSecSettings()
{
	subsector_t	*ss;
	line_t		*ld;
	seg_t		*sg;
	int			i, k, secnum, count = 0;

	// Modified v2.1.00 build_34
	// Care has to be taken as this routine can cause unwanted changes

	// SpecSec normally set (but not always) with trick floors or
	// maps with lines crossing on trick floors (e.g. 10sector map25)
	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		int		k, osec, valid = 0;

		if (ss->ssflags & SSF_NOCHANGE
			|| ss->floorsec >= 0
			|| ss->ceilsec  >= 0
			|| ssecInvalid(ss))	// v220-5 test added
		{
			continue;
		}

		secnum = subfn_CheckIfSpecSec(ss, &osec);

		if (secnum == -1)
		{
			if (osec >= 0 && (ss->sector->specFlags & 0xFFFFFC) == 0x11888C)	// SF_MOVFLOOR|SF_MOVINVALID type
			{
				SetOverride(ss, osec, 23036);
			}
			continue;
		}

		if (secnum != ss->sector - sectors)
		{
			line_t	*lc;
			int		d;

			sg = &segs[ss->firstline];

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if ( !(lc = sg->linedef)  || !sg->backsector) continue;

				if (sg->frontsector == ss->sector
					&& sg->backsector - sectors == secnum)
				{
					valid |= 1;
				}
				if (sg->frontsector == sg->backsector)
				{
					if (sg->frontsector - sectors == secnum)
					{
						valid |= 2;
					}
					if (sg->frontsector == ss->sector)
					{
						valid |= 4;
					}
				}
				// v220-6 test added - upper walls with missing tex may need to be drawn
				else if (sg->frontsector != sg->backsector
					&& sg->linedef
					&& (d = sg->frontsector->ceilingheight - sg->backsector->ceilingheight) )
				{
					if (   (sg->frontsector == &sectors[secnum] && d > 4*FRACUNIT && sg->backsector->ceilingpic  != skyflatnum)
						|| (sg->backsector  == &sectors[secnum] && d < 4*FRACUNIT && sg->frontsector->ceilingpic != skyflatnum))
					{
						if (!subfn_FindLineCrossing(sg->linedef, LC_NORMAL))	// .e.g. 10sector map17 L1921
							valid |= 8;
					}
				}
			}
		}

		if ( (valid & 7) != 7 && !LS_AllSameSecLine(&sectors[secnum]) )
		{
			// v220-5 1-11-2009
			if (   secnum >= 0
				&& sectors[secnum].specFlags & SF_NOLOTEX
				&& sectors[secnum].floorheight < ss->sector->floorheight)
			{
				secnum = -1;
			}

			// tests added v2.0.01 build_26
			if (   secnum < 0
				|| sectors[secnum].ceilingheight <= ss->sector->ceilingheight
				|| sectors[secnum].floorheight < ss->sector->floorheight
				|| ss->sector->ceilingpic == sectors[secnum].ceilingpic
				|| ss->sector->ceilingpic == skyflatnum)
			{
				// v220-6 add ceilingheight test
				if ( !(valid & 8) )
					Setr3dsscs(ss, secnum, 23037);	// set only if line crossing
				// v220-33 - added
				else if (secnum >= 0
					&& !(sectors[secnum].specFlags & SF_NOLOTEX)
					&& ss->sector - sectors != secnum
					&& ss->sector->ceilingheight - ss->sector->floorheight <= 2*FRACUNIT)
				{
					Setr3dsscs(ss, secnum, 23037);
					if (abs(sectors[secnum].floorheight - ss->sector->floorheight) <= 2*FRACUNIT) goto bypass_floorsec;
				}

				// Note: set even if same as ss->sector.
				Setr3dssfs(ss, secnum, 23037);
bypass_floorsec:;

				// v220-5 check validity of any textureless lines set to R3D_00
				//        These, if any exist, would have been set from 30025
				if (   secnum >= 0
					&& (ss->sector->specFlags & 0xC) == 0xC
					&& ss->sector - sectors != secnum)
				{
					line_t	*store[128];
					seg_t	*sg = &segs[ss->firstline];
					int		c = 0;

					validcount++;

					for (k = 0; k < ss->linecount; k++, sg++)
					{
						if ( !(ld = sg->linedef)
							|| origTEX[ld - lines]
							|| ld->frontsector == ld->backsector
							|| ld->validcount == validcount)
						{
							continue;
						}

						ld->validcount = validcount;

						if (sg->backsector && sg->backsector - sectors == secnum && ld->specType == R3D_00 && c < 128)
						{
							store[c++] = ld; continue;
						}

						c = 0; break; 
					}

					for (k = 0; k < c; k++)
					{
						Forcetype(store[k], R3D_01, 23037);
					}
				}
			}
		}
	}

	ma_inf.sscount3 = count;

	CLR_hitline();

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		if (ss->ssflags & SSF_NOCHANGE) continue;

		if (!(ss->floorsec >= 0 && ss->floorsec == ss->ceilsec))
		{
			ss->ssflags &= ~SSF_SECTSET;

			for (k = 0, sg = &segs[ss->firstline]; k < ss->linecount; k++, sg++)
			{
				if ((ld = sg->linedef) != NULL &&
					ChkLptr(ld) &&
					!CmpLptr(ld, ss->sector - sectors))
				{
					hitline[ld - lines] = 1;
					ss->ssflags |= SSF_SECTSET;
				}
			}
		}
	}

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ChkLptr(ld) && !hitline[i])
		{
			Setspec(ld, -1, 23037);
		}
	}
}
