//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_occlude.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//
// DESCRIPTION: sets skyfix and checks for conditions that could
//              cause occlusion errors in the renderer.
//
//-----------------------------------------------------------------------------

#include <memory.h>

#include "r3d_MAINspec.h"
#include "r3d_compat.h"
#include "m_misc.h"
#include "p_maputil.h"
#include "r_model.h"
#include "r_things.h"

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE -----------------------------------------------------------------------

static void subfn_SetSkyWallFlags()
{
	sector_t	*sec;
	int			i, k;

	SET_countsec(0);

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		// Find lines with textures that will need a skywall.
		if (sec->ceilingpic == skyflatnum &&
			sec->floorpic == skyflatnum &&
			sec->ceilingheight - sec->floorheight <= FRACUNIT &&
			!(sec->specFlags & SF_MOVSECTOR))
		{
			countsec[i] = 1;
		}
	}

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		RDBOOL cond = sec->ceilingpic == skyflatnum &&
			sec->ceilingheight - sec->floorheight <= FRACUNIT &&
			!(sec->specFlags & SF_MOVSECTOR);

		sector_t *s;

		if (cond && (sec->specFlags & 0x3001C) == 0x1C)
		{
			RDBOOL valid = true;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t		*ld = sec->lines[k];
				sector_t	*fs = ld->frontsector, *bs = ld->backsector;

				if (bs && !(origTEX[ld - lines]))
				{
					if (    fs->floorheight == bs->floorheight
						&& (fs->ceilingheight == fs->floorheight
						||  bs->ceilingheight == bs->floorheight))
					{
						Forcetype(sec->lines[k], R3D_00, 34001);
					}
					else if ( (s = sec == fs ? bs : fs) )	// $v67 - affected Eternal map23 sec 527, 528
					{
						if ( !(sec->specFlags & (SF_LINESCLOSED|SF_MULTILOOP)
							&& sec->ceilingheight - sec->floorheight <= FRACUNIT))
						{
							SetSkywall(sec->lines[k], 34002);
						}

						if (   ld->specType != R3D_01
							&& s->floorpic == skyflatnum
							|| ((s->ceilingpic==skyflatnum)^(sec->ceilingpic==skyflatnum)) )
						{
							Forcetype(sec->lines[k], R3D_01, 34002);
						}
					}
				}
			}
		}
		else if (!secdone[i] && cond)
		{
			int valid = false;

			for (k = 0; k < sec->linecount; k++)
			{
				if (sec->lines[k]->ml_flags & ML_TWOSIDED &&
					(countsec[sec->lines[k]->frontsector - sectors] ||
					 countsec[sec->lines[k]->backsector - sectors]))
				{
					valid = true;
					break;
				}
			}

			for (k = 0; k < sec->linecount; k++)
			{
				line_t * ld = sec->lines[k];

				if (origTEX[ld - lines] & 0x11 &&
					!(origTEX[ld - lines] & 0x44))
					continue;

				if (ld->ml_flags & ML_TWOSIDED &&
					ld->specType != R3D_01 &&
					(valid || !origTEX[ld - lines]) &&
					!LineHasMovSectors(ld) &&
					ld->frontsector->ceilingheight !=	//change rev32
					ld->backsector->ceilingheight &&
					ld->frontsector->ceilingpic ==
					ld->backsector->ceilingpic)
				{
					Settype(ld, R3D_00, 34003);
				}
			}
		}
	}
}

static RDBOOL subfn_Trace(line_t *li)
{
	vertex_t *v1 = li->v1, *v2 = li->v2;
	line_t	 *ld;
	int		 i, c = 0, d, repeat = true;

	SET_countsec(0);
	hitline[li - lines] = 1;

	while(repeat)
	{
		repeat = false;

		for (i = 0, ld = lines; i < numlines; i++, ld++)
		{
			if (!(ld->ml_flags & ML_TWOSIDED) &&
				!hitline[i] &&
				ld->v1->x == v2->x &&
				ld->v1->y == v2->y)
			{
				v2 = ld->v2;
				hitline[ld - lines] = 1;
				countsec[ld->frontsector - sectors] = 1;
				c++;
				repeat = true;	// start again
			}
		}
	}

	if (v1 == v2)	// Check loop is closed
	{
		// v220-6 add limits for multisec loop occlusion override.
		//
		// This is fairly arbitrary. Going on what's out there then
		// ultramarine uses 33 1's lines connecting 12 sectors
		// where occlusion or skywalls are not wanted with one
		// of its structures whereas idmgwcon.wad map07has a loop
		// with a chain of 1037 lines embracing 331 sectors where
		// the removal of occlusion is not wanted.
		//
		// From these limits of 64 lines and 32 sectors have been chosen
		//
		if (c < 64)
		{
			for (d = i = 0; i < numsectors; i++)
			{
				if (countsec[i]) d++;
			}

			if (d < 32)
			{
				for (i = 0; i < numsectors; i++)
				{
					if (countsec[i])
						SetNOocclude(&sectors[i], 34004);
				}
			}
		}
		return true;
	}
	return false;
}

static RDBOOL SecIsClosed(int secnum, RDBOOL *alltwosided)
{
	sector_t *sec = &sectors[secnum];
	int i, c1, skyheightdiff = false;
	
	if (sec->linecount < 3) return false;

	CLR_vtxcheck();

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (!(ld->ml_flags & ML_TWOSIDED))
		{
			*alltwosided = false;
		}
		else
		{
			sector_t *s = ld->frontsector == sec ? ld->backsector : ld->frontsector;

			if (!skyheightdiff)
			{
				if (ld->frontsector->ceilingpic == skyflatnum &&
					abs(sec->ceilingheight - s->ceilingheight) > 64*FRACUNIT)
				{
					skyheightdiff = true;
				}
			}
		}

		vtxcheck[ld->v1 - vertexes]++;
		vtxcheck[ld->v2 - vertexes]++;
	}

	if (!skyheightdiff) return false;

	CLR_hitline();

	for (i = 0, c1 = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (vtxcheck[ld->v1 - vertexes] != 2 ||
			vtxcheck[ld->v2 - vertexes] != 2)
		{
			return false;
		}

		if (ld->ml_flags & ML_TWOSIDED
			&& ld->frontsector->ceilingpic != ld->backsector->ceilingpic)
		{
			hitline[ld - lines] = 1;
			c1++;
		}
	}

	if (c1 > 3)
	{
		int ring_count, bbox[4];

		for (i = 0; i < sec->linecount; i++)
		{
			int lnum = sec->lines[i] - lines;

			if (hitline[lnum] == 1
				&& subfn_CheckForRing(hitline, sec, lnum, &ring_count, bbox, NULL, NULL)
				&& LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, bbox, -1, MSIA_TESTAREA))
			{
				return false;
			}
		}
	}
	return true;
}

static RDBOOL skyflatType(sector_t *sec)
{
	if ((sec->specFlags & 0x20018) == 0x20018)
	{
		int i;

		for (i = 0; i < sec->linecount; i++)
		{
			line_t *ld = sec->lines[i];

			if (ld->spflags & SPF_CONTROLLINE) continue;

			if (   origTEX[ld - lines] & 0x22
				|| !(ld->ml_flags & ML_TWOSIDED)
				|| ld->frontsector == ld->backsector
				|| !Vtxs2(ld)
				|| ld->frontsector->ceilingpic != skyflatnum
				|| ld->backsector->ceilingpic != skyflatnum)
			{
				return false;
			}
		}
		return true;
	}
	return false;
}

// Check for skyflat sectors that should not occlude
static void subfn_Occlude()
{
	sector_t *sec;
	int		 i;

	CLR_hitline();

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		int alltwosided = true;

		if (sec->ceilingpic != skyflatnum || sec->skyfix <= 0)
			continue;

		if ( skyflatType(sec) )
		{
			SetNOocclude(sec, 34005);
			continue;
		}

		if (   sec->ceilingheight - sec->floorheight > 8*FRACUNIT
			&& SecIsClosed(i, &alltwosided)
			&& !alltwosided)
		{
			line_t	*li = NULL;
			int		k, bbox[4], ht = sec->ceilingheight;
			RDBOOL	setflag = false;

			M_ClearBox(bbox);

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];
				if (ld->ml_flags & ML_TWOSIDED)
				{
					M_AddLineToBox (bbox, ld);

					if (ld->frontsector->ceilingheight != ht ||
						ld->backsector->ceilingheight != ht)
						setflag = true;

					if (origTEX[ld - lines] == 0x01 ||
						origTEX[ld - lines] == 0x10)
					{
						setflag = false;
						break;
					}
				}
			}

			if (!setflag) continue;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->ml_flags & ML_TWOSIDED) continue;

				if (ld->bbox[BOXLEFT ] <= bbox[BOXLEFT ] ||
					ld->bbox[BOXRIGHT] >= bbox[BOXRIGHT] ||
					ld->bbox[BOXBASE ] <= bbox[BOXBASE ] ||
					ld->bbox[BOXTOP  ] >= bbox[BOXTOP  ])
				{
					li = NULL;
					break;
				}
				li = ld;
			}
			if (li) subfn_Trace(li);
		}
	}
}

//
// LS_CheckSkyfix
//
// 1) Look for adjacent sky sectors and set all such adjacent sectors
//    to the heighest found by adjusting skyfix. 
//
// 2) Things can be added by map authors that extend above surrounding wall
//    heights.  Check for these and set skyfix appropriately.
//

static RDBOOL ssecMatchSeg(sector_t *sec, subsector_t *ss, seg_t *sg)
{
	int i;

	for (i = 0; i < ss->linecount; i++, sg++)
	{
		if (sg->frontsector == sec) return true;
	}
	return false;
}

static void iterateLines(int lnum)
{
	ssline_t	*sl = &sslines[lnum];
	int			i, k;

	for (i = 0; i < sl->num; i++)
	{
		subsector_t *ss = &subsectors[sl->list[i]];

		if (ss->ssvalidcount != ss_validcount)
		{
			ss->ssvalidcount = ss_validcount;

			if (ss->sector->ceilingpic == skyflatnum)
			{
				seg_t	*sg = &segs[ss->firstline];
				line_t	*ld;

				hitssec[hitssecptr++] = ss - subsectors;

				for (k = 0; k < ss->linecount; k++, sg++)
				{
					if ( (ld = sg->linedef) && ld->validcount != validcount)
					{
						ld->validcount = validcount;

						if (   ld->ml_flags & ML_TWOSIDED
							&& ld->frontsector->ceilingpic == skyflatnum
							&& ld->backsector->ceilingpic == skyflatnum)
						{
							iterateLines(ld - lines);
						}
					}
				}
			}
		}
	}
}

static void subfn_CheckSkyfix()
{
	sector_t	*sec;
	line_t		*ld;
	int			c, i, k, count = 0;
	sector_t	**seclist = (sector_t**)M_Malloc(numsectors * sizeof(sector_t*), 0);
	int			seccount;
	int			setSideskyTex;

	SET_countsec(0);

	for (i = 0, sec = sectors; i < numsectors; sec++, i++)
	{
		if (secspflags[i] & SECSPF_NOSKYFIX
			||
			(sec->ceilingpic == skyflatnum
			&& (sec->specFlags & 0x20008) == 0x20008)
			||
			(sec->ceilingpic == skyflatnum
			&&   sec->specFlags & SF_MASKEDDOOR
			&& !(sec->specFlags & SF_MOVLOWER)) )
		{
			countsec[i] = 1;
		}
	}

	validcount++;
	ss_validcount++;

	// Iterate all connected sky sectors
	for (c = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		int ht = DDMININT;

		seccount = hitssecptr = 0;

		if (ld->validcount != validcount)
		{
			ld->validcount = validcount;

			if (   ld->ml_flags & ML_TWOSIDED
				&& ld->frontsector->ceilingpic == skyflatnum
				&& ld->backsector->ceilingpic == skyflatnum)
			{
				iterateLines(i);
			}
		}

		if (!hitssecptr) continue;
		
		setSideskyTex = -1;

		for (k = 0; k < hitssecptr; k++)
		{
			int secnum = subsectors[hitssec[k]].sector - sectors;

			if (!countsec[secnum])
			{
				// v220-3 add sidesky inheritance
				if (sectors[secnum].sidesky & PL_SKYFLAT)
					setSideskyTex = sectors[secnum].sidesky;

				countsec[secnum] |= 2;
				seclist[seccount++] = &sectors[secnum];
			}
		}

		if (seccount)
		{
			for (k = 0; k < seccount; k++)
			{
				int m;

				countsec[seclist[k] - sectors] &= ~2;

				// v220-5 Sometimes the subsectors don't link. This test attempts to
				//        catch any adjoining sky sectors that might have been missed.
				//        This test is only done if seccount is > 1. The whole purpose
				//        of walking the subsectors is to prevent changing skyheights
				//        where they are not linked but a sector has been split.
				if (seccount > 1)
				{
					for (m = 0; m < seclist[k]->linecount; m++)
					{
						line_t *li = seclist[k]->lines[m];

						if (li->ml_flags & ML_TWOSIDED)
						{
							sector_t *s = li->frontsector->ceilingheight > li->backsector->ceilingheight ?
								li->frontsector : li->backsector;

							if (   s->ceilingpic == skyflatnum
								&& s->ceilingheight > ht
								&& !countsec[s - sectors])
							{
								if (s->heightsec >= 0 &&
									sectors[s->heightsec].ceilingheight > ht)
								{
									ht = sectors[s->heightsec].ceilingheight;
								}
								else if (s->ceilingheight > ht)
								{
									ht = s->ceilingheight + s->skyfix;
								}
							}
						}
					}
				}

				if (seclist[k]->heightsec >= 0 &&
					sectors[seclist[k]->heightsec].ceilingheight > ht)
				{
					ht = sectors[seclist[k]->heightsec].ceilingheight;
				}
				else if (seclist[k]->ceilingheight + seclist[k]->skyfix > ht)
				{
					ht = seclist[k]->ceilingheight + seclist[k]->skyfix;
				}
			}

			for (k = 0; k < seccount; k++)
			{
				if (setSideskyTex != -1 && !(seclist[k]->sidesky & PL_SKYFLAT))
					seclist[k]->sidesky = setSideskyTex;

				if (seclist[k]->ceilingheight < ht)
				{
					seclist[k]->skyfix = ht - seclist[k]->ceilingheight;
				}
			}
		}
		setSideskyTex = -1;
	}

	M_Free(seclist);
}

static void subfn_SetCandidateSkyWallFlags()
{
	sector_t	*fs, *bs;
	line_t		*ld;
	int			i;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (  !(ld->ml_flags & ML_TWOSIDED)
			|| (fs = ld->frontsector) == (bs = ld->backsector)
			|| LineHasMovSectors(ld)
			|| fs->heightsec >= 0
			|| bs->heightsec >= 0)
		{
			continue;
		}

		if (   fs->ceilingheight - fs->floorheight > 0
			&& bs->ceilingheight - bs->floorheight > 0)
		{
			// v220-5 E3M6 e.g line L6, L665 an invalid low tex prevents drawing a skywall
			if (   origTEX[i] & 0x11
				&& !(origTEX[i] & 0x44)
				&& fs->ceilingpic == skyflatnum
				&& bs->ceilingpic == skyflatnum
				&& ld->mtflags & R3D_INVLOTEX
				&& !(ld->mtflags & R3D_MISTEXANY))
			{
				SetNoskywall(ld - lines, 34007);
			}

			if (!origTEX[i])
			{
				if (   (!(fs->specFlags & SF_MULTILOOP) && fs->ceilingpic == skyflatnum && fs->ceilingheight <= fs->floorheight)
					|| (!(bs->specFlags & SF_MULTILOOP) && bs->ceilingpic == skyflatnum && bs->ceilingheight <= bs->floorheight) )
				{
					SetSkywall(ld, 34006);

					if (ld->specType != R3D_01) Forcetype(ld, R3D_01, 34006);
				}
				else
				if (   ld->specType == R3D_01
					&& bs->ceilingpic == skyflatnum && bs->floorpic == skyflatnum
					&& fs->ceilingpic == skyflatnum && fs->floorpic == skyflatnum
					// rev97
					&& fs->ceilingheight - fs->floorheight <= MAX_SECHEIGHT
					&& bs->ceilingheight - bs->floorheight <= MAX_SECHEIGHT)
				{
					SetSkywall(ld, 34007);
				}
				continue;
			}
		}

		if (fs->ceilingpic != skyflatnum || bs->ceilingpic != skyflatnum)
		{
			continue;
		}

		if ( ( origTEX[i] == 0x04
			&& bs->ceilingheight == bs->floorheight
			&& bs->ceilingheight + bs->skyfix < fs->ceilingheight + fs->skyfix)
			||
			(  origTEX[i] == 0x40
			&& fs->ceilingheight == fs->floorheight
			&& fs->ceilingheight + fs->skyfix < bs->ceilingheight + bs->skyfix) )
		{
			ld->tyflags |= TYF_SKYWALLCANDIDATE;
		}
	}
}

// R3D - All sorts of shenanigans is required given
//       the info height could be a tissue of lies.
static void subfn_initSkyfixWithSprites()
{
	modeldef_t	*mf;
	mobj_t		*mo;
	sector_t	*sec;
	int			i, temp;

	SET_countsec(0);

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (sec->ceilingpic != skyflatnum) continue;

		// Process things in the sector.
		for (mo = sec->thinglist; mo; mo = mo->snext)
		{
			// b220-1 Don't adjust skyfix with pickups on a lowering floor
			if ( !(mo->flags & MF_SPECIAL
				&& mo->subsector->sector->specFlags & SF_MOVLOWER) )
			{
				// v220-5 revised
				temp = 0;

				if (loadModels && R_CheckModelFor(mo, &mf, NULL) > 0)
					temp = mf->zmax - mf->zmin + mf->zoff;
				else
					temp = GetSpriteMaxHeight(mo->sprite);

				if (temp)
				{
					// This may have a scripted offset or be ceiling spawned
					if (mo->z > mo->subsector->sector->floorheight)
						temp += (mo->z - mo->subsector->sector->floorheight);

					if (temp > countsec[i]) countsec[i] = temp;
				}
			}
		}
	}

	// adjust skyfix with mobj heights
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   countsec[i]
			&& sec->ceilingheight + sec->skyfix - sec->floorheight < countsec[i])
		{
			sec->skyfix = countsec[i] - (sec->ceilingheight - sec->floorheight);
		}
	}
}

static RDBOOL hasHiTex(int ssecidx, sector_t *sec)
{
	subsector_t *ss = &subsectors[ssecidx];
	seg_t		*sg = &segs[ss->firstline];
	int			i, txn;

	for (i = 0; i < ss->linecount; i++, sg++)
	{
		line_t *ld = sg->linedef;

		if (ld == NULL
			|| !(ld->ml_flags & ML_TWOSIDED)) continue;

		txn = origTEX[ld - lines];

		if (txn & 0x22 || txn == 0x05 || txn == 0x50)
		{
			return true;
		}
		if (txn == 0x01 || txn == 0x10)
		{
			if (ld->frontsector == sec || ld->backsector == sec)
				return true;
		}
	}
	return false;
}

void subfn_SetChkSkyFixFlags()
{
	sector_t	*sec;
	int			i;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (sec->specFlags & SF_SSLINETRUE
			&& sec->ceilingpic == skyflatnum)
		{
			RDBOOL	nht = (sec->ceilingheight - sec->floorheight) <= FRACUNIT
						&& !(sec->specFlags & SF_MOVFLOOR);
			RDBOOL	mov = (sec->specFlags & SF_MOVCEIL) > 0;

			int	k, c1 = 0, c2 = 0;

			if (!nht && !mov) continue;

			CLR_hitline();

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (!(ld->ml_flags & ML_TWOSIDED))
				{
					hitline[ld - lines] = 1;
					c1++;
				}
				else if ((nht || !(origTEX[ld - lines] & 0x66))
					&& ld->frontsector != ld->backsector
					&& ld->frontsector->ceilingpic
					== ld->backsector->ceilingpic)
				{
					hitline[ld - lines] = 1;
					c2++;
				}
			}

			if (!c1 || !c2 || c1 + c2 < 3) continue;

			for (k = 0; k < sec->linecount; k++)
			{
				int	ring_count, linnb, bbox[4];
				int m, n, p, lnum = sec->lines[k] - lines;

				if (hitline[lnum] == 1
					&& subfn_CheckForRing(hitline, sec, lnum, &ring_count, bbox, &linnb, NULL)
					&& linnb
					&& LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, bbox, -1, MSIA_NONE)
					&& !LS_LinesInSSECarray(hitline, LSS_NONE, NULL))
				{
					for (m = 0; m < sec->linecount; m++)
					{
						line_t	 *ld = sec->lines[m];

						if (hitline[ld - lines] == 4)
						{
							hitline[ld - lines] = 8;

							if (ld->ml_flags & ML_TWOSIDED)
							{
								ssline_t *sl = &sslines[ld - lines];

								for (n = 0; n < sl->num; n++)
								{
									subsector_t *ss = &subsectors[sl->list[n]];
									seg_t		*sg = &segs[ss->firstline];

									for (p = 0; p < ss->linecount; p++, sg++)
									{
										if (sg->linedef && sg->linedef == ld)
										{
											sg->sgflags |= R3DSEGF_CHKSKYFIX;
										}
									}
								}
							}
						}
					}
					for (m = 0; m < sec->linecount; m++)
					{
						lnum = sec->lines[m] - lines;

						if (hitline[lnum] == 4)
							hitline[lnum] = 8;
					}
				}
			}
		}
	}
}

// Check sky height's correctly set with midtex same sector sided lines.
// The routine only accounts for one occurrence in a map but given that
// this type of map fault is as rare as hens' teeth then the extra code
// to extend it to multiple occurrences was not considered worthwile,
static void subfn_setTrueCeilHeights()
{
	sector_t *sec, *s = NULL;
	int		 c, i, k, m, txn, sf = 0;

	for (c = i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (sec->ceilingpic == skyflatnum
			&& (sec->specFlags & 0x30808) == 0x808)
		{
			int fht = sec->floorheight;
			int cht = sec->ceilingheight;
			int nht = 0, txht;

			if (fht >= cht || fht > 32000*FRACUNIT) continue;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (   ld->ml_flags & ML_TWOSIDED
					&& ld->frontsector == ld->backsector
					&& origTEX[ld - lines] & 0x22)
				{
					int		lnum = ld - lines;
					RDBOOL	valid = true;

					for (m = 0; m < c; m++)
					{
						if (linearray[m] == lnum) { valid = false; break; }
					}

					if (valid) linearray[c++] = lnum;

					for (m = 0; m < 2; m++)
					{
						if (   (txn = sides[ld->sidenum[m]].midtexture)
							&& (txht = texArray[txn]->height) > nht)
						{
							nht = txht<<16;
						}
					}
				}
			}

			if (nht && nht + fht - cht > sf)
			{
				sf = nht + fht - cht;
				s = sec;
			}
		}
	}

	if (s)
	{
		sector_t *cs = XS_GetBoundary(s, NULL);
		int		 box[4];

		sf = s->ceilingheight + sf;

		if (sf > s->ceilingheight + s->skyfix) s->skyfix = sf - s->ceilingheight;

		if (!c || !cs) return;

		SET_countsec(0);

		// Find all connected sky sectors

		M_ClearBox(box);

		for (i = 0; i < cs->linecount; i++)
		{
			line_t *ld = cs->lines[i];

			if (!(ld->spflags & SPF_CONTROLLINE))
				M_AddLineToBox(box, ld);
		}
		
		LS_SubsecsInBox(box);

		for (i = 0; i < hitssecptr; i++)
		{
			if (subsectors[hitssec[i]].sector->ceilingpic == skyflatnum)
				countsec[subsectors[hitssec[i]].sector - sectors] = true;
		}

		for (i = 0; i < numsectors; i++)
		{
			if (countsec[i])
			{
				s = &sectors[i];

				if (sf > s->ceilingheight + s->skyfix)
					s->skyfix = sf - s->ceilingheight;
			}
		}
	}
}

void LS_CheckSkyfix()
{
	subfn_setTrueCeilHeights();
	subfn_SetSkyWallFlags();
	subfn_initSkyfixWithSprites();
	subfn_CheckSkyfix();
	subfn_Occlude();
	subfn_SetCandidateSkyWallFlags();
	subfn_SetChkSkyFixFlags();
}

// added v220-16
// v220-22 SSF_CTRLONLY added
void PS_CheckSubsecOcclude()
{
	subsector_t *ss;
	seg_t		*sg;
	line_t		*ld;
	byte		txn;
	int			i, k, m;

	//-------------------------------

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->sidenum[1] >= 0 && (txn = origTEX[i]) & 0x66)
		{
			if (   (txn & 0x02 && texArray[sides[ld->sidenum[0]].midtexture]->txf_flags & TXF_MASKED)
				|| (txn & 0x04 && texArray[sides[ld->sidenum[0]].toptexture]->txf_flags & TXF_MASKED)
				|| (txn & 0x20 && texArray[sides[ld->sidenum[1]].midtexture]->txf_flags & TXF_MASKED)
				|| (txn & 0x40 && texArray[sides[ld->sidenum[1]].toptexture]->txf_flags & TXF_MASKED) ) 
			{
				ssline_t *sl = &sslines[i];

				for (k = 0; k < sl->num; k++)
				{
					if (!(subsectors[sl->list[k]].ssflags & SSF_NOOCCLUDE))
						subsectors[sl->list[k]].ssflags |= SSF_NOOCCLUDE;
				}
			}
		}

		if (   ld->frontsector == ld->backsector
			&& !(lineflags[i] & LFG_3DTYPE)
			&& !origTEX[i]
			&& ld->specType != R3D_3F)
		{
			Settype(ld, R3D_3F, 34010);
		}
	}

	// v220-33 check for flat ring sectors
	for (i = 0; i < numsectors; i++)
	{
		if (  !secdone[i]
			&& sectors[i].ceilingpic == skyflatnum
			&& sectors[i].floorheight == sectors[i].ceilingheight
			&& (sectors[i].specFlags & 0x3001C) == 0x1C)
		{
			sector_t *os =  NULL, *ref = NULL;

			for (k = 0; k < sectors[i].linecount; k++)
			{
				ld = sectors[i].lines[k];

				if (ld->specType != R3D_00) goto next_sector;
				if (!ld->backsector) goto next_sector;
				if (ld->frontsector == ld->backsector) continue;
				os = ld->frontsector - sectors == i ? ld->backsector : ld->frontsector;
				if (os->ceilingpic != skyflatnum) goto next_sector;

				if (!ref)
					ref = os;
				else if (os != ref)
				{
					if (os->ceilingheight != ref->ceilingheight)  goto next_sector;
					if (os->floorheight != ref->floorheight)  goto next_sector;
				}
			}

			if (os)
			{
				for (k = 0; k < sectors[i].linecount; k++)
				{
					ld = sectors[i].lines[k];
					Settype(ld, R3D_3F, 34010);
				}

				hitssecptr = 0;

				for (k = 0; k < numsubsectors; k++)
				{
					if (subsectors[k].sector - sectors == i)
						hitssec[hitssecptr++] = k;
				}

				// Check all ssec segs' frontsectors point to sector i
				for (k = 0; k < hitssecptr; k++)
				{
					ss = &subsectors[hitssec[k]];

					for (m = 0, sg = &segs[ss->firstline]; m < ss->linecount; m++, sg++)
					{
						if (sg->frontsector && sg->frontsector - sectors != i)
							goto next_sector;
					}
				}

				for (k = 0; k < numsubsectors; k++)
				{
					if (subsectors[k].sector - sectors == i)
					{
						subsectors[k].ssflags |= (SSF_NOOCCLUDE|SSF_USEFSEC);
						SetOverride(&subsectors[k], os - sectors, 34011);
					}
				}

				Setsecdone(&sectors[i], 34012);
				Setsecdone(os, 34012);
				if (ref) Setsecdone(ref, 34012);
			}
		}
next_sector:;
	}

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		for (k = 0, sg = &segs[ss->firstline]; k < ss->linecount; k++, sg++)
		{
			if ((sg->linedef) && sg->linedef->spflags & SPF_CONTROLLINE)
			{
				ss->ssflags |= SSF_CTRLONLY;
				break;
			}
			// v220-33 add trans test
			if (sg->linedef->spflags & (SPF_MIDTRANSLUCENT|SPF_LOWTRANSLUCENT))
			{
				ss->ssflags |= SSF_NOOCCLUDE;
				break;
			}
		}

		if ( !(ss->ssflags & SSF_NOOCCLUDE) )
		{
			if (ss->ceilsec >= 0 || ss->sector->heightsec >= 0)
			{
				subsectors[i].ssflags |= SSF_NOOCCLUDE;
			}
		}
	}
}
