//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2009 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_script_reader.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
//
// DESCRIPTION: script reader from file or lump
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "r3d_compat.h"

#include "con_console.h"
#include "dd_wad.h"
#include "m_misc.h"

// MACROS ------------------------------------------------------------------

#define SCPBUFFER 1024

// TYPES -------------------------------------------------------------------

static struct scp_t {
	char	*iden;
	char	*base;
	char	*ptr;
	char	*end;
	uint	len;
	ushort	lcount;
	RDBOOL	last_string;
	RDBOOL	quoted;
} sv;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern RDBOOL EMAPINFO_script;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static char str_buf[SCPBUFFER + 2];
static char *stt_str;
static char msgbuf[SCPBUFFER + 128];
static char srbuff[SCPBUFFER];
static int  *error_return;
static RDBOOL nextstrtoend;

// CODE --------------------------------------------------------------------

// Can be called by the script reader
void SR_SCPErr(char *message, ...)
{
	if (error_return)
	{
		*error_return = true;
		return;
	}

	if(message == NULL)
	{
		strcpy(srbuff, "Bad syntax.");
	}
	else
	{
		va_list argptr;

		va_start(argptr, message);
		_vsnprintf(srbuff, SCPBUFFER - 2, message, argptr);
		va_end(argptr);
			
		srbuff[SCPBUFFER - 1] = 0;  
	}

	sprintf(msgbuf, "Script Error in \"%s\" line %i:\n%s\n", sv.iden, sv.lcount, srbuff);

	SR_EndScript();
	Con_ErrorD(msgbuf);
}

void SR_SCPWarn(char *message, ...)
{
	if(message == NULL)
	{
		strcpy(srbuff, "Bad syntax.");
	}
	else
	{
		va_list argptr;

		va_start(argptr, message);
		_vsnprintf(srbuff, SCPBUFFER - 2, message, argptr);
		va_end(argptr);
			
		srbuff[SCPBUFFER - 1] = 0;  
	}
	// b211-0 \n added
	sprintf(msgbuf, "Script problem in \"%s\" line %i: %s\n", sv.iden, sv.lcount, srbuff);

	pr_OFdirect(msgbuf);
}

static void subfn_InitSCP(char *name, int *abort_on_err)
{
	error_return = abort_on_err;

	sv.iden = name;
	sv.ptr  = sv.base;
	sv.end  = sv.base + sv.len;
	sv.lcount = 1;
	sv.last_string = false;
	sv.quoted = false;
	nextstrtoend = false;
}

static RDBOOL checkSeparators()
{
	// leading spaces, tabs and crud
	while (sv.ptr < sv.end && (*sv.ptr == ' ' || *sv.ptr <= 0x09))
	{
		sv.ptr++;
	}

	// Comment from /* to */
	if (*sv.ptr == '/' && *(sv.ptr + 1) == '*')
	{
		sv.ptr += 2;

		while (sv.ptr < sv.end)
		{
			if (*sv.ptr == 0xd && *(sv.ptr + 1) == 0xa)
			{
				sv.lcount++;
				sv.ptr++;
			}
			else
			if (*sv.ptr == '\n')
			{
				sv.lcount++;
			}
			else
			if (*sv.ptr == '*' && *(sv.ptr + 1) == '/')
			{
				sv.ptr += 2; break;
			}
			sv.ptr++;
		}
	}

	// Comment to end of line
	if (*sv.ptr == ';' || *sv.ptr == '#' || (*sv.ptr == '/' && *(sv.ptr + 1) == '/'))
	{
		while (sv.ptr < sv.end && *(++sv.ptr) != '\n');
	}

	return sv.ptr >= sv.end;
}

static char *subfn_NextSTR()
{
	int		len;
	RDBOOL	inquote;

	str_buf[0] = 0;

	if (sv.ptr >= sv.end) return NULL;

	sv.quoted = false;

	// v220-4 deal with 0xd and 0xa
	if (*sv.ptr == 0xd && *(sv.ptr + 1) == 0xa)
	{
		sv.ptr += 2;
		if (sv.ptr >= sv.end) return NULL;
		sv.lcount++;
		return str_buf;
	}

	if (*sv.ptr == '\n')
	{
		if (++sv.ptr >= sv.end) return NULL;
		sv.lcount++;
		return str_buf;
	}

	if (checkSeparators()) return NULL;

	if ( (inquote = *sv.ptr == '"') )
	{
		if (++sv.ptr >= sv.end) return NULL;
	}

	stt_str = sv.ptr;

	if (inquote)
	{
		while (sv.ptr < sv.end)
		{
			if (*sv.ptr == '"')
			{
				sv.quoted = true; break;
			}
			sv.ptr++;
		}
	}
	else if (*sv.ptr == '{' || *sv.ptr == '}')
	{
		sv.ptr++;
	}
	else
	{
		while (*sv.ptr != '\n'		// line end
			&& *sv.ptr > 0x09		// ignored if less than 0x10
			&& !(*sv.ptr == ' ' && !nextstrtoend)		// following space
			&& *sv.ptr != '{'		// brackets..
			&& *sv.ptr != '}'
			&& *sv.ptr != '"'		// quote
			&& *sv.ptr != ';'		// comments..
			&& *sv.ptr != '#'
			&& !(*sv.ptr == '/' 
			&& (*(sv.ptr + 1) == '/' || *(sv.ptr + 1) == '*')))
		{
			if (++sv.ptr >= sv.end) break;
		}
	}
	
	nextstrtoend = false;
	
	// v220-4 revamped to deal with 0xd and 0xa
	if ( (len = (int)(sv.ptr - stt_str)) > 1024)
	{
		SR_SCPErr("line length exceeds 1024 chars");
		len = 0;
	}
	else if (len)
	{
		memcpy(str_buf, stt_str, len);

		if (str_buf[len - 1] == 0xa) len -= 1;
		if (len > 0 && str_buf[len - 1] == 0xd) len -= 1;
	}

	str_buf[len] = 0;

	if (sv.quoted && sv.ptr < sv.end) sv.ptr++;

	return str_buf;
}

char *SR_NextSTR(RDBOOL notend)
{
	if (sv.last_string)
	{
		sv.last_string = false;
		if (str_buf[0])
			return str_buf;
	}

	while(subfn_NextSTR())
	{
		if (str_buf[0] || sv.quoted)
		{
			if (!strcmp(str_buf, ",")) continue;

			if (strcmp(str_buf, "=") )
			{
				return str_buf;	// if str_buf != "="
			}
			
			// "=" together with EMAPINFO means include
			// all following spaces to end of line
			nextstrtoend = EMAPINFO_script;
		}
	}

	// EOF
	if (notend) return str_buf;	// return empty str_buf
	return NULL;
}

void SR_KeepString()
{
	sv.last_string = true;
}

int SR_NumForString(RDBOOL *scperror)
{
	char	numstr[128];
	char	*stopstring;
	// Rule out base 8
	int		base = !strnicmp(str_buf, "0x", 2) ? 16 : 10;
	int		L, num;

	strcpy(numstr, str_buf);

	while( (L = strlen(numstr)) > 1 && (numstr[L - 1] == ' ' || numstr[L - 1] == ',') )
		numstr[L - 1] = 0;
	
	num = (int)(strtoul(numstr, &stopstring, base));

	if (*stopstring != 0)
	{
		if (scperror)
		{
			*scperror = true;
			return 0;
		}
		SR_SCPErr("Invalid numeric string \"%s\"", numstr);
		return 0;
	}

	if (scperror) *scperror = false;

	return num;
}

float SR_FloatForString(RDBOOL *scperror)
{
	char	*stopstring;
	double	num = strtod(str_buf, &stopstring);

	if (*stopstring != 0)
	{
		if (scperror)
		{
			*scperror = true;
			return 0;
		}

		SR_SCPErr("Invalid float string \"%s\"", str_buf);
		return 0.0f;
	}

	if (scperror) *scperror = false;

	return (float)num;
}

float SR_NextFloat()
{
	SR_NextSTR(true);
	return SR_FloatForString(NULL);
}

int SR_NextNum()
{
	SR_NextSTR(true);
	return SR_NumForString(NULL);
}

int SR_NextBOOL()
{
	int num;
	
	SR_NextSTR(true);
	num = SR_NumForString(NULL);

	if (num != 0 && num != 1)
	{
		SR_SCPErr("entry must be 0 or 1");
		return 0;
	}

	return num ? true : false;
}

void SR_EndScript(void)
{
	if (!sv.base) return;		// Not open

	M_unlockCACHE(sv.base);
	sv.base = NULL;
}

// lump number loads
void SR_LoadLump(int lump, char *idname)
{
	SR_EndScript();

	sv.base = (char*)M_lumpCACHE(lump, &sv.len, MLC_LOCK, MLC_ERR);
	subfn_InitSCP(idname, NULL);
}

// Script must be open - no checks are made
void SR_RewindScript()
{
	sv.ptr = sv.base;
	sv.lcount = 1;
	sv.last_string = false;
	sv.quoted = false;
	nextstrtoend = false;
}

// v220-6 transfer data direct (cache must be used, however)
RDBOOL loadDATAdirect(void *data, uint size, char *name)
{
	SR_EndScript();

	sv.len = size;
	sv.base = (char*)M_LockedCACHE(sv.len);

	memcpy(sv.base, data, sv.len);
	subfn_InitSCP(name, NULL);
	return true;
}

// file load. Prime subfn_InitSCP() with idname and error address
//  idname must be set but the error address can be NULL
//  This could be made more flexible but does for the moment
RDBOOL SR_LoadFile(char *fname, char *idname, int *abort_on_err)
{
	DFILE	*f;

	SR_EndScript();

	if (!(f = F_OpenFile(fname, "rb")))
	{
		if (abort_on_err) *abort_on_err = true;
		return false;
	}

	F_Seek(f, 0, SEEK_END);
	sv.len = F_Tell(f);
	F_Seek(f, 0, SEEK_SET);

	sv.base = (char*)M_LockedCACHE(sv.len);	// Allocate memory

	F_Seek(f, 0, SEEK_SET);
	F_Read(sv.base, sv.len, f);		// Load the file into memory
	F_Close(f);

	subfn_InitSCP(idname, abort_on_err);
	return true;
}

// pwad lumpname loads - no error if lump not found
RDBOOL SR_LoadPWADSCPLump(int lump)
{
	SR_EndScript();

	if (lump < 0) return false;

	sv.base = (char*)M_lumpCACHE(lump, &sv.len, MLC_LOCK, MLC_ERR);

	subfn_InitSCP(lumpinfo[lump].name, NULL);
	return true;
}

// pwad lumpname loads - no error if lump not found
RDBOOL SR_LoadPWADLumpName(char *lumpname)
{
	int	lump = W_CheckPWADNumForName(lumpname);

	SR_EndScript();

	if (lump < 0 || (ge_inf.nomapinfo && strstr(lumpname, "MAPINFO")) ) return false;

	sv.base = (char*)M_lumpCACHE(lump, &sv.len, MLC_LOCK, MLC_ERR);

	subfn_InitSCP(lumpname, NULL);
	return true;
}

// v220-4 Added to get values from a generic script lump
RDBOOL SR_GetInfoFromScript(char *sname, char *keyname, char *keyword, RDBOOL t1, int l1n, int *l1v, RDBOOL t2, int l2n, int *l2v)
{
	RDBOOL result = false;

	if (SR_LoadPWADLumpName(sname))
	{
		char	*nextstr;
		int		e, i, n = 0;

		while( (nextstr = SR_NextSTR(false)) )
		{
			if (!_stricmp(nextstr, keyname))
			{
				for (i = 0; i < l1n; i++)
				{
					result = false;

					if (t1 == 0)	// pwad lump
					{
						if ( (nextstr = SR_NextSTR(true)) )
						{
							if ( (l1v[i] = W_CheckNumForName(nextstr)) < 0)
							{
								break;
							}
							result = true;
						}
					}
					else			// number
					{
						l1v[i] = SR_NumForString(&e);
						if (e) { result = false; break; }
						result = true;
					}
				}

				if (!keyword || !l2v) break;
				
				nextstr = SR_NextSTR(true);

				if (!_stricmp(nextstr, "{"))
				{
					RDBOOL found = false;

					while( (nextstr = SR_NextSTR(false)) )
					{
						if ( !nextstr || !_stricmp(nextstr, "}"))
							break;

						if (!_stricmp(nextstr, keyword))
						{
							found = true;

							for (i = 0; i < l2n; i++)
							{
								result = false;

								if (t2 == 0)	// pwad lump
								{
									if (  !(nextstr = SR_NextSTR(true))
										|| (l2v[i] = W_CheckNumForName(nextstr)) < 0)
									{
										break;
									}
									result = true;
								}
								else			// number
								{
									nextstr = SR_NextSTR(true);
									l2v[i] = SR_NumForString(&e);
									if (e) break;
									result = true;
								}
							}

							break;
						}
					}

					if (found) break;
				}
				else
				{
					result = false; break;
				}
			}
		}

		SR_EndScript();
	}

	return result;
}
