//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_slopes.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//
// DESCRIPTION: Slopes
//  Translates map definitons.
//  Calculates z vertex values for subsector's nodes.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <string.h>
#include <math.h>

#include "r3d_MAINspec.h"
#include "r3d_slopes.h"
#include "r3d_compat.h"
#include "m_misc.h"
#include "m_bams.h"
#include "p_maputil.h"
#include "sys_system.h"

// MACROS ------------------------------------------------------------------

#define		SLOPE_INCREMENT  32
#define		VTX_INCREMENT	128	// Make 4*SLOPE_INCREMENT

// TYPES -------------------------------------------------------------------

typedef struct {
	float	x, y;
	int		ptr;
} sslistv_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern int	R3D_edit;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int			slopenum, maxslopenum;
slope_t		*slopes;

int			numsidarray, maxnumsidarray;
slsid_t		*slsidarray;

ssecvtxz_t	*ssecvtxz;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static int	ssecvtxnum, maxssecvtxnum;
static char equate[] = {"\nEDIT MODE: If this does not equate then ensure the sector's lines' vertexes have been merged.\n"
						"R3D does not merge vertexes in edit mode to preserve the wad's line, side and sector numbering.\n"
						"Use a map editor to do this before running the map again with R3D in edit mode.\n"};

// CODE --------------------------------------------------------------------

// Initialise slope array, ssec vtxz array
void Init_SlopeArrays()
{
	R3D_slopetypes = false;

	slopenum = 0;
	maxslopenum = 0;
	numsidarray = 0;
	maxnumsidarray = 0;

	slopes = NULL;
	slsidarray = NULL;

	// This is always initialised as the first ptr is used (points to zeros)
	ssecvtxz = (ssecvtxz_t*)W_Malloc(sizeof(ssecvtxz_t), LEVELID, 25);
	ssecvtxz[0].z[0] = 0;
	ssecvtxz[0].z[1] = 0;

	maxssecvtxnum = 0;	// Deliberately set to 0 (not 1) to prevent odd numbered array size change
	ssecvtxnum = 1;		// Don't change ssecvtxz[0]
}

static RDBOOL PrecisionVector(vertex_t *v1, vertex_t *v2, vertex_t *r1, vertex_t *r2, int *x, int *y)
{
	double	ax = (double)r1->x  , ay = (double)r1->y;
	double	bx = (double)r2->x  , by = (double)r2->y;
	double	cx = (double)v1->x  , cy = (double)v1->y;
	double	dx = cx + (double)(v1->x - v2->x), dy = cy + (double)(v1->y - v2->y);
	double	dv = (bx - ax)*(dy - cy) - (by - ay)*(dx - cx);

	// If dv is zero then lines have same slope (and therefore cannot intersect)
	if (dv)
	{
		double	r = ((ay - cy)*(dx - cx) - (ax - cx)*(dy - cy)) / dv;
		double	xd = (ax + r*(bx - ax));
		double	yd = (ay + r*(by - ay));

		if      (xd >  32767*FRACUNIT) *x = DDMAXINT;
		else if (xd < -32768*FRACUNIT) *x = DDMININT;
		else *x = (int)xd;

		if      (yd >  32767*FRACUNIT) *y = DDMAXINT;
		else if (yd < -32768*FRACUNIT) *y = DDMININT;
		else *y = (int)yd;

		return true;
	}
	return false;
}

static RDBOOL StandardLineVector(subslope_t *sub, int ax, int ay, int *x, int *y)
{
	double dx = sub->dx, dy = sub->dy;
	double cosfrac = sub->cosfrac, sinfrac = sub->sinfrac;
	double dv = cosfrac * dy + sinfrac * dx;

	// If dv zero then lines have same slope (and therefore cannot intersect)
	if (dv)
	{
		double r = (((double)ay - (double)sub->v1[VY]) * dx - ((double)ax - (double)sub->v1[VX]) * dy) / dv;

		*x = ax + (int)(r * cosfrac);
		*y = ay - (int)(r * sinfrac);

		return true;
	}
	return false;
}

static int pointOffsetZ(int stptr, int xp, int yp, int hi)
{
	if (stptr >= 0 && slopes[stptr].slope[hi].slopeheight)
	{
		subslope_t	*sub = &slopes[stptr].slope[hi];
		int			x, y;

		if (!sub->dx)
		{
			if (yp <= sub->bbox[BOXTOP ] &&
				yp >= sub->bbox[BOXBASE])
			{
				int ht = sub->slopeheight * abs(FixedDiv(yp - sub->v1[VY], sub->dy));
				return hi ? -ht : ht;
			}
		}

		else if (!sub->dy)
		{
			if (xp <= sub->bbox[BOXRIGHT] &&
				xp >= sub->bbox[BOXLEFT ])
			{
				int ht = sub->slopeheight * abs(FixedDiv(xp - sub->v1[VX], sub->dx));
				return hi ? -ht : ht;
			}
		}

		else if (StandardLineVector(sub, xp, yp, &x, &y))
		{
			if (x <= sub->bbox[BOXRIGHT] &&
				x >= sub->bbox[BOXLEFT ] &&
				y <= sub->bbox[BOXTOP  ] &&
				y >= sub->bbox[BOXBASE ])
			{
				double dx = x - sub->v1[VX], dy = y - sub->v1[VY];
				int val = (int)sqrt(dx*dx + dy*dy);
				int ht = sub->slopeheight * FixedDiv(val, sub->len);
				
				return hi ? -ht : ht;
			}
		}
	}
	return 0;
}

static int pointOffset(int stptr, int xp, int yp, int hi)
{
	if (stptr >= 0 && slopes[stptr].slope[hi].slopeheight)
	{
		subslope_t	*sub = &slopes[stptr].slope[hi];
		int			x, y;

		if (!sub->dx)
		{
			if (yp < sub->bbox[BOXTOP ] &&
				yp > sub->bbox[BOXBASE])
			{
				int ht = sub->slopeheight * abs(FixedDiv(yp - sub->v1[VY], sub->dy));
				return hi ? -ht : ht;
			}
		}

		else if (!sub->dy)
		{
			if (xp < sub->bbox[BOXRIGHT] &&
				xp > sub->bbox[BOXLEFT ])
			{
				int ht = sub->slopeheight * abs(FixedDiv(xp - sub->v1[VX], sub->dx));
				return hi ? -ht : ht;
			}
		}

		else if (StandardLineVector(sub, xp, yp, &x, &y))
		{
			if (x < sub->bbox[BOXRIGHT] &&
				x > sub->bbox[BOXLEFT ] &&
				y < sub->bbox[BOXTOP  ] &&
				y > sub->bbox[BOXBASE ])
			{
				double dx = x - sub->v1[VX], dy = y - sub->v1[VY];
				int val = (int)sqrt(dx*dx + dy*dy);
				int ht = sub->slopeheight * FixedDiv(val, sub->len);
				
				return hi ? -ht : ht;
			}
		}
	}
	return 0;
}

static int pointHeightSet(subslope_t *sub, int xp, int yp)
{
	int ht = 0;

	if (sub->slopeheight)
	{
		int x, y;

		if (!sub->dx)
		{
			ht = sub->slopeheight * abs(FixedDiv(yp - sub->v1[VY], sub->dy));
		}
		else if (!sub->dy)
		{
			ht = sub->slopeheight * abs(FixedDiv(xp - sub->v1[VX], sub->dx));
		}
		else if (StandardLineVector(sub, xp, yp, &x, &y))
		{
			double dx = x - sub->v1[VX], dy = y - sub->v1[VY];
			int val = (int)sqrt(dx*dx + dy*dy);

			ht = sub->slopeheight * FixedDiv(val, sub->len);
		}
	}

	return ht;
}

static void setTRIWALLsegs(sector_t *sec, int hi)
{
	int			i, k, m, n, svnum = 0, svmaxnum = 32;
	sslistv_t	*svarray = (sslistv_t*)M_Malloc(svmaxnum * sizeof(sslistv_t), 68);

	// Set up an array of vertex heights from the ssec verts list
	// The x, y values will also match with any segs on the other
	// side such as the segs attached to any LNF2_LOWTRIWALL flagged
	// lines where the tex to be rendered is on the other side.

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		zvertex_t	*vtx = ss->verts;

		for (k = 0; k < ss->numverts; k++, vtx++)
		{
			if (vtx->zptr > 0)
			{
				if (svnum >= svmaxnum)
				{
					svmaxnum += 32;
					svarray = (sslistv_t*)M_Realloc(svarray, svmaxnum * sizeof(sslistv_t), 38);
				}

				svarray[svnum  ].x = vtx->x;
				svarray[svnum  ].y = vtx->y;
				svarray[svnum++].ptr = vtx->zptr;
			}
		}
	}

	// Set the triwall seg vertex heights
	for (i = 0; i < sec->linecount; i++)
	{
		ssline_t *sl;
		int txn = origTEX[sec->lines[i] - lines];
		int testtxn = txn & 0x22;
		int lowall, hiwall, triwall = (testtxn == 0x02 || testtxn == 0x20);

		testtxn = txn & 0x11;
		lowall  = (testtxn == 0x01 || testtxn == 0x10);
		testtxn = txn & 0x44;
		hiwall  = (testtxn == 0x04 || testtxn == 0x40);

		if (!triwall && !lowall && !hiwall) continue;

		sl = &sslines[sec->lines[i] - lines];

		for (k = 0; k < sl->num; k++)
		{
			subsector_t *ss = &subsectors[sl->list[k]];
			seg_t		*sg = &segs[ss->firstline];

			for (m = 0; m < ss->linecount; m++, sg++)
			{
				line_t *li = sg->linedef;

				if (li && (li->frontsector == sec || li->backsector == sec))
				{
					sslistv_t	*sv = svarray;

					for (n = 0; n < svnum; n++, sv++)
					{
						if (sv->x == sg->fv1[VX] && sv->y == sg->fv1[VY])
						{
							sg->zptr1 = sv->ptr;
						}
						else
						if (sv->x == sg->fv2[VX] && sv->y == sg->fv2[VY])
						{
							sg->zptr2 = sv->ptr;
						}
					}
				}
			}
		}
	}

	M_Free(svarray);
}

static slsid_t *getTriVerts(sector_t *sec, int *vsnum, int *type, int hi)
{
	slsid_t	 *sld = NULL;
	int		 i, k, vn[3];

	(*type) = -1;

	memset(vn, -1, sizeof(vn));

	// Get the three vertexes
	for (i = 0; i < 3; i++)
	{
		int		vn0 = sec->lines[i]->v1 - vertexes;
		int		vn1 = sec->lines[i]->v2 - vertexes;

		if (vn[0] < 0 && vn[1] < 0)
		{
			vn[0] = vn0, vn[1] = vn1;
		}
		else if (vn[0] != vn0 && vn[1] != vn0)
		{
			vn[2] = vn0; break;
		}
		else if (vn[0] != vn1 && vn[1] != vn1)
		{
			vn[2] = vn1; break;
		}
	}

	for (i = 0; i < 3; i++)
	{
		line_t	*ld = sec->lines[i];
		int		sidenum = ld->frontsector == sec ?
					ld->sidenum[0] : ld->sidenum[1];
		int offset;

		sld = slsidarray;

		for (k = 0; k < numsidarray; k++, sld++)
		{
			if (sld->sidenum != sidenum) continue;

			offset = hi ? sld->offsethigh : sld->offsetlow;

			if (offset)
			{
				vsnum[0] = ld->v1 - vertexes;
				vsnum[1] = ld->v2 - vertexes;
				vsnum[2] = -1;

				*type = (sld->type[hi] == 1) ? 1 : 0;
				break;
			}
		}
		if ((*type) >= 0) break;
	}

	if ((*type) < 0) return NULL;

	for (i = 0; i < 3; i++)
	{
		if (vn[i] != vsnum[0] &&
			vn[i] != vsnum[1])
		{
			vsnum[2] = vn[i];
		}
	}

	return sld;
}

static slsid_t *getQuadVerts(sector_t *sec, int *vsnum, int hi)
{
	int		 i, k, m;

	for (i = 0; i < 4; i++)
	{
		slsid_t		*sld = slsidarray;
		line_t		*ld = sec->lines[i];
		int			side = !(ld->frontsector == sec);
		int			sidenum = ld->sidenum[side];
		int			offset;

		for (k = 0; k < numsidarray; k++, sld++)
		{
			if (sld->sidenum != sidenum) continue;

			offset = hi ? sld->offsethigh : sld->offsetlow;

			if (offset)
			{
				vertex_t *v1 = ld->v1, *v2 = ld->v2;

				vsnum[0] = v1 - vertexes;
				vsnum[1] = v2 - vertexes;

				for (m = 0; m < 4; m++)
				{
					if (m != i)
					{
						line_t	*li = sec->lines[m];
						int		x, y;

						if (   li->v1 != v1 && li->v1 != v2
							&& li->v2 != v1 && li->v2 != v2)
						{
							if (PrecisionVector(ld->v1, ld->v2, li->v1, li->v2, &x, &y))
							{
								PrintProgInfo("SLOPE SEC %i MAP ERROR: is not a trapezium. line %i should be parallel with line %i\n%s",
									sec - sectors, ld - lines, li - lines, R3D_edit ? equate : "" );
								ma_inf.slopeerror++;
								return NULL;
							}
							vsnum[2] = li->v1 - vertexes;	// Doesn't matter which vertex is used
							return sld;
						}
					}
				}
			}
		}
	}
	return NULL;
}

static void SetZheightPtr(zvertex_t *vtx, float height, int hi)
{
	ssecvtxz_t *zh;

	if (!vtx->zptr)
	{
		if (ssecvtxnum >= maxssecvtxnum)
		{
			maxssecvtxnum += VTX_INCREMENT;
			// The new assignment must be cleared
			ssecvtxz = (ssecvtxz_t*)W_ReallocCLR(ssecvtxz, maxssecvtxnum * sizeof(ssecvtxz_t), LEVELID, 21);
		}
		vtx->zptr = ssecvtxnum++;
	}

	zh = &ssecvtxz[vtx->zptr];

	if (hi) zh->z[1] = -height; else zh->z[0] = height;
}

static void subfn_SetSlopeVtxHeights(sector_t *sec, int hi, int analyse)
{
	vertex_t	*vs1, *vs2, *vs3, vs4;
	slsid_t		*sld;
	int			vsnum[3];
	int			fht = sec->floorheight;
	int			slnum = sec->linecount;
	float		height;
	int			i, k, type, looplist[4];

	if (slnum == 3)
	{
		if ( !(sld = getTriVerts(sec, vsnum, &type, hi)) ) return;
	}
	else if (slnum == 4)
	{
		if ( !(sld = getQuadVerts(sec, vsnum, hi)) ) return;
		
		if (sld->type[hi] != 1)
		{
			PrintProgInfo("SLOPE SEC %i MAP ERROR: should have a $B side tex\n%s", sec - sectors, R3D_edit ? equate : "" );
			ma_inf.slopeerror++;
			return;
		}
		type = 1;
	}
	else
	{
		PrintProgError("Program error [7]: slnum %i", slnum);
	}

	vs1 = &vertexes[vsnum[0]];
	vs2 = &vertexes[vsnum[1]];
	vs3 = &vertexes[vsnum[2]];

	height = hi ? (float)sld->offsethigh : (float)sld->offsetlow;

	if (!height) return;

	for (i = 0; i < slnum; i++) looplist[i] = sec->lines[i] - lines;

	// Rather than muck about the subsecs are found by looking at
	// the area bounded.  Probably slower than providing a specific
	// routine based on the sslines array but hardly noticeable.
	if (LS_SetSSECSinLoop(sec, NULL, slnum, looplist, MSIA_NONE))
	{
		// For point height calcs
		subslope_t	*sub;
		double		angle;
		int			xi, yi;
		int			gradient;
		RDBOOL		error = false;

		// Check line errors
		for (i = 0; i < hitssecptr; i++)
		{
			subsector_t *ss = &subsectors[hitssec[i]];
			seg_t		*sg = &segs[ss->firstline];

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if (sg->linedef && sg->frontsector != sec)
				{
					PrintProgInfo("SLOPE SEC %i MAP ERROR: has an incorrect lineside sector\n%s", sec-sectors, R3D_edit ? equate : "" );
					ma_inf.slopeerror++;
					error = true;
					break;
				}
			}
		}

		if (error) return;

		if (sec->stptr < 0)
		{
			if (slopenum >= maxslopenum)
			{
				maxslopenum += SLOPE_INCREMENT;

				// The new portion must be cleared
				slopes = (slope_t*)W_ReallocCLR(slopes, maxslopenum * sizeof(slope_t), LEVELID, 22);
			}

			sec->stptr = slopenum;

			sub = &slopes[slopenum++].slope[hi];
		}
		else
		{
			sub = &slopes[sec->stptr].slope[hi];
		}

		sub->dx = vs1->x - vs2->x;
		sub->dy = vs1->y - vs2->y;

		vs4.x = vs3->x;
		vs4.y = vs3->y;

		if (!sub->dx)
		{
			vs4.x += FRACUNIT;
		}
		else if (!sub->dy)
		{
			vs4.y += FRACUNIT;
		}
		else
		{
			angle = atan2(FIX2DBL(sub->dx), FIX2DBL(sub->dy));
			sub->cosfrac = (int)(cos(angle)*FRACUNIT);
			sub->sinfrac = (int)(sin(angle)*FRACUNIT);
			vs4.x += sub->cosfrac;
			vs4.y -= sub->sinfrac;
		}

		// height vector
		sub->v1[VX] = vs3->x;
		sub->v1[VY] = vs3->y;
		sub->slopeheight = (int)height;

		if (PrecisionVector(vs1, vs2, vs3, &vs4, &xi, &yi))
		{
			float	mapangle;

			if (!type)
			{
				sub->v1[VX] = xi;
				sub->v1[VY] = yi;
				sub->v2[VX] = vs3->x;
				sub->v2[VY] = vs3->y;
			}
			else
			{
				sub->v2[VX] = xi;
				sub->v2[VY] = yi;
			}

			sub->dx = sub->v1[VX] - sub->v2[VX];
			sub->dy = sub->v1[VY] - sub->v2[VY];

			// Calculate the direction of the slope in the map
			mapangle = R_PointToAngle2(sub->v1[VX], sub->v1[VY], sub->v2[VX],
				sub->v2[VY])/(float)ANGMAX * 360 - 90;

			if (mapangle < 0) mapangle += 360;

			sub->mapangle = (int)mapangle;

			{
				double	dx = sub->dx, dy = sub->dy;

				sub->len = (int)sqrt(dx*dx + dy*dy);

				dx = sub->len;
				dy = height*FRACUNIT;

				gradient = (int)(fabs(dy/dx * FRACUNIT));

				if (!hi)
				{
					// set slopeangles for THING orientation on a low slope
					double gradangle = (-180/PI * atan2(dy, dx));

					sec->slopeangles[0] = (float)gradangle;
					sec->slopeangles[1] = (float)(mapangle/180 * PI);

					// Reduce gradient if stepup <= 24
					if (height < 25 && gradient > (FRACUNIT>>3))
					{
						gradient = FRACUNIT>>3;
					}
				}
			}

			sub->gradient = gradient;

			if (sub->dx && sub->dy)
			{
				angle = atan2(FIX2DBL(sub->dx), FIX2DBL(sub->dy));

				sub->cosfrac = (int)(cos(angle)*FRACUNIT);
				sub->sinfrac = (int)(sin(angle)*FRACUNIT);
			}

			sub->bbox[BOXLEFT ] = MIN_OF(sub->v1[VX], sub->v2[VX]);
			sub->bbox[BOXRIGHT] = MAX_OF(sub->v1[VX], sub->v2[VX]);
			sub->bbox[BOXBASE ] = MIN_OF(sub->v1[VY], sub->v2[VY]);
			sub->bbox[BOXTOP  ] = MAX_OF(sub->v1[VY], sub->v2[VY]);
		}
		else
		{
			PrintProgError("Vector creation failed sec %i\n",sec - sectors);
		}

		for (i = 0; i < hitssecptr; i++)
		{
			subsector_t *ss  = &subsectors[hitssec[i]];
			zvertex_t	*vtx = ss->verts;

			for (k = 0; k < ss->numverts; k++, vtx++)
			{
				int xD = (int)(vtx->x*FRACUNIT);
				int yD = (int)(vtx->y*FRACUNIT);

				if (xD == vs1->x && yD == vs1->y)
				{
					if (type)
					{
						SetZheightPtr(vtx, height, hi);
					}
				}
				else if (xD == vs2->x && yD == vs2->y)
				{
					if (type)
					{
						SetZheightPtr(vtx, height, hi);
					}
				}
				else if (xD == vs3->x && yD == vs3->y)
				{
					if (!type)
					{
						SetZheightPtr(vtx, height, hi);
					}
				}
				else
				{
					float res = FIX2FLT(pointHeightSet(sub, xD, yD));
					if (res)
					{
						SetZheightPtr(vtx, res, hi);
					}
				}
			}
		}
		
		setTRIWALLsegs(sec, hi);

		if ( hi) sec->specFlags2 |= SF2_HISLOPE;
		if (!hi) sec->specFlags2 |= SF2_LOSLOPE;

		if (analyse)
		{
			Setsecdone3(sec, 45001);

			for (i = 0; i < sec->linecount; i++)
			{
				line_t *ld = sec->lines[i];

				ld->lnflags2 |= LNF2_SLOPESECLN;
				Settype(ld, R3D_3D, 45001);
				SetLFGSecSet(ld, 45001);
				lineflags[ld - lines] |= LFG_3DTYPE;
			}
		}
	}
	else
	{
		PrintProgInfo("SLOPE SEC %i MAP ERROR: malformed or not closed\n%s", sec-sectors, R3D_edit ? equate : "" );
		ma_inf.slopeerror++;
	}
}

static int getAdefHeight(sector_t *s, int csecptr, int sidenum, int hi)
{
	vertexowner_t	*own;
	vertex_t		*vop = NULL;
	sector_t		*slp = NULL;
	line_t			*chk = NULL;
	int				i, k, secnum;

	if (s->linecount != 3) return 0;

	// Find the opposing vertex
	for (i = 0; i < 3; i++)
	{
		line_t *ld = s->lines[i];

		if (ld->sidenum[0] == sidenum || ld->sidenum[1] == sidenum)
		{
			chk = ld; break;
		}
	}

	if (!chk) return 0;

	for (i = 0; i < 3; i++)
	{
		line_t *ld = s->lines[i];

		if (ld == chk) continue;

		if (ld->v1 != chk->v1 && ld->v1 != chk->v2) { vop = ld->v1; break; }
		if (ld->v2 != chk->v1 && ld->v2 != chk->v2) { vop = ld->v2; break; }
	}

	if (!vop) return 0;

	own = &vertexowners[vop - vertexes];

	for (i = 0; i < own->num; i++)
	{
		sector_t *sec = &sectors[own->list[i]];
		int		 snum = sec - sectors, found = false;

		if (sec == s || sec->linecount != 3) continue;

		for (k = 0; k < csecptr; k++)
		{
			if (countsec[k] == snum) { found = true; break; }
		}

		if (!found) continue;

		for (k = 0; k < 3; k++)
		{
			line_t *ld = sec->lines[k];

			if (ld != chk && ld->ml_flags & ML_TWOSIDED
				&& !origTEX[ld - lines]
				&& (ld->v1 == vop || ld->v2 == vop)
				&& (ld->frontsector == s || ld->backsector == s))
			{
				slp = sec; break;
			}
		}
		if (slp) break;
	}

	if (!slp) return 0;

	secnum = slp - sectors;

	for (i = 0; i < slp->linecount; i++)
	{
		slsid_t	*sld;
		line_t	*ld = slp->lines[i];
		int		side = ld->frontsector == slp ? 0 : 1;
		int		snum = ld->sidenum[side];

		if (snum < 0 || snum == sidenum) continue;

		for (k = 0, sld = slsidarray; k < numsidarray; k++, sld++)
		{
			if (sld->sidenum == snum)
			{
				if ( hi && sld->offsethigh && sld->offsethigh <= 0x7FFF) return sld->offsethigh;
				if (!hi && sld->offsetlow  && sld->offsetlow  <= 0x7FFF) return sld->offsetlow;
			}
		}
	}

	return 0;
}

// Look for textures with B lines defined on the opposite side
static void swapTextures(int sidenum, RDBOOL hi)
{
	sector_t *sec;
	line_t	 *ld;
	side_t	 *sideA = &sides[sidenum], *sideB = NULL;
	int		 i, side = 1;

	if (  (!hi && sideA->texFlags & T_LOW)
		|| (hi && sideA->texFlags & T_HI))
	{
		return;
	}

	// Find the other side
	for (i = 0, sec = sideA->sector; i < sec->linecount; i++)
	{
		ld = sec->lines[i];

		if (ld->sidenum[0] == sidenum)
		{
			if (ld->sidenum[1] >= 0) sideB = &sides[ld->sidenum[1]];
			break;
		}
		if (ld->sidenum[1] == sidenum)
		{
			sideB = &sides[ld->sidenum[0]];
			side = 0;
			break;
		}
	}

	if (!sideB) return;

	if (!hi && sideB->texFlags & T_LOW)
	{
		sideA->bottomtexture = sideB->bottomtexture;
		sideB->bottomtexture = 0;
		sideA->texFlags |=  T_LOW;
		sideB->texFlags &= ~T_LOW;
		origTEX[ld - lines] &= ~0x11;
		if (side) origTEX[ld - lines] |= 0x10; else origTEX[ld - lines] |= 0x01;
	}

	if (hi && sideB->texFlags & T_HI)
	{
		sideA->toptexture = sideB->toptexture;
		sideB->toptexture = 0;
		sideB->texFlags |=  T_HI;
		sideA->texFlags &= ~T_HI;
		origTEX[ld - lines] &= ~0x44;
		if (side) origTEX[ld - lines] |= 0x40; else origTEX[ld - lines] |= 0x04;
	}
}

// Public Functions =================================================================

// point heights must use unsmoothed values

int SL_PointHeightZ(sector_t *sec, int xp, int yp, int hi)
{
	int secht = hi ? sec->ceilingheight : sec->floorheight;

	return pointOffsetZ(sec->stptr, xp, yp, hi) + secht;
}

int SL_PointHeight(sector_t *sec, int xp, int yp, int hi)
{
	int secht = hi ? sec->ceilingheight : sec->floorheight;

	return pointOffset(sec->stptr, xp, yp, hi) + secht;
}

// Returns slope;s z height for vector from any point to sec (no bounds test)
int SL_XYHeight(int stptr, int xp, int yp, int hi)
{
	int ht = (stptr >= 0) ? pointHeightSet(&slopes[stptr].slope[hi], xp, yp) : 0;

	return hi ? -ht : ht;
}

// v220-4
float SL_PointHeightFP(sector_t *sec, float xp, float yp, int hi)
{
	int secht = hi ? sec->ceilingheight : sec->floorheight;

	return FIX2FLT(pointOffset(sec->stptr, (int)(xp*FRACUNIT), (int)(yp*FRACUNIT), hi) + secht);
}

// end point heights

// v220-33 texture heights must use smoothed values

int SL_LowCeilHeight(sector_t *s)
{
	if (s->stptr < 0) return s->smoothceil;

	return s->smoothceil - slopes[s->stptr].slope[1].slopeheight * FRACUNIT;
}

int SL_HighFloorHeight(sector_t *s)
{
	if (s->stptr < 0) return s->smoothfloor;

	return s->smoothfloor + slopes[s->stptr].slope[0].slopeheight * FRACUNIT;
}

// end texture heights

// Functions called on loading map

// Deals with map slope info read in P_LoadSideDefs2()
void LS_SetTiltOffsets(int analyse)
{
	slsid_t	 *sld;
	seg_t	 *sg;
	byte	 *typemarker;
	int		 i, k, csecptr = 0;
	const char	 *err = "SLOPE MAP ERROR : sector ";

	if (!R3D_slopetypes || !numsidarray) return;

	SetProgHeader("SLOPES error report;\n\n");

	// If an analyser file's been loaded we'll need to init some arrays.
	if (!analyse)
	{
		R3D_Set_sslinesArray();

		// subfn_SetSlopeVtxHeights() with int looplist[4] calls
		// LS_SetSSECSinLoop() which in turn calls LS_MarkSubsecsInLoop()
		// This uses linearray which means it must be at least the same
		// size as looplist.

		// Actually linearray only needs 4 but what the heck.
		linearray = (int  *)M_Malloc(numlines      * sizeof(int ), 69);
		hitssec	  = (int  *)M_Malloc(numsubsectors * sizeof(int ), 70);
		ssecarray = (byte *)M_Malloc(numsubsectors * sizeof(byte), 71);
		countsec  = (int  *)M_Malloc(numsectors    * sizeof(int ), 72);
	}

	typemarker = (byte*)M_Calloc(numsectors, 298);

	for (i = 0, sld = slsidarray; i < numsidarray; i++, sld++)
	{
		sector_t *s = sides[sld->sidenum].sector;
		int		 height, secnum = s - sectors;
		RDBOOL	 addsec = true;

		if (sld->offsethigh == 0xFFFF || sld->offsetlow == 0xFFFF)
		{
			PrintProgInfo("%s%i side %i's height exceeds %i\n", err, secnum, sld->sidenum, 0x7FFF);
			ma_inf.slopeerror++;
			continue;
		}

		if (secnum < 0 || secnum >= numsectors)
		{
			M_Free(typemarker);
			PrintProgError("Program error [8]: side %i\n", i);
		}

		if (s->linecount != 3 && s->linecount != 4)
		{
			PrintProgInfo("%s%i must be a triangle or a trapezium\n", err, secnum);
			ma_inf.slopeerror++;
			continue;
		}

		if (sld->type[1] == 1 && sld->offsethigh == 0xFFFE)
		{
			sector_t *other = NULL;
			line_t	 *chk;

			for (k = 0; k < s->linecount; k++)
			{
				chk = s->lines[k];

				if (chk->sidenum[0] == sld->sidenum) {other = chk->backsector ; break; }
				if (chk->sidenum[1] == sld->sidenum) {other = chk->frontsector; break; }
			}

			if (!other)
			{
				PrintProgInfo("%s%i line %i has an auto lower height of 0 with a one-sided line!\n", err, secnum, chk - lines);
				ma_inf.slopeerror++;
				continue;
			}

			height = (s->ceilingheight - other->ceilingheight) >> FRACBITS;

			if (height <= 0)
			{
				PrintProgInfo("%s%i line %i's other sector must have a lower ceiling height where using an autoheight setting of 0\n", err, secnum, chk - lines);
				ma_inf.slopeerror++;
				continue;
			}

			sld->offsethigh = height;
		}

		if (sld->type[0] == 1 && sld->offsetlow == 0xFFFE)
		{
			sector_t *other = NULL;
			line_t	 *chk;

			for (k = 0; k < s->linecount; k++)
			{
				chk = s->lines[k];

				if (chk->sidenum[0] == sld->sidenum) {other = chk->backsector ; break; }
				if (chk->sidenum[1] == sld->sidenum) {other = chk->frontsector; break; }
			}

			if (!other)
			{
				PrintProgInfo("%s%i line %i has an auto raise height of 0 with a one-sided line!\n", err, secnum, chk - lines);
				ma_inf.slopeerror++;
				continue;
			}

			height = (other->floorheight - s->floorheight) >> FRACBITS;

			if (height <= 0)
			{
				PrintProgInfo("%s%i line %i's other sector must have a higher floor height where using an autoheight setting of 0\n", err, secnum, chk - lines);
				ma_inf.slopeerror++;
				continue;
			}

			sld->offsetlow = height;
		}

		for (k = 0; k < csecptr; k++)
		{
			if (countsec[k] == secnum)
			{
				int	errnum = 0;

				if (sld->type[0] && typemarker[k] & 4) errnum = 1;
				if (sld->type[1] && typemarker[k] & 8) errnum = 2;

				if (errnum)
				{
					PrintProgInfo("%s%i has more than one %s definition\n", err, secnum, errnum == 1 ? "low" : "high");

					ma_inf.slopeerror++;
				}

				if (sld->type[0] && typemarker[csecptr] & 1) typemarker[csecptr] |= 4;
				if (sld->type[1] && typemarker[csecptr] & 2) typemarker[csecptr] |= 8;

				addsec = false;
				break;
			}
		}

		if (!addsec) continue;

		if (sld->type[0]) typemarker[csecptr] |= 1;
		if (sld->type[1]) typemarker[csecptr] |= 2;

		countsec[csecptr++] = secnum;
	}

	M_Free(typemarker);

	// Set auto $A heights
	for (i = 0, sld = slsidarray; i < numsidarray; i++, sld++)
	{
		sector_t *s = sides[sld->sidenum].sector;
		int height;

		if (sld->type[1] == 0 && sld->offsethigh == 0xFFFE)
		{
			if ( (height = getAdefHeight(s, csecptr, sld->sidenum, 1)) == 0)
			{
				PrintProgInfo("%s%i $A auto height can find no adjacent $B defined sloped ceiling sector\n", err, s - sectors);

				ma_inf.slopeerror++;
				continue;
			}
			sld->offsethigh = height;
		}

		if (sld->type[0] == 0 && sld->offsetlow == 0xFFFE)
		{
			if ( (height = getAdefHeight(s, csecptr, sld->sidenum, 0)) == 0)
			{
				PrintProgInfo("%s%i $A auto height can find no adjacent $B defined sloped floor sector\n", err, s - sectors);

				ma_inf.slopeerror++;
				continue;
			}
			sld->offsetlow = height;
		}
	}

	ma_inf.slopecount = csecptr;

	for (i = 0; i < csecptr; i++)
	{
		sector_t	*sec = &sectors[countsec[i]];
		
		subfn_SetSlopeVtxHeights(sec, LOSLOPE, analyse);
		subfn_SetSlopeVtxHeights(sec, HISLOPE, analyse);
	}

	// Set triwall line flags
	for (i = 0, sg = segs; i < numsegs; i++, sg++)
	{
		if (sg->linedef && origTEX[sg->linedef - lines] & 0x22)
		{
			if (ssecvtxz[sg->zptr1].z[0] != ssecvtxz[sg->zptr2].z[0])
			{
				sg->linedef->lnflags2 |= LNF2_LOWTRIWALL;
			}
			if (ssecvtxz[sg->zptr1].z[1] != ssecvtxz[sg->zptr2].z[1])
			{
				sg->linedef->lnflags2 |= LNF2_HITRIWALL;
			}
		}
	}

	// Check if texture swap required
	for (i = 0; i < numsides; i++)
	{
		if (sides[i].lsiflags & LS_LOSLOPEB) swapTextures(i, 0);
		if (sides[i].lsiflags & LS_HISLOPEB) swapTextures(i, 1);
	}

	if (!analyse)
	{
		M_Free(ssecarray); ssecarray = NULL;
		M_Free(hitssec)  ; hitssec   = NULL;
		M_Free(linearray); linearray = NULL;
		M_Free(countsec) ; countsec  = NULL;
	}

	if (ma_inf.slopeerror) Sys_Quit(false);

	ResetProgHeader();
}

// Set up flags following analysis
void LS_SetSectorSlopeFlags()
{
	int i;

	// Set up any midpoint z heights required
	for (i = 0; i < numsubsectors; i++)
	{
		if (subsectors[i].sector->specFlags2 & (SF2_HISLOPE|SF2_LOSLOPE))
		{
			subsector_t	*ss = &subsectors[i];
			int			flags = ss->sector->specFlags2;
			int			x = (int)(65536.0f*ss->midpoint.x);
			int			y = (int)(65536.0f*ss->midpoint.y);
			float		z;

			if (flags & SF2_LOSLOPE)
			{
				z = FIX2FLT(pointOffset(ss->sector->stptr, x, y, LOSLOPE));
				SetZheightPtr(&ss->midpoint, z, LOSLOPE);
			}

			if (flags & SF2_HISLOPE)
			{
				z = FIX2FLT(pointOffset(ss->sector->stptr, x, y, HISLOPE));
				SetZheightPtr(&ss->midpoint, -z, HISLOPE);
			}
		}
	}
}

void LS_CheckSlopeAnalysis()
{
	line_t	*ld;
	int		i;

	if (!R3D_maptype) return;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->ml_flags & ML_TWOSIDED
			&& ld->frontsector != ld->backsector
			&& !(lineflags[i] & LFG_3DTYPE) )		// v220-8
		{
			if (ld->lnflags2 & (LNF2_LOWTRIWALL|LNF2_HITRIWALL))
			{
				ld->specType = R3D_00;
			}
			else if (!origTEX[i]
				&& ld->lnflags2 & LNF2_SLOPESECLN)
			{
				ld->specType = R3D_01;
			}
		}
	}
}
