//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2013 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_traceGen.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
//
// DESCRIPTION: a general trace routine
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdlib.h>
#include <string.h>
#include <search.h>

#include "con_bar.h"
#include "r3d_MAINspec.h"
#include "r3d_trace.h"

#include "m_misc.h"

// MACROS ------------------------------------------------------------------

#define NO_CONDITION			0
#define ELIMINATE_SINGLELINE	1
#define DONT_RESET_SSECARRAY	2

// TYPES -------------------------------------------------------------------

typedef struct
{
	fixed_t x, y, dx, dy;
} divline_t;

divline_t	dref, dvn;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

int P_DivlineSide(fixed_t x, fixed_t y, const divline_t *node);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int		ssec_midx, ssec_midy;
int		hitssecptr;

char	*tg_markline;
char	*tg_marklinemem;
char	*tg_specline;
char	*tg_linedone;
byte	*tg_ssecflags;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static int trace_count;

static int	larray_count;

static seg_t	*rcv_sg;
static vertex_t *rcv_v1, *rcv_start;

static int secnum;
static int line_count;
static int abox[4];

static int fill_secnum;
static int success;

static int base_secnum;
static int genfill_count;

static int ref_bbox[4];
static vertex_t ref_v1, ref_v2;

// For debug
static int start_boundsline;
static int global_callnum;
static int global_print;
static int testn = 335;

// CODE --------------------------------------------------------------------

static void refLine(linev_t *lv, int index)
{
	ref_v1.x = ssec_midx;
	ref_v1.y = ssec_midy;
	ref_v2.x = lv[index].midx;
	ref_v2.y = lv[index].midy;

	dref.x = ref_v2.x;
	dref.y = ref_v2.y;
	dref.dx = ref_v2.x - ref_v1.x;
	dref.dy = ref_v2.y - ref_v1.y;

	ref_bbox[BOXLEFT ] = MIN_OF(ref_v1.x, ref_v2.x);
	ref_bbox[BOXRIGHT] = MAX_OF(ref_v1.x, ref_v2.x);
	ref_bbox[BOXBASE ] = MIN_OF(ref_v1.y, ref_v2.y);
	ref_bbox[BOXTOP  ] = MAX_OF(ref_v1.y, ref_v2.y);
}

static RDBOOL checkTestLineIntersect(linev_t *lv, int lnum)
{
	vertex_t	*v1, *v2;
	line_t		*line = &lines[lnum];

	if (   line->bbox[BOXLEFT ] > ref_bbox[BOXRIGHT]
		|| line->bbox[BOXRIGHT] < ref_bbox[BOXLEFT ]
		|| line->bbox[BOXBASE ] > ref_bbox[BOXTOP  ]
		|| line->bbox[BOXTOP  ] < ref_bbox[BOXBASE ])
	{
		return false; // quick bounds box rejection
	}

	v1 = line->v1;
	v2 = line->v2;

	if (P_DivlineSide(v1->x, v1->y, &dref) ==
		P_DivlineSide(v2->x, v2->y, &dref))
	{
		return false;	// line isn't crossed
	}
	
	dvn.dx = v2->x - (dvn.x = v1->x);
	dvn.dy = v2->y - (dvn.y = v1->y);
	
	if (P_DivlineSide(ref_v1.x, ref_v1.y, &dvn) ==
		P_DivlineSide(ref_v2.x, ref_v2.y, &dvn))
	{
		return false;	// line isn't crossed
	}

	return true;
}

static float subfn_SegRelToMidpoint(float v1[2], float v2[2], float mx, float my)
{
	// The dot product.
	return	(v1[VY] - v2[VY])*(v1[VX] - mx) + (v2[VX] - v1[VX])*(v1[VY] - my);
}

static int traceSecFacing(linev_t *lv, int index, float mx, float my)
{
	float res = subfn_SegRelToMidpoint(lv[index].v1, lv[index].v2, mx, my);

	if (res > 0) return lv[index].fsec;
	if (res < 0) return lv[index].bsec;

	return -2;	// Can't be determined
}

// v220-22 rewritten
RDBOOL subfn_LiesInTrueArea(linev_t	*lv, int lcount, int refsec, float mx, float my)
{
	int		i, k, secn;

	for (i = 0; i < lcount; i++)
	{
		int	valid = true;

		if (!lv[i].mark) continue;

		lv[i].test = 0;

		secn = traceSecFacing(lv, i, mx, my);

		if (refsec != secn) continue;

		// OK. Check that a line cast from the ssec's midpoint to the
		// test line's midpoint does not cross another boundary line.
		// If it does then the ssec is outside the area.

		refLine(lv, i);
		lv[i].test = 1;
		
		for (k = 0; k < lcount; k++)
		{
			if (lv[k].mark && i != k)
			{
				if (checkTestLineIntersect(&lv[i], linearray[k]))
				{
					valid = false;
					break;
				}
			}
		}

		// does not intersect another boundary line
		if (valid) 	return true;
	}

	return false;
}

static void subfn_TestSubsectors(linev_t *lv, int type)
{
	subsector_t *ss;
	int i, k, ssec_valid = true;

	// Mark lines of interest i.e. all boundary
	// lines contained in the test area.
	for (i = 0; i < larray_count; i++)
	{
		int k = linearray[i];

		lv[i].mark = (tg_markline[k] == 1);

		if (lv[i].mark || type == FGA_LFSEC)
			continue;

		if (lv[i].bsec >= 0)
		{
			if (lv[i].bsec != secnum &&
				lv[i].fsec != secnum)
				continue;
		}
		else if (lv[i].fsec != secnum)
			continue;

		if (lines[k].bbox[BOXLEFT ] >= abox[BOXLEFT ] &&
			lines[k].bbox[BOXRIGHT] <= abox[BOXRIGHT] &&
			lines[k].bbox[BOXBASE ] >= abox[BOXBASE ] &&
			lines[k].bbox[BOXTOP  ] <= abox[BOXTOP  ])
		{
			lv[i].mark = 1;
		}
	}

	// We need to know if the line's sectors are on the
	// inside of the bound area or the outside.
	// Pick a point outside the box limits.  If it thinks
	// it's inside then the sector's on the outside of the box.

	if (abox[BOXLEFT ] > -32758 * FRACUNIT)
		ssec_midy = ssec_midx = abox[BOXLEFT ] - 10*FRACUNIT;
	else
	if (abox[BOXRIGHT] <  32757 * FRACUNIT)
		ssec_midy = ssec_midx = abox[BOXRIGHT] + 10*FRACUNIT;
	else
	if (abox[BOXBASE] > -32758 * FRACUNIT)
		ssec_midy = ssec_midx = abox[BOXBASE ] - 10*FRACUNIT;
	else
	if (abox[BOXTOP ] <  32757 * FRACUNIT)
		ssec_midy = ssec_midx = abox[BOXTOP  ] + 10*FRACUNIT;
	else
		ssec_valid = false;

	if (ssec_valid
		&& subfn_LiesInTrueArea(lv, larray_count, secnum, FIX2FLT(ssec_midy), FIX2FLT(ssec_midx)))
	{
		success = false;	// thinks it's in the box when outside
		return;
	}

	if (type == MSIA_TESTAREA)
	{
		success = true; return;
	}

	// Predone rather than in the following loop - v2.1.00 build_35
	for(i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		if (!ssecarray[i])
		{
			ssec_midx = (int)(ss->midpoint.x * FRACUNIT);
			ssec_midy = (int)(ss->midpoint.y * FRACUNIT);

			if (ssec_midx <= abox[BOXLEFT ] ||	// Quick bounds test for rejection
				ssec_midx >= abox[BOXRIGHT] ||
				ssec_midy <= abox[BOXBASE ] ||
				ssec_midy >= abox[BOXTOP  ] ||
				!subfn_LiesInTrueArea(lv, larray_count, secnum, ss->midpoint.x, ss->midpoint.y))
			{
				ssecarray[i] = 1;
			}
		}
	}

	for(i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		if (!ssecarray[i])
		{
			switch(type)
			{
			case FGA_SECTOR:
				SetOverride(ss, fill_secnum, global_callnum);
				break;
			case FGA_LFSEC:
				Setr3dsslfsec(ss, fill_secnum, 19019);
				break;
			case FGA_SSFLOORSEC:
				if (ss->floorsec < 0)
					Setr3dssfs(ss, fill_secnum, 19026);
				break;
			case FGA_SSCEILSEC:
				if (ss->ceilsec >= 0)
				{
					if (   sectors[ss->ceilsec].ceilingheight
						!= sectors[fill_secnum].ceilingheight)
					{
						Setr3dsslcsec(ss, ss->ceilsec, 19022);
						Setr3dsscs(ss, fill_secnum, 19023);
					}
				}
				else if (ss->sector->ceilingheight
					> sectors[fill_secnum].ceilingheight)
				{
					Setr3dsscs(ss, fill_secnum, 19024);
				}
				break;
			case FGA_SSOVRSEC:
				SetOverride(ss, fill_secnum, 19025);
				if (ss->sector) hitssec[hitssecptr++] = i;
				break;
			default:
				break;
			}
			ssecarray[i] = 1;
		}
	}

	// Now check if tg_linedone == 2
	for(i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		if (ss->floorsec < 0)
		{
			seg_t	*sg = &segs[ss->firstline];
			RDBOOL	found = false;

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if (sg->linedef != NULL &&
					tg_linedone[sg->linedef - lines] == 2)
				{
					Setr3dssfs(ss, secnum, 19027);
					found = true;
					break;
				}
			}

			if (found)
			{
				sector_t *s = &sectors[secnum];
				
				sg = &segs[ss->firstline];

				for (k = 0; k < ss->linecount; k++, sg++)
				{
					if (sg->linedef
						&& sg->linedef->frontsector == s
						&& sg->linedef->backsector == s)
					{
						Setr3dsscs(ss, secnum, 19028);
						break;
					}
				}
			}
		}
	}
}

// Checks to see if ring has already been tested
// (problem caused by potentially shared lines).
static RDBOOL RingCheck()
{
	int i;
	
	int lcount = 0, mmcount = 0;	// for ring tested check

	for (i = 0; i < larray_count; i++)
	{
		int lnum = linearray[i];

		if (tg_markline[lnum] == 1)
		{
			if (tg_marklinemem[lnum]) mmcount++;
			tg_marklinemem[lnum] = 1;
			lcount++;	
		}
	}

	return (lcount != mmcount);	// If false ring already tested
}

static void subfn_MarkLinesTestedGen()
{
	int i;

	for (i = 0; i < larray_count; i++)
	{
		int k = linearray[i];

		if (tg_markline[k] == 1)
		{
			tg_markline[k] = 4;	// Assume unique
			
			if (lines[k].ml_flags & ML_TWOSIDED
				&& !tg_specline[k]
				&& !hitline[k]
				&& lines[k].frontsector != lines[k].backsector
				&& IsLiquidFloorFlat(lines[k].frontsector)
				&& IsLiquidFloorFlat(lines[k].backsector))
			{
				tg_markline[k] = 0;	// Possible shared line
			}
		}
	}
}

static int __cdecl lineArraySorter(const void *e1, const void *e2)
{
	int lc1 = lines[*(int*) e1].frontsector->linecount;
	int lc2 = lines[*(int*) e2].frontsector->linecount;
	if (lc1 > lc2) return -1;
	if (lc1 < lc2) return  1;
	return 0;
}

static void subfn_SetLArrayVars(linev_t *lv)
{
	line_t	*ld;
	int		i;

	for (i = 0; i < larray_count; i++)
	{
		ld = &lines[linearray[i]];

		lv[i].v1[VX] = FIX2FLT(ld->v1->x);
		lv[i].v1[VY] = FIX2FLT(ld->v1->y);
		lv[i].v2[VX] = FIX2FLT(ld->v2->x);
		lv[i].v2[VY] = FIX2FLT(ld->v2->y);

		lv[i].fsec = ld->frontsector - sectors;
		lv[i].bsec = ld->backsector ? ld->backsector - sectors : -1;
		lv[i].midx = meanV(ld->v1->x, ld->v2->x);
		lv[i].midy = meanV(ld->v1->y, ld->v2->y);
	}
}

static line_t	*rcv_li;

static RDBOOL subfn_TraceBoundary(int lnum)
{
	sector_t *s = &sectors[secnum];
	vertex_t *v    = lines[lnum].v1;
	vertex_t *vchk = lines[lnum].v2;
	int isclosed = false, repeat = true;
	int i, n;

	// -----------------------

	tg_markline[lnum] = 1;
	trace_count = 1;
	
	M_ClearBox(abox);
	M_AddLineToBox(abox, rcv_li);

	start_boundsline = lnum;

	while (repeat)
	{
		repeat = false;
		for (n = 0; n < larray_count; n++)
		{
			i = linearray[n];
			rcv_li = &lines[i];

			if (!(rcv_li->frontsector == s
				  || rcv_li->backsector == s))
				  continue;

			if (!tg_markline[i])
			{
				if (lines[i].v1 == v)
				{
					v = rcv_li->v2;
					M_AddLineToBox(abox, rcv_li);
					tg_markline[i] = 1;
					trace_count++;
					repeat = true;
					break;
				}
				else if (lines[i].v2 == v)
				{
					v = rcv_li->v1;
					M_AddLineToBox(abox, rcv_li);
					tg_markline[i] = 1;
					trace_count++;
					repeat = true;
					break;
				}
			}
		}
	}

	if (v == vchk)
	{
		isclosed = RingCheck();
	}
	else
	{
		for (n = 0; n < larray_count; n++)
		{
			if (tg_markline[linearray[n]] == 1)
				tg_markline[linearray[n]] = 0;
		}
	}

	return isclosed;
}

static void subfn_Trace2(int type)
{
	linev_t	*lv;
	int		i;

	if (larray_count <= 2) return;

	lv = (linev_t*)M_Calloc(larray_count*sizeof(linev_t), 299);

	// Put linedef's with a sector with highest
	// linecounts at start of the list.
	qsort(linearray, larray_count, sizeof(int), lineArraySorter);

	subfn_SetLArrayVars(lv);

	for (i = 0; i < larray_count; i++)
	{
		int k = linearray[i];

		Con_Progress(i, PBARF_PERCENT); // $progress XI

		if (tg_markline[k]) continue;

		rcv_li = &lines[k];

		switch(type)
		{
		case FGA_SECTOR:
			secnum = fill_secnum;
			break;
		case FGA_LFSEC:
		case FGA_SSCEILSEC:
		case FGA_SSFLOORSEC:
		case FGA_SSOVRSEC:
			secnum = base_secnum;
			break;
		default:
			break;
		}

		if (subfn_TraceBoundary(k))
		{
			success = true;
			subfn_TestSubsectors(lv, type);	// Mark all subsectors in this group
			subfn_MarkLinesTestedGen();	// Mark off lines used
		}
	}

	M_Free(lv);
}

static void subfn_SetArrays(byte *userArray, RDBOOL condition)
{
	int i, k;

	CLR_hitline();

	memset(tg_linedone		, 0, numlines);
	memset(tg_markline		, 0, numlines);
	memset(tg_marklinemem	, 0, numlines);
	memset(tg_specline		, 0, numlines);
	memset(tg_ssecflags	, 0, numsubsectors);

	SET_linearray(0);
	larray_count = 0;

	if (condition == DONT_RESET_SSECARRAY) return;

	hitssecptr = 0;

	for (i = 0; i < numsubsectors; i++)
	{
		ssecarray[i] = 0;

		if (condition == ELIMINATE_SINGLELINE)
		{
			subsector_t *ss = &subsectors[i];
			int	count = 0;

			rcv_sg = &segs[ss->firstline];
			CLR_hitline();

			// Eliminate ssecs with only one sg->linedef ref
			for (k = 0; k < ss->linecount; k++, rcv_sg++)
			{
				if (rcv_sg->linedef)
				{
					// Is it a boundary line?
					if (userArray[rcv_sg->linedef - lines])
					{
						count = 2;	// This must be allowed
						break;
					}
					count++;
					if (count > 1) break;
				}
			}
			if (count < 2)
				ssecarray[i] = 1;
		}
	}
}

static RDBOOL hasTEXzero(sector_t *sec)
{
	int i;

	for (i = 0; i < sec->linecount; i++)
	{
		if (sec->lines[i]->ml_flags & ML_TWOSIDED)
		{
			if (origTEX[sec->lines[i] - lines] & 0x01
				&& !(sides[sec->lines[i]->sidenum[0]].bottomtexture))
				return true;
			if (origTEX[sec->lines[i] - lines] & 0x10
				&& !(sides[sec->lines[i]->sidenum[1]].bottomtexture))
				return true;
		}
	}
	return false;
}

// This is only called from subfn_SecArea() in r3d_secarbiter
// conarray must have been preset
RDBOOL LS_SetSectorArea(sector_t *s, int fillnum, int callnum)
{
	int i;

	global_callnum = callnum;
	global_print = false;

	subfn_SetArrays(conarray, NO_CONDITION);

	fill_secnum = fillnum;
	success = false;

	for (i = 0; i < s->linecount; i++)
	{
		int k = s->lines[i] - lines;
		if (conarray[k]) // Set by calling function
		{
			linearray[larray_count++] = k;
			tg_markline[k] = 0;
		}
		else
			tg_markline[k] = -1;
	}
	subfn_Trace2(FGA_SECTOR);
	return success;
}

// This routine ignores any enclosed boundary lines.
// It will set values determined by <type> for all subsecs
// contained in each loop of lines marked in conarray which
// must, therefore, be set prior to calling the function.

// Type FGA_LFSEC		sets lfsec,
//      FGA_SSFLOORSEC	sets ss->floorsec
//		FGA_SSCEILSEC	sets ss->ceilsec,
//		FGA_SSOVRSEC	sets ss->floorsec/ss->ceilsec,
RDBOOL LS_FillGenArea(sector_t *s, int fillnum, int type, int callnum)
{
	int i, cond =  type != FGA_SSCEILSEC;	// rev56 (was true)

	if ((unsigned)type > FGA_SSOVRSEC)
		Con_Error("Program error [10]: LS_FillGenArea from %i", callnum);

	global_callnum = callnum;
	global_print = false;

	subfn_SetArrays(conarray, cond ? ELIMINATE_SINGLELINE : NO_CONDITION);
	fill_secnum = fillnum;
	base_secnum = s - sectors;

	success = false;
	genfill_count = 0;

	// First pass traces all loops
	for (i = 0; i < s->linecount; i++)
	{
		int lnum = s->lines[i] - lines;

		if (conarray[lnum]) // Set by calling function
		{
			linearray[larray_count++] = lnum;
			tg_markline[lnum] = 0;
		}
		else
			tg_markline[lnum] = -1;
	}
	subfn_Trace2(type);
	return success;
}
