//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2013 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_traceLim.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 17 (v220-17) 27/Oct/2012 gmj
// Revision 2.2.0 build 18 (v220-18) 05/Jan/2013 gmj
// Revision 2.2.0 build 20 (v220-20) 31/Mar/2013 gmj
//
// DESCRIPTION: finds all separated groups of subsectors
//
// NOTE: In certain rare cases a subsector's lines are only owned by
//       that particular subsector. This means it will never get
//       assigned to a viewgroup. For this reason viewgroup 0 is always
//       rendered rather than trying through yet more code to establish
//       what viewgroup actually owns it.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------
#include <stdlib.h>
#include <string.h>
#include <search.h>

#include "con_bar.h"

#include "r3d_MAINspec.h"
#include "r3d_compat.h"
#include "r3d_edit.h"
#include "m_misc.h"
#include "p_maputil.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

int		playerGNUM(int num);
int		getTPgnum(int secidx);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static int trace_count, ssec_count, group_count;

static int larray_count;

// Used by subfn_IterateSubsectors - locals not
// used to prevent unnecessary memory loading.
static subsector_t *rcv_ss;

// Used by subfn_TraceBoundary
static line_t	*rcv_li;
static seg_t	*rcv_sg;
static vertex_t *rcv_v1;
static int		rcv_k;
static int		lgroup_num;
static int		lgroup_max;

// Used by both subfn_IterateSubsectors and subfn_TraceBoundary
static int rcv_i;

// Used to store line numbers of faulty lines
static int badlinemaxcount = 0;
static int *badlinemem = NULL;

// CODE --------------------------------------------------------------------

static void getLinesInGroup(int vg, byte *pgn)
{
	subsector_t	*ss;
	line_t		*ld;
	seg_t		*sg;
	int			i, k, n, t;
	int			vgp;

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		if (ss->viewgroup != vg) continue;

		sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if ((ld = sg->linedef) && ld->validcount != validcount)
			{
				ld->validcount = validcount;

				if (!ld->special) continue;
				
				t = ld->tag;	// v220-20 tag 0 allowed for TELEPORTMAN checks

				switch(ld->special)
				{
				default: break;

				// Check for player teleports to TELEPORTMAN
				case 39:  case 97:  case 174: case 195:
				case 207: case 208: case 209: case 210:
					for (n = -1; (n = P_FindSectorFromTag(t, n)) >= 0;)
					{
						if ( (vgp = getTPgnum(n)) >= 0) pgn[vgp] = true;
					}
					break;

				// Check for player line to line teleporters
				case 243: case 244: case 262: case 263:
					if (t)
					{
						for (n = -1; (n = P_FindLineFromLineTag(&lines[ld - lines], n)) >= 0;)
						{
							if ( (vgp = subsectors[sslines[n].list[0]].viewgroup) >= 0)
								pgn[vgp] = true;
						}
					}
					else
					{
						pr_OFinsert("MAP ERROR: L%i sp %i has no tag value set\n", ld - lines, ld->special);
						ma_inf.maperror++;
					}
					break;
				}
			}
		}
	}
}

// v220-18 (rewritten) mark any subsectors the player will never see
static void subfn_SetControlLines()
{
	thinker_t	*th;
	int			i, k, vgp;
	int			repeat = true;
	byte		*pgn, *pgp;
	int			pgc;

	//---------------------------------------------
	
	pgn = (byte*)calloc(lgroup_max, 1);
	pgp = (byte*)calloc(lgroup_max, 1);

	// v220-9 Add ssecs groups with TPUNDER/OVERFAKE things
	for (th = thinkercap.next ; th != &thinkercap ; th = th->next)
	{
		mobj_t *iter = (mobj_t*)th;

		if (   iter->thinker.function == (void (__cdecl *)(void))P_MobjThinker
			&& (iter->type == MT_TPUNDERFAKE || iter->type == MT_TPOVERFAKE))
		{
			vgp = iter->subsector->viewgroup;
			pgn[vgp] = true;
		}
	}

	// v220-20 add scripted dual spot lines
	if (R3D_ScriptedMap)
	{
		line_t	*ld;

		for (i = 0; i < NUM7400SPOTS; i++)
		{
			if ((ld = th7400[i].refline) && ld->tag && ld->special)
			{
				int n, t = ld->tag;

				switch(ld->special)
				{
				default: break;

				// Check for player teleports to TELEPORTMAN
				case 39:  case 97:  case 174: case 195:
				case 207: case 208: case 209: case 210:
					for (n = -1; (n = P_FindSectorFromTag(t, n)) >= 0;)
					{
						if ( (vgp = getTPgnum(n)) >= 0) pgn[vgp] = true;
					}
					break;

				// Check for player line to line teleporters
				case 243: case 244: case 262: case 263:
					for (n = -1; (n = P_FindLineFromLineTag(&lines[ld - lines], n)) >= 0;)
					{
						if ( (vgp = subsectors[sslines[n].list[0]].viewgroup) >= 0)
							pgn[vgp] = true;
					}
					break;
				}
			}
		}
	}

	for (i = 0; i < dmcount; i++)
	{
		subsector_t *ss = R_PointInSubsector(dmarray[i].x << FRACBITS, dmarray[i].y << FRACBITS);

		vgp = ss->viewgroup;
		pgn[vgp] = true;
	}

	vgp = playerGNUM(0);
	pgp[vgp] = true;

	if (vgp < 0) Con_ErrorD("MAP is unuseable - check player start area");

	pr_OFinsert("\nSubsec groups: pvg %i", vgp);

	validcount++;

	while(repeat)
	{
		repeat = false;

		pgp[vgp] = true;
		getLinesInGroup(vgp, pgn);

		for (i = 0; i < lgroup_max; i++)
		{
			if (pgn[i] && !pgp[i])
			{
				vgp = i;
				repeat = true;
				break;
			}
		}
	}

	for (i = 0, pgc = 0; i < lgroup_max; i++)
	{
		if (pgp[i]) pgc++;
	}

	pr_OFinsert(" [%i:%i]\n", pgc, lgroup_max - 1);

	for (i = 0; i < numsubsectors; i++)
	{
		if ( ((vgp = subsectors[i].viewgroup) && !pgp[vgp]) )
		{
			subsector_t *ss = &subsectors[i];
			seg_t		*sg = &segs[ss->firstline];
			line_t		*ld;
			
			if (!R3D_edit) ss->ssflags |= SSF_NORENDER;

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if ( (ld = sg->linedef) )
				{
					if ( !(ld->spflags & SPF_CONTROLLINE) )
					{
						SetControlline(ld, 3012);
					}

					secspflags[ld->frontsector - sectors] |= SECSPF_CTRLLINES;

					if (ld->backsector)
					{
						secspflags[ld->backsector - sectors] |= SECSPF_CTRLLINES;
					}
				}
			}
		}
	}

	M_Free(pgn);
	M_Free(pgp);
}

// --------------------------------------------------------------------

static RDBOOL subfn_IterateBoundary()
{
	rcv_v1 = rcv_li->v1;

	countsec[rcv_li->frontsector - sectors] = 1;
	hitline[rcv_li - lines] = 1;
	trace_count++;

	for (rcv_i = 0; rcv_i < larray_count; rcv_i++)
	{
		if (!hitline[linearray[rcv_i]])
		{
			rcv_li = &lines[linearray[rcv_i]];
			if (rcv_li->v2->x == rcv_v1->x &&
				rcv_li->v2->y == rcv_v1->y)
				return true;
		}
	}
	return false;
}

static RDBOOL subfn_IterateSsecHelper()
{
	seg_t *seg = &segs[rcv_ss->firstline];
	int i;
	RDBOOL valid = false;

	for (i = 0; i < rcv_ss->linecount; i++, seg++)
	{
		if (seg->linedef != NULL &&
			!(hitline[seg->linedef - lines] & 3))
		{
			countsec[seg->frontsector - sectors] = 1;

			if (!hitline[seg->linedef - lines])	// Zero - so no backsector
			{
				rcv_li = seg->linedef;
				while (subfn_IterateBoundary());
			}
			else
			{
				if (seg->backsector)
					countsec[seg->backsector - sectors] = 1;
				hitline[seg->linedef - lines] = 1;
			}
			valid = true;	// Don't break - ensure all lines are tested
		}
	}

	return valid;
}

static RDBOOL subfn_IterateSubsectors()
{
	for(rcv_i = 0; rcv_i < numsubsectors; rcv_i++)
	{
		rcv_ss = &subsectors[rcv_i];

		if (!ssecarray[rcv_i] && countsec[rcv_ss->sector - sectors])
		{
			rcv_sg = &segs[rcv_ss->firstline];

			for (rcv_k = 0; rcv_k < rcv_ss->linecount; rcv_k++, rcv_sg++)
			{
				if (rcv_sg->linedef != NULL &&
					hitline[rcv_sg->linedef - lines] == 1)
				{
					rcv_ss->viewgroup = group_count;
					ssecarray[rcv_i] = 1;
					ssec_count++;
					if (subfn_IterateSsecHelper())
						return true;
				}
			}
		}
	}
	return false;
}

static void subfn_MarkLinesTestedLim()
{
	int		i;

	for (i = 0; i < numlines; i++)
	{
		if (hitline[i] == 1) hitline[i] = 2;
	}
}

static int lineChecker(int *line_error)
{
	int		i, count = 0, badhdr = 0;

	//------------------------------------

	CLR_hitline();
	CLR_vtxcheck();

	// v220-10 ignore one's lines with disconnected end
	for (i = 0; i < numlines; i++)
	{
		if (lines[i].backsector == NULL)
		{
			int		vn1 = lines[i].v1 - vertexes;
			int		vn2 = lines[i].v2 - vertexes;
			byte	vnum1 = vtxarray[vn1];
			byte	vnum2 = vtxarray[vn2];

			if (   vtxarray[vn1] == 1 && vertexowners[vn1].num == 1
				&& vtxarray[vn2] == 1 && vertexowners[vn2].num == 1)
			{
				hitline[i] = 4;
			}
			else
			if (vtxarray[vn1] == 1 && vertexowners[vn1].num == 1)
			{
				if (vtxarray[vn2] > 1) vtxarray[vn2]--;
				hitline[i] = 4;
			}
			else
			if (vtxarray[vn2] == 1 && vertexowners[vn2].num == 1)
			{
				if (vtxarray[vn1] > 1) vtxarray[vn1]--;
				hitline[i] = 4;
			}

			if (hitline[i])
			{
				if (ma_inf.badsslines >= badlinemaxcount)
				{
					badlinemaxcount += 64;
					badlinemem = (int*)W_Realloc(badlinemem,
						badlinemaxcount * sizeof(int), STATICID, 0);
				}

				badlinemem[ma_inf.badsslines++] = i;

				if (R3D_edit)
				{
					if (!badhdr) pr_OFdirect("\n Bad 1's lines [unconnected]:\n");
					pr_OFinsert("%7i", i);
					if (!(++badhdr % 8)) pr_OFdirect("\n");
				}
			}
		}
	}

	if (badhdr) pr_OFdirect("\n");
	if (badhdr % 8) pr_OFdirect("\n");

	for (i = 0; i < numlines; i++)
	{
		if (lines[i].backsector == NULL)
		{
			vtxcheck[lines[i].v1 - vertexes]++;
			vtxcheck[lines[i].v2 - vertexes]++;
		}
	}

	// Set larray_count
	larray_count = 0;

	for (i = 0; i < numlines; i++)
	{
		if (!hitline[i])
		{
			if (lines[i].ml_flags & ML_TWOSIDED)
			{
				hitline[i] = 4;
			}
			else
			{
				linearray[larray_count++] = i;
			}
		}
	}

	return count;
}

static int max_group;

static void mergeGroups(int gn1, int gn2)
{
	int gnmax = MAX_OF(gn1, gn2);
	int gnmin = MIN_OF(gn1, gn2);
	int i;

	max_group = gnmax;

	pr_OFinsert("Merging groups %i %i\n",gn1, gn2);

	for (i = 0; i < numsubsectors; i++)
	{
		if (subsectors[i].viewgroup == gnmax)
			subsectors[i].viewgroup = gnmin;
	}

	for (i = 0; i < numsubsectors; i++)
	{
		if (subsectors[i].viewgroup > gnmax)
		{
			subsectors[i].viewgroup--;

			if (subsectors[i].viewgroup > max_group)
				max_group = subsectors[i].viewgroup;
		}
	}

	lgroup_num--;
}

static int __cdecl traceArraySorter(const void *e1, const void *e2)
{
	int lc1 = lines[*(int*) e1].frontsector->linecount;
	int lc2 = lines[*(int*) e2].frontsector->linecount;
	if (lc1 > lc2) return -1;
	if (lc1 < lc2) return  1;
	return 0;
}

void TL_TraceBoundary()
{
	int		i, k, linecheck, secnum;
	int		count = 0, barpos = 5;
	RDBOOL	line_error = false;

	group_count = 1;
	lgroup_max = 1;

	SET_linearray(0);
	
	Con_TimingStart();
	
	CLR_ssecarray();

	linecheck = lineChecker(&line_error);

	// b211-0 Added
	if (ma_inf.badsslines > 1)
	{
		RDBOOL endret = false;

		for (i = 0; i < ma_inf.badsslines; i++)
		{
			line_t	*ld = &lines[badlinemem[i]];

			if (!(ld->ml_flags & ML_BADLINE))
			{
				line_t	*lc = subfn_FindCoincidentLine(ld, 4);

				if (lc)
				{
					if (M_AccurateLength(lc) > M_AccurateLength(ld))
						ld->ml_flags |= ML_BADLINE;
					else
						lc->ml_flags |= ML_BADLINE;
				}
			}
		}

		for (i = 0; i < ma_inf.badsslines; i++)
		{
			int ln1 = badlinemem[i];

			for (k = i + 1; k < ma_inf.badsslines; k++)
			{
				int ln2 = badlinemem[k];

				if (  (lines[ln1].v1->x == lines[ln2].v1->x
					&& lines[ln1].v1->y == lines[ln2].v1->y
					&& lines[ln1].v2->x == lines[ln2].v2->x
					&& lines[ln1].v2->y == lines[ln2].v2->y)
					||
					  (lines[ln1].v2->x == lines[ln2].v1->x
					&& lines[ln1].v2->y == lines[ln2].v1->y
					&& lines[ln1].v1->x == lines[ln2].v2->x
					&& lines[ln1].v1->y == lines[ln2].v2->y))
				{
					int fl1 = lines[ln1].ml_flags & ML_TWOSIDED;
					int fl2 = lines[ln2].ml_flags & ML_TWOSIDED;

					if (!fl1 && !fl2)
					{
						if (R3D_edit)
						{
							pr_OFinsert("   1's lines %i [fs %i] and %i [fs %i] are superimposed\n",
								ln1, lines[ln1].frontsector - sectors,
								ln2, lines[ln2].frontsector - sectors);

							endret = true;
						}

						// The trace is fixed if they both have the same frontsector.
						if (lines[ln1].frontsector == lines[ln2].frontsector)
							hitline[ln2] = 4;
						else
							line_error = true;
					}
				}
			}
		}
		if (endret) pr_OFdirect("\n");
	}

	// Put linedef's with a sector with highest
	// linecounts at start of the list.
	qsort(linearray, larray_count, sizeof(int), traceArraySorter);

	for (i = 0; i < larray_count; i++)
	{
		if ((int)(20.0f * ((float)i/(float)larray_count)) > barpos)
		{
			barpos += 5;
			Con_Progress(5, 0); // $progress XH 3
		}

		if (hitline[linearray[i]]) continue;

		trace_count = 0;
		SET_countsec(0);

		rcv_li = &lines[linearray[i]];
		secnum = rcv_li->frontsector - sectors;

		while (subfn_IterateBoundary()) ;	// loop ends here

		// Mark all subsectors in this group
		ssec_count = 0;

		while (subfn_IterateSubsectors()) ;	// loop ends here

		// Mark off lines used
		subfn_MarkLinesTestedLim();

		if (!ssec_count) continue;

		group_count++;
	}

	for (i = 0, count = 0; i < numlines; i++)
	{
		if (hitline[i] == 2) count++;
	}

	if (numlines > count)
	{
		ma_inf.sslinecheck = count;
	}

	for (i = 0; i < numsubsectors; i++)
	{
		if (subsectors[i].viewgroup < 0)
		{
			subsectors[i].viewgroup = 0;
		}
	}

	CLR_hitline();

	if (!R3D_edit)
	{
		for (i = 0; i < ma_inf.badsslines; i++)
		{
			if (   !subfn_FindLineConnected(&lines[badlinemem[i]], 200)
				|| !subfn_FindLineConnected(&lines[badlinemem[i]], 201))
			{
				hitline[badlinemem[i]] = 1;
			}
		}
	}

	for (i = 0; i < numlines; i++)
	{
		ssline_t *sl = &sslines[i];
		int		 gn1 = hitline[i] ? 0 : -1;

		for (k = 0; k < sl->num; k++)
		{
			subsector_t *ss = &subsectors[sl->list[k]];

			if (gn1 < 0)
			{
				gn1 = ss->viewgroup;
			}
			else if (ss->viewgroup != gn1)
			{
				mergeGroups(gn1, ss->viewgroup);
				break;
			}
		}
	}

	// Set the line group num array
	for (i = 0, count = 0; i < numsubsectors; i++)
	{
		int k, gr_num;

		if (!ssecarray[i]) count++;

		if ((gr_num = subsectors[i].viewgroup) >= 0)
		{
			rcv_ss = &subsectors[i];
			rcv_sg = &segs[rcv_ss->firstline];

			for (k = 0; k < rcv_ss->linecount; k++, rcv_sg++)
			{
				if (rcv_sg->linedef != NULL)
					lgrpnum[rcv_sg->linedef - lines] = gr_num;
			}
		}
	}

	ma_inf.nogroup = count;

	lgroup_num = lgroup_max = group_count;

	subfn_SetControlLines();
}
