//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2011 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_traceReversed.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
//
// DESCRIPTION: finds all subsectors in a line loop but adds functionality
//              in that the lines can have the same sector each side.  The
//              outer line's side has a dummy sector inserted to allow the
//              trace to work.  The correct sector is restored on exit.
//              This also means that a line loop having the same sector
//              facing outwards but with different sectors on the inner
//              sides can also be traced by substitution.  This routine was
//              also intended to eventually replace r3D_traceNormal but, as
//              ever, time ran out.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdlib.h>
#include <string.h>
#include <search.h>

#include "con_bar.h"
#include "r3d_MAINspec.h"
#include "r3d_trace.h"

#include "m_misc.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static int	larray_count;

static seg_t	*rcv_sg;
static vertex_t *rcv_v1, *rcv_start;

static int refsec;
static int flag_test;
static int secnum;
static int abox[4];

// CODE --------------------------------------------------------------------

static RDBOOL subfn_TestSubsectors1(linev_t	*lv, int flags)
{
	subsector_t *ss;
	int i, testxy_valid = true;

	// Mark lines of interest i.e. all boundary
	// lines contained in the test area.
	for (i = 0; i < larray_count; i++)
	{
		lv[i].mark = true;
	}

	// We need to know if the line's sectors are on the inside of
	// the bound area or the outside.  Since the main algo expects
	// the sectors to be facing in, then, in order to use it to test
	// the obverse then x and y must be set to the same value (and
	// must, of course, be outside the box).  Since the box may be
	// on the map limits 4 tests are provided where one should pass.

	if (abox[BOXLEFT ] > -32758 * FRACUNIT)
		ssec_midy = ssec_midx = abox[BOXLEFT ] - 10*FRACUNIT;
	else
	if (abox[BOXRIGHT] <  32757 * FRACUNIT)
		ssec_midy = ssec_midx = abox[BOXRIGHT] + 10*FRACUNIT;
	else
	if (abox[BOXBASE ] > -32758 * FRACUNIT)
		ssec_midy = ssec_midx = abox[BOXBASE ] - 10*FRACUNIT;
	else
	if (abox[BOXTOP  ] <  32757 * FRACUNIT)
		ssec_midy = ssec_midx = abox[BOXTOP  ] + 10*FRACUNIT;
	else
		testxy_valid = false;

	if (   testxy_valid
		&& subfn_LiesInTrueArea(lv, larray_count, refsec, FIX2FLT(ssec_midy), FIX2FLT(ssec_midx)))
	{
		return false;	// the box lines' sectors are facing out.
	}

	if (flags & MSIA_TESTAREA) return true;

	for(i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		if (!ssecarray[i])
		{
			ssec_midx = (int)(ss->midpoint.x * FRACUNIT);
			ssec_midy = (int)(ss->midpoint.y * FRACUNIT);

			// Quick bounds test for rejection
			if (ssec_midx > abox[BOXLEFT ] &&
				ssec_midx < abox[BOXRIGHT] &&
				ssec_midy > abox[BOXBASE ] &&
				ssec_midy < abox[BOXTOP  ])
			{
				if (subfn_LiesInTrueArea(lv, larray_count, refsec, ss->midpoint.x, ss->midpoint.y))
				{
					// v220-7 check ceiling and floor flats are both skyflats
					if (   flag_test & MSIA_SKY
						&& !(ss->sector->floorpic == skyflatnum
						&& ss->sector->ceilingpic == skyflatnum))
					{
						return false;
					}

					if (ss->sector) hitssec[hitssecptr++] = i;
					ssecarray[i] = 1;
				}
			}
		}
	}

	return hitssecptr > 0;
}

static void subfn_SetLArrayVars(linev_t *lv, int flags, int replacesec)
{
	line_t	*ld;
	int		i;

	refsec = (flags & (MSIA_REVERSED|MSIA_TESTINNER)) ? numsectors : secnum;

	for (i = 0; i < larray_count; i++)
	{
		ld = &lines[linearray[i]];

		lv[i].v1[VX] = FIX2FLT(ld->v1->x);
		lv[i].v1[VY] = FIX2FLT(ld->v1->y);
		lv[i].v2[VX] = FIX2FLT(ld->v2->x);
		lv[i].v2[VY] = FIX2FLT(ld->v2->y);

		lv[i].bsec = (ld->ml_flags & ML_TWOSIDED ? ld->backsector - sectors : -1);
		lv[i].fsec = ld->frontsector - sectors;

		// When reversed give the line side inside the loop a unique sector number
		if (flags & MSIA_REVERSED)
		{
			if (lv[i].fsec != secnum)
			{
				lv[i].fsec = refsec;
			}

			if (lv[i].bsec != secnum)
			{
				lv[i].bsec = refsec;
			}
		}

		if (flags & MSIA_TESTINNER)
		{
			if (lv[i].fsec == secnum)
			{
				lv[i].bsec = refsec;
			}

			if (lv[i].bsec == secnum)
			{
				lv[i].fsec = refsec;
			}
		}

		if (replacesec >= 0)
		{
			if (lv[i].fsec != secnum && lv[i].bsec != secnum)
			{
				if (lv[i].fsec == replacesec) lv[i].fsec = secnum;
				if (lv[i].bsec == replacesec) lv[i].bsec = secnum;
			}
		}

		lv[i].midx = meanV(ld->v1->x, ld->v2->x);
		lv[i].midy = meanV(ld->v1->y, ld->v2->y);
	}
}

static int subfn_Trace1(int flags, int replacesec)
{
	linev_t	*lv;
	int		ret;

	if (!larray_count) return false;

	// This can be called from r3d_slopes.c even if an analyser
	// file's been loaded.  So it must be alloced locally.
	lv = (linev_t*)M_Calloc(larray_count*sizeof(linev_t), 307);

	subfn_SetLArrayVars(lv, flags, replacesec);

	ret = subfn_TestSubsectors1(lv, flags);	// Mark all subsectors in this group

	M_Free(lv); lv = NULL;

	return ret;
}

// This is used where the sector's lines are within the sector itself.
// It should only be called with pre-marked, complete single loops.
// If array is NULL then linearray is assumed to be set.
RDBOOL LS_MarkSubsecsInAreaGeneral(sector_t *s, byte *array, int loopptr, int *looplist, int *bbox, int rsec, int flags)
{
	int i;

	memcpy(abox, bbox, sizeof(abox));

	hitssecptr = 0;

	CLR_ssecarray();

	secnum = s - sectors;

	larray_count = 0;	// default
	flag_test = flags;

	if (!array && !looplist)	// Indicates linearray already set up - rev170s
	{
		larray_count = loopptr;
	}
	else if (looplist)
	{
		for (i = 0; i < loopptr; i++)
		{
			if (array[looplist[i]] == 4) // Set by calling function
				linearray[larray_count++] = looplist[i];
		}
	}
	else
	{
		for (i = 0; i < s->linecount; i++)
		{
			line_t	*ld = s->lines[i];

			if (array[ld - lines] == 4) // Set by calling function
				linearray[larray_count++] = ld - lines;
		}
	}

	return subfn_Trace1(flags, rsec);
}

RDBOOL inList(int snum, int loopcnt)
{
	int i;

	for (i = 0; i < loopcnt; i++)
	{
		if (snum == sidelist[i])
			return true;
	}
	return false;
}

// Uses linearray so this must be set up before calling
RDBOOL LS_MarkSubsecsInMultiSecLoop(int chkptr, int *chksec, int loopcnt, int *bbox, int *other, RDBOOL usesidelist, int flags)
{
	linev_t	*lv;
	line_t	*ld;
	int		c, i, k;
	RDBOOL	ret;

	if (other) *other = -1;
	hitssecptr = 0;
	CLR_ssecarray();

	if ( (larray_count = loopcnt) <= 0) return false;

	refsec = numsectors;
	memcpy(abox, bbox, sizeof(abox));

	lv = (linev_t*)M_Calloc(larray_count*sizeof(linev_t), 308);

	for (c = i = 0; i < larray_count; i++)
	{
		RDBOOL found = false;

		ld = &lines[linearray[i]];

		lv[i].v1[VX] = FIX2FLT(ld->v1->x);
		lv[i].v1[VY] = FIX2FLT(ld->v1->y);
		lv[i].v2[VX] = FIX2FLT(ld->v2->x);
		lv[i].v2[VY] = FIX2FLT(ld->v2->y);

		if (usesidelist)
		{
			if (inList(ld->sidenum[0], loopcnt))
			{
				if (flags & MSIA_REVERSED)
				{
					lv[i].fsec = chksec ? chksec[0] : refsec + 1;
					lv[i].bsec = refsec;
				}
				else
				{
					lv[i].fsec = refsec;
					lv[i].bsec = chksec ? chksec[0] : refsec + 1;
				}
			}
			else
			{
				if (flags & MSIA_REVERSED)
				{
					lv[i].fsec = refsec;
					lv[i].bsec = chksec ? chksec[0] : refsec + 1;
				}
				else
				{
					lv[i].fsec = chksec ? chksec[0] : refsec + 1;
					lv[i].bsec = refsec;
				}
			}
		}
		else
		{
			lv[i].bsec = (ld->ml_flags & ML_TWOSIDED ? ld->backsector - sectors : -1);
			lv[i].fsec = ld->frontsector - sectors;

			for (k = 0; k < chkptr; k++)
			{
				if (lv[i].fsec == chksec[k])
				{
					if (flags & MSIA_MSREVERSED)
						lv[i].bsec = refsec;
					else
						lv[i].fsec = refsec;
					found = true;
					break;
				}
			}

			if (lv[i].bsec >= 0)
			{
				for (k = 0; k < chkptr; k++)
				{
					if (lv[i].bsec == chksec[k])
					{
						if (flags & MSIA_MSREVERSED)
							lv[i].fsec = refsec;
						else
							lv[i].bsec = refsec;
						break;
					}
				}
			}
			else if (!found)
			{
				lv[i].bsec = refsec; c++;
			}

			if (other && *other != -2)
			{
				int flags = LS_TestLineTex(ld);

				if (flags & (LT_MISHI|LT_MISLO))
				{
					int sec = lv[i].fsec == refsec ?
						ld->frontsector - sectors : ld->backsector - sectors;

					if      (*other ==  -1) *other = sec;
					else if (*other != sec) *other = -2;
				}
			}
		}

		lv[i].midx = meanV(ld->v1->x, ld->v2->x);
		lv[i].midy = meanV(ld->v1->y, ld->v2->y);
	}

	if (c == larray_count) { M_Free(lv); return false; }

	ret = subfn_TestSubsectors1(lv, flags);	// Mark all subsectors in this group

	M_Free(lv);

	return ret;
}

// v220-10
// Uses preset linearray and sidelist
// uses different sidelist scheme
RDBOOL LS_MarkSsecsUsingSidelist(int loopcnt, int *bbox, int flags)
{
	linev_t	*lv;
	line_t	*ld;
	int		i;
	RDBOOL	ret;

	hitssecptr = 0;

	if ( (larray_count = loopcnt) < 3) return false;

	memcpy(abox, bbox, sizeof(abox));
	CLR_ssecarray();

	lv = (linev_t*)M_Calloc(larray_count*sizeof(linev_t), 308);

	refsec = numsectors;

	for (i = 0; i < larray_count; i++)
	{
		RDBOOL found = false;

		ld = &lines[linearray[i]];

		lv[i].v1[VX] = FIX2FLT(ld->v1->x);
		lv[i].v1[VY] = FIX2FLT(ld->v1->y);
		lv[i].v2[VX] = FIX2FLT(ld->v2->x);
		lv[i].v2[VY] = FIX2FLT(ld->v2->y);

		if (ld->sidenum[0] == sidelist[i])
		{
			if (flags & MSIA_REVERSED)
			{
				lv[i].fsec = refsec + 1;
				lv[i].bsec = refsec;
			}
			else
			{
				lv[i].fsec = refsec;
				lv[i].bsec = refsec + 1;
			}
		}
		else if (ld->sidenum[1] == sidelist[i])
		{
			if (flags & MSIA_REVERSED)
			{
				lv[i].fsec = refsec;
				lv[i].bsec = refsec + 1;
			}
			else
			{
				lv[i].fsec = refsec + 1;
				lv[i].bsec = refsec;
			}
		}
		else
			Con_ErrorD("Program error: LS_MarkSubsecsInMultiSecLoop2");

		lv[i].midx = meanV(ld->v1->x, ld->v2->x);
		lv[i].midy = meanV(ld->v1->y, ld->v2->y);
	}

	ret = subfn_TestSubsectors1(lv, flags);	// Mark all subsectors in this group

	M_Free(lv);

	return ret;
}

// Used where all the lines in the looplist form a single loop
// b211-0 Revised
RDBOOL LS_MarkSubsecsInLoop(sector_t *s, int loopcnt, int *looplist, int *bbox, int flags, int *seccnt)
{
	sector_t	*rs = s;
	int			i, k;
	int			chksec[128];
	int			chkseccnt = 0;

	hitssecptr = 0;

	CLR_ssecarray();

	secnum = s - sectors;

	// If looplist is NULL then it is assumed that linearray has been set up;
	if (looplist)
	{
		larray_count = 0;

		for (i = 0; i < loopcnt; i++)
		{
			linearray[larray_count++] = looplist[i];
		}
	}
	else
	{
		larray_count = loopcnt;
	}

	if (flags & MSIA_REVERSED)
	{
		for (i = 0; i < larray_count; i++)
		{
			line_t	*li;

			li = &lines[linearray[i]];

			if (li->ml_flags & ML_TWOSIDED)
			{
				int		n = li->frontsector == s ?
							li->backsector - sectors : li->frontsector - sectors;
				RDBOOL	found = false;

				for (k = 0; k < chkseccnt; k++)
				{
					if (chksec[k] == n) {found = true; break; }
				}

				if (!found)
				{
					if (chkseccnt >= 128) Con_ErrorD("LS_MarkSubsecsInLoop: chkseccnt exceeded");
					chksec[chkseccnt++] = n;
				}
			}
		}

		if (seccnt) *seccnt = chkseccnt - 1;	// Note: - 1 to eliminate test sector

		return LS_MarkSubsecsInMultiSecLoop(chkseccnt, chksec, larray_count, bbox, NULL, false, flags);
	}

	// if flag != MSIA_REVERSED and seccnt required then this purely
	// checks the line's sectors.  It is not used by the main code
	if (seccnt)
	{
		sector_t *fs = NULL, *bs = NULL;

		for (i = 0; i < loopcnt; i++)
		{
			sector_t *cf = lines[linearray[i]].frontsector;
			sector_t *cb = lines[linearray[i]].backsector;

			if (!cb) continue;

			if (cf != s)
			{
				int		n = cf - sectors;
				RDBOOL	found = false;

				for (k = 0; k < chkseccnt; k++)
				{
					if (chksec[k] == n) {found = true; break; }
				}

				if (!found && chkseccnt < 128)
				{
					chksec[chkseccnt++] = n;
				}
			}
			if (cb != s)
			{
				int		n = cb - sectors;
				RDBOOL	found = false;

				for (k = 0; k < chkseccnt; k++)
				{
					if (chksec[k] == n) {found = true; break; }
				}

				if (!found && chkseccnt < 128)
				{
					chksec[chkseccnt++] = n;
				}
			}
		}
		*seccnt = chkseccnt;	// Note test sector not counted
		chkseccnt = 0;
	}

	memcpy(abox, bbox, sizeof(abox));

	return subfn_Trace1(flags, -1);
}
