//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_wad.c
//
// First created (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 29 (v220-29) 22/Nov/2014 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//
//
// DESCRIPTION:	Node builder select
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <windows.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>
#include <direct.h>
#include <process.h>
#include <sys/timeb.h>
#include <time.h>
#include <errno.h>

#include "doom\d_englsh.h"
#include "con_console.h"
#include "dd_main.h"
#include "dd_wad.h"
#include "m_misc.h"
#include "sys_system.h"
#include "r3d_compat.h"


// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

#define NUM_CHECK_LUMPS	8

static const char *level_lumps[NUM_CHECK_LUMPS] =
{
	"THINGS",
	"LINEDEFS",
	"SIDEDEFS",
	"VERTEXES",
	"SEGS", 
	"SSECTORS",
	"NODES",
	"SECTORS"
};

typedef struct {
	void *ptr;
	uint sz;
} dataptr_t;

dataptr_t *dptrs;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void initNodesFile(void);
void updateNodesBuild(byte *sidedata, uint timewrite);
byte *NodesData(char *name, uint *size);
byte *getSIDElump(uint *size);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

static int changeIWADtoPWAD(char *fd, char *outfile);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern char	new_mapid[8];
extern	int	R3D_edit;		// Set to 1 if edit is active

// PUBLIC DATA DEFINITIONS -------------------------------------------------

RDBOOL		haslevels;
RDBOOL		levelExMxExists;
checkwad_t	cwinf;
int			epihaspwad[10];

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static byte		*sidedata;
static RDBOOL	nodesmissing;

// CODE --------------------------------------------------------------------


//=============================================================================
//  subfn_CheckWAD routines: Checks for levels - R3D
//=============================================================================

static RDBOOL CharIsNum(char s)
{
	return (s >= '0' && s <= '9');
}


void getDIR_entry(DFILE *f, lump_header_t *lumpinf, char *buff, char *fname, int *err)
{
	if (F_Read(lumpinf, sizeof(lump_header_t), f) != sizeof(lump_header_t))
	{
		if (err) *err = true;
		return;
	}

	if (err) *err = false;
	strncpy(buff, lumpinf->name, 8);
}

RDBOOL subfn_CheckWAD(char *fname, uint *csm, int *type)
{
	wad_header_t  header;
	lump_header_t lumpinf;

	int		c, i, k, p;
	int		directory_errors = 0;
	int		num_entries;
	int		leveltype, error;
	int		epi, num;
	
	char	entry[10];
	char	szName[MAXFNAME];
	char	szExt[MAXEXT];

	
	DFILE *f = F_OpenFile(fname, "b");	// changed to DFILE v220-6
	
	//----------------------------------------------------------------
	
	if (f == NULL) Con_Error("ERROR: %s\n%s", fname, strerror(errno));

	*csm = 0;
	*type = TYPE_UNKNOWN;	// Until we know better
	entry[8] = 0;

	haslevels = false;

	_splitpath(fname, NULL, NULL, szName, szExt);

	if (_stricmp(".WAD", szExt)) return false;	// not a .wad file
	
	// read the WAD header
	if ((i = F_Read(&header, sizeof(header), f)) != sizeof(header)
		|| !(num_entries = header.num_entries))	// nothing to read
	{    
		F_Close(f); 
		Con_Error("CheckWAD  : %s header read error", fname);
		return false; 
	}
	
	if (!strncmp(header.type, "PWAD", 4))
	{
		*type = TYPE_PWAD;
	}
	else if (!strncmp(header.type, "IWAD", 4))
	{
		RDBOOL	freedoom = !_stricmp(szName, "FREEDOOM1") || !_stricmp(szName, "FREEDOOM2")  || !_stricmp(szName, "FREEDM");
		RDBOOL	hacx = !strcmp(szName, "HACX");
		char	buff[1028];
		char	id_buf[26];
		i;

		F_Seek(f, 12, SEEK_SET);

		if (F_Read(id_buf, 24, f) != 24) { F_Close(f); return 3; }

		// v220-29 check for a hacx iwad
		if (!strcmp(id_buf, "HacX Registered v1.2"))
		{
			if (!hacx)
			{
				F_Close(f);
				Con_Error("Renamed HACX iwads are not supported");
			}
			hacx = true;
		}
		else
		{
			id_buf[24] = 0;

			if ( (!strcmp(id_buf, " DeuTex 4.4.902 OJM 94 ")) )
			{
				if (!freedoom)
				{
					F_Close(f);
					Con_Error("Renamed FREEDOOM iwads are not supported");
				}
			}
		}

		if (hacx || freedoom)
		{
			F_Close(f);

			sprintf(buff,
				"Risen3D is not set up to support the %s.wad\n"
				"but it can be converted for use with Risen3D\n\n"
				"The %s.wad will be saved as %s.bak\n"
				"and the modified file named to %sPWAD.wad\n\n",
				szName, szName, szName, szName);

			if (Sys_CriticalMessage(buff, true))
			{
				char	outfile[MAXFNAMEX];
				char	msg[512];
				int		error;
				uint	pos;

				strcpy(outfile, fname);
				pos = strlen(outfile) - 3;
				outfile[pos] = 0;
				strcat(outfile, "PWAD.wad");

				// delete any existing PWAD file
				if ((_access(outfile, 4 )) != -1)
				{
					pr_OFinsert("deleting %s\n", outfile);
					remove(outfile);
				}

				outfile[pos] = 0;
				strcat(outfile, "bak");

				// delete any existing .bak file
				if ((_access(outfile, 4 )) != -1)
				{
					pr_OFinsert("deleting %s\n", outfile);
					remove(outfile);
				}

				error = changeIWADtoPWAD(fname, outfile);

				if (error)
				{
					sprintf(msg, 
						"Failed to convert %s\nERROR: %s",
						fname,
						error == 1 ? "failed to open file for reading" :
						error == 2 ? "insufficient memory" :
						error == 3 ? "failed to read data" :
						error == 4 ? "failed to rename file" :
						error == 5 ? "failed to open file for writing" :
						error == 6 ? "failed to write data" : "unknown");
				}
				else
				{
					sprintf(msg, 
							"%s.wad has been saved as a PWAD\n"
							"to allow it to be run with Risen3D.\n"
							"On pressing OK Risen3D will exit\n"
							"Just restart and run %sPWAD.wad",
							szName, szName);
				}

				Sys_MessageBox(msg, error != 0);
			}

			Sys_Quit(false);
		}
		else
		{
			*type = TYPE_IWAD;
		}
	}

	if (*type == TYPE_UNKNOWN)
	{
		F_Close(f); 
		Con_Error("CheckWAD  : %s is neither an IWAD or a PWAD", fname);
		return false; 
	}

	num = -1;

	F_Seek(f, header.dir_start, SEEK_SET);

	for (p = 0; p < num_entries; p++)
	{   
		getDIR_entry(f, &lumpinf, entry, fname, &error);	// read the entry

		if (error)
		{
			directory_errors++; continue;
		}

		(*csm) += lumpinf.length;

		for(k = 0; k < 8; k++) (*csm) += entry[k];

		// check if a level name
		if (( (!cwinf.iwadisdoom2 || *type == TYPE_IWAD)
			&& !entry[4]
			&& entry[0] == 'E'
			&& entry[2] == 'M' 
			&& CharIsNum(entry[1])
			&& CharIsNum(entry[3])) 
			||
			( (cwinf.iwadisdoom2 || *type == TYPE_IWAD)
			&& !entry[5]
			&& !strncmp(entry, "MAP", 3) 
			&& CharIsNum(entry[3])
			&& CharIsNum(entry[4])))
		{
			char mapname[6];

			strcpy (mapname, entry);

			if (entry[0] == 'M')	// MAPxx
			{
				num = (entry[3] - '0')*10 + (entry[4] - '0');

				if (num < 1 || num > 99) continue;

				leveltype = 1;
			}
			else						// ExMx
			{
				if ((epi = entry[1] - '0') < 1 || epi > 9) continue;
				if ((num = entry[3] - '0') < 0 || num > 9) continue;

				num += epi*10;
				leveltype = 0;
			}
			
			// check the following lumps are present and in the correct order
			// v220-33 revised following change first put in in v220-32
			for (c = 0 ; p < num_entries && c < NUM_CHECK_LUMPS; p++) 
			{
				getDIR_entry(f, &lumpinf, entry, fname, NULL);
				
nextCcheck:;
				if (!strcmp(entry, level_lumps[c]))
				{
					if (c == 4 || c == 5 || c == 6)
					{
						if (R3D_edit && !lumpinf.length)
						{
							if (ge_inf.devmap)
								pr_OFinsert("NOTE: %s %s %.8s is zero length\n",
									*type == TYPE_IWAD ? "IWAD" : "PWAD", mapname, entry);
							nodesmissing = true;
						}
					}
					c++;
				}
				else
				{
					// v220-16 don't care about missing segs, ssecs or nodes
					if (c == 4 || c == 5 || c == 6)
					{
						nodesmissing = true;
						c++; goto nextCcheck;
					}
					else
					{
						c = 0; break;	// required lump not found
					}
				}
			}
			
			if (c != NUM_CHECK_LUMPS) continue;	// Required lumps not found

			if (*type == TYPE_PWAD)
			{
				if (leveltype)
				{
					haslevelMapxx[num] = true;
				}
				else
				{
					haslevelExMx[num] = true;
					levelExMxExists = true;
					epihaspwad[epi] = true;
				}
				
				hasPWADlevel[num] = true;

				haspwads = true;
				cwinf.numlevels  += 1;         // +1 level found
				cwinf.pwadisdoom2 = leveltype; // type of level DOOM or DOOM2
			}
			else
			{
				hasIWADlevel[num] = true;
			}

			haslevels = true;
		}
	}

	F_Close(f);

	if (directory_errors)
	{
		char szFname[256];

		_splitpath(fname, NULL, NULL, szFname, NULL);
		pr_OFinsert("WARNING   : %s.wad has a malformed directory\n", szFname);
	}

	// Note: slot 0 reserved for the IWAD csm
	if (*type == TYPE_IWAD)
	{
		pwadchecksum[0] = *csm;
		return haslevels > 0;
	}

	// csm only stored if the pwad has levels
	// These are used to check savegames
	if (haslevels)
	{
		for (k = 1; k <= MAX_PWADS; k++)
		{
			if (!pwadchecksum[k])
			{
				pwadchecksum[k] = *csm;
				break;
			}
		}

		*csm = 0;
	}
	
	return true;
}

// Separated out to help make the code more readable
static void updateSIDES(byte *data, char *mapwad, uint sz)
{
	mapsidedef_t	*ptr = (mapsidedef_t*)data;
	side_t			*sd = sides;
	short			v;
	int				i, cx = 0, cy = 0;

	//Con_Message("numsides %i mlsides %i\n", numsides, sz/sizeof(mapsidedef_t));

	for (i = 0; i < numsides; i++, sd++, ptr++)
	{
		if (ptr->textureoffset != (v = (short)(sd->textureoffset>>16)) )
		{
			ptr->textureoffset = v; cx++;
		}
		if (ptr->rowoffset != (v = (short)(sd->rowoffset>>16)) )
		{
			ptr->rowoffset = v; cy++;
		}
	}

	sidedata = (byte*)malloc(sz);
	memcpy(sidedata, data, sz);

	pr_OFinsert(
		"%i texture x-offset and %i y-offset changes were found\n",
		cx, cy);

	pr_OFrepeat("\n", "=", 64, 1);
}

// v220-6
static int changeIWADtoPWAD(char *fd, char *outfile)
{
	if ( (_access(outfile, 4)) == -1L )
	{
		FILE *f;
		uint sz;
		byte *p;

		if ( !(f = fopen(fd, "rb")) ) return 1;

		fseek(f, 0, SEEK_END);
		sz = ftell(f);

		if (!(p = (byte*)malloc(sz))) { fclose(f); return 2; }

		rewind(f);
		if (fread(p, 1, sz, f) != sz) { fclose(f); free(p); return 3; }
		fclose(f);

		if (rename(fd, outfile)) { free(p); return 4; }

		strcpy(outfile, fd);
		outfile[strlen(outfile) - 4] = 0;
		strcat(outfile, "PWAD.wad");

		if ( !(f = fopen(outfile, "wb")) ) { free(p);  return 5; }
		
		p[0] = 'P';
		if (fwrite(p, sz, 1, f) != 1) { fclose(f); free(p); return 6; }
		fclose(f);
		free(p);
	}

	return 0;
}

// v220-6
void modifyWAD(int changes)
{
	lump_header_t lumpinfin;
	lump_header_t *lumpinf;
	wad_header_t  header;
	
	DFILE	*in = NULL;
	DFILE	*out = NULL;

	RDBOOL	bak_exists;
	RDBOOL	leanwad;

	char	outfile[MAXFNAME];
	char	lname[10];
	char	buff[1536];
	char	*mapwad;

	byte	*data = NULL;
	byte	*filedata = NULL;
	
	uint	dir_start;
	uint	filepos;
	uint	current_pos;
	int		num_entries, num_new_entries;
	int		map_set = -1;
	int		error = 0;
	int		p;

	//----------------------------------------------------------------

	if (!changes || !strlen(new_mapid) || !ma_inf.map_lumpnum)
	{
		pr_OFdirect("No editing changes were made\n");
		return;
	}

	// find the file being edited
	mapwad = filenameRecords[lumpinfo[ma_inf.map_lumpnum].modidx].filename;

	strcpy(outfile, mapwad);
	outfile[strlen(outfile) - 3] = 0;
	strcat(outfile, "bak");

	bak_exists = (_access(outfile, 4)) != -1L;

	//-----------------------------------------------
	// Confirm action

	pr_OFrepeat("\n", "=", 64, 1);

	buff[0] = 0;

	if (ma_inf.mapsides != numsides)
	{
		strcpy(buff, "EDIT PROBLEM: sides were packed.\n"
			" This means the changes cannot be saved.\n"
			" The changes you made can be viewed in edit.out.\n"
			" and set using a map editor.");

		pr_OFdirect(buff);
		Sys_MessageBox(buff, false);
		pr_OFrepeat("\n", "=", 64, 1);
	
		return;
	}
	
	sprintf(buff,
		"Texture offset changes were made.\n\n"
		"Changes plus the normal nodes built for %s will be saved to;\n   %s\n\n"
		"The existing file will be saved as;\n   %s\n%s\n"
		"Save changes?",
		new_mapid, mapwad, outfile,
		bak_exists ?
		"\nNOTE: a backup already exists and will be overwritten" : "");

	if (!Sys_CriticalMessage(buff, true))
		return;
	
	//-----------------------------------------------
	// Create and open backup

	// check mapwad exists - just in case some lulu has tabbed out and deleted it
	if ( (_access(mapwad, 0)) == -1)
	{
		error = 1;
	}
	else if (system(NULL))
	{
		if (bak_exists) remove(outfile);

		// To preserve dates without a lot of mucking around use the system call
		sprintf(buff, "copy \"%s\" \"%s\"", mapwad, outfile);
		_flushall();

		if (system(buff) || errno)
			error = 2;
		// Entries will be read from the backup
		else if ( (in = F_Open(outfile, "b")) == NULL)
			error = 3;
	}
	else
		error = 4;

	if (error)
	{
		Con_Error("EDIT: failed to %s",
			error == 1 ? "find wad" : error == 2 ? "create backup" : error == 3 ?  "open backup" : "find cmd.exe");
	}

	//-----------------------------------------------
	// Check the backup is valid

	in = F_Open(outfile, "b");

	if (F_Read(&header, sizeof(header), in) != sizeof(header))
	{    
		Con_Error("INFILE %s : header read error", mapwad);
	}

	if (   strncmp(header.type, "PWAD", 4)	// not a pwad
		|| !(num_entries = header.num_entries) )
	{
		Con_Error("INFILE %s : not a valid pwad", mapwad);
	}

	//-----------------------------------------------

	// v220-27 moved here
	initNodesFile();	// may not return

	leanwad = ArgExists("-lean") || nodesmissing;

	// Get lump data using the backup directory
	lumpinf = (lump_header_t *)W_Calloc(num_entries * sizeof(lump_header_t), STATICID, 0);
	dptrs = (dataptr_t *)W_Calloc(num_entries * sizeof(dataptr_t), STATICID, 0);
	
	dir_start = header.dir_start;
	F_Seek(in, dir_start, SEEK_SET);

	filepos = sizeof(wad_header_t);

	lname[8] = 0;

	// read the directory entries
	// no GL_ lumps for any map, will be copied
	// reject or blockmap lumps will be set to 0 for the input map
	// if -lean was put on the command line then segs, ssectors and nodes will be stored empty 
	for (num_new_entries = p = 0; p < num_entries; p++)
	{   
		if (F_Read(&lumpinfin, sizeof(lump_header_t), in) == sizeof(lump_header_t))
		{
			byte	*data;
			uint	sz = 0;
			RDBOOL	mustread = false;

			if (!strncmp(lumpinfin.name, "GL_", 3)) // i.e. anything but
				continue;

			memcpy(lname, lumpinfin.name, 8);
			strncpy(lumpinf[num_new_entries].name, lname, 8);

			if (map_set < 0)
			{
				if (!_stricmp(lname, new_mapid))
					map_set = p;
				mustread = true;
			}
			else if (!_stricmp(lname, "THINGS"))
			{
				if (p != map_set + 1)
					map_set = -1;
				mustread = true;
			}
			else if (map_set >= 0)
			{
				if (leanwad)
				{
					// copy lumps from nodes file
					if (   !strcmp(lname, "LINEDEFS")
						|| !strcmp(lname, "SIDEDEFS")
						|| !strcmp(lname, "VERTEXES")
						|| !strcmp(lname, "SECTORS"))
					{
						if ( !(data = NodesData(lname, &sz)) )
							Con_Error("NODES FILE : failed to get \"%s\"", lname);

						if (!_stricmp(lname, "SIDEDEFS"))
							updateSIDES(data, mapwad, sz);

						dptrs[num_new_entries].ptr = data;
						dptrs[num_new_entries].sz = sz;
					}
					else
					// Lumps that are not needed by R3D
					if (   !stricmp(lname, "SEGS")
						|| !stricmp(lname, "SSECTORS")
						|| !stricmp(lname, "NODES")
						|| !strcmp(lname, "BLOCKMAP")
						|| !strcmp(lname, "REJECT"))
					{
						if (dptrs[num_new_entries].ptr)
						{
							static firsttime = true;

							if (firsttime)
							{
								firsttime = false;
								pr_OFinsert("NOTE: SSEGS, SSECTORS, NODES, REJECT, BLOCMAP\n"
											"These are not needed by Risen3D. The labels have been kept as this allows the\n"
											"wad to be loaded into an editor (despite even these not being required by R3D).\n\n");
							}

							pr_OFinsert("NOTE: \"%s\" was made empty\n", lname);
						}

						M_Free(dptrs[num_new_entries].ptr);
						dptrs[num_new_entries].sz = 0;
						dptrs[num_new_entries].ptr = NULL;
					}
					else
					{
						map_set = -1;
					}
				}
				else
				{
					// copy lumps from nodes file
					if (   !strcmp(lname, "LINEDEFS")
						|| !strcmp(lname, "SIDEDEFS")
						|| !strcmp(lname, "VERTEXES")
						|| !strcmp(lname, "SECTORS"))
					{
						if ( !(data = NodesData(lname, &sz)) )
							Con_Error("NODES FILE : failed to get \"%s\"", lname);

						if (!_stricmp(lname, "SIDEDEFS"))
							updateSIDES(data, mapwad, sz);

						dptrs[num_new_entries].ptr = data;
						dptrs[num_new_entries].sz = sz;
					}
					else
					if (   !stricmp(lname, "SEGS")
						|| !stricmp(lname, "SSECTORS")
						|| !stricmp(lname, "NODES"))
					{
						// These need to be constructed from the nodes build
						if ( !(data = NodesData(lname, &sz)) )
							Con_Error("NODES FILE : failed to get \"%s\"", lname);

						dptrs[num_new_entries].ptr = data;
						dptrs[num_new_entries].sz = sz;
					}
					else
					// Probably empty anyway and don't care if they are not 
					if (   !strcmp(lname, "BLOCKMAP")
						|| !strcmp(lname, "REJECT"))
					{
						M_Free(dptrs[num_new_entries].ptr);
						dptrs[num_new_entries].sz = 0;
						dptrs[num_new_entries].ptr = NULL;
					}
					else
					{
						map_set = -1;
					}
				}
			}

			if (map_set < 0 || mustread)
			{
				data = NULL;
				current_pos = F_Tell(in);
				F_Seek(in, lumpinfin.start, SEEK_SET);

				if ( (sz = lumpinfin.length) )
				{
					data = (byte*)malloc(sz);
					F_Read(data, sz, in);
				}

				dptrs[num_new_entries].ptr = data;
				dptrs[num_new_entries].sz = sz;

				F_Seek(in, current_pos, SEEK_SET);
			}

			lumpinf[num_new_entries].start = filepos;
			lumpinf[num_new_entries++].length = sz;
			filepos += sz;
		}
	}

	if ( (out = F_OpenFileForWrite(mapwad, "b")) == NULL)
	{
		for (p = 0; p < num_new_entries; p++)
		{
			if (dptrs[p].ptr) M_Free(dptrs[p].ptr);
		}
		Con_Error("EDIT ERROR: Could not open %s for writing", mapwad);
	}

	// write the header
	F_Seek(out, 0, SEEK_SET);
	strncpy(header.type, "PWAD", 4);
	header.num_entries = num_new_entries;
	header.dir_start = filepos;
	F_FileWrite(&header, sizeof(header), out, -8);

	for (p = 0; p < num_new_entries; p++)
	{
		if (dptrs[p].ptr)
		{
			F_FileWrite(dptrs[p].ptr, dptrs[p].sz, out, -8);
			M_Free(dptrs[p].ptr);
		}
	}

	if ((current_pos = F_Tell(out)) != filepos)
		Con_Message("ftell %i filepos %i\n", current_pos, filepos);

	// write the directory
	F_FileWrite(lumpinf, sizeof(lump_header_t) * num_new_entries, out, -9);

	F_Close(in);
	F_Close(out);
	W_Free(lumpinf, STATICID, 0);
	W_Free(dptrs, STATICID, 0);

	if (sidedata)
	{
		char	fname[MAXFNAMEX];
		char	szFname[MAXFNAME];
		long	hFile;
		struct	_finddata_t finddata;
		uint	timewrite = 0;
		
		_splitpath(mapwad, NULL, NULL, szFname, NULL);

		strcpy(fname, mapwad);

		fname[strlen(mapwad) - (strlen(szFname) + 4)] = 0;
		
		if (!_chdir(fname)) 
		{
			sprintf(fname, "%s.wad", szFname);

			if((hFile = _findfirst(fname, &finddata)) != -1)
			{
				timewrite = finddata.time_write;
				_findclose( hFile );
			}
		}
		
		_chdir(ddRuntimeDir.path);

		updateNodesBuild(sidedata, timewrite);
	}

	M_Free(sidedata);
}

// v220-10
void changeFileDate(char *filename)
{
	char	mbuf[2048];
	FILE	*file;

	//---------------------------------------

	sprintf(mbuf, "PROBLEM; %s\n"
				  "The pwad file's time is in advance of your PC's clock time\n"
				  "Unless this is corrected save files cannot be used\n\n"
				  "Do you wish to change the pwad's file create time?", filename);

	if (!Sys_CriticalMessage(mbuf, false))
	{
		pr_OFdirect("User aborted time correction.\n");
		return;
	}

	if ( (file = fopen(filename, "rb")) != NULL)
	{
		char		*filebuff;
		_fsize_t	size;

		fseek(file, 0, SEEK_END);
		size = ftell(file);
		rewind(file);
		filebuff = (char*)malloc(size);
		if (filebuff)
			fread(filebuff, 1, size, file);
		fclose(file);

		if (!filebuff) Con_ErrorD("Insufficient memory to save file");

		if ( (file = fopen(filename, "wb")) != NULL)
		{
			fwrite(filebuff, 1, size, file);
			fclose(file);
			M_Free(filebuff); return;
		}

		M_Free(filebuff);
	}

	sprintf(mbuf, "PROBLEM; %s\n"
				  "The file's date could not be changed. Check to see if it is set\n"
				  "to read only and, if so, unset it before restarting.\n", filename);

	Sys_MessageBox(mbuf, true);
}
