//-----------------------------------------------------------------------------
//
// Copyright (C) 1998 - 2005 id Software, Inc., Jaakko Kernen
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r_data.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified/extended for R3D
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//  Cacheing code revisions
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//
// Based on: Doom, Doomsday v1.7.8
//
//
// DESCRIPTION:	Base level rendering refresh functions
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include "de_base.h"
#include "de_console.h"
#include "de_system.h"
#include "de_network.h"
#include "de_refresh.h"
#include "de_graphics.h"
#include "de_misc.h"
#include "de_render.h"

#include "doom/p_local.h"

#include "r3d_MAINspec.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

typedef struct
{
	short	originx;
	short	originy;
	short	patch;
	short	stepdir;
	short	colormap;
} mappatch_t;

typedef struct
{
	char		name[8];
	short		masked;	
	byte		ScaleX;				// [RH] Scaling (8 is normal)
	byte		ScaleY;				// [RH] Same as above
	short		width;
	short		height;
	byte		columndirectory[4];	// OBSOLETE
	short		patchcount;
	mappatch_t	patches[1];
} maptexture_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

RDBOOL	CheckPatch(patch_t *patch, int origx, int origy, int texwidth, uint size);
RDBOOL	R_CompareIWADPatchCounts(char *name, int pc);
void	p2swd(char *msg, ...);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

lumptexinfo_t *lumptexinfo;
unknown_t	unknownname;
flat_t		**flatsarray;

int			numtextures;
texture_t	**texArray;
texscale_t	*texscales;
int			*texturetranslation;	// for global animation

// Used to resize arrays eliminating any new textures generated for 3D lines.
int			numstandardtextures;
int			blacktexnum;
int			notexTEXNUM;

// Glowing textures are always rendered fullbright.

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

#pragma optimize("g", off)

// v220-3
void R_InitFlatsArray()
{
	flatsarray = (flat_t**)W_Calloc(sizeof(flat_t*) * numlumps, STATICID, 214);
}

// v220-5 used to check/report entry errors with the pnames list
void addPnameErrorToList(char *name)
{
	static int		count  = 0;
	static size_t	length = 0;
	static char		*list = NULL;

	if (name)
	{
		length += 10;
		list = (char*)W_ReallocCLR(list, length, STATICID, 0);
		memcpy(&list[count*10], name, 8);
		count++;
	}
	else if (count)
	{
		int i;

		pr_OFdirect ("\nWAD marker errors =======\n\n");

		for (i = 0; i < count; i++)
			Con_Message("PATCH \"%s\"\n", &list[i*10]);

		pr_OFrepeat ("\n", "=", 25, 2);
	}
}

//===========================================================================
// R_ChangeFlatGlow
//	v220-33 Glow management.
//===========================================================================
void R_ChangeFlatGlow()
{
	if (r_bloodglow == bloodglowcurrent_state) return;
	
	bloodglowcurrent_state = r_bloodglow;

	for (int i = 0; i < numlumps; i++)
	{
		if (flatsarray[i] && flatsarray[i]->flflags & (FLT_SLIME|FLT_BLOOD|FLT_WATER))
		{
			if (r_bloodglow)
				flatsarray[i]->flflags |= FLT_GLOW;
			else
				flatsarray[i]->flflags &= ~FLT_GLOW;
		}
	}
}

//===========================================================================
// R_GetFlatPtr
//	Flat management.
//===========================================================================
flat_t *R_GetFlat(int lumpnum)
{
	flat_t *f;

	if (lumpnum < 0 || lumpnum >= numlumps)
	{
		Con_Error("Invalid flat lumpnum %i", lumpnum);
	}

	if (flatsarray[lumpnum]) return flatsarray[lumpnum];

	// Check dodgy wad construction
	if (lumpinfo[lumpnum].groupMarker != LGT_FLATS)
	{
		int		width, height;
		int		type = DoomLumpIdentify(lumpnum, &width, &height, DLI_GRAPHIC);

		if (type == LMP_FLAT || type == LMP_PNG || type == LMP_JPG)
		{
			lumpinfo[lumpnum].groupMarker = LGT_FLATS;
		}
		else
		// Eliminate special case: console flat
		if (   type != LMP_PATCH
			&& !(!strcmp(lumpinfo[lumpnum].name, "#CONBAK#")
			&& lumpinfo[lumpnum].handle == RISEN3DWADhandle))
		{
			Con_Error("Invalid flat lumpnum %i\n"
				"%8s %s %s %s : w %i h %i size %i",
				lumpnum,
				lumpinfo[lumpnum].name,
				typedesc[lumpinfo[lumpnum].lumptype],
				typedesc[type],
				grpdesc[lumpinfo[lumpnum].groupMarker],
				width, height, lumpinfo[lumpnum].size);
		}
	}

	f = flatsarray[lumpnum] = (flat_t*)W_Calloc(sizeof(flat_t), STATICID, 214);

	f->lump = lumpnum;
	f->flat_translation = lumpnum;
	f->flflags = 0;

	// Set the default width and height
	f->width  = 64;
	f->height = 64;

	memset(f->color.rgb, 0xff, 3);
	f->detail.tex = 0;
	f->group = 0;
	f->gltex = 0;

	return f;
}

int R_GetFlatFlags(int lumpnum)
{
	return R_GetFlat(lumpnum)->flflags;
}

// Passive
RDBOOL R_CheckFlatFlag(int lumpnum, int value)
{
	return (R_GetFlatFlags(lumpnum) & value) > 0;
}

// Active
RDBOOL R_CheckFlatGlowFlag(int lumpnum)
{
	if (!(r_texglow && BDMFX) || outsidefog || (underwaterfog && underfakefloor))
		return false;

	return (R_GetFlatFlags(lumpnum) & FLT_GLOW) > 0;
}

void R_SetNewTexMemory(int extrapatches, float scale, int width, int height)
{
	size_t		size;

	texturetranslation = (int*)W_ReallocCLR(texturetranslation, (numtextures + 2) * sizeof(int), STATICID, 305);
	texArray  = (texture_t**)W_ReallocCLR(texArray, (numtextures + 1) * sizeof(texture_t*), STATICID, 306);
	texscales = (texscale_t*)W_ReallocCLR(texscales, (numtextures + 1) * sizeof(texscale_t), STATICID, 307);

	texscales[numtextures].texscalex = texscales[numtextures].texscaley = scale;

	texturetranslation[numtextures] = numtextures;

	size = sizeof(texture_t) + sizeof(texpatch_t) * extrapatches;
	texArray[numtextures] = (texture_t*)W_Calloc(size, STATICID, 108);

	numtextures++;
}

//===========================================================================
// R_InitTextures
//	Initializes the texture list with the textures from the world map.
//  Modified for R3D
//===========================================================================

void R_InitTextures (void)
{
	maptexture_t *mtexture;
	texture_t	 *texture;
	mappatch_t	 *mpatch;
	texpatch_t	 *patch;

	int			i , j;
	int			*maptex, *maptex1;
	char		name[10];	// Needs 9 but made 10 to align
	char		*names, *name_p;
	int			*patchlookup;
	int			nummappatches;
	int			offset;
	uint		maxoff;
	int			numtextures1;
	int			*directory;
	RDBOOL		noscalebmp = ArgExists("-noscalebmp");

	int			*maptex2 = NULL, numtextures2 = 0;
	uint		maxoff2 = 0;

	float		dotinc, dot = 0.f;

	char *special_lumps[] =
	{
		"#NOTEXT#",
		"#BLACK#",
		NULL
	};

	//---------------------------------

	name[8] = 0;

	// Load the patch names from the PNAMES lump.
	names         = (char*)M_CacheLumpNum(W_GetNumForName("PNAMES"), NULL, true);
	nummappatches = LONG(*(int*)names);
	name_p        = names + 4;
	patchlookup   = (int*)M_Malloc(nummappatches * sizeof(*patchlookup), 93);

	for (i = 0; i < nummappatches; i++)
	{
		strncpy (name, name_p + i*8, 8);
		name[8] = 0;

		// Check if within HI_START or TX_START markers
		patchlookup[i] = W_CheckTextureNumForName(name);

		if (patchlookup[i] < 0)	// Not a hires texture
		{
			patchlookup[i] = W_CheckTextureMissingNumForName(name);

			if (patchlookup[i] >= 0 && lumpinfo[patchlookup[i]].groupMarker != 1)
				addPnameErrorToList(name);
		}
	}

	M_Free(names);

	// Load the map texture definitions from TEXTURE1/2.
	i = W_GetNumForName ("TEXTURE1");
	maptex = maptex1 = (int*)M_CacheLumpNum(i, &maxoff, false);
	numtextures1	 = *maptex;
	directory		 = maptex + 1;

	if ((i = W_CheckNumForName ("TEXTURE2")) != -1)
	{
		maptex2 = (int*)M_CacheLumpNum (i, &maxoff2, true);
		numtextures2 = *maptex2;
	}

	numtextures = numtextures1 + numtextures2;

	// Assign memory including any specials found.
	// Translation table for global animation.
	texturetranslation = (int*)W_Calloc((numtextures + 1) * sizeof(int), STATICID, 109);
	// The main texture array
	texArray  = (texture_t**)W_Calloc(numtextures * sizeof(texture_t*), STATICID, 110);
	// texscale_t parms could go in texture_t but have left it as is.
	texscales = (texscale_t*)W_Calloc(numtextures * sizeof(texscale_t), STATICID, 111);

	p2swd("Loading      : %i textures.\n", numtextures);

	dotinc = numtextures/72.f;

	for (i = 0 ; i < numtextures ; i++, directory++)
	{
		if (i > (int)dot)
		{
			pr_OFdirect(".");
			dot += dotinc;
		}

		texturetranslation[i] = i;

		if (i == numtextures1)
		{	// Start looking in second texture file.
			maptex = maptex2;
			maxoff = maxoff2;
			directory = maptex+1;
		}

		offset = *directory;
		if ((uint)offset > maxoff)
			Con_ErrorD("R_InitTextures: bad texture directory");

		mtexture = (maptexture_t *) ( (byte *)maptex + offset);

		if (mtexture->patchcount < 1)
		{
			Con_Error ("R_InitTextures: invalid patchcount %i texnum %i",
				mtexture->patchcount, i);
			dot = 10000.f;
		}

		texture = texArray[i] = (texture_t*)W_Calloc(sizeof(texture_t) 
			+ sizeof(texpatch_t)*(mtexture->patchcount-1), STATICID, 123);

		texture->width		= mtexture->width;
		texture->height		= mtexture->height;
		texture->patchcount	= mtexture->patchcount;
		texscales[i].texscalex = texscales[i].texscaley = 1.0f;

		if (!noscalebmp)
		{
			byte sx = mtexture->ScaleX != 8 ? mtexture->ScaleX : 0;
			byte sy = mtexture->ScaleY != 8 ? mtexture->ScaleY : 0;
			texscales[i].texscalex = sx ? 8.0f/(float)sx : 1.0f;
			texscales[i].texscaley = sy ? 8.0f/(float)sy : 1.0f;
		}

		memcpy (texture->name, mtexture->name, 8);
		mpatch = &mtexture->patches[0];
		patch  = &texture->patches[0];

		for (j = 0; j < texture->patchcount; j++, mpatch++, patch++)
		{
			int		p, n = mpatch->patch;

			if (n < 0 || n >= nummappatches)
				Con_ErrorD("Invalid patch: Check the correct IWAD is being used");

			if ( (p = patchlookup[n]) < 0 || p >= numlumps)
			{
#ifdef _DEBUG
				p2swd("\nR_InitTextures: Missing patch in texture %s", texture->name);
#endif
				texture->patchcount = 0;
				ge_inf.missingpatch++;
				dot = 10000.f;
			}
			// v220-5 Check the patch is valid
			else if (DoomLumpIdentify(p, NULL, NULL, DLI_GRAPHIC) == LMP_PATCH)
			{
				uint	size;
				patch_t	*ptr = (patch_t*)M_CacheLumpNum(p, &size, true);

				if (!CheckPatch(ptr, ptr->xoffset, ptr->yoffset, texture->width, size))
				{
#ifdef _DEBUG
					p2swd("\nR_InitTextures: Invalid texture %s", texture->name);
#endif

					texture->patchcount = 0;
					ge_inf.invalidpatch++;
					dot = 10000.f;
				}
				M_Free(ptr);
			}

			if (!texture->patchcount) break;

			patch->patch = p;

			if (j == 0)
			{
				n = texture->patchcount;
				texture->groupMarker = lumpinfo[p].groupMarker;
			}

			patch->originx = mpatch->originx;
			patch->originy = mpatch->originy;
		}
	}

	pr_OFdirect("\n");

	M_Free(maptex1);
	M_Free(maptex2);
	M_Free(patchlookup);

	// v220-4
	if (ge_inf.missingpatch || ge_inf.invalidpatch)
	{
extern RDBOOL bypassAdvisory;
		char msg[256];
		
		pr_OFdirect ("\n");
		Sys_AdvisoryHeader("SYSTEM ADVISORY ");

		if (ge_inf.invalidpatch)
		{
			sprintf(msg, "ERROR: pwad has invalid patches\n\n");
		}
		else
		{
			sprintf(msg,
				"ERROR: pwad has %i missing patches\n"
				"This probably means the wrong iwad is being used or you have\n"
				"mixed pwads in a pwads folder for different games. Either\n"
				"remove any surrogate pwad or change the iwad being used.\n",
				ge_inf.missingpatch);
		}

		pr_OFdirect (msg);
		bypassAdvisory = true;
#ifdef _DEBUG
		if (ge_inf.invalidpatch)
		{
			Sys_MessageBox(msg, true);
			Sys_AdvisoryHeader(NULL);
			Sys_Quit(false);
		}

		if (ge_inf.missingpatch && !Sys_CriticalMessage(msg, false))
		{
			pr_OFdirect ("User chose to exit\n\n");
			Sys_AdvisoryHeader(NULL);
			Sys_Quit(false);
		}
#else
		if (ge_inf.missingpatch > 10 || ge_inf.invalidpatch)
		{
			Sys_MessageBox(msg, true);
			Sys_AdvisoryHeader(NULL);
			Sys_Quit(false);
		}
#endif

		pr_OFdirect ("User chose to continue\n\n");
		Sys_AdvisoryHeader(NULL);
	}

	// Check for LGT_TX types, these are not in the PNAMES list
	for (i = 0; i < numlumps; i++)
	{
		if (lumpinfo[i].groupMarker == LGT_TX || lumpinfo[i].groupMarker == LGT_HI)
		{
			int width, height;
			int	type = DoomLumpIdentify(i, &width, &height, DLI_GRAPHIC);

			if (type == LMP_PATCH || type == LMP_PNG)
			{
				// Realloc arrays for new texture definition.
				R_SetNewTexMemory(0, 1.0f, width, height);

				texture = texArray[numtextures - 1];

				patch = &texture->patches[0];
				patch->patch   = i;
				patch->originx = 0;
				patch->originy = 0;

				texture->width  = width;
				texture->height = height;
				texture->patchcount	= 1;
				texture->groupMarker = lumpinfo[i].groupMarker;
				strncpy(texture->name, lumpinfo[i].name, 8);
				texture->txf_flags |= TXF_NOHIRES;
			}
		}
	}
	// Temporarily store current number of textures
	numstandardtextures = numtextures;
	
	// Check for special_lumps in the Risen3D.wad.
	for (i = 0; special_lumps[i]; i++)
	{
		int width, height;
		int lumpnum = W_CheckRisen3DWADNumForName(special_lumps[i], true);
		int	type = DoomLumpIdentify(lumpnum, &width, &height, DLI_GRAPHIC);

		if (type == LMP_PATCH || type == LMP_PNG)
		{
			if (!strcmp(special_lumps[i], "#NOTEXT#"))
				notexTEXNUM = numtextures;
			if (!strcmp(special_lumps[i], "#BLACK#"))
				blacktexnum = numtextures;

			// Add new texture definition.
			R_SetNewTexMemory(0, 1.0f, width, height);

			texture = texArray[numtextures - 1];

			patch = &texture->patches[0];
			patch->patch = lumpnum;
			patch->originx = 0;
			patch->originy = 0;

			texture->width  = width;
			texture->height = height;
			texture->patchcount	= 1;
			strncpy(texture->name, special_lumps[i], 8);
			texture->txf_flags |= TXF_NOHIRES;
			texture->groupMarker = LGT_PNAMES;
		}
		else
			Con_ErrorD("Risen3D.wad is outdated or damaged");
	}

	if (!notexTEXNUM || !blacktexnum)
		Con_ErrorD("Error Risen3D.wad: wrong version in use");

	// Set up switches and anims
	P_InitPicAnims();

	// Used by routine that deallocates transient 3D scaled textures.
	numstandardtextures = numtextures;
}

//===========================================================================
// R_TextureFlags
//===========================================================================
short R_TextureFlags(int texture)
{
	return texArray[texturetranslation[texture]]->txf_flags;
}

//===========================================================================
// R_InitLumpTexInfo
//===========================================================================
void R_InitLumpTexInfo(void)
{
	// Allocate one info per lump.  Use win mem list R3D
	lumptexinfo = (lumptexinfo_t*)W_Calloc(sizeof(lumptexinfo_t) * numlumps, STATICID, 112);
}

//===========================================================================
// R_InitTranslationTables
//===========================================================================
void R_InitTranslationTables(void)
{
	byte	*transLump;
	int		i, num, entries = 3 * 7;

	// Allocate default translation tables
	translationtables = (byte*)W_Calloc(256 * entries + 255, STATICID, 10);
	translationtables = (byte*)(((int)translationtables+255) & ~255);

	// check for a specific lump
	if ((num = W_CheckNumForName("trantbl0")) >= 0)
	{
		for (i = 0; i < entries; i++)
		{
			if (num + i < numlumps)
			{
				// Use trantbl0 to overwrite the default tables
				transLump = (byte*)M_CacheLumpNum(num + i, NULL, true);
				memcpy(translationtables + i*256, transLump, 256);
				M_Free(transLump);
			}
			else
			{
				Con_Message("Bad number of trantbl0 entries.\n");
				break;
			}
		}
	}
	else
	{
		// translate just the 16 green colors
		for (i = 0 ; i < 256 ; i++)
		{
			if (i >= 0x70 && i<= 0x7f)
			{
				// map green ramp to gray, brown, red
				translationtables[i    ] = 0x60 + (i&0xf);
				translationtables[i+256] = 0x40 + (i&0xf);
				translationtables[i+512] = 0x20 + (i&0xf);
			}
			else
			{
				// Keep all other colors as is.
				translationtables[i] = translationtables[i+256] 
					= translationtables[i+512] = i;
			}
		}
	}
}

//===========================================================================
// AddUnknownName - R3D: Store list of names with no corresponding lump
//  v220-4 revised
//===========================================================================
RDBOOL AddUnknownName(char *name)
{
	static int	unknownslots = 0;

	char		name8[10];
	int			i;

	memset (name8, 0, 9);
	strncpy(name8, name, 8);

	for(i = 0; i < unknownname.count; i++)
	{
		if (!_stricmp(&unknownname.name[i * 9], name8))
			return true;
	}

	if ((unknownname.count + 1) >= unknownslots)
	{
		unknownslots += 32;

		unknownname.name = (char*)W_ReallocCLR(unknownname.name, 9 * unknownslots, STATICID, 218);
		unknownname.flat_substituted = (int*)W_ReallocCLR(unknownname.flat_substituted, unknownslots * sizeof(RDBOOL), STATICID, 218);
	}

	memcpy(&unknownname.name[9 * unknownname.count++], name8, 9);

	return false;
}

int wallfornamecount = 0;

//===========================================================================
// AddFlatForWallName - R3D: Store list of flats using wall textures
//===========================================================================
static RDBOOL AddFlatForWallName(char *name, int lumpnum)
{
	typedef struct {
		char name[9];
	} wallflat_t;

	static wallflat_t *wallflats;

	static int wallnameslots = 0;

	char	name8[9];
	int		i, found = false;

	//----------------------------------

	memset(name8, 0, 9);
	strncpy(name8, name, 8);

	for(i = 0; i < wallfornamecount; i++)
	{
		if (!_stricmp(wallflats[i].name, name8))
			return true;
	}

	if ((wallfornamecount + 1) >= wallnameslots)
	{
		wallnameslots += 32;

		wallflats = (wallflat_t*)W_Realloc(wallflats,
			sizeof(wallflat_t) * wallnameslots, STATICID, 219);
	}

	// Store the name
	memcpy(wallflats[wallfornamecount++].name, name8, 9);

	return false;
}

//===========================================================================
// R_FlatNumForName - modified for R3D
//===========================================================================
int	R_FlatNumForName (char *name)
{
	int i = W_CheckFlatNumForName(name);

	if (i < 0)
	{
		if ((i = R_CheckTextureNumForName(name)) >= 0)
		{
			int	num = texArray[i]->patches[0].patch;

			AddFlatForWallName(name, num);

			return num;
		}

		// v220-5 check for any unset due to pname or marker errors
		for (i = 0; i < numlumps; i++)
		{
			if (!_stricmp(lumpinfo[i].name, name))
			{
				if (lumpinfo[i].lumptype == LMP_PATCH) return i;

				if (lumpinfo[i].lumptype == LMP_UNDEF)
				{
					int type = DoomLumpIdentify(i, NULL, NULL, DLI_GRAPHIC);

					if (type == LMP_FLAT || type == LMP_PATCH || type == LMP_PNG)
					{
						addPnameErrorToList(name);
						return i;
					}
				}
			}
		}

		if ((i = W_CheckRisen3DWADNumForName("#NOFLAT#", true)) < 0) // jav use this one
			Con_Error("R_FlatNumForName: neither %.8s nor #NOFLAT# found", name);

		if (!AddUnknownName(name))	// b211-0 Add line feed
			Con_Message("R_FlatNumForName: %8.8s not found, #NOFLAT# substituted\n", name);
	}
	return i;
}

//===========================================================================
// R_CheckTextureNumForName - modified for R3D
//===========================================================================
int	R_CheckTextureNumForName(char *name)
{
	int		i;

	if (!name[0] || (name[0] == '-' && !name[1]))		// no texture marker
		return 0;

	for (i = 0 ; i < numtextures ; i++)
	{
		if ((texArray[i]->groupMarker == LGT_TX	|| texArray[i]->groupMarker == LGT_HI)// These take precedence
			&& !strnicmp (texArray[i]->name, name, 8) )
			return i;
	}
		
	for (i = 0 ; i < numtextures ; i++)
	{
		if (!strnicmp (texArray[i]->name, name, 8) )
			return i;
	}
		
	return -1;
}

//===========================================================================
// R_CheckDetailTexNumForName - modified for R3D
//===========================================================================
int	R_CheckDetailTexNumForName(char *name)
{
	int		i;

	if (name[0] == '-' && !name[1])		// no texture marker
		return 0;
		
	for (i = 0 ; i < numtextures ; i++)
	{
		if (!strnicmp (texArray[i]->name, name, 8)
			&& lumpinfo[texArray[i]->patches[0].patch].groupMarker != LGT_DETAIL)
		{
			return i;
		}
	}
		
	return -1;
}

int W_CheckLumpNumForName(char *name)
{
	lumpinfo_t *lump_p = lumpinfo + numlumps; // Scan backwards

	while (--lump_p >= lumpinfo)
	{
		if (!strnicmp(lump_p->name, name, 8))
		{
			return lump_p - lumpinfo;
		}
	}

	return -1;
}

//===========================================================================
// Add_NewTexture - v220-33: Add a new wall texture
//===========================================================================
int Add_NewTexture(char *name, int lumpnum, int *type)
{
	texture_t	*texture;
	texpatch_t	*patch;
	int			width, height; 

	*type = DoomLumpIdentify(lumpnum, &width, &height, DLI_GRAPHIC);

	if ( !(*type >= LMP_PNG && *type <= LMP_PATCH) )
		return -1;

	// Realloc memory for new texture definition.
	R_SetNewTexMemory(0, 1.0f, width, height);
	newtexcount++;	// Keep track of number of textures added

	texture = texArray[numtextures - 1];
	texturetranslation[numtextures - 1] = numtextures - 1;

	patch = &texture->patches[0];
	patch->patch = lumpnum;
	patch->originx = 0;
	patch->originy = 0;

	texture->width  = width;
	texture->height = height;
	texture->patchcount	= 1;
	strncpy(texture->name, name, 8);
	texture->txf_flags = TXF_NOHIRES|TXF_DELETE;

	if (*type == LMP_FLAT) texture->txf_flags |= TXF_FLAT;

	return numtextures - 1;
}

//===========================================================================
// R_TextureNumForName - modified for R3D
//  v220-6 skytex switch added. Sunder.wad has custom RSKY1/2 tetures
//         but they have not been put in the pnames list by the editor.
//===========================================================================
int	R_TextureNumForName (char *name, RDBOOL skytex)
{
	int		zero = R_CheckTextureNumForName("AASHITTY");	// v220-31
	int		n = R_CheckTextureNumForName(name);
	RDBOOL	found = false;

	if (n < 0)
	{
		if (!skytex)
		{
			if (W_CheckColormapForName(name) >= 0)
			{
				Con_Message("R_TextureNumForName: \"%8.8s\" colour map not on a 242 linedef\n", name);
				return blacktexnum;
			}

			if (!strnicmp(name, "TRAN", 4))
				return blacktexnum;

			if (zero >= 0 && !strnicmp(name, "0", 1))
				return zero;
		}

		if (skytex || (found = AddUnknownName(name)) == false)
		{
			int type, lump = W_CheckLumpNumForName(name);

			// Is there a flat with this name?
			if (lump >= 0
				&&
				(  skytex
				|| !(lumpinfo[lump].groupMarker == LGT_FLATS
				||   lumpinfo[lump].groupMarker == LGT_PNAMES
				||   lumpinfo[lump].groupMarker == LGT_SPRITES)) )
			{
				n = Add_NewTexture(name, lump, &type);
				AddUnknownName(name);
				// Note: unknownname.count will have been incremented
				unknownname.flat_substituted[unknownname.count - 1] = true;
				if (lumpinfo[lump].groupMarker == LGT_FLATS)
					unknownname.flat_count++;
			}
		}
	}

	if (n < 0) n = R3D_edit ? notexTEXNUM : blacktexnum;
		
	return n;
}

//===========================================================================
// R_TextureNameForNum
//===========================================================================
char* R_TextureNameForNum(int num)
{
	if (num < 0 || num >= numtextures)
		Con_Error("Texture number %i out of range %i", num, numtextures);
		
	return texArray[num]->name;
}

//===========================================================================
// R_LumpNameForNum
//===========================================================================
char* R_LumpNameForNum(int num)
{
	if (num < 0 || num >= numlumps)
		Con_Error("Lump number %i out of range %i", num, numlumps);
		
	return lumpinfo[num].name;
}

#pragma optimize("", on)
