//-----------------------------------------------------------------------------
//
// Copyright (C) 2005 Jaakko Kernen
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r_main.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified/extended for R3D
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:	Base level rendering refresh functions
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <math.h>

#include "de_base.h"
#include "de_console.h"
#include "de_system.h"
#include "de_network.h"
#include "de_render.h"
#include "de_refresh.h"
#include "de_graphics.h"
#include "de_audio.h"
#include "de_misc.h"

#include "r_data.h"
#include "doom/p_local.h"

#include "r3d_compat.h"
#include "r3d_movsec.h"
#include "r3d_edit.h"

// MACROS ------------------------------------------------------------------

// $smoothplane: Maximum speed for a smoothed plane.
#define MAX_SMOOTH_PLANE_MOVE	(64*FRACUNIT)

#define FAKELIMIT			0x10000
#define FAKEVIEWLIMIT		0x40000

// TYPES -------------------------------------------------------------------

typedef struct viewer_s {
	fixed_t		x, y, z;
	angle_t		angle;
	float		pitch;
} viewer_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

RDBOOL	G_SetScriptedFog(sector_t *s);
void	G_ResetScriptedFog(void);
void	CheckEditWarc(void);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

fixed_t playermaxmove = MAXMOVE;

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern int	cpossubsec;
extern int	cpossec;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int			viewangleoffset = 0;
int			validcount = 1;	// increment every time a check is made
fixed_t		viewx, viewy, viewz;
fixed_t		viewxOffset = 0, viewyOffset = 0, viewzOffset = 0;
angle_t		viewangle;
float		viewpitch;			// player->lookdir, global version
fixed_t		viewcos, viewsin;

int			extralight;			// bumped light from gun blasts
int			skyflatnum;
int			wasskyflatnum;
int			viewSCRheight = 200;

plrparm_t	ppt;	// R3D
int			plr_viewgroup;

RDBOOL	resetNextViewer = true;

double		dbViewx;
double		dbViewy;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static viewer_t lastSharpView[2];
static float	lastSharpFrameTime;

// CODE --------------------------------------------------------------------

//===========================================================================
// R_Init
//	One-time initialization of the refresh daemon. Called by DD_Main.
//	GL has not yet been inited.
//===========================================================================
void R_Init(void)
{
	r_detail = has_detail;
	R_InitFlatsArray();
	R_InitLumpTexInfo();
	R_InitTextures();
}

void R_Init2(void)
{
	skyflatnum = R_FlatNumForName("F_SKY1");
	wasskyflatnum = skyflatnum;

	if (   (ge_inf.use_sky_box & GE_SKYBOXCHECK) == GE_SKYBOXUSE
		&& lumpinfo[skyflatnum].handle != IWADhandle)
	{
		ge_inf.use_sky_box &= ~GE_SKYBOXUSE;
		ge_inf.use_sky_box |= GE_NOSKYBOXMESSAGE;
	}

	R_InitModels();
	R_InitTranslationTables();
	Def_PostInit();	// Check for detail textures
}

//===========================================================================
// R_InterpolateViewer
//===========================================================================
void R_InterpolateViewer
	(viewer_t *start, viewer_t *end, viewer_t *out, float pos)
{
	float inv = 1 - pos;

	out->x = (fixed_t)(inv*start->x + pos*end->x);
	out->y = (fixed_t)(inv*start->y + pos*end->y);
	out->z = (fixed_t)(inv*start->z + pos*end->z);
	out->angle = (angle_t)(start->angle + pos*((int)end->angle - (int)start->angle));
	out->pitch = inv*start->pitch + pos*end->pitch;
}

//===========================================================================
// R_SetViewPos
//===========================================================================
void R_SetViewPos(viewer_t *v)
{
	// v220-20 (double) values used in the clipper
	viewx = v->x; dbViewx = (double)viewx;
	viewy = v->y; dbViewy = (double)viewy;
	viewz = v->z;
	viewangle = v->angle;
	viewpitch = v->pitch;

	fvx = FIX2FLT(viewx);
	fvy = FIX2FLT(viewz);
	fvz = FIX2FLT(viewy);
}

//===========================================================================
// R_CheckViewerLimits
//	The elements whose difference is too large for interpolation will be 
//	snapped to the sharp values.
//===========================================================================
void R_CheckViewerLimits(viewer_t *src, viewer_t *dst)
{
	if (   abs(dst->x - src->x) > MAXMOVE
		|| abs(dst->y - src->y) > MAXMOVE) 
	{
		src->x = dst->x;
		src->y = dst->y;
		src->z = dst->z;
	}
	if(abs((int)dst->angle - (int)src->angle) >= ANG45) 
		src->angle = dst->angle;
}

//===========================================================================
// R_SetupFrame
//	Prepare rendering the view of the given player.
//===========================================================================
static void R_SetupFrame(ddplayer_t *player)
{
	viewer_t	sharpView, smoothView;
	float		nowTime = Sys_GetTimef();
	float		timePos, mt, tm;
	int			i, n, tableAngle;
	int			*ptr;

	dplookdir = player->lookdir;

	if (R3D_edit)
	{
		cpossubsec = ppt.viewplrmo->subsector - subsectors;
		cpossec = ppt.view_subsector->sector - sectors;
	}

	sharpView.angle = player->mo->angle + viewangleoffset;

	sharpView.pitch = player->lookdir;
	sharpView.x = player->mo->x + viewxOffset;
	sharpView.y = player->mo->y + viewyOffset;
	sharpView.z = player->viewz + viewzOffset;

	// Check that the viewz doesn't go too high or low.
	if(sharpView.z > player->mo->ceilingz - 4*FRACUNIT)
		sharpView.z = player->mo->ceilingz - 4*FRACUNIT;
	if(sharpView.z < player->mo->floorz + 4*FRACUNIT)
		sharpView.z = player->mo->floorz + 4*FRACUNIT;

	if (resetNextViewer || stats.fpsnum < 35) // b211-0
	{
		resetNextViewer = false;

		// Just view from the sharp position.
		R_SetViewPos(&sharpView);
		lastSharpFrameTime = nowTime;
		memcpy(&lastSharpView[0], &sharpView, sizeof(viewer_t));
		memcpy(&lastSharpView[1], &sharpView, sizeof(viewer_t));
		mpos.tm = mpos.mt = 0;

		// $smoothplane: Reset the plane height trackers.
		ptr = movsec_list;		// b211-0 Use the R3D movsec list

		for (i = 0, n = *ptr++; i < n; i++, *ptr++)
		{
			sector_t	 *sc = &sectors[*ptr];

			sc->visceilingoffset = sc->visflooroffset = 0;

			// Reset the old Z values.
			sc->oldfloor  [0] = sc->oldfloor  [1] = sc->floorheight;
			sc->oldceiling[0] = sc->oldceiling[1] = sc->ceilingheight;

			sc->smoothfloor = sc->floorheight;		// b211-0
			sc->smoothceil  = sc->ceilingheight;	// b211-0
		}

		// v220-5 water mod smoothing
		for (i = 0; i < watersec_count; i++)
		{
			sector_t	 *sc = &sectors[wsec[i].secn];

			sc->visflooroffset = 0;

			// Reset the old Z values.
			sc->oldfloor[0] = sc->oldfloor[1] = wsec[i].floorht;
			sc->smoothfloor = wsec[i].floorht;
		}
		
		// v220-4 Reset the remaining scroll values
		for (i = 0; i < scsm.fnum; i++)		// floor
		{
			sectors[scsm.f[i].s].flooroffx = scsm.f[i].oldx = scsm.f[i].dx = scsm.f[i].sdx;
			sectors[scsm.f[i].s].flooroffy = scsm.f[i].oldy = scsm.f[i].dy = scsm.f[i].sdy;
		}

		for (i = 0; i < scsm.cnum; i++)		// ceiling
		{
			sectors[scsm.c[i].s].ceiloffx = scsm.c[i].oldx = scsm.c[i].dx = scsm.c[i].sdx;
			sectors[scsm.c[i].s].ceiloffy = scsm.c[i].oldy = scsm.c[i].dy = scsm.c[i].sdy;
		}

		for (i = 0; i < scsm.snum; i++)		// side
		{
			sides[scsm.s[i].s].textureoffset = scsm.s[i].oldx = scsm.s[i].dx = scsm.s[i].sdx;
			sides[scsm.s[i].s].rowoffset     = scsm.s[i].oldy = scsm.s[i].dy = scsm.s[i].sdy;
		}
	}
	else 
	{
		static int	oldgtic = 0;
		sector_t	*sc;
		int			diff;

		if(gametic != oldgtic)
		{

			// The game tic has changed which means we have an updated
			// sharp camera position. The position, however, is at the 
			// beginning of the tic and we are most likely not at a sharp
			// tic boundary in time. We will update lastSharpFrameTime
			// to store the time of the new sharp position and move the
			// viewer positions one step back in the buffer. The effect
			// of this will be that [0] is the previous sharp position
			// and [1] is the current one.
			oldgtic = gametic;
			lastSharpFrameTime = (float)((int)nowTime);	// Lose the fractional part
			memcpy(&lastSharpView[0], &lastSharpView[1], sizeof(viewer_t));
			memcpy(&lastSharpView[1], &sharpView       , sizeof(viewer_t));
			R_CheckViewerLimits(lastSharpView, &sharpView);

			// $smoothplane: Roll the height tracker buffers.
			ptr = movsec_list;		// b211-0

			for (i = 0, n = *ptr++; i < n; i++, ptr++)
			{
				sc = &sectors[*ptr];

				// v220-5 Don't smooth moving sectors which are modulated here.
				//        This will be done by the water mod smoothing
				if (sc->specFlags & SF_MOVFLOOR && !(sc->specFlags2 & SF2_NOFAKESMOOTH))
				{
					sc->oldfloor[0] = sc->oldfloor[1];
					sc->oldfloor[1] = sc->floorheight;

					diff = abs(sc->oldfloor[0] - sc->oldfloor[1]);

					if (diff >= MAX_SMOOTH_PLANE_MOVE)
					{
						// Too fast: make an instantaneous jump.
						sc->oldfloor[0] = sc->oldfloor[1];
					}
				}

				if (sc->specFlags & SF_MOVCEIL)
				{
					sc->oldceiling[0] = sc->oldceiling[1];
					sc->oldceiling[1] = sc->ceilingheight;

					diff = abs(sc->oldceiling[0] - sc->oldceiling[1]);

					if (diff >= MAX_SMOOTH_PLANE_MOVE ||	// Too fast
						sc->planes[PLN_CEILING].speed >		// Check won't go through floor
						sc->ceilingheight - sc->floorheight)
					{
						// Make an instantaneous jump.
						sc->oldceiling[0] = sc->oldceiling[1];
					}
				}
			}

			// v220-5 water mod smoothing
			for (i = 0; i < watersec_count; i++)
			{
				sc = &sectors[wsec[i].secn];

				sc->oldfloor[0] = sc->oldfloor[1];
				sc->oldfloor[1] = wsec[i].floorht;

				diff = abs(sc->oldfloor[0] - sc->oldfloor[1]);

				if (diff >= MAX_SMOOTH_PLANE_MOVE)
				{
					// Too fast: make an instantaneous jump.
					sc->oldfloor[0] = sc->oldfloor[1];
				}
			}

			// v220-4 smooth floor scroll
			for (i = 0; i < scsm.fnum; i++)
			{
				scsm.f[i].oldx = scsm.f[i].dx; scsm.f[i].dx = scsm.f[i].sdx;
				scsm.f[i].oldy = scsm.f[i].dy; scsm.f[i].dy = scsm.f[i].sdy;
			}

			// v220-4 smooth ceiling scroll
			for (i = 0; i < scsm.cnum; i++)
			{
				scsm.c[i].oldx = scsm.c[i].dx; scsm.c[i].dx = scsm.c[i].sdx;
				scsm.c[i].oldy = scsm.c[i].dy; scsm.c[i].dy = scsm.c[i].sdy;
			}

			// v220-4 smooth wall scroll
			for (i = 0; i < scsm.snum; i++)
			{
				scsm.s[i].oldx = scsm.s[i].dx; scsm.s[i].dx = scsm.s[i].sdx;
				scsm.s[i].oldy = scsm.s[i].dy; scsm.s[i].dy = scsm.s[i].sdy;
			}
		}

		if (nowTime - lastSharpFrameTime > 1.0f)
		{
			// This'll synchronize our frame timing with the game tics.
			lastSharpFrameTime = nowTime - 1.0f;
		}

		// Calculate the smoothed camera position, which is somewhere between
		// the previous and current sharp positions. This introduces slight
		// delay (max. 1/35 sec) to the movement of the smoothed camera.
		timePos = (float)(nowTime - lastSharpFrameTime);

		R_InterpolateViewer(lastSharpView, &sharpView, &smoothView, timePos);
		R_SetViewPos(&smoothView);

		mt = 1.0f - timePos, tm = timePos - 1.0f;
		mpos.mt = mt, mpos.tm = tm;

		// $smoothplane: Set the visible offsets.
		ptr = movsec_list;		// b211-0

		for (i = 0, n = *ptr++; i < n; i++, ptr++)
		{
			sc = &sectors[*ptr];

			if (sc->specFlags & SF_MOVFLOOR && !(sc->specFlags2 & SF2_NOFAKESMOOTH))
			{
				sc->visflooroffset = (fixed_t)(sc->oldfloor[0]*mt + sc->floorheight*tm);
				sc->smoothfloor = sc->floorheight + sc->visflooroffset;	// b211-0
			}

			if (sc->specFlags & SF_MOVCEIL)
			{
				sc->visceilingoffset = (fixed_t)(sc->oldceiling[0]*mt + sc->ceilingheight*tm);
				sc->smoothceil = sc->ceilingheight + sc->visceilingoffset;	// b211-0
			}
		}

		// v220-5 water mod smoothing
		for (i = 0; i < watersec_count; i++)
		{
			sc = &sectors[wsec[i].secn];

			sc->visflooroffset = (fixed_t)(sc->oldfloor[0]*mt + wsec[i].floorht*tm);
			sc->smoothfloor = wsec[i].floorht + sc->visflooroffset;
		}

		// v220-4 smooth floor scroll
		for (i = 0; i < scsm.fnum; i++)
		{
			sectors[scsm.f[i].s].flooroffx = scsm.f[i].sdx + (fixed_t)(scsm.f[i].oldx*mt + scsm.f[i].sdx*tm);
			sectors[scsm.f[i].s].flooroffy = scsm.f[i].sdy + (fixed_t)(scsm.f[i].oldy*mt + scsm.f[i].sdy*tm);
		}

		// v220-4 smooth ceiling scroll
		for (i = 0; i < scsm.cnum; i++)
		{
			sectors[scsm.c[i].s].ceiloffx = scsm.c[i].sdx + (fixed_t)(scsm.c[i].oldx*mt + scsm.c[i].sdx*tm);
			sectors[scsm.c[i].s].ceiloffy = scsm.c[i].sdy + (fixed_t)(scsm.c[i].oldy*mt + scsm.c[i].sdy*tm);
		}

		// v220-4 smooth wall scroll
		for (i = 0; i < scsm.snum; i++)
		{
			sides[scsm.s[i].s].textureoffset = scsm.s[i].sdx + (fixed_t)(scsm.s[i].oldx*mt + scsm.s[i].sdx*tm);
			sides[scsm.s[i].s].rowoffset     = scsm.s[i].sdy + (fixed_t)(scsm.s[i].oldy*mt + scsm.s[i].sdy*tm);
		}
	}

	P_UpdateDynLights();	// v220-31
	ppt.fake_floorheight = ppt.viewer_fake_sector >= 0 ?
		sectors[ppt.viewer_fake_sector].smoothfloor : DDMININT;

	if (player->mo)
	{
		// v220-31 reduce near clip if player's dead in case near to a line
		if (player->mo->health <= 0) reduceNearClip = true;

		// v220-31 check player's in map (mainly for development purposes)
		//         can give false result if outside a sloped boundary line
		//         until vacating the extent of the ssec's box.
		ppt.inmap = player->mo->x >= player->mo->subsector->ssbox[BOXLEFT]
				 && player->mo->x <= player->mo->subsector->ssbox[BOXRIGHT]
				 && player->mo->y >= player->mo->subsector->ssbox[BOXBASE]
				 && player->mo->y <= player->mo->subsector->ssbox[BOXTOP];
	}
	// v220-35 check vdoor links
	if (vdoornum)
	{
		for (i = 0; i < vdoornum; i++)
		{
			if (vdoormo[i][0] && vdoormo[i][1])
				vdoormo[i][1]->z = vdoormo[i][0]->z;
		}
	}

	extralight = player->extralight;
	tableAngle = viewangle>>ANGLETOFINESHIFT;
	viewsin = finesine[tableAngle];
	viewcos = finecosine[tableAngle];
	validcount++;
}

//===========================================================================
// P_FakeSectorCheck - R3D
//  Prevents rendering problem when viewz is at water level in fake sector.
//  Marginally inefficient also setting 'waterlevel' here but who'd notice?
//===========================================================================
static void P_FakeSectorCheck(int noclip)
{
	RDBOOL	fogset = false;
	int		fsec;

	fakefilter = 0;

	underwater = underfakefloor = abovefakeceil = onsurface = 0;
	waterlevel = flying ? 4 : 0;	// default
	playermaxmove = MAXMOVE;

	if (ppt.viewer_fake_sector >= 0)
	{
		sector_t	*sc = &sectors[ppt.viewer_fake_sector];
		fixed_t		ht = sc->smoothfloor;
		fixed_t		diff = viewz - ht;
		fixed_t		factor = FRACUNIT;

		underfakefloor = viewz <= ht;

		if (abs(diff) < FAKEVIEWLIMIT)
		{
			reduceNearClip = true;

			// Add hysteresis of FAKELIMIT - ensures viewz will not be at level
			// where the ceiling/floor is between the top and bottom of the viewport.
			if (diff > 0)
			{
				viewz = ht + FAKELIMIT; // head above fake ceiling
				underfakefloor = false;
			}
			else
			{
				viewz = ht - FAKELIMIT; // head below
				underfakefloor = true;
			}
			fvy = FIX2FLT(viewz);
		}

		if (underfakefloor)
		{
			if (R3D_ScriptedMap)
				fogset = G_SetScriptedFog(sc);
		}
		else if (viewz > sc->smoothceil
			&& ppt.view_subsector->sector->ceilingheight > sc->ceilingheight)
		{
			abovefakeceil = viewz >= sc->smoothceil;
		}

		fakefilter = underfakefloor ? sc->bottommap : abovefakeceil ? sc->topmap : sc->midmap;

		// Set onsurface, underwater and waterlevel

		// If flying is active then 4 will be returned unless condition 1 to 3 applies.
		// Water heights: 1 = waist, 2 = chest, 3 = eyes, 4 = flying

		if (underfakefloor)
		{
			// If there's a colormap or the ceil flat is a splash 
			// type then we assume the player is underwater.
			if (fakefilter || sc->specFlags2 & SF2_HASWATERFLATS)
			{
				underwater = true;
				waterlevel = waterdynamics ? 3 : 0;
				factor = 8*FRACUNIT;

				// Apply a filter if none exists
				if (!fakefilter && !R3D_ScriptedMap)
					fakefilter = 0x80004fa5;	// Use WATERMAP default
			}
		}
		else if (waterdynamics
			&& (diff = ppt.fake_floorheight - ppt.viewplrmo->z) > 0
			&& R_CheckFlatFlag(ppt.view_subsector->ssfloorpic, FLT_SPLASH))
		{
			waterlevel = (diff < (ppt.viewplrmo->height>>1)) ? 1 : 2;

			if (waterlevel == 1)
			{
				factor = 4*FRACUNIT;
				onsurface = diff < 8*FRACUNIT && ppt.viewplrmo->z - sc->smoothfloor > 16*FRACUNIT;
			}
			else
				factor = 6*FRACUNIT;
		}

		if (waterlevel & 0x3 && !noclip)
			playermaxmove = FixedDiv(MAXMOVE, factor);

		// v220-31 Allow player to fly out of a fake sector when waterdynamics is active
		if (flying && (!waterdynamics || !(waterlevel & 0x2) || noclip))
			waterlevel = 4;
	}
	else if ( ((fsec = ppt.view_subsector->floorsec) >= 0 || (fsec = ppt.view_subsector->lfsec) >= 0)
		&& R_CheckFlatFlag(sectors[fsec].floorpic, FLT_SPLASH))
	{
		int ht = sectors[fsec].floorheight;

		if (  (viewz >  ht -  0x70000 || (ppt.viewplrmo->health <= 0 && ht - ppt.view_subsector->sector->floorheight <= plrviewheight) )	// v220-33 don't go under floor on death
			&& viewz < (ht += 0x50000)
			&& ht < ppt.view_subsector->sector->ceilingheight)
		{
			viewz = ht;
			fvy = FIX2FLT(viewz);
			reduceNearClip = true;

			if (!flying) waterlevel = waterdynamics ? 2 : 0;

			if (waterlevel == 2)
				playermaxmove = FixedDiv(MAXMOVE, 6*FRACUNIT);
		}
	}
	// v220-10 check view not near to floor limit
	else if (ppt.viewplrmo->dplayer->deltaviewheight < 0)
	{
		if (viewz < ppt.view_subsector->sector->floorheight + 5*FRACUNIT)
		{
			viewz = ppt.view_subsector->sector->floorheight + 5*FRACUNIT;
			fvy = FIX2FLT(viewz);
		}
	}

	if (!waterdynamics && waterlevel != 4) waterlevel = 0;

	// v220-3 underwater fog support
	if (underwaterfog)
		GL_UseFog(underfakefloor);

	if (!fogset) G_ResetScriptedFog();
}

//===========================================================================
// R_RenderPlayerView
//	Draw the view of the player inside the view window.
//===========================================================================
void R_RenderPlayerView(ddplayer_t *player, int noclip)
{
	// Setup for rendering the frame.
	R_SetupFrame(player);

	// Adjust viewz if necessary, set filter
	P_FakeSectorCheck(noclip);

	if (R3D_edit || ge_inf.devmap) CheckEditWarc();

	PG_InitForNewFrame();
	Rend_PrepareList();
	GL_RenderFrame();
}
