//-----------------------------------------------------------------------------
//
// Copyright (C) 2005 Jaakko Kernen, Jack Vermeulen
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r_model.c
//  Code reworked to speed up loading
//  R_ShutdownModels - removed gmj
//	All allocated memory uses the new win mem routines in place of
//  malloc and is automatically released on exit
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified/extended for Risen3D to speed up model loading
//  and track/report errors in user ded file entries.
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//	shininess distance test removed
//
// Based on: Doomsday v1.7.8
//
// DESCRIPTION:   MD2/DMD loading and setup
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdlib.h>
#include <io.h>
#include <direct.h>
#include <ctype.h>
#include <math.h>

#include "de_base.h"
#include "de_console.h"
#include "de_system.h"
#include "de_graphics.h"
#include "de_refresh.h"
#include "de_play.h"
#include "de_misc.h"

#include "r3d_compat.h"
#include "r3d_edit.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

typedef struct 
{
	float pos[3];
}
vector_t;

// 12-23-2005 put paths individually so we can track better
typedef struct modelpath_s
{
	char  *pathname;
} modelpath_t;

typedef struct skinpath_s
{
    int   loaded;         // 1 = already loaded, 2 = not found
    int   dupcount;       // count of times this is used (FYI only)
    int   idx;            // skin index
    char  *skinname;      // skinname
    char  *modelname;     // modelname
	char  *fullpathname;  // pathname usually same as modelfilename+skinname

} skinpath_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

RDBOOL chkSkinIndex(model_t *mdl, int skin);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern mvertex_t	*mvertices;
extern int			maxnumModelVtx;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// The dummy is used for model zero.

model_t	dummy = { true };

model_t	*modellist[MAX_MODELS] = { &dummy }; 

float	rModelAspectMod = 1/1.2f; //.833334f;
int		totalskins;
int		maxmodelDefs;			// 12-12-2005 v220-35 made global

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static int			skincount = 0;
static int			maxskincount = 0;
static skinpath_t	*skinpathlist = NULL;

static char			modelDirectory[MAXPATH];

static RDBOOL		renorm = false;

static int			modelpathcount = 0;
static modelpath_t	modelpathlist[MAX_MODELPATHS];

static modeldef_t	*modelDefs = NULL;		// 12-12-2005 confusing name changed
static int			nummodelDefs;			// 12-12-2005

static	char *filenameptr;

static RDBOOL		model_warn;

#define NUMVERTEXNORMALS	162

static const float avertexnormals[NUMVERTEXNORMALS][3] = {
#include "anorms.h"
};

static int	mdefCount = 0;

static int missing_modelcount = 0;
static ddstring_t missing_models;

static int missing_skincount = 0;
static ddstring_t missing_skins;

static ddstring_t missing_paths;

// CODE --------------------------------------------------------------------

//===========================================================================
//	Calculate vertex normals. Only called if renorm is true
//===========================================================================
static void R_VertexNormals(model_t *mdl)
{
	int tris = mdl->lodInfo[0].numTriangles;
	int verts = mdl->info.numVertices;
	int i, k, j, n, cnt;
	vector_t *normals, norm;
	dmd_vertex_t *list;
	dmd_triangle_t *tri;

	normals = (vector_t*)M_Malloc(sizeof(vector_t) * tris, 94);

	// Calculate the normal for each vertex

	for (i = 0; i < mdl->info.numFrames; i++)
	{
		list = mdl->frames[i].vertices;

		for (k = 0; k < tris; k++)
		{
			tri = mdl->lods[0].triangles + k;

			// 1) Calculate the surface normal and store the result to *normals[k].pos
			M_PointCrossProduct(list[tri->vertexIndices[0]].vertex,
	                			list[tri->vertexIndices[2]].vertex,
				                list[tri->vertexIndices[1]].vertex, normals[k].pos);

			// 2) Normalise the results
			M_Normalize(normals[k].pos);
		}

		for (k = 0; k < verts; k++)
		{
			memset(&norm, 0, sizeof(norm));

			for (j = 0, cnt = 0; j < tris; j++)
			{
				tri = mdl->lods[0].triangles + j;

				for (n = 0; n < 3; n++)
                {
					if (tri->vertexIndices[n] == k)
					{
						for (cnt++, n = 0; n < 3; n++)
							norm.pos[n] += normals[j].pos[n];

						break;
					}
                }
			}

			if (cnt)	// Expected to be true
			{
				// Calculate the average
				for (n = 0; n < 3; n++) norm.pos[n] /= cnt;

				M_Normalize(norm.pos);  // Normalize it.
				
				memcpy(list[k].normal, norm.pos, sizeof(norm.pos));
			}
		}
	}

	M_Free(normals);
}


//===========================================================================
//	Packed: pppppppy yyyyyyyy. Yaw is on the XY plane.
//===========================================================================
static void UnpackVector(unsigned short packed, float vec[3])
{
	float yaw   = (float)((packed & 511)/512.0f * 2*PI);
	float pitch = (float)(((packed >> 9)/127.0f - 0.5f) * PI);
	float cosp  = (float) cos(pitch);

	vec[VX] = (float) cos(yaw) * cosp;
	vec[VY] = (float) sin(yaw) * cosp;
	vec[VZ] = (float) sin(pitch);	
}


static void storeModelname(char *mpath, char *path)
{
	if (!mpath || _stricmp(mpath, path))
	{
		mpath = (char*)W_Realloc(mpath, strlen(path) + 1, MODELID, 100);
		strcpy(mpath, path);
	}
}

//===========================================================================
//	Check model's been prepared by looking in list  Redone 12-23-2005 jav, gmj
//  if found Returns the index into modellist and the fullpathname
//  if not found, allocates a new one and return new index
//===========================================================================
static int R_GetModel_inList(char *modelname, char* fullpathname, BOOL allocate)
{
	int	 i;
	char szmodelname[MAXFNAME];
	char szext      [MAXEXT];
	
	//-------------------------------------------
	
	// split model with a path into actual model name so that any path is removed
	// this ensures that it's just the model name we store, since a path messes this up
	// (somehow this got removed - via an undo) 1-8-2006
	
	Dir_ConvertSlashes(modelname);								// consistency here
	
	_splitpath(modelname, NULL, NULL, szmodelname, szext);
	strcat(szmodelname, szext);
	
	for(i = 0; i < MAX_MODELS; i++)
	{
		if (   modellist[i] // there is a model
			&& modellist[i]->fullpathName
			&& !_stricmp(modellist[i]->modelName, szmodelname))	// same file
		{
			strcpy(fullpathname, modellist[i]->fullpathName);	// return the full name
			return i;
		}
	}
	
	if (!allocate) return -1;
	
	// does not exist, initialize a new one if allocate is requested	
	for (i = 0; i < MAX_MODELS; i++)
	{
		if (!modellist[i])									// first open slot
		{
			// allocate a new entry
			modellist[i] = (model_t*)W_Calloc(sizeof(model_t), MODELID, 101);
			
			modellist[i]->modelName = (char*)W_Calloc(strlen(szmodelname) + 1, MODELID, 102);
			strcpy(modellist[i]->modelName, szmodelname);	// save the model name
			
			if (fullpathname)	// save the full pathname (usually null first time)
				storeModelname(modellist[i]->fullpathName, fullpathname);
			
			ge_inf.models_count++;
			return i;
		}
	}

	return -1;	// Out of models or just not found
}

//===========================================================================
// Free all memory allocated to skin paths jav 1-8-2006, gmj
//===========================================================================

static void R_FreeSkinPathList(void)
{
    int i;

	for(i = 0; i < skincount; i++)
    {
        M_Free(skinpathlist[i].skinname);     // free skinname
        M_Free(skinpathlist[i].modelname);    // free modelname
        M_Free(skinpathlist[i].fullpathname); // free full path(M_Free allows NULL)
    }
    
	ge_inf.skincount  = skincount;
    
    skincount=0;
}

//===========================================================================
// R_GetSkinName_inList  1-8-2006 rewritten jav, gmj
//	see if skin was aready done
//  if found returns the index into skinpathlist and the fullpathname
//  if not found, allocates a new one and return new index
//===========================================================================
static int R_GetSkinName_inList(char *skinname, char* modelname, char* fullpathname)
{
	int	 i;
    
	//------------------

    Dir_ConvertSlashes(skinname);								// consistency here

	for (i = 0; i < skincount; i++)
    {
        if (!_stricmp(skinpathlist[i].skinname , skinname )		// same skinname
        &&  !_stricmp(skinpathlist[i].modelname, modelname))	// same modelname
        {
            strcpy(fullpathname, skinpathlist[i].fullpathname);	// return the full name
            
            skinpathlist[i].dupcount++;							// count how many times we looked it up

			return i;
        }
    }

	// does not exist, initialize a new one
	if (skincount >= maxskincount)
	{
		maxskincount += 512;
		skinpathlist = (skinpath_t*)M_Realloc(skinpathlist, maxskincount * sizeof(skinpath_t), 42);
	}

     skinpathlist[skincount].skinname  = (char*)M_Calloc(strlen(skinname ) + 1, 315);
     skinpathlist[skincount].modelname = (char*)M_Calloc(strlen(modelname) + 1, 316);

     strcpy(skinpathlist[skincount].skinname , skinname );	// save the skin name
     strcpy(skinpathlist[skincount].modelname, modelname);	// save the model name
     
     skinpathlist[skincount].fullpathname = NULL;
     skinpathlist[skincount].loaded   = 0;    
     skinpathlist[skincount].dupcount = 0;
     skinpathlist[skincount++].idx    = -1;    
    
     return skincount-1;
}
//===========================================================================
// 
//===========================================================================
static void *AllocAndLoad(DFILE *file, int offset, int len)
{
	void *ptr = W_Malloc(len, MODELID, 103);

	F_Seek(file, offset, SEEK_SET);
	F_Read(ptr, len, file);

	return ptr;
}

//===========================================================================
// 
//===========================================================================
static char *findEXT(char *filename)
{
	char ext[MAXEXT];

	_splitpath(filename, NULL, NULL, NULL, ext);

	if (!ext[0]) return NULL;

	return &filename[strlen(filename) - strlen(ext)];
}

//===========================================================================
// R_FindModelFile  redone 12-23-2005, 1-4-2006 1-10-2006 jav, gmj
//
//  removed white space count dir paths (garbage code)
//
//  if isModel = 1
//     look for DMD or MD2 DMD overrides MD2
//     technically users do not have to put DMD or MD2 on model names
//  else
//     look for file given (GL_LoadImage and particles)
//
//  always absolute paths and findfirst
//
//	Returns true if the file was found.
//===========================================================================

static int R_FindModelFile(char *modelpath, char *szinfilename, char *szoutfilename)
{
	char	filebuf[MAXPATHNAMESIZE];
	char	currentpath[MAXPATH];
	
	char	szdir     [MAXDIR];
	char	szfilename[MAXFNAME];
	char	szext     [MAXEXT];
	
	char	szmodelname[MAXFNAME];
	
	char	foundfile[2][MAXPATH]={0,0};
	
	long	hFile;    
	struct	_finddata_t finddata;
	
	// in case of duplicate names this is the priority of names
	
	char	*formats[2] = { ".dmd", ".md2" }; 
	char	*extfile;
	char	*ext;
	
	int		i;
	int		modelpasscount = 0;			// general model directory index
	int		found          = false;
	
	//==============================================
	
	if(!szinfilename || !szinfilename[0]) return false;
	
	Dir_ConvertSlashes(szinfilename);   // 12-23-2005
	
	// get extension now 12-23-2005
	_splitpath(szinfilename, NULL, szdir, szmodelname, szext);
	
	szoutfilename[0]= 0;                // indicate nothing
	
	if (!szext[0])						// check for extension
	{
		strcpy(szoutfilename, ".");		// suppress error messages for pure paths (at caller location)
		return false;                   // no extension
	}
	
	strcpy(filebuf, modelpath);          // 1-4-2006 jav first model path
	strcpy(szfilename, szinfilename);    // 1-4-2006 jav filename
	
	// see if the filename comes with a path   (revised 1-8-2006)
	// normally not done, user probably do this to speed up load a tiny bit if sub model has not been defined
	// but this is at the expense of maintenance issue, so not a good idea
	// could also fix this in DED reading and give this model the right path name - this is easier for now
	if (szdir[0])                       // a dir name exists only if there was a path given
	{
		// if this path includes the base path then copy
		// otherwise extend the path

		if (!strnicmp(szdir, "md2", 3))
		{
			strcpy(filebuf, szdir);					// create first path as given
		}
		else
		{
			strcat(filebuf, szdir);					// add to basic model path given
		}
		
		strcpy(szfilename, szmodelname);
		strcat(szfilename, szext);					// put extension back on filename
	}
	
	// create extension independent search for model name  ".*"
	if ( (extfile = findEXT(szfilename)) != NULL) *extfile = 0;
	strcat(szfilename, ".*");					// generic find
	
	for(;;)
	{
		if(!_stricmp(filebuf,ddRuntimeDir.path)) continue;	// ignore if only root directory

		M_PrependBasePath(filebuf, currentpath);			// add base path to subpath
		strcpy(filebuf,currentpath);						// fully qualified path
		strcat(filebuf, szfilename);						// + filename (.* or actual)
		
		if((hFile = _findfirst(filebuf, &finddata)) != -1)	// found something
		{   
			do
			{
				if ( (ext = findEXT(finddata.name)) )		// find extension
				{	
					for (i = 0; i < 2; i++)					// does it match one of the extension we support?
					{
						if (!_stricmp(ext,formats[i]) )		// matches extension
						{
							strcpy(foundfile[i],filebuf);	// current search name
							extfile = findEXT(foundfile[i]);// find extension
							*extfile=0;						// at the period (always there)
							strcat(extfile,formats[i]);		// + extension found
							
							found++;
							break;
						}
					}
					if (foundfile[0][0]) break;				// DMD overrides all 1-4-2006
				}
			}
			while(!_findnext(hFile, &finddata));
			
			_findclose( hFile );
			
			if (found)
			{  
				// in case of duplicates, pick in priority the one we want
				
				for (i = 0; i < 2; i++)
				{
					if (foundfile[i][0])
					{
						strcpy(szoutfilename, foundfile[i]);    // copy to caller's ptr
						break;                                  // all done
					}
				}
				break;
			}
		}
		
		if (modelpasscount >= modelpathcount) break;             // all paths searched so break
		
		strcpy(filebuf,modelpathlist[modelpasscount].pathname);  // Try the next path
		
		modelpasscount++;
	} 
	
	// if this prints with c_moddebug disabled then user made a mistake since
	// it should be found at once (model path is incorrect)

	if (compat[c_moddebug]		// User aid: if set always print out results
		|| modelpasscount > 0)  // always tell if it took more than 1 search to find (path mistake or order problem)
	{
		char	composite[MAXPATHNAMESIZE];

		strcpy (composite, modelpath);
		strcat (composite, szinfilename);

		if (Str_TokenMatch(&missing_paths, composite))
			return found;
	
		Str_AppendComma(&missing_paths, composite);

		if (found)
		{
			Con_Message(
				"WARNING specified path invalid ---------\n"
				"%3d calls for ; %s\n"
				"  found in    ; %s\n",
				modelpasscount, composite, szoutfilename);
		}
		else
		{
			Con_Message(
				"ERROR file name not found ---------\n"
				"%3d calls for ; %s\n"
				,modelpasscount, composite);
		}
		model_warn = true;
	}

	return found;
}

//===========================================================================
//
//===========================================================================
static void R_MissingModel(const char *filename)
{
	if (Str_TokenMatch(&missing_models, filename))
		return;
	
    missing_modelcount++;
	Str_AppendComma(&missing_models, filename);
}

//===========================================================================
// 12-23-2005 jav redone
//===========================================================================
static int R_ExpandSkinName(char *skin, char *expandedname, char *modelfilename) // 12-23-2005 path fix
{
	int		modelpasscount =  0;             // general model directory 
	int		found          =  0;
	int		i;
	
	// in case of duplicate names this is the search priority for extensions
	char    *formats[] = { ".jpg", ".png", ".tga", ".pcx", NULL };
	
	// 1 for each type that's possible required since duplicate graphics are often found. 
	char	foundfile  [4][MAXPATH] = {0, 0, 0, 0};	// 12-29-2005
	
	char	currentPath	[MAXPATH];
	char	filebuf		[MAXPATH];
	char	skinbuf		[MAXPATH];
	
	char	szdir		[MAXPATH];	// b211-0
	char	szext		[MAXEXT];
	
	long    hFile;    
	struct  _finddata_t finddata;
	
	directory_t mydir;
	
	//---------------------------------------
	
	strcpy(currentPath, modelfilename);
	
	R_GetModel_inList(modelfilename, currentPath, false); // just see if it's there already, returns full path in currentpath
	
	// directory from modelfile name
	
	Dir_FileDir(currentPath, &mydir);    // get first directory to search from model filename
	
	expandedname[0] = 0;                 // default failure return
	
	// make sure we get just the name and no paths attached (a few mistakes by users)
	
	Dir_ConvertSlashes(skin);
	
	_splitpath(skin, NULL, szdir, skinbuf, szext);
	
	strcat(skinbuf,".*");				// extension independent find

	// note works with absolute directories now (make sure that relative is already resolved before this point
	
	strcpy(filebuf, mydir.path);		// start with model directory path
	
    // 1-9-2006 put in same stuff as in models to catch skins with paths
    // if skin has a pathname then that is used for first search dir

    if (szdir[0])                    // a dir name exists only if there was a path given
    {
        // if this path is "md2" prepend with base
        // otherwise extend the modeldirectory with the path given

        if (!strnicmp(szdir,"md2",3))
        {
			M_PrependBasePath(szdir, filebuf); // add base path to subpath
        }
        else
        // use model path + path given
        {
            strcpy(filebuf,	modelDirectory);
            strcat(filebuf,"\\");
            strcat(filebuf,szdir);              // + basic model path given
        }
    }

    // Works with absolute directories but ensure that
	// relative ones are resolved before this point

	for(;;)
	{
		if(_stricmp(filebuf,ddRuntimeDir.path))	// ignore only root directory  
		{
			strcpy(currentPath,filebuf);		// current directory
			strcat(filebuf, skinbuf);			// skinname 
			
			if((hFile = _findfirst(filebuf, &finddata)) != -1)	// found something
			{  
				do
				{
					char *ext = findEXT(finddata.name);			// find extension
					
					if (ext)
					{  
						// does it match one of the extension we support
						for (i = 0; formats[i]; i++)
						{
							if (!_stricmp(ext, formats[i]))
							{
								strcpy(foundfile[i],currentPath);   // current path
								strcat(foundfile[i],finddata.name); // name found
								found++;
								break;
							}
						}
						if (foundfile[0][0]) break;	// PNG overrides all 12-29-2005
					}
				}
				while(!_findnext(hFile, &finddata));
				
				_findclose( hFile );
				
				if (found)
				{  
					// in case of duplicates, pick in priority the one we want
					for (i = 0; formats[i]; i++)
					{
						if (foundfile[i][0])
						{
							strcpy(expandedname, foundfile[i]);
							break;								// all done
						}
					}
					break;
				}
			}
		}
		
		if (modelpasscount >= modelpathcount) break;			// all paths done
		
		strcpy(filebuf,modelpathlist[modelpasscount].pathname);	// current path
		
		modelpasscount++;
	}
	
	return found;
}

//===========================================================================
//	Registers a new skin name	12-23-2005 jav redone to track missing
//===========================================================================
static int R_RegisterSkin(char *skin, char *modelfilename, char *fullpathout)
{
	int idx;
	int found;
    int index;
	
	// Has a skin name been provided?
	
	if (!skin[0]) return -1;

	if ( (index = R_GetSkinName_inList(skin, modelfilename, fullpathout)) < 0)
		return -1;
	
	if (skinpathlist[index].loaded == 2) return -1;		// already found missing
	
	if (skinpathlist[index].loaded == 1)				// already found
	{
		return skinpathlist[index].idx;    
	}
	// not loaded yet

	// next automatically gets PNG, TGA, PCX (in that order)
	
	found = R_ExpandSkinName(skin, fullpathout, modelfilename);
	
    if (!strnicmp(fullpathout, "md2", 3))
    {
		char realpath[MAXPATHNAMESIZE];
		strcpy(realpath, ddRuntimeDir.path);
		strcat(realpath, fullpathout);
		strcpy(fullpathout, realpath);
	}

	skinpathlist[index].fullpathname = (char*)M_Calloc(strlen(fullpathout) + 1, 317);

	strcpy(skinpathlist[index].fullpathname, fullpathout);	// save path name found

	if (found && (idx = GL_GetSkinTexIndex(fullpathout)) >= 0)
	{		
		skinpathlist[index].loaded = 1;    
		skinpathlist[index].idx    = idx;
		return idx;
	}

	// Not found or failed to get skin index i.e. idx < 0
	{
		char workfile[1024];
		int	 len;

		skinpathlist[index].loaded = 2;	// indicate missing or bad

		strcpy(workfile, skin);			// input string (for clarity)

		len = strlen(workfile);
		if (len > 4 && workfile[len - 4] == '.')
			workfile[len - 4] = 0;

		strcat(workfile, " : ");
		strcat(workfile, modelfilename);	// output string
		
		missing_skincount++;
	
		Str_AppendComma(&missing_skins, workfile);
	}

#if 0 // debug only: show skins processed
		Con_Message("R_RegisterSkin %3d %32s m=%32s path=%s\n",idx,skin,modelfilename,fullpathout);
#endif
	
	return -1;
}

//===========================================================================
// v220-31 Read buf for the last occurrence of a '\' or '/' char and
//         return a ptr to the next char or buf if neither is found.
//===========================================================================
static char *namePtr(char *buf)
{
	char *ptr1 = strrchr(buf, '\\');
	char *ptr2 = strrchr(buf, '/');

	if (!ptr1 && !ptr2) return buf;
	if ( ptr1 && !ptr2) return ++ptr1;
	if (!ptr1 &&  ptr2) return ++ptr2;
	if ( ptr2 > ptr1)   return ++ptr2;
	return ++ptr1;
}

//===========================================================================
//
//===========================================================================
static void R_LoadModelMD2(DFILE *file, model_t *mdl)
{
	md2_header_t oldhd;
	dmd_header_t *hd = &mdl->header;
	dmd_info_t *inf   = &mdl->info;
	byte *frames;
	int i, k, c;

	// Read the header

	F_Read(&oldhd, sizeof(oldhd), file);
	
	// Convert it to DMD

	hd->magic   = MD2_MAGIC;
	hd->version = 8;
	hd->flags   = 0;
	mdl->vertexUsage  = NULL;

	inf->skinWidth    = oldhd.skinWidth;
	inf->skinHeight   = oldhd.skinHeight;
	inf->frameSize    = oldhd.frameSize; 
	inf->numLODs      = 1;
	inf->numSkins     = oldhd.numSkins;
	inf->numTexCoords = oldhd.numTexCoords;
	inf->numVertices  = oldhd.numVertices;
	inf->numFrames    = oldhd.numFrames;
	inf->offsetSkins  = oldhd.offsetSkins;
	inf->offsetTexCoords = oldhd.offsetTexCoords;
	inf->offsetFrames    = oldhd.offsetFrames;
	inf->offsetLODs      = oldhd.offsetEnd; // Doesn't exist

	mdl->lodInfo[0].numTriangles     = oldhd.numTriangles;
	mdl->lodInfo[0].numGlCommands    = oldhd.numGlCommands;
	mdl->lodInfo[0].offsetTriangles  = oldhd.offsetTriangles;
	mdl->lodInfo[0].offsetGlCommands = oldhd.offsetGlCommands;

	inf->offsetEnd = oldhd.offsetEnd;

	// Allocate and load the various arrays

	mdl->texCoords = (dmd_textureCoordinate_t*)AllocAndLoad(file, 
                                  inf->offsetTexCoords, 
		                          sizeof(md2_textureCoordinate_t) * inf->numTexCoords);

	mdl->lods[0].triangles = (dmd_triangle_t*)AllocAndLoad(file, 
	                                      mdl->lodInfo[0].offsetTriangles,
		                                  sizeof(md2_triangle_t) * mdl->lodInfo[0].numTriangles);

	// The frames need to be unpacked

	frames = (byte*)AllocAndLoad(file, inf->offsetFrames, 
		                        inf->frameSize * inf->numFrames);

	mdl->frames = (dmd_frame_t*)W_Malloc(sizeof(dmd_frame_t) * inf->numFrames, MODELID, 104);

	for(i = 0; i < inf->numFrames; i++)
	{
		md2_packedFrame_t *pfr = (md2_packedFrame_t*) (frames + inf->frameSize * i);
		md2_triangleVertex_t *pVtx;
		dmd_vertex_t *vtx;
		memcpy(mdl->frames[i].name, pfr->name, sizeof(pfr->name));
		mdl->frames[i].vertices = (dmd_vertex_t*)W_Malloc(sizeof(dmd_vertex_t) * inf->numVertices, MODELID, 105); 

		// v220-4 preset prior to model rendering
		if (inf->numVertices > maxnumModelVtx) maxnumModelVtx = inf->numVertices;

		// Translate each vertex.

		for(k = 0, vtx = mdl->frames[i].vertices, pVtx = pfr->vertices;
			k < inf->numVertices; k++, vtx++, pVtx++)
		{
			memcpy(vtx->normal, avertexnormals[pVtx->lightNormalIndex],	sizeof(float) * 3);
			for(c = 0; c < 3; c++)
			{
				vtx->vertex[c] = pVtx->vertex[c] * pfr->scale[c] + 
					pfr->translate[c];
			}

            vtx->vertex[2] *= rModelAspectMod;	// Aspect mod reversal
		}
	}

	W_Free(frames, MODELID, 512);

	mdl->lods[0].glCommands = (int*)AllocAndLoad(file,
		                                   mdl->lodInfo[0].offsetGlCommands,
                                    	   sizeof(int) * mdl->lodInfo[0].numGlCommands);

	// Load skins

	mdl->skins = inf->numSkins ? (dmd_skin_t*)W_Calloc(sizeof(dmd_skin_t) * inf->numSkins, MODELID, 106) : 0;

	if (!mdl->skins) pr_OFinsert("ERROR: model %s has no skins\n", filenameptr);

	F_Seek(file, inf->offsetSkins, SEEK_SET);

	char buf[66];
	memset(buf, 0, 66);	// In case some dodgy entry is read ensure a terminating NULL

	for (i = 0; i < inf->numSkins; i++)
    {
		int count = F_Read(buf, 64, file);
		if (count) strcpy(mdl->skins[i].name, namePtr(buf));
    }
}

//===========================================================================
//
//===========================================================================
static void R_LoadModelDMD(DFILE *file, model_t *mo)
{
	dmd_chunk_t chunk;
	char *temp;
	dmd_info_t *inf = &mo->info;
	int i, k, c;
	
	// Read the chunks.

	F_Read(&chunk, sizeof(chunk), file);

	while(chunk.type != DMC_END)
	{
		switch(chunk.type)
		{
		case DMC_INFO: // Standard DMD information chunk.
			F_Read(inf, chunk.length, file);
			break;
			
		default:
			// Just skip all unknown chunks // jav why not just seek past them ? (todolater)
			temp = (char*)M_Malloc(chunk.length, 95);
			F_Read(temp, chunk.length, file);
			M_Free(temp);
		}

		// Read the next chunk header

		F_Read(&chunk, sizeof(chunk), file);
	}
	
	// Allocate and load in the data.

	mo->skins = (dmd_skin_t*)W_Calloc(sizeof(dmd_skin_t) * inf->numSkins, MODELID, 107);
	F_Seek(file, inf->offsetSkins, SEEK_SET);
	
	char buf[66];

	memset(buf, 0, 66);	// In case some dodgy entry is read ensure a terminating NULL

	for (i = 0; i < inf->numSkins; i++)
    {
		int count = F_Read(buf, 64, file);
		strcpy(mo->skins[i].name, namePtr(buf));
    }

	mo->texCoords = (dmd_textureCoordinate_t*)AllocAndLoad(file, inf->offsetTexCoords, sizeof(dmd_textureCoordinate_t) * inf->numTexCoords);
	temp       = (char*)AllocAndLoad(file, inf->offsetFrames, inf->frameSize * inf->numFrames);
	mo->frames = (dmd_frame_t*)W_Malloc(sizeof(dmd_frame_t) * inf->numFrames, MODELID, 108);

	for(i = 0; i < inf->numFrames; i++)
	{
		dmd_packedFrame_t *pfr = (dmd_packedFrame_t*)(temp + inf->frameSize * i);
		dmd_packedVertex_t *pVtx;
		dmd_vertex_t *vtx;

		memcpy(mo->frames[i].name, pfr->name, sizeof(pfr->name));
		mo->frames[i].vertices = (dmd_vertex_t*)W_Malloc(sizeof(dmd_vertex_t) * inf->numVertices, MODELID, 109); 
		
		// v220-4 preset prior to model rendering
		if (inf->numVertices > maxnumModelVtx) maxnumModelVtx = inf->numVertices;

        // Translate each vertex
		
        for(k = 0, vtx = mo->frames[i].vertices, pVtx = pfr->vertices;
			k < inf->numVertices; k++, vtx++, pVtx++)
		{
			UnpackVector(pVtx->normal, vtx->normal);

			for(c = 0; c < 3; c++)
			{
				vtx->vertex[c] = pVtx->vertex[c] * pfr->scale[c] + 
					pfr->translate[c];
			}
			
            // Aspect mod reversal
			
            vtx->vertex[2] *= rModelAspectMod;
		}
	}

	W_Free(temp, MODELID, 110);

	F_Seek(file, inf->offsetLODs, SEEK_SET);
	F_Read(mo->lodInfo, sizeof(dmd_levelOfDetail_t) * inf->numLODs, file);
	
	for(i = 0; i < inf->numLODs; i++)
	{
		mo->lods[i].triangles = (dmd_triangle_t*)AllocAndLoad(file,
                                         	mo->lodInfo[i].offsetTriangles,
                                 			sizeof(dmd_triangle_t) * mo->lodInfo[i].numTriangles);

		mo->lods[i].glCommands = (int*)AllocAndLoad(file, 
		                                      mo->lodInfo[i].offsetGlCommands,
                                             sizeof(int) * mo->lodInfo[i].numGlCommands);
	}

	// Determine vertex usage at each LOD level.
	// This code assumes there will never be more than 8 LOD levels.
	// v220-4 in practice it appears always to be 1
	if (inf->numLODs > 1)
	{
		mo->vertexUsage = (char*)W_Calloc(inf->numVertices, MODELID, 111);

		for(i = 0; i < inf->numLODs; i++)
		{
			for(k = 0; k < mo->lodInfo[i].numTriangles; k++)
			{
				for(c = 0; c < 3; c++)
				{
					mo->vertexUsage[mo->lods[i].triangles[k].vertexIndices[c]] |= 1 << i;
				}
			}
		}
	}
}

//===========================================================================
//	Finds the existing model or loads in a new one.  12-23-2005 jav redone
//===========================================================================
static int R_LoadModel(char* modelpath, char *szmodelfilename)
{
	int		i, index;
	model_t	*mdl;
	DFILE	*file;
	
	char	buf[MAXPATHNAMESIZE];      // jav sizes
	char    fullpathname[MAXPATHNAMESIZE]={0};
	
	
	//-------------------------------------------
	
	if(!szmodelfilename[0]) return 0;				// No model specified
	
	// check if it's already there
	index = R_GetModel_inList(szmodelfilename, fullpathname, false);
	
	if (index < 0)									// not found yet
	{  
		// get new index and store name now (avoids duplicate missing so done here)
		
		index = R_GetModel_inList(szmodelfilename, fullpathname, true); 
		
		if (index < 0) 
		{
			Con_Message("Too many models %s\n",szmodelfilename);
			return 0;								// default to first
		}
		
		mdl = modellist[index];						// where it is stored
		
		// fullpathname on return contains full qualified path
		
		if (!R_FindModelFile(modelpath,szmodelfilename, fullpathname))
		{   
			// check not already reported
			
			if (!mdl->loaded
				&&  !fullpathname[0] )				// not just a path that was passed (which is weird)
			{
				R_MissingModel(szmodelfilename);	// lists missing model
			}
			mdl->loaded = 2;						// indicate missing model
			return 0;
		}
		
		mdl->fullpathName = (char*)W_Realloc(mdl->fullpathName, strlen(fullpathname) + 1, MODELID, 112);
		strcpy(mdl->fullpathName, fullpathname);	// store the fullname found
	}
	
	mdl = modellist[index];
	if(mdl->loaded == 2)  return 0;					// just a missing model
	
	if(mdl->loaded) 
	{
#ifdef _DEBUG
		if (verbose) pr_OFinsert("model already loaded %32s %s\n",szmodelfilename, fullpathname);
#endif
		return index;								// Already loaded.
	}
	
	// debug
	
	if (verbose) 
	{
		pr_OFinsert("R_LoadModel %32s %s\n",szmodelfilename, fullpathname);
	}
	
	// Try to open the model file (should find - this can be reduced by keeping handles in structure
	
	if((file = F_Open(fullpathname, "rb")) == NULL)
	{
		// should not occur! this is a mistake in the search code
		
		Con_Message("R_LoadModel %s not found\n", fullpathname);
		return 0;
	}

	filenameptr = fullpathname;
	
	// Now we can load in the data	
	F_Read(&mdl->header, sizeof(mdl->header), file);
	
	switch(mdl->header.magic)
	{
	case MD2_MAGIC:		// MD2
		F_Rewind(file);
		R_LoadModelMD2(file, mdl);
		break;

	case DMD_MAGIC:		// DMD
		R_LoadModelDMD(file, mdl);
		break;

	default:			// Bad magic! Cancel the loading.
		modellist[index] = 0;
		F_Close(file);
		Con_Message("R_LoadModel %s not MD2 or DMD data?\n", fullpathname);
		return 0;
	}
	
	// Determine the full texture name path using the name and the model path	
	for(i = 0; i < mdl->info.numSkins; i++)
	{
		strcpy(buf, mdl->skins[i].name);
		mdl->skins[i].id = R_RegisterSkin(buf, fullpathname, mdl->skins[i].name);
	}

	if (renorm) R_VertexNormals(mdl);		// If renorm recalculate vertex normals
	
	// We're done.
	mdl->loaded = 1;

	storeModelname(mdl->fullpathName, fullpathname);
	
	F_Close(file);
	
	return index;
}

//===========================================================================
//
//===========================================================================
int R_ModelFrameNumForName(int modelnum, char *fname)
{
	int			i;
	model_t		*mdl;

	if(!modelnum) return 0;
	
    mdl = modellist[modelnum];

	for(i = 0; i < mdl->info.numFrames; i++)
	{
		if(!_stricmp(mdl->frames[i].name, fname)) return i;
	}
	return 0;
}

//===========================================================================
//	Returns the appropriate modeldef for the given state.
//    R3D modified to work defensively.
//===========================================================================
modeldef_t *GetStateModel(state_t *st, int select)
{
	modeldef_t *modef = NULL;

	if (st)
	{
		if (st->stflags & ST_NOMODEL) return NULL;

		if (!(st->stflags & (ST_HASMODEL|ST_NOMODEL)) )
		{
			// Note - order must not be changed
			if (  !st->model
				|| st->sprite < 0
				|| st->sprite >= defs.count.sprites.num
				|| (sprites[st->sprite].sprflags & SPR_NOMODEL))
			{
				st->stflags |= ST_NOMODEL;
				return NULL;
			}

			st->stflags |= ST_HASMODEL;
		}

		// Choose the correct selector, or selector zero if the given 
		// one not available.
		if ( (modef = (modeldef_t*) st->model) && select)
		{
			modeldef_t	*mdef;
			int			mosel = select & DDMOBJ_SELECTOR_MASK;

			for (mdef = modef; mdef; mdef = mdef->selectnext)
			{
				if (mdef->select == mosel) return mdef;
			}
		}
	}

	return modef;
}

//===========================================================================
//
//===========================================================================
static dmd_frame_t *R_GetDMDFrame(int model, int frame)
{
	return modellist[model]->frames + frame;
}

//===========================================================================
// Sets x, y, z bounds
//===========================================================================
static void R_GetModelBounds(int model, int frame, float min[3], float max[3])
{
	dmd_frame_t		*mframe = R_GetDMDFrame(model, frame);
	model_t			*mo = modellist[model];
	dmd_vertex_t	*v;
	int				i, k;

	if (!mframe) Con_ErrorD("R_GetModelBounds: bad model/frame");

	for (i = 0, v = mframe->vertices; i < mo->info.numVertices; i++, v++)
	{
		for (k = 0; k < 3; k++)
		{
			if (!i || v->vertex[k] < min[k]) min[k] = v->vertex[k];
			if (!i || v->vertex[k] > max[k]) max[k] = v->vertex[k];
		}
	}
}

//===========================================================================
//	Height range, really ("horizontal range" comes to mind...).
//===========================================================================
static float R_GetModelHRange(int model, int frame, float *top, float *bottom)
{
	float min[3], max[3];

	R_GetModelBounds(model, frame, min, max);

	*top    = max[VZ];
	*bottom = min[VZ];

	return max[VZ] - min[VZ]; 
}

// v220-4
static void R_CalcModelDimension(modeldef_t *mf, float *radius, float *minheight, float *maxheight)
{
	float			min[3], max[3];
	float			minz = .0f, maxz = .0f;
	float			minr = .0f, maxr = .0f;
	float			rax, ray, rbx, rby;
	int				i;
	RDBOOL			isset;

	for (i = 0, isset = false, minz = maxz = 0; i < MAX_FRAME_MODELS; i++)
	{
		if (!mf->sub[i].model) continue;
		
		R_GetModelBounds(mf->sub[i].model, mf->sub[i].frame, min, max);

		rax = min[VX] * mf->scale[VX], ray = min[VY] * mf->scale[VZ];
		rbx = max[VX] * mf->scale[VX], rby = max[VY] * mf->scale[VZ];

		if (!isset)
		{
			minz = min[VZ];
			maxz = max[VZ];
			minr = MIN_OF(rax, ray);
			maxr = MAX_OF(rbx, rby);
			isset = true;
		}
		else
		{
			if (max[VZ] > maxz) maxz = max[VZ];
			if (min[VZ] < minz) minz = min[VZ];
			if (MAX_OF(rbx, rby) > maxr) maxr = MAX_OF(rbx, rby);
			if (MIN_OF(rax, ray) < minr) minr = MIN_OF(rax, ray);
		}
	}

	*radius = (maxr - minr) * 0.5f;
	*minheight = minz * mf->scale[VY];
	*maxheight = maxz * mf->scale[VY];
}

//===========================================================================
// v220-3  Calculate model's min and max heights, inc offset, and max radius
//===========================================================================
static void R_SetModelMaxDimensions(modeldef_t	*mf)
{
	float		radius, minheight, maxheight;

	R_CalcModelDimension(mf, &radius, &minheight, &maxheight);

	mf->zmin	= (fixed_t)(minheight * FRACUNIT);
	mf->zmax	= (fixed_t)(maxheight * FRACUNIT);
	mf->xyradius= (fixed_t)(radius * FRACUNIT);
	mf->zoff	= (fixed_t)(mf->offset[VY] * FRACUNIT);
}

//===========================================================================
//	Scales the given model so that it'll be 'destHeight' units tall.
//	The measurements are based on submodel zero. The scaling is done
//	uniformly! 
//===========================================================================
static void R_ScaleModel(modeldef_t *mf, float destHeight, float offset)
{
	submodeldef_t	*smf = &mf->sub[0];
	float			top, bottom, height;
	float			scale;
	int				i;

	if(!smf->model) return;	// No model to scale!

	// Find the top and bottom heights.

	height = R_GetModelHRange(smf->model, smf->frame, &top, &bottom);
	if(!height) height = 1;
	
	scale = destHeight / height;

	for(i = 0; i < 3; i++) mf->scale[i] = scale;
	mf->offset[VY] = -bottom*scale + offset;
}

//===========================================================================
//
//===========================================================================
static void R_ScaleModelToSprite(modeldef_t *mf, int sprite, int frame)
{
	spritedef_t	*spr = sprites + sprite;

	if (!(spr->sprflags & SPR_NOMODEL))
	{
		int lump = spr->spriteframes[frame & FF_FRAMEMASK].lump[0];    // first view

		int	off = spritelumps[lump].slyoffset - spritelumps[lump].height;

		if(off < 0) off = 0;
		R_ScaleModel(mf, (float)spritelumps[lump].height, (float)off);
	}
}

//===========================================================================
//
//===========================================================================
static float R_GetModelVisualRadius(modeldef_t *mf)
{
	float min[3], max[3];

	if(!mf->sub[0].model) return 0;

	// Use the first submodel's bounds
	R_GetModelBounds(mf->sub[0].model, mf->sub[0].frame, min, max);

	// Half of the average of width and depth.
	return (mf->scale[VX]*(max[VX] - min[VX]) + mf->scale[VZ]*(max[VZ]-min[VZ]))/3.5f;
}

//===========================================================================
//	Create a new modeldef or find an existing one. There can be only one
//	model definition associated with a state/intermark pair. 
//===========================================================================
modeldef_t *R_GetModelDef(int state, float intermark, int select)
{
	int			i;
	modeldef_t	*md;

	if(state < 0 || state >= count_states.num) return NULL; // Not a valid state.

	// First try to find an existing model

	for(i = 0; i < nummodelDefs; i++)
    {
		if(modelDefs[i].state     == &states[state] 
		&& modelDefs[i].intermark == intermark
		&& modelDefs[i].select    == select) 
		{
			// This is an exact match; use it

			modelDefs[i].order = mdefCount++;
			return modelDefs + i;
		}
    }
	
	// This is impossible, but checking won't hurt...

	if(nummodelDefs >= maxmodelDefs) return NULL;

	md = modelDefs + nummodelDefs++;
	memset(md, 0, sizeof(*md));

	// Set initial data

	md->order     = mdefCount++;
	md->state     = &states[state];
	md->intermark = intermark;
	md->select    = select;
	return md;
}

//===========================================================================
// R_SetupModel  jav- revisions to skip empty data
//
//	Creates a modeldef based on the given DED info.
//	A pretty straightforward operation. No interlinks are set yet.
//	Autoscaling is done and the scale factors set appropriately.
//	After this has been called for all available Model DEDs, each
//	State that has a model will have a pointer to the one with the
//	smallest intermark (start of a chain). 
//===========================================================================
static void R_SetupModel(ded_model_t *def)
{
	char dummybuf[MAXPATH];  // full path name not required here, just for func (simplify)

	modeldef_t     *modef;
	int            common_flags      = Def_EvalFlags(defs.model_flags);
	int            model_scope_flags = Def_EvalFlags(def->flags) | common_flags;

	ded_submodel_t *subdef;
	submodeldef_t  *sub;

	int    i, k;
    int    statenum = Def_GetStateNum(def->state);
	float  min[3];
    float  max[3];

    char*  extmodel;

	//-------------------------------------------

	if(statenum < 0)
	{
		Con_Message("R_SetupModel: Undefined state '%s'\n", def->state);
		return; 
	}

	modef = R_GetModelDef(statenum + def->off, def->intermark, def->selector);

	if(!modef) 
	{
		Con_Message("R_SetupModel: Invalid: %s +%i\n", def->state, def->off);
		return; // Can't get a modef, quit!
	}

	// Init modef info (state & intermark already set).

	modef->def   = def;
	modef->flags = model_scope_flags;

	for(i = 0; i < 3; i++)
	{
		modef->offset[i] = def->offset[i];
		modef->scale[i]  = def->scale[i];
	}

	modef->offset[VY] += defs.model_offset;	// Common Y axis offset.
	modef->scale[VY]  *= defs.model_scale;	// Common Y axis scaling.
	modef->resize      = def->resize;
	modef->skintics    = def->skintics;

	if ( !(modef->scale[VX] || modef->scale[VY] || modef->scale[VZ]) )
	{
		Con_Message("R_SetupModel: model with state %s has no scale\n", def->state);
		return;	// Scaled to 0??
	}

	for(i = 0; i < 2; i++)
	{
		modef->interrange[i] = def->interrange[i];
	}

	if(modef->skintics < 1) modef->skintics = 1;
	
	// Submodels.

    for(i = 0, subdef = def->sub, sub = modef->sub; 
		i < MAX_FRAME_MODELS; 
		i++, subdef++, sub++)
	{
		sub->model = 0;

        if (subdef->filename.path[0] == 0) continue;

        // 1-4-2006 some "modelname" are not model names at all, just a path?

        extmodel = findEXT(subdef->filename.path);	// find extension
        if (!extmodel) continue;					// it's not a model 

        sub->model = R_LoadModel(def->model_path,subdef->filename.path);  // load the model
        if(!sub->model) continue;

		sub->frame = R_ModelFrameNumForName(sub->model, subdef->frame);
		sub->framerange = subdef->framerange;

		// Frame range must always be greater than zero.
		if(sub->framerange < 1) sub->framerange = 1;

		// Submodel-specific flags cancel out model-scope flags!
		sub->flags = model_scope_flags ^ Def_EvalFlags(subdef->flags);
		sub->skin  = subdef->skin;
		sub->skinrange = subdef->skinrange;

		// Skin range must always be greater than zero.
		if(sub->skinrange < 1) sub->skinrange = 1;

		// Offset within the model.
		for(k = 0; k < 3; k++) sub->offset[k] = subdef->offset[k];

		sub->alpha     = (byte) (subdef->alpha * 255);
		sub->shinyskin = R_RegisterSkin(subdef->shinyskin,subdef->filename.path,dummybuf); // always use a buf vs NULL = no full path return required
	}

	// Do scaling, if necessary

	if(modef->resize)
	{
		R_ScaleModel(modef, modef->resize, modef->offset[VY]);
	}
	else 
    if(modef->sub[0].flags & MFF_AUTOSCALE)
	{
		int sprNum   = Def_GetSpriteNum(def->sprite.id);
		int sprFrame = def->spriteframe;

		/*
		if(sprNum < 0) // No sprite ID given?
		{
			sprNum   = modef->state->sprite;
			sprFrame = modef->state->frame;
		}
        
		R_ScaleModelToSprite(modef, sprNum, sprFrame);
		*/
		// v220-31 replace with
		if (sprNum >= 0 || (sprNum = modef->state->sprite) >= 0)
		{
			R_ScaleModelToSprite(modef, sprNum, sprFrame);
		}
	}

	if(!modef->state->model)		// Associate this modeldef with its state
	{
		// No modef; use this
		modef->state->model = modef;
	}
	else							// Must check intermark; smallest wins!
	{
		modeldef_t *other = (modeldef_t*) modef->state->model;

		if (modef->intermark <= other->intermark	// Should never be ==
			&& modef->select == other->select
			|| modef->select < other->select)		// Smallest selector?
        {
			modef->state->model = modef;
        }
	}

	if (   modef->flags & MFF_PARTICLE_SUB1			// Particle offset?
		&& modef->sub[1].model)
	{
		R_GetModelBounds(modef->sub[1].model, modef->sub[1].frame, min, max);

		for(i = 0; i < 3; i++)	// Apply the various scalings and offsets.
		{
			// The coordinate systems are mixed up badly...
			// MD2:        +Z is up
			// Offset/DGL: +Y is up
			// Game:       +Z is up

			k = i==1 ? 2 : i==2 ? 1 : 0;

			modef->ptcoffset[i] =
				((max[k] + min[k])/2
				+ modef->sub[1].offset[i]) * modef->scale[i] 
				+ modef->offset[i];
		}
	}
	else
	{
		memset(modef->ptcoffset, 0, sizeof(modef->ptcoffset));
	}

	// Calculate visual radius for shadows
	if (def->shadowradius && !R3D_ScriptedMap)
		modef->visualradius = def->shadowradius;
	else
		modef->visualradius = R_GetModelVisualRadius(modef);
}

//===========================================================================
// R_FindGraphicFile  Rewritten 1-10-2006 jav, gmj
//
// input name is usually a full path and it should find it right away
// if not, searches all the model paths for a texture
//
// looks only for exactly the file name given 
//
// NOTE: general search loop is done just in case we want to add a bit
//       more generic searching later as in other code.
//
//	Returns true if the file was found.
//===========================================================================

int R_FindGraphicFile(char *szinfilename, char *szoutfilename)
{
	char filebuf[MAXPATHNAMESIZE];
	char currentpath[MAXPATH];
	
	char szdrive   [MAXDRIVE];
	char szdir     [MAXDIR];
	char szfilename[MAXFNAME];
	char szext     [MAXEXT];
	
	long    hFile;    
	struct  _finddata_t finddata;
	
	int    modelpasscount = 0;			// general model directory index
	int    found          = false;
	
	//-------------------------------------------------
	
	if(!szinfilename || !szinfilename[0]) return false;
	
	Dir_ConvertSlashes(szinfilename);   
	
	_splitpath(szinfilename, szdrive, szdir, szfilename, szext);  
	
	szoutfilename[0] = 0;				// indicate nothing
	
	strcpy(filebuf,szdrive);			// drive
	strcat(filebuf,szdir);				// + first texture path
	
	for(;;)
	{
		if(_stricmp(filebuf,ddRuntimeDir.path))			// ignore only root directory 
		{ 
			M_PrependBasePath(filebuf, currentpath);	// add base path to subpath
			strcpy(filebuf, currentpath);				// fully qualified path
			strcat(filebuf, szfilename);				// + filename
			strcat(filebuf, szext);						// + extension
			
			if((hFile = _findfirst(filebuf, &finddata)) != -1)   // found something
			{   
				strcpy(szoutfilename,filebuf);			// current search name found
				found=1;
				
				_findclose( hFile );
				
				break;
			}
		}
		
		// Try a new path
		
		if (modelpasscount >= modelpathcount) break;             // all paths done
		
		strcpy(filebuf,modelpathlist[modelpasscount].pathname);  // next path
		
		modelpasscount++;
	} 
	
	// if this prints with c_moddebug disabled then user made a mistake since
	// it should be found at once (model path is incorrect)
	
	if (   compat[c_moddebug]	// User aid: if set always print out results
		|| modelpasscount > 0)  // always tell if it took more than 1 search to find (path mistake or order problem)
	{
		if (found)
		{
			Con_Message("%2d found   texture=%-32s %s\n", modelpasscount, szinfilename, szoutfilename);  
		}
		else
		{
			Con_Message("%2d missing texture=%s\n",modelpasscount, szinfilename); 
		}
	}

	return found;
}

//==========================================================================
//	States must be initialized prior to calling
//==========================================================================
void R_InitModels(void)
{
	int			i, k, minsel;
	float		minmark;
	modeldef_t	*me;
	modeldef_t	*other;
	modeldef_t	*closest;
	RDBOOL		error = false;
	
	char buf[300];
	
	if (   !CFX					// No core effects then no models
		|| !modelpathcount		// In case none set
		|| ge_inf.noobjects)	// Edit mode (quick select) option
	{
		return;
	}

	model_warn = false;

	memset(&missing_models, 0, sizeof(ddstring_t));
	memset(&missing_skins , 0, sizeof(ddstring_t));
	memset(&missing_paths , 0, sizeof(ddstring_t));
	
	strcpy(buf, ddRuntimeDir.path);		// root/
	strcat(buf, "md2");					// root/MD2
	
	if ( (_access(buf, 0)) == -1)
	{
		Con_Message("MD2 Model directory %s not found\n",buf); // jav show the complete path
		return;
	}
	
	renorm = (ArgCheck("-renorm") > 0);	// Renormalizing?

	// There can't be more modelDefs than there are DED Models.
	// There can be fewer, though.
	
	maxmodelDefs = defs.count.models.num;
	modelDefs    = (modeldef_t*)W_Malloc(sizeof(modeldef_t) * maxmodelDefs, MODELID, 113);
	nummodelDefs = 0;

	if (!defs.model_scale) defs.model_scale = 1.2f;	// b211-0
	
	Con_Printf("R_InitModels : Initializing MD2 models.\n");
	Con_InitProgress("R_InitModels: Initializing models...", maxmodelDefs);
	
	// Clear the modef pointers of all States.

	for(i = 0; i < count_states.num; i++) 
	{
		states[i].model = NULL;
	}
	
	// Read in the model files and their data, 'n stuff.
	// Use the latest definition available for each sprite ID.
	
	Con_TimingStart();
	
	for(i = 0; i < maxmodelDefs; i++) 
	{
		Con_Progress(i, PBARF_PERCENT);		// check for 5 percent jumps
		
		R_SetupModel(defs.models + i);		// 12-23-2005 include path
	}

	byte *tempcount = (byte*)calloc(skincount, 1);

	for (i = totalskins = 0; i < skincount; i++)
    {
		if (tempcount[i]) continue;

		char *ptr = skinpathlist[i].skinname;
		totalskins++;
		tempcount[i] = 1;

		for (k = i; k < skincount; k++)
		{
			if (!tempcount[k] && !stricmp(ptr, skinpathlist[k].skinname))
			{
				tempcount[k] = 1;
			}
		}
	}

	free(tempcount);
	
	timecache[ITC_MODLOAD] = Con_TimingEnd(NULL);
	
	// Report missing files (should have no duplicates)
	
	if (missing_models.length)
	{
		char *token;
		
		Con_Message("ERROR: %d models not found:\n",missing_modelcount);
		
		i = 1;
		token = strtok(missing_models.str, ",");
		
		while(token)
		{
			Con_Message("%4i)  %s\n", i++, token);
			token = strtok(NULL, ",");
		}
	}
	
    // Report Missing skins (per model)
	// v220-7 revised
	if (missing_skins.length)
	{
		char *token;
		
		Con_Message("ERROR: %d skins not found (jpg, png, tga or pcx):\n", missing_skincount);
		Con_Message("   NUM SKIN NAME : MODEL\n");
		
		i = 1;
		token = strtok(missing_skins.str, ",");
		
		while(token)
		{
			Con_Message("%4i)  %s\n", i++, token);
			token = strtok(NULL, ",");
		}

		error = true;
	}
	
	Con_HideProgress();

	// Create interlinks. Note that the order in which the defs were loaded
	// is important. We want to allow "patch" definitions, right?
	
	// For each modeldef we will find the "next" def.
	
	for(i = 0, me = modelDefs; i < nummodelDefs; i++, me++)
	{
		minmark = 2; // max = 1, so this is "out of bounds".
		closest = NULL;
		
		for(k = 0, other = modelDefs; k < nummodelDefs; k++, other++)
		{
			// Same state and a bigger order are the requirements
			
			if(other->state == me->state 
				&& other->order     > me->order       // Loaded after me.
				&& other->intermark > me->intermark
				&& other->intermark < minmark)
			{
				minmark = other->intermark;
				closest = other;
			}
		}
		me->internext = closest;
	}
	
	// Create selectlinks
	
	for(i = 0, me = modelDefs; i < nummodelDefs; i++, me++)
	{
		minsel = DDMAXINT;
		closest = NULL;

		// Start scanning from the next definition.
		for(k = 0, other = modelDefs; k < nummodelDefs; k++, other++)
		{
			// Same state and a bigger order are the requirements.
			if (   other->state == me->state 
				&& other->order  > me->order          // Loaded after me.
				&& other->select > me->select
				&& other->select < minsel
				&& other->intermark >= me->intermark)
			{
				minsel = other->select;
				closest = other;
			}
		}
		me->selectnext = closest;
	}

	if (model_warn) Con_Message("-----------------\n");

	// v220-4
	for(i = 0, me = modelDefs; i < nummodelDefs; i++, me++)
	{	R_SetModelMaxDimensions(me);
	}

	// the following are no longer needed
	Str_Free(&missing_models);
	Str_Free(&missing_skins);
	Str_Free(&missing_paths);
	R_FreeSkinPathList();

	if (   error
		&& (ge_inf.devmap || stricmp(analysename, "crueld")) )	// CruelD backward compatibility
	{
		char buff[256];

		sprintf(buff,	"Missing model skins. If continuing affected models will be skinned white.\n");

		if (!Sys_CriticalMessage(buff, true))
			Sys_Quit(0);
	}

	// v220-4 preset mvertices used by rend_models.c
	// Added this scheme to prevent constant alloc/dealloc - R3D
	mvertices = (mvertex_t*)W_Realloc(mvertices, sizeof(mvertex_t) * maxnumModelVtx, STATICID, 224);
}

//===========================================================================
// R_AddModelPath jav 12-23-2005 revised completely to get duplicates out
//                               (old code did not remove duplicates)
//
//	Appends or prepends a new path to the list of model search paths.
//	-modeldir prepends
//   DED files append search paths.
//===========================================================================

void R_AddModelPath(char *addPath, RDBOOL append)
{
	int		i, p;
	char    pathBuf[MAXPATH];

	strcpy(pathBuf,addPath);

	// convert slashes, remove trailing blanks make sure it ends with a slash
	Dir_ValidDir(pathBuf);

	if (!_stricmp(pathBuf,ddRuntimeDir.path)) return;		// ignore root directory 

	// Check if path has already been added to the list	
	for (i = 0; i < modelpathcount; i++)
	{
		if (!_stricmp(modelpathlist[i].pathname, pathBuf))	// duplicate
			return;
	}

#if 0 // debug only: show model paths
	Con_Message("R_AddModelPath: %d %s\n",append,pathBuf);
#endif

	if (modelpathcount >= MAX_MODELPATHS)	// Out of paths? unlikely.....
		Con_Error("number of model paths exceeds %d", MAX_MODELPATHS);

	// allocate a new path (use realloc - gmj)
	modelpathlist[modelpathcount].pathname = (char*)W_Realloc(
		modelpathlist[modelpathcount].pathname, strlen(pathBuf) + 1, MODELID, 114);

	if (!append)	// Paths are added to top of list if called from dd_defns
	{
		// shift all the names down 1		
		for (i = modelpathcount - 1; i >= 0; i--)
			strcpy(modelpathlist[i+1].pathname, modelpathlist[i].pathname);
		
		p = 0;	// Add model path to top of list
	}
	else
		p = modelpathcount;	// Append model path to list

	strcpy(modelpathlist[p].pathname, pathBuf);

	modelpathcount++;
}

//===========================================================================
// v220-3 average the sprite frame z-heights
//  Used with torches (see call to this function)
//  The source z offset from state to state is averaged to stop 'bounce'.
//===========================================================================
void R_SetModelTorchCentre(mobj_t *mo)
{
	spritedef_t		*sprdef;
	spriteframe_t	*sprframe;
	modeldef_t		*mf = (modeldef_t*) mo->state->model;
	state_t			*st, *st1;
	float			a = 0.0f;
	int				splnum, count = 0;

	if (!mf || mo->zcenter) return;

	st = st1 = &states[mo->info->spawnstate];

	sprdef = &sprites[mo->sprite];

	while(1)
	{
		float yoffset = 0.0f;

		if (!st) break;

		if (st->light)
		{
			ded_light_t *def = (ded_light_t*) st->light;

			if (def->yoffset) yoffset = def->yoffset;		
		}

		sprframe = &sprdef->spriteframes[st->frame & FF_FRAMEMASK];

		if (sprframe->rotate == 1)
		{
			int i = (R_PointToAngle(mo->x, mo->y) - mo->angle + (unsigned)0x90000000) >> 29;

			if ((unsigned)a >= (unsigned)sprdef->numframes) break;

			splnum = sprframe->lump[i];
		}
		else
		{
			splnum = sprframe->lump[0];
		}

		if (splnum < 0 || splnum >= numspritelumps) break;

		a += (spritelumps[splnum].slyoffset - yoffset);

		count++;

		if ( st->nextstate <= 0 || (st = &states[st->nextstate]) == st1) break;
	}

	if (count > 1)
	{
		a /= (float)count;
		mo->zcenter = a;
		mo->flags3 |= MF3_HASCENTRE;
	}
}

//===========================================================================
//
//===========================================================================
RDBOOL R_CheckModelExists(mobj_t *mo)
{
	return GetStateModel(mo->state, mo->selector) != NULL;
}

//===========================================================================
//	Checks if a model exists for mo based on its state and tics.
//  Sets current modef and also nextmodef (based on interlinks).
//===========================================================================
float R_CheckModelFor(mobj_t *mo, modeldef_t **modef, modeldef_t **nextmodef)
{
	state_t *st = mo->state;
	float	interp =  1.0f;

	// By default there are no models.
	if (nextmodef) *nextmodef = NULL;

	if ( (*modef = GetStateModel(st, mo->selector)) == NULL)
	{
		return -interp; // [-1] indicates no model available.
	}

	// v220-5 if nextmodef unset then call was just to get modef
	if (!nextmodef) return interp;

	// Calculate the currently applicable intermark.
	if (mo->tics > 0 && st->tics >= mo->tics )
		interp -= (mo->tics / (float)st->tics);

	// First find the modef for the interpoint.
	// Intermark is 'stronger' than interrange.

	// Scan interlinks.
	while((*modef)->internext && (*modef)->internext->intermark <= interp)
    {
		*modef = (*modef)->internext;
    }

	// Scale to the modeldef's interpolation range.
	interp = (*modef)->interrange[0] + interp *
			((*modef)->interrange[1] - (*modef)->interrange[0]);

	// What would be the next model? Check interlinks first.
	if ((*modef)->internext) 
    {
		*nextmodef = (*modef)->internext;
    }
	else if (st->nextstate > 0) // Check next state.
	{
		state_t *it = states + st->nextstate;

        // Find the appropriate state based on interrange
		if ((*modef)->interrange[1] < 1)
		{
			modeldef_t	*mdit = NULL;
			int			max = 20; // Let's not be here forever...
			
			// Current modef doesn't interpolate to the end, find the
			// proper destination modef (it isn't just the next one).
			// Scan the states that follow (and interlinks of each).
			while (max-- && it->nextstate > 0)
			{
				if (it->model)
				{
					if (!(mdit = GetStateModel(it, mo->selector)) )
					{
						if (hasDEH)
							Con_ErrorD( "This wad is heavily dehacked and is unsuitable\n"
										"for use with the 'forced models' option\n\n"
										"Untick the 'Force Model Override' option in the\n"
										"Launcher before restarting");
						else
							Con_ErrorD( "PROBLEM : Missing models\n\n"
										"Please ensure all the most recent model packs available\n"
										"from our site have been correctly installed before\n"
										"restarting");
					}

					if (!mdit->interrange[0]) break;
				}

				// Scan interlinks, then go to the next state.
				if (mdit || (mdit = GetStateModel(it, mo->selector)) )
				{
					if (mdit->internext)
					{
						for(;;)
						{
							mdit = mdit->internext;

							if (!mdit) break;
					
                			if (mdit->interrange[0] <= 0.f) // A new beginning?
							{
								*nextmodef = mdit;
								return interp;
							}
						}
					}
				}

				it = states + it->nextstate;
			}
		}

		*nextmodef = GetStateModel(it, mo->selector);
	}

	return interp;
}

//===========================================================================
//
//===========================================================================
int R_CheckModelSkins(modeldef_t *modef)
{
	model_t	*mdl;
	int		c, i, sub;
	
	// Precache this
	for (sub = 0, c = 0; sub < MAX_FRAME_MODELS; sub++)
	{
		if ( (mdl = modellist[modef->sub[sub].model]) )
		{
			// Load all skins
			for (i = 0; i < mdl->info.numSkins; i++)
			{
				if (!GL_PrepareSkin(mdl, i, NULL)) c++;
			}

			GL_PrepareShinySkin(modef, sub);
		}
	}
	return c;
}

//===========================================================================
// v220-5
//===========================================================================
int R_PrecacheSkinsForMoType(int type)
{
	mobjinfo_t	*mi;
	modeldef_t	*modef;
	model_t		*mdl;
	int			i, c, sub, k;

	if (type < 0 || type >= defs.count.mobjs.num) return 0;

	mi = &mobjinfo[type];

	// Check through all the model definitions.
	for (c = i = 0, modef = modelDefs; i < nummodelDefs; i++, modef++)
	{
		if (!modef->state) continue;

		if (stateowners[modef->state - states] != mi)
			continue;

		// Precache this.
		for (sub = 0; sub < MAX_FRAME_MODELS; sub++)
		{
			if (!modef->sub[sub].model) continue;

			mdl = modellist[modef->sub[sub].model];

			// Load all skins.
			for (k = 0; k < mdl->info.numSkins; k++)
			{
				if (GL_PrepareSkin(mdl, k, NULL)) c++;
			}

			GL_PrepareShinySkin(modef, sub);
		}
	}

	return c;
}

//===========================================================================
//
//===========================================================================
RDBOOL R_ModelHasSkins(modeldef_t *modef)
{
	model_t	*mdl;
	int		i, sub;
	
	for (sub = 0; sub < MAX_FRAME_MODELS; sub++)
	{
		if ( (mdl = modellist[modef->sub[sub].model]) )
		{
			for (i = 0; i < mdl->info.numSkins; i++)
			{
				if (chkSkinIndex(mdl, i))
				{
					return true;
				}
			}
		}
	}
	return false;
}

//===========================================================================
//	v220-9
//===========================================================================
static void R_Modelflags(int flag)
{
	modeldef_t	*modef;
	char		*nextstr = SR_NextSTR(true);
	int			i, k;

	// Check through all the model definitions.
	for (i = 0, modef = modelDefs; i < nummodelDefs; i++, modef++)
	{
		if (modef->state && !stricmp(modef->def->state, nextstr))
		{
			modef->flags |= flag;

			if (ge_inf.devmap)
				pr_OFinsert("MATCHED STATE %s mo->f %.8X\n", nextstr, modef->flags);

			for (k = 0; k < MAX_FRAME_MODELS; k++)
			{
				if (modef->sub[k].model)
					modef->sub[k].flags |= flag;
			}
		}
	}
}

//===========================================================================
//	v220-33
//===========================================================================
void R_PreCacheModelSkins() // $precache
{
	modeldef_t	*modef;
	model_t		*mdl;
	int			c, i, k, m;

	// Check through all the model definitions.
	for (c = i = 0, modef = modelDefs; i < nummodelDefs; i++, modef++)
	{
		for (k = 0; k < MAX_FRAME_MODELS; k++)
		{
			if ( (mdl = modellist[modef->sub[k].model]) )
			{
				for (m = 0; m < mdl->info.numSkins; m++)
				{
					if (chkSkinIndex(mdl, m))
					{
						if (GL_PrepareSkin(mdl, m, NULL))
						{
							Con_Progress(++c, PBARF_PERCENT);
						}
					}
				}
			}
		}
	}
}

extern char	new_mapid[8];

// v220-9
void checkSCP_ModifyFlags()
{
	if (ge_inf.hasscp_defn)
	{
		char		fname[512];
		char		*nextstr;

		sprintf(fname, "%sData\\SCP_modify\\%s.scp", ddRuntimeDir.path, analysename);

		if (!SR_LoadFile(fname, "SCP_MODIFY", NULL)) return;

		if (R3D_edit)
		{
			Con_Message("WARNING: .scp files are not loaded in edit mode\n");
			SR_EndScript();
			return;
		}

		while( (nextstr = SR_NextSTR(false)) )
		{
			if (!_stricmp(nextstr, "thingflag_noocclude"))
			{
				int	ednum, mapmatch;
				
				nextstr = SR_NextSTR(true);
				mapmatch = !stricmp(nextstr, new_mapid);
				ednum = SR_NextNum();
				
				if (mapmatch)
				{
					thinker_t	*thinker = thinkercap.next;
					mobj_t		*mo;

					while (thinker != &thinkercap)
					{
						if (   thinker->function == (void (__cdecl *)(void))P_MobjThinker
							&& (mo = (mobj_t *) thinker)->info->doomednum == ednum)
						{
							if (ge_inf.devmap) pr_OFinsert("MATCHED thingflag_noocclude: %i T%i\n", ednum, MAPNUM(mo));

							mo->flags2 |= MF2_NOOCCLUDE;
						}

						thinker = thinker->next;
					}
				}
			}
			else
			if (!_stricmp(nextstr, "setNearClip"))
			{
				thinker_t	*thinker = thinkercap.next;
				mobj_t		*mo;
				int			dmednum = SR_NextNum();
				
				while (thinker != &thinkercap)
				{
					if (   thinker->function == (void (__cdecl *)(void))P_MobjThinker
						&& (mo = (mobj_t *) thinker)->info->doomednum == dmednum)
					{
						P_UnlinkThing(mo);
						mo->flags |= MF_NOCLIP;
						mo->flags3 |= MF3_NEARCLIP;
						mo->radius = mo->info->radius = R_MaxRadius(mo, false, 3);
						P_LinkThing(mo, (mo->flags & MF_NOSECTOR   ? 0 : DDLINK_SECTOR) |
										(mo->flags & MF_NOBLOCKMAP ? 0 : DDLINK_BLOCKMAP));

						if (ge_inf.devmap) pr_OFinsert("MATCHED setNearClip: %i T%i rad %i\n", dmednum, MAPNUM(mo), mo->radius>>16);
					}

					thinker = thinker->next;
				}
			}
			else
			if (!_stricmp(nextstr, "modelflag_alphanormal"))
			{
				R_Modelflags(MFF_ALPHANORMAL);
			}
			else
			if (!_stricmp(nextstr, "modelflag_nomaxz"))
			{
				R_Modelflags(MFF_NO_DISTANCE_CHECK);
			}
			else
			if (!_stricmp(nextstr, "line3D_offset"))
			{
				int	lnum, offset, mapmatch;

				nextstr = SR_NextSTR(true);
				mapmatch = !stricmp(nextstr, new_mapid);
				
				lnum = SR_NextNum();
				offset = SR_NextNum();

				if (offset < 0 || offset > 99999)
					SR_SCPErr("offset %i not in range 0 to 99999", offset);

				if (mapmatch)
				{
					if (lnum < 0 || lnum >= numlines || lines[lnum].specType < R3D_40)
						SR_SCPErr("line %i is not a 3D line", lnum);
					else
						lines[lnum].d3d->offset3D = offset*FRACUNIT;
				}
			}
			else if (!_stricmp(nextstr, "actionSpot_makestatic"))
			{
				int	t78num, mapmatch, dmednum, height;
				
				nextstr = SR_NextSTR(true);
				mapmatch = !stricmp(nextstr, new_mapid);
				t78num = SR_NextNum();
				height = SR_NextNum() * FRACUNIT;
				dmednum = SR_NextNum();
				
				if (mapmatch)
				{
					thinker_t	*thinker = thinkercap.next;
					mobj_t		*mo;
					int			num = t78num - 7800;

					if (num < 0 || num > 99)
					{
						if (num < 0) num = t78num - 6800;

						if (num < 0 || num > 99)
							Con_Error("Invalid Tnum %i used with actionSpot_makestatic", t78num);

						num += 100;
					}

					if ( !(th7800[num].setflags & T_SET) )
						Con_Error("Tnum %i used with actionSpot_makestatic not found in map", t78num);

					if (ge_inf.devmap) pr_OFinsert("MATCHED actionSpot_makestatic: %i\n", t78num);

					th7800[num].options &= ~(MTF_EASY|MTF_NORMAL|MTF_HARD);
					th7800[num].options |= MTF_NORMAL;

					while (thinker != &thinkercap)
					{
						if (   thinker->function == (void (__cdecl *)(void))P_MobjThinker
							&& (mo = (mobj_t *) thinker)->info->doomednum == dmednum
							&& MAPNUM(mo) == t78num)
						{
							if (ge_inf.devmap) pr_OFinsert("MATCHED doomednum %i T%i\n", dmednum, t78num);

							mo->floorz = mo->dropoffz = mo->z = height;
							mo->flags3 |= (MF3_STATIC|MF3_NOROTATE);
						}

						thinker = thinker->next;
					}
				}
			}
			else if (!_stricmp(nextstr, "thing_Ignore"))
			{
				thinker_t	*thinker = thinkercap.next;
				mobj_t		*mo;
				int			dmednum = SR_NextNum();
				
				while (thinker != &thinkercap)
				{
					if (   thinker->function == (void (__cdecl *)(void))P_MobjThinker
						&& (mo = (mobj_t *) thinker)->info->doomednum == dmednum)
					{
						mo->ddflags |= DDF_IGNORE;
						if (ge_inf.devmap) pr_OFinsert("MATCHED thing_Ignore: %i type %i\n", dmednum, MAPNUM(mo));
					}

					thinker = thinker->next;
				}
			}
		}

		SR_EndScript();
	}
}
