//-----------------------------------------------------------------------------
//
// Copyright (C) 2018  G. Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: rend_slopes.cpp
//
// First written (v220-33) 08/Jul/2018 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// DESCRIPTION:
//	Slope functions either new or revised and moved here
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdlib.h>

#include "de_base.h"
#include "de_render.h"
#include "r3d_MAINspec.h"
#include "r3d_slopes.h"
#include "gl_tex.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

rendpolyq_t *Get_Quad(seg_t *sg);
void rend_UpdateSide(side_t *ls, short pos, int flatn);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern rendpolyq_t *quad;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

void rendSlopesInfill(seg_t *seg, side_t *sid, short midtex,
					  float xoff, float roff, float segoff)
{
	sector_t	*frontsec = seg->frontsector;
	sector_t	*backsec = seg->backsector;
	line_t		*ldef = seg->linedef;

	fixed_t		fhtf, fhtb, chtf, chtb, diff;
	RDBOOL		back = !midtex;	// side being viewed

	// swap the sides so that sd1 points to the side being viewed
	side_t	*sd1 = back ? &sides[ldef->sidenum[1]] : &sides[ldef->sidenum[0]];
	side_t	*sd2 = back ? &sides[ldef->sidenum[0]] : &sides[ldef->sidenum[1]];

	if (midtex || (midtex = sd2->midtexture) )	// should be
	{
		// check for infill
		fhtf = sd1->lsiflags & LS_LOSLOPEB ? SL_HighFloorHeight(frontsec) : frontsec->smoothfloor;
		fhtb = sd2->lsiflags & LS_LOSLOPEB ? SL_HighFloorHeight(backsec) : backsec->smoothfloor;
		chtf = sd1->lsiflags & LS_HISLOPEB ? SL_LowCeilHeight(frontsec) : frontsec->smoothceil;
		chtb = sd2->lsiflags & LS_HISLOPEB ? SL_LowCeilHeight(backsec) : backsec->smoothceil;

		// A midtex on side 0 either being faced or not suggests an infill loslope midtex may
		// be needed. The map must be constructed so that where an infill tex is needed that
		// this is always put on the side facing the viewer and that this is the front side.

		if (!back && fhtb > (diff = MIN_OF(fhtf, chtf)))
		{
			quad = Get_Quad(seg);
			quad->tex = GL_PrepareTexture(midtex);
			quad->texoffx = segoff + xoff*ctex->texscalex;
			quad->texoffy = roff*ctex->texscaley;
			if (ldef->ml_flags & (ML_DONTPEGBOTTOM|ML_DONTPEGTOP))
				quad->texoffy += (ctex->rendh - FIX2FLT(fhtb - diff));
			quad->texw = ctex->rendw;
			quad->texh = ctex->rendh;
			quad->detail = rendtexdetail;
			quad->bottom = FIX2FLT(diff);
			quad->top = FIX2FLT(fhtb);
			rend_UpdateSide(sid, (byte)(LS_LOWTEX), 0);
		}
		else
		if (back && fhtf > (diff = MIN_OF(fhtb, chtb)))
		{
			quad = Get_Quad(seg);
			quad->tex = GL_PrepareTexture(midtex);
			quad->texoffx = segoff + xoff*ctex->texscalex;
			quad->texoffy = roff*ctex->texscaley;
			if (ldef->ml_flags & (ML_DONTPEGBOTTOM|ML_DONTPEGTOP))
				quad->texoffy += (ctex->rendh - FIX2FLT(fhtf - diff));
			quad->texw = ctex->rendw;
			quad->texh = ctex->rendh;
			quad->detail = rendtexdetail;
			quad->bottom = FIX2FLT(diff);
			quad->top = FIX2FLT(MIN_OF(fhtf, chtb));
			rend_UpdateSide(sid, (byte)(LS_LOWTEX), 0);
		}

		// As above but for hislopes.
		if (!back && chtb < (diff = MAX_OF(fhtb, chtf)))
		{
			quad = Get_Quad(seg);
			quad->tex = GL_PrepareTexture(midtex);
			quad->texoffx = segoff + xoff*ctex->texscalex;
			quad->texoffy = roff*ctex->texscaley;
			if (ldef->ml_flags & (ML_DONTPEGBOTTOM|ML_DONTPEGTOP))
				quad->texoffy += (ctex->rendh - FIX2FLT(diff - chtb));
			quad->texw = ctex->rendw;
			quad->texh = ctex->rendh;
			quad->detail = rendtexdetail;
			quad->bottom = FIX2FLT(chtb);
			quad->top = FIX2FLT(diff);
			rend_UpdateSide(sid, (byte)(LS_HITEX), 0);
		}
		else
		if (back && chtf > (diff = MAX_OF(fhtf, chtb)))
		{
			quad = Get_Quad(seg);
			quad->tex = GL_PrepareTexture(midtex);
			quad->texoffx = segoff + xoff*ctex->texscalex;
			quad->texoffy = roff*ctex->texscaley;
			if (ldef->ml_flags & (ML_DONTPEGBOTTOM|ML_DONTPEGTOP))
				quad->texoffy += (ctex->rendh - FIX2FLT(chtf - diff));
			quad->texw = ctex->rendw;
			quad->texh = ctex->rendh;
			quad->detail = rendtexdetail;
			quad->bottom = FIX2FLT(diff);
			quad->top = FIX2FLT(chtf);
			rend_UpdateSide(sid, (byte)(LS_HITEX), 0);
		}
	}
}

void rendSlopesInfill2(seg_t *seg, side_t *sid, short midtex,
					  float xoff, float roff, float segoff)
{
	// check for infill
	sector_t	*fsec = seg->frontsector;
	sector_t	*bsec = seg->backsector;

	fixed_t fhtf = sid->lsiflags & LS_LOSLOPEB ? SL_HighFloorHeight(fsec) : fsec->smoothfloor;
	fixed_t fhtb = bsec->smoothfloor;
	fixed_t chtf = sid->lsiflags & LS_HISLOPEB ? SL_LowCeilHeight(fsec) : fsec->smoothceil;
	fixed_t chtb = bsec->smoothceil;

	if (sid->lsiflags & LS_LOSLOPEB && fhtb > fhtf)	// if not then, unless equal, there is a map error
	{
		fixed_t	slopeht = fhtb - fhtf;

		quad = Get_Quad(seg);
		quad->top = FIX2FLT(fhtb);
		quad->bottom = quad->top - FIX2FLT(slopeht);

		if (midtex)
		{
			quad->tex = GL_PrepareTexture(midtex);
		}
		else
		{
			if (fsec->floorpic == skyflatnum)
			{
				quad->flags |= RPF_SKY_MASK|RPF_QUADSET;
				goto test1;
			}
			else
			{
				quad->tex = GL_PrepareFlat(fsec->floorpic);
			}
		}

		quad->texw = ctex->rendw;
		quad->texh = ctex->rendh;
		quad->detail = rendtexdetail;
		quad->texoffx = segoff + xoff*ctex->texscalex;
		quad->texoffy = roff*ctex->texscaley;

		rend_UpdateSide(sid, LS_LOWTEX, midtex);
	}

test1:;
	if (!(sid->lsiflags & LS_HISLOPEB) && chtf < chtb)
	{
		quad = Get_Quad(seg);
		quad->top = FIX2FLT(chtb);
		quad->bottom = FIX2FLT(chtf);

		if (sid->toptexture)
		{
			quad->tex = GL_PrepareTexture(sid->toptexture);
		}
		else
		{
			if (fsec->ceilingpic == skyflatnum)
			{
				quad->flags |= RPF_SKY_MASK|RPF_QUADSET;
				goto test2;
			}
			else
			{
				quad->tex = GL_PrepareFlat(fsec->ceilingpic);
			}
		}

		quad->texw = ctex->rendw;
		quad->texh = ctex->rendh;
		quad->detail = rendtexdetail;
		quad->texoffx = segoff + xoff*ctex->texscalex;
		quad->texoffy = roff*ctex->texscaley;

		rend_UpdateSide(sid, LS_HITEX, sid->toptexture);
	}
	else
	if (sid->lsiflags & LS_HISLOPEB && chtf < chtb)
	{
		fixed_t	slopeht = chtf - chtb;

		if (slopeht < chtf - chtb)
		{
			quad = Get_Quad(seg);

			if (midtex)
			{
				quad->tex = GL_PrepareTexture(midtex);
			}
			else
			{
				if (fsec->ceilingpic == skyflatnum)
				{
					quad->flags |= RPF_SKY_MASK|RPF_QUADSET;
					quad->top = FIX2FLT(chtf);
					goto test2;
				}
				else
				{
					quad->tex = GL_PrepareFlat(fsec->ceilingpic);
				}
			}

			quad->texw = ctex->rendw;
			quad->texh = ctex->rendh;
			quad->detail = rendtexdetail;
			quad->texoffx = segoff + xoff*ctex->texscalex;
			quad->texoffy = roff*ctex->texscaley;
			quad->top = FIX2FLT(fsec->ceilingheight);
			quad->bottom = quad->top + FIX2FLT(slopeht);

			rend_UpdateSide(sid, LS_HITEX, midtex);
		}
	}
	else
	if (!(sid->lsiflags & LS_HISLOPEB) && chtf > chtb)
	{
		quad = Get_Quad(seg);
		quad->top = FIX2FLT(chtf);
		quad->bottom = FIX2FLT(chtb);

		if (sid->toptexture)
		{
			quad->tex = GL_PrepareTexture(sid->toptexture);
		}
		else
		{
			if (fsec->ceilingpic == skyflatnum)
			{
				quad->flags |= RPF_SKY_MASK|RPF_QUADSET;
				return;
			}
			else
			{
				quad->tex = GL_PrepareFlat(fsec->ceilingpic);
			}
		}
		quad->texw = ctex->rendw;
		quad->texh = ctex->rendh;
		quad->detail = rendtexdetail;
		quad->texoffx = segoff + xoff*ctex->texscalex;
		quad->texoffy = roff*ctex->texscaley;
		if (!(seg->linedef->ml_flags & ML_DONTPEGTOP))
			quad->texoffy += (ctex->rendh - FIX2FLT(chtf - chtb));	// Normal alignment to bottom

		rend_UpdateSide(sid, LS_HITEX, sid->toptexture);
	}
	else	// special condition for backward compatibility
	if (sid->toptexture && sid->lsiflags & LS_HISLOPEB && chtf > fhtf && chtb == fhtb)
	{
		quad = Get_Quad(seg);
		quad->tex = GL_PrepareTexture(sid->toptexture);
		quad->texw = ctex->rendw;
		quad->texh = ctex->rendh;
		quad->detail = rendtexdetail;
		quad->texoffx = segoff + xoff*ctex->texscalex;
		quad->texoffy = roff*ctex->texscaley;
		quad->top = FIX2FLT(chtf);
		quad->bottom = FIX2FLT(fhtf);

		rend_UpdateSide(sid, LS_HITEX, sid->toptexture);
	}

test2:;

	if (!(sid->lsiflags & LS_LOSLOPEB) && fhtf < fhtb)
	{
		quad = Get_Quad(seg);
		quad->top = FIX2FLT(bsec->floorheight);
		quad->bottom = FIX2FLT(fsec->floorheight);

		if (sid->bottomtexture)
		{
			quad->tex = GL_PrepareTexture(sid->bottomtexture);
		}
		else
		{
			if (fsec->floorpic == skyflatnum)
			{
				quad->flags |= RPF_SKY_MASK|RPF_QUADSET;
				return;
			}
			else
			{
				quad->tex = GL_PrepareFlat(fsec->floorpic);
			}
		}

		quad->texw = ctex->rendw;
		quad->texh = ctex->rendh;
		quad->detail = rendtexdetail;
		quad->texoffx = segoff + xoff*ctex->texscalex;
		quad->texoffy = roff*ctex->texscaley;
		if (seg->linedef->ml_flags & ML_DONTPEGBOTTOM)
			quad->texoffy += quad->texh - FIX2FLT(chtf - fhtb);

		rend_UpdateSide(sid, LS_LOWTEX, sid->bottomtexture);	// v220-35 was LS_HITEX
	}
}
