//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: s_environ.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified for R3D
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//
// Based on: Doomsday v1.7.8
//
// DESCRIPTION:  Calculation of the aural properties of sectors
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <ctype.h>
#include <string.h>
#include <eax.h>

#include "sys_timer.h"

#include "de_base.h"
#include "de_play.h"
#include "de_refresh.h"
#include "de_audio.h"

#include "con_console.h"
#include "con_bar.h"
#include "m_args.h"
#include "m_misc.h"

#include "m_winmem.h"
#include "r3d_compat.h"
#include "r3d_utils.h"

// MACROS ------------------------------------------------------------------

#define DIV255 0.003921569f

enum // Texture types
{
	TEXTYPE_UNKNOWN,
	TEXTYPE_METAL,
	TEXTYPE_ROCK,
	TEXTYPE_WOOD,
	TEXTYPE_CLOTH
};

// TYPES -------------------------------------------------------------------

typedef struct
{
	uint	space;
	int		volume;
	int		decay;
	int		damping;
	float	x;
	float	y;

} subreverb_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static byte *revtype = NULL;
static int  rev_count;

// CODE --------------------------------------------------------------------

static byte getType(char *name)
{
	if(!_stricmp(name, "Metal")) return TEXTYPE_METAL;
	if(!_stricmp(name, "Rock" )) return TEXTYPE_ROCK;
	if(!_stricmp(name, "Cloth")) return TEXTYPE_CLOTH;
	if(!_stricmp(name, "Wood" )) return TEXTYPE_WOOD;
	return TEXTYPE_UNKNOWN;
}

static int S_GetTypeFromTexnum(int txn)
{
	if (!revtype)
	{
		ded_tenviron_t *env;
		int		i, k;

		revtype = (byte*)W_Calloc(numstandardtextures, STATICID, 0);

		for (i = 0, env = defs.tenviron; i < defs.count.tenviron.num; i++, env++)
		{
			byte type = getType(env->id);

			for (k = 0; k < env->count.num; k++)
			{
				int num = R_CheckTextureNumForName(env->textures[k].str);

				if (num >= 0 && num < numstandardtextures)
					revtype[num] = type;
			}
		}
	}
		
	return (int)revtype[txn];
}

//===========================================================================
// S_CalcSectorReverbs
//	Calculate the reverb settings for each sector.
//  b220-0 reworked
//  Note: must be run even if sound is disabled as it could be stored to
//  an analyser file which could then later be loaded with sound enabled
//===========================================================================
void S_CalcSectorReverbs()
{
	subreverb_t *sub_reverb, *rev;
	subsector_t	*sub;
	sector_t	*sec;
	seg_t		*seg;

	float	bbox[4];
	float	total;
	float	metal;
	float	rock;
	float	wood;
	float	cloth;
	float	x, y;
	float	*box;

	RDBOOL	conbar_active = false;

	int		conbarnum = numlines / 100;
	int		i, c, ht, type;
	int		*listsec;
	int		ibox[4];
	int     sn;

	byte	*lut = ArgExists("-lut") ? (byte*)M_Calloc(numstandardtextures, 0) : NULL;
	byte	*marksec = (byte*)M_Calloc(numsectors, 0);

	//-------------------------------------------------

	timecache[ITC_SECREVERB] = Sys_GetSeconds();

	if (ma_inf.reverbthings && ArgExists("-debug_scp"))
	{
		RDBOOL	thingset = false;

		for (c = 0; c < MAX_REVERBTHING; c++)
		{
			if (reverbThing[c].seccount)
			{
				thingset = true;


				pr_OFinsert("\n======= REVERB SEC with Thing type %i thing count %i sec count %i\n",
					reverbThing[c].num, reverbThing[c].thcount, reverbThing[c].seccount);

				for (i = 0; i < reverbThing[c].seccount; i++)
				{
					if (!(i % 16)) pr_OFdirect("\n");
					pr_OFinsert(" %i",reverbThing[c].secnum[i]);
				}

				pr_OFdirect ("\n");
			}
		}
		if (thingset) pr_OFdirect ("\nEnd Reverb Things =========\n\n");
	}

	// Allocate memory for the temporary subsector array
	sub_reverb = (subreverb_t*)M_Calloc(sizeof(subreverb_t) * numsubsectors, 0);
	rev_count = 0;

	// Draw conbar (R3D used revised functions)
	if (conbarnum > 100) conbarnum = 100;
	if (conbarnum > 30)
	{
		Con_InitProgress("CALCULATING: sector reverb values...", numsectors);
		Con_TimingStart();
		conbar_active = true;
	}

	// Set up 3D lift sectors
	for (c = 0, sec = sectors; c < numsectors; c++, sec++)
	{
		if (marksec[c]) continue;

		if ( (listsec = sec->listsec) )
		{
			M_ClearBox(ibox);

			while ((sn = *(int*)listsec++) >= 0)
			{
				sector_t *s = &sectors[sn];

				marksec[sn] = 1;

				for (i = 0; i < s->linecount; i++)
				{
					line_t *ld = s->lines[i];

					if (ld->specType >= R3D_40)
					{
						M_AddLineToBox(ibox, ld);
					}
				}
			}

			listsec = sec->listsec;

			while ((sn = *(int*)listsec++) >= 0)
			{
				sector_t *s = &sectors[sn];
				
				box = s->sibounds;

				for (i = 0; i < 4; i++) box[i] = FIX2FLT(ibox[i]);

				// Correct sound origin
				s->soundorg.x = meanV(ibox[BOXRIGHT], ibox[BOXLEFT]);
				s->soundorg.y = meanV(ibox[BOXTOP  ], ibox[BOXBASE]);
			}
		}
	}

	// R3D - Check for invalid sectors
	for (c = 0, sec = sectors; c < numsectors; c++, sec++)
	{
		if (marksec[c]) continue;

		box = sec->sibounds;

		if (   !sec->linecount
			|| (!SecCtrl(c)
			&& (box[BOXLEFT] >= box[BOXRIGHT] || box[BOXBASE] >= box[BOXTOP])) )
		{
			sec->specFlags |= SF_IGNORE;	// Mark as unusable
			sec->specFlags2 |= SF2_NOMOVSOUND;
			sec->reverb[SRD_PRESET ] = EAX_ENVIRONMENT_GENERIC; 
			sec->reverb[SRD_VOLUME ] = 0; 
			sec->reverb[SRD_DECAY  ] = 0;
			sec->reverb[SRD_DAMPING] = 0;
			memset(box, 0, sizeof(box));
		}
	}

	M_Free(marksec);	

	// First determine each subsector's individual characteristics.
	for (c = 0, sub = subsectors; c < numsubsectors; c++, sub++)
	{
		if (!sub->sector->linecount || sub->sector->specFlags & SF_IGNORE)
		{
			continue;
		}

		total = metal = rock = cloth = wood = 0;

		rev = &sub_reverb[rev_count++];

		rev->x = sub->midpoint.x;
		rev->y = sub->midpoint.y;

		x  = FIX2FLT(sub->ssbox[BOXRIGHT] - sub->ssbox[BOXLEFT]);
		y  = FIX2FLT(sub->ssbox[BOXTOP  ] - sub->ssbox[BOXBASE]);
		ht = (sub->sector->ceilingheight - sub->sector->floorheight) >> FRACBITS;

		// Space is approx = the area of the subsector's bounding box and sector height.
		rev->space = (uint)(ht * x * y);

		for(i = 0, seg = &segs[sub->firstline]; i < sub->linecount; i++, seg++)
		{
			short texn;

			if (seg->sidedef && (texn = seg->sidedef->midtexture) > 0)
			{
				RDBOOL valid = texn < numstandardtextures;
				type = valid ? S_GetTypeFromTexnum(texn) : TEXTYPE_UNKNOWN;

				total += seg->length;

				// The texture of the seg determines its type.
				// If unknown [0] then treated as wood.
				if (type == TEXTYPE_UNKNOWN)
				{
					wood += seg->length;
					if (lut && valid) lut[texn] = true;
				}
				else if (type == TEXTYPE_WOOD)
					wood += seg->length;
				else if (type == TEXTYPE_METAL)
					metal += seg->length;
				else if (type == TEXTYPE_ROCK)
					rock += seg->length;
				else if (type == TEXTYPE_CLOTH)
					cloth += seg->length;
			}
		}

		if(!total) continue; // No lines with a mid tex found

		metal /= total;
		rock  /= total;
		wood  /= total;
		cloth /= total;

		// Volume.
		i = (int) ( metal*255 + rock*200 + wood*80 + cloth*5 );
		if(i < 0) i = 0;
		if(i > 255) i = 255;
		rev->volume = i;

		// Decay time.
		i = (int) ( metal*255 + rock*160 + wood*50 + cloth*5 );
		if(i < 0) i = 0;
		if(i > 255) i = 255;
		rev->decay = i;

		// High frequency damping.
		i = (int) ( metal*25 + rock*100 + wood*200 + cloth*255 );
		if(i < 0) i = 0;
		if(i > 255) i = 255;
		rev->damping = i;
	}
	
	for (c = 0, sec = sectors; c < numsectors; c++, sec++)
	{
		float	spaceScatter;
		float	space = 0;
		uint	sectorSpace;

		if (sec->specFlags & SF_IGNORE) continue;

		memcpy(bbox, sec->sibounds, sizeof(bbox));

		ht = (abs(sec->ceilingheight - sec->floorheight)) >> FRACBITS;

		sectorSpace = (uint)(ht * (bbox[BOXRIGHT] - bbox[BOXLEFT]) * (bbox[BOXTOP] - bbox[BOXBASE]));

		bbox[BOXLEFT ] -= 128;
		bbox[BOXRIGHT] += 128;
		bbox[BOXBASE ] -= 128;
		bbox[BOXTOP  ] += 128;

		for(i = 0, rev = sub_reverb; i < rev_count; i++, rev++)
		{
			// Is this subsector close enough?
			if (rev->x > bbox[BOXLEFT] && rev->x < bbox[BOXRIGHT] &&
				rev->y < bbox[BOXTOP ] && rev->y > bbox[BOXBASE ])
			{
				rev = &sub_reverb[i];

				space += rev->space;
				sec->reverb[SRD_VOLUME ] += rev->volume  * DIV255 * rev->space;
				sec->reverb[SRD_DECAY  ] += rev->decay   * DIV255 * rev->space;
				sec->reverb[SRD_DAMPING] += rev->damping * DIV255 * rev->space;
			}
		}

		if (space)
		{
			spaceScatter = sectorSpace / space;
			// These three are weighted by the space.
			sec->reverb[SRD_VOLUME ] /= space;
			sec->reverb[SRD_DECAY  ] /= space;
			sec->reverb[SRD_DAMPING] /= space;
		}
		else // Use standard defaults
		{
			sec->reverb[SRD_PRESET ] = EAX_ENVIRONMENT_GENERIC; 
			sec->reverb[SRD_VOLUME ] = .2f;
			sec->reverb[SRD_DECAY  ] = .4f;
			sec->reverb[SRD_DAMPING] =  1;
			continue;
		}

		// If the space is scattered, the reverb effect lessens.
		space /= (spaceScatter > .8 ? 10 : spaceScatter > .6 ? 4 : 1);

		// Scale the reverb space to a reasonable range, so that very small
		// tends to 0 and .99 is very large. 1.0 is only for open areas.
		if ( (space /= 120e6) > .99f) space = .99f;
		
		if (   sec->ceilingpic == skyflatnum
			|| sec->floorpic   == skyflatnum)
		{
			if (space > .5) 
				sec->reverb[SRD_VOLUME] = 1.0f; // Full volume.
			else 
				// For small open sectors the reverb is fixed.
				sec->reverb[SRD_VOLUME] = .5f;

			space = 1.0f;
		}
		else
		{
			// If no sky flats then add 25% more audible reverb.
			sec->reverb[SRD_VOLUME] += space / 4;
		}

		if (sec->reverb[SRD_DECAY] > .5 && space < .2)
			// This much decay needs at least the Generic environment.
			space = .2f;

		sec->reverb[SRD_PRESET] = (float)(
			  space >=  1? EAX_ENVIRONMENT_PLAIN
			: space >= .8? EAX_ENVIRONMENT_CONCERTHALL
			: space >= .6? EAX_ENVIRONMENT_AUDITORIUM
			: space >= .4? EAX_ENVIRONMENT_CAVE
			: space >= .2? EAX_ENVIRONMENT_GENERIC
			: EAX_ENVIRONMENT_ROOM);

		if (sec->reverb[SRD_VOLUME] < 0.0f) sec->reverb[SRD_VOLUME] = 0.0f;
		if (sec->reverb[SRD_VOLUME] > 1.0f) sec->reverb[SRD_VOLUME] = 1.0f;

		// Transform the values
		// Reverb decay.
		sec->reverb[SRD_DECAY] = (sec->reverb[SRD_DECAY] - .5f)*1.5f + 1.0f;
		if (sec->reverb[SRD_DECAY] < .1f) sec->reverb[SRD_DECAY] = .1f;

		// Damping.
		sec->reverb[SRD_DAMPING] = 1.1f * (1.2f - sec->reverb[SRD_DAMPING]);
		if (sec->reverb[SRD_DAMPING] < .1f) sec->reverb[SRD_DAMPING] = .1f;

		// Update conbar
		if (conbar_active) Con_Progress(c, PBARF_PERCENT);
	}

	// Design aid: Print out any midtextures undefined in defs Texture Environment
	if (lut)
	{
		Con_Message("======= LUT\n");
		for (i = 0; i < numstandardtextures; i++)
		{
			if (lut[i])
				Con_Message("%s\n", R_TextureNameForNum(i));
		}
		Con_Message("END LUT =======\n");
		M_Free(lut);
	}

	// Deal with closed door sectors
	for (c = 0, sec = sectors; c < numsectors; c++, sec++)
	{
		if (sec->ceilingheight <= sec->floorheight && sec->specFlags & SF_MOVSECTOR)
		{
			for (i = 0; i < sec->linecount; i++)
			{
				line_t *ld = sec->lines[i];
				sector_t *s = ld->frontsector == sec ? ld->backsector : ld->frontsector;

				if (s && !(s->specFlags & SF_MOVSECTOR))
				{
					memcpy(sec->reverb, s->reverb, sizeof(sec->reverb));
					break;
				}
			}
		}
	}

	M_Free(sub_reverb);

	// R3D - Return time to complete
	if (conbar_active) Con_HideProgress();

	timecache[ITC_SECREVERB] = Sys_GetSeconds() - timecache[ITC_SECREVERB];
}

