//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: s_logic.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified for R3D
//  NOTE: Uses W_Calloc[LEVELID], so this has to be inited for every map.
//  (Done via S_LevelChange()).
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//
// Based on: Doomsday v1.7.8
//
// DESCRIPTION:  The Logical Sound Manager. Tracks all currently playing
//               world sounds, regardless of whether Sfx is available
//               or if the sounds are actually audible to anyone.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <string.h>

#include "de_base.h"
#include "de_console.h"
#include "de_audio.h"

#include "m_winmem.h"
#include "sys_timer.h"

// MACROS ------------------------------------------------------------------

// The logical sounds hash table uses sound IDs as keys.
#define LOGIC_HASH_SIZE		64

#define PURGE_INTERVAL		2000	// 2 seconds

// TYPES -------------------------------------------------------------------

typedef struct logicsound_s {
	struct logicsound_s *next, *prev;
	int		id;
	int		HSFflags;
	mobj_t	*origin;
	uint	length;
	uint	endTime;	
} logicsound_t;

typedef struct logichash_s {
	logicsound_t *first, *last;
} logichash_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static logichash_t logicHash[LOGIC_HASH_SIZE];

// CODE --------------------------------------------------------------------

// Initialize the Logical Sound Manager for a new map.
void Sfx_InitLogical(void)
{
	// Memory in the hash table is LEVELID, so this is acceptable.
	memset(logicHash, 0, sizeof(logicHash));
}

// Return a pointer the logical sounds hash.
static logichash_t *Sfx_LogicHash(int soundid)
{
	return &logicHash[ (uint) soundid % LOGIC_HASH_SIZE ];
}

// Create and link a new logical sound hash table node.
static logicsound_t *Sfx_CreateLogical(int soundid)
{
	logichash_t  *hash = Sfx_LogicHash(soundid);
	logicsound_t *node = (logicsound_t*)W_Calloc(sizeof(logicsound_t), LEVELID, 46);

	if(hash->last)
	{
		hash->last->next = node;
		node->prev = hash->last;
	}

	hash->last = node;
	if(!hash->first) hash->first = node;

	return node;
}

// Unlink and destroy a logical sound hash table node.
static void Sfx_DestroyLogical(logicsound_t *node)
{
	logichash_t *hash = Sfx_LogicHash(node->id);
	
	if(hash->first == node) hash->first = node->next;
	if(hash->last  == node) hash->last  = node->prev;
	if(node->next) node->next->prev = node->prev;
	if(node->prev) node->prev->next = node->next;

	W_Free(node, LEVELID, 47);
}

// The sound is entered into the list of playing sounds. Called when a 
// 'world class' sound is started, regardless of whether it's actually
// started on the local system.
void Sfx_StartLogical(int id, mobj_t *origin)
{
	if (!sfx_avail) return;

	logicsound_t	*node, *next, *newnode = NULL;
	uint			length;
	int				soundid = id & ~HSF_FLAG_MASK;
	int				hsf_flags = id & HSF_FLAG_MASK;

	BEGIN_COP;

	for (node = Sfx_LogicHash(soundid)->first; node; node = next)
	{
		next = node->next;

		if (   node->origin == origin
			&& node->id == soundid
			&& node->HSFflags == hsf_flags)
		{
			newnode = node;	// already exists
			break;
		}
	}

	if (!newnode)
	{
		length = Sfx_GetSoundLength(soundid);

		if (!length)
		{	// This is not a valid sound.
			END_COP;
			sounds[soundid].flags |= SFX_BADSOUND;
			return;
		}
		
		newnode = Sfx_CreateLogical(soundid);
		newnode->id       = soundid;
		newnode->HSFflags = hsf_flags;
		newnode->origin   = origin;
		newnode->length   = length;
	}

	newnode->endTime = Sys_GetRealTime() + newnode->length;

	END_COP;
}

// The sound is removed from the list of playing sounds. Called whenever
// a sound is stopped, regardless of whether it was actually playing on 
// the local system. Returns the number of sounds stopped.
// Sending id == 0 && origin == NULL will stop everything
int Sfx_StopLogical(int id, mobj_t *origin)
{
	logicsound_t *node, *next;
	int stopCount = 0;
	int i, soundid = id & ~HSF_FLAG_MASK;

	BEGIN_COP;

	if (soundid)
	{
		for (node = Sfx_LogicHash(soundid)->first; node; node = next)
		{
			next = node->next;

			if (node->id == soundid && node->origin == origin)
			{
				Sfx_DestroyLogical(node);
				stopCount++;
			}
		}
	}
	else
	{
		// Browse through the entire hash.
		for(i = 0; i < LOGIC_HASH_SIZE; i++)
		{
			for(node = logicHash[i].first; node; node = next)
			{
				next = node->next;

				if (!origin || node->origin == origin)
				{
					Sfx_DestroyLogical(node);
					stopCount++;
				}
			}
		}
	}

	END_COP;

	return stopCount;
}

// Remove stopped logical sounds from the hash.
void Sfx_PurgeLogical(void)
{
	static uint		lastTime = 0;
	uint			i, nowTime = Sys_GetRealTime();
	logicsound_t	*node, *next;

	// Check that it's time to purge
	if (nowTime - lastTime < PURGE_INTERVAL) return;	// No

	BEGIN_COP;

	lastTime = nowTime;

	// Check all sounds in the hash.
	for(i = 0; i < LOGIC_HASH_SIZE; i++)
	{
		for (node = logicHash[i].first; node; node = next)
		{
			next = node->next;

			if ( !(node->HSFflags & (HSF_REPEAT|HSF_SUSPEND))
				&& node->endTime < nowTime)
			{
				Sfx_DestroyLogical(node);	// This has stopped.
			}
		}
	}

	END_COP;
}

// b220-0
RDBOOL Sfx_SetNodeSuspend(int id, mobj_t *origin)
{
	logicsound_t *node = (Sfx_LogicHash(id))->first;

	for (; node; node = node->next)
	{
		if (node->id == id && node->origin == origin)
		{
			if (Sfx_CacheHit(id))
			{
				node->HSFflags |= HSF_SUSPEND;
				node->endTime = Sys_GetRealTime() + node->length;
				return true;
			}

			node->HSFflags &= ~HSF_SUSPEND;
			break;
		}
	}
	return false;
}

// b220-0
RDBOOL Sfx_DelNodeSuspend(int id, mobj_t *origin)
{
	logicsound_t *node = (Sfx_LogicHash(id))->first;

	for (; node; node = node->next)
	{
		if (   node->id == id
			&& node->origin == origin
			&& node->HSFflags & (HSF_SUSPEND|HSF_REPEAT))
		{
			node->HSFflags &= ~(HSF_SUSPEND|HSF_REPEAT);
			return true;
		}
	}
	return false;
}

// origin is expected to be set
RDBOOL Sfx_IsPlaying(int id, mobj_t *origin)
{
	logicsound_t	*node;
	uint			nowTime = Sys_GetRealTime();
	int				i, soundid = id & ~HSF_FLAG_MASK;

	if (!origin) return false;

	if (id)
	{
		node = (Sfx_LogicHash(soundid))->first;

		for (; node; node = node->next)
		{
			if (node->id == soundid && node->origin == origin)
			{
				if (   node->endTime > nowTime
					|| node->HSFflags & (HSF_REPEAT|HSF_SUSPEND))	// b220-0
				{
					return true;	// In use
				}
			}
		}
	}
	else
	{
		// Check if the origin is playing any sound.
		for (i = 0; i < LOGIC_HASH_SIZE; i++)
		{
			for (node = logicHash[i].first; node; node = node->next)
			{
				if (   node->origin == origin
					&& (node->endTime > nowTime
					|| node->HSFflags & HSF_REPEAT))
				{
					return true;	// In use
				}
			}
		}
	}

	// The sound was not found.
	return false;
}
