//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 Jaakko Kernen
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: s_main.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified for R3D
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
//  wav_volume removed
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//
// Based on: Doomsday v1.7.8
//
// DESCRIPTION:  Interface to Sfx modules,
//               High-level (and exported) sound control.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <math.h>

#include "de_base.h"
#include "de_console.h"
#include "de_network.h"
#include "de_system.h"
#include "de_graphics.h"
#include "de_audio.h"
#include "de_misc.h"
#include "r_main.h"
#include "sound_var.h"
#include "common_vars.h"

#include "doom/info.h"
#include "doom/sndidx.h"
#include "doom/p_local.h"
#include "doom/mbf_player.h"

#include "r3d_openGL.h"
#include "r3d_compat.h"
#include "r3d_sfx.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void	Music_VolumeSet(void);
void	S_InitSNDScript(void);
void	S_InitMUSLIST(char *path);
void	p2swd(char *msg, ...);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

byte	sound_info = false;
int		soundstarts[MAXLINESOUNDS];	// unattenuated line triggered sounds

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

//===========================================================================
// S_Init
//	Main sound system initialization. Inits both the Sfx and Mus modules.
//	Returns true if there were no errors.
//===========================================================================
RDBOOL S_Init(void)
{
	ge_inf.rolandResetLmp = -1;

	pr_OFdirect("\n===========INITIALISING SOUND:\n");

	// There is no sound or music if set
	if (compat[c_nosound])
	{
		pr_OFdirect("\nAll sound has been disabled by the user.\n");
		pr_OFdirect("\n==============================\n");
		return false;
	}

	if (compat[c_nomusic])
	{
		pr_OFdirect("Music is disabled.\n");
	}
	else
	{
		FModInit();

		if (MC_MidiInit())
		{
			ge_inf.rolandResetLmp = W_CheckRisen3DWADNumForName("GSRESET", true);

			if (DoomLumpIdentify(ge_inf.rolandResetLmp, NULL, NULL, DLI_SOUND) != LMP_MIDI)
			{
				ge_inf.rolandResetLmp = -1;
			}
		}
		else if (!ge_inf.usefmodmidi)
			pr_OFdirect("Error        : failed to initialise WinMM midi player\n");

		pr_OFdirect("Checking music list.\n");
		S_InitMUSLIST(ddRuntimeDir.path);
	}

	Sfx_Init();

	pr_OFinsert("Sound Effects: Init %s.\n", sfx_avail ? "succeeded" : "failed");

	if (sfx_avail) S_InitSNDScript(); else compat[c_nosound] = true;

	pr_OFrepeat(NULL, "=", 29, 1);

	return sfx_avail;
}

//===========================================================================
// S_LevelChange
//	Called before loading a map
//===========================================================================
void S_LevelChange(void)
{
	Sfx_InitLogical();	// Stop everything in the LSM.
	Sfx_LevelChange();	// Clear unwanted sound channels
	memset(soundstarts, 0, MAXLINESOUNDS*sizeof(int));
}

//===========================================================================
// S_Shutdown
//	Shutdown the whole sound system (Sfx + Mus).
//===========================================================================
void S_Shutdown(void)
{
	Sfx_Shutdown();
	FModShutdown();
	MC_MidiShutdown();		// v220-13 $midi
}

//===========================================================================
// S_Reset
//	Stop all channels and music, delete the entire sample cache.
//===========================================================================
void S_Reset(void)
{
	Sfx_Reset();
	S_StopMusic();
}

//===========================================================================
// S_StartFrame
//===========================================================================
void S_StartFrame(void)
{
	Music_VolumeSet();	// b220-1

	// Update all channels (freq, 2D:pan,volume, 3D:position,velocity).
	Sfx_StartFrame();

	// Remove stopped sounds from the LSM.
	Sfx_PurgeLogical();
}

extern RDBOOL	allowrefresh;

static RDBOOL soundValid(int sound_id)
{
	int sound;

	return sfx_avail
		&& allowrefresh
		&& !(sound_id & HSF_SNDTHING)
		&& (sound = sound_id & ~HSF_FLAG_MASK) > 0
		&&  sound < count_sounds.num
		&& !(sounds[sound].flags & SFX_BADSOUND);
}

// returns false if sound cannot be heard
RDBOOL S_CheckSoundDistance(mobj_t *origin, float *fixedPos, float attenuation)
{
	ma_inf.sound_distance =
		P_MobjSoundDistancef(origin, fixedPos, attenuation);

	// If origin is too far then it's pointless playing the sound.
	return (ma_inf.sound_distance < SOUND_MAX_DISTANCE);
}

//===========================================================================
// S_LocalSoundAtVolumeFrom
//	Plays a sound on the local system. A public interface.
//	Origin and fixedpos can be both NULL, in which case the sound is
//	played in 2D and centered.
//	Flags can be included in the sound ID number (DDSF_*).
//	Returns nonzero if a sound was started.
//===========================================================================
int S_LocalSoundAtVolumeFrom(
				int soundIdAndFlags,
				mobj_t *origin,
				float *fixedPos,
				float volume,
				float attenuation)
{
	sfxsample_t	*sample;
	float		freq = 1.0f;
	int			soundId, sflags, sfx_flags = 0;

	if (  !soundValid(soundIdAndFlags)
		|| volume <= 0) // Pointless playing it
	{
		return false;
	}

	if (volume > 1.0f) volume = 1.0f;	// Shouldn't be - insurance only

	soundId = soundIdAndFlags & ~HSF_FLAG_MASK;
	sflags  = sounds[soundId].flags;

	// R3D
	if (   sflags & SFX_BADSOUND
		|| (sflags & SFX_MENUONLY && !(soundIdAndFlags & HSF_MENUSOUND)))
	{
		return false;
	}
	
	if (soundIdAndFlags & HSF_NO_ATTENUATION) attenuation = 0;
	if (soundIdAndFlags & HSF_AMBIENT) sfx_flags |= SFX_AMBIENT;

	// R3D A repeating buffer must not have the SFX_DONT_STOP flag set
	if (soundIdAndFlags & HSF_REPEAT)
		sfx_flags |= SFX_REPEAT;
	else
	{
		if (soundIdAndFlags & SFX_DONT_STOP) sfx_flags |= SFX_DONT_STOP;
		else	// v220-33 added
		if (soundId == sfx_barexp) sfx_flags |= SFX_DONT_STOP;
	}

	if (soundIdAndFlags & HSF_AMBIENT)
		sfx_flags |= SFX_AMBIENT;

	// R3D
	if (soundIdAndFlags & HSF_MENUSOUND)
	{
		sfx_flags |= SFX_MENUONLY;
		ma_inf.sound_distance = 0;
	}
	else
	{
		if (!ppt.viewplrmo) return false;
		
		if (soundIdAndFlags & HSF_HALFATTN) attenuation = 0.5f;

		if ((origin && origin != ppt.viewplrmo) || fixedPos)
		{
			// distances with repeating sounds have already been checked
			if (   !(sfx_flags & SFX_REPEAT)
				&& !S_CheckSoundDistance(origin, fixedPos, attenuation))
			{
				return false;
			}
		}
		else
		{
			ma_inf.sound_distance = SOUND_MAX_DISTANCE;
		}
	}

	// Load the sample.
	if ( !(sample = Sfx_Cache(soundId)) ) return false;

	// Random frequency alteration
	if (sflags & (SFX_RANDOM_SHIFT|SFX_RANDOM_SHIFT2) && cfg.pitch)
	{
		// don't use with non-attenuated or repeating sounds
		if (attenuation && !(soundIdAndFlags & HSF_REPEAT))
		{
			// Multipliers chosen to match original sound code
			if (sflags & SFX_RANDOM_SHIFT)
				freq += (M_FRandom() - M_FRandom()) * ( 7.0f / 255.0f);
			else	// SFX_RANDOM_SHIFT2
				freq += (M_FRandom() - M_FRandom()) * (15.0f / 255.0f);
		}
	}

	// If the sound has an exclusion group all of the same
	// emitter's iterations of this sound will stop.
	// E.g. chainsaw
	// NOTE: global exclude removed since if a global sound is being
	//       played then only one of any one type is allowed - R3D
	if (sounds[soundId].group)
	{
		Sfx_StopSoundGroup(sounds[soundId].group, origin);
	}

	// Let's play it.
	return Sfx_StartSound(
		sample,
		volume,
		freq,
		origin,
		fixedPos,
		attenuation,
		ma_inf.sound_distance,
		sfx_flags);
}

//===========================================================================
// S_FootstepSound	v220-20
//===========================================================================
int S_FootstepSound(int sndid, float vol, int shiftup)
{
	if (!soundValid(sndid)) return false;

	sfxsample_t	*sample = Sfx_Cache(sndid);

	if (!sample) return false;

	// v220-4 $stepsounds
	// v220-33 If player is dropping into water make gloop sound complete
	return Sfx_StartSound(
		sample,
		sndid == r3ds.jumpdn_id ? cfg.footstepvol : vol,
		1.f + (shiftup ? (M_FRandom() *.1f) : (M_FRandom() * -.1f)),
		ppt.viewplrmo,
		NULL,
		1.f,
		0.f,
		sfx_gloop && ppt.viewplrmo->z > ppt.viewplrmo->floorz ? SFX_DONT_STOP : 0);
}

//===========================================================================
// S_LocalSoundAtVolume
//	Plays a sound on the local system.
//	This is a public sound interface.
//	Returns nonzero if a sound was started.
//===========================================================================
int S_LocalSoundAtVolume(int sound_id, mobj_t *origin, float volume)
{
	if (sound_id & HSF_REPEAT)
		Con_Error("ERROR: S_LocalSoundAtVolume sent repeating sound [id %i]",
				sound_id & ~HSF_REPEAT);

	return S_LocalSoundAtVolumeFrom(sound_id, origin, NULL, volume, DEF_ATTN);
}

//===========================================================================
// S_LocalSound
//	This is a public sound interface.
//	Returns nonzero if a sound was started.
//===========================================================================
int S_LocalSound(int sound_id, mobj_t *origin)
{
	if (sound_id & HSF_REPEAT)
		Con_Error("ERROR: S_LocalSound sent repeating sound [id %i]",
				sound_id & ~HSF_REPEAT);

	// Play local sound at max volume
	return S_LocalSoundAtVolumeFrom(sound_id, origin, NULL, 1.0f, DEF_ATTN);
}

//===========================================================================
// S_LocalSoundAtPos
//	This is a public sound interface.  Coords sent in fixedPos
//	Returns nonzero if a sound was started.
//===========================================================================
int S_LocalSoundAtPos(int sound_id, float *fixedPos, float volume)
{
	if (sound_id & HSF_REPEAT)
		Con_Error("ERROR: S_LocalSoundAtPos sent repeating sound [id %i]",
				sound_id & ~HSF_REPEAT);

	return S_LocalSoundAtVolumeFrom(sound_id, NULL, fixedPos, volume, DEF_ATTN);
}

//===========================================================================
// S_LocalMenuSound
//  All 2D unattenuated menu sounds are directed here. In addition
//	this is used where a nulled pwad sound has been substituted.
//  Sounds can be nulled by wad authors when they don't want it to
//  play in the game.  At load time any such sounds are substituted
//  for use in the menu (i.e. equivalent named iwad sound). R3D
//===========================================================================
int S_LocalMenuSound(int sound_id)
{
	return S_LocalSound(sound_id|HSF_MENUSOUND, NULL);
}

// v220-4 (Used for ui button click)
int S_LocalMenuSoundAtVolume(int sound_id, float vol)
{
	return S_LocalSoundAtVolume(sound_id|HSF_MENUSOUND, NULL, vol);
}

//=========================================================================
// S_StartSound
//=========================================================================
int S_StartSound(int sound_id, mobj_t *origin)
{
	if (!soundValid(sound_id)) return false;

	// test origin as both could be unset
	if (   origin && origin == ppt.viewplrmo
		&& (origin->flags & MF_NOCLIP || origin->health <= 0))
	{
		if (   (sound_id & ~HSF_FLAG_MASK) == sfx_oof
			|| (sound_id & ~HSF_FLAG_MASK) == sfx_noway)
		{
			return false;
		}
	}

	if (sound_id & HSF_NO_ATTENUATION) origin = NULL;
	
	return S_LocalSound(sound_id, origin);
}

//=========================================================================
// S_StartSoundAtVolume
//	Play a sound at the origin relative to the listener.
//  Returns nonzero if a sound was started.
//=========================================================================
int S_StartSoundAtVolume(int sound_id, mobj_t *origin,
									float volume, float attenuation)
{
	if (sound_id & HSF_REPEAT)
		return S_SoundRepeat(sound_id, origin, volume, attenuation);

	if (!soundValid(sound_id)) return false;

	if (sound_id & HSF_NO_ATTENUATION) origin = NULL;

	S_StopSound(sound_id, origin);

	if (sound_id & HSF_CANSUSPEND)
		Sfx_StartLogical(sound_id, origin);

	return S_LocalSoundAtVolumeFrom(sound_id, origin, NULL, volume,
															attenuation);
}

//=========================================================================
// S_StartSoundAtVolumeNoStop
//	Play a sound at the origin relative to the listener.
//  Returns nonzero if a sound was started.
//=========================================================================
int S_StartSoundAtVolumeNoStop(int sound_id, mobj_t *origin,
									float volume, float attenuation)
{
	if (sound_id & HSF_FLAG_MASK) return false;

	return S_LocalSoundAtVolumeFrom(sound_id, origin, NULL, volume,
															attenuation);
}

//=========================================================================
// S_SoundRepeat
//  b211-0 Both starts and stops repeating sounds
//  An origin is expected
//=========================================================================
int S_SoundRepeat(int sound_id, mobj_t *origin,
									float volume, float attenuation)
{
	RDBOOL valid_dist;

	if (!soundValid(sound_id)) return false;

	if (    !(sound_id & HSF_REPEAT)
		|| (!(sound_id & HSF_NO_ATTENUATION) && !origin))
	{
		return false;
	}

	if (sound_id & HSF_NO_ATTENUATION)
	{
		valid_dist = true; origin = NULL;
		ma_inf.sound_distance = 0;
	}
	else
	{
		valid_dist = S_CheckSoundDistance(origin, NULL, attenuation);
	}


	if (!S_IsPlaying(sound_id, origin))
	{
		if (!valid_dist) return false;

		Sfx_StartLogical(sound_id, origin);
		return S_LocalSoundAtVolumeFrom(sound_id, origin, NULL,
													volume, attenuation);
	}
	else if (!valid_dist || !Sfx_CheckChannel(sound_id, origin))
	{
		S_StopSound(sound_id, origin);
		return false;
	}

	// Update the cache start time to avoid problems
	// caused by the sound getting unloaded
	Sfx_CacheHit(sound_id & ~HSF_FLAG_MASK);

	return true;
}

//=========================================================================
// S_ConsoleSound
//	Play a player sound. Only the specified player will hear it.
//  R3D - Was true but target_console now ignored
//	Returns nonzero if a sound was started (always).
//=========================================================================
int S_ConsoleSound(int sound_id, mobj_t *origin, int target_console)
{
	if (!soundValid(sound_id)) return false;

	S_LocalSound(sound_id, origin);

	return true;
}

//===========================================================================
// S_StopSound
//	If sound_id == 0, then stops all sounds of the origin.
//	If origin == NULL, stops all sounds with the ID.
//	Otherwise both ID and origin must match.
//
//  Modified to use an explicit array - R3D
//
//===========================================================================
int S_StopSound(int sound_id, mobj_t *emitter)
{
	if (!soundValid(sound_id)) return false;

	Sfx_StopSound(sound_id, emitter);
	Sfx_StopLogical(sound_id, emitter);	// Notify the LSM.

	return false;
}

//===========================================================================
// S_StopAllSounds - R3D
//  Stops all sounds belonging to the origin
//  If origin is NULL then all sounds will be stopped
//===========================================================================
int S_StopAllSounds(mobj_t *origin)
{
	if (sfx_avail)
	{
		Sfx_StopSound(0, origin);
		Sfx_StopLogical(0, origin);	// Notify the LSM.
	}
	return false;
}

RDBOOL Sfx_IsPlaying(int id, mobj_t *origin);

//===========================================================================
// S_IsPlaying
//	Returns true if an instance of the sound is playing with the given
//	emitter. If sound_id is zero, returns true if the source is emitting
//	any sounds. An exported function.
//===========================================================================
RDBOOL S_IsPlaying(int sound_id, mobj_t *emitter)
{
	if (!soundValid(sound_id)) return false;

	return Sfx_IsPlaying(sound_id, emitter);
}

//===========================================================================
// S_StartMusicNum
//	Start a def song based on the index to defs.music.
//===========================================================================
int S_StartMusicNum(int id)
{
	if (id >= 0 || id < defs.count.music.num)
		return Music_GetMUS(defs.music + id);

	Music_Stop();

	return false;
}

//===========================================================================
// S_StartMusic
//	Returns true if the song is found. 
//===========================================================================
int S_StartMusic(char *musicid)
{
	if (!compat[c_nomusic])
	{
		int idx = Def_GetMusicNum(musicid);
		
		if(idx < 0)
		{
			Con_Message("S_StartMusic: song %s not defined.\n", musicid);
			return false;
		}
		return S_StartMusicNum(idx);
	}

	return false;
}

//===========================================================================
// S_StopMusic
//	Stops playing a song.
//===========================================================================
void S_StopMusic(void)
{
	if (!compat[c_nomusic]) Music_Stop();
}
