//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2012 G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: s_mus.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified for R3D
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
//  revised for mixer removal
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
//  All sound revised (including in dd_main.c, s_mus.c, sys_fmod.c,
//        sys_midi.c, f_infine.c, p_saveg.c, p_setup.c, wi_stuff.c)
//
// Based on: Doomsday v1.7.8
//
// DESCRIPTION:  midi/cd interface
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <string.h>
#include <io.h>
#include <direct.h>

#include "de_base.h"
#include "de_audio.h"
#include "de_console.h"
#include "de_system.h"
#include "de_misc.h"

#include "r3d_mus2mid.h"
#include "r3d_compat.h"

// MACROS ------------------------------------------------------------------

enum {
	PM_NONE,
	PM_MAP,
	PM_LIST,
	PM_LUMP,
	PM_FILE,
	PM_DEFN
};


// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

int		S_GetLISTnum(char *ptr);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern char		midiTempFname[256];

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int currentMapMusLumpnum = -1;

// v220-13
int global_looped;

char	*pspNone = "NONE";
char	*playStatus = pspNone;
char	szFNAME[42];
char	music_playing[96];
char	spare;

// Setting these variables is enough to adjust the volumes.
// S_StartFrame() will call the actual routines to change the volume
// when there are changes.
int		sfx_volume = 128;
int		mid_volume = 128;
int		ext_volume = 128;
int		pos_slider = 0;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static RDBOOL	mustplay = false;

static int		old_mid_volume	= -1;
static int		old_ext_volume	= -1;


// CODE --------------------------------------------------------------------

char *getFNAME(char *path)
{
	static char	out[514];

	char szfname[MAXFNAME];
	char szdrv[MAXDRIVE];
	char szext[MAXEXT];

	_splitpath(path, szdrv, NULL, szfname, szext);

	if (strlen(szext) > 4) return NULL;
	if (strlen(szfname) > 40) szfname[40] = 0;

	strcpy(szFNAME, szfname);

	// v220-9 if path has no drive then assume related to runtime path
	if (!szdrv[0])
	{
		strcpy(out, ddRuntimeDir.path);

		if (path[0] == '\\' || path[0] == '/')
		{
			strcat(out, path + 1);
		}
		else
		{
			strcat(out, path);
		}
	}
	else
	{
		strcpy(out, path);
	}

	return out;
}

//===========================================================================
// Check for volume change b220-1
//===========================================================================
void Music_VolumeSet()
{
	MC_WinMidiTime();	// v220-13 $midi

	if (old_ext_volume != ext_volume)
	{
		old_ext_volume = ext_volume;
		FModVolumeSet(ext_volume);
	}

	if (old_mid_volume != mid_volume)
	{
		old_mid_volume = mid_volume;
		MC_MidiVolumeSet(mid_volume);
	}
}

//===========================================================================
// Stops the music
//===========================================================================

void Music_Stop()
{
	playStatus = pspNone;
	FModStop();
	MC_MidiStop();
	midi_selected = false;
	music_playing[0] = 0;
	music_status = 0;
	mustplay = true;
	sld_musicpos.value = 0.f;
	sld_musicpos.max = 0.f;
	musicpos = 0xFFFFFFFF;
}

static RDBOOL createTempMIDIfile(int lump, char *name)
{
	FILE	*f;
	size_t	psize;
	void	*ptr = M_lumpCACHE(lump, &psize, MLC_LOCK, MLC_NOERR);

	if ( (f = fopen(name, "wb" )) == NULL) return false;

	fwrite(ptr, psize, 1, f);
	fclose(f);

	M_FreelumpCACHE();

	return true;
}

// v220-13 Only called from ui and only if midi file is active
void playTempMIDIResetFile()
{
	extern char  *midi_mname;
	extern char  lastmidifname[256];
	static char	 lastMidiName[42];
	static char	 replayMidiName[256];
	char		 resetName[256];

	if (ge_inf.rolandResetLmp < 0) return;

	strcpy(lastMidiName, midi_mname);
	strcpy(replayMidiName, lastmidifname);

	MC_MidiStop();

	sprintf(resetName, "%sr3dmdrst", ddRuntimeDir.path);

	if (createTempMIDIfile(ge_inf.rolandResetLmp, resetName))
	{
		int		tmp = global_looped;

		global_looped = 0;

		// Allow time for midi instrument to reset before restarting
		if (MC_PlayMidiFile(resetName, "")) Sys_Sleep(1000);

		global_looped = tmp;
	}

	MC_MidiStop();
	MC_PlayMidiFile(replayMidiName, lastMidiName);
}

RDBOOL CheckAndPlayFile(int lumpnum, char *calledfrom)
{
	static char	status[32];
	int		type, num;

	if (compat[c_nomusic]) return false; // disabled
#ifdef _DEBUG
	Con_Printf("CheckAndPlayFile: called from %s [lumpnum %i]\n", calledfrom, lumpnum);
#endif
	
	if (  lumpnum >= 0
		&& (gamestate != GS_LEVEL && (num = S_GetLISTnum(lumpinfo[lumpnum].name)) )
		&& PlayMusListExt(num))
	{
		return true;
	}

	Music_Stop();

	currentMapMusLumpnum = lumpnum;	// v220-3

	if (lumpnum < 0)
	{
		playStatus = pspNone;
		return false;	// No music set
	}

	playStatus = lumpinfo[lumpnum].handle == IWADhandle ? "IWAD" : "PWAD" ;
	strcpy(status, playStatus);

	type = Mus_IsMUSorMIDILump(lumpnum);
	
	switch(type)
	{
	case M2M_MIDI:
	case M2M_RMI:
		if (lumpinfo[lumpnum].size > 255)
		{
			char *mname = lumpinfo[lumpnum].name;

			if (createTempMIDIfile(lumpnum, midiTempFname))
			{
				midi_selected = true;
				strcat(status, " MIDI");
				playStatus = status;

				if (MC_PlayMidiFile(midiTempFname, mname))
				{
					mustplay = false;
					return true;
				}

				return false;
			}
		}
		break;
	case M2M_MUS:
		if (ConvertMusToMidi(lumpnum, midiTempFname))
		{
			char *mname = lumpinfo[lumpnum].name;

			midi_selected = true;
			strcat(status, " MUS");
			playStatus = status;
			
			if (MC_PlayMidiFile(midiTempFname, mname))
			{
				mustplay = false;
				return true;
			}

			return false;
		}
		break;
	case M2M_BADLUMP: return false;	// graphics or something else
	default:
		if (FModPlay(NULL, NULL, lumpnum, "CheckAndPlayFile", calledfrom))
		{
			strcat(status, " SONG");
			playStatus = status;
			mustplay = false;
			return true;
		}
		break;
	}

	strcat(status, " LUMP UNSET");
	playStatus = status;
	
	return false;
}

//===========================================================================
//	The lump may contain non-MUS data. Returns true if successful.
//===========================================================================
int Music_GetMUS(ded_music_t *def)
{
	if (   def->path.path[0]
		&& PlayMapinfoFile(def->path.path))
	{
		return true;
	}

	return CheckAndPlayFile(W_CheckNumForName(def->lumpname), "Music_GetMUS");
}

//===========================================================================
// Used by INFINE A valid lumpnum must be sent
//===========================================================================
int Music_PlayInter(int lumpnum)
{
	return CheckAndPlayFile(lumpnum, "Music_PlayInter");
}

//===========================================================================
// Used by MT_MUSICSOURCE/PM MAP/INFINE	R3D
//===========================================================================
void Music_PlayLump(int lumpnum)
{
	if (compat[c_nomusic]) return;	// No way of playing whatever it is.

	if (currentMapMusLumpnum != lumpnum)
	{
		// Remember lumpnum. We might want to go back
		// to this plus needs to be stored in a savegame.
		currentMapMusLumpnum = lumpnum;

		// MT_MUSICSOURCE does not override list music
		if (!_strnicmp(playStatus, "LIST", 4)) return;
	}

	CheckAndPlayFile(lumpnum, "Music_PlayLump");
}


// Used with external files.
// Expects full file path and name
// This is just a quick initial check it does not
// fully verify the file is a mdid/riff type.
//
// returns -1 if the file was not found or was a MUS type
//          0 if opened OK
//          1 if a midi or riff file
//
static int checkIfMidiFile(char *msg, char *path)
{
	DFILE		*f = F_OpenFile(path, "rb");
	WAVEhdr_t	*WAVEHdr;
	byte		data[128];
	int			ret = -1;

	if (!f)
	{
		pr_OFinsert("%s %s was not found\n", msg, path);
		return ret;
	}

	ret = 0;
	memset(data, 0, 128);

	// Check if a MUS type
	if (F_Read(data, sizeof(MUShdr_t) + 4, f) > sizeof(MUShdr_t) + 1)
	{
		void	*ptr = (void*)data;
		MUShdr_t musHdr = *(MUShdr_t*) ptr;

		if (    DWORDSWAP(musHdr.ID) == MUS_HDR_ID
			&&  musHdr.muschannels 
			&&  musHdr.muschannels < MIDI_MAXNUMCHANS)
		{
			int len;

			F_Rewind(f);
			F_Seek(f, 0, SEEK_END);
			len = F_Tell(f);
			
			if (musHdr.scoreLen + musHdr.scoreStart <= len)
			{
				pr_OFinsert("%s %s\n External MUS files are not supported\n", msg, path);
				F_Close(f);
				return -1;
			}
		}
	}

	F_Rewind(f);

	if (F_Read(data, sizeof(WAVEhdr_t) + 4, f) > sizeof(WAVEhdr_t) + 1)
	{
		void	*ptr = (void*)data;

		WAVEHdr = (WAVEhdr_t*)ptr;

		if (   !strncmp(WAVEHdr->riff,"RIFF",4)
			&& !strncmp(WAVEHdr->wave,"WAVEfmt ",8) )
		{
			// do nothing. This is a wav file
		}
		else
		if (!strncmp(WAVEHdr->riff,"MThd",4))
		{
			ret = 1;
		}
		else
		if (   !strncmp(WAVEHdr->riff,"RIFF",4)
			&& !strncmp(WAVEHdr->wave,"RMIDdata",8))
		{
			ret = 1;
		}
	}

	F_Close(f);
	return ret;
}

static int checkFileExternal(char *msg, char *path)
{
	int type = checkIfMidiFile(msg, path);

	if (type < 0) return false;

	Music_Stop();

	if (   ( type && MC_PlayMidiFile(path, szFNAME))
		|| (!type && FModPlay(path, szFNAME, 0, "checkFileExternal", "")) )
	{
		playStatus = !strncmp(msg, "LIST", 4) ? "LIST" : type ? "MIDI" : "FILE";
		midi_selected = type;
		mustplay = false;
		return true;
	}

	return false;
}

//===========================================================================
// PlayMapinfoFile - b211-0
//===========================================================================
RDBOOL PlayMapinfoFile(char *fname)
{
	if (compat[c_nomusic]) return false;

	if (!checkFileExternal("Map Info File", getFNAME(fname)) )
	{
		Con_Printf("PlayMapinfoFile: Dropped %s\n", fname);
		return false;
	}
	return true;
}

//===========================================================================
// PlayMapinfoFile - b211-0
//===========================================================================
RDBOOL PlayScriptedFile(char *fname)
{
	if (compat[c_nomusic]) return false;

	if (!checkFileExternal("Scripted File", getFNAME(fname)) )
	{
		Con_Printf("PlayMusListExt: Dropped %s\n", fname);
		return false;
	}
	return true;
}

//===========================================================================
// PlayMusListExt - R3D
//===========================================================================
int PlayMusListExt(int idx)
{
	if (!compat[c_nomusic])
	{
		musiclist_t *ml = &musicLIST[idx];

		playStatus = "NONE";

		if (ml->mlist_type == MLIST_FILE)
		{
			// first 4 chars must be LIST
			if (checkFileExternal("LIST File", getFNAME(ml->fname)) )
				return true;
			Con_Printf("PlayMusListExt: Dropped %i %s\n", idx, getFNAME(ml->fname));
		}
	}

	return false;
}

//===========================================================================
// CCmdPlayMusic
//===========================================================================
int CCmdPlayMusic(int argc, char **argv)
{
	int		i, type = PM_NONE;

	if (compat[c_nomusic])
	{
		Con_Message("music not enabled.\n");
		return true;
	}

	if (gamestate != GS_LEVEL)
	{
		Con_Message("Music can only be set when in game\n");
		return true;
	}

	if (argc == 2 || argc == 3)
	{
		if (     !_stricmp(argv[1], "map" )) type = PM_MAP;
		else if (!_stricmp(argv[1], "list")) type = PM_LIST;
		else if (!_stricmp(argv[1], "lump")) type = PM_LUMP;
		else if (!_stricmp(argv[1], "file")) type = PM_FILE;
		else if (argc == 2) type = PM_DEFN;
	}

	if (argc == 2)
	{
		if (type == PM_MAP)
		{
			int lump = currentMapMusLumpnum >= 0 ?
					currentMapMusLumpnum : mapinfo->muslump;

			Music_Stop();

			if (lump >= 0 && CheckAndPlayFile(lump, "CCmdPlayMusic"))
				return true;

			Con_Message("Map music not set\n");
			return false;
		}
		else if (type == PM_LIST)
		{
			if (PlayMusListExt(W_GetMapIndex(mapinfo->id)))
			{
				ge_inf.UseList = 3;
				return true;
			}

			Con_Message("No list entries defined.\n");
			return false;
		}
		else if (type == PM_DEFN)
		{
			if ((i = Def_GetMusicNum(argv[1])) >= 0)
			{
				if (Music_GetMUS(&defs.music[i]))
					return true;
				Con_Message("'%s' could not be played\n", argv[1]);
			}
			else
			{
				Con_Message("'%s' not found in def files\n", argv[1]);
			}
			return false;
		}
	}

	if (argc == 3)
	{
		if (type == PM_LUMP)
		{
			return CheckAndPlayFile(W_CheckNumForName(argv[2]), "CCmdPlayMusic");
		}
		if (type == PM_FILE)
		{
			char buf[300];
			char *ptr;

			M_TranslatePath(argv[2], buf);
			ptr = getFNAME(buf);

			return checkFileExternal("PM File", ptr);
		}
	}

	if (argc >= 2)
		Con_Message("command(s) unrecognised\n");

	Con_Message("Usage:\n  %s <music-def>\n", argv[0]);
	Con_Message("  %s lump <lumpname>\n", argv[0]);
	Con_Message("  %s file <path/filename>\n", argv[0]);
	Con_Message("  %s list\n", argv[0]);
	Con_Message("  %s map (returns to default map music)\n", argv[0]);
	return true;
}

//===========================================================================
// CCmdStopMusic
//===========================================================================
int CCmdStopMusic(int argc, char **argv)
{
	if (!compat[c_nomusic]) Music_Stop();

	return true;
}

//===========================================================================
// CCmdRebuildMidi - R3D
//===========================================================================
int CCmdRebuildMidi(int argc, char **argv)
{
	if (!compat[c_nomusic] && currentMapMusLumpnum >= 0)
	{
		return CheckAndPlayFile(currentMapMusLumpnum, "CCmdRebuildMidi");
	}

	return false;
}

//===========================================================================
// CCmdToggleList
//===========================================================================
int CCmdToggleList(int argc, char **argv)
{
	if ((gamestate == GS_LEVEL) && ge_inf.UseList & 1)
	{
		ge_inf.UseList = ge_inf.UseList & 2 ? 1 : 3;
		S_LevelMusic();
	}
	return true;
}
