//-----------------------------------------------------------------------------
//
// Copyright (C) 1999 Randy Heit
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: s_sound.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Includes partial support for ZDOOM SNDINFO script
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 14 (v220-14) 01/May/2012 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: ZDOOM v1.22 December 12, 1999 (dis. marked v1.23beta33)
//
// DESCRIPTION: Ambient sound loader/handler
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdlib.h>
#include <string.h>

#include "r3d_compat.h"

#include "doom/mbf_player.h"
#include "p_data.h"

#include "con_console.h"
#include "dd_wad.h"
#include "net_main.h"
#include "p_maputil.h"
#include "s_main.h"
#include "m_misc.h"
#include "s_logic.h"
#include "s_cache.h"

#include "r3d_utils.h"

// MACROS ------------------------------------------------------------------

#define MAXLOGIC	128	// Max number of aliases supported

#define AS_ACTIVE	1
#define AS_TRIGGER	2

// ambientsnd_t type flags
#define AMB_POINT		1
#define AMB_RANDOM		2
#define AMB_PERIODIC	3
#define AMB_CONTINUOUS	4
#define AMB_WORLD		0x10
#define AMB_ONCE		0x20

enum SICommands
{
	SI_Ambient,
	SI_World,
	SI_Random,
	SI_PlayerReserve,
	SI_PlayerSound,
	SI_PlayerSoundDup,
	SI_PlayerCompat,
	SI_PlayerAlias,
	SI_Alias,
	SI_Limit,
	SI_PitchShift,
	SI_PitchShiftRange,
	SI_Map,
	SI_Registered,
	SI_ArchivePath,
	SI_IfDoom
};

// TYPES -------------------------------------------------------------------

typedef struct
{
	byte		type;			// type of ambient sound
	byte		as_flags;		// activation flags
	int			periodmin;		// # of tics between repeats
	int			periodmax;		// max # of tics for random ambients
	int			sound_id;
	float		volume;			// relative volume of sound
	float		attenuation;	// attenuation
	char		lumpname[10];	// lumpname of sound to play
	int			lgcnum;			// link to lgcalias
	mobj_t		*mo;
} ambientsnd_t;

typedef struct
{
	char	alias[32];
	char	lumpname[10];
	int		sound_no;
}logicalsnd_t;

typedef struct
{
	char	*alias;
	int		sound_no;
}aliases_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

musiclist_t musicLIST[MAX_LIST_ENTRIES];
char		*listName;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// Order must match SICommands
static const char *SICommandStrings[] =
{
	"$ambient",
	"$world",
	"$random",
	"$playerreserve",
	"$playersound",
	"$playersounddup",
	"$playercompat",
	"$playeralias",
	"$alias",
	"$limit",
	"$pitchshift",
	"$pitchshiftrange",
	"$map",
	"$registered",
	"$archivepath",
	"$ifdoom",
	NULL
};

static aliases_t aliases[] =
{
	"weapons/sawup",		sfx_sawup,
	"weapons/sawidle",		sfx_sawidl,
	"weapons/sawfull",		sfx_sawful,
	"weapons/sawhit",		sfx_sawhit,
	"weapons/pistol",		sfx_pistol,
	"weapons/shotgf",		sfx_shotgn,
	"weapons/shotgr",		sfx_sgcock,
	"weapons/sshotf",		sfx_dshtgn,
	"weapons/sshoto",		sfx_dbopn,
	"weapons/sshotc",		sfx_dbcls,
	"weapons/sshotl",		sfx_dbload,
	"weapons/chngun",		sfx_chgun,
	"weapons/rocklx",		sfx_barexp,
	"weapons/rocklf",		sfx_rlaunc,
	"weapons/plasmaf",		sfx_plasma,
	"weapons/bfgf",			sfx_bfg,
	"plats/pt1_strt",		sfx_pstart,
	"plats/pt1_stop",		sfx_pstop,
	"plats/pt1_mid",		sfx_stnmov,
	"doors/dr1_open",		sfx_doropn,
	"doors/dr1_clos",		sfx_dorcls,
	"doors/dr2_open",		sfx_bdopn,
	"doors/dr2_clos",		sfx_bdcls,
	"world/barrelx",		sfx_None,
	"world/drip",			sfx_None,
	"world/watersplash",	sfx_None,
	"world/sludgegloop",	sfx_None,
	"world/lavasizzle",		sfx_None,
	"misc/secret",			sfx_getpow,
	"misc/w_pkup",			sfx_wpnup,	// Pickup weapon
	"misc/p_pkup",			sfx_getpow,	// Pickup powerup
	"misc/i_pkup",			sfx_itemup,	// Pickup item
	"misc/k_pkup",			sfx_itemup,	// Pickup key
	"misc/spawn	",			sfx_telept,	// Item respawn
	"misc/chat",			sfx_radio,	// Doom 2 chat sound
	"misc/chat2",			sfx_tink,	// Chat sound for everything else
	NULL,					0
};

static ambientsnd_t *Ambients[MAXAMBIENT];

static int		logical_count;
static int		world_count;
static int		world_ambients[MAXAMBIENT];
static int		*ambtimers;

// CODE --------------------------------------------------------------------

int S_GetLISTnum(char *ptr)
{
	static const char *interMus[] = {
		"D_BUNNY",
		"D_INTER",
		"D_INTRO",
		"D_INTROA",
		"D_VICTOR",
		"D_READ_M",
		"D_EVIL",
		"D_DM2INT",
		"D_DM2TTL",
		NULL
	};

	int i;

	//-------------------

	if (!ptr) return 0;
	
	if (M_ValidateID(ptr)) return W_GetMapIndex(ptr);

	for (i = 0; interMus[i]; i++)
	{
		if (!stricmp(interMus[i], ptr))
			return 100 + i;
	}

	return 0;
}

static RDBOOL checkIfMap(char *id)
{
	int num;

	if (strlen(id) == 5)
	{
		return !strnicmp(id, "map", 3)
			&& (num = atoi(&id[3])) > 0
			&& num <= 99;
	}

	return strlen(id) == 4
		&& (id[0] == 'E' || id[0] == 'e')
		&& (id[2] == 'M' || id[2] == 'm')
		&& (num = id[1] - '0') > 0 && num <= 9
		&& (num = id[3] - '0') > 0 && num <= 9;
}

// v220-13 modified
void S_InitMUSLIST(char *path)
{
	if (compat[c_nomusic]) return;

	// Set music list path.
	if (ArgCheckWith("-extmus", 1))
	{
		musiclist_t *ml;
		char	buf[512];
		char	list[64];
		char	*nextstr;
		char	*ptr = ArgNext();	// List name
		int		i, len, count = 0;

		char prefix[] = { "LIST: " };
		char szName[256];
		uint uLimit = 27;

		//------------------------

		if (!ptr || *ptr == '-' || !(len = strlen(ptr)) || len > 511) return;

		strcpy(buf, ptr);

		// Try and deal with home brewed batch file anomalies
		// Top and tail spaces (for names that were enclosed in quotes)
		ptr = buf + len - 1;
		while(ptr >= buf && *ptr == ' ') *(ptr--) = 0;
		if ( !(len = strlen(buf)) ) return;
		ptr = buf;
		while(*ptr == ' ') ptr++;
		if ( !(len = strlen(ptr)) ) return;

		if (len > 63)
			Con_Error("LIST ERROR: list name length exceeds max allowed of 63;\n%s", ptr);

		strcpy(list, ptr);

		// Quick *.bat hack for those who like to fuck about using a batch
		// file rather than use the Launcher or short cuts. Batch files
		// cause  to get changed to  for example. Thus Lorcn becomes
		// Lorcn. This is the only character covered at present.
		//
		for (i = 0; i < len; i++) if (list[i] == '') list[i] = '';

		sprintf(buf, "%sr3dlaunch\\music\\%s.r3m", path, list);

		if (!SR_LoadFile(buf, "MUSIC FILE", NULL))
			Con_Error("LIST ERROR: \"%s\" not found;\n", buf);

		listName = (char*)W_Malloc(uLimit + strlen(prefix) + 1, STATICID, 48);

		strcpy(listName, prefix);

		_splitpath(buf, NULL, NULL, szName, NULL);

		if (strlen(szName) > uLimit)
		{
			szName[uLimit - 2] = 0;
			strcat(szName, "..");
		}

		strcat(listName, szName);

		while( (nextstr = SR_NextSTR(false)) )
		{
			RDBOOL	isValid = true;
			int		num = S_GetLISTnum(nextstr);

			if (!num)
			{
				if (checkIfMap(nextstr))
					isValid = false;
				else
					Con_Error("LIST ERROR: \"%s\" in list \"%s\" is invalid;\n", nextstr, list);
			}

			ml = &musicLIST[num];
			
			if ( !(nextstr = SR_NextSTR(false)) ) break;

			// Check if an entry was made for this id..
			if (checkIfMap(nextstr) || S_GetLISTnum(nextstr))
			{
				// No, this is the next id, so continue
				SR_KeepString();
				continue;
			}

			if (!_stricmp(nextstr, "CD"))
			{
				Con_Error("LIST ERROR: \"%s\" CD no longer supported;\n", nextstr);
			}
			else if (isValid)
			{
				uint sz;
				
				M_PrependBasePath(nextstr, buf);

				if ( (_access(buf, 0)) == -1)
				{
					Con_Message("LIST ERROR: \"%s\" not found;\n", nextstr);
					continue;
				}

				sz = strlen(buf);
				ml->mlist_type = MLIST_FILE;
				ml->fname = (char*)W_Malloc(sz + 1, STATICID, 49);
				strcpy(ml->fname, buf);
				count++;
			}
		}

		if (!count)
			pr_OFinsert("MUSIC LIST %s has no entries ======\n", szName);
		else
			pr_OFinsert("MUSIC LIST %s Successfully read %i entries ======\n", szName, count);

		SR_EndScript();
	}
}

// v220-14
void S_SetLISTge_inf(int listnum, int lumpnum)
{
	if (   ge_inf.usepwadmusic
		&& lumpnum >= 0
		&& lumpnum < numlumps
		&& lumpinfo[lumpnum].handle != IWADhandle)
	{
		ge_inf.UseList = 0;
	}
	else if (listnum >= 0
		&& listnum < MAX_LIST_ENTRIES
		&& musicLIST[listnum].mlist_type)
	{
		ge_inf.UseList = 3;
	}
	else
	{
		ge_inf.UseList = 0;
	}
}

void S_InitMUSScript(char *mapid)
{
	int		num, lumpnum, lastlump = 0;
	RDBOOL	inMap = false;
	RDBOOL	muscript = W_CheckNumForName("MUSINFO") != -1;
	char	*nextstr;

	memset(music_array, 0, sizeof(music_array));

	if (!muscript) return;

	while ((lumpnum = W_FindScriptLump("MUSINFO", &lastlump)) != -1)
	{
		SR_LoadLump(lumpnum, "MUSINFO");

		while( (nextstr = SR_NextSTR(false)) )
		{
			// v220-31 bug fix
			char *ptr = nextstr;

			if (*ptr == '$') ptr++;

			if (M_ValidateID(ptr) && (inMap = !_stricmp(ptr, mapid)))
			{
				nextstr = SR_NextSTR(true);
			}

			if (inMap)	// b211-0
			{
				num = SR_NumForString(NULL);

				// Check number in range
				if (num < 1 || num >= MAX_MUS_ENTRIES)
					SR_SCPErr("Number %i not in range 1 to %i", num, MAX_MUS_ENTRIES - 1);

				nextstr = SR_NextSTR(true);

				if ( (lumpnum = W_CheckNumForName(nextstr)) < 0)
				{
					SR_SCPErr("Unknown MUS lump : %s", nextstr);
				}

				music_array[num] = lumpnum;

				// v220-13  Reading the file means these may stay in memory
				//          making for faster loading when needed.
				{
					uint	size = lumpinfo[lumpnum].size;
					byte	*p;

					if (   size > 0x4000	// don't bother if < 16384B
						&& size < 0x2000000	// don't try if > 32MB
						&& (p = (byte*)malloc(size)) )
					{
						read(lumpinfo[lumpnum].handle, (void*)p, size);
						free(p);
					}
				}
			}
		}

		SR_EndScript();
	}
}

void S_InitAMBIENTS()
{
	// v220-6 changed to LEVELID
	ambtimers = (int*)W_Calloc(numthings * sizeof(int), LEVELID, 90);
}

static int subfn_StrMatch(char *instr, const char **strings)
{
	int i;

	for(i = 0; *strings != NULL; i++)
	{
		if (!_stricmp(instr, *strings++))
			return i;
	}
	return -1;
}

static char *subfn_gotoNextIden()
{
	char *nextstr;

	while( (nextstr = SR_NextSTR(false)) )
	{
		if (*nextstr == '$') return nextstr;
	}
	return NULL;
}

// This is only run if sfx_avail == true and then only once	
void S_InitSNDScript()
{
	ambientsnd_t *as;
	logicalsnd_t *lgcalias;
	int			 i, found, lump;
	int			 logical_preset = 0;
	char		 *nextstr;
	int			 nextnum;

	logical_count  = 0;
	world_count    = 0;

	if (   compat[c_nosound]
		|| ArgCheck("-nosndinfo")
		|| (lump = W_CheckPWADNumForName("SNDINFO")) < 0)
	{
		return;
	}

	Con_Message("Initializing SNDINFO Script\n");

	// Set up the logical alias array
	lgcalias = (logicalsnd_t*)W_Calloc(MAXLOGIC * sizeof(logicalsnd_t), STATICID, 142);

	for (i = 0; aliases[i].alias; i++)
	{
		int soundnum = aliases[i].sound_no;

		strcpy(lgcalias[i].alias, aliases[i].alias);
		strncpy(lgcalias[i].lumpname, R3D_GetNameForSoundID(soundnum), 8);
		lgcalias[i].sound_no = soundnum;
		logical_count++; logical_preset++;
	}

	SR_LoadLump(lump, "SNDINFO");

	while( (nextstr = SR_NextSTR(false)) )
	{
start:;
		if (*nextstr == '$')
		{
			int		type = subfn_StrMatch(nextstr, SICommandStrings);

			switch (type)
			{
// SI_Ambient
			case SI_Ambient:
				nextnum = SR_NextNum() - 1;

				// Check number in range
				if (nextnum < 0 || nextnum >= MAXAMBIENT)
					SR_SCPErr("Number %i not in range 1 to %i", MAXAMBIENT, nextnum + 1);

				if (Ambients[nextnum])
					SR_SCPErr("Ambient index %i already allocated", nextnum + 1);

				nextstr = SR_NextSTR(true);				// Get logical alias

				for (i = 0; i < logical_count; i++)		// Find logical alias
				{
					if (!_stricmp(lgcalias[i].alias, nextstr))
						break;
				}

				if (i == logical_count)
					SR_SCPErr(" Logical alias %s undefined.", nextstr);

				if (lgcalias[i].sound_no < 0)
				{
					nextstr = subfn_gotoNextIden();
					if (nextstr) goto start;
					break;
				}

				// Create new ambient
				as = (ambientsnd_t*)W_Calloc(sizeof(ambientsnd_t), STATICID, 143);
				Ambients[nextnum] = as;
				as->as_flags |= AS_ACTIVE;	// Set active by default
				strncpy(as->lumpname, lgcalias[i].lumpname, 8);
				as->sound_id = lgcalias[i].sound_no;
				as->sound_id |= HSF_AMBIENT;
				as->lgcnum = i;

				as->attenuation = 1.0f;		// default

				nextstr = SR_NextSTR(true);

				if (!_stricmp(nextstr, "POINT"))
				{
					RDBOOL	scperror;

					as->type = AMB_POINT;	// Don't need this - set by the ambient value
					nextstr = SR_NextSTR(true);

					as->attenuation = SR_FloatForString(&scperror);

					if (!scperror)
					{
						if (as->attenuation >= 0.1f)
						{
							if (as->attenuation > MAX_ATTN)
								as->attenuation = MAX_ATTN;
						}
						else
						{
							as->attenuation = 0;
						}

						nextstr = SR_NextSTR(true);
					}
					else
					{
						as->attenuation = 1;
					}
				}
				// World is an optional keyword - also don't need as->type = SURROUND
				else if (!_stricmp(nextstr,"SURROUND") || !_stricmp(nextstr,"WORLD"))
				{
					nextstr = SR_NextSTR(true);
				}

				if (!_stricmp(nextstr,"CONTINUOUS"))
				{
					as->type = AMB_CONTINUOUS;
				}
				else if (!_stricmp(nextstr,"RANDOM"))
				{
					int min, max;

					min = (int)(SR_NextFloat() * TICRATE);
					max = (int)(SR_NextFloat() * TICRATE);

					if (min > max) SR_SCPWarn("RANDOM sound MIN > MAX value; Values have been swapped");

					as->type = AMB_RANDOM;

					as->periodmin = min <= max ? min : max;
					as->periodmax = min >= max ? min : max;
				}
				else if (!_stricmp(nextstr,"PERIODIC"))
				{
					as->type = AMB_PERIODIC;
					as->periodmin = (int)(SR_NextFloat() * TICRATE);
				}
				else if (!_stricmp(nextstr,"ONCE"))
				{
					as->type |= AMB_ONCE;
					as->as_flags &= ~AS_ACTIVE;	// Special trigger type
				}
				else
				{
					SR_SCPErr("Ambient type %s undefined.", nextstr);
				}

				as->volume = SR_NextFloat();
				if      (as->volume > 1.0f) as->volume = 1.0f;
				else if (as->volume <    0) as->volume = 0.f;

				break;
// SI_World
			case SI_World:
				if (world_count >= MAXAMBIENT)
					SR_SCPErr("world ambients exceeds max of %i", MAXAMBIENT);

				// This is the referencing logical group name
				// Not currently used - needed to identify different groups
				nextstr = SR_NextSTR(true);

				nextstr = SR_NextSTR(true);

				if (!_stricmp(nextstr, "BBOX"))
				{
					// This has been removed
					SR_NextNum();
					SR_NextNum();
					SR_NextNum();
					SR_NextNum();
					nextstr = SR_NextSTR(true);
				}

				if (strcmp(nextstr, "{")) SR_SCPErr("Missing {");

				while ( (nextstr = SR_NextSTR(false)) )
				{
					int k, num = 0;

					if (!strcmp(nextstr, "}")) break;

					for (i = 0; i < logical_count; i++)
					{
						if (!_stricmp(lgcalias[i].alias, nextstr))
						{
							num = i; break;
						}
					}

					if (i == logical_count)
						SR_SCPErr("Alias \"%s\" not found", nextstr);

					for (i = 0, k = -1; i < MAXAMBIENT; i++)
					{
						if (Ambients[i] && Ambients[i]->lgcnum == num)
						{
							k = i; break;
						}
					}

					if (k < 0)
						SR_SCPErr("No Ambient linked to alias \"%s\" exists", nextstr);

					as = Ambients[k];

					// Build list of worlds
					if ( (as->type & 0x7) != AMB_RANDOM)
						SR_SCPErr("Grouped world ambients must be random type.");

					for (num = i = 0; i < world_count; i++)
					{
						if (world_ambients[i] == k)
						{
							SR_SCPWarn("alias %s already used", nextstr);
							num = 1;
							break;
						}
					}

					if (!num)
					{
						world_ambients[world_count++] = k;
						as->type |= AMB_WORLD;	// Indicate a world type
						as->attenuation = 0;
					}
				}
				break;

			default:
				// Unsupported types
				// This also filters out any Heretic or Hexen command
				// types none of which are supported by Risen3D.
				SR_SCPWarn("type %s unsupported.", nextstr);	// v220-8 bug fix: was %i

				if ( (nextstr = subfn_gotoNextIden()) )
					goto start;
				break;
			}
		}
		else // Must be a logical name
		{
			char	firstdef[32];
			int		soundnum;

			// The alias (first entry on line - can be up to 31 chars)
			if (strlen(nextstr) > 31)
				SR_SCPErr("Alias exceeds 31 chars");

			strncpy(firstdef, nextstr, 32);

			nextstr = SR_NextSTR(true);

			// The logical name
			if (strlen(nextstr) > 8)
				SR_SCPErr("Logical name exceeds 8 chars");

			found = false;

			// First check if a known alias
			for (i = 0; i < logical_count; i++)
			{
				if (!_stricmp(lgcalias[i].alias, firstdef))
				{
					if (i < logical_preset
						&& !R3D_ChangeSoundUsingAlias(lgcalias[i].sound_no, nextstr))
					{
						SR_SCPWarn("lump \"%s\" not found or invalid - sound will use doom default", nextstr);
					}
					found = true;	// Already exists if we get here
					break;
				}
			}

			if (!found)
			{
				if (logical_count >= MAXLOGIC)
					Con_Error("Max logical aliases %i exceeded", MAXLOGIC);

				strncpy(lgcalias[logical_count].alias   , firstdef ,31);
				strncpy(lgcalias[logical_count].lumpname, nextstr, 8);

				if ( !(soundnum = R3D_AddNewSound(nextstr)) )
						SR_SCPWarn("\"%s\" not found or invalid.", nextstr);

				lgcalias[logical_count++].sound_no = soundnum;
			}
		}
	}

	SR_EndScript();

	W_Free(lgcalias, STATICID, 534);

	Con_Message("%i Ambients defined\n", logical_count - logical_preset);
}

void S_CheckLineTriggeredAmbients()
{
	thinker_t	*th;
	line_t		*ld;
	mobj_t		*iter;
	RDBOOL		debug_scp = ArgExists("-debug_scp");

	if (!logical_count) return;

	for (th = thinkercap.next ; th != &thinkercap ; th = th->next)
	{
		iter = (mobj_t*)th;

		if (iter->thinker.function == (void (__cdecl *)(void))P_MobjThinker
			&& iter->type == MT_TRIGAMBIENT)
		{
			sector_t	*s = iter->subsector->sector;
			int			i, x = iter->x, y = iter->y;

			for (i = 0; i < s->linecount; i++)
			{
				if ( (ld = s->lines[i])->special
					&& PointIsOnLine(x, y, ld, 1))
				{
					int num = TIDNUM(iter);

					if (num >= 0 && num < MAXAMBIENT && Ambients[num])
					{
						if (debug_scp)
							Con_Message("MT_TRIGAMBIENT dm %4i num %3i line %5i special %4i snd %8.8s\n",
										MAPNUM(iter),num,ld-lines,ld->special,Ambients[num]->lumpname);

						Ambients[num]->as_flags &= ~AS_ACTIVE;
						ld->lnflags2 |= LNF2_AMBIENTTRIG;
						ld->ambNumLink = (byte)num;
					}
				}
			}
		}
	}
}

// v220-35
void S_SetLineTriggeredAmbient(int num, int from)
{
	if (num >= 0 && num < MAXAMBIENT && Ambients[num])
	{
		Ambients[num]->as_flags |= AS_ACTIVE;
	}
}

// v220-35
void S_StartAmbient(int num)
{
	if (num >= 0 && num < MAXAMBIENT && Ambients[num])
	{
		Ambients[num]->as_flags |= AS_ACTIVE;
	}
}

void S_StopAmbient(mobj_t *mo, int num)
{
	if (num >= 0 && num < MAXAMBIENT && Ambients[num])
	{
		S_StopSound(Ambients[num]->sound_id, Ambients[num]->mo);
		Ambients[num]->as_flags &= ~AS_ACTIVE;
	}
}

int S_GetAmbientCount()
{
	return logical_count;
}

static int getAmbTimerVal(ambientsnd_t *as, RDBOOL addrnd)
{
	int rnd = addrnd ?
		(int)(M_FRandom() * (float)(as->periodmax - as->periodmin)) : 0;

	if (rnd < 0) rnd = 0;

	return rnd + as->periodmin + Sfx_GetSoundLength(as->sound_id)*TICRATE/1000;
}
// b211-0 revised
void S_CheckAmbientSound(mobj_t *mo)
{
	ambientsnd_t	*as;
	int				ambient_num = TIDNUM(mo);
	int				snd_id;
	int				mapnum;

	if (   !sfx_avail
		|| ambient_num < 0
		|| ambient_num >= MAXAMBIENT
		|| !(as = Ambients[ambient_num])
		|| !(as->as_flags & AS_ACTIVE)		// line trigger flag
		|| (mapnum = MAPNUM(mo)) >= numthings)
	{
		return;
	}

	if ((as->type & 0x7) == AMB_RANDOM || as->type == AMB_PERIODIC)
	{
		if (as->type & AMB_WORLD && world_count > 0)
		{
			static int last_leveltime = 0;
			ambientsnd_t *now;
			int			 i;

			// We only want to check this once per tick
			if (last_leveltime == leveltime) return;
			
			last_leveltime = leveltime;

			// Don't play a world ambient immediately on starting
			if (ma_inf.world_starttime) return;

			//  Check the status of world ambients
			for (i = 0; i < world_count; i++)
			{
				now = Ambients[world_ambients[i]];

				// check it's not timing out
				if (   now->mo
					&& ambtimers[MAPNUM(now->mo)] > 0)
				{
					ambtimers[MAPNUM(now->mo)]--;
					return;
				}
			}

			// None are timing out so select one at random and play it.
			i = M_RandomD() % world_count;
			as = Ambients[world_ambients[i]];
			as->mo = mo;
			ambtimers[mapnum] = getAmbTimerVal(as, true);
		}
		else if (ambtimers[mapnum] > 0)
		{
			ambtimers[mapnum]--; return;
		}
	}

	snd_id = as->sound_id;

	if (!(as->type & AMB_WORLD))
	{
		if (as->type == AMB_PERIODIC)
		{
			 ambtimers[mapnum] = getAmbTimerVal(as, false);
		}
		else if (as->type == AMB_RANDOM)
		{
			ambtimers[mapnum] = getAmbTimerVal(as, true);
		}
		else if (as->type == AMB_CONTINUOUS)
		{
			ambtimers[mapnum] = 0;
			snd_id |= HSF_REPEAT;
		}
		else if (as->type == AMB_ONCE)
		{
			ambtimers[mapnum] = 0;
			as->as_flags &= ~AS_ACTIVE;
		}
		else
			return;
	}

	if (!as->attenuation)
	{
		snd_id |= HSF_NO_ATTENUATION;
	}

	if (snd_id & HSF_REPEAT)
	{
		if (!(leveltime & 0x7))
			 S_SoundRepeat(snd_id, mo, as->volume, as->attenuation);
	}
	else
		S_StartSoundAtVolume(snd_id, mo, as->volume, as->attenuation);
}

int	S_ExistsAmbient(int num)
{
	return (num >= 0 && num <= MAXAMBIENT && Ambients[num]);
}

// Stop all ambients owned by a mobj referenced with ambient_num
// irrespective of the owner.
void S_KillAmbient(int ambient_num)
{
	ambientsnd_t *as;

	if (   sfx_avail
		&& ambient_num >= 0
		&& ambient_num < MAXAMBIENT
		&& (as = Ambients[ambient_num]) != NULL
		&& as->as_flags & AS_ACTIVE)
	{
		thinker_t	*th = thinkercap.next;
		mobj_t		*mo;

		as->as_flags &= ~AS_ACTIVE;

		// All thinkers must be scanned as there could be
		// multiple instances of a particular ambient
		while(th != &thinkercap)
		{
			if (   th->function == (void (__cdecl *)(void))P_MobjThinker
				&& (mo = (mobj_t*)th)->type == MT_SOUNDSOURCE
				&& TIDNUM(mo) == ambient_num)
			{
				S_StopSound(as->sound_id, as->attenuation ? mo : NULL);
			}
			th = th->next;
		}
	}
}

// R3D HACK for codsound
RDBOOL chkCODambient(int tnum)
{
	static struct codaliassnd_t 
	{
		char	*truename;
		int		type;
		float	atten;
		float	vol;
		float	pmin, pmax;

	} codaliases[] = {
		// This is all guess work as no information seems to exist
		// regarding the COD engine.  Listening to the sounds they
		// all seem to be continuous types. I have upped the atten-
		// uation values (distance over which they are heard) for
		// some.  The quality is not great for DSBUZZ (nasty click)
		// and DSDRIP (a lot of background hiss).

		// NAME     TYPE            ATTN    VOL
		"DSBUZZ",	AMB_CONTINUOUS,	3.0f,	0.3f,	0,	0,	// 4988
		"DSCOMP",	AMB_CONTINUOUS,	2.0f,	0.7f,	0,	0,	// 4989
		"DSDRIP",	AMB_CONTINUOUS,	2.0f,	0.6f,	0,	0,	// 4990
		"DSFIRE",	AMB_CONTINUOUS,	1.5f,	0.8f,	0,	0,	// 4991
		"DSLAVA",	AMB_CONTINUOUS,	1.5f,	0.3f,	0,	0,	// 4992
		"DSRIVER",	AMB_CONTINUOUS,	1.0f,	0.8f,	0,	0,	// 4993
		"DSWFALL",	AMB_CONTINUOUS,	1.7f,	0.8f,	0,	0,	// 4994
		"DSWIND",	AMB_CONTINUOUS,	1.0f,	0.8f,	0,	0,	// 4995
		"DSWIND1",	AMB_CONTINUOUS,	1.0f,	0.8f,	0,	0	// 4996
	};

	int sound_id;
	int num = count_sounds.num;

	char *name;

	if (tnum < 0 || tnum > 8) return false;

	// Has it already been setup?
	if (Ambients[tnum]) return true;

	name = codaliases[tnum].truename;

	if ( (sound_id = R3D_AddNewSound(name)) > 0)
	{
		// Seems ok - set up the sounds
		ambientsnd_t *as = (ambientsnd_t*)W_Calloc(sizeof(ambientsnd_t), STATICID, 144);

		Ambients[tnum] = as;
		memset(as->lumpname, 0, sizeof(as->lumpname));
		strncpy(as->lumpname, name, 8);
		as->sound_id	= sound_id;
		as->sound_id |= HSF_AMBIENT;
		as->attenuation	= codaliases[tnum].atten;
		as->volume		= codaliases[tnum].vol;
		as->type		= codaliases[tnum].type;
		as->periodmin	= (int)(codaliases[tnum].pmin * TICRATE);
		as->periodmax	= (int)(codaliases[tnum].pmax * TICRATE);
		as->as_flags   |= AS_ACTIVE;

		if (count_sounds.num != num)
			Con_Message("Adding COD ambient %s\n",name);
		return true;
	}
	return false;
}

// Save/loadgame as_flags routines
// b211-0 Calling code must free asflags
byte *S_FetchASFLAGS()
{
	int		i;
	byte	*asflags = (byte*)M_Calloc(MAXAMBIENT, 0);

	for (i = 0; i < MAXAMBIENT; i++)
	{
		if (Ambients[i])
			asflags[i] = Ambients[i]->as_flags;
	}
	return asflags;
}

void S_RestoreASFLAGS(byte *asflags)
{
	int i;

	for (i = 0; i < MAXAMBIENT; i++)
	{
		if (Ambients[i])
			Ambients[i]->as_flags = asflags[i];
	}
}

void printoutAmbients()
{
	char *ambtype[] = { "", "POINT", "RANDOM", "PERIODIC", "CONTINUOUS", "WORLD", "ONCE", "" };
	int	 i, num;

	pr_OFinsert("\nAMBIENTS FOUND %i\n", ambientfound);

	for (i = 0; i < MAXAMBIENT; i++)
	{
		if (ambientcount[i] && Ambients[i])
		{
			if (Ambients[i]->type & 0x10) num = 5;
			else
			if (Ambients[i]->type & 0x20) num = 6;
			else
				num = Ambients[i]->type;

			pr_OFinsert("ambient %i count %2i num %i name %8s   type %s\n",
				14001 + i, ambientcount[i], num, Ambients[i]->lumpname, ambtype[num]);
		}
		else if (ambientcount[i])
		{
			pr_OFinsert("ambient %i found in the map but no Ambient defintion has been assigned\n",
				14001 + i);
		}
	}
	pr_OFinsert("\n");
}
