//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: s_wav.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified for R3D
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//
// Based on: Doomsday v1.7.8
//
// DESCRIPTION:  A 'bare necessities' WAV loader
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <windows.h>
#include <mmsystem.h>

#include "de_base.h"
#include "de_console.h"
#include "de_system.h"
#include "m_winmem.h"
#include "sys_sfxd.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

typedef struct riff_hdr_s {
	char	id[4];  			// identifier string = "RIFF"
	uint	len;    			// remaining length after this header
} riff_hdr_t;

typedef struct chunk_hdr_s {	// CHUNK 8-byte header
	char	id[4];				// identifier, e.g. "fmt " or "data"
	uint	len;				// remaining chunk length after header
} chunk_hdr_t;					// data bytes follow chunk header

typedef struct wav_format_s {
	WORD	wFormatTag;			// Format category
	WORD	wChannels;			// Number of channels
	uint	dwSamplesPerSec;    // Sampling rate
	uint	dwAvgBytesPerSec;   // For buffer estimation
	WORD	wBlockAlign;		// Data block size
	WORD	wBitsPerSample;		// Sample size
} wav_format_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

//===========================================================================
// WRead
//===========================================================================
static void WRead(void **ptr, void **dest, int length)
{
	*dest = *ptr;
	*ptr = (char*) *ptr + length;
}

//===========================================================================
// WAV_MemoryLoad
//	If successful the returned samp->sdata contains the wave data else NULL.
//	It is the caller's responsibility to free the buffer (using M_Free).
//  Only single channel WAV files are supported.
//  All parameters must be passed, no NULLs are allowed.
//===========================================================================

void WAV_MemoryLoad(byte *data, byte *end, sfxsample_t *samp)
{
	wav_format_t	*wave_format = NULL;
	chunk_hdr_t		*riff_chunk;
	riff_hdr_t		*riff_header;

	samp->sdata = NULL;

	// Read the RIFF header.
	WRead((void **) &data, (void **) &riff_header, sizeof(*riff_header));

	// Confirm type (should have already been identified before calling)
	if (strncmp((char*)data, "WAVE", 4) || strncmp(riff_header->id, "RIFF", 4))
	{
		Con_Message("WAV_MemoryLoad: Not WAVE data.\n"); return;
	}

	data += 4;

	// Start readin' the chunks, baby!
	while (data < end)
	{
		// Read next chunk header.
		WRead((void **) &data, (void **) &riff_chunk, sizeof(*riff_chunk));
				
		// What have we got here?
		if (!strncmp(riff_chunk->id, "fmt ", 4))
		{
			// Read format chunk.
			WRead((void**)&data, (void**)&wave_format, sizeof(*wave_format));

			// Check that it's a format we know how to read.
			if(wave_format->wFormatTag != WAVE_FORMAT_PCM)
			{
				Con_Message("WAV_MemoryLoad: Unsupported format.\n"); break;
			}

			if (wave_format->wChannels != 1)
			{
				Con_Message("WAV_MemoryLoad: Too many channels (only mono supported).\n"); break;
			}

			// Check bps is valid
			if (   wave_format->wBitsPerSample != 8
				&& wave_format->wBitsPerSample != 16)
			{
				Con_Message("WAV_MemoryLoad: Not a 8/16 bit WAVE.\n"); break;
			}

			// Now we know some information about the sample.
			samp->bytesper = wave_format->wBitsPerSample == 8 ? 1 : 2;
			samp->rate = wave_format->dwSamplesPerSec;

			// If the next two bytes are 0 skip them. This means a cbSize
			// entry of 0 was inserted when the file was created - gmj
			if (!(short)(*data)) data += 2;
		}
		else if (!strncmp(riff_chunk->id, "data", 4))
		{
			if (wave_format && riff_chunk->len > wave_format->wBlockAlign)
			{
				samp->sdata = M_Malloc(riff_chunk->len, 97);
				samp->numsamples = riff_chunk->len / wave_format->wBlockAlign;
				memcpy(samp->sdata, data, riff_chunk->len);
			}
			else
				Con_Message("WAV_MemoryLoad: Malformed WAVE data.\n");

			break;
		}
		else
		{
			// Unknown chunk, just skip it.
			data += riff_chunk->len;
		}
	}
}

