//-----------------------------------------------------------------------------
//
// Copyright (C) 2005 Jaakko Kernen
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: ui_panel.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  modified for use with R3D (cut down mainly)
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
//	The mixer has been abandoned (not supported in Vista) and all playing
//  of music is now handled by FMOD (god bless FMOD).  Thus the master
//  volume setting has been removed and other changes made as required.
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
//  extensively revised
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
//  unsupported code removed. Flash intensity and aspect changes added
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 31 (v220-31) 14/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:  Control Panel (opened with the "panel" command)
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include "de_base.h"
#include "de_console.h"
#include "de_graphics.h"
#include "de_ui.h"
#include "s_main.h"

#include "r3d_compat.h"
#include "r3d_edit.h"

// MACROS ------------------------------------------------------------------

#define NUM_CP_BUTTONS	11
#define NUMITEMS(x)		(sizeof(x)/sizeof(uidata_listitem_t))
#define RES(x, y)		(x | (y<<16))
#define CPG_DISPLAY		2		// the meta value of the display page
#define CPG_VIDEO		3		// the meta value of the video page
#define CPG_GRAPHICS	7		// the meta value of the graphics page
#define CPG_LIGHTING	8		// the meta value of the lighting page
#define HELP_OFFSET		(current_w*290/1000)

// TYPES -------------------------------------------------------------------

typedef struct cvarbutton_s
{
	char	active;
	char	*cvarname;
} 
cvarbutton_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void setHZvalues(int Hz_val, int num);
char *getHZstr(int Hz_val);

extern DGLubyte	*screen_buffer;

// v220-4 vars to allow ui panel to be opened from DD_GameLoop()
//        rather than as soon as the command is actioned.
RDBOOL	open_panel_request;
char	open_panelat[64];

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

float datAlign(float step, float value, int change);

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

void	CP_ClosePanel(ui_object_t *ob);
void	CP_ChooseGroup(ui_object_t *ob);
void	CP_DrawBorder(ui_object_t *ob);
void	CP_DrawBox(ui_object_t *ob);
void	CP_CvarButton(ui_object_t *ob);
void	CP_CvarList(ui_object_t *ob);
void	CP_CvarEdit(ui_object_t *ob);
void	CP_CvarSlider(ui_object_t *ob);
int		CP_KeyGrabResponder(ui_object_t *ob, event_t *ev);
void	CP_KeyGrabDrawer(ui_object_t *ob);
void	CP_DetailReset(ui_object_t *ob);
void	CP_MidiReset(ui_object_t *ob);
void	CP_MidiRebuild(ui_object_t *ob);
void	CP_ToggleList(ui_object_t *ob);
void	CP_ResolutionInfo(ui_object_t *ob);
void	CP_DesktopInfo(ui_object_t *ob);
void	CP_ResolutionList(ui_object_t *ob);
void	CP_SetRes(ui_object_t *ob);

void	UI_ChangeMidiDev(ui_object_t *ob);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern char		Hz_str[3][16];
extern char		aspect_str[8];

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int	panel_32;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static char	panel_buttons[NUM_CP_BUTTONS] = { true }; // The first is active.
static char	panel_sv_password[100];
static int	nMode_ptr;
static ui_object_t *lastResObject;

cvarbutton_t cvarbuttons[] =
{
	{ 0, "scp-usemapmusic" },
	{ 0, "aggressive-flush" },
	{ 0, "con-var-silent" },
	{ 0, "con-dump" },
	{ 0, "con-key-show" },
	{ 0, "input-mouse-y-inverse" }, 
	{ 0, "input-mouse-x-disable" },
	{ 0, "input-mouse-y-disable" },
	{ 0, "input-key-show-scancodes" },
	{ 0, "sound-info" },
	{ 0, "sound-footsteps" },
	{ 0, "sound-weapon" },
	{ 0, "rend-splash" },
	{ 0, "rend-tex-detail" },
	{ 0, "rend-sharp" },
	{ 0, "rend-light" },
	{ 0, "rend-light-torch" },
	{ 0, "rend-light-sine" },
	{ 0, "rend-light-vtx" },
	{ 0, "rend-light-default" },
	{ 0, "rend-addlolight" },
	{ 0, "rend-glow" },
	{ 0, "rend-bloodslime" },
	{ 0, "rend-glow-wall" },
	{ 0, "rend-sky-full" },
	{ 0, "rend-skybox" },
	{ 0, "rend-smoke" },
	{ 0, "rend-shadow" },
	{ 0, "rend-water-trans" },
	{ 0, "toggle_ogg" },
	{ 0, "vid-fps" },
	{ 0, "vid-bpp" },
	{ 0, "vid-vsync" },
	{ 0, "vid-aspect" },
	{ 0, 0 }
};

uidata_listitem_t *lstit_mididev;
uidata_list_t lst_mididev = {
	NULL, 0, 0, "midi-dev"};

uidata_listitem_t lstit_distance[] = {
	{ "1000\tFastest", 0 },
	{ "2000", 1 },
	{ "3000", 2 },
	{ "6000\tDefault", 3 },
	{ "Never Change\tBest Quality", 4 }
};
uidata_list_t lst_distance = { 
	lstit_distance, 5, NUMITEMS(lstit_distance), "rend-model-distance"
};

uidata_listitem_t lstit_con_completion[] = {
	{ "List with values", 0 },
	{ "Cycle through", 1 }
};
uidata_list_t lst_con_completion = { 
	lstit_con_completion, 2, NUMITEMS(lstit_con_completion), "con-completion"
};

// v220-3 window resolutions
uidata_listitem_t *lstit_resolutionW;
uidata_list_t lst_resolutionW;

// v220-3 fullscreen resolutions
uidata_listitem_t *lstit_resolutionF;
uidata_list_t lst_resolutionF;

uidata_slider_t sld_vid_bright		= { -.5f, .7f, 0, .01f, true , "vid-bright" };
uidata_slider_t sld_vid_contrast	= { .5f, 1.5f, 0, .01f, true , "vid-contrast" };
uidata_slider_t sld_vid_gamma		= { .5f, 1.5f, 0, .01f, true , "vid-gamma" };

uidata_slider_t sld_vid_cap			= { 4, 11, 0, 1, false , "vid-cap" };

uidata_slider_t sld_flash_damage	= {   0, 1.0f, 0, .05f, true, "flash-damage" };
uidata_slider_t sld_flash_pickup	= {   0, 1.0f, 0, .05f, true, "flash-pickup" };

uidata_slider_t sld_vid_compress	= {   0,  140, 0,    1, false, "vid-compress" };
uidata_slider_t sld_vid_darken		= {1024, 8192, 0,  512, true , "vid-distance-darken" };

uidata_slider_t sld_sound_volume	= { 0, 255, 0, 1, false, "vol-sfx" };
uidata_slider_t sld_step_volume		= { 0.1f, 1.f, 0.1f, 0.1f, true, "vol-step" };
uidata_slider_t sld_splash_volume	= { 0.1f, 1.f, 0.1f, 0.1f, true, "vol-splash" };
uidata_slider_t sld_ext_volume		= { 0, 255, 0, 1, false, "vol-ext" };
uidata_slider_t sld_midi_volume		= { 0, 255, 0, 1, false, "vol-midi" };

uidata_slider_t sld_musicpos		= { 0, 0, 0, .1f, true, "pos_slider" };
uidata_slider_t sld_mus_compress	= { .5f, 2.f, 0, .1f, true, "mus-compress" };
uidata_slider_t sld_reverb_volume	= { 0, 1, 0, .1f, true , "sound-reverb-volume" };

uidata_slider_t sld_playerdrag		= { 0, 10, 0, 1, false, "player-drag", "Doom default" };
uidata_slider_t sld_keywait1		= { 5, 35, 0, 1, false, "input-key-wait1" };
uidata_slider_t sld_keywait2		= { 1, 10, 0, 1, false, "input-key-wait2" };

uidata_slider_t sld_opacity			= { 150, 255, 0, 1, false, "rend-opacity" };
uidata_slider_t sld_shadow_dark		= { 0, 1, 0, .1f, true, "rend-shadow-darkness" };
uidata_slider_t sld_base_fov		= { 80, 110,  0, 1, false, "rend-base-fov" };
uidata_slider_t sld_shadow_far		= { 0, 3000, 0, 50, false, "rend-shadow-far" };
uidata_slider_t sld_shadow_radmax	= { 0, 128, 0, 1, false, "rend-shadow-radius-max" };
uidata_slider_t sld_light_bright	= { .5f, .9f, 0, .05f, true, "rend-light-bright" };
uidata_slider_t sld_light_ambient	= { 0, 255, 0, 1, false, "rend-light-ambient" };

uidata_slider_t sld_anisotropic		= { 1, 16, 0, 1, false, "rend-anisotropic" };
uidata_slider_t sld_detail_scale	= { .1f, 32, 0, .1f, true, "rend-tex-detail-scale" };
uidata_slider_t sld_detail_strength	= { 0, 2, 0, .01f, true, "rend-tex-detail-strength" };
uidata_slider_t sld_detail_far		= { 1, 1000, 0, 10, true, "rend-tex-detail-far" };
uidata_slider_t sld_model_lights	= { 0, 10, 0, 1, false, "rend-model-lights" };

uidata_slider_t sld_particle_rate	= { .5f, 2, 0, .1f, true, "rend-particle-rate" };
uidata_slider_t sld_particle_diffuse = { 1, 4, 0, .1f, true, "rend-particle-diffuse" };

uidata_slider_t sld_hud_size  = {  1,   5, 1, 1, false, "hud-size" };
uidata_slider_t sld_con_alpha = {  0, 100, 0, 5, false, "con-alpha" };
uidata_slider_t sld_con_light = {  0, 100, 0, 5, false, "con-light" };
uidata_slider_t sld_txt_delay = { 1.0f, 3.0f, 0, .1f, true, "scp-txtdelay", "MIN" };
// v220-33
uidata_slider_t sld_zbias = { .0001f, .0022f, .0001f, .0001f, true, "set-zbias" };

ui_page_t	page_panel;

ui_object_t ob_panel[] =
{
	{ UI_BUTTON2,	1,	UIF_LEFT_ALIGN,	  0,  55, 250, 60,	"Display",		UIButton_Drawer, UIButton_Responder, 0,	CP_ChooseGroup, &panel_buttons[0] },
	{ UI_BUTTON2,	1,	UIF_LEFT_ALIGN,	  0, 130, 250, 60,	"Video",		UIButton_Drawer, UIButton_Responder, 0,	CP_ChooseGroup, &panel_buttons[1] },
	{ UI_BUTTON2,	1,	UIF_LEFT_ALIGN,	  0, 205, 250, 60,	"Stress Settings",
																			UIButton_Drawer, UIButton_Responder, 0,	CP_ChooseGroup, &panel_buttons[2] },
	{ UI_BUTTON2,	1,	UIF_LEFT_ALIGN,	  0, 280, 250, 60,	"Audio",		UIButton_Drawer, UIButton_Responder, 0,	CP_ChooseGroup, &panel_buttons[3] },
	{ UI_BUTTON2,	1,	UIF_LEFT_ALIGN,	  0, 355, 250, 60,	"Input",		UIButton_Drawer, UIButton_Responder, 0,	CP_ChooseGroup, &panel_buttons[4] },
	{ UI_BUTTON2,	1,	UIF_LEFT_ALIGN,	  0, 455, 250, 60,	"Graphics",		UIButton_Drawer, UIButton_Responder, 0,	CP_ChooseGroup, &panel_buttons[5] },
	{ UI_BUTTON2,	1,	UIF_LEFT_ALIGN,	 30, 530, 220, 60,	"Lighting",		UIButton_Drawer, UIButton_Responder, 0,	CP_ChooseGroup, &panel_buttons[6] },
	{ UI_BUTTON2,	1,	UIF_LEFT_ALIGN,	 30, 605, 220, 60,	"Textures",		UIButton_Drawer, UIButton_Responder, 0,	CP_ChooseGroup, &panel_buttons[7] },
	{ UI_BUTTON2,	1,	UIF_LEFT_ALIGN,	  0, 705, 250, 60,	"HUD-Scripted Map",
																			UIButton_Drawer, UIButton_Responder, 0,	CP_ChooseGroup, &panel_buttons[8] },
	{ UI_BUTTON2,	1,	UIF_LEFT_ALIGN,	  0, 780, 250, 60,	"Console",		UIButton_Drawer, UIButton_Responder, 0,	CP_ChooseGroup, &panel_buttons[9] },
	{ UI_BUTTON,	0,	0,				  0, 940, 250, 60,	"Close Panel (Esc)",
																			UIButton_Drawer, UIButton_Responder, 0,	CP_ClosePanel },
	{ UI_BOX,		0,	0,				280,  55, 720,945,	"",			CP_DrawBorder },

// Display
	{ UI_META,		2 },
	{ UI_TEXT,		0,	0,				280,   0,   0, 50,	"DISPLAY ADJUSTMENTS - Try first by altering just the gamma", UIText_Drawer },
	{ UI_TEXT,		0,	0,				300,  70,   0, 55,	"Gamma correction", UIText_Drawer },
	{ UI_TEXT,		0,	0,				510,  70,   0, 55,	"[1 - optimum]", UIText_Drawer },
	{ UI_SLIDER,	0,	0,				680,  70, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_vid_gamma },
	{ UI_TEXT,		0,	0,				300, 140,   0, 55,	"Display contrast", UIText_Drawer },
	{ UI_TEXT,		0,	0,				510, 140,   0, 55,	"[1 - optimum]", UIText_Drawer },
	{ UI_SLIDER,	0,	0,				680, 140, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_vid_contrast },
	{ UI_TEXT,		0,	0,				300, 210,   0, 55,	"Display brightness", UIText_Drawer },
	{ UI_TEXT,		0,	0,				510, 210,   0, 55,	"[0 - optimum]", UIText_Drawer },
	{ UI_SLIDER,	0,	0,				680, 210, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_vid_bright },
	{ UI_SCREEN,	0,	0,				290, 285, 700,700,	"",	UI_DrawScreen },

// Video
	{ UI_META,		3 },
	{ UI_TEXT,		0,	UIEF_SETVSYNC,	280,   0,   0, 50,	"VIDEO OPTIONS", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_CHKVSYNC,	300,  65,   0, 55,	"Enable vertical sync", UIText_Drawer },
	{ UI_BUTTON2,	0,	UIEF_HASVSYNC,	520,  65,  70, 55,	"vid-vsync", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },
	{ UI_TEXT,		0,	UIEF_VIDCAP,	640,  65,   0, 55,	"Cap", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_VIDCAP,	700,  65, 200, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_vid_cap },
	{ UI_TEXT,		0,	0,				300, 130,   0, 55,	"Display FPS counter", UIText_Drawer },
	{ UI_BUTTON2,	0,	0,				520, 130,  70, 55,	"vid-fps",	UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },

	{ UI_BOX,		0,	0,				300, 180,   0, 55,	"", CP_DesktopInfo },
	{ UI_TEXT,		0,	0,				300, 210,   0, 55,	"The window list has already been set to match Hz.", UIText_Drawer },
	{ UI_TEXT,		0,	0,				300, 240,   0, 55,	"For a window: bpp, below, must match the desktop bpp", UIText_Drawer },
	{ UI_BOX,		0,	UIF_DEFSCREEN,	300, 270,   0, 55,	"", CP_ResolutionInfo },
	{ UI_BOX,		0,	0,				300, 300,   0, 55,	"",	CP_ResolutionInfo },

	{ UI_TEXT,		0,	0,				360, 350,   0, 50,	"WINDOW", UIText_Drawer },
	{ UI_TEXT,		0,	0,				670, 350,   0, 50,	"FULLSCREEN", UIText_Drawer },
	{ UI_BUTTON2,	0,	UIEF_ASPECT|UIF_ASPECT,
										880, 350,  70, 55,	"vid-aspect", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },

	{ UI_LIST,		0,	UIF_RES_LISTW,	340, 400, 300,350,	"",	UIList_Drawer, UIList_Responder, UIList_Ticker, CP_ResolutionList, &lst_resolutionW },
	{ UI_LIST,		0,	UIF_RES_LISTF,	650, 400, 300,350,	"",	UIList_Drawer, UIList_Responder, UIList_Ticker, CP_ResolutionList, &lst_resolutionF },

	{ UI_BUTTON2,	0,	UIF_BPP,		580, 780, 140, 55,	"vid-bpp", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },
	{ UI_BUTTON,	0,	UIF_SET_RES,	340, 860, 610, 60,	"",	UIButton_Drawer, UIButton_Responder, 0, CP_SetRes },
	{ UI_BUTTON,	0,	UIF_SET_DEF,	340, 930, 300, 60,	"Set as the default", UIButton_Drawer, UIButton_Responder, 0, CP_SetRes },
	{ UI_BUTTON,	0,	UIF_RESET_DEF,	650, 930, 300, 60,	"Reset to default", UIButton_Drawer, UIButton_Responder, 0, CP_SetRes },

// Stress Settings
	{ UI_META,		4 },
	{ UI_TEXT,		0,	0,				280,   0,   0, 50,	"STRESS OPTIONS: Adjust for quality/speed", UIText_Drawer },
	{ UI_TEXT,		0,	0,				300,  50,   0, 60,	"FIELD OF VIEW  Note: values above 90 will impact frame rates", UIText_Drawer },
	{ UI_TEXT,		0,	0,				300, 100,   0, 55,	"Setting for 4:3 ratio - auto adjusted for wider aspect ratio", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_DEFFOV,	300, 180,   0, 55,	"", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_SETFOV,	680, 180, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_base_fov },

	{ UI_TEXT,		0,	UIEF_TXTMODVIS,	300, 270,   0, 55,	"NOTE: All decoration models have a lo-res alternative", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_MODVIS,	300, 300,   0, 55,	"that automatically gets substituted at a distance of 1000.", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_MODVIS,	300, 330,   0, 55,	"Where no lo-res type is currently available (all monster", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_MODVIS,	300, 360,   0, 55,	"models for example) then the list can be used to change", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_MODVIS,	300, 390,   0, 55,	"to its sprite at the selected distance instead:", UIText_Drawer },
	{ UI_LIST,		0,	UIEF_MODVIS,	300, 450, 360,350,	"",	UIList_Drawer, UIList_Responder, UIList_Ticker, CP_CvarList, &lst_distance },

	{ UI_TEXT,		0,	UIEF_PARTICLES,	300, 700,   0, 60,	"PARTICLES - Set to defaults for optimum results", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_PARTICLES,	300, 770,   0, 55,	"Spawn rate factor", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_PARTICLES,	540, 770,   0, 55,	"[default 1]", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_PARTICLES,	680, 770, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_particle_rate },
	{ UI_TEXT,		0,	UIEF_PARTICLES,	300, 840,   0, 55,	"Near diffusion factor", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_PARTICLES,	540, 840,   0, 55,	"[default 3]", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_PARTICLES,	680, 840, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_particle_diffuse },

// Audio
	{ UI_META,		5 },
	{ UI_TEXT,		0,	0,				280,   0,   0, 50,	"AUDIO OPTIONS", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_TXTSOUND,	300,  70,   0, 55,	"Sound effects volume", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_SOUND,		680,  70, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_sound_volume },
	{ UI_TEXT,		0,	UIEF_SOUND,		300, 130,   0, 55,	"General reverb strength", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_SOUND,		680, 130, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_reverb_volume },

	{ UI_TEXT,		0,	UIEF_SOUND,		300, 192,   0, 50,	"Show status of channels", UIText_Drawer },
	{ UI_BUTTON2,	0,	UIEF_SOUND,		600, 192,  70, 50,	"sound-info", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },

	{ UI_TEXT,		0,	UIEF_TXTFOOTSTEP, 300,252,  0, 50,	"Footstep sounds", UIText_Drawer },
	{ UI_BUTTON2,	0,	UIEF_FOOTSTEP,	600, 252,  70, 80,	"sound-footsteps", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },
	{ UI_TEXT,		0,	UIEF_SPLASH,	680, 246,   0, 55,	"step vol", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_SPLASH,	800, 252, 180, 40,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_step_volume },
	{ UI_TEXT,		0,	UIEF_SPLASH,	680, 290,   0, 55,	"splash vol", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_SPLASH,	800, 300, 180, 40,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_splash_volume },

	{ UI_TEXT,		0,	UIEF_TXTWEAPSND,300, 352,   0, 50,	"Weapon ready sounds", UIText_Drawer },
	{ UI_BUTTON2,	0,	UIEF_WEAPSND,	600, 352,  70, 50,	"sound-weapon", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },

	{ UI_TEXT,		0,	UIEF_FMODMUSIC,	300, 412,   0, 55,	"FMOD music volume", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_FMODMUSIC,	680, 412, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_ext_volume },
	{ UI_TEXT,		0,	UIEF_MIDMUSIC,	300, 412,   0, 55,	"WinMM MIDI volume", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_MIDMUSIC,	680, 412, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_midi_volume },

	{ UI_TEXT,		0,	UIEF_TXTMUSIC,	300, 472,   0, 55,	"", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_SLDMUSIC|UIF_NOVALUES,	300, 520, 680, 30,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_musicpos },
	{ UI_TEXT,		0,	UIEF_MUSIC,		300, 570,   0, 50,	"", UIText_Drawer, 0,  0, CP_ToggleList},	// text string set elsewhere
	{ UI_BUTTON,	0,	UIEF_TOGGLE,	860, 570, 110, 50,	"", UIButton_Drawer, UIButton_Responder, 0, CP_ToggleList },
	{ UI_LIST,		0,	UIEF_MIDIDEV,	300, 630, 360,350,	"",	UIList_Drawer, UIList_Responder, UIList_Ticker, CP_CvarList, &lst_mididev },
	{ UI_BUTTON,	0,	UIEF_MIDIDEV,	700, 630, 110, 50,	"CHANGE", UIButton_Drawer, UIButton_Responder, 0, UI_ChangeMidiDev },

	{ UI_BUTTON,	0,	UIEF_MIDIRESET,	700, 730, 220, 55,	"GSReset [Roland]", UIButton_Drawer, UIButton_Responder, 0, CP_MidiReset },

	{ UI_TEXT,		0,	UIEF_MIDI,		490, 800,   0, 50,	"---- MUS to MIDI OPTION ----", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_MIDI,		300, 860,   0, 55,	"MUS vol compress (1=No Change)", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_MIDI,		680, 860, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_mus_compress },
	{ UI_TEXT,		0,	UIEF_MIDI,		300, 925,   0, 55,	"Rebuild Mus2Midi", UIText_Drawer },
	{ UI_BUTTON,	0,	UIEF_MIDI,		680, 925, 110, 55,	"Rebuild", UIButton_Drawer, UIButton_Responder, 0, CP_MidiRebuild },

// Input options
	{ UI_META,		6 },
	{ UI_TEXT,		0,	0,				280,   0,   0, 50,	"INPUT OPTIONS", UIText_Drawer },
	{ UI_TEXT,		0,	0,				300,  70,   0, 55,	"Player deceleration", UIText_Drawer },
	{ UI_SLIDER,	0,	0,				680,  70, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_playerdrag },
	{ UI_TEXT,		0,	UIEF_TXTMOUSE,	300, 160,   0, 55,	"Invert mouse Y axis", UIText_Drawer },
	{ UI_BUTTON2,	0,	UIEF_MOUSE,		680, 160,  70, 55,	"input-mouse-y-inverse", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },
	{ UI_TEXT,		0,	UIEF_MOUSE,		300, 220,   0, 55,	"Disable mouse X axis", UIText_Drawer },
	{ UI_BUTTON2,	0,	UIEF_MOUSE,		680, 220,  70, 55,	"input-mouse-x-disable", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },
	{ UI_TEXT,		0,	UIEF_MOUSE,		300, 280,   0, 55,	"Disable mouse Y axis", UIText_Drawer },
	{ UI_BUTTON2,	0,	UIEF_MOUSE,		680, 280,  70, 55,	"input-mouse-y-disable", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },
	{ UI_TEXT,		0,	0,				300, 660,   0, 55,	"Key repeat delay", UIText_Drawer },
	{ UI_TEXT,		0,	0,				540, 660,   0, 55,	"[default 15]", UIText_Drawer },
	{ UI_SLIDER,	0,	0,				680, 660, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_keywait1 },
	{ UI_TEXT,		0,	0,				300, 730,   0, 55,	"Key repeat rate", UIText_Drawer },
	{ UI_TEXT,		0,	0,				540, 730,   0, 55,	"[default 4]", UIText_Drawer },
	{ UI_SLIDER,	0,	0,				680, 730, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_keywait2 },
	{ UI_TEXT,		0,	0,				300, 880,   0, 55,	"Print scancodes of pressed keys", UIText_Drawer },
	{ UI_BUTTON2,	0,	0,				680, 880,  70, 55,	"input-key-show-scancodes", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },

// Graphics options
	{ UI_META,		7 },
	{ UI_TEXT,		0,	0,				280,  0,   0, 50,	"MAIN GRAPHICS OPTIONS", UIText_Drawer },

	{ UI_TEXT,		0,	UIEF_SCPNOSKY,	300, 100,  0, 55,	"Use skybox", UIText_Drawer },
	{ UI_BUTTON2,	0,	UIEF_NOSKYBOX,	680, 100, 70, 55,	"rend-skybox", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },
	{ UI_TEXT,		0,	UIEF_TXTSPLASH, 300, 160,   0, 55,	"Render bullet splashes", UIText_Drawer },
	{ UI_BUTTON2,	0,	UIEF_SPLASH,	680, 160,  70, 55,	"rend-splash", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },

	// $translucent_water
	{ UI_TEXT,		0,	UIEF_TXTTRANSLUC,300,220,   0, 55,	"Translucent water", UIText_Drawer },
	{ UI_BUTTON2,	0,	UIEF_TXTTRANS2,	680, 220,  70, 55,	"rend-water-trans", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },
	{ UI_TEXT,		0,	UIEF_TRANSONLY,	300, 280,   0, 55,	"Water opacity", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_TRANSONLY,	515, 280,   0, 55,	"[default 170]", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_TRANSONLY,	680, 280, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_opacity },

	{ UI_BUTTON2,	0,	UIEF_SMOKE,		680, 450,  70, 55,	"rend-smoke", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },

	{ UI_TEXT,		0,	UIEF_TXTSHADOW,	300, 510,   0, 55,	"Solid objects cast shadows", UIText_Drawer },
	{ UI_BUTTON2,	0,	UIEF_SHADOW,	680, 510,  70, 55,	"rend-shadow", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },

	{ UI_TEXT,		0,	UIEF_OBSHADOW,	300, 570,   0, 55,	"#darkness factor", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_OBSHADOW,	515, 570,   0, 55,	"[def 0.50]", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_OBSHADOW,	680, 570, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_shadow_dark },
	{ UI_TEXT,		0,	UIEF_OBSHADOW,	300, 630,   0, 55,	"#visible distance", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_OBSHADOW,	515, 630,   0, 55,	"[def 1000]", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_OBSHADOW,	680, 630, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_shadow_far },
	{ UI_TEXT,		0,	UIEF_OBSHADOW,	300, 690,   0, 55,	"#shadow radius", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_OBSHADOW,	515, 690,   0, 55,	"[def 80]", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_OBSHADOW,	680, 690, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_shadow_radmax },

	{ UI_TEXT,		0,	0,				300, 800,   0, 55,	"Set zbias", UIText_Drawer },
	{ UI_TEXT,		0,	0,				515, 800,   0, 55,	"[def 0.0010]", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_ZBIAS,		680, 800, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_zbias },
	{ UI_TEXT,		0,	0,				300, 880,   0, 55,	"Aggressive flush (e.g. needed by NVidia GTX1080)", UIText_Drawer },
	{ UI_BUTTON2,	0,	0,				880, 880,  70, 50,	"aggressive-flush", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },

// Lighting	modified v220-31/v220-33/v220-35
	{ UI_META,		8 },
	{ UI_TEXT,		0,	0,				280,   0,   0, 50,	"Graphics Options: Lighting", UIText_Drawer },

	{ UI_TEXT,		0,	0,				300,  70,   0, 55,	"Damage flash level", UIText_Drawer },
	{ UI_TEXT,		0,	0,				520,  70,   0, 55,	"[1=Doom]", UIText_Drawer },
	{ UI_SLIDER,	0,	0,				680,  70, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_flash_damage },
	{ UI_TEXT,		0,	0,				300, 134,   0, 55,	"Pickup flash level", UIText_Drawer },
	{ UI_TEXT,		0,	0,				520, 134,   0, 55,	"[1=Doom]", UIText_Drawer },
	{ UI_SLIDER,	0,	0,				680, 134, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_flash_pickup },
	{ UI_TEXT,		0,	0,				300, 198,   0, 55,	"Light Compress", UIText_Drawer },
	{ UI_TEXT,		0,	0,				520, 198,   0, 55,	"[Optimum 0]", UIText_Drawer },
	{ UI_SLIDER,	0,	0,				680, 198, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_vid_compress },
	{ UI_TEXT,		0,	UIEF_TXTDARKEN,	300, 262,   0, 55,	"Darken limit", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_DARKEN,	520, 262,   0, 55,	"[Default 3072]", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_DARKEN|UIF_INT,680,262,300,55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_vid_darken },

	{ UI_BOX,		0,	0,				290, 340, 700,220,	"", CP_DrawBox },
	{ UI_TEXT,		0,	UIEF_TXTLTORCH,	300, 350,   0, 55,	"Torch light (no vertex)", UIText_Drawer },
	{ UI_BUTTON2,	0,	UIEF_LTORCH,	750, 350,  70, 45,	"rend-light-torch", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },
	{ UI_TEXT,		0,	UIEF_LVTX,		300, 400,   0, 55,	"Vertex distance (no torch)", UIText_Drawer },
	{ UI_BUTTON2,	0,	UIEF_LVTX,		750, 400,  70, 45,	"rend-light-vtx", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },
	{ UI_TEXT,		0,	UIEF_LSINE,		300, 450,   0, 55,	"Non-Linear", UIText_Drawer },
	{ UI_BUTTON2,	0,	UIEF_LSINE,		750, 450,  70, 45,	"rend-light-sine", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },
	{ UI_TEXT,		0,	UIEF_TXTLDEFS,	300, 500,   0, 55,	"On exit use these as defaults for all maps", UIText_Drawer },
	{ UI_BUTTON2,	0,	UIEF_LDEFS,		780, 500,  40, 45,	"rend-light-default", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },

	{ UI_TEXT,		0,	UIEF_TXTSETGLOW,300, 570,   0, 55,	"Enable glow flats", UIText_Drawer },
	{ UI_BUTTON2,	0,	UIEF_SETGLOW,	500, 570,  70, 45,	"rend-glow", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },
	{ UI_TEXT,		0,	UIEF_EDITINH,	600, 570,   0, 55,	"+Blood-Slime-Water", UIText_Drawer },
	{ UI_BUTTON2,	0,	UIEF_EDITINH,	860, 570,  70, 45,	"rend-bloodslime", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },
	{ UI_TEXT,		0,	UIEF_EDITINH,	300, 620,   0, 55,	"Floor/ceiling glow on walls", UIText_Drawer },
	{ UI_BUTTON2,	0,	UIEF_EDITINH,	680, 620,  70, 45,	"rend-glow-wall", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },
	{ UI_TEXT,		0,	0,				300, 690,   0, 55,	"Brighten low lighting", UIText_Drawer },
	{ UI_BUTTON2,	0,	0,				680, 690,  70, 45,	"rend-addlolight", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },

	{ UI_TEXT,		0,	UIEF_TXTMODONLY,300, 766,   0, 55,	"Enable dynamic lights", UIText_Drawer },
	{ UI_BUTTON2,	0,	UIEF_MODONLY,	680, 766,  70, 45,	"rend-light", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },
	{ UI_TEXT,		0,	UIEF_TXTLIGHTS,	300, 816,   0, 55,	"Max model dyn lights", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_DLIGHTS,	520, 816,   0, 55,	"[Default 6]", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_DLIGHTS,	680, 816, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_model_lights },

	{ UI_TEXT,		0,	UIEF_DLITLEVEL,	300, 880,   0, 55,	"Dyn light level", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_DLITLEVEL,	520, 880,   0, 55,	"[Default 0.7]", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_DLITLEVEL,	680, 880, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_light_bright },

	{ UI_TEXT,		0,	UIEF_TXTAMBIENT,300, 950,   0, 55,	"Ambient light level edit mode", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_AMBIENT,	680, 940, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_light_ambient },

// Textures
	{ UI_META,		9 },
	{ UI_TEXT,		0,	0,				280,   0,   0, 50,	"Graphics Options: Textures", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_NOANISO,	300, 220,   0, 55,	"Anisotropic filtering", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_NOANISO,	510, 220, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_anisotropic },
	{ UI_TEXT,		0,	UIEF_CHANGE,	300, 290,   0, 60,	"CHANGES REQUIRE TEXTURE RELOADING.", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_CHANGE,	300, 330,   0, 60,	"This will be done automatically after closing the panel", UIText_Drawer },
	{ UI_TEXT,		0,	0,				300, 480,	0, 55,	"Sharpen lo-res sprites?", UIText_Drawer },
	{ UI_BUTTON2,	0,	0,				600, 480,  70, 55,	"rend-sharp", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },
	{ UI_TEXT,		0,	0,				400, 620,   0, 55,	"DETAIL TEXTURES - Note: Only used with lo-res textures", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_TXTNODETAIL,300,670,   0, 55,	"Enable detail textures", UIText_Drawer },
	{ UI_BUTTON2,	0,	UIEF_TXTNODET2,	600, 670,  70, 55,	"rend-tex-detail", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },
	{ UI_TEXT,		0,	UIEF_NODETAIL,	300, 730,   0, 55,	"Detail texture scaling factor", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_NODETAIL,	600, 730, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_detail_scale },
	{ UI_TEXT,		0,	UIEF_NODETAIL,	300, 790,   0, 55,	"Detail texture contrast", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_NODETAIL,	600, 790, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_detail_strength },
	{ UI_TEXT,		0,	UIEF_NODETAIL,	300, 850,   0, 55,	"Detail texture max distance", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_NODETAIL,	600, 850, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_detail_far },
	{ UI_TEXT,		0,	UIEF_NODETAIL,	300, 920,   0, 60,	"Reset detail defaults", UIText_Drawer },
	{ UI_BUTTON,	0,	UIEF_NODETAIL,	600, 920, 110, 60,	"Reset", UIButton_Drawer, UIButton_Responder, 0, CP_DetailReset },

// Scripted map and hud options
	{ UI_META,		10 },
	{ UI_TEXT,		0,	0,				300,  70,   0, 50,	"SCRIPTED MAP OPTIONS", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_TXTSCPALLOW,300,140,   0, 55,	"CURRENT MAP IS NOT SCRIPTED - OPTIONS UNAVAILABLE", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_SCPALLOW,	300, 140,   0, 55,	"Text fade delay factor", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_SCPALLOW,	540, 140,   0, 55,	"[default 1.5]", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_SCPALLOW,	680, 140, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_txt_delay },
	{ UI_TEXT,		0,	UIEF_SCPALLOW,	300, 210,   0, 55,	"SCP music overrides list", UIText_Drawer },
	{ UI_BUTTON2,	0,	UIEF_SCPALLOW,	680, 210,  70, 55,	"scp-usemapmusic", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },
	{ UI_TEXT,		0,	0,				300, 410,   0, 50,	"HUD OPTIONS [default 1 - Risen3D]", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_TXTHUDSIZE,300, 470,   0, 55,	"HUD type", UIText_Drawer },
	{ UI_TEXT,		0,	UIEF_TXTHUDTYPE,440, 470,   0, 55,	"", UIText_Drawer },
	{ UI_SLIDER,	0,	UIEF_SCPINH,	680, 470, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_hud_size },

// Console options
	{ UI_META,		11 },
	{ UI_TEXT,		0,	0,				280,   0,   0, 55,	"CONSOLE OPTIONS", UIText_Drawer },
	{ UI_TEXT,		0,	0,				300,  80,   0, 55,	"Background opacity", UIText_Drawer },	// was translucency
	{ UI_TEXT,		0,	0,				540,  80,   0, 55,	"[default 75]", UIText_Drawer },
	{ UI_SLIDER,	0,	0,				680,  80, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_con_alpha },
	{ UI_TEXT,		0,	0,				300, 160,   0, 55,	"Background light", UIText_Drawer },
	{ UI_TEXT,		0,	0,				540, 160,   0, 55,	"[default 50]", UIText_Drawer },
	{ UI_SLIDER,	0,	0,				680, 160, 300, 55,	"",	UISlider_Drawer, UISlider_Responder, UISlider_Ticker, CP_CvarSlider, &sld_con_light },
	{ UI_TEXT,		0,	0,				300, 310,   0, 55,	"Command completion with Tab", UIText_Drawer },
	{ UI_LIST,		0,	UIF_LIST,		680, 310, 300,120,	"",	UIList_Drawer, UIList_Responder, UIList_Ticker, CP_CvarList, &lst_con_completion },
	{ UI_TEXT,		0,	0,				300, 500,   0, 55,	"Silent console variables", UIText_Drawer },
	{ UI_BUTTON2,	0,	0,				680, 500,  70, 55,	"con-var-silent", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },
	{ UI_TEXT,		0,	0,				300, 580,   0, 55,	"Dump messages to Risen3D.out", UIText_Drawer },
	{ UI_BUTTON2,	0,	0,				680, 580,  70, 55,	"con-dump", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },
	{ UI_TEXT,		0,	0,				300, 700,   0, 55,	"CONSOLE ACTIVATION KEY", UIText_Drawer },
	{ UI_TEXT,		0,	0,				300, 780,   0, 40,	"Focus the box, press a key.", UIText_Drawer },
	{ UI_FOCUSBOX,	0,	0,				680, 780,  70, 55,	"con-key-activate",	CP_KeyGrabDrawer, CP_KeyGrabResponder },
#ifdef _DEBUG
	{ UI_TEXT,		0,	0,				300, 860,   0, 55,	"Show ASCII codes of pressed keys", UIText_Drawer },
	{ UI_BUTTON2,	0,	0,				680, 860,  70, 55,	"con-key-show", UIButton_Drawer, UIButton_Responder, 0, CP_CvarButton },
#endif

	{ UI_NONE }
};

// CODE --------------------------------------------------------------------

//===========================================================================
// CP_ClosePanel
//===========================================================================
void CP_ClosePanel(ui_object_t *ob)
{
	UI_End();
}

//===========================================================================
// CP_ChooseGroup
//===========================================================================
void CP_ChooseGroup(ui_object_t *ob)
{
	int i;

	memset(panel_buttons, 0, sizeof(panel_buttons));
	UI_FlagGroup(ob_panel, 1, UIF_ACTIVE, UIFG_CLEAR);
	*(char*) ob->data = true;
	ob->flags |= UIF_ACTIVE;
	// Hide/show the option controls.
	for(i = 0; i < NUM_CP_BUTTONS; i++)
		UI_FlagGroup(ob_panel, 2 + i, UIF_HIDDEN, !panel_buttons[i]);
}

//===========================================================================
// CP_DrawBorder
//===========================================================================
void CP_DrawBorder(ui_object_t *ob)
{
	int b = UI_BORDER/2;

	UI_DrawRect(ob->x, ob->y, ob->w, ob->h, b, 
		UI_COL(UIC_BRD_HI), UI_COL(UIC_BRD_MED), UI_COL(UIC_BRD_LOW));
	UI_Gradient(ob->x + b, ob->y + b, ob->w - 2*b, ob->h - 2*b,
		UI_COL(UIC_BG_MEDIUM), 0, .6f, 0);
}

void CP_DrawBox(ui_object_t *ob)
{
	int b = UI_BORDER/2;

	UI_DrawRect(ob->x, ob->y, ob->w, ob->h, b, 
		UI_COL(UIC_BRD_WHI), UI_COL(UIC_BRD_WMED), UI_COL(UIC_BRD_WLOW));
}

static void UI_checkLightingButton(uint flags)
{
	ui_object_t *list = ob_panel;
	ui_object_t *obf = NULL;

	for(; list->type; list++)
	{
		if (   list->group == CPG_LIGHTING
			&& list->responder == UIButton_Responder
			&& (list->flags & 0xff) == flags)
		{
			obf = list; break;
		}
	}

	if (obf)
	{
		cvarbutton_t *cb = NULL;

		if (!obf->data)
		{
			// Find the cvarbutton representing this one.
			for (int i = 0; cvarbuttons[i].cvarname; i++)
			{
				if (!strcmp(obf->text, cvarbuttons[i].cvarname))
				{
					obf->data = &cvarbuttons[i];
					break;
				}
			}
		}

		cb = (cvarbutton_t *)obf->data;
		cb->active = (byte)(Con_GetByte(cb->cvarname) != 0);
		cvar_t *var = Con_GetVariable(cb->cvarname);

		if (var && cb->active)
		{
			obf->flags &= ~UIF_ACTIVE;
			strcpy(obf->text, "No");
			cb->active = 0;
			*(char*)var->ptr = 0;
		}
	}
}

//===========================================================================
// CP_CvarButton
//===========================================================================
void CP_CvarButton(ui_object_t *ob)
{
	cvarbutton_t *cb = (cvarbutton_t *)ob->data;
	cvar_t *var = Con_GetVariable(cb->cvarname);

	if (!var) return;

	if (ob->flags & UIF_ASPECT)
	{
		strcpy(ob->text, cb->active ? "4:3" : aspect_str);
	}
	else
	if (ob->flags & UIF_BPP)
	{
		strcpy(ob->text, cb->active ? "32 bpp" : "16 bpp");
		cMode.t_bpp = cb->active ? 32 : 16;
		panel_32 = cb->active;

		ob = UI_FindObject(ob_panel, CPG_VIDEO, UIF_RES_LISTW);

		if (cMode.w_initialbpp == cMode.t_bpp)
			ob->flags &= ~UIF_DISABLED;
		else
			ob->flags |= UIF_DISABLED;

		if (lastResObject) CP_ResolutionList(lastResObject);
	}
	else
	{
		strcpy(ob->text, cb->active ? "Yes" : "No");
	}

	// Change state.
	*(char*)var->ptr = cb->active;

	if (cb->active)
	{
		if ((ob->flags & 0xFF) == UIEF_LTORCH)
		{
			UI_checkLightingButton(UIEF_LVTX);
		}
		else if ((ob->flags & 0xFF) ==  UIEF_LVTX)
		{
			UI_checkLightingButton(UIEF_LTORCH);
		}
	}
}

//===========================================================================
// CP_CvarList
//===========================================================================
void CP_CvarList(ui_object_t *ob)
{
	uidata_list_t	*list = (uidata_list_t*)ob->data;
	cvar_t			*var  = Con_GetVariable((char*)list->data);
	int				value = ((uidata_listitem_t*)list->items)[list->selection].data;

	if (UI_CheckListOb(ob)) return;

	if(list->selection < 0) return; // Hmm?

	if (   !strcmp((const char*)list->data, "rend-tex-maskmipmap")
		|| !strcmp((const char*)list->data, "rend-tex-mipmap"))
	{
		Con_Executef(false, "%s force %i", list->data, value);
	}
	
	if (var) *(int*) var->ptr = value;
}

//===========================================================================
// CP_CvarEdit
//===========================================================================
void CP_CvarEdit(ui_object_t *ob)
{
	uidata_edit_t *ed = (uidata_edit_t*)ob->data;
	Con_SetString((char*)ed->data, ed->ptr);
}

//===========================================================================
// CP_CvarSlider
//===========================================================================
void CP_CvarSlider(ui_object_t *ob)
{
	uidata_slider_t *slid = (uidata_slider_t*)ob->data;
	cvar_t *var = Con_GetVariable((char*)slid->data);

	if(!var) return;

	// b211-0 align to step
	slid->value = datAlign(slid->step, slid->value, 0);

	if(var->type == CVT_FLOAT)
	{
		*(float*) var->ptr = slid->value;

		if (   slid == &sld_vid_bright
			|| slid == &sld_vid_gamma
			|| slid == &sld_vid_contrast)
		{
			update_gamma = true;
		}
	}
	else if(var->type == CVT_INT)
	{
		if ((ob->flags & 0xFF) == UIEF_NOANISO)
		{
			if (slid->value > maxAniso)
				slid->value = maxAniso;
		}
		*(int*) var->ptr = (int) (slid->value);
	}
	else if(var->type == CVT_BYTE)
		*(byte*) var->ptr = (byte) (slid->value);

	if ((ob->flags & 0xFF) == UIEF_SETFOV) ui_setfov = true;
}

//===========================================================================
// CP_KeyGrabResponder
//===========================================================================
int CP_KeyGrabResponder(ui_object_t *ob, event_t *ev)
{
	if (ev->data1 == DDKEY_ESCAPE) return false;

	if (   (ev->type == ev_mousebdown && UI_MouseInside(ob))
		|| (ev->type == ev_keydown && IS_ACTKEY(ev->data1)) )
	{
		// We want the focus.
		return true; 
	}

	// Only does something when has the focus.
	if(!(ob->flags & UIF_FOCUS)) return false;

	if(ev->type == ev_keydown)
	{
		Con_SetInteger(ob->text, ev->data1);
		// All keydown events are eaten. 
		// Note that the UI responder eats all Tabs!
		return true;
	}
	return false;
}

//===========================================================================
// CP_KeyGrabDrawer
//===========================================================================
void CP_KeyGrabDrawer(ui_object_t *ob)
{
	RDBOOL sel = (ob->flags & UIF_FOCUS) != 0;
	float alpha = (ob->flags & UIF_DISABLED? .2f : 1);
	char buf[80];
	byte key = Con_GetByte(ob->text);

	UI_DrawRectEx((float)ob->x, (float)ob->y, ob->w, ob->h,
		(float)UI_BAR_BORDER, 
		UI_COL(sel? UIC_BRD_LOW : UIC_BRD_HI), 
		UI_COL(UIC_BRD_MED), 
		UI_COL(sel? UIC_BRD_HI : UIC_BRD_LOW),
		alpha);
	UI_Gradient(ob->x + UI_BAR_BORDER, ob->y + UI_BAR_BORDER,
		ob->w - UI_BORDER, ob->h - UI_BAR_BORDER,
		UI_COL(UIC_SHADOW), 0, 1, 0);
	sprintf(buf, "%i", key);
	UI_TextOutEx(buf, ob->x + ob->w/2, ob->y + ob->h/2, true, true,
		UI_COL(UIC_TEXT), alpha);
}

//===========================================================================
// CP_TexReset
//===========================================================================
void CP_TexReset(ui_object_t *ob)
{
	Con_Execute("texreset", false);

	FR_SetFont(1);	// v220-6 reset the font to the correct size
}

//===========================================================================
//
//===========================================================================
void CP_DetailReset(ui_object_t *ob)
{
	sld_detail_far.value = detailMaxDist = 256.0f;
	sld_detail_strength.value = detailFactor = .5f;
	sld_detail_scale.value = detailScale = 4.0f;
}


//===========================================================================
// CP_MidiReset
//===========================================================================
void CP_MidiReset(ui_object_t *ob)
{
void playTempMIDIResetFile();
	if (ge_inf.rolandResetLmp >= 0) playTempMIDIResetFile();
}

//===========================================================================
// CP_ToggleList
//===========================================================================
void CP_ToggleList(ui_object_t *ob)
{
void UI_Set_showStatus();
	Con_Execute("togglelist", false);
	UI_Set_showStatus();
}

//===========================================================================
// CP_MidiRebuild
//===========================================================================
void CP_MidiRebuild(ui_object_t *ob)
{
	Con_Execute("midirebuild", false);
}

//===========================================================================
// CP_DesktopInfo
//===========================================================================
void CP_DesktopInfo(ui_object_t *ob)
{
	char buf[80];

	sprintf(buf, "Desktop: %i x %i %s %i bpp",
						cMode.max_width, cMode.max_height,
						Hz_str[0],
						cMode.w_initialbpp);

	UI_TextOutEx(buf, ob->x, ob->y + ob->h/2, false, true, UI_COL(UIC_TEXT), 1);
}

static char *CP_SetPosString(int *ptr)
{
	static char	buf[32];

	if (*(int*)(ptr + 3))
		strcpy(buf, "fullscreen");
	else if (*(int*)ptr || *(int*)(ptr + 1))
		sprintf(buf, "window at %i x %i", *(int*)ptr, *(int*)(ptr + 1));
	else
		strcpy(buf, "window centered");

	return buf;
}

//===========================================================================
// CP_ResolutionInfo
//===========================================================================
void CP_ResolutionInfo(ui_object_t *ob)
{
	char buf[80];

	if (ob->flags & UIF_DEFSCREEN)
		sprintf(buf, "Current default: %i x %i  %s %i bpp %s",
			cMode.d_width,
			cMode.d_height,
			Hz_str[1],
			cMode.d_bpp,
			CP_SetPosString(&cMode.d_xpos));
	else
		sprintf(buf, "Current settings: %i x %i  %s %i bpp %s",
			cMode.width,
			cMode.height,
			Hz_str[2],
			cMode.bpp,
			CP_SetPosString(&cMode.xpos));

	UI_TextOutEx(buf, ob->x, ob->y + ob->h/2, false, true, UI_COL(UIC_TEXT), 1);
}

//===========================================================================
// CP_ResolutionList
//  v220-3 revisions to allow control panel change from window <-> fullscreen
//===========================================================================
void CP_ResolutionList(ui_object_t *ob)
{
	uidata_list_t *list = (uidata_list_t*)ob->data;
	RDBOOL	set_tex = true;
	RDBOOL	is_win  = false;

	//-------------------------

	lastResObject = ob;
	
	nMode_ptr = ((uidata_listitem_t*)list->items)[list->selection].data2;

	memcpy(&cMode.t_width, &modeNlist[nMode_ptr], SZNMODE);
	
	cMode.t_bpp = panel_32 ? 32 : 16;
	cMode.t_fullscreen = !(ob->flags & UIF_RES_LISTW && cMode.t_bpp == cMode.w_initialbpp);

	if (ob->flags & UIF_RES_LISTW && !nMode_ptr && !cMode.t_width)
	{
		ob->flags &= ~UIF_FOCUS;
		set_tex = false;
		is_win = !list->selection;
	}

	ob = UI_FindObject(ob_panel, CPG_VIDEO, UIF_SET_RES);
	ob->flags &= ~UIF_DISABLED;

	if (set_tex)
	{
		if (cMode.t_width)
		{
			sprintf(ob->text, "Change to %i x %i %s %s",
				cMode.t_width,
				cMode.t_height,
				getHZstr(cMode.t_freq),
				CP_SetPosString(&cMode.t_xpos));

			valid_screen_ratio = cMode.t_fullscreen
				&& (unsigned)cMode.t_width == scr_max_width
				&& (unsigned)cMode.t_height == scr_max_height
				&& screen_ratio != (4.0f/3.0f);
		}
		else
		{
			ob->flags |= UIF_DISABLED;
			ob->flags &= ~UIF_FOCUS;
			set_tex = false;
		}
	}
	else if (is_win)
	{
		strcpy(ob->text, "CUSTOM unset - ALT_TAB out to allow drag/resize");
	}
	else
	{
		ob->flags |= UIF_DISABLED;
		ob->flags &= ~UIF_FOCUS;
		set_tex = false;
	}

	if (!set_tex || !memcmp(&cMode.t_width, &cMode.width, SZWDATA))
		ob->flags |= UIF_DISABLED;

	ob = UI_FindObject(ob_panel, CPG_VIDEO, UIF_SET_DEF);
	ob->flags &= ~UIF_DISABLED;

	if (!set_tex || !memcmp(&cMode.t_width, &cMode.d_width, SZWDATA))
		ob->flags |= UIF_DISABLED;

	ob = UI_FindObject(ob_panel, CPG_VIDEO, UIF_RESET_DEF);
	ob->flags &= ~UIF_DISABLED;

	if (!memcmp(&cMode.width, &cMode.d_width, SZWDATA))
		ob->flags |= UIF_DISABLED;
}

void UI_ChangeResolution(int w, int h, RDBOOL c)
{
	UI_End();	// Resolution can only be adjusted if the ui's closed
	automap_active = false;	// The automap will have to be restarted
	GL_ChangeResolution(w, h, c);
	if (R3D_edit) cfg.xhairType = 1;
}

//===========================================================================
// CP_SetRes
//  v220-3 revisions to allow control panel change from window <-> fullscreen
//===========================================================================
void CP_SetRes(ui_object_t *ob)
{
	int flags = UIHZ_CURRENT|UIHZ_DEFAULT;

	if (!lastResObject) return;

	if (ob->flags & UIF_SET_DEF)
	{
		if (memcmp(&cMode.width, &cMode.t_width, SZWDATA))
		{
			memcpy(&cMode.width, &cMode.t_width, SZWDATA);
			cMode.w_setdefaults = true;
		}
		else
		{
			memcpy(&cMode.d_width, &cMode.t_width, SZWDATA);
			setHZvalues(cMode.freq, flags);
			ob->flags |= UIF_DISABLED;
			ob->flags &= ~UIF_FOCUS;
			return;
		}
	}
	else if (ob->flags & UIF_RESET_DEF)
	{
		memcpy(&cMode.width, &cMode.d_width, SZWDATA);
	}
	else	// UIF_SET_RES
	{
		memcpy(&cMode.width, &cMode.t_width, SZWDATA);
		flags = UIHZ_CURRENT;
	}
	
	cMode.w_checkwin = 1;
	UI_ChangeResolution(cMode.width, cMode.height, !cMode.xpos && !cMode.ypos);
}

//===========================================================================
// CP_FindHover
//	Returns the option, if any, the mouse is currently hovering on. The
//	check is based on the coordinates of the Text object.
//===========================================================================
ui_object_t *CP_FindHover(void)
{
	ui_object_t *ob;
	
	for(ob = ob_panel; ob->type; ob++)
	{
		if(ob->flags & UIF_HIDDEN
			|| ob->type != UI_TEXT
			|| ob->group < 2
			|| ob->relx < 280) continue;
		// Extend the detection area to the right edge of the screen.
		if(UI_MouseInsideBox(ob->x, ob->y, current_w, ob->h))
		{
			return ob;
		}
	}

	return NULL;
}

//===========================================================================
// CP_Ticker
//	Track the mouse and move the documentation window as needed.
//===========================================================================
void CP_Ticker(ui_page_t *page)
{
	// Normal ticker actions first.
	UIPage_Ticker(page);
}

//===========================================================================
// CP_Drawer
//===========================================================================
void CP_Drawer(ui_page_t *page)
{
	// Call the regular drawer.
	UIPage_Drawer(page);
}

static int UI_ListFindItem(ui_object_t *ob, int data_value, int flag)
{
	uidata_list_t	*dat = (uidata_list_t*)ob->data;
	int				i, n;

	for (i = 0; i < dat->count; i++)
	{
		if(((uidata_listitem_t*)dat->items)[i].data == data_value)
		{
			if (ob->flags & flag)
			{
				n = ((uidata_listitem_t*)dat->items)[i].data2;

				if (modeNlist[n].freq == (uint)cMode.freq)
					return i;
			}
			else
			{
				return i;
			}
		}
	}
	return -1;
}

// v220-3 Set the resolution list with all valid values
void UI_SetDisplayLists()
{
	static RDBOOL firsttime = true;

	ui_object_t		*obf = UI_FindObject(ob_panel, CPG_VIDEO, UIF_RES_LISTF);
	ui_object_t		*obw = UI_FindObject(ob_panel, CPG_VIDEO, UIF_RES_LISTW);
	uidata_list_t	*datw = (uidata_list_t*)obw->data;
	uidata_list_t	*datf = (uidata_list_t*)obf->data;
	dispmode_t		*dplist = modeNlist;
	int				dxsc = cMode.max_width;
	int				dysc = cMode.max_height;
	int				caption = cMode.szCaption;
	int				border  = cMode.szBorder;
	int				i, k, f, w;

	//--------------------------

	if (!firsttime) return;

	firsttime = false;

	lstit_resolutionW = (uidata_listitem_t*)W_Calloc(modeNcount * sizeof(uidata_listitem_t), STATICID, 114);
	lstit_resolutionF = (uidata_listitem_t*)W_Calloc(modeNcount * sizeof(uidata_listitem_t), STATICID, 115);

	datw->items = lstit_resolutionW;
	datf->items = lstit_resolutionF;

	// Set the custom field
	((uidata_listitem_t*)datw->items)[0].data2 = 0;
	sprintf(((uidata_listitem_t*)datw->items)[0].text, "CUSTOM");

	for (i = f = 0, k = w = 1; k < modeNcount; i++, k++)
	{
		int		width  = dplist[k].width;
		int		height = dplist[k].height;
		float	ratio = (float)width/(float)height;

		((uidata_listitem_t*)datf->items)[i].data = width|(height << 16);
		((uidata_listitem_t*)datf->items)[i].data2 = k;

		// When using ATI cards any old rubbish gets stuffed in the display lists
		// by the ATI driver, irrespective of what the card will actually support.
		// This removes the most ridiculous cases in the fullscreen list - v220-10
		if (ratio < 1.34f || ratio <= screen_ratio)
		{
			((uidata_listitem_t*)datf->items)[f].data  = dplist[k].width|(dplist[k].height << 16);
			((uidata_listitem_t*)datf->items)[f].data2 = k;

			sprintf(((uidata_listitem_t*)datf->items)[f++].text, "%i x %i   %s",
				width, height, getHZstr(dplist[k].freq));
		}

		if (dplist[k].freq == (uint)cMode.w_initialfreq)
		{
			int xoff = (dxsc - width  - border) / 2;
			int yoff = (dysc - height - border - caption) / 2;

			if (   xoff + width + border >= cMode.max_width
				|| yoff + height + border + caption + 20 >= cMode.max_height)
			{
				continue;
			}

			((uidata_listitem_t*)datw->items)[w].data  = dplist[k].width|(dplist[k].height << 16);
			((uidata_listitem_t*)datw->items)[w].data2 = k;

			sprintf(((uidata_listitem_t*)datw->items)[w++].text, "%i x %i   %s",
				dplist[k].width, dplist[k].height, getHZstr(dplist[k].freq));
		}
	}

	datf->count = f;
	datw->count = w;
}

// v220-3 modified to deal with new screen resolution lists
// v220-4 sync to DD_GameLoop()
void M_Close(void);

int CCmdOpenPanel(int argc, char **argv)
{
	if (!ui_active)
	{
		open_panel_request = true;
		open_panelat[0] = 0;

		if (argc == 2)
			strcpy(open_panelat, argv[1]);
	}

	return true;
}

int UI_OpenPanel()
{
	uidata_list_t	*list;
	ui_object_t		*ob, *foc;
	int				i, k;

	//-----------------------------

	open_panel_request = false;
	S_LocalMenuSound(sfx_swtchn);

	GL_FreeScreenBuffer();
	screen_buffer = GL_GrabScreen();

	if (menu_active) M_Close();
	UI_InitPage(&page_panel, ob_panel);
	page_panel.ticker = CP_Ticker;
	page_panel.drawer = CP_Drawer;
	ui_use43 = cfg.use43;

	if (!open_panelat[0])
	{
		// Choose the group that was last visible.
		for (i = 0; i < NUM_CP_BUTTONS; i++)
		{
			if (panel_buttons[i])
			{
				if (i == NUM_CP_BUTTONS - 1)	// Don't select "Console" - R3D
				{
					panel_buttons[i] = 0;
					i = 0;
				}
				CP_ChooseGroup(foc = ob_panel + i);
			}
		}
	}
	else
	{
		// With an argument, choose the appropriate group.
		foc = NULL;

		for (i = 0; i < NUM_CP_BUTTONS; i++)
		{
			if(!_stricmp(ob_panel[i].text, open_panelat))
			{
				CP_ChooseGroup(foc = ob_panel + i);
				break;
			}
		}

		if(!foc) CP_ChooseGroup(foc = ob_panel);
	}
	
	panel_32 = cMode.bpp == 32 ? 1 : 0;

	lst_resolutionW.numtodisplay = 12;
	lst_resolutionF.numtodisplay = 12;

	// Set cvarbutton data pointers.
	// This is only done the first time "panel" is issued.
	for(i = 0, ob = ob_panel; ob_panel[i].type; i++, ob++)
	{
		if(ob->action == CP_CvarButton)
		{
			char n[16], y[16];

			if (ob->flags & UIF_ASPECT)
			{
				strcpy(n, aspect_str), strcpy(y, "4:3");
			}
			else
			if (ob->flags & UIF_BPP)
			{
				strcpy(n, "16 bpp"), strcpy(y, "32 bpp");
			}
			else
			{
				strcpy(n, "No"), strcpy(y, "Yes");
			}

			if(!strcmp(ob->text, y) || !strcmp(ob->text, n))
			{
				// This button has already been initialized.
				// Update value?
				if(ob->data)
				{
					cvarbutton_t *cvb = (cvarbutton_t*)ob->data;
					cvb->active = (byte)(Con_GetByte(cvb->cvarname) != 0);
					strcpy(ob->text, cvb->active? y : n);
				}
				continue;
			}

			// Find the cvarbutton representing this one.
			for(k = 0; cvarbuttons[k].cvarname; k++)
			{
				if(!strcmp(ob->text, cvarbuttons[k].cvarname))
				{
					ob->data = &cvarbuttons[k];
					cvarbuttons[k].active 
						= (byte)(Con_GetByte(cvarbuttons[k].cvarname) != 0);
					strcpy(ob->text, cvarbuttons[k].active? y : n);
					break;
				}
			}
		}

		if(ob->action == CP_CvarList)
		{
			list = (uidata_list_t*)ob->data;
			// Choose the correct list item based on the value of the cvar.
			list->selection = UI_ListFindItem(ob, Con_GetInteger((char*)list->data), 0);
		}

		if(ob->action == CP_CvarEdit)
		{
			uidata_edit_t *ed = (uidata_edit_t*)ob->data;
			strncpy(ed->ptr, Con_GetString((char*)ed->data), ed->maxlen);
		}

		if(ob->action == CP_CvarSlider)
		{
			uidata_slider_t *slid = (uidata_slider_t*)ob->data;
			slid->value = slid->floatmode? Con_GetFloat((char*)slid->data)
				: Con_GetInteger((char*)slid->data);
		}
	}
	
	UI_SetDisplayLists();

	ob = UI_FindObject(ob_panel, CPG_VIDEO, UIF_RES_LISTW);

	// Set the custom data field
	list = (uidata_list_t*)ob->data;
	((uidata_listitem_t*)list->items)[0].data = modeNlist[0].width|(modeNlist[0].height << 16);
	
	lastResObject = NULL;

	ob = UI_FindObject(ob_panel, CPG_VIDEO, UIF_SET_RES);
	ob->flags |= UIF_DISABLED;
	ob = UI_FindObject(ob_panel, CPG_VIDEO, UIF_SET_DEF);
	ob->flags |= UIF_DISABLED;
	ob = UI_FindObject(ob_panel, CPG_VIDEO, UIF_RESET_DEF);
	ob->flags |= UIF_DISABLED;
	if (cMode.bpp != cMode.w_initialbpp)
	{
		ob = UI_FindObject(ob_panel, CPG_VIDEO, UIF_RES_LISTW);
		ob->flags |= UIF_DISABLED;
	}

	UI_Init();

	UI_SetPage(&page_panel);

	if (foc) UI_Focus(foc);
	return true;
}
