//------------------------------------------------------------------------
// LEVEL : Level structure read/write functions.
//------------------------------------------------------------------------
//
//  GL-Friendly Node Builder (C) 2000-2002 Andrew Apted
//
//  Based on `BSP 2.3' by Colin Reed, Lee Killough and others.
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  Revision 2.2.0 build 6 (v220-6) 19/Apr/2010 gmj
//   Modified for internal use. See the bsp_ReadMe.txt in the bsp dir
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
//------------------------------------------------------------------------

#include "rd_system.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <math.h>
#include <limits.h>
#include <assert.h>

#include "rd_blockmap.h"
#include "rd_level.h"
#include "rd_node.h"
#include "rd_seg.h"
#include "rd_structs.h"
#include "rd_util.h"
#include "rd_wad.h"

#include "dd_types.h"
#include "rd_shared.h"	// R3D

void pr_OFinsert(char *message, ...);
void pr_OFdirect(char *msg);

#define GL_MARKER	0x80000000

#define ALLOC_BLKNUM  1024
#define POLY_BOX_SZ		10

BOOL doing_normal;

// per-level variables

static BOOL normal_exists;

#define LEVELARRAY(TYPE, BASEVAR, NUMVAR)  \
    static TYPE ** BASEVAR = NULL;  \
    int NUMVAR = 0;


// v220-6 Note: sector_t replaced with sect_t, seg_t with nseg_t
//        vertex_t with vert_t, linedaf_t with ldef_t and node_t with nodb_t
//        also sectors with sectrs
//        not because this was necessary but to avoid confusion

LEVELARRAY(vert_t	, vertices,   num_vertices)
LEVELARRAY(ldef_t	, linedefs,   num_linedefs)
LEVELARRAY(sidedef_t, sidedefs,   num_sidedefs)
LEVELARRAY(sect_t	, sectrs,     num_sectors)
LEVELARRAY(nseg_t	, nsegs,      num_segs)
LEVELARRAY(subsec_t	, subsecs,    num_subsecs)
LEVELARRAY(nodb_t	, nodes,      num_nodes)
LEVELARRAY(nodb_t	, stale_nodes,num_stale_nodes)
LEVELARRAY(wall_tip_t,wall_tips,  num_wall_tips)

int num_normal_vert = 0;
int num_gl_vert = 0;
int num_complete_seg = 0;

int doing_factor;

// v220-6 map identifiers for specific modifications
//        The idea of modifying values prior to building nodes was
//        originally conceived just for the Doom2 iwad map faults
//        as they are pretty dreadful. It became extended to also
//        deal with Eternal III as this seems to be the one and
//        only pwad where faults cannot be retrofixed.
extern int doom2Map;
extern int mapEternal;
extern int verbose;

static int GL_extended;

/* ----- allocation routines ---------------------------- */

#define ALLIGATOR(TYPE, BASEVAR, NUMVAR)  \
{  \
  if ((NUMVAR % ALLOC_BLKNUM) == 0)  \
  {  \
    BASEVAR = (TYPE**)UtilRealloc(BASEVAR, (NUMVAR + ALLOC_BLKNUM) *   \
        sizeof(TYPE *));  \
  }  \
  BASEVAR[NUMVAR] = (TYPE *) UtilCalloc(sizeof(TYPE));  \
  NUMVAR += 1;  \
  return BASEVAR[NUMVAR - 1];  \
}


vert_t *NewVertex(void)
  ALLIGATOR(vert_t, vertices, num_vertices)

ldef_t *NewLinedef(void)
  ALLIGATOR(ldef_t, linedefs, num_linedefs)

sidedef_t *NewSidedef(void)
  ALLIGATOR(sidedef_t, sidedefs, num_sidedefs)

sect_t *NewSector(void)
  ALLIGATOR(sect_t, sectrs, num_sectors)

nseg_t *NewSeg(void)
  ALLIGATOR(nseg_t, nsegs, num_segs)

subsec_t *NewSubsec(void)
  ALLIGATOR(subsec_t, subsecs, num_subsecs)

nodb_t *NewNode(void)
  ALLIGATOR(nodb_t, nodes, num_nodes)

nodb_t *NewStaleNode(void)
  ALLIGATOR(nodb_t, stale_nodes, num_stale_nodes)

wall_tip_t *NewWallTip(void)
  ALLIGATOR(wall_tip_t, wall_tips, num_wall_tips)


/* ----- free routines ---------------------------- */

#define FREEMASON(TYPE, BASEVAR, NUMVAR)  \
{  \
  int i;  \
  for (i=0; i < NUMVAR; i++)  \
    UtilFree(BASEVAR[i]);  \
  if (BASEVAR)  \
    UtilFree(BASEVAR);  \
  BASEVAR = NULL; NUMVAR = 0;  \
}


void FreeVertices(void)
  FREEMASON(vert_t, vertices, num_vertices)

void FreeLinedefs(void)
  FREEMASON(ldef_t, linedefs, num_linedefs)

void FreeSidedefs(void)
  FREEMASON(sidedef_t, sidedefs, num_sidedefs)

void FreeSectors(void)
  FREEMASON(sect_t, sectrs, num_sectors)

void FreeSegs(void)
  FREEMASON(nseg_t, nsegs, num_segs)

void FreeSubsecs(void)
  FREEMASON(subsec_t, subsecs, num_subsecs)

void FreeNodes(void)
  FREEMASON(nodb_t, nodes, num_nodes)

void FreeStaleNodes(void)
  FREEMASON(nodb_t, stale_nodes, num_stale_nodes)

void FreeWallTips(void)
  FREEMASON(wall_tip_t, wall_tips, num_wall_tips)


/* ----- lookup routines ------------------------------ */

#define LOOKERUPPER(BASEVAR, NUMVAR, NAMESTR)  \
{  \
  if (index < 0 || index >= NUMVAR)  \
    FatalError("No such %s number #%d", NAMESTR, index);  \
    \
  return BASEVAR[index];  \
}

vert_t *LookupVertex(int index)
  LOOKERUPPER(vertices, num_vertices, "vertex")

ldef_t *LookupLinedef(int index)
  LOOKERUPPER(linedefs, num_linedefs, "linedef")
  
sidedef_t *LookupSidedef(int index)
  LOOKERUPPER(sidedefs, num_sidedefs, "sidedef")
  
sect_t *LookupSector(int index)
  LOOKERUPPER(sectrs, num_sectors, "sector")
  
nseg_t *LookupSeg(int index)
  LOOKERUPPER(nsegs, num_segs, "seg")
  
subsec_t *LookupSubsec(int index)
  LOOKERUPPER(subsecs, num_subsecs, "subsector")
  
nodb_t *LookupNode(int index)
  LOOKERUPPER(nodes, num_nodes, "node")

nodb_t *LookupStaleNode(int index)
  LOOKERUPPER(stale_nodes, num_stale_nodes, "stale_node")


/* ----- reading routines ------------------------------ */


//
// CheckForNormalNodes
//
BOOL CheckForNormalNodes(void)
{
	lump_t *lump;
	
	/* Note: an empty NODES lump can be valid */
	if (FindLevelLump("NODES") == NULL)
		return false;
	
	lump = FindLevelLump("SEGS");
	
	if (! lump || lump->length == 0 || CheckLevelLumpZero(lump))
		return false;
	
	return true;
}

//
// GetVertices
//
void GetVertices(void)
{
	int i, k, count=-1;
	raw_vertex_t *raw;
	lump_t *lump = FindLevelLump("VERTEXES");
	
	if (lump)
		count = lump->length / sizeof(raw_vertex_t);
		
	if (!lump || count == 0)
		FatalError("Couldn't find any Vertices");
	
	raw = (raw_vertex_t *) lump->data;
	
	for (i=0; i < count; i++, raw++)
	{
		vert_t *vert = NewVertex();
		
		vert->x = (float) (short)(raw->x);
		vert->y = (float) (short)(raw->y);

		// v220-35 fix long unnoticed vertex placement error in Doom2 map02
		if (doom2Map == 2 && i == 273)
			vert->y = 1208.f;
		// v220-6 This misplaced vertex causes L1493 to cross L1490
		//        which, put bluntly, buggers the nodes build
		else if (mapEternal == 28 && i == 1268)
			vert->x = 681.f;

		for (k = 0; k < numvertexfix; k++)
		{
			if (vertexfix[k].vtxnum == i)
			{
				vert->x = (float) (short)(vertexfix[k].vx);
				vert->y = (float) (short)(vertexfix[k].vy);
			}
		}
		
		vert->index = i;
	}
	
	num_normal_vert = num_vertices;
	num_gl_vert = 0;
	num_complete_seg = 0;
}

//
// GetSectors
//
void GetSectors(void)
{
	int i, count=-1;
	raw_sector_t *raw;
	lump_t *lump = FindLevelLump("SECTORS");
	
	if (lump)
		count = lump->length / sizeof(raw_sector_t);
	
	if (!lump || count == 0)
		FatalError("Couldn't find any Sectors");
	
	raw = (raw_sector_t *) lump->data;
	
	for (i=0; i < count; i++, raw++)
	{
		sect_t *sector = NewSector();
		
		sector->floor_h = (short)(raw->floor_h);
		sector->ceil_h  = (short)(raw->ceil_h);
		
		memcpy(sector->floor_tex, raw->floor_tex, sizeof(sector->floor_tex));
		memcpy(sector->ceil_tex,  raw->ceil_tex,  sizeof(sector->ceil_tex));
		
		sector->light = (ushort)(raw->light);
		sector->special = (ushort)(raw->special);
		sector->tag = (short)(raw->tag);
		
		sector->coalesce = (sector->tag >= 900 && sector->tag < 1000) ? true : false;
		
		/* sector indices never change */
		sector->index = i;
		
		/* Note: rej_* fields are handled completely in reject.c */
	}
}

//
// GetSidedefs
//
void GetSidedefs(void)
{
	int i, count=-1;
	raw_sidedef_t *raw;
	lump_t *lump = FindLevelLump("SIDEDEFS");
	
	if (lump)
		count = lump->length / sizeof(raw_sidedef_t);
	
	if (!lump || count == 0)
		FatalError("Couldn't find any Sidedefs");
	
	raw = (raw_sidedef_t *) lump->data;
	
	for (i=0; i < count; i++, raw++)
	{
		sidedef_t *side = NewSidedef();
		
		side->sector = ((ushort)(raw->sector) == 0xFFFF) ? NULL :
		LookupSector((ushort)(raw->sector));
		
		if (side->sector)
			side->sector->ref_count++;
		
		side->x_offset = (short)(raw->x_offset);
		side->y_offset = (short)(raw->y_offset);
		
		memcpy(side->upper_tex, raw->upper_tex, sizeof(side->upper_tex));
		memcpy(side->lower_tex, raw->lower_tex, sizeof(side->lower_tex));
		memcpy(side->mid_tex,   raw->mid_tex,   sizeof(side->mid_tex));
		
		/* sidedef indices never change */
		side->index = i;
	}
}

char global_name[9];

// $R3Dgwa
int R3D_ChkLineConnected(int lnum, int vnum, unsigned char *vtxarray1)
{
	int i;

	for (i = 0; i < num_linedefs; i++)
	{
		ldef_t *line = linedefs[i];

		if ((line->start->index == vnum ||
			 line->end->index == vnum) &&
			i != lnum)
		{
			if (vtxarray1[line->start->index] == 2 &&
				vtxarray1[line->end->index] == 2)
				return false;
			return true;
		}
	}
	return true;	//happy compiler
}

// R3D v220-33 scripted NODESfix
void R3D_CheckNODESfix()
{
	static char *nderr = "NODESfix: script error ";
	static char *ndabort = ", script abandoned\n";

	sidedef_t *sd1, *sd2;
	int i;

	for (i = 0; i <= currentnodefix; i++)
	{
		ldef_t *line = linedefs[nodefix[i].line];

		if (nodefix[i].side1sec < 0)
		{
			pr_OFinsert("%sside1 must be given a valid sector%s", nderr, ndabort);
			return;
		}

		if (nodefix[i].ndflags & ND_NEWSIDE2 && nodefix[i].side2sec < 0)
		{
			pr_OFinsert("%sside2 must be given a valid sector%s\n", nderr, ndabort);
			return;
		}

		if ((line->right && nodefix[i].ndflags & ND_NEWSIDE1) || (line->left && nodefix[i].ndflags & ND_NEWSIDE2))
		{
			pr_OFinsert("%san existing sidedef cannot be replaced%s\n", nderr, ndabort);
			return;
		}
	}

	for (i = 0; i <= currentnodefix; i++)
	{
		ldef_t *line = linedefs[nodefix[i].line];

		if (nodefix[i].ndflags & ND_CLEAR2S)
		{
			line->flags &= ~LINEFLAG_TWO_SIDED;
			line->flags |= LINEFLAG_IMPASSABLE;
			line->two_sided = false;
		}

		if (nodefix[i].ndflags & ND_SWAPSIDES)
		{
			sd1 = line->left;
			sd2 = line->right;
			if (!sd1) nodefix[i].ndflags |= ND_NEWSIDE1;
			line->right = sd1 ? sd1 : NewSidedef();
			line->left = sd2;
			sd1 = line->right;
			sd2 = line->left;
		}
		else
		{
			sd1 = nodefix[i].ndflags & ND_NEWSIDE1 ? NewSidedef() : line->right;
			sd2 = nodefix[i].ndflags & ND_NEWSIDE2 ? NewSidedef() : line->left;
		}

		if (!sd1) FatalError("NODESfix: side1 must be set");

		sd1->sector = LookupSector(nodefix[i].side1sec); 

		if (sd2)
			sd2->sector = nodefix[i].side2sec >= 0 ? LookupSector(nodefix[i].side2sec) : NULL;

		if (nodefix[i].ndflags & ND_NEWSIDE2)
		{
			if (!sd2) FatalError("NODESfix: newside asked for but side2 is unset");
			sd2->index = num_sidedefs - 1;
			sd2->sector->ref_count++;
			line->left = sd2;
			line->flags |= LINEFLAG_TWO_SIDED;
			line->two_sided = true;
			line->left->ref_count++;
			line->left->sector = sd2->sector;
		}

		if (nodefix[i].ndflags & ND_NEWSIDE1)
		{
			sd1->index = num_sidedefs - 1;
			sd1->sector->ref_count++;
			line->right = sd1;
			line->right->ref_count++;
			line->right->sector = sd1->sector;
		}

		line->right->sector = sd1->sector;
		
		// Set the textures *****************************
		if (stricmp(nodefix[i].sd1hi, "NOCHANGE"))
		{
			if (stricmp(nodefix[i].sd1hi, "NONE"))
				memcpy(sd1->upper_tex, nodefix[i].sd1hi, sizeof(sd1->upper_tex));
			else
				sd1->upper_tex[0] = '-';
		}
		
		if (stricmp(nodefix[i].sd1mid, "NOCHANGE"))
		{
			if (stricmp(nodefix[i].sd1mid, "NONE"))
				memcpy(sd1->mid_tex, nodefix[i].sd1mid, sizeof(sd1->mid_tex));
			else
				sd1->mid_tex[0] = '-';
		}
		
		if (stricmp(nodefix[i].sd1low, "NOCHANGE"))
		{
			if (stricmp(nodefix[i].sd1low, "NONE"))
				memcpy(sd1->lower_tex, nodefix[i].sd1low, sizeof(sd1->lower_tex));
			else
				sd1->lower_tex[0] = '-';
		}
		
		if (sd2)	// should be
		{
			if (stricmp(nodefix[i].sd2hi, "NOCHANGE"))
			{
				if (stricmp(nodefix[i].sd2hi, "NONE"))
					memcpy(sd2->upper_tex, nodefix[i].sd2hi, sizeof(sd2->upper_tex));
				else
					sd2->upper_tex[0] = '-';
			}
			
			if (stricmp(nodefix[i].sd2mid, "NOCHANGE"))
			{
				if (stricmp(nodefix[i].sd2mid, "NONE"))
					memcpy(sd2->mid_tex, nodefix[i].sd2mid, sizeof(sd2->mid_tex));
				else
					sd2->mid_tex[0] = '-';
			}
			
			if (stricmp(nodefix[i].sd2low, "NOCHANGE"))
			{
				if (stricmp(nodefix[i].sd2low, "NONE"))
					memcpy(sd2->lower_tex, nodefix[i].sd2low, sizeof(sd2->lower_tex));
				else
					sd2->lower_tex[0] = '-';
			}
		}
		// ************************************************
	}
}

// $R3Dgwa
void R3D_SpecialFX(void)
{
	unsigned char *vtxarray1, *vtxarray2;
	int		i;
	uint	size = 0;
		
	vtxarray1 = (byte*)UtilCalloc(num_vertices);
	vtxarray2 = (byte*)UtilCalloc(num_vertices);
	
	for (i = 0; i < num_linedefs; i++)
	{
		ldef_t *line = linedefs[i];
		
		if (line->right && !line->left)
		{
			vtxarray1[line->start->index]++;
			vtxarray1[line->end->index]++;
		}
		vtxarray2[line->start->index]++;
		vtxarray2[line->end->index]++;
	}
	
	for (i = 0; i < num_linedefs; i++)
	{
		ldef_t *line = linedefs[i];
		
		if (line->right && !line->left)
		{
			unsigned char vnum1 = vtxarray1[line->start->index];
			unsigned char vnum2 = vtxarray1[line->end->index];
			unsigned char vnum3 = vtxarray2[line->start->index];
			unsigned char vnum4 = vtxarray2[line->end->index];

			int test1 = (vnum1 > 2 && vnum1 % 2) && vnum1 == vnum2;

			if (!test1)
			{
				if ((vnum1 == 3 && vnum2 == 2 && vnum4 == 2) ||
					(vnum1 == 2 && vnum2 == 3 && vnum3 == 2))
				{
					int vnum = vnum1 == 2 ? line->start->index : line->end->index;
					test1 = R3D_ChkLineConnected(i, vnum, vtxarray1);
				}
			}
			
			if (test1)
			{
				sidedef_t *side = NewSidedef();
				
				side->sector = line->right->sector;
				side->sector->ref_count++;
				
				memset(side->upper_tex, 0, sizeof(side->upper_tex));
				memset(side->mid_tex  , 0, sizeof(side->mid_tex));
				memset(side->lower_tex, 0, sizeof(side->lower_tex));
				
				side->upper_tex[0] = '-';
				side->lower_tex[0] = '-';
				side->mid_tex[0] = '-';
				
				side->index = num_sidedefs - 1;
				
				line->left = side;
				line->left->ref_count++;
				line->left->sector = side->sector;
				line->left->on_special = 0;
				line->two_sided = true;
				line->flags |= LINEFLAG_TWO_SIDED;
				line->is_special = true;

				if (verbose)
					pr_OFinsert("MAP FIX: 1's LINE %i has been made 2 sided\n", i);
			}
		}
	}
	
	UtilFree(vtxarray1);
	UtilFree(vtxarray2);
}

//
// GetLinedefs
//

// v220-6 modification code ========================================

// Add a same sector line in sector 0 from vtx 1707 to 1708
static void ModifyEternalMap10(int count)
{
	ldef_t *line;
	int    i;
		
	line = NewLinedef();
	
	line->start = LookupVertex(1707);
	line->end   = LookupVertex(1708);
	line->start->ref_count++;
	line->end->ref_count++;

	line->type = 0;
	line->tag = 0;
	line->is_precious = false;
	line->is_special = false;
	line->zero_len = false;
	line->two_sided = true;
	line->flags = LINEFLAG_TWO_SIDED;
	
	line->index = count;

	// set the sidedefs
	for (i = 0; i < 2; i++)
	{
		sidedef_t *side = NewSidedef();
		
		if (i) line->right = side; else line->left = side;
		
		side->index = num_sidedefs - 1;

		side->sector = LookupSector(0);
		side->sector->ref_count++;
		
		memset(side->upper_tex, 0, sizeof(side->upper_tex));
		memset(side->mid_tex  , 0, sizeof(side->mid_tex));
		memset(side->lower_tex, 0, sizeof(side->lower_tex));
		
		side->upper_tex[0] = '-';
		side->lower_tex[0] = '-';
		side->mid_tex[0] = '-';
		
		side->ref_count++;
	}
}

// The following Doom2 fixes are specifically restricted to these maps
// as all three are horrible errors and should not be encouraged in pwads.

// v220-6 Fix the line problem in Doom2 map18
static void ModifyDoom2Map18()
{
	sidedef_t *sd1 = LookupSidedef(536);
	sidedef_t *sd2 = LookupSidedef(545);
	ldef_t    *ld1 = LookupLinedef(451);
	ldef_t    *ld2 = LookupLinedef(459);

	if (   ld1->right == sd1
		&& ld2->right == sd2
		&& ld1->right->upper_tex[0] != '-'
		&& ld1->right->mid_tex  [0] == '-'
		&& ld1->right->lower_tex[0] == '-'
		&& ld1->left->upper_tex [0] == '-'
		&& ld1->left->mid_tex   [0] == '-'
		&& ld1->left->lower_tex [0] == '-')
	{
		ld1->flags &= ~LINEFLAG_TWO_SIDED;
		memcpy(ld1->right->mid_tex, ld1->right->upper_tex, sizeof(ld1->right->mid_tex));
		ld2->flags &= ~LINEFLAG_TWO_SIDED;
		memcpy(ld2->right->mid_tex, ld2->right->upper_tex, sizeof(ld2->right->mid_tex));
		pr_OFdirect("Lines 451 and 459 made one sided\n");
	}
}

// v220-6 Fix the line 636 problem in Doom2 map19
static void ModifyDoom2Map19(raw_linedef_t *raw)
{
	raw_linedef_t *tmp = &raw[636];

	if (tmp->start == 517 && tmp->end == 520)
	{
		tmp->start = 520;
		tmp->end = 517;
		pr_OFdirect("Swapped VTXS L636\n");
	}
}

// v220-6 Fix the sidedef sector problem in Doom2 map21
static void ModifyDoom2Map21()
{
	sidedef_t *sd1 = LookupSidedef(183);
	sidedef_t *sd2 = LookupSidedef(220);
	sect_t    *sec = LookupSector(50);

	if (sd1->sector == sec && sd2->sector == sec)
	{
		sd1->sector = LookupSector(49);
		sd2->sector = sd1->sector;
		pr_OFdirect("Set sides 183 and 220's sectors from 50 to 49\n");
	}
}

// v220-7 Fix 2's line problem in Doom2 map22
static void ModifyDoom2Map22()
{
	uint	ln[] = { 530, 539, 542, 543, 544, 545, 548 };
	uint	sn[] = { 771, 786, 790, 792, 794, 796, 800 };
	ldef_t	*ld1;
	int		i;

	// Check all parms match
	for (i = 0; i < 7; i++)
	{
		ld1 = LookupLinedef(ln[i]);

		if (   ld1->right != LookupSidedef(sn[i])
			|| ld1->left  != LookupSidedef(sn[i] + 1) )
		{
			return;	// Mismatch
		}
	}

	pr_OFdirect("Lines made one sided:");

	for (i = 0; i < 7; i++)
	{
		ld1 = LookupLinedef(ln[i]);
		ld1->flags &= ~LINEFLAG_TWO_SIDED;
		pr_OFinsert(" %i", ln[i]);
	}

	pr_OFdirect("\n");
}

// v220-6 end modification code ===============================

void GetLinedefs(void)
{
	raw_linedef_t	*raw;
	lump_t			*lump = FindLevelLump("LINEDEFS");
	int				i, count = -1;

	//-------------------------

	if (lump)
		count = lump->length / sizeof(raw_linedef_t);
	
	if (count <= 0)
		FatalError("Couldn't find any Linedefs");
	
	raw = (raw_linedef_t*) lump->data;
	
	if (doom2Map == 19) ModifyDoom2Map19(raw);	// v220-6
	
	for (i=0; i < count; i++, raw++)
	{
		ldef_t *line;
		
		vert_t *start = LookupVertex((ushort)(raw->start));
		vert_t *end   = LookupVertex((ushort)(raw->end));
		
		start->ref_count++;
		end->ref_count++;
		
		line = NewLinedef();

		// v220-7 Fix line length errors in Doom2 map30
		// Doubtless missed with a 320x200 screen res but nowadays sticks
		// out like a sore thumb especially when using hi-res textures
		if (doom2Map == 30)
		{
			int fixed = false;

			if (start->x == 2488 && (start->y == 1406 || start->y == 1408 || start->y == 1410))
			{
				start->x += 8;
				fixed = true;
			}

			if (end->x == 2488 && end->y == 1406)
			{
				end->x += 8;
				fixed = true;
			}
			
			if (fixed) pr_OFinsert("Fixed line %i's length\n", i);
		}
		
		line->start = start;
		line->end   = end;
		
		/* check for zero-length line */
		line->zero_len = (fabs(start->x - end->x) < DIST_EPSILON) && 
			(fabs(start->y - end->y) < DIST_EPSILON);
		
		line->flags = (ushort)(raw->flags);
		line->type  = (ushort)(raw->type);
		line->tag   = (short)(raw->tag);
		
		// $r3d v220-35
		if (doom2Map == 27 && i == 342)
		{
			line->flags |= LINEFLAG_TWO_SIDED;
			line->flags &= ~LINEFLAG_IMPASSABLE;
		}

		line->two_sided = (line->flags & LINEFLAG_TWO_SIDED) ? true : false;
		line->is_precious = (line->tag >= 900 && line->tag < 1000) ? 
			true : false;
		
		// Extended =======================================
		{
			unsigned short sd1 = (ushort)(raw->sidedef1);
			unsigned short sd2 = (ushort)(raw->sidedef2);
			
			line->right = NULL;
			line->left  = NULL;
			
			if (sd1 != 0xFFFF) line->right = LookupSidedef(sd1);
			if (sd2 != 0xFFFF) line->left = LookupSidedef(sd2);
		}
		// ================================================
		
		// Line side check - fortunately these errors are pretty rare
		if (!line->right && line->left)
		{
			line->right = line->left;
			line->flags &= ~LINEFLAG_TWO_SIDED;
		}
		
		if (line->right)
		{
			line->right->ref_count++;
			line->right->on_special |= (line->type > 0) ? 1 : 0;
		}
		
		if (line->left)
		{
			line->left->ref_count++;
			line->left->on_special |= (line->type > 0) ? 1 : 0;
		}
		
		line->index = i;
	}

	if (doom2Map == 18) ModifyDoom2Map18();			// v220-6
	else if (doom2Map == 21) ModifyDoom2Map21();			// v220-6
	else if (doom2Map == 22) ModifyDoom2Map22();			// v220-6
	else if (mapEternal == 10) ModifyEternalMap10(count);	// v220-6
	else R3D_CheckNODESfix();
	
	if (doing_normal && num_linedefs > 0) R3D_SpecialFX();	// $R3Dgwa
}

//
// GetStaleNodes
//
void GetStaleNodes(void)
{
	int i, count=-1;
	raw_node_t *raw;
	lump_t *lump = FindLevelLump("NODES");
	
	if (!lump
		|| (count = lump->length / sizeof(raw_node_t)) < 5)
		return;
	
	raw = (raw_node_t *) lump->data;
	
	// All the nodes must be pre-allocated as they
	// contain internal references to each other.
	for (i=0; i < count; i++)
	{
		NewStaleNode();
	}
	
	for (i=0; i < count; i++, raw++)
	{
		nodb_t *nd = LookupStaleNode(i);
		
		nd->x  = (short)(raw->x);
		nd->y  = (short)(raw->y);
		nd->dx = (short)(raw->dx);
		nd->dy = (short)(raw->dy);
		
		nd->index = i;
		
		// Note: we only need the nodes
		if (((ushort)(raw->right) & 0x8000) == 0)
		{
			nd->r.node = LookupStaleNode((ushort)(raw->right));
		}
		
		if (((ushort)(raw->left) & 0x8000) == 0)
		{
			nd->l.node = LookupStaleNode((ushort)(raw->left));
		}
	}
}


/* ----- analysis routines ----------------------------- */

static int __cdecl VertexCompare(const void *p1, const void *p2)
{
	int vert1 = ((const ushort *) p1)[0];
	int vert2 = ((const ushort *) p2)[0];
	
	vert_t *A = vertices[vert1];
	vert_t *B = vertices[vert2];
	
	if (vert1 == vert2)
		return 0;
	
	if ((int)A->x != (int)B->x)
		return (int)A->x - (int)B->x; 
	
	return (int)A->y - (int)B->y;
}

static int __cdecl SidedefCompare(const void *p1, const void *p2)
{
	int comp;
	
	int side1 = ((const ushort *) p1)[0];
	int side2 = ((const ushort *) p2)[0];
	
	sidedef_t *A = sidedefs[side1];
	sidedef_t *B = sidedefs[side2];
	
	if (side1 == side2)
		return 0;
	
	// don't merge sidedefs on special lines
	if (A->on_special || B->on_special)
		return side1 - side2;
	
	if (A->sector != B->sector)
	{
		if (A->sector == NULL) return -1;
		if (B->sector == NULL) return +1;
		
		return (A->sector->index - B->sector->index);
	}
	
	if ((int)A->x_offset != (int)B->x_offset)
		return A->x_offset - (int)B->x_offset;
	
	if ((int)A->y_offset != B->y_offset)
		return (int)A->y_offset - (int)B->y_offset;
	
	// compare textures
	
	comp = memcmp(A->upper_tex, B->upper_tex, sizeof(A->upper_tex));
	if (comp) return comp;
	
	comp = memcmp(A->lower_tex, B->lower_tex, sizeof(A->lower_tex));
	if (comp) return comp;
	
	comp = memcmp(A->mid_tex, B->mid_tex, sizeof(A->mid_tex));
	if (comp) return comp;
	
	// sidedefs must be the same
	return 0;
}

static void DetectDuplicateVertices(void)
{
	int i;
	ushort *array = (ushort*)UtilCalloc(num_vertices * sizeof(ushort));
	
	for (i = 0; i < num_linedefs; i++)
	{
		ldef_t *line = linedefs[i];

		if (!(line->flags & LINEFLAG_TWO_SIDED))
		{
			vertices[line->start->index]->canbemerged = 1;
			vertices[line->end->index]->canbemerged = 1;
		}
	}
	
	// sort array of indices
	for (i=0; i < num_vertices; i++)
		array[i] = i;
	
	qsort(array, num_vertices, sizeof(ushort), VertexCompare);
	
	// now mark them off
	for (i=0; i < num_vertices - 1; i++)
	{
		// duplicate ?
		if (VertexCompare(array + i, array + i+1) == 0)
		{
			vert_t *A = vertices[array[i]];
			vert_t *B = vertices[array[i+1]];

			if (A->canbemerged && B->canbemerged)
			{
				// found a duplicate !
				if (A->equiv)
					B->equiv = A->equiv;
				else
					B->equiv = A;
			}
		}
	}
	
	UtilFree(array);
}

static void DetectDuplicateSidedefs(void)
{
	int i;
	ushort *array = (ushort*)UtilCalloc(num_sidedefs * sizeof(ushort));
	
	// sort array of indices
	for (i=0; i < num_sidedefs; i++)
		array[i] = i;
	
	qsort(array, num_sidedefs, sizeof(ushort), SidedefCompare);
	
	// now mark them off
	for (i=0; i < num_sidedefs - 1; i++)
	{
		// duplicate ?
		if (SidedefCompare(array + i, array + i+1) == 0)
		{
			sidedef_t *A = sidedefs[array[i]];
			sidedef_t *B = sidedefs[array[i+1]];
			
			// found a duplicate !
			if (A->equiv)
				B->equiv = A->equiv;
			else
				B->equiv = A;
		}
	}
	
	UtilFree(array);
}

// R3Dgwa v220-25 add xline support
#include "r3d_xline.h"
#include "m_winmem.h"	// R3D

extern int	use_xl_lines;

static void findXlines(byte *lmem)
{
	int			*cptr = (int *)calloc(num_linedefs, sizeof(int));
	int			*lptr = (int *)calloc(num_linedefs, sizeof(int));
	int			num_xlfound = 0;
	int			i, m, n;

	// identify same sector lines
	for (i = m = n = 0; i < num_linedefs; i++)
	{
		ldef_t *L = linedefs[i];

		if (   L->right
			&& L->left
			&& L->right->sector == L->left->sector
			&& !L->right->sector->special
			&& !L->is_special
			&& !L->zero_len
			&& L->right->lower_tex[0] == '-'
			&& L->left->lower_tex[0] == '-'
			&& L->right->upper_tex[0] == '-'
			&& L->left->upper_tex[0] == '-')
		{
			if (   L->right->sector->tag
				&& L->right->mid_tex[0] == '-'
				&& L->left->mid_tex[0] == '-')
			{
				cptr[n++] = i;
			}
			else
			if (   !(L->flags & 0x0243)	// NOT ML_PASSUSE/ML_SOUNDBLOCK/ML_BLOCKMONSTERS/ML_BLOCKING
				&& !L->tag
				&& !L->right->sector->tag
				&& L->right->mid_tex[0] != '-'
				&& L->left->mid_tex[0] != '-'
				&& L->right->y_offset == L->left->y_offset
				&& !strncmp(L->right->mid_tex, L->left->mid_tex, 8))
			{
				lptr[m++] = i;
			}
		}
	}

	if (m > 1 && n > 1)
	{
		typedef struct { float	bbox[4]; } fbox_t;

		fbox_t		*fbox1, *fbox2;
		ldef_t		*L;
		float		*b1, *b2;
		int			j;

		fbox1 = (fbox_t*)malloc(m * sizeof(fbox_t));
		fbox2 = (fbox_t*)malloc(n * sizeof(fbox_t));

		// create bounding boxes for lines
		for (i = 0; i < m; i++)
		{
			L = linedefs[lptr[i]];

			fbox1[i].bbox[BLEFT ] = L->start->x < L->end->x ? L->start->x : L->end->x;
			fbox1[i].bbox[BRIGHT] = L->start->x < L->end->x ? L->end->x   : L->start->x;
			fbox1[i].bbox[BBASE ] = L->start->y < L->end->y ? L->start->y : L->end->y;
			fbox1[i].bbox[BTOP  ] = L->start->y < L->end->y ? L->end->y   : L->start->y;
		}

		for (i = 0; i < n; i++)
		{
			L = linedefs[cptr[i]];

			fbox2[i].bbox[BLEFT ] = L->start->x < L->end->x ? L->start->x : L->end->x;
			fbox2[i].bbox[BRIGHT] = L->start->x < L->end->x ? L->end->x   : L->start->x;
			fbox2[i].bbox[BBASE ] = L->start->y < L->end->y ? L->start->y : L->end->y;
			fbox2[i].bbox[BTOP  ] = L->start->y < L->end->y ? L->end->y   : L->start->y;
		}

		for (i = 0; i < m; i++)
		{
			b1 = fbox1[i].bbox; 

			for (j = 0; j < n; j++)
			{
				b2 = fbox2[j].bbox;

				// check if an lptr line overlaps a cptr line
				if (   b1[BLEFT ] > b2[BRIGHT] || b1[BRIGHT] < b2[BLEFT ]
					|| b1[BBASE ] > b2[BTOP  ] || b1[BTOP  ] < b2[BBASE ])
				{
					continue;
				}

				// midtex line crosses a textureless line
				lmem[lptr[i]] = true;
				num_xlfound++;
				break;
			}
		}

		free(fbox1);
		free(fbox2);
	}

	free(cptr);
	free(lptr);

	if (num_xlfound)
	{
		xlinemem = (xlinemem_t*)W_Calloc(num_xlfound * sizeof(xlinemem_t), LEVELID, 0);
		xlisrendered = (byte*)W_Malloc(num_xlfound, LEVELID, 0);
	}
}

// NOTE; v220-25 all xline vars are cleared in p_setup
//       This must be done prior to building any nodes.
static void PruneLinedefs(void)
{
	int		i;
	int		new_num;
	int		num_lzero = 0;

	// R3Dgwa xline --------------
	xlinemem_t	*lm;
	byte		*lmem = NULL;

	if (use_xl_lines)
	{
		lmem = (byte*)calloc(num_linedefs, sizeof(byte));
		
		findXlines(lmem);
	}
	// ---------------------------

	// scan all linedefs
	for (i=0, new_num=0; i < num_linedefs; i++)
	{
		ldef_t *L = linedefs[i];
		
		// handle duplicated vertices
		while (L->start->equiv)
		{
			L->start->ref_count--;
			L->start = L->start->equiv;
			L->start->ref_count++;
		}
		
		while (L->end->equiv)
		{
			L->end->ref_count--;
			L->end = L->end->equiv;
			L->end->ref_count++;
		}
		
		// handle duplicated sidedefs
		while (L->right && L->right->equiv)
		{
			L->right->ref_count--;
			L->right = L->right->equiv;
			L->right->ref_count++;
		}
		
		while (L->left && L->left->equiv)
		{
			L->left->ref_count--;
			L->left = L->left->equiv;
			L->left->ref_count++;
		}
		
		// remove zero length lines
		if (L->zero_len)
		{
			L->start->ref_count--;
			L->end->ref_count--;
			num_lzero++;
			
			UtilFree(L);
			continue;
		}

		if (use_xl_lines)
		{
			// R3Dgwa - remove xline sector lines
			if (lmem[i])
			{
				lm = &xlinemem[num_xlsame++];

				strncpy(lm->midtex, L->right->mid_tex, 8);
				lm->midtex[8] = 0;
				lm->startx = L->start->x;
				lm->starty = L->start->y;
				lm->endx = L->end->x;
				lm->endy = L->end->y;
				lm->y_off = (float)L->right->y_offset;
				lm->x_off[0] = (float)L->right->x_offset;
				lm->x_off[1] = (float)L->left->x_offset;
				lm->secn = L->right->sector->index;
				lm->flags = L->flags;

				L->start->ref_count--;
				L->end->ref_count--;
				UtilFree(L);
				continue;
			}
		}
		
		L->index = new_num;
		linedefs[new_num++] = L;
	}

	if (lmem) free(lmem);

	if ( (num_linedefs = new_num) == 0)
		FatalError("Couldn't find any Linedefs");
	
	if (num_lzero)
		pr_OFinsert("Pruned %d zero-length linedefs\n", num_lzero);
}

static void PruneVertices(void)
{
	int i;
	int new_num;
	int unused = 0;
	
	// scan all vertices
	for (i=0, new_num=0; i < num_vertices; i++)
	{
		vert_t *V = vertices[i];
		
		if (V->ref_count < 0)
			InternalError("Vertex %d ref_count is %d", i, V->ref_count);
		
		if (V->ref_count == 0)
		{
			if (V->equiv == NULL)
				unused++;
			
			UtilFree(V);
			continue;
		}
		
		V->index = new_num;
		vertices[new_num++] = V;
	}
	
	if (new_num < num_vertices)
	{
		int dup_num = num_vertices - new_num - unused;
		
		if (unused > 0)
			pr_OFinsert("Pruned %d unused vertices\n", unused);
		
		if (dup_num > 0)
			pr_OFinsert("Pruned %d duplicate vertices\n", dup_num);
		
		num_vertices = new_num;
	}
	
	if (new_num == 0)
		FatalError("Couldn't find any Vertices");
	
	num_normal_vert = num_vertices;
}

static void PruneSidedefs(void)
{
	int i;
	int new_num;
	int unused = 0;
	
	// scan all sidedefs
	for (i=0, new_num=0; i < num_sidedefs; i++)
	{
		sidedef_t *S = sidedefs[i];
		
		if (S->ref_count < 0)
			InternalError("Sidedef %d ref_count is %d", i, S->ref_count);
		
		if (S->ref_count == 0)
		{
			if (S->sector)
				S->sector->ref_count--;
			
			if (S->equiv == NULL)
				unused++;
			
			UtilFree(S);
			continue;
		}
		
		S->index = new_num;
		sidedefs[new_num++] = S;
	}
	
	if (new_num < num_sidedefs)
	{
		int dup_num = num_sidedefs - new_num - unused;
		
		if (unused > 0)
			pr_OFinsert("Pruned %d unused sidedefs\n", unused);
		
		if (dup_num > 0)
			pr_OFinsert("Pruned %d duplicate sidedefs\n", dup_num);
		
		num_sidedefs = new_num;
	}
	
	if (new_num == 0)
		FatalError("Couldn't find any Sidedefs");
}

static void PruneSectors(void)
{
	int i;
	int new_num;
	
	// scan all sectrs
	for (i=0, new_num=0; i < num_sectors; i++)
	{
		sect_t *S = sectrs[i];
		
		if (S->ref_count < 0)
			InternalError("Sector %d ref_count is %d", i, S->ref_count);
		
		if (S->ref_count == 0)
		{
			UtilFree(S);
			continue;
		}
		
		S->index = new_num;
		sectrs[new_num++] = S;
	}
	
	if (new_num < num_sectors)
	{
		pr_OFinsert("Pruned %d unused sectors\n", num_sectors - new_num);
		num_sectors = new_num;
	}
	
	if (new_num == 0)
		FatalError("Couldn't find any Sectors");
}

static int TransformSegDist(const nseg_t *seg)
{
	float sx = seg->side ? seg->linedef->end->x : seg->linedef->start->x;
	float sy = seg->side ? seg->linedef->end->y : seg->linedef->start->y;
	
	return (int) ceil(ComputeDist(seg->start->x - sx, seg->start->y - sy));
}

static int TransformAngle(angle_g angle)
{
	int result;
	
	result = (int)(angle * 65536.0 / 360.0);
	
	if (result < 0)
		result += 65536;
	
	return (result & 0xFFFF);
}

static int __cdecl SegCompare(const void *p1, const void *p2)
{
	const nseg_t *A = ((const nseg_t **) p1)[0];
	const nseg_t *B = ((const nseg_t **) p2)[0];
	
	if (A->index < 0)
		InternalError("Seg %p never reached a subsector !", A);
	
	if (B->index < 0)
		InternalError("Seg %p never reached a subsector !", B);
	
	return (A->index - B->index);
}


/* ----- vertex routines ------------------------------- */

static void VertexAddWallTip(vert_t *vert, float dx,
							 float dy, sect_t *left, sect_t *right)
{
	wall_tip_t *tip = NewWallTip();
	wall_tip_t *after;
	
	tip->angle = ComputeAngle(dx, dy);
	tip->left  = left;
	tip->right = right;
	
	// find the correct place (order is increasing angle)
	for (after=vert->tip_set; after && after->next; after=after->next)
	{ }
	
	while (after && tip->angle + ANG_EPSILON < after->angle) 
		after = after->prev;
	
	// link it in
	tip->next = after ? after->next : vert->tip_set;
	tip->prev = after;
	
	if (after)
	{
		if (after->next)
			after->next->prev = tip;
		
		after->next = tip;
	}
	else
	{
		if (vert->tip_set)
			vert->tip_set->prev = tip;
		
		vert->tip_set = tip;
	}
}

static void CalculateWallTips(void)
{
	int i;
	
	for (i=0; i < num_linedefs; i++)
	{
		ldef_t *line = linedefs[i];
		
		float x1 = line->start->x;
		float y1 = line->start->y;
		float x2 = line->end->x;
		float y2 = line->end->y;
		
		sect_t *left  = (line->left)  ? line->left->sector  : NULL;
		sect_t *right = (line->right) ? line->right->sector : NULL;
		
		VertexAddWallTip(line->start, x2-x1, y2-y1, left, right);
		VertexAddWallTip(line->end,   x1-x2, y1-y2, right, left);
	}
}

//
// NewVertexFromSplitSeg
//
vert_t *NewVertexFromSplitSeg(nseg_t *seg, float x, float y)
{
	vert_t *vert = NewVertex();
	
	vert->x = x;
	vert->y = y;
	
	vert->ref_count = seg->partner ? 4 : 2;
	
	vert->index = num_gl_vert | GL_MARKER;
	
	num_gl_vert++;
	
	// compute wall_tip info
	
	VertexAddWallTip(vert, -seg->pdx, -seg->pdy, seg->sector, 
		seg->partner ? seg->partner->sector : NULL);
	
	VertexAddWallTip(vert, seg->pdx, seg->pdy,
		seg->partner ? seg->partner->sector : NULL, seg->sector);
	
	// create a duplex vertex if needed
	
	if (doing_normal)
	{
		vert->normal_dup = NewVertex();
		
		vert->normal_dup->x = x;
		vert->normal_dup->y = y;
		vert->normal_dup->ref_count = vert->ref_count;
		
		vert->normal_dup->index = num_normal_vert;
		num_normal_vert++;
	}
	
	return vert;
}

//
// NewVertexDegenerate
//
vert_t *NewVertexDegenerate(vert_t *start, vert_t *end)
{
	float dx = end->x - start->x;
	float dy = end->y - start->y;
	
	float dlen = (float)ComputeDist(dx, dy);
	
	vert_t *vert = NewVertex();
	
	vert->ref_count = start->ref_count;
	
	if (doing_normal)
	{
		vert->index = num_normal_vert;
		num_normal_vert++;
	}
	else
	{
		vert->index = num_gl_vert | GL_MARKER;
		num_gl_vert++;
	}
	
	// compute new coordinates
	
	vert->x = start->x;
	vert->y = start->x;
	
	if (dlen == 0)
		InternalError("NewVertexDegenerate: bad delta !");
	
	dx /= dlen;
	dy /= dlen;
	
	while ((int)vert->x == (int)start->x && 
		(int)vert->y == (int)start->y)
	{
		vert->x += dx;
		vert->y += dy;
	}
	
	return vert;
}

//
// VertexCheckOpen
//
int VertexCheckOpen(vert_t *vert, float dx, float dy,
    sect_t ** left_sec, sect_t ** right_sec)
{
	wall_tip_t *tip;
	
	angle_g angle = ComputeAngle(dx, dy);
	
	*left_sec = *right_sec = NULL;
	
	// first check whether there's a wall_tip that lies in the exact
	// direction of the given direction (which is relative to the
	// vertex).
	
	for (tip=vert->tip_set; tip; tip=tip->next)
	{
		if (fabs(tip->angle - angle) < ANG_EPSILON)
		{
			// yes, found one
			
			*left_sec  = tip->left;
			*right_sec = tip->right;
			
			return false;
		}
	}
	
	// OK, now just find the first wall_tip whose angle is greater than
	// the angle we're interested in.  Therefore we'll be on the RIGHT
	// side of that wall_tip.
	
	for (tip=vert->tip_set; tip; tip=tip->next)
	{
		if (angle + ANG_EPSILON < tip->angle)
		{
			// found it
			*left_sec = *right_sec = tip->right;
			
			return (tip->right != NULL);
		}
		
		if (! tip->next)
		{
			// no more tips, thus we must be on the LEFT side of the tip
			// with the largest angle.
			
			*left_sec = *right_sec = tip->left;
			
			return (tip->left != NULL);
		}
	}
	
	InternalError("Vertex %d has no tips !", vert->index);
	return false;
}

void WarnLimit(char *entry, int num, int limit)
{
	PrintWarn("Number of %s (%d)"
		" has OVERFLOWED the limit (%d)\n",
		entry, num, limit);

	MarkLevelFailed();
}

/* ----- writing routines ------------------------------ */

void PutVertices()
{
	lump_t	*lump;
	int		count, i;
	int		limit = GL_extended ? 0x7FFFF : 0xFFFF;	// $R3Dgwa
	
	lump = CreateLevelLump("VERTEXES");
	
	for (i=0, count=0; i < num_vertices; i++)
	{
		raw_vertex_t raw;
		vert_t *vert = vertices[i];
		
		if (!(vert->index & GL_MARKER))
		{
			raw.x = (short)((int)vert->x);
			raw.y = (short)((int)vert->y);
			
			AppendLevelLump(lump, &raw, sizeof(raw));
			
			count++;
		}
	}
	
	if (count >= limit)	// $R3Dgwa
		WarnLimit("vertices", count, 0xFFFE);
	
	if (count != num_normal_vert)
		InternalError("PutVertices miscounted (%d != %d)",
		count, num_normal_vert);
}

void PutV2Vertices(char *v2_magic)
{
	int count, i;
	lump_t *lump;
	
	lump = CreateGLLump("GL_VERT");
	
	AppendLevelLump(lump, v2_magic, 4);
	
	for (i=0, count=0; i < num_vertices; i++)
	{
		raw_v2_vertex_t raw;
		vert_t *vert = vertices[i];
		
		if (vert->index & GL_MARKER)
		{
			raw.x = (int)(vert->x * 65536.0);
			raw.y = (int)(vert->y * 65536.0);
			
			AppendLevelLump(lump, &raw, sizeof(raw));
			
			count++;
		}
	}
	
	if (count >= 0xFFFF)
	{
		WarnLimit("GL vertices", count, 0xFFFE);
	}
	
	if (count != num_gl_vert)
		InternalError("PutV2Vertices miscounted (%d != %d)", count,
		num_gl_vert);
}

void PutSectors(void)
{
	int i;
	lump_t *lump = CreateLevelLump("SECTORS");
	
	for (i=0; i < num_sectors; i++)
	{
		raw_sector_t raw;
		sect_t *sector = sectrs[i];
		
		raw.floor_h = (short)(sector->floor_h);
		raw.ceil_h  = (short)(sector->ceil_h);
		
		memcpy(raw.floor_tex, sector->floor_tex, sizeof(raw.floor_tex));
		memcpy(raw.ceil_tex,  sector->ceil_tex,  sizeof(raw.ceil_tex));
		
		raw.light = (ushort)(sector->light);
		raw.special = (ushort)(sector->special);
		raw.tag   = (short)(sector->tag);
		
		AppendLevelLump(lump, &raw, sizeof(raw));
	}
	
	if (num_sectors >= 0x8000)
	{
		WarnLimit("sectors", num_sectors, 0x7FFF);
	}
}

void PutSidedefs(void)
{
	int i;
	lump_t *lump = CreateLevelLump("SIDEDEFS");
	
	for (i=0; i < num_sidedefs; i++)
	{
		raw_sidedef_t raw;
		sidedef_t *side = sidedefs[i];
		
		raw.sector = (side->sector == NULL) ? (ushort)(0xFFFF) :
		(ushort)(side->sector->index);
		
		raw.x_offset = (short)(side->x_offset);
		raw.y_offset = (short)(side->y_offset);
		
		memcpy(raw.upper_tex, side->upper_tex, sizeof(raw.upper_tex));
		memcpy(raw.lower_tex, side->lower_tex, sizeof(raw.lower_tex));
		memcpy(raw.mid_tex,   side->mid_tex,   sizeof(raw.mid_tex));
		
		AppendLevelLump(lump, &raw, sizeof(raw));
	}
	
	if (num_sidedefs >= 0xFFFF)
	{
		WarnLimit("sidedefs", num_sidedefs, 0xFFFE);
	}
}

void PutLinedefs(void)
{
	int i;
	lump_t *lump = CreateLevelLump("LINEDEFS");
	
	for (i=0; i < num_linedefs; i++)
	{
		raw_linedef_t raw;
		ldef_t *line = linedefs[i];
		
		raw.start = (ushort)(line->start->index);
		raw.end   = (ushort)(line->end->index);
		
		raw.flags = (ushort)(line->flags);
		raw.type  = (ushort)(line->type);
		raw.tag   = (short)(line->tag);
		
		raw.sidedef1 = line->right ? (ushort)(line->right->index) : 0xFFFF;
		raw.sidedef2 = line->left  ? (ushort)(line->left->index)  : 0xFFFF;
		
		if (line->is_special)	//R3Dgwa
		{
			raw.flags &= ~LINEFLAG_TWO_SIDED;
		}
		
		AppendLevelLump(lump, &raw, sizeof(raw));
	}
	
	if (num_linedefs >= 0xFFFF)
	{
		WarnLimit("linedefs", num_linedefs, 0xFFFE);
	}
}

void PutSegs(void)
{
	int i, count;
	lump_t *lump = CreateLevelLump("SEGS");
	
	// sort segs into ascending index
	qsort(nsegs, num_segs, sizeof(nseg_t *), SegCompare);
	
	for (i=0, count=0; i < num_segs; i++)
	{
		raw_seg_t raw;
		nseg_t *seg = nsegs[i];
		
		// ignore minisegs and degenerate segs
		if (!seg->linedef || seg->degenerate)
			continue;
		
		raw.start   = (ushort)(seg->start->index);
		raw.end     = (ushort)(seg->end->index);
		raw.angle   = (ushort)(TransformAngle(seg->p_angle));
		raw.linedef = (ushort)(seg->linedef->index);
		raw.flip    = (ushort)(seg->side);
		raw.dist    = (ushort)(TransformSegDist(seg));
		
		AppendLevelLump(lump, &raw, sizeof(raw));
		
		count++;
	}
	
	if (count >= 0xFFFF)
	{
		WarnLimit("segs", count, 0xFFFE);
	}
	
	if (count != num_complete_seg)
		InternalError("PutSegs miscounted (%d != %d)", count,
		num_complete_seg);
}

void PutGLSegs(void)
{
	lump_t	*lump = CreateGLLump("GL_SEGS");
	int		i, count;
	int		limit = GL_extended ? 0x7FFFF : 0xFFFE;	// $R3Dgwa
	
	// sort segs into ascending index
	qsort(nsegs, num_segs, sizeof(nseg_t *), SegCompare);
	
	for (i=0, count=0; i < num_segs; i++)
	{
		nseg_t *seg = nsegs[i];

		if (seg->degenerate) continue; // ignore degenerate segs
		
		if (GL_extended)
		{
			raw_gl_seg4_t raw;
			
			raw.start = (uint)(seg->start->index);	// Extended
			raw.end   = (uint)(seg->end->index);	// Extended
			raw.side  = (ushort)(seg->side);
			
			if (seg->linedef)
				raw.linedef = (ushort)(seg->linedef->index);
			else
				raw.linedef = (ushort)(0xFFFF);
			
			// Extended (partner not required)
			
			AppendLevelLump(lump, &raw, sizeof(raw));
			
			count++;
		}
		else
		{
			raw_gl_seg_t raw;
			unsigned int ind1 = seg->start->index;
			unsigned int ind2 = seg->end->index;

			if (ind1 & GL_MARKER) ind1 |= 0x8000;
			if (ind2 & GL_MARKER) ind2 |= 0x8000;

			raw.start = (ushort)(ind1);
			raw.end   = (ushort)(ind2);
			raw.side  = (ushort)(seg->side);
			
			if (seg->linedef)
				raw.linedef = (ushort)(seg->linedef->index);
			else
				raw.linedef = (ushort)(0xFFFF);
			
			if (seg->partner)
				raw.partner = (ushort)(seg->partner->index);
			else
				raw.partner = (ushort)(0xFFFF);

			
			AppendLevelLump(lump, &raw, sizeof(raw));
			
			count++;
		}
	}
	
	if (count >= limit)	// $R3Dgwa
		WarnLimit("GL segs", count, limit);
	
	if (count != num_complete_seg)
		InternalError("PutGLSegs miscounted (%d != %d)", count,
		num_complete_seg);
}

void PutSubsecs()
{
	int i;
	lump_t *lump;
	
	lump = CreateGLLump("GL_SSECT");

	for (i=0; i < num_subsecs; i++)
	{
		if (GL_extended)
		{
			raw_subsec4_t raw;
			subsec_t *sub = subsecs[i];
			
			raw.num   = (ushort)(sub->seg_count);
			raw.first = (uint)(sub->seg_list->index);	// Extended
			
			AppendLevelLump(lump, &raw, sizeof(raw));
		}
		else
		{
			raw_subsec_t raw;
			subsec_t *sub = subsecs[i];
			
			raw.num   = (ushort)(sub->seg_count);
			raw.first = (ushort)(sub->seg_list->index);	// Extended
			
			AppendLevelLump(lump, &raw, sizeof(raw));
		}
	}
	
	if (num_subsecs >= 0xFFFF)
	{
		WarnLimit("GL subsectors", num_subsecs, 0xFFFE);
	}
}

static int node_cur_index;

static void PutOneNodeGL4(nodb_t *node, lump_t *lump)
{
	raw_node4_t raw;
	
	if (node->r.node)
		PutOneNodeGL4(node->r.node, lump);
	
	if (node->l.node)
		PutOneNodeGL4(node->l.node, lump);
	
	node->index = node_cur_index++;
	
	raw.x  = (short)(node->x);
	raw.y  = (short)(node->y);
	raw.dx = (short)(node->dx / (node->too_long ? 2 : 1));
	raw.dy = (short)(node->dy / (node->too_long ? 2 : 1));
	
	raw.b1.minx = (short)(node->r.bounds.minx);
	raw.b1.miny = (short)(node->r.bounds.miny);
	raw.b1.maxx = (short)(node->r.bounds.maxx);
	raw.b1.maxy = (short)(node->r.bounds.maxy);
	
	raw.b2.minx = (short)(node->l.bounds.minx);
	raw.b2.miny = (short)(node->l.bounds.miny);
	raw.b2.maxx = (short)(node->l.bounds.maxx);
	raw.b2.maxy = (short)(node->l.bounds.maxy);
	
	// Extended ===================================
	if (node->r.node)
		raw.right = (uint)(node->r.node->index);
	else if (node->r.subsec)
		raw.right = (uint)(node->r.subsec->index | GL_MARKER);
	else
		InternalError("Bad right child in node %d", node->index);
	
	if (node->l.node)
		raw.left = (uint)(node->l.node->index);
	else if (node->l.subsec)
		raw.left = (uint)(node->l.subsec->index | GL_MARKER);
	else
		InternalError("Bad left child in node %d", node->index);
	// ============================================
	
	AppendLevelLump(lump, &raw, sizeof(raw));
}

static void PutOneNode(nodb_t *node, lump_t *lump)
{
	raw_node_t raw;
	
	if (node->r.node)
		PutOneNode(node->r.node, lump);
	
	if (node->l.node)
		PutOneNode(node->l.node, lump);
	
	node->index = node_cur_index++;
	
	raw.x  = (short)(node->x);
	raw.y  = (short)(node->y);
	raw.dx = (short)(node->dx / (node->too_long ? 2 : 1));
	raw.dy = (short)(node->dy / (node->too_long ? 2 : 1));
	
	raw.b1.minx = (short)(node->r.bounds.minx);
	raw.b1.miny = (short)(node->r.bounds.miny);
	raw.b1.maxx = (short)(node->r.bounds.maxx);
	raw.b1.maxy = (short)(node->r.bounds.maxy);
	
	raw.b2.minx = (short)(node->l.bounds.minx);
	raw.b2.miny = (short)(node->l.bounds.miny);
	raw.b2.maxx = (short)(node->l.bounds.maxx);
	raw.b2.maxy = (short)(node->l.bounds.maxy);
	
	if (node->r.node)
		raw.right = (ushort)(node->r.node->index);
	else if (node->r.subsec)
		raw.right = (ushort)(node->r.subsec->index | 0x8000);
	else
		InternalError("Bad right child in node %d", node->index);
	
	if (node->l.node)
		raw.left = (ushort)(node->l.node->index);
	else if (node->l.subsec)
		raw.left = (ushort)(node->l.subsec->index | 0x8000);
	else
		InternalError("Bad left child in node %d", node->index);
	
	AppendLevelLump(lump, &raw, sizeof(raw));
}

void PutNodes(char *name, int do_gl, nodb_t *root)
{
	lump_t	*lump;
	int		limit = GL_extended ? 0x7FFFF : 0x8000;	// $R3Dgwa
	
	if (do_gl)
		lump = CreateGLLump(name);
	else
		lump = CreateLevelLump(name);
	
	node_cur_index = 0;
	
	if (root)
	{
		if (GL_extended)
			PutOneNodeGL4(root, lump);
		else
			PutOneNode(root, lump);
	}
	
	if (node_cur_index >= limit)	// $R3Dgwa
	{
		if (do_gl)
			WarnLimit("GL nodes", node_cur_index, limit);
		else
			WarnLimit("nodes", node_cur_index, limit);
	}
	
	if (node_cur_index != num_nodes)
		InternalError("PutNodes miscounted (%d != %d)",
		node_cur_index, num_nodes);
}


/* ----- whole-level routines --------------------------- */

//
// LoadLevel
//
void LoadLevel(void)
{
	const char *level_name = GetLevelName();

	strncpy(global_name, level_name, 8);
	global_name[8] = 0;
	
	normal_exists = CheckForNormalNodes();
	
	doing_normal = !cur_info->gwa_mode
				&& (cur_info->force_normal
					|| (!cur_info->no_normal && !normal_exists));
/*  $glbsp_nomsg
	if (doing_normal)
		pr_OFinsert("Building Normal and GL nodes on %s\n", level_name);
	else
		pr_OFinsert("Building GL nodes for %s\n", level_name);*/
	
	GetVertices();
	GetSectors();
	GetSidedefs();
	GetLinedefs();
	
	if (   !cur_info->choose_fresh
		&& !doing_normal
		&& normal_exists
		&& num_sectors > 5
		&& num_linedefs > 100)
	{
		pr_OFinsert("Using original nodes as requested\n");
		GetStaleNodes();
	}

// v220-35 lines >= 0xF000 overload the pruning code
	int change_noprune = false;

	if (!cur_info->no_prune && num_linedefs >= 61440)
	{
void Con_InitProgress(const char *title, int full);

		change_noprune = true;
		Con_InitProgress("WAIT building glbsp game nodes - This could take a while", 100); // $progress XA Start
	}
// end noprune check
	
	if (doing_normal && !cur_info->no_prune && !change_noprune)
	{
		DetectDuplicateVertices();
		
		if (  !cur_info->no_prune
			&& cur_info->pack_sides)
		{
			DetectDuplicateSidedefs();
		}
		
		PruneLinedefs();
		PruneVertices();
		PruneSidedefs();
		
		if (   !cur_info->no_prune
			&& !cur_info->keep_sect)
		{
			PruneSectors();
		}
	}
	
	CalculateWallTips();
}

//
// FreeLevel
//
void FreeLevel(void)
{
	FreeVertices();
	FreeSidedefs();
	FreeLinedefs();
	FreeSectors();
	FreeSegs();
	FreeSubsecs();
	FreeNodes();
	FreeStaleNodes();
	FreeWallTips();
}

//
// SaveLevel
//
void SaveLevel(nodb_t *root_node, BOOL extended)
{
	char *type[] = { "gNd2", "gNd4" };

	GL_extended = extended;

/*  $glbsp_nomsg
	pr_OFinsert("\nSaving using format %s\n", type[GL_extended]);
*/	
	PutV2Vertices(type[GL_extended]);
	PutGLSegs();
	PutSubsecs();
	PutNodes("GL_NODES", true, root_node);
		
	if (doing_normal)
	{
		RoundOffBspTree(root_node);
		NormaliseBspTree(root_node);
		
		PutVertices();
		PutSectors();
		PutSidedefs();
		PutLinedefs();
		
		if (!GL_extended)
		{
			PutSegs();
			PutNodes("NODES", false, root_node);
		}
	}
}

