//------------------------------------------------------------------------
// LEVEL : Level structures & read/write functions.
//------------------------------------------------------------------------
//
//  GL-Friendly Node Builder (C) 2000-2002 Andrew Apted
//
//  Based on `BSP 2.3' by Colin Reed, Lee Killough and others.
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  Revision 2.2.0 build 6 (v220-6) 19/Apr/2010 gmj
//   Modified for internal use. See the bsp_ReadMe.txt in the bsp dir
//
//------------------------------------------------------------------------

#ifndef __GLBSP_LEVEL_H__
#define __GLBSP_LEVEL_H__

#include "rd_structs.h"
#include "rd_wad.h"


struct nodb_s;
struct sect_s;
struct superblock_s;


// a wall_tip is where a wall meets a vertex
typedef struct wall_tip_s
{
  // link in list.  List is kept in ANTI-clockwise order.
  struct wall_tip_s *next;
  struct wall_tip_s *prev;
  
  // angle that line makes at vertex (degrees).
  angle_g angle;

  // sectors on each side of wall.  Left is the side of increasing
  // angles, right is the side of decreasing angles.  Either can be
  // NULL for one sided walls.
  struct sect_s *left;
  struct sect_s *right;
}
wall_tip_t;


typedef struct vert_s
{
  // coordinates
  float x, y;

  // vertex index.  Always valid after loading and pruning of unused
  // vertices has occurred.  For GL vertices, bit 15 will be set.
  int index;

  int canbemerged;	// $R3Dgwa

  // reference count.  When building normal node info, unused vertices
  // will be pruned.
  int ref_count;

  // usually NULL, unless this vertex occupies the same location as a
  // previous vertex.  Only used during the pruning phase.
  struct vert_s *equiv;

  // set of wall_tips
  wall_tip_t *tip_set;

  // contains a duplicate vertex, needed when both normal and V2 GL
  // nodes are being built at the same time (this is the vertex used
  // for the normal segs).  Normally NULL.  Note: the wall tips on
  // this vertex are not created.
  struct vert_s *normal_dup;
}
vert_t;


typedef struct sect_s
{
  // sector index.  Always valid after loading & pruning.
  int index;

  // allow segs from other sectors to coexist in a subsector
  int coalesce;

  // reference count.  When building normal nodes, unused sectors will
  // be pruned.
  int ref_count;

  // heights
  int floor_h, ceil_h;

  // textures
  char floor_tex[8];
  char ceil_tex[8];

  // attributes
  int light;
  int special;
  int tag;

  // used when building REJECT table.  Each set of sectors that are
  // isolated from other sectors will have a different group number.
  // Thus: on every 2-sided linedef, the sectors on both sides will be
  // in the same group.  The rej_next, rej_prev fields are a link in a
  // RING, containing all sectors of the same group.
  int rej_group;

  struct sect_s *rej_next;
  struct sect_s *rej_prev;

} sect_t;


typedef struct sidedef_s
{
  // adjacent sector.  Can be NULL (invalid sidedef)
  sect_t *sector;

  // offset values
  int x_offset, y_offset;

  // texture names
  char upper_tex[8];
  char lower_tex[8];
  char mid_tex[8];
  
  // sidedef index.  Always valid after loading & pruning.
  int index;

  // reference count.  When building normal nodes, unused sidedefs will
  // be pruned.
  int ref_count;

  // usually NULL, unless this sidedef is exactly the same as a
  // previous one.  Only used during the pruning phase.
  struct sidedef_s *equiv;

  // this is true if the sidedef is on a special line.  We don't merge
  // these sidedefs together, as they might scroll, or change texture
  // when a switch is pressed.
  int on_special;

} sidedef_t;


typedef struct ldef_s
{
  // link for list
  struct ldef_s *next;

  vert_t *start;    // from this vertex...
  vert_t *end;      // ... to this vertex

  sidedef_t *right;   // right sidedef
  sidedef_t *left;    // left sidede, or NULL if none

  // line is marked two-sided
  char two_sided;

  // prefer not to split
  char is_precious;

  //$R3Dgwa Identify single sided line needing a subsector
  char is_special;

  // zero length (line should be totally ignored)
  char zero_len;

  int flags;
  int type;
  int tag;

  // Hexen support
  int specials[5];

  // linedef index.  Always valid after loading & pruning of zero
  // length lines has occurred.
  int index;

} ldef_t;


typedef struct nseg_s
{
  // link for list
  struct nseg_s *next;

  vert_t *start;   // from this vertex...
  vert_t *end;     // ... to this vertex

  // linedef that this seg goes along, or NULL if miniseg
  ldef_t *linedef;

  // adjacent sector, or NULL if invalid sidedef or miniseg
  sect_t *sector;

  // 0 for right, 1 for left
  int side;

  // seg on other side, or NULL if one-sided.  This relationship is
  // always one-to-one -- if one of the segs is split, the partner seg
  // must also be split.
  struct nseg_s *partner;

  // seg index.  Only valid once the seg has been added to a
  // subsector.  A negative value means it is invalid -- there
  // shouldn't be any of these once the BSP tree has been built.
  int index;

  // when 1, this seg has become zero length (integer rounding of the
  // start and end vertices produces the same location).  It should be
  // ignored when writing the SEGS or V1 GL_SEGS lumps.  [Note: there
  // won't be any of these when writing the V2 GL_SEGS lump].
  int degenerate;
 
  // the superblock that contains this seg, or NULL if the seg is no
  // longer in any superblock (e.g. now in a subsector).
  struct superblock_s *block;

  // precomputed data for fast calculations
  float psx, psy;
  float pex, pey;
  float pdx, pdy;

  float p_length;
  float p_angle;
  float p_para;
  float p_perp;

  // linedef that this seg initially comes from.  For "real" segs,
  // this is just the same as the `linedef' field above.  For
  // "minisegs", this is the linedef of the partition line.
  ldef_t *source_line;

} nseg_t;


typedef struct subsec_s
{
  // list of segs
  nseg_t *seg_list;

  // count of segs
  int seg_count;

  // subsector index.  Always valid, set when the subsector is
  // initially created.
  int index;

  // approximate middle point
  float mid_x;
  float mid_y;
}
subsec_t;


typedef struct bbox_s
{
  int minx, miny;
  int maxx, maxy;
}
bbox_t;


typedef struct child_s
{
  // child node or subsector (one must be NULL)
  struct nodb_s *node;
  subsec_t *subsec;

  // child bounding box
  bbox_t bounds;
}
child_t;


typedef struct nodb_s
{
  int x, y;     // starting point
  int dx, dy;   // offset to ending point

  // right & left children
  child_t r;
  child_t l;

  // node index.  Only valid once the NODES or GL_NODES lump has been
  // created.
  int index;

  // the node is too long, and the (dx,dy) values should be halved
  // when writing into the NODES lump.
  int too_long;
}
nodb_t;


typedef struct superblock_s
{
  // parent of this block, or NULL for a top-level block
  struct superblock_s *parent;

  // coordinates on map for this block, from lower-left corner to
  // upper-right corner.  Pseudo-inclusive, i.e (x,y) is inside block
  // if and only if x1 <= x < x2 and y1 <= y < y2.
  int x1, y1;
  int x2, y2;

  // sub-blocks.  NULL when empty.  [0] has the lower coordinates, and
  // [1] has the higher coordinates.  Division of a square always
  // occurs horizontally (e.g. 512x512 -> 256x512 -> 256x256).
  struct superblock_s *subs[2];

  // number of real segs and minisegs contained by this block
  // (including all sub-blocks below it).
  int real_num;
  int mini_num;

  // list of segs completely contained by this block.
  nseg_t *segs;
}
superblock_t;

#define SUPER_IS_LEAF(s)  \
    ((s)->x2-(s)->x1 <= 256 && (s)->y2-(s)->y1 <= 256)


/* ----- Level data arrays ----------------------- */

extern int num_vertices;
extern int num_linedefs;
extern int num_sidedefs;
extern int num_sectors;
extern int num_segs;
extern int num_subsecs;
extern int num_nodes;
extern int num_stale_nodes;

extern int num_normal_vert;
extern int num_gl_vert;
extern int num_complete_seg;


/* ----- function prototypes ----------------------- */

// allocation routines
vert_t		*NewVertex(void);
ldef_t		*NewLinedef(void);
sidedef_t	*NewSidedef(void);
sect_t		*NewSector(void);
nseg_t		*NewSeg(void);
subsec_t	*NewSubsec(void);
nodb_t		*NewNode(void);
nodb_t		*NewStaleNode(void);

// lookup routines
vert_t		*LookupVertex(int index);
ldef_t		*LookupLinedef(int index);
sidedef_t	*LookupSidedef(int index);
sect_t		*LookupSector(int index);
nseg_t		*LookupSeg(int index);
subsec_t	*LookupSubsec(int index);
nodb_t		*LookupNode(int index);
nodb_t		*LookupStaleNode(int index);

// check whether the current level already has normal nodes
int CheckForNormalNodes(void);

// load all level data for the current level
void LoadLevel(void);

// free all level data
void FreeLevel(void);

// save the newly computed NODE info etc..
void SaveLevel(nodb_t *root_node, BOOL extended);

// return a new vertex (with correct wall_tip info) for the split that
// happens along the given seg at the given location.
//
vert_t *NewVertexFromSplitSeg(nseg_t *seg, float x, float y);

// return a new end vertex to compensate for a seg that would end up
// being zero-length (after integer rounding).  Doesn't compute the
// wall_tip info (thus this routine should only be used _after_ node
// building).
//
vert_t *NewVertexDegenerate(vert_t *start, vert_t *end);

// check whether a line with the given delta coordinates and beginning
// at this vertex is open.  Returns 1 if open, or 0 if closed.  The
// sectors that lie on the left & right side of the given line are
// also determined (NULL if the area is void space).
//
int VertexCheckOpen(vert_t *vert, float dx, float dy,
    sect_t ** left_sec, sect_t ** right_sec);


#endif /* __GLBSP_LEVEL_H__ */
