//------------------------------------------------------------------------
// STRUCT : Doom structures, raw on-disk layout
//------------------------------------------------------------------------
//
//  GL-Friendly Node Builder (C) 2000-2002 Andrew Apted
//
//  Based on `BSP 2.3' by Colin Reed, Lee Killough and others.
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  Revision 2.2.0 build 6 (v220-6) 19/Apr/2010 gmj
//   Modified for internal use. See the bsp_ReadMe.txt in the bsp dir
//
//------------------------------------------------------------------------

#ifndef __GLBSP_STRUCTS_H__
#define __GLBSP_STRUCTS_H__

#include "rd_system.h"


/* ----- The wad structures ---------------------- */

// wad header

typedef struct raw_wad_header_s
{
  char type[4];

  uint num_entries;
  uint dir_start;
}
raw_wad_header_t;


// directory entry

typedef struct raw_wad_entry_s
{
  uint start;
  uint length;

  char name[8];
}
raw_wad_entry_t;


// blockmap

typedef struct raw_blockmap_header_s
{
  short x_origin, y_origin;
  short x_blocks, y_blocks;
}
raw_blockmap_header_t;


/* ----- The level structures ---------------------- */

typedef struct raw_vertex_s
{
  short x, y;
}
raw_vertex_t;

typedef struct raw_v2_vertex_s
{
  int x, y;
}
raw_v2_vertex_t;


typedef struct raw_linedef_s
{
  ushort start;		// from this vertex...
  ushort end;		// ... to this vertex
  ushort flags;	
  ushort type;		// linedef type (special actually)
  short tag;
  ushort sidedef1;	// right sidedef
  ushort sidedef2;	// left sidedef (only if this line adjoins 2 sectors)
}
raw_linedef_t;

typedef struct raw_hexen_linedef_s
{
  ushort start;        // from this vertex...
  ushort end;          // ... to this vertex
  ushort flags;        // linedef flags (impassible, etc)
  byte  type;         // linedef type
  byte  specials[5];  // hexen specials
  ushort sidedef1;     // right sidedef
  ushort sidedef2;     // left sidedef
}
raw_hexen_linedef_t;

#define LINEFLAG_IMPASSABLE		1	// $R3Dgwa ML_BLOCKING
#define LINEFLAG_TWO_SIDED		4	// $R3Dgwa ML_TWOSIDED


#define HEXTYPE_POLY_START		1
#define HEXTYPE_POLY_EXPLICIT	5


typedef struct raw_sidedef_s
{
  short x_offset;  // X offset for texture
  short y_offset;  // Y offset for texture

  char upper_tex[8];  // texture name for the part above
  char lower_tex[8];  // texture name for the part below
  char mid_tex[8];    // texture name for the regular part

  ushort sector;    // adjacent sector
}
raw_sidedef_t;


typedef struct raw_sector_s
{
  short floor_h;   // floor height
  short ceil_h;    // ceiling height

  char floor_tex[8];  // floor texture
  char ceil_tex[8];   // ceiling texture

  ushort light;     // light level (0-255)
  ushort special;   // special behaviour (0 = normal, 9 = secret, ...)
  short tag;       // sector activated by a linedef with same tag
}
raw_sector_t;


typedef struct raw_thing_s
{
  short x, y;      // position of thing
  short angle;     // angle thing faces (degrees)
  ushort type;      // type of thing
  ushort options;   // when appears, deaf, etc..
}
raw_thing_t;


/* ----- The BSP tree structures ----------------------- */

#pragma pack(2)

typedef struct raw_bbox_s
{
  short maxy, miny;
  short minx, maxx;
}
raw_bbox_t;

typedef struct raw_seg_s
{
  ushort start;     // from this vertex...
  ushort end;       // ... to this vertex
  ushort angle;     // angle (0 = east, 16384 = north, ...)
  ushort linedef;   // linedef that this seg goes along
  ushort flip;      // true if not the same direction as linedef
  ushort dist;      // distance from starting point
}
raw_seg_t;


typedef struct raw_gl_seg_s
{
  ushort start;      // from this vertex...
  ushort end;        // ... to this vertex
  ushort linedef;    // linedef that this seg goes along, or -1
  ushort side;       // 0 if on right of linedef, 1 if on left
  ushort partner;    // partner seg number, or -1
}
raw_gl_seg_t;


typedef struct raw_subsec_s
{
  ushort num;     // number of Segs in this Sub-Sector
  ushort first;   // first Seg
}
raw_subsec_t;


// Used for looking up nodes in unextended maps
typedef struct raw_node_s
{
  short x, y;         // starting point
  short dx, dy;       // offset to ending point
  raw_bbox_t b1, b2;     // bounding rectangles
  ushort right, left;  // children: Node or SSector (if high bit is set)
}
raw_node_t;

//========================================================

// GL4 extended types
typedef struct raw_node4_s
{
  short x, y;
  short dx, dy;
  raw_bbox_t b1, b2;
  uint right, left;	// Extended
}
raw_node4_t;

typedef struct raw_subsec4_s
{
  ushort num;
  uint first;	// Extended
}
raw_subsec4_t;

typedef struct raw_gl_seg4_s
{
  uint start;	// Extended
  uint end;		// Extended
  ushort linedef;
  ushort side;
}
raw_gl_seg4_t;

#pragma pack()

#endif /* __GLBSP_STRUCTS_H__ */
