//------------------------------------------------------------------------
// WAD : WAD read/write functions.
//------------------------------------------------------------------------
//
//  GL-Friendly Node Builder (C) 2000-2002 Andrew Apted
//
//  Based on `BSP 2.3' by Colin Reed, Lee Killough and others.
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  Revision 2.2.0 build 6 (v220-6) 19/Apr/2010 gmj
//   Modified for internal use. See the bsp_ReadMe.txt in the bsp dir
//
//------------------------------------------------------------------------

#include "rd_system.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <math.h>
#include <limits.h>
#include <errno.h>

#include "rd_blockmap.h"
#include "rd_level.h"
#include "rd_node.h"
#include "rd_seg.h"
#include "rd_structs.h"
#include "rd_util.h"
#include "rd_wad.h"

#include "sys_file.h"

void pr_OFinsert(char *message, ...);

#define APPEND_BLKSIZE  256
#define LEVNAME_BUNCH   20
#define IWAD			0
#define PWAD			1

DFILE *in_file = NULL;
DFILE *out_file = NULL;

// current wad info
static wad_t wad;


/* ---------------------------------------------------------------- */


#define NUM_LEVEL_LUMPS  8
#define NUM_GL_LUMPS     4

static const char *level_lumps[NUM_LEVEL_LUMPS]=
{
	"LINEDEFS",
	"SIDEDEFS",
	"VERTEXES",
	"SEGS", 
	"SSECTORS",
	"NODES",
	"SECTORS",
	"REJECT",
};

static const char *gl_lumps[NUM_GL_LUMPS]=
{
	"GL_VERT",
	"GL_SEGS",
	"GL_SSECT",
	"GL_NODES"
};


//
// CheckMagic
//
static int CheckMagic(const char type[4])
{
	return (!strncmp(type, "PWAD", 4) ||
			!strncmp(type, "IWAD", 4));
}


//
// CheckLevelName
//
static int CheckLevelName(const char *name)
{
	int n;
	
	if (strlen(name) > 5)
		return false;
	
	for (n=0; n < wad.num_level_names; n++)
	{
		if (strcmp(wad.level_names[n], name) == 0)
			return true;
	}
	
	return false;
}


//
// CheckLevelLumpName
//
// Tests if the entry name is one of the level lumps.
//
static int CheckLevelLumpName(const char *name)
{
	int i;
	
	for (i=0; i < NUM_LEVEL_LUMPS; i++)
	{
		if (strcmp(name, level_lumps[i]) == 0)
			return true;
	}
	
	return false;
}

//
// Level name helper
//
static void AddLevelName(const char *name)
{
	if ((wad.num_level_names % LEVNAME_BUNCH) == 0)
	{
		wad.level_names = (const char **) UtilRealloc((void *)wad.level_names,
			(wad.num_level_names + LEVNAME_BUNCH) * sizeof(const char *));
	}
	
	wad.level_names[wad.num_level_names] = UtilStrDup(name);
	wad.num_level_names++;
}


//
// NewLump
//
// Create new lump.  `name' must be allocated storage.
//
static lump_t *NewLump(char *name)
{
	lump_t *cur;
	
	cur = (lump_t*)UtilCalloc(sizeof(lump_t));
	
	cur->name = name;
	cur->start = cur->new_start = -1;
	cur->flags = 0;
	cur->length = 0;
	cur->space = 0;
	cur->data = NULL;
	cur->level_list = NULL;
	cur->level_gl_list = NULL;
	cur->level_buddy = NULL;
	
	return cur;
}


//
// FreeLump
//
static void FreeLump(lump_t *lump)
{
	// free level lumps, if any
	if (lump->flags & LUMP_IS_LEVEL)
	{
		while (lump->level_list)
		{
			lump_t *head = lump->level_list;
			lump->level_list = head->next;
			
			// the ol' recursion trick... :)
			FreeLump(head);
		}
	}
	
	// free GL level lumps, if any
	if (lump->flags & LUMP_IS_GL_LEVEL)
	{
		while (lump->level_gl_list)
		{
			lump_t *head = lump->level_gl_list;
			lump->level_gl_list = head->next;
			
			// the ol' recursion trick... :)
			FreeLump(head);
		}
	}
	
	// check `data' here, since it gets freed in WriteLumpData()
	if (lump->data)
		UtilFree(lump->data);
	
	UtilFree(lump->name);
	UtilFree(lump);
}


//
// ReadHeader
//
// Returns true if successful, or false if there was a problem (in
// which case the error message as been setup).
//
static BOOL ReadHeader(const char *filename)
{
	int len;
	raw_wad_header_t header;
	char strbuf[1024];
	
	len = F_Read(&header, sizeof(header), in_file);
	
	if (len != sizeof(header))
	{
		sprintf(strbuf, "Trouble reading wad header for %s : %s", 
			filename, strerror(errno));
		
		GlbspFree(cur_comms->message);
		cur_comms->message = GlbspStrDup(strbuf);
		
		return false;
	}
	
	if (! CheckMagic(header.type))
	{
		sprintf(strbuf, "%s does not appear to be a wad file : bad magic", 
			filename);
		
		GlbspFree(cur_comms->message);
		cur_comms->message = GlbspStrDup(strbuf);
		
		return false;
	}
	
	wad.kind = (header.type[0] == 'I') ? IWAD : PWAD;
	
	wad.num_entries = (uint)(header.num_entries);
	wad.dir_start   = (uint)(header.dir_start);
	
	// initialise stuff
	wad.dir_head = NULL;
	wad.dir_tail = NULL;
	wad.current_level = NULL;
	wad.level_names = NULL;
	wad.num_level_names = 0;
	
	return true;
}


//
// ReadDirEntry
//
static void ReadDirEntry(void)
{
	int len;
	raw_wad_entry_t entry;
	lump_t *lump;
	
	len = F_Read(&entry, sizeof(entry), in_file);
	
	if (len != sizeof(entry))
		FatalError("Trouble reading wad directory");
		
	lump = NewLump(UtilStrNDup(entry.name, 8));
	
	lump->start = (uint)(entry.start);
	lump->length = (uint)(entry.length);
		
	// link it in
	lump->next = NULL;
	lump->prev = wad.dir_tail;
	
	if (wad.dir_tail)
		wad.dir_tail->next = lump;
	else
		wad.dir_head = lump;
	
	wad.dir_tail = lump;
}

//
// DetermineLevelNames
//
static void DetermineLevelNames(void)
{
	lump_t *L, *N;
	int i;
	
	for (L=wad.dir_head; L; L=L->next)
	{
		// Check if the next four lumps after the
		// current lump match the level-lump names.
		for (i=0, N=L->next; (i < 4) && N; i++, N=N->next)
			if (strcmp(N->name, level_lumps[i]) != 0)
				break;

		if (i != 4) continue;
			
		// check for invalid name and duplicate levels
		if (strlen(L->name) > 5)
			PrintWarn("Bad level name `%s' in wad (too long)\n", L->name);
		else if (CheckLevelName(L->name))
			PrintWarn("Level name `%s' found twice in wad\n", L->name);
		else
			AddLevelName(L->name);
	}
}

//
// ProcessDirEntry
//
static void ProcessDirEntry(lump_t *lump)
{
	// mark the lump as `ignorable' when in GWA mode.
	// v220-6 needed in all cases
/*	if (cur_info->gwa_mode)
		lump->flags |= LUMP_IGNORE_ME;*/
	
	// --- LEVEL MARKERS ---
	
	if (CheckLevelName(lump->name))
	{
		// OK, start a new level
		lump->flags |= LUMP_COPY_ME;
		lump->flags |= LUMP_IS_LEVEL;
		wad.current_level = lump;
		
		// link it in
		lump->next = NULL;
		lump->prev = wad.dir_tail;
		
		if (wad.dir_tail)
			wad.dir_tail->next = lump;
		else
			wad.dir_head = lump;
		
		wad.dir_tail = lump;
		
		return;
	}
	
	// --- LEVEL LUMPS ---
	
	if (wad.current_level)
	{
		if (CheckLevelLumpName(lump->name))
		{
			// check for duplicates
			if (FindLevelLump(lump->name))
			{
				PrintWarn("Duplicate entry `%s' ignored in %s\n",
					lump->name, wad.current_level->name);
				
				FreeLump(lump);
				wad.num_entries--;
				
				return;
			}
			
			// mark it to be loaded
			lump->flags |= LUMP_READ_ME;
			
			// link it in
			lump->next = wad.current_level->level_list;
			lump->prev = NULL;
			
			if (lump->next)
				lump->next->prev = lump;
			
			wad.current_level->level_list = lump;
			return;
		}
		
		// OK, non-level lump.  End the previous level.
		
		wad.current_level = NULL;
	}
	
	// --- ORDINARY LUMPS ---
	
	if (CheckLevelLumpName(lump->name))
		PrintWarn("Level lump `%s' found outside any level\n", lump->name);
	
	// maybe load data
	lump->flags |= LUMP_COPY_ME;
	
	// link it in
	lump->next = NULL;
	lump->prev = wad.dir_tail;
	
	if (wad.dir_tail)
		wad.dir_tail->next = lump;
	else
		wad.dir_head = lump;
	
	wad.dir_tail = lump;
}

//
// ReadDirectory
//
static void ReadDirectory(void)
{
	int i;
	int total_entries = wad.num_entries;
	lump_t *prev_list;
	
	F_Seek(in_file, wad.dir_start, SEEK_SET);
	
	for (i=0; i < total_entries; i++)
	{
		ReadDirEntry();
	}
	
	DetermineLevelNames();
	
	// finally, unlink all lumps and process each one in turn
	
	prev_list = wad.dir_head;
	wad.dir_head = wad.dir_tail = NULL;
	
	while (prev_list)
	{
		lump_t *cur = prev_list;
		prev_list = cur->next;
		
		ProcessDirEntry(cur);
	}
}


//
// ReadLumpData
//
static void ReadLumpData(lump_t *lump)
{
	int len;
	
	if (lump->length == 0)
		return;
	
	lump->data = UtilCalloc(lump->length);
	
	F_Seek(in_file, lump->start, SEEK_SET);
	
	len = F_Read(lump->data, lump->length, in_file);
	
	if (len != lump->length)
	{
		if (wad.current_level)
			PrintWarn("Trouble reading lump `%s' in %s\n",
			lump->name, wad.current_level->name);
		else
			PrintWarn("Trouble reading lump `%s'\n", lump->name);
	}
	
	lump->flags &= ~LUMP_READ_ME;
}


//
// ReadAllLumps
//
// Returns number of entries read.
//
static int ReadAllLumps(void)
{
	lump_t *cur, *L;
	int count = 0;
	
	for (cur=wad.dir_head; cur; cur=cur->next)
	{
		count++;
		
		if (cur->flags & LUMP_READ_ME)
			ReadLumpData(cur);
		
		if (cur->flags & LUMP_IS_LEVEL)
		{
			for (L=cur->level_list; L; L=L->next)
			{
				count++;
				
				if (L->flags & LUMP_READ_ME)
					ReadLumpData(L);
			}
		}
	}
	
	return count;
}


//
// CountLumpTypes
//
// Returns number of entries with matching flags.  Used to predict how
// many ReadLumpData() or WriteLumpData() calls will be made (for the
// progress bars).
//
static int CountLumpTypes(int flag_mask, int flag_match)
{
	lump_t *cur, *L;
	int count = 0;
	
	for (cur=wad.dir_head; cur; cur=cur->next)
	{
		if ((cur->flags & flag_mask) == flag_match)
			count++;
		
		if (cur->flags & LUMP_IS_LEVEL)
		{
			for (L=cur->level_list; L; L=L->next)
				if ((L->flags & flag_mask) == flag_match)
					count++;
		}
		
		if (cur->flags & LUMP_IS_GL_LEVEL)
		{
			for (L=cur->level_gl_list; L; L=L->next)
				if ((L->flags & flag_mask) == flag_match)
					count++;
		}
	}
	
	return count;
}


/* ---------------------------------------------------------------- */


//
// WriteHeader
//
static void WriteHeader(void)
{
	int len;
	raw_wad_header_t header;
	
	switch (wad.kind)
	{
	case IWAD:
		strncpy(header.type, "IWAD", 4);
		break;
		
	case PWAD:
		strncpy(header.type, "PWAD", 4);
		break;
	}
	
	header.num_entries = (uint)(wad.num_entries);
	header.dir_start   = (uint)(wad.dir_start);
	
	len = F_FileWrite(&header, sizeof(header), out_file, -2);
	
	if (len != 1)
		PrintWarn("Trouble writing wad header [-2]\n");
}


//
// CreateGLMarker
//
lump_t *CreateGLMarker(lump_t *level)
{
	lump_t *cur;
	char name_buf[16];
	
	sprintf(name_buf, "GL_%s", level->name);
	
	cur = NewLump(UtilStrDup(name_buf));
	cur->flags = LUMP_IS_GL_LEVEL;
	
	// link it in
	cur->next = level->next;
	cur->prev = level;
	
	if (cur->next)
		cur->next->prev = cur;
	
	level->next = cur;
	level->level_buddy = cur;
	
	return cur;
}


//
// SortLumps
//
// Algorithm is pretty simple: for each of the names, if such a lump
// exists in the list, move it to the head of the list.  By going
// backwards through the names, we ensure the correct order.
//
static void SortLumps(lump_t ** list, const char **names, int count)
{
	int i;
	lump_t *cur;
	
	for (i=count-1; i >= 0; i--)
	{
		for (cur=(*list); cur; cur=cur->next)
		{
			if (strcmp(cur->name, names[i]) != 0)
				continue;
			
			// unlink it
			if (cur->next)
				cur->next->prev = cur->prev;
			
			if (cur->prev)
				cur->prev->next = cur->next;
			else
				(*list) = cur->next;
			
			// add to head
			cur->next = (*list);
			cur->prev = NULL;
			
			if (cur->next)
				cur->next->prev = cur;
			
			(*list) = cur;
			
			// continue with next name (important !)
			break;
		}
	}
}

//
// RecomputeDirectory
//
// Calculates all the lump offsets for the directory.
//
static void RecomputeDirectory(void)
{
	lump_t *cur, *L;
	
	wad.num_entries = 0;
	wad.dir_start = sizeof(raw_wad_header_t);
	
	// run through all the lumps, computing the `new_start' fields, the
	// number of lumps in the directory, the directory starting pos, and
	// also sorting the lumps in the levels.
	
	for (cur=wad.dir_head; cur; cur=cur->next)
	{
		if (cur->flags & LUMP_IGNORE_ME)
			continue;
		
		cur->new_start = wad.dir_start;
		
		wad.dir_start += cur->length;
		wad.num_entries++;
		
		if (cur->flags & LUMP_IS_LEVEL)
		{
			SortLumps(&cur->level_list, level_lumps, NUM_LEVEL_LUMPS);
			
			for (L=cur->level_list; L; L=L->next)
			{
				if (L->flags & LUMP_IGNORE_ME)
					continue;
				
				L->new_start = wad.dir_start;
				
				wad.dir_start += L->length;
				wad.num_entries++;
			}
		}
		
		if (cur->flags & LUMP_IS_GL_LEVEL)
		{
			SortLumps(&cur->level_gl_list, gl_lumps, NUM_GL_LUMPS);
			
			for (L=cur->level_gl_list; L; L=L->next)
			{
				if (L->flags & LUMP_IGNORE_ME)
					continue;
				
				L->new_start = wad.dir_start;
				
				wad.dir_start += L->length;
				wad.num_entries++;
			}
		}
	}
}


//
// WriteLumpData
//
static void WriteLumpData(lump_t *lump)
{
	int len;
		
	if (F_Tell(out_file) != lump->new_start)
		PrintWarn("Consistency failure writing %s (%08X, %08X\n", 
		lump->name, F_Tell(out_file), lump->new_start);
	
	if (lump->length == 0)
		return;
	
	if (lump->flags & LUMP_COPY_ME)
	{
		lump->data = UtilCalloc(lump->length);
		
		F_Seek(in_file, lump->start, SEEK_SET);
		
		len = F_Read(lump->data, lump->length, in_file);
		
		if (len != lump->length)
			PrintWarn("Trouble reading lump %s to copy\n", lump->name);
	}
	
	len = F_FileWrite(lump->data, lump->length, out_file, -3);
	
	if (len != 1)
		PrintWarn("Trouble writing lump %s [-3]\n", lump->name);
	
	UtilFree(lump->data);
	
	lump->data = NULL;
}


//
// WriteAllLumps
//
// Returns number of entries written.
//
static int WriteAllLumps(void)
{
	lump_t *cur, *L;
	int count = 0;
	
	for (cur=wad.dir_head; cur; cur=cur->next)
	{
		if (cur->flags & LUMP_IGNORE_ME)
			continue;
		
		WriteLumpData(cur);
		count++;
		
		if (cur->flags & LUMP_IS_LEVEL)
		{
			for (L=cur->level_list; L; L=L->next)
			{
				if (L->flags & LUMP_IGNORE_ME)
					continue;
				
				WriteLumpData(L);
				count++;
			}
		}
		
		if (cur->flags & LUMP_IS_GL_LEVEL)
		{
			for (L=cur->level_gl_list; L; L=L->next)
			{
				if (L->flags & LUMP_IGNORE_ME)
					continue;
				
				WriteLumpData(L);
				count++;
			}
		}
	}
	
	return count;
}


//
// WriteDirEntry
//
static void WriteDirEntry(lump_t *lump)
{
	int len;
	raw_wad_entry_t entry;
	
	strncpy(entry.name, lump->name, 8);
	
	entry.start = (uint)(lump->new_start);
	entry.length = (uint)(lump->length);
	
	len = F_FileWrite(&entry, sizeof(entry), out_file, -4);
	
	if (len != 1)
		PrintWarn("Trouble writing wad directory [-4]\n");
}


//
// WriteDirectory
//
// Returns number of entries written.
//
static int WriteDirectory(void)
{
	lump_t *cur, *L;
	int count = 0;
	
	if (F_Tell(out_file) != wad.dir_start)
		PrintWarn("Consistency failure writing lump directory "
		"(%08X,%08X)\n", F_Tell(out_file), wad.dir_start);
	
	for (cur=wad.dir_head; cur; cur=cur->next)
	{
		if (cur->flags & LUMP_IGNORE_ME)
			continue;
		
		WriteDirEntry(cur);
		count++;
		
		if (cur->flags & LUMP_IS_LEVEL)
		{
			for (L=cur->level_list; L; L=L->next)
			{
				if (cur->flags & LUMP_IGNORE_ME)
					continue;
				
				WriteDirEntry(L);
				count++;
			}
		}
		
		if (cur->flags & LUMP_IS_GL_LEVEL)
		{
			for (L=cur->level_gl_list; L; L=L->next)
			{
				if (cur->flags & LUMP_IGNORE_ME)
					continue;
				
				WriteDirEntry(L);
				count++;
			}
		}
	}
	
	return count;
}


/* ---------------------------------------------------------------- */

//
// CreateLevelLump
//
lump_t *CreateLevelLump(const char *name)
{
	lump_t *cur;
	
	// already exists ?
	for (cur=wad.current_level->level_list; cur; cur=cur->next)
	{
		if (strcmp(name, cur->name) == 0)
			break;
	}
	
	if (cur)
	{
		if (cur->data)
			UtilFree(cur->data);
		
		cur->data = NULL;
		cur->length = 0;
		cur->space  = 0;
		
		return cur;
	}
	
	// nope, allocate new one
	cur = NewLump(UtilStrDup(name));
	
	// link it in
	cur->next = wad.current_level->level_list;
	cur->prev = NULL;
	
	if (cur->next)
		cur->next->prev = cur;
	
	wad.current_level->level_list = cur;
	
	return cur;
}


//
// CreateGLLump
//
lump_t *CreateGLLump(const char *name)
{
	lump_t *cur;
	lump_t *gl_level;
	
	// create GL level marker if necessary
	if (! wad.current_level->level_buddy)
		CreateGLMarker(wad.current_level);
	
	gl_level = wad.current_level->level_buddy;
	
	// check if already exists
	for (cur=gl_level->level_gl_list; cur; cur=cur->next)
	{
		if (strcmp(name, cur->name) == 0)
			break;
	}
	
	if (cur)
	{
		if (cur->data)
			UtilFree(cur->data);
		
		cur->data = NULL;
		cur->length = 0;
		cur->space  = 0;
		
		return cur;
	}
	
	// nope, allocate new one
	cur = NewLump(UtilStrDup(name));
	
	// link it in
	cur->next = gl_level->level_gl_list;
	cur->prev = NULL;
	
	if (cur->next)
		cur->next->prev = cur;
	
	gl_level->level_gl_list = cur;
	
	return cur;
}


//
// AppendLevelLump
//
void AppendLevelLump(lump_t *lump, void *data, int length)
{
	if (length == 0)
		return;
	
	if (lump->length == 0)
	{
		lump->space = MAX(length, APPEND_BLKSIZE);
		lump->data = UtilCalloc(lump->space);
	}
	else if (lump->space < length)
	{
		lump->space = MAX(length, APPEND_BLKSIZE);
		lump->data = UtilRealloc(lump->data, lump->length + lump->space);
	}
	
	memcpy(((char *)lump->data) + lump->length, data, length);
	
	lump->length += length;
	lump->space  -= length;
}


//
// CountLevels
//
int CountLevels(void)
{
	lump_t *cur;
	int result = 0;
	
	for (cur=wad.dir_head; cur; cur=cur->next)
	{
		if (cur->flags & LUMP_IS_LEVEL)
			result++;
	}
	
	return result;
}

//
// FindNextLevel
//
int FindNextLevel(void)
{
	lump_t *cur;
	
	if (wad.current_level)
		cur = wad.current_level->next;
	else
		cur = wad.dir_head;
	
	while (cur && !(cur->flags & LUMP_IS_LEVEL) )
		cur=cur->next;
	
	wad.current_level = cur;
	
	return (cur != NULL);
}

//
// GetLevelName
//
const char *GetLevelName(void)
{
	if (!wad.current_level)
		InternalError("GetLevelName: no current level");
	
	return wad.current_level->name;
}

//
// FindLevelLump
//
lump_t *FindLevelLump(const char *name)
{
	lump_t *cur = wad.current_level->level_list;
	
	while (cur && (strcmp(cur->name, name) != 0))
		cur=cur->next;
	
	return cur;
}

//
// CheckLevelLumpZero
//
int CheckLevelLumpZero(lump_t *lump)
{
	int i;
	
	if (lump->length == 0)
		return true;
	
	for (i=0; i < lump->length; i++)
	{
		if (((byte *)lump->data)[i])
			return false;
	}
	
	return true;
}

//
// ReadWadFile
//
glbsp_ret_e ReadWadFile(const char *filename)
{
	int check;
	
	// open input wad file & read header
	in_file = F_OpenFile(filename, "b");
	
	if (!in_file)
	{
		pr_OFinsert("Cannot open WAD file %s : %s", filename, 
			strerror(errno));
		
		return GLBSP_E_ReadError;
	}
	
	if (!ReadHeader(filename))
	{
		F_Close(in_file);
		return GLBSP_E_ReadError;
	}
	
/*  $glbsp_nomsg
	pr_OFinsert("Reading %d lumps from intemp.wad\n", wad.num_entries);
*/	
	// read directory
	ReadDirectory();
	
	// now read lumps
	check = ReadAllLumps();
	
	if (check != wad.num_entries)
		PrintWarn("Read directory count consistency failure (%d,%d)\n",
		check, wad.num_entries);
	
	wad.current_level = NULL;
	
	return GLBSP_E_OK;
}


//
// WriteWadFile
//
glbsp_ret_e WriteWadFile(const char *filename)
{
	int check1, check2;
	
	RecomputeDirectory();

	// create output wad file & write the header
	if (!(out_file = F_OpenFileForWrite(filename, "b")))
	{
		pr_OFinsert("Cannot open output %s file: %s", filename, strerror(errno));
		return GLBSP_E_WriteError;
	}
	
	WriteHeader();
	
	// now write all the lumps to the output wad
	check1 = WriteAllLumps();
	
	// finally, write out the directory
	check2 = WriteDirectory();
	
	if (check1 != wad.num_entries || check2 != wad.num_entries)
		PrintWarn("Write directory count consistency failure (%d,%d,%d\n",
		check1, check2, wad.num_entries);
/*  $glbsp_nomsg
	else
		pr_OFinsert("Stored %i lumps\n", wad.num_entries);
*/
	return GLBSP_E_OK;
}


//
// CloseWads
//
void CloseWads(void)
{
	int i;
	
	if (in_file ) F_Close(in_file);
	if (out_file) F_Close(out_file);
	
	/* free directory entries */
	while (wad.dir_head)
	{
		lump_t *head = wad.dir_head;
		wad.dir_head = head->next;
		
		FreeLump(head);
	}
	
	/* free the level names */
	if (wad.level_names)
	{
		for (i=0; i < wad.num_level_names; i++)
			UtilFree((char *) wad.level_names[i]);
		
		UtilFree((void *)wad.level_names);
		wad.level_names = NULL;
	}
}


//
// MarkLevelFailed
//
void MarkLevelFailed(void)
{
	wad.current_level->flags |= LUMP_FAILED_LEVEL;
}


//
// ReportFailedLevels
// Simplified as only handles one level $R3Dgwa
//
int ReportFailedLevels(void)
{
	lump_t *cur;
	
	for (cur = wad.dir_head; cur; cur = cur->next)
	{
		if (   cur->flags & LUMP_IS_LEVEL				
			&& cur->flags & LUMP_FAILED_LEVEL)
		{
			pr_OFinsert("\nATTENTION: LEVEL FAILED TO BUILD !!\n");
			return 1;
		}
	}
	
	return 0;
}

